extern basicdoc;
/* DOCUMENT basic package
 *
 *   Basic UNIX like functions and some other very
 *   general purpose ones.
 *
 *   2004-10/NJW
 *   2009-11-13/NJW, updated with lettername

      SiGn            fullpath        plural          
      _cp             gen_log_entry   priarr          
      _lettername     gen_log_init    prmat           
      _str2x          get_time        prstrarr        
                                      prtmat
      _wrdig          gettimetag      pwd             
      add13           ghost           radec2sxg       
      adstring        grep            random_i        
      back            groupn          rdbinfil        
      bit_extract     gt              reldif          
      bitpatt2arr     hex2deci        rem_elem        
      cat             indexnearest    remove13        
      check_time      indices         rm              
      cmp             insert          selection       
      cp              inter_step      set_time        
      cronjob         is_dir          setbatch        
      deci2hex        lbico           setup_dir       
      delta           lettername      sexa2deci       
      diag_elems      lfacul          sixty           
      docuc           ls              splitfname      
      docud           lt              stch2num        
      docuf           mkdirp          t2s             
      docui           more            tdosunix        
      docum           mv              ten             
      docup           n_idls          uniq            
      docus           n_yoricks       uniq2d          
      dwhere          nan2zero        vrest           
      facul           near            vsave           
      filesize        near2pi         whereany        
      filter_common   near360         wherenan        
      filter_done     nextbatch       wherenone       
      filter_done1    pathname        wrdig           
      fndeci          pfiles_path     yjour           
      fswap           wstyles         is_linux
*/

timestamp, TSTARTYORICK;

/* Function insert */

func insert( new_elem, old_arr, position )
/* DOCUMENT new_arr = insert( new_elem, old_arr, position )

     Inserts 'new_elem' into the existing array 'old_arr' so that
     the returned result 'new_arr' has the new element at index 'position',
     i.e. new_arr(position) == new_elem.

   2008-08-19/NJW
*/
{
   if( position <= 0 ) position += numberof(old_arr)+1;

   if( position == 1 ) {
      res = new_elem;
      grow, res, old_arr;
   } else if( position == numberof(old_arr)+1 ) {
      res = old_arr;
      grow, res, new_elem;
   } else {
      res = old_arr(1:position-1);
      grow,res,new_elem,old_arr(position:0);
   }

   return res;

}

/* Function uniq */

func uniq( x )
/* DOCUMENT idx = uniq( x )

   Returns the indices of the unique elements of the
   one-dimensional array 'x'.
   'x' will not be sorted here so
       >  uniq([1,2,1]) returns [1,2,3]
       >  uniq([1,1,2]) returns [1,3]

   2007-06-19/NJW
   2009-10-27/NJW faster because 'grow' now operates on blocks

*/
{
   n = numberof(x);
   if( n == 0 ) return [];
   rsize = 500;
   res = array(long,rsize);
   res(1) = 1;
   k = 1;
   for(i=2;i<=n;i++) {
      if( x(i-1) != x(i) ) {
         res(++k) = i;
         if( k == rsize ) {
            grow, res, res;
            rsize *= 2;
         }
      }
   }
   return res(1:k);
}

/* Function fswap */

func fswap( &a, &b )
/* DOCUMENT fswap, (>)a, (>)b

   returns the value of 'a' in 'b' and vice versa

   2007-02-06/NJW
*/
{ x = a; a = b; b = x; }

/* Function fullpath */

func fullpath( filename )
/* DOCUMENT res = fullpath( filename )

   Returns the entire path to the named file(s).

*/
{
   if( is_scalar(filename) ) return _fullpath( filename );
   n = numberof( filename );
   res = array(string,n);
   for( i = 1; i <= n; i++ ) res(i) = _fullpath( filename(i) );
   return res;
}

func _fullpath( filename )
/* DOCUMENT res = _fullpath( filename )

     Returns the entire path to the named file.

     2007-01-29/NJW
*/
{
   local dir, fname;

   if( strlen(get_env("OSTYPE")) == 0 ) { // Windows OS

      if( strpart(filename,2:3) == ":/" ) return filename;

      if( filename == "." ) return strpart( get_cwd(), 1:-1);

      if( strpart(filename,1:2) == "./" ) {
         filename = strpart(filename, 3:0 );
      }
      return get_cwd()+filename;

   } else { // Linux or Solaris OS

      if( strpart(filename,1:1) == "/" ) return filename;

      if( filename == "." ) return strpart( get_cwd(), 1:-1);

      if( filename == ".." ) {
         cwd = get_cwd();
         cwd = cwd == "/" ? cwd : rem_slash(cwd);
         splitfname, cwd, dir, fname;
         return dir;
      }

      if( strpart(filename,1:2) == "./" ) {
         filename = strpart(filename, 3:0 );
         return get_cwd()+filename;
      }

      if( strpart(filename,1:3) == "../" ) {
         filename = strpart(filename, 4:0 );
         splitfname, rem_slash(get_cwd()), dir, fname;
         return dir+"/"+filename;
      }
 
      // a relative filename has been given
      return get_cwd()+filename;

   }
}

/* Function pathname */

func pathname( filename )
/* DOCUMENT res = pathname( filename )

    Returns the full path to the file, excluding the file part.
    The argument 'filename' may be a scalar or an array of strings.

*/
{
   fp = fullpath( filename );
   p = strpos( fp, "/", rev=1 );
   for( i = 1; i <= numberof(fp); i++ ) fp(i) = strpart( fp(i), 1:p(i)-1 );
   return fp;
}

/* Function cat */

func cat( file1, file2, outfile )
/* DOCUMENT cat, file1, file2, outfile
   Works as UNIX csh command 'cat'; puts the text files: 'file1' and 'file2'
   into 'outfile'.

   If 'outfile' is void then the output will be directed to the screen.
   If 'file2' is void then it will be skipped.

   2004-09-09/NJW

*/

{
   f1 = open( file1 );
   f2 = structof(file2) == string ? open( file2 ) : [] ;
   fout = structof(outfile) == string ? open( outfile, "w" ) : [];

   while( line = rdline(f1) ) write, fout, format="%s\n", line;
   if( ! is_void(f2) ) while( line = rdline(f2) ) write, fout, format="%s\n", line;
   close, f1;
   if( ! is_void(f2) ) close, f2;
   if( ! is_void(fout) ) close, fout;

}

/* Function cp */

func cp( file_from, file_to )
/* DOCUMENT cp, file_from, file_to

   Works almost like the Unix/Linux 'cp' command.
   'file_from' cannot be a directory.

   Examples:  cp,"./*","mysubdir";  will copy all files
                  from this directory to directory: mysubdir

   2008-12-11/NJW
*/
{
   local basename, dirname;

   // 'file_from' may not be a directory

   if( is_dir(file_from) ) {
      write,format="%s is a directory\n", file_from;
      return;
   }

   // if 'file_to' is a directory then check the
   // possibility for wild cards i.e. multi-copying

   if( is_dir(file_to) ) {
      splitfname, file_from, dirname, basename;
      filelist = file_search( basename, dirname );
      n = numberof( filelist );
      if( n == 0 ) {
         write,format="%s not found\n", file_from;
         return;
      }
      for( i = 1; i <= n; i++ ) {
         splitfname, filelist(i), dirname, basename;
         filename = file_to+"/"+basename;
         _cp, filelist(i), filename;
      }
      return;
   }

   // from here it is a single file copy

   if( !file_test( file_from ) ) {
      write,format="%s not found\n", file_from;
      return;
   }
   // 'file_to' must be a file name, get the directory
   // where it resides
   splitfname, file_to, dirname, basename;
   if( !file_test(dirname) ) {
      write,format="Directory: %s not found\n", dirname;
      return;
   }
   _cp, file_from, file_to;
   return;
}

/* Function _cp */

func _cp( file1, outfile )
/* DOCUMENT    _cp, file1, outfile
   Works as UNIX csh command 'cp'; copies file1 into outfile
   but only single file copying and 'outfile' cannot be a
   directory.

   2004-09-09/NJW
   2007-08-11/NJW, updated to operate on binary files
   2008-12-11/NJW, updated to see all files as binary

*/
{
   f1 = open( file1, "rb" );
   fout = open( outfile, "wb" );
   address = 0;
   v = array(char,1);

   while( _read( f1, address, v) == 1 ) _write, fout, address++, v;

   close, f1;
   close, fout;

   remove, outfile+"L"; // An unwanted log file is produced

}

/* Function cmp */

func cmp( file1, file2 )
/* DOCUMENT  res = cmp( file1, file2 )

   Works as UNIX csh command 'cmp'; compares file1 with file2
   and reports on the first differing byte.

   The byte number returned starts counting with 1 (one);
   Returning 0 (zero) indicates identical files.
   Returning -1 tells that file1 is bigger than file2
   Returning -2 tells that file2 is bigger than file1

   2012-01-02/NJW

*/
{
   f1 = open( file1, "rb" );
   f2 = open( file2, "rb" );

   //  Check for differing sizes

   a = sizeof(f1) - sizeof(f2);
   if( a ) {
      close, f1; close, f2;
      r = a > 0 ? -1 : -2;
      return r;
   }

   //  Well, they have the same size, check byte by byte

   address = 0;
   v = array(char,1);
   w = array(char,1);

   while( _read( f1, address, v) == 1 ) {
      _read, f2, address++, w;
      if( v(1) != w(1) ) {
         r = address;
         close, f1; close, f2;
         return r;
      }
   }

   close, f1; close, f2;
   return 0; // A perfect match

}


/* Function grep */

func grep( str, dir, sel=, filelist=, nocap= )
/* DOCUMENT  list = grep( str, dir, sel=, filelist=, nocap= )

     Returns a list of findings of string 'str' in the files
     in directory 'dir' (defaults to current directory).

     A file name selection can be given with keyword 'sel'.

     Keyword 'filelist' can contain a string array with a
     file list that then will override the directory selection.

     Keyword 'nocap' will cause it to disregard upper/lower case.

*/

{

   if( numberof(filelist) ) {
      files = filelist;
   } else {
      if( !is_void(sel) ) {
         files = file_search( sel, dir );
      } else {
         files = file_search( "*", dir );
      }
   }

   if( nocap ) str = strlowcase(str);

   n = numberof(files);
   if( n == 0 ) return [];

   first = 1;
   for( i = 1; i <= n; i++ ) {
      f = open( files(i),"r", 1 );
      if( !f ) continue;
      s = swrite(format="// file = %s", files(i));
      fw = 0;
      while( line = rdline(f) ) {
         if( nocap ) cline = strlowcase(line);
         else cline = line;
         if( strmatch( cline, str ) ) {
            if( fw == 0 ) {
               fw = 1;
               if( first ) {
                  result = s;
                  first = 0;
               } else {
                  grow, result, s;
               }
            }
            grow, result, line;
         }
      }
      close, f;
   }
   return result;

}

/* Function nan2zero */

func nan2zero( arr )
/* DOCUMENT newarr = nan2zero( arr )

   Uses the standard function ieee_test (in 'ieee.i') to
   test for "not normal" i.e. NaN,-Inf,or Inf in array 'arr'.
   The returned array has the same type and size as the input array.

   2006-01-23/NJW

*/
{

   require,"ieee.i";

   newarr = arr;
   wnan = ieee_test(arr);
   w = where( wnan != 0 ); // Rather a test for 'normal'
   if( numberof(w) > 0 ) newarr(w) = 0;

   return newarr;

}

/* Function wherenan */

func wherenan( arr, not= )
/* DOCUMENT idx = wherenan( arr, not= )

   similar to "where( arr is NaN )"
   or - when keyword 'not' is set -
   "where( arr is not NaN )"

   2010-02-16/NJW
*/
{

   require,"ieee.i";
   wnan = ieee_test(arr);
   if( not ) return where( wnan == 0 );
   return where( wnan != 0 );

}

/* Function t2s */

func t2s( filename )
/* DOCUMENT t2s,filename

   "Tab to space" function.
   Convert the tab character ('\t' or ASCII '9') to
   the appropriate number of spaces.

*/
{

   require, "idlx.i";

   text = read_slist( filename );
   nt = numberof( text );
   for( i = 1; i<= nt; i++ ) {
      line = text(i);
      len = strlen( line );
      newline = "";
      pos = 0;
      for( j = 1; j <= len; j++ ) {
         c = strpart( line, j:j );
         if( c == "\t" ) {
            nsp = 8 - (pos % 8);
            for( k = 1; k <= nsp; k++ ) newline += " ";
            pos += nsp;
         } else {
            newline += c;
            pos++;
         }
      }
      text(i) = newline;
   }
   write_slist, filename, text;

}

/* Function inter_step */

func inter_step( y, x, xp )
/* DOCUMENT yp = inter_step( y, x, xp )

   Returns the value y_i where x_i is the largest x_i smaller than xp

   2005-01-30/NJW

   Note [2018-03-09/NJW]: this function is not particularly
                          useful so it should probably be removed.

*/

{

   w = where( x <= xp );
   if( numberof(w) > 0 ) {
      return y(w(0));
   } else {
      return y(1);
   }

}

func back( filename, del= )
/* DOCUMENT back, filename, del=

   copies filename to filename.nnn where nnn is one larger than
   all other instances of filename.xxx

   When keyword 'del' is set then a move in stead of a copy is done.

   2007-08-10/NJW
*/
{

   local dirname, basename;

   if( !file_test( filename ) ) {
      write,format="BACK: %s was not found\n", filename;
      return;
   }

   splitfname, filename, dirname, basename;

   list = file_search(basename+"*", dirname);
   n = numberof(list);
   len = strlen( basename );
   list = list(sort(list));
   //+ write,format="%i files considered:\n",n;
   //+ write,list;
   next = 1;
   if( n > 1 ) {
      for(i=2;i<=n;i++) {
         if( strlen(list(i)) < len+4 ) continue; // not part of 'family'
         if( strpart(list(i),-3:-3) != "." ) continue; // not part of 'family'
         s = strpart(list(i),-2:0);
         if( !is_digit(s) ) continue; // not part of 'family'
         //+ write,format="Evaluating %s\n", list(i);
         k = 0;
         sread,format="%d",s,k;
         if( k+1 > next ) next = k+1;
      }
   }
   s = swrite(format=".%03i",next);
   comm = del ? mv : cp;
   comm, filename, filename+s;
   write,format="Copied into: %s\n", filename+s;
   if( file_test(filename+s+"L") ) remove, filename+s+"L";
}

/* Function bit_extract */

func bit_extract( m, pwr2=, bitn= )
/* DOCUMENT bit = bit_extract( m, pwr2=, bitn= )

   Returns a (long) array of same dimension as 'm'
   indicating whether the bit specified by one of the keywords
   is set.

   pwr2 is one of: 1, 2, 4, 8, 16, ...
   bitn is one of: 1,2,3,4, ...

   pwr2 = 2^(bitn-1)

   2008-01-02/NJW
*/
{
   m = long(m);

   if( pwr2 ) {
      bitn = 0;
      while( pwr2 ) {
         pwr2 = pwr2>>1;
         bitn++;
      }
   } else if( !bitn ) {
      error,"BIT_EXTRACT missing keyword";
   }
   if( bitn <= 0 ) error,"BIT_EXTRACT illegal value of bitn";
   //+ write,format="bitn = %i\n", bitn;

   return (m>>(bitn-1))%2;

}

/* Function bitpatt2arr */

func bitpatt2arr( m )
/* DOCUMENT arr = bitpatt2arr( m )

   Returns bitpattern in an array
   with least significant bit first.

   Argument: can be of type long, int, or char.

   Example:   bitpatt2arr(22) -> [0,1,1,0,1]

   SEE ALSO: bitpatt2arr8

   2010-01-08/NJW
*/
{
   if( m == 0 ) return [0];
   r = [];
   while( m > 0 ) {
      grow, r, m%2;
      m = m>>1;
   }
   return r;
}

/* Function bitpatt2arr8 */

func bitpatt2arr8( m )
/* DOCUMENT arr = bitpatt2arr8( m )

   Returns bitpattern in an array with at least 8 elements
   and least significant bit first.

   Argument: can be of type long, int, or char.

   Example:   bitpatt2arr8(22) -> [0,1,1,0,1,0,0,0]

   SEE ALSO: bitpatt2arr

   2017-04-24/NJW based on bitpatt2arr
*/
{

 /*************
  *if( m == 0 ) return array(0,8);
  *r = [];
  *while( m > 0 ) {
  *   grow, r, m%2;
  *   m = m>>1;
  *}
 **************/
   r = bitpatt2arr(m);
   while( numberof(r) < 8 ) r = grow( r, 0 );
   return r;
}

/* Function filter_done */

func filter_done( done_list, list )
/* DOCUMENT diff_list = filter_done( done_list, list )

   Return an array with values from "list" except
   those that are found in "done_list"

   The functionality is parallel to the C-program "filter_done.c"

   SEE ALSO: wherenone which has the same functionality.
             filter_done1 that is supposed to be faster

    2007-09-25/NJW
    2008-01-09/NJW updated to operate on all types of arrays
*/

{

   if( is_void(done_list) ) return list;

   if( is_void(list) ) return list;

   if( typeof(done_list) != typeof(list) ) {
      error,"FILTER_DONE mismatching argument types";
   }

   // Step through "list"

   ndone_list = numberof(done_list);
   nlist = numberof(list);

   outlist = [];
   for( i = 1; i <= nlist; i++ ) {

      w = where( done_list  ==  list(i) );
      nw = numberof(w);
      if( nw  ==  0 ) {
         grow, outlist, list(i);
      }

   }

   return outlist;

}

/* Function filter_common */

func filter_common( list1, list2 )
/* DOCUMENT common_list = filter_common( list1, list2 )

   Return an array with values that are found both in "list2"
   and in "list1"

   The functionality is parallel to the shell script "filter_common"

    2008-01-09/NJW
    2009-10-27/NJW improved (faster) algorithm
*/
{
   if( is_void(list1) ) return [];
   if( is_void(list2) ) return [];
   if( typeof(list1) != typeof(list2) ) {
      error,"FILTER_DONE mismatching argument types";
   }

   // sort arrays and keep only unique elements

   list1w = list1(sort(list1));
   list2w = list2(sort(list2));
   list1w = list1w(uniq(list1w));
   list2w = list2w(uniq(list2w));

   // only keep the overlap parts

   w = where( list1w >= list2w(1) );
   if( numberof(w) ) list1w = list1w(w);
   else return [];
   w = where( list2w >= list1w(1) );
   if( numberof(w) ) list2w = list2w(w);
   else return [];
   w = where( list1w <= list2w(0) );
   if( numberof(w) ) list1w = list1w(w);
   else return [];
   w = where( list2w <= list1w(0) );
   if( numberof(w) ) list2w = list2w(w);
   else return [];

   // prepare stepping through both arrays

   nlist1w = numberof(list1w);
   nlist2w = numberof(list2w);

   i = j = 1;
   osize = 500;
   outlist = array(structof(list1w),osize);

   k = 0;
   while( i <= nlist1w && j <= nlist2w ) {

      // stepping while indices 'i' and 'j' point to
      // array elements to be compared

      if( list1w(i) == list2w(j) ) {
         outlist(++k) = list1w(i);
         if( k == osize ) {
            grow, outlist, outlist;
            osize *= 2;
         }
         i++; j++;
      } else if( list1w(i) > list2w(j) ) {
         j++;
      } else {
         i++;
      }

   }

   return k == 0 ? [] : outlist(1:k);

}

/* Function near */

func near( x, a, epsilon, rel= )
/* DOCUMENT answer = near( x, a, epsilon, rel= )

   Result of x > a - epsilon & x < a + epsilon

   When 'rel' (for relative) is set:

   Result of x > a - epsilon*max(abs(x),abs(a))  &  x < a + epsilon*max(abs(x),abs(a))

   2008-01-30/NJW, 2015-04-22/NJW added keyword 'rel'
*/
{
   if( rel ) {
      maxxa = max( abs(a), abs(x) );
      return x > a - epsilon*maxxa & x < a + epsilon*maxxa;
   } else {
      return x > a - epsilon & x < a + epsilon;
   }
}

/* Function near360 */

func near360( x, a, epsilon )
/* DOCUMENT answer = near360( x, a, epsilon )

   Result of abs(zero360(x)-zero360(a)) < epsilon |
      360 - abs(zero360(x)-zero360(a)) < epsilon

   2011-01-18/NJW
*/
{
   zz = abs(zero360(x)-zero360(a));
   return zz < epsilon | 360.0 - zz < epsilon;
}

/* Function splitfname */

func splitfname( filenamepath, &dirname, &basename )
/* DOCUMENT  splitfname, filenamepath, &dirname, &basename

    separates a file path into directory name and file name
    using the '/' character as separator. If it does no exist
    in the filenamepath then '.' is given as the directory name.

   2006-02-09/NJW

*/
{
   // Remove leading and trailing blanks
   filenamepath = strtrim(filenamepath);
   // Make sure that separator '/' is not multiple anywhere
   filenamepath = strcompress(filenamepath,st="/");
   // Remove trailing '/' if present
   if( strpart(filenamepath, 0:0) == "/" ) filenamepath = strpart(filenamepath,1:-1);

   p = strpos( filenamepath, "/", 0, rev=1); // reversed search direction
   if( p < 1 ) {
      dirname = ".";
      basename = filenamepath;
   } else {
      if( p == 1 ) {
         dirname = "/";
         basename = strpart(filenamepath,2:);
      } else {
         dirname = strpart(filenamepath, 1:p-1);
         basename = strpart(filenamepath,p+1:);
      }
   }
}

/* Function set_time */

func set_time( a )
/* DOCUMENT set_time

   Defines an external variable 'LIMIT_TIME_0' with the current number of
   seconds since 1970-01-01. This is used by function 'check_time' that will
   compare the current time with this time origin.

   SEE ALSO: check_time

   2008-11-01/NJW

*/
{
   extern LIMIT_TIME_0;
   local starttime;

   timestamp, starttime; // returns the number of seconds since Jan 1, 1970
                         // in 'starttime'

   LIMIT_TIME_0 = starttime;

}

/* Function check_time */

func check_time( period, unit= )
/* DOCUMENT within_time_period = check_time( period, unit= )

   Returns 1 (yes) as long as the time elapsed is shorter than
   'period'. The default time unit is hours but it can be set
   to 's' (seconds), 'm' (minutes) in stead by keyword 'unit'.
   (unit='h' will work as well).

   The time origin (saved in external 'LIMIT_TIME_0') must have
   been set by a call of 'set_time'
   or 'check_time()' that returns 1 and resets the time origin.
   If 'LIMIT_TIME_0' has not been defined then it will be set to
   the current time and '1' will be returned.

   SEE ALSO: set_time

   2008-11-01/NJW
*/
{
   extern LIMIT_TIME_0;
   local nowtime;

   if( is_void(LIMIT_TIME_0) && is_void(period) ) {
      set_time;
      return 1;
   }

   timestamp, nowtime;

   if( is_void(unit) ) {
      per = period * 3600.;
   } else if( unit == 's' ) {
      per = double(period);
   } else if( unit == 'm' ) {
      per = period * 60.;
   } else if( unit == 'h' ) {
      per = period * 3600.;
   } else error,"Illegal unit keyword";

   if( double(nowtime - LIMIT_TIME_0) < per ) {
      return 1;
   } else return 0;

}

/* Function get_time */

func get_time( void )
/* DOCUMENT time_in_sec = get_time( )

   Returns number of seconds since the last call of 'set_time'.

   The time origin (saved in external 'LIMIT_TIME_0') must have
   been set by a call of 'set_time'.

   SEE ALSO: set_time

   2009-07-12/NJW
*/
{
   extern LIMIT_TIME_0;
   local nowtime;

   if( is_void(LIMIT_TIME_0) ) {
      write,"set_time must be called first";
      return [];
   }

   timestamp, nowtime;

   return nowtime - LIMIT_TIME_0;

}

/* Function setbatch */

func setbatch( orig_arr, begin= )
{
   extern Orig_batch_arr, Num_batch_arr, Last_batch_index;

   Orig_batch_arr = orig_arr;
   if( is_void(begin) ) {
      Last_batch_index = 0;
   } else {
      if( begin < 1 ) begin = 1;
      Last_batch_index = begin - 1;
   }
   Num_batch_arr = numberof( orig_arr );
}

/* Function nextbatch */

func nextbatch( number )
{
   extern Orig_batch_arr, Num_batch_arr, Last_batch_index;

   i = Last_batch_index + 1;
   j = Last_batch_index + number;
   if( i > Num_batch_arr ) return [];
   if( j > Num_batch_arr ) j = Num_batch_arr;
   Last_batch_index = j;
   return Orig_batch_arr(i:j);

}

/* Function setup_dir */

func setup_dir( file, type= )
/* DOCUMENT setup_dir, file, type=

   if( type=='f' ) // default
      Assumes that 'file' is a file name (not a directory).
      Checks the existence of the directory where it is
      supposed to reside and create it if missing.
   if( type=='d' )
      Check the existence of the directory and create it if
      missing.
   Directory creation by Yorick command: mkdirp

   Default for type : 'f'

   2008-12-01/NJW
*/
{
   local filedir, filename;

   if( is_void(type) ) type = 'f';

   if( type == 'f' ) {
      file = fullpath(file);
      splitfname, file, filedir, filename;
   } else {
      filedir = file;
   }
   mkdirp, filedir;
}

/* Function mkdirp */

func mkdirp(dir)
/* DOCUMENT mkdirp, directory_name
     Create DIRECTORY_NAME, creating any missing parent directories
     (like UNIX utility mkdir -p).  Unlike mkdir, signals error if
     the creation is unsuccessful.  If DIRECTORY_NAME already exists
     and is a directory, mkdirp is a no-op.
   SEE ALSO: mkdir
*/
{
   dir = strtrim(dir, 2, blank="/") + "/";
   list = strfind("/", dir, n=1024);  /* assume <1024 components in dir */
   i = list(1:-1:2);
   list = i(where((i>0) & (list(2:0:2)>0)));
   for (i=numberof(list) ; i>=1 ; i--) {
      name = strpart(dir, [0,list(i)]);
      if (lsdir(name) != 0) break;
   }
   for (i++ ; i<=numberof(list) ; i++)
       mkdir, strpart(dir, [0,list(i)]);
   if (lsdir(dir) == 0) error, "mkdirp: failed to create "+dir;
}

/* Function wrdig */

func wrdig( x, ndec, sci= )
/* DOCUMENT wrdig, x, ndec, sci=
     prints scalar with indicated number of decimals
     Keyword 'sci' forces (scientific) e-format.

    SEE ALSO: atof, ftoa, f2scienota (all in idlx.i)
*/
{
   if( is_scalar(x) ) {
      write,format="%s\n", _wrdig( x, ndec, sci=sci);
   } else if( numberof(x) <= 6 ) { // Single line
      write,format="%s","[";
      for( i = 1; i <= numberof(x); i++ ) {
         post = i == numberof(x) ? "]" : ",";
         write,format="%s%s", _wrdig( x(i), ndec, sci=sci), post;
      }
      write,format="%s","\n";
   } else { // Column format
      for( i = 1; i <= numberof(x); i++ ) {
          write,format="%5i %s\n", i, _wrdig( x(i), ndec, sci=sci );
      }
   }
}

func _wrdig( x, ndec, sci= )
{
   if( is_void(ndec) ) ndec = 1;
   if( sci ) {
      fmt = "%."+swrite(format="%ie", ndec);
   } else {
      fmt = "%."+swrite(format="%if", ndec);
   }
   return swrite(format=fmt, x);
}


/* Function fndeci */

func fndeci( x, ndec )
/* DOCUMENT res = fndeci( x, ndec )

    Returns a double representation of 'x' with 'ndec' decimals

   2013-01-09/NJW
*/
{
   if( ndec < 0 ) error,"No negative number of decimals is accepted!";
   if( x < 0 ) {
      s = -1.0;
      ax = -x;
   } else {
      s = 1.0;
      ax = x;
   }
   f = 10.^ndec;
   return s * (floor(ax*f + 0.5)) / f;
}


/* Function rm */

func rm( filename, inq= )
/* DOCUMENT rm, filename, inq=

   Similar to the Unix/Linux 'rm' command
   Keyword 'inq' to be set for inquries

    2008-12-11/NJW
*/
{
   local basename, dirname;

   splitfname, filename, dirname, basename;

   list = file_search( basename, dirname );
   n = numberof(list);

   if( n == 0 ) {
      write,format="%s not found\n", filename;
      return;
   }

   for( i = 1; i <= n; i++ ) {
      ans = "y";
      if( inq ) read,prompt="Remove "+list(i)+" ? ... ", ans;
      if( ans == "y" ) remove, list(i);
   }
}

/* Function mv */

func mv( file_from, file_to, inq= )
/* DOCUMENT mv, file_from, file_to, inq=

   Similar to the Unix/Linux 'mv' command
   Unlike 'rename' mv accepts 'file_to' as a directory and
   then a file with same basename as 'file_from' will be
   created in that directory.
   If 'file_to' is a directory then wild cards may be used
   in 'file_from'.
   Keyword 'inq' for inquiry i.e. a prompt for user approval.

   SEE ALSO: rename, cp, _cp

   2008-12-10/NJW
*/
{
   local basename, dirname;

   // 'file_from' may not be a directory

   if( is_dir(file_from) ) {
      write,format="%s is a directory\n", file_from;
      return;
   }

   // if 'file_to' is a directory then check the
   // possibility for wild cards i.e. multi-moving

   if( is_dir(file_to) ) {
      splitfname, file_from, dirname, basename;
      filelist = file_search( basename, dirname );
      n = numberof( filelist );
      if( n == 0 ) {
         write,format="%s not found\n", file_from;
         return;
      }
      for( i = 1; i <= n; i++ ) {
         ans="y"; if(inq)read,prompt="Move "+filelist(i)+" ? ... ", ans;
         if( ans == "y" ) {
            splitfname, filelist(i), dirname, basename;
            filename = file_to+"/"+basename;
            _cp, filelist(i), filename;
            remove, filelist(i);
         }
      }
      return;
   }

   // from here it is a single file move

   if( !file_test( file_from ) ) {
      write,format="%s not found\n", file_from;
      return;
   }
   // 'file_to' must be a file name, get the directory
   // where it resides
   splitfname, file_to, dirname, basename;
   if( !file_test(dirname) ) {
      write,format="Directory: %s not found\n", dirname;
      return;
   }
   ans="y"; if(inq)read,prompt="Move "+file_from+" ? ... ", ans;
   if( ans == "y" ) {
      _cp, file_from, file_to;
      remove, file_from;
   }
   return;
}

/* Function is_dir */

func is_dir( pathname )
/* DOCUMENT res = is_dir( pathname )

   returns 1 if 'pathname' is a directory, 0 if it is a file,
   and void if it does not exist.

   2008-12-10/NJW
*/
{
   local dir, subdirs, fname;

   if( pathname == "." ) return 1;
   if( pathname == ".." ) return 1;
   if( !file_test( pathname ) ) return [];

   splitfname, pathname, dir, fname;
   lsdir, dir, subdirs;
   return anyof(fname==subdirs);
}

/* Function priarr */

func priarr( i_arr )
/* DOCUMENT priarr, i_arr

   Print integer array on screen on one line
*/
{
   for(i=1;i<=numberof(i_arr);i++) {
      write,format=" %i", i_arr(i);
   }
   write,"";
}


/* Function prstrarr */

func prstrarr( stream, str_arr, senu=, all= )
/* DOCUMENT prstrarr, stream, str_arr, senu=, all=

   Print string array on screen or to stream, one line per element
   except zero length strings.

   Keywords:
     senu     (bool) for adding line numbers.
      all     (bool) to include also zero length strings
*/
{
   if( structof(stream) == string ) {
      str_arr = stream;
      stream = [];
   }

   if( senu ) {
      for(i=1;i<=numberof(str_arr);i++) {
         if(strlen(str_arr(i)) || all) write,stream,format="%5i  %s\n", i, str_arr(i);
      }
   } else {
      for(i=1;i<=numberof(str_arr);i++) {
         if(strlen(str_arr(i)) || all) write,stream,format="%s\n", str_arr(i);
      }
   }
}

/* Function sixty */

func sixty( val_in )
/* DOCUMENT res = sixty
 *+
 * NAME:
 *       SIXTY()
 * PURPOSE:
 *       Converts a decimal number to sexigesimal.
 * EXPLANATION:
 *       Reverse of the TEN() function.
 *
 * CALLING SEQUENCE:
 *       X = SIXTY( SCALAR )
 *
 * INPUTS:
 *       SCALAR -- Decimal quantity.
 * OUTPUTS:
 *       Function value returned = floating real vector of three elements,
 *       sexigesimal equivalent of input decimal quantity.
 *       A negative number is signified by making the first non-zero
 *       element of the output vection negative.
 *
 * PROCEDURE:
 *       Mostly involves checking arguments and setting the sign.
 *
 * EXAMPLE:
 *       If x = -0.345d then sixty(x) = [0.0, -20.0, 24.0]
 *
 * MODIFICATION HISTORY:
 *       Written by R. S. Hill, STX, 19-OCT-87
 *       Output changed to single precision.  RSH, STX, 1/26/88
 *       Accept single element vector   W. Landsman   Sep. 1996
 *       Converted to IDL V5.0   W. Landsman   September 1997
 */
{

   ss = abs(3600.0 * val_in );
   mm = abs(60.0 * val_in );
   dd = abs( val_in );
   result = array(double,3);
   result(1) = double(int(dd));
   result(2) = double(int(mm - 60.0 * result(1)));
   result(3) = double(ss - 3600. * result(1) - 60.0 * result(2));
   if(  val_in  <  0.0 ) {
      if( result(1)  !=  0 ) {
         result(1) = -result(1);
      } else if( result(2)  !=  0 ) {
         result(2) = -result(2);
      } else {
         result(3) = -result(3);
      }
   }

   return result;
}
/* Function radec2sxg */

func radec2sxg(ra, dec, &ihr, &imin, &xsec, &ideg, &imn, &xsc)
/* DOCUMENT radec2sxg, >ra, >dec, >ihr, >imin, >xsec, >idet, >imn, >xsc

 *Convert from decimal to hrs,min
 *+
 * NAME:
 *       RADEC2SXG
 * PURPOSE:
 *       To convert right ascension and declination from decimal degrees to
 *       sexigesimal hours (for( R.A.)  and degrees( for Dec.).
 * CALLING SEQUENCE:
 *       radec,ra,dec,ihr,imin,xsec,ideg,imn,xsc
 * INPUTS:
 *       ra   - right ascension in decimal DEGREES, scalar or vector
 *       dec  - declination in decimal DEGREES, scalar or vector, same number
 *              of elements as RA
 * OUTPUTS:
 *       ihr  - right ascension hours   (INTEGER*2)
 *       imin - right ascension minutes (INTEGER*2)
 *       xsec - right ascension seconds  (REAL*4 or REAL*8)
 *       ideg - declination degrees (INTEGER*2)
 *       imn  - declination minutes (INTEGER*2)
 *       xsc  - declination seconds (REAL*4 or REAL*8)
 * RESTRICTIONS:
 *       RADEC does minimal parameter checking.
 * REVISON HISTORY:
 *       Written by B. Pfarr, STX, 4/24/87
 *-
 * Compute RA
 *
 */
{
   dr = pi / 180.0;
   ra = zero2pi(ra*dr)/dr;

   ihr = toint(ra/15.);
   xmin = abs(ra*4.0-ihr*60.0);
   imin = toint(xmin);
   xsec = (xmin-imin)*60.0;
   //
   // Compute Dec
   //
   ideg= toint(dec);
   xmn = abs(dec-ideg)*60.0;
   imn = toint(xmn);
   xsc = (xmn-imn)*60.0;
   //
   // Now test for the special case of zero degrees
   //
   zero_deg = (ideg  ==  0) & (dec  <  0);
   imn = imn - 2*imn*toint(zero_deg*(imn  !=  0));
   xsc = xsc - 2*xsc*zero_deg*(imn  ==  0);
   return;
}

/* Function ten */

func ten(dd,mm,ss)
/* DOCUMENT res = ten( dd, mm, ss)

   Returns degrees, minutes, seconds as decimal number.
   dd, mm, and ss may be arrays

*/
{

   if( is_void(dd) ) error,"TEN: bad input";
   if( is_void(mm) ) {
      return double(dd);
   }
   vector = [double(dd),double(mm)];
   if( !is_void(ss) ) {
      grow, vector, double(ss);
   }
   dms = dimsof(vector);
   if(dms(1)==1) {
      n_arr = 1;
      nv = dms(2);
      vector = reform(vector,1,dms(2));
   } else {
      n_arr = dms(2);
      nv = dms(3);
   }
   facs = [1.0,60.0,3600.0];
   res = array(double,n_arr);

   for( j = 1; j <= n_arr; j++ ) {
      sgn = anyof( vector(j,) < 0.0 ) ? -1.0 : 1.0;
      vector(j,) = abs(vector(j,));
      decim = vector(j,1);
      for(i=2;i<=nv;i++) decim += vector(j,i)/facs(i);
      res(j) = decim*sgn;
   }
   return n_arr > 1 ? res : res(1);
}

/* Function adstring */

func adstring(ra, dec, precision, truncate=)
/* DOCUMENT  res = adstring(ra, dec, precision, truncate=)

   Return RA and Dec as character string(s) in sexagesimal format.
   EXPLANATION:
         RA and Dec should be entered as
         two separate vectors (or scalars).  One can also specify the precision
         of the declination in digits after the decimal point.

   INPUTS:

         RA     - Right ascension in decimal degrees, numeric scalar or vector
         DEC    - Declination in decimal degrees, numeric scalar or vector

   OPTIONAL INPUT:
         PRECISION  - Integer scalar (0-4) giving the number of digits after the
                 decimal of DEClination.   The RA is automatically 1 digit more.
                 If no PRECISION parameter is passed, a
                 precision of 1 for both RA and DEC is returned to maintain
                 compatibility with past ADSTRING functions.    Values of
                 precision larger than 4 will be truncated to 4.    If
                 PRECISION is 3 or 4, RA and Dec should be input as
                 double precision.
   OPTIONAL INPUT KEYWORD:
         /TRUNCATE - if( set, ) the last displayed digit in the output is
                 truncated in precision rather than rounded.   This option is
                 useful if( ADSTRING() is used to form an official IAU name
                 (see http://vizier.u-strasbg.fr/Dic/iau-spec.htx) with
                 coordinate specification.   The IAU name will typically be
                 be created by applying STRCOMPRESS/REMOVE) after the ADSTRING()
                 call, e.g.
                strcompress( adstring(ra,dec,0,/truncate), /remove)   ;IAU format
   OUTPUT:
         RESULT - Character string(s) containing HR,MIN,SEC,DEC,MIN,SEC formatted
                 as ( 2I3,F5.(p+1),2I3,F4.p ) where p is the PRECISION
                 parameter.    If only a single scalar is supplied it is
                 converted to a sexigesimal string (2I3,F5.1).

   EXAMPLE:
         (1) Display CRVAL coordinates in a FITS header, H

         IDL> crval = sxpar(h,"CRVAL*")  ;Extract 2 element CRVAL vector (degs)
         IDL> write, adstring(crval)     ;Print CRVAL vector sexigesimal format

         (2)  write,adstring(30.42,-1.23,1)  ==>  " 02 01 40.80  -01 13 48.0"
              write,adstring(30.42,+0.23)    ==>  " 02 01 40.8   +00 13 48.0"
              write,adstring(+0.23)          ==>  "+00 13 48.0"

         (3) The first two calls in (2) can be combined in a single call using
             vector input
                write,adstring([30.42,30.42],[-1.23,0.23], 1)
   PROCEDURES CALLED:
         FSTRING(), RADEC, SIXTY()

 */
{

   dr = pi / 180.0;

   nra = numberof(ra);
   if( numberof(dec) != nra ) error,"ADSTRING error, mismatch in array lengths";
   if( is_scalar(ra) ) ra = [ra];
   if( is_scalar(dec) ) dec = [dec];

   ra = zero2pi(ra*dr)/dr;

   if( anyof( dec < -90.0 | dec > 90.0 ) ) error,"adstring, dec values outside 90 deg";

   local ihr, imin, xsec, ideg, imn, xsc;

   radec2sxg, ra, dec, ihr, imin, xsec, ideg, imn, xsc;

   if( is_void(precision) ) precision = 2;
   if( precision < 0 ) precision = 0;
   if( precision > 5 ) precision = 4;

   if( !truncate ) {
      roundsec = [59.5,59.95,59.995,59.9995,59.99995,59.999995];
      carry = where(xsec > roundsec(precision+2));
      Ncarry = numberof(carry);
      if( Ncarry > 0 ) {
         imin(carry)++;
         xsec(carry) = 0.0;
         mcarry = where(imin(carry) == 60);
         Nmcarry = numberof(mcarry);
         if( Nmcarry > 0 ) {
            ic = carry(mcarry);
            ihr(ic) = (ihr(ic) + 1) % 24;
            imin(ic) = 0;
         }
      }
   } else {
      xsec = (long(xsec*10^(precision+1)))/10.0^(precision+1);
   }

/*
 *      Deal with Right Ascension
 */

   secfmt = swrite(format="%%0%i.%if", 3+precision+1, precision+1);
   result = array(string, nra);
   for(i = 1; i <= nra; i++ ) {
      result(i) = swrite(format="%3.2i%3.2i "+secfmt, ihr, imin, xsec );
   }

/*
 *  Deal with declination
 */

   if( precision == 0 ) {

      secfmt = "%2.2i" ;
      if( !truncate ) {
         xsc = floor(xsc+0.5);
         carry = where(xsc == 60);
         Ncarry = numberof(carry);
         if( Ncarry > 0 ) {
            xsc(carry) = 0;
            imn(carry) = imn(carry) + 1;
         }
      }

   } else {

      secfmt = swrite(format="%%0%i.%if", 3+precision, precision);
      if( !truncate ) {
         ixsc = long(xsc + 0.5/10^precision);
         carry = where(ixsc >= 60);
         Ncarry = numberof(carry);
         if( Ncarry > 0 ) {
            xsc(carry) = 0.;
            imn(carry) = imn(carry) + 1;
         }
      } else {
         xsc = (long(xsc*10^precision))/10.0^precision;
      }

   }

   pos = dec >= 0.0;
   carry = where(imn == 60);
   Ncarry = numberof(carry);
   if( Ncarry > 0  ) {
      ideg(carry) = ideg(carry) - 1 + 2*pos(carry);
      imn(carry) = 0;
   }

   deg = array(string,nra);
   for(i = 1; i <= nra; i++ ) {
      deg(i) = swrite(format="%3.2i", ideg(i));
   }
   zero = where(ideg == 0);
   Nzero = numberof(zero);
   if( Nzero > 0 ) {
       negzero = where( dec < 0);
       Nneg = numberof(negzero);
       if( Nneg > 0 ) {
          ineg = zero(negzero);
          deg(ineg) = "-00";
          if( is_scalar(imn) ) {
             imn = abs(imn);
             xsc = abs(xsc);
          } else {
             imn(ineg) = abs(imn(ineg));
             xsc(ineg) = abs(xsc(ineg));
          }
       }
   }

   ipos = where(pos);
   Npos = numberof(ipos);
   if( Npos > 0 ) deg(ipos) =  "+" + strtrim(deg(ipos));

   if( precision == 0 ) {
      for( i = 1; i <= nra; i++ ) {
         result(i) += " "+deg(i)+swrite(format="%3.2i "+secfmt, \
                      imn(i), long(floor(xsc(i)+0.5)));
      }
   } else {
      for( i = 1; i <= nra; i++ ) {
         result(i) += " "+deg(i)+swrite(format="%3.2i "+secfmt, imn(i), xsc(i));
      }
   }

   return nra == 1 ? result(1) : result;

}

/* Function facul */

func facul( i, j )
/* DOCUMENT res = facul( i [,j] )

    returns i!/(j-1)!
    i.e.:   i * (i-1) * (i-2) * ... * j
    where j < i.

    If 'j' is not given it is assumed to be one.

    2009-04-15/NJW
*/
{
   if( i < 2 ) return 1;
   if( is_void(j) ) j = 1;
   if( i == j ) return i;
   if( i == j-1 ) return 1;
   if( i < j+1 ) error,"Bad input to function 'facul'";
   r = j;
   for(k=j+1;k<=i;k++) r *= k;
   return r;
}

/* Function lfacul */

func lfacul( i, j )
/* DOCUMENT res = lfacul( i [,j] )

    returns log( i!/(j-1)! )
    i.e.:   log(i * (i-1) * (i-2) * ... * j)
    where j < i.

    If 'j' is not given it is assumed to be one.

    2010-05-26/NJW, based on 'facul'
*/
{
   if( i < 2 ) return 0.0;
   if( is_void(j) ) j = 1;
   if( i == j ) return log(i);
   if( i == j-1 ) return 0.0;
   if( i < j+1 ) error,"Bad input to function 'lfacul'";
   r = log(j);
   for(k=j+1;k<=i;k++) r += log(k);
   return r;
}

/* Function lbico */

func lbico( n, m )
/* DOCUMENT res = lbico( n, m )

                ( m )
     returns log(   )  = log( n! / ( m! * (n-m)!) )
                ( n )

     i.e. the natural logarithm of the binomial coefficient.

      SEE ALSO: bico (in 'gamma.i' under yorick-2.1/i)
                lfacul

*/
{
   if( m > n ) {
      write,"Warning: lbico called with m > n - returning -1.0";
      return -1.0;
   }
   return lfacul(n,m+1) - lfacul(n-m);
}

/* Function rem_elem */

func rem_elem( arr, i )
/* DOCUMENT new_arr = rem_elem( arr, i )

   Return an array of same type as 'arr' but with
   element(s) 'i' removed. 'i' may be an array.
   'arr' can have several dimensions but the action
   is on the last index.

   Returns void ([]) if all elements are to be removed.

*/
{
   if( numberof(i) == 0 ) return arr;
   new_arr = arr;
   dms = dimsof(arr);
   n = dms(0);
   // if last dimension is 1 and 1 is included in 'i' then return void
   if( n == 1 && nallof(i-1) ) return [];
   k = 0;
   for(j=1;j<=n;j++) {
      // if no match then keep the array element
      if( allof(i-j) ) new_arr(..,++k) = arr(..,j);
   }
   if( k == 0 ) return []; // 'k' was never updated so nothing remains
   return new_arr(..,1:k);
}

/* Function filesize */

func filesize( filename )
/* DOCUMENT number_of_bytes = filesize( filename )
*/
{
   f = open(filename,"rb");
   s = sizeof(f);
   close,f;
   return s;
}

/* Function pfiles_path */

func pfiles_path( filename )
/* DOCUMENT parfile = pfiles_path( parfile_name )

   Returns the complete path to the parameter file given
   by 'parfile_name'. Searches first the current directory,
   then the directories given in the environment variable
   PFILES separated by colons but it stops at the (first)
   semicolon.

   Returns void ([]) if file is not found.

   2009-07-09/NJW, updated 2017-03-21/NJW
*/
{
   // First search current directory

   if( file_test( filename ) ) return fullpath( filename );

   // If not here then get the environment variable 'PFILES'

   pfiles = get_env("PFILES");
   if( strlen(pfiles) == 0 ) {
      write,"Warning, PFILES is not set";
      return [];
   }
   // remove part following ';'
   pfiles = strsplit( pfiles,";")(1);
   str = strsplit( pfiles, ":");
   n_str = numberof(str);
   for( i = 1; i <= n_str; i++ ) {
      pth = str(i)+"/"+filename;
      if( file_test(pth) ) return pth;
   }
   return [];
}

/* Function delta */

func delta( arr )
/* DOCUMENT darr = delta( arr )

   Returns the pairwise difference between successive
   elements. The first one is copied to keep the
   number of elements.

*/
{
   d = arr - shift(arr,-1);
   d(1) = d(2);
   return d;
}

/* Function struc */

func struc( str )
/* DOCUMENT structure = struc( char_type )

   Returns the structure that corresponds to 'char_type'
   's' : string, 'i' : integer, 'j' : long,
   'f' : float, 'd' : double, 'c' : char

   e.g. a = array(struc('d'),10)
   will produce an array of doubles.

*/
{
   if( typeof(str) == "string" ) {
      cstr = (*pointer(str))(1);
   } else if( typeof(str) == "char" ) {
      cstr = str;
   } else error,"Bad argument for STRUC";

   if( cstr == 's' ) {
      return string;
   } else if( cstr == 'i' ) {
      return int;
   } else if( cstr == 'j' ) {
      return long;
   } else if( cstr == 'f' ) {
      return float;
   } else if( cstr == 'd' ) {
      return double;
   } else if( cstr == 'c' ) {
      return char;
   } else return [];
}

/* Function _str2x */

func _str2x( str, dtype )
/* DOCUMENT value = _str2x( str, dtype )

   Converts a string in 'str' to a value of
   type given by 'dtype':

   's' : string, 'i' : integer, 'j' : long,
   'f' : float, 'd' : double, 'c' : char
*/
{
   if( dtype == 's' ) {
      return str;
   } else if( dtype == 'i') {
      return int(atoi(str));
   } else if( dtype == 'j') {
      return atoi(str);
   } else if( dtype == 'f') {
      return float(atof(str));
   } else if( dtype == 'd') {
      return atof(str);
   } else if( dtype == 'c') {
      return char(atoi(str));
   } else return [];
}

/* Function ytar */

/*****************************************************
 *   A tar system operated by Yorick
 *****************************************************/

func ytar( option, ytarfilename, list_of_files_or_filename, safe= )
/* DOCUMENT ytar, option, ytarfilename, list_of_files_or_filename, safe=

   A tar like function that only operates on text files.

   Option  "c"  create archive
           "l"  list archive contents
           "x"  extract files or named file
                 Accepts space or comma separated list with
                 exact file names
        A  "v"  can be added for verbosity

   Keyword  safe : When set a backup will be done before replacing
                   if the file to be extracted already exists
*/
{
   if( 1 != strmatch(option,"c")+strmatch(option,"l")+strmatch(option,"x") ) {
      write,"Illegal option for ytar";
      return;
   }
   v = strmatch(option,"v"); // flag for verbosity


   if( strmatch(option,"c") ) {
      nfiles = numberof(list_of_files_or_filename);
      f = open(ytarfilename,"w");
      write,f,format="%%YTAR%% Header %s\n", ndate(3);
      for( i = 1; i <= nfiles; i++ ) {
         text = rdfile(list_of_files_or_filename(i));
         if( !(ntext=numberof(text)) ) continue;
         if(v) write,format="Adding %s (%i lines)\n", list_of_files_or_filename(i), ntext;
         write,f,format="%%FILE%% %s\n", list_of_files_or_filename(i);
         for( j = 1; j <= ntext; j++ ) write,f,format="%s\n",text(j);
      }
      close,f;
   } else if( strmatch(option,"x") ) {
      text = rdfile(ytarfilename);
      // test file contents
      if( strpart(text(1),1:6) != "%YTAR%" ) {
         write,"Not a ytar file - quit";
         return;
      }
      // see if named files are to be extracted
      nxlist = 0;
      if( !is_void(list_of_files_or_filename) ) {
         llist = strcharrepl( list_of_files_or_filename," ",",");
         xlist = strsplit(llist,",");
         nxlist = numberof(xlist);
      }
      ntext = numberof(text);
      file_is_open = 0;
      for( j = 2; j <= ntext; j++ ) {
         if( strpart(text(j),1:6) == "%FILE%" ) {
            if( file_is_open ) {
               close, f;
               file_is_open = 0;
            }
            filename = strtrim(strpart(text(j),7:0));
            ok = 1;
            if( nxlist ) { // see if this is wanted
               ok = anyof( filename == xlist );
            }
            if( ok ) {
               if( safe ) back, filename;
               if(v) write,"Creating "+filename;
               f = open(filename,"w");
               file_is_open = 1;
            }
         } else {
            if( ok ) write,f,format="%s\n", text(j);
         }
      }
      close,f;
   } else if( strmatch(option,"l") ) {
      text = rdfile(ytarfilename);
      // test file contents
      if( strpart(text(1),1:6) != "%YTAR%" ) {
         write,"Not a ytar file - quit";
         return;
      }
      ntext = numberof(text);
      for( j = 2; j <= ntext; j++ ) {
         if( strpart(text(j),1:6) == "%FILE%" ) {
            filename = strtrim(strpart(text(j),7:0));
            write,format="%s\n", filename;
         }
      }
   } else {
      write,"Illegal option - quit";
   }
}

/* Function bytar */

/*****************************************************
 *   A binary tar system operated by Yorick
 *****************************************************/

func bytar( option, ytarfilename, list_of_files_or_filename, safe= )
/* DOCUMENT bytar, option, ytarfilename, list_of_files_or_filename, safe=

   A tar like function with limited functionality  that operates
   on all kinds of files.

   Option  "c"  create archive
           "l"  list archive contents
           "x"  extract files or named file
                 Accepts space or comma separated list with
                 exact file names
           A "v" can be added for verbosity

   Keyword  safe : When set a backup will be done before replacing
                   if the file to be extracted already exists

   The archive is formatted as

      %BYTAR% Header YYYY-MM-DDTHH:MM:SS {filename number_of_bytes file_content}

   where {...} is repeated as many times as required.

*/
{
   c = char(0);
   cheader = array(char,35);
   nb = int(0);

   // check the integrity of the option
   nsum = 0;
   if( strmatch(option,"c") ) nsum++;
   if( strmatch(option,"l") ) nsum++;
   if( strmatch(option,"x") ) nsum++;
   if( nsum != 1 ) {
      write,"Illegal option for bytar";
      return;
   }
   v = strmatch(option,"v"); // flag for verbosity

   if( strmatch(option,"c") ) {
      nfiles = numberof(list_of_files_or_filename);
      f = open(ytarfilename,"wb");
      signature = swrite(format="%%BYTAR%% Header %s", ndate(3));
      csig = *pointer(signature);
      address = 0;
      _write,f,address,csig;
      address += sizeof(csig);
      for( i = 1; i <= nfiles; i++ ) {
         nbytes = filesize(list_of_files_or_filename(i));
         if( !nbytes ) continue;
         if(v)write,format="%s (%i)\n", list_of_files_or_filename(i), nbytes;
         cname = *pointer(list_of_files_or_filename(i));
         _write,f,address,cname;     address += sizeof(cname);
         _write,f,address,nbytes;    address += sizeof(nbytes);
         fi = open(list_of_files_or_filename(i),"rb");
         for( j = 0; j < nbytes; j++ ) {
            _read,fi,j,c;
            _write,f,address+j,c;
         }
         close,fi;
         address += nbytes;
      }
      close,f;
      remove, ytarfilename+"L";

   } else if( strmatch(option,"x") ) {

      nbtar = filesize(ytarfilename);
      f = open(ytarfilename,"rb");
      address = 0;
      _read,f,address,cheader;   address += sizeof(cheader);
      header = string(&cheader);
      // test file signature
      if( strpart(header,1:15) != "%BYTAR% Header " ) {
         write,"Not a bytar file - quit";
         close,f;
         return;
      }
      fdate = strpart(header,16:34);
      write,format="Date of bytar file: %s\n", fdate;

      // see if named files are to be extracted
      nxlist = 0;
      if( !is_void(list_of_files_or_filename) ) {
         llist = strcharrepl( list_of_files_or_filename," ",",");
         xlist = strsplit(llist,",");
         nxlist = numberof(xlist);
      }
      file_is_open = 0;

      while( address < nbtar-1 ) {

         // get filename of next file in archive
         cname = [];
         do {
            _read,f,address++,c;
            grow,cname,c;
         } while( c );
         filename = string(&cname);

         _read,f,address,nb; address += sizeof(nb);

         // set 'ok' flag if this file is to be extracted

         ok = 1;
         if( nxlist ) { // see if this is wanted
            ok = anyof( filename == xlist );
         }
         if( ok ) {
            if(v)write,format="%s (%i)\n", filename, nb;
            if( safe ) {
               if( file_test(filename) ) back, filename;
            }
            fi = open(filename,"wb");
            for( j = 0; j < nb; j++ ) {
               _read,f,address+j,c;
               _write,fi,j,c;
            }
            close,fi;
            remove,filename+"L";
         }
         address += nb;
      }
      close,f;

   } else if( strmatch(option,"l") ) {

      nbtar = filesize(ytarfilename);
      f = open(ytarfilename,"rb");
      address = 0;
      _read,f,address,cheader;   address += sizeof(cheader);
      header = string(&cheader);
      // test file signature
      if( strpart(header,1:15) != "%BYTAR% Header " ) {
         write,"Not a bytar file - quit";
         close,f;
         return;
      }
      fdate = strpart(header,16:34);
      write,format="Date of bytar file: %s\n", fdate;

      while( address < nbtar-1) {

         // get filename of next file in archive
         cname = [];
         do {
            _read,f,address++,c;
            grow,cname,c;
         } while( c );
         filename = string(&cname);

         _read,f,address,nb; address += sizeof(nb);
         write,format="%s (%i)\n", filename, nb;

         address += nb;
      }
      close,f;


   } else {

      write,"Illegal option - quit";

   }

}

/* Function lettername */

func lettername( n, nlet, lc= )
/* DOCUMENT res = lettername( n, nlet, lc= )

   Returns a string (or an array of strings) with letter representation
   of the integer number 'n' (that can be an array). The number of letters
   is 'nlet'.
   Default is basis 'A' but lower case can be selected with keyword 'lc'

   'n' must be larger than zero.

   2009-11-13/NJW
*/
{
   if( anyof(n <= 0) ) error,"Bad argument, n must be > 0";
   nn = numberof(n);
   if( nn > 1 ) {
      s = array(string,nn);
      for( i = 1; i <= nn; i++ ) s(i) = _lettername( n(i), nlet, lc=lc );
   } else {
      s = _lettername( n, nlet, lc=lc );
   }
   return s;
}

func _lettername( n, nlet, lc= )
{
   if( n > 26^nlet ) error,"_LETTERNAME n is too large";
   cA = lc ? *pointer("a") : *pointer("A");
   la = array(int,nlet);
   for(i=1;i<nlet;i++) {
      la(i) = (n-1)/26^(nlet-i);
      n -= la(i)*26^(nlet-i);
   }
   la(i) = n-1;
   c = char(la)+cA(1);
   grow,c,cA(2);
   return string(&c);
}

func _letternamex( n, nlet )
/* DOCUMENT str = _letternamex( n, nlet )

   Return n'th string starting with "0000" (size depends
   on 'nlet' in the 62 period sequence 0-9,A-Z,a-z
   2009-11-18/njw
*/
{
   if( n > 62^nlet ) error,"_LETTERNAMEX n is too large";
   la = array(int,nlet);
   ca = array(char,nlet);
   for(i=1;i<nlet;i++) {
      la(i) = (n-1)/62^(nlet-i);
      n -= la(i)*62^(nlet-i);
      ca(i) = _tochar(la(i));
   }
   la(i) = n-1;
   ca(i) = _tochar(la(i));
   grow,ca,0;
   return string(&ca);
}
func _letternamey( n, nlet )
/* DOCUMENT str = _letternamey( n, nlet )

   Return n expressed in a base 62 number system
   using the symbol sequence: 0-9,A-Z,a-z
   2009-11-18/njw
*/
{
   if( n >= 62^nlet ) error,"_LETTERNAMEY n is too large";
   la = array(int,nlet);
   ca = array(char,nlet);
   for(i=1;i<nlet;i++) {
      la(i) = n/62^(nlet-i);
      n -= la(i)*62^(nlet-i);
      ca(i) = _tochar(la(i));
   }
   la(i) = n;
   ca(i) = _tochar(la(i));
   grow,ca,0;
   return string(&ca);
}
func _tochar( n )
{
   if( n > 61 || n < 0 ) return '?';
   ca = 'a'; cA = 'A'; c0 = '0';
   if( n <= 9 ) return char(c0 + n);
   if( n <= 35 ) return char(cA + n - 10);
   return char(ca + n - 36);
}

/* Function pwd */

func pwd( void )
/* DOCUMENT pwd
     or  res = pwd()

    First version:  Prints the current directory name
    Second version: Returns the current directory name
*/
{
   if( am_subroutine() ) write, get_cwd();
   else return get_cwd();
}

/* Function indices */

func indices( arr, j )
/* DOCUMENT indices( arr, j )

    If 'arr' is scalar then return 1
    If 'arr' is 1D then
       If 'j' is scalar then return [j]
       Else return j(1)
    If 'arr' is 2D or more then
       If 'j' is scalar return array with
           indices of 'arr'
       If 'j' is a vector and has as many elements as
           'arr' has dimensions then return single index

    2010-01-07/NJW
*/
{
   dms = dimsof(arr);
   if( dms(1) == 0 ) return 1;
   ndms = dms(1);
   jdms = dimsof(j);
   if( dms(1) == 1 ) {
      if( jdms(1) == 0 ) return [j];
      if( jdms(1) == 1 && jdms(2) == 1 ) return j(1);
      error,"INDICES ##1##";
   }

   // When you come here then 'arr' has 2 or more dimensions
   // so decide it 'j' is a scalar or has as many dimensions
   // as 'arr'

   if( jdms(1) != 0 && jdms(2) != ndms ) error,"INDICES ##2##";

   far = gar = dms(1:-1);
   far(1) = 1;
   for( k = 3; k <= dms(1); k++ ) {
      gar(k:0) = dms(k-1);
      gar(1:k-1) = 1;
      far *= gar;
   }

   // Case for determining single index

   if( jdms(1) == 1 ) {

      return sum((j-1)*far)+1;

   } else {  // Case for determining index vector

      idx = array(int,dms(1));
      id = j-1;
      for( k = dms(1); k >= 1; k-- ) {
         idx(k) = id/far(k) + 1;
         id -= (idx(k)-1)*far(k);
      }
      return idx;
   }
}

/* Function indexnearest */

func indexnearest( arr, value )
/* DOCUMENT indexnearest( arr, value )

   Returns the index (or indices) of 'arr' where
   the value is nearest to 'value'

*/
{
   return where(min(abs(arr-value)) == abs(arr-value));
}

/* Function is_scalar */

func is_scalar( arg )
{
    d = dimsof(arg);
    return numberof(d) == 1 && d(1) == 0;
}

/* Function ls */

func ls( arg1, arg2, type=, width= )
/* DOCUMENT ls, dir, sel, type=, width=
            ls, sel, type=, width=
            ls, type=, width=

   First version: Displays files in directory 'dir' using   
                  selection string 'sel'
   Second version: Displays files in current directory using   
                  selection string 'sel'
   Third version: Displays files in current directory without
                  any selection

   When called as a function e.g. list = ls(dir, sel, type= )
   a list of files is returned and the screen message is suppressed.

   'type' can be "f" or "d" (files or directories)
           default is both
   'width' is terminal width (default: 80)
*/
{
   local subdirs;

   if( is_void(arg2) ) {
      dir = "."; // second or third version
      sel = arg1;
   } else {
      if( is_void(arg1) ) {
         write,"Illegal arguments for 'ls'";
         return;
      } else {
         dir = arg1; // first version
         sel = arg2;
      }
   }
   if( is_void(width) ) width = 80;
   if( typeof(sel) == "string" ) { // search by 'file_search'

      files = file_search( sel, dir );
      nfiles = numberof(files);
      if( nfiles ) {
         w = where(strpart(files,1:2)=="./");
         if( numberof(w) ) files(w) = strpart(files,3:0);
      }
      nsubd = 0;
      type = "f"; // avoid directories, override user''s choice

   } else { //   search by 'lsdir'

      files = lsdir( dir, subdirs );
      if( structof(files) == long ) {
         write,"No such directory: "+dir;
         return []
      }
      nfiles = numberof(files);
      // remove files where names begin with "."
      if( nfiles ) {
         w = where( strpart(files,1:1) != "." );
         if( (nfiles = numberof(w)) ) files = files(w);
      }
      nsubd  = numberof(subdirs);
      if( nsubd ) {
         subdirs += "/";
         w = where( strpart(subdirs,1:1) != "." );
         if( (nsubd = numberof(w)) ) subdirs = subdirs(w);
      }
   }

   ntyp = 3;
   if( typeof(type) == "string" ) {
      if( type == "f" ) ntyp = 1; 
      if( type == "d" ) ntyp = 2; 
   }
   if( ntyp == 1 && nfiles == 0 ) {
      write,"No plain files found";
      return;
   }
   if( ntyp == 2 && nsubd == 0 ) {
      write,"No directories found";
      return;
   }
   list = [];
   if( ntyp%2 == 1 ) grow, list, files;
   if( ntyp   >= 2 ) grow, list, subdirs;

   if( am_subroutine() ) {
      if( !is_void(list) ) {
         maxlen = max(strlen(list));
         ncol = maxlen > width ? 1 : width / (maxlen+2);
         arrange_words_in_columns,list,ncol,maxlen+2;
      } else write," < no files found >";
   }
   return list;
}

/* Function more */

func more( filename, nl= )
/* DOCUMENT more, filename, nl=

     displays the contents of the file on the screen.

     Keyword 'nl' gives the number of lines per display
                  (defaults to 20).
                  Setting it to zero causes the entire file
                  to be written.
*/
{
   if( is_void(nl) ) nl = 20;
   if( nl == 0 ) nl = 1000000;
   ans = "";
   text = rdfile(filename);
   ntext = numberof(text);
   nchunks = (ntext-1)/nl + 1;
   for( i = 1; i <= nchunks; i++ ) {
      i1 = (i-1)*nl + 1;
      i2 = i*nl;
      if( i2 > ntext ) i2 = ntext;
      prstrarr,text(i1:i2);
      if( i2 == ntext ) {
         write,format="      *** Done! *** line: %i, %.1f %% ***\n", i2, (100.*i2)/ntext;
      } else {
         write,format="      *** line: %i, %.1f %% ***\n", i2, (100.*i2)/ntext;
      }
      if( i < nchunks ) {
         // read,prompt=" ...",ans;
         ans = rdline(prompt=" ... ");
         if( ans == "x" || ans == "q" ) break;
      }
   }
}

/* Function vi */

func vi( filename )
/* DOCUMENT vi, filename

   Wrapper to call the 'vi' editor
*/
{
   system,"vi "+filename;
}

/* Function whereany */

func whereany( arr, target )
/* DOCUMENT idx = whereany( arr, target)

   Returns the list of indices where 'arr' matches any
   of the values in 'target'
*/
{
   if( !(na = numberof(arr)) ) return [];
   if( !(nb = numberof(target)) ) return [];
   //+ tic,2; is = sort(arr); write,"whereany: sort(arr) takes "+ftoa(tac(2),ndec=3)+" s";
   is = sort(arr);
   as = arr(is);
   //+ tic,2; js = sort(target); write,"whereany: sort(target) takes "+ftoa(tac(2),ndec=3)+" s";
   target = target(sort(target));
   target = target(uniq(target)); // no need for repetitions in 'target'
   nb = numberof(target);
   idx = array(long,na);
   k = 0;
   j = 1; // index in 'target' array
   //+ tic,2;
   // --- Loop over all elements in 'arr' testing against target value
   // --- that is increased i.e. stepping to the next one, whenever the
   // --- 'arr' value surpasses it.
   for( i = 1; i <= na; i++ ) {
      loopstart:
      if( as(i) < target(j) ) continue;
      if( as(i) == target(j) ) {
         idx(++k) = is(i);
         continue;
      } else { //    as(i) > target(j)
         j++;  // move forward in 'target'
         if( j > nb ) break; // we're done
         goto loopstart; // with updated 'j' as(i) may match
      }
   }
   //+ write,"whereany: Loop takes "+ftoa(tac(2),ndec=3)+" s";
   if( k ) {
      idx = idx(1:k);
      return idx(sort(idx));
   } else return [];
}

/* Function n_yoricks */

func n_yoricks( a )
/* DOCUMENT n_processes = n_yoricks()

   Returns the current number of yorick processes running.

   Works only on Unix/Linux systems.
*/
{
   user = get_env("USER");
   return atoi(rdfile(popen("ps -u "+user+" | grep yorick | wc -l",0))(1));
}

/* Function n_idls */

func n_idls( a )
/* DOCUMENT n_processes = n_idls()

   Returns the current number of IDL processes running.

   Works only on Unix/Linux systems.
*/
{
   user = get_env("USER");
   return atoi(rdfile(popen("ps -u "+user+" | grep idl | wc -l",0))(1));
}

/* Function docu */

/************************************************************

    Functions for PostScript image (picture) documentation
    in the PostScript file itself.

    2011-11-24/NJW

************************************************************/

extern docudoc;
/* DOCUMENT Functions for image documentation

    docui       Initalization of next documentation file
    docuf       Insert name of used function into the documentation file
    docud       Insert name of data file into the documentation file
    docuc       Insert name of file with code into the documentation file
    docup       Update PostScript file with documentation
    docum       Prints current doc. file on terminal
    docus       Insert string or string array into the documentation file
    docut       Append text file to the documentation file
    docux       Extract documentation lines from PS-file

    Example sequence:
      > docui, gn=1; // Start a new documentation file and include
                     // the filename in the text
      > docuc,"my_bunch_of_functions.i";
      > docuf,"butanification";
      > docus,"used on the data file:";
      > docud,"measurements.fits";
      > docum; // Check the current documentation file
      > docup, "yplot_0033.ps"; // Append documentation
      > text = docux( "yplot_0033.ps" ); // Retrieve documentation
*/

/* Function docui */

func docui( gn= )
{
   extern Docu_filename;
   Docu_filename = get_next_filename("docu_????.txt");
   write,"External: Docu_filename is now "+Docu_filename;
   if( gn ) {
      f = open( Docu_filename,"w" );
      write,f,format="%s","\n";
      write,f,format="Documentation from %s\n", fullpath(Docu_filename);
      write,f,format="%s","\n";
      close, f;
   }
}
func docum( void )
{
   extern Docu_filename;
   if( is_void(Docu_filename) ) {
      write,"No documentation file has been defined";
      return;
   }
   prstrarr,rdfile(Docu_filename);
}
func docuf( funcname )
{
   extern Docu_filename;
   if( is_void(Docu_filename) ) docui; // Create if missing
   f = open( Docu_filename,"a" );
   write,f,format="Function name: %s\n", funcname;
   close,f;
}
func docus( instring )
{
   extern Docu_filename;
   if( is_void(Docu_filename) ) docui; // Create if missing
   f = open( Docu_filename,"a" );
   for( i = 1; i <= numberof(instring); i++ ) write,f,format="%s\n", instring(i);
   close,f;
}
func docud( filename )
{
   extern Docu_filename;
   if( is_void(Docu_filename) ) docui; // Create if missing
   f = open( Docu_filename,"a" );
   write,f,format="Data file name: %s\n", filename;
   close,f;
}
func docut( filename )
{
   extern Docu_filename;
   if( !file_test(filename) ) {
      write,"Did not find "+filename;
      return;
   }
   if( is_void(Docu_filename) ) docui; // Create if missing
   docus, rdfile(filename);
}
func docuc( codefile )
{
   extern Docu_filename;
   if( is_void(Docu_filename) ) docui; // Create if missing
   f = open( Docu_filename,"a" );
   write,f,format="Code file: %s\n", codefile;
   close,f;
}
func docup( psfilename )
{
   extern Docu_filename;
   if( is_void(psfilename) ) {
      write,"Ups, you forgot the psfilename";
      return;
   }
   if( !file_test(psfilename) ) {
      write,"Did not find "+psfilename;
      return;
   }
   if( is_void(Docu_filename) ) docui; // Create if missing
   f = open( psfilename,"a" );
   text = rdfile( Docu_filename );
   for( i = 1; i <= numberof(text); i++ ) {
      write,f,format="%%docu%% %s\n", text(i);
   }
   close,f;
}
func docux( psfilename )
{
   extern Docu_filename;
   if( is_void(psfilename) ) {
      write,"Ups, you forgot the psfilename";
      return [];
   }
   if( !file_test(psfilename) ) {
      write,"Did not find "+psfilename;
      return [];
   }
   lines = rdfile( psfilename );
   a = strfind( "%docu%", lines );
   w = where( a(1,) == 0 & a(2,) == 6 );
   nw = numberof( w );
   if( numberof( w ) ) return strpart(lines(w),8:);
   // start from position#8 to avoid initial blank
   return [];
}

/* Function cleanl */

func cleanl( void )
/* DOCUMENT cleanl

    Remove all files in current directory whose names end with 'L'.

*/
{
   list = file_search("*L");
   if( (n = numberof(list)) ) {
      for( i = 1; i <= n; i++ ) remove,list(i);
      write,"Removed "+itoa(n)+" files.";
   } else {
      write,"Nothing to do.";
   }
}

/* Function ghost */

func ghost( void )
/* DOCUMENT hostname = ghost()
         or  ghost

    If the hostname is composed of several terms like
    tesla.spacecenter.dk then only the first part ('tesla')
    is returned.
*/
{
   host = get_env("HOST");
   if( am_subroutine() ) {
      write,host;
   } else return strsplit(host,".")(1);
}

/* Function prtmat */

func prtmat( arr, fmt=, stream= )
/* DOCUMENT  prtmat, arr, fmt=, stream=
          res = prtmat( arr, fmt= )

    Nice print of the transpose of 'arr'.

    See documentation for 'prmat'
*/
{
   prmat, transpose(arr), fmt=fmt, stream=stream;
}

/* Function prmat */

func prmat( arr, fmt=, stream= )
/* DOCUMENT prmat, arr, fmt=, stream=
       res = prmat( arr, fmt= )

   First version (subroutine):
     Nice print to 'stream' of a 2D matrix
   Second version (function);
     Returns string array with a 2D matrix

     'fmt' is for a single number e.g. "%8.3f" or "%13.5e"
     Default for float is "%11.3e", default for int is "%8i".

     If 'fmt' is given then it will decide the whether int's or floats
     are printed. Else the type of 'arr' will determine.
     'fmt' is saved as an external ('Fmt_prmat') so that subsequent calls of 'prmat'
     will reuse the value unless the keyword is given again.
*/
{
   extern Fmt_prmat;

   if( !is_void(fmt) ) {
      Fmt_prmat = fmt; // save for future use in the same session
   } else {
      if( !is_void(Fmt_prmat) ) fmt = Fmt_prmat; // restore saved format
   }
   dms = dimsof( arr );
   if( dms(1) != 2 ) error,"Not a 2D array";
   // define 'natural' type
   tp = typeof(arr);
   tp = tp == "long" || tp == "int" ? 0 : 1;

   // check if 'fmt' forces a type
   if( typeof(fmt) == "string" ) {
      if( strpart(fmt, 0:0 ) == "i" ) tp = 0;
      else tp = 1;
   } else {
      fmt = tp == 1 ? "%11.3e" : "%8i";
   }
   ty = tp == 1 ? double : long;

   // extract size of format indicator
   flen = "";
   for( p = 2; ; p++ ) {
      ss = strpart(fmt,p:p);
      if( !is_digit(ss) ) break;
      flen += ss;
   }
   hyphens = "";
   for(j=1;j<=atoi(flen);j++) hyphens += "-";

   if( am_subroutine() ) { // write to 'stream'
      write,stream,format="%s","    ";
      for(j=1;j<=dms(3);j++) write,stream,format="%"+flen+"i",j;
      write,stream,format="\n%s","    ";
      for(j=1;j<=dms(3);j++) write,stream,format="%s",hyphens;
      write,stream,"";
      for( i = 1; i <= dms(2); i++ ) {
         write,stream,format="%3i:", i;
         for( j = 1; j <= dms(3); j++ ) write,stream,format=fmt, ty(arr(i,j));
         write,stream,"";
      }
   } else {                // return string array
      res = array( string, dms(2)+2 );
      res(1) = "    ";
      for(j=1;j<=dms(3);j++) res(1) += swrite(format="%"+flen+"i",j);
      res(2) = "    ";
      for(j=1;j<=dms(3);j++) res(2) += swrite(format="%s",hyphens);
      for( i = 1; i <= dms(2); i++ ) {
         res(i+2) = swrite(format="%3i:",i);
         for( j = 1; j <= dms(3); j++ ) res(i+2) += swrite(format=fmt, ty(arr(i,j)));
      }
      return res;
   }

}

/* Function add13 */

func add13( file_in, file_out )
/* DOCUMENT add13, file_in, file_out

     Insert a CR (13) before each occurrence of LF (10)
     in a similar way for Unix to Dos file conversion.
 
     SEE ALSO remove13 and tdosunix.
*/
{
   if( !file_test(file_in) ) error,"Did not find "+file_in;

   f_in = open( file_in,"rb" );
   f_out = open( file_out,"wb" );

   i = 0;  j = 0;

   v = array(char,1);
   v13 = array(char(13),1);
   address_in = 0;
   address_out = 0;
   while( _read( f_in, address_in++, v ) == 1 ) {
      i++;
      if( v(1) == 10 ) {
         _write,f_out, address_out++, v13;
         j++;
      }
      _write,f_out,address_out++, v;
      j++;
   }
   close, f_in;
   close, f_out;
   write,format="%i bytes input, %i bytes output\n", i, j;
   remove, file_out+"L";
}

/* Function remove13 */

func remove13( file_in, file_out )
/* DOCUMENT remove13, file_in, file_out

    Remove all occurrences of CR (13)
    in a similar way as DOS to Unix file conversion.
 
     SEE ALSO add13 and tdosunix.
*/
{
   if( !file_test(file_in) ) error,"Did not find "+file_in;

   f_in = open( file_in,"rb" );
   f_out = open( file_out,"wb" );

   i = 0;  j = 0;

   v = array(char,1);
   address_in = 0;
   address_out = 0;
   while( _read( f_in, address_in++, v ) == 1 ) {
      i++;
      if( v(1) != 13 ) {
         _write,f_out, address_out++, v;
         j++;
      }
   }
   close, f_in;
   close, f_out;
   write,format="%i bytes input, %i bytes output\n", i, j;
   remove, file_out+"L";
}

/* Function tdosunix */

func tdosunix( filename )
/* DOCUMENT tdosunix, filename
      or    res = tdosunix( filename )

     "Test DOS or UNIX"
  
     Count the number of occurrences of '10' and '13'
     both as singles and as pairs in a file.
     On the basis of that decide if file is in DOS style (res = 1)
     or in UNIX style (res = 0).
  
     Prints counts of 10 and 13 bytes (and pairs) when called as a
     subroutine. If called as a function 1 or 0 will be returned.
  
     Returns -1 and prints the counts if in doubt.
  
     SEE ALSO add13 and remove13.
 */
{
   i10 = 0;
   i13 = 0;
   i1310 = 0;
   i1013 = 0;
   p10 = 0;
   p13 = 0;

   if( !file_test(filename) ) error,"Did not find "+filename;
   f = open( filename, "rb" );

   v = array(char,1);

   address = 0;

   while( _read( f, address++, v ) == 1 ) {
      if( v(1) == 13 ) {
         i13++;
         if( p10 ) i1013++;
         p13 = 1;
      } else {
         if( v(1) != 10 ) p13 = 0;
      }
      if( v(1) == 10 ) {
         i10++;
         if( p13 ) i1310++;
         p10 = 1;
      } else {
         if( v(1) != 13 ) p10 = 0;
      }
   }
   close,fp;
   if( am_subroutine() ) {
      write,format="   In file %s:\n", filename;
      write,format="Found %i occurrences of LF (10)\n", i10;
      write,format="Found %i occurrences of CR (13)\n", i13;
      write,format="Found %i occurrences of pairs of CR+LF (13+10)\n", i1310;
      write,format="Found %i occurrences of pairs of LF+CR (10+13)\n", i1013;
   } else {
      if( i1310 == i10 ) return 1; // Definitely in DOS style
      if( i1310 == 0 ) return 0; // Definitely in UNIX style
      write,format="Found %i occurrences of LF (10)\n", i10;
      write,format="Found %i occurrences of CR (13)\n", i13;
      write,format="Found %i occurrences of pairs of CR+LF (13+10)\n", i1310;
      write,format="Found %i occurrences of pairs of LF+CR (10+13)\n", i1013;
      return -1; // Indeterminate
   }
}

/* Function is_link */

func is_link( filename )
/* DOCUMENT res = is_link( filename )

   Returns 1 if 'filename' is a symbolic link
   0 otherwise.

   This function is only useful for Linux/Unix operative systems.

   Uses the 'targetname' script.

   2012-04-30/NJW
*/
{
   OSTYPE = get_env("OSTYPE");
   if( strlen(OSTYPE) == 0 ) error,"IS_LINK is not appropriate for this OS";

   target = rdline(popen("targetname "+filename,0));
   ffilename = fullpath(filename);
   ftarget = fullpath(target);
   if( ftarget == ffilename ) return 0;
   else return 1;

}

/* Function reldif */

func reldif( x, y )
/* DOCUMENT r = reldif( x, y )

    Returns max( abs(x/y), abs(y/x) )

    So neither x nor y can be zero. Makes no sense if of opposite sign.

*/
{
   x = double(x);
   y = double(y);
   if( nallof(x) ) error,"Called reldif with x == 0";
   if( nallof(y) ) error,"Called reldif with y == 0";
   return max( abs(x/y), abs(y/x) );
}

/* Function gettimetag */

func gettimetag( filename, fmt= )
/* DOCUMENT res = gettimetag( filename, fmt= )

     Returns the modification time of the mentioned file.

     Keyword 'fmt':
            1 (default) : long,   number of seconds since Jan 1, 1970
            2           : string, Date string YYYY-MM-DD
            3           : string, Dattim string YYYY-MM-DDTHH:MM:SS

     Based on the shell function 'date'.

     2012-10-09/NJW
*/
{
   if( get_env("OSTYPE") != "linux" ) {
      write,"gettimetag is not implemented for your OS: "+get_env("OSTYPE");
      return [];
   }
   if( fmt == 2 ) return rdline(popen("date -r "+filename+" +%F",0));
   if( fmt == 3 ) return rdline(popen("date -r "+filename+" +%FT%T",0));
   return atoi(rdline(popen("date -r "+filename+" +%s",0)));
}

/* Function cronjob */

func cronjob( sys_string, interval, ntimes=, until=, pt=, y= )
/* DOCUMENT cronjob, sys_string, interval, ntimes=, until=, pt=, y=

 When keyword 'y' has not been set:
    Executes system,sys_string with a timeinterval of
    'interval' seconds. When keyword 'ntimes' is given then
    the command will be repeated this number of times.
    When keyword 'until' is given, then the command is repeated
    until this number of seconds have elapsed since the
    function was called. A second option for 'until' is
    to give a string like "100m" i.e. number followed by unit.
    "s", "m", "h", "d" (seconds, minutes, hours, days) are
    accepted.

 When keyword 'y' has been set:
    'sys_string' is then interpreted as a string array with yorick
    statements. They are saved to a temporary file and executed
    according to the instructions (see above).

    Keyword 'pt' (print time) will cause the time to be printed
                 before the system call.

    2012-11-12/NJW
*/
{
   local tstart, tnow;
   nti = 0;
   timestamp, tstart;
   if( until ) {
      if( typeof(until) == "string" ) {
         uni = strpart(until,0:0);
         until = atoi(strpart(until,1:-1));
         if( uni == "s" ) {
         } else if( uni == "m" ) {
            until *= 60;
         } else if( uni == "h" ) {
            until *= 3600;
         } else if( uni == "d" ) {
            until *= 86400;
         } else error,"Illegal unit";
      }
      tstop = tstart + until;
   }

   // Save yorick commands to temporary file

   tmp_file = ranstr(20)+".tmp";
   write_slist,tmp_file, sys_string;

   do {
      if( y ) {
         require, tmp_file;
      } else {
         if(pt) write,ndate(3);
         system, sys_string;
      }
      nti++;
      timestamp, tnow;
      pause, interval*1000; // convert to ms
      done = 0;
      if( ntimes ) done = nti > ntimes;
      if( until ) done = tnow >= tstop;
   } while( !done );

   if( y ) remove, tmp_file;
}

/* Function vsave */

func vsave( a, file )
/* DOCUMENT vsave, a, file

    Uses yorick's 'save' command to save variable 'a' to
    the binary file 'file' (under the name of 'a').

    SEE ALSO: vrest (for restoring)
*/
{
   stream = createb(file);
   save,stream,a;
   close,stream;
}

/* Function vrest */

func vrest( &a, file )
/* DOCUMENT vrest, >a, file

    Uses yorick's 'restore' command to restore variable 'a' from
    the binary file 'file' (under the name of 'a').

    SEE ALSO: vsave (for saving a variable)
*/
{
   stream = openb(file);
   restore,stream,a;
   close,stream;
}

/* Function plural */

func plural( n, isare=, waswere= )
/* DOCUMENT ending = plural( n, isare=, waswere= )

     returns "s" if n == 0 or abs(n) > 1, "" otherwise i.e. when abs(n) == 1

     When keyword 'isare' is set:
     Returns "is" when abs(n) == 1, "are" when n == 0 or abs(n) > 1

     When keyword 'waswere' is set:
     Returns "was" when abs(n) == 1, "were" when n == 0 or abs(n) > 1

*/
{
   if( isare ) {
      return abs(n) == 1 ? "is" : "are" ;
   } else if( waswere ) {
      return abs(n) == 1 ? "was" : "were" ;
   } else {
      return abs(n) == 1 ? "" : "s" ;
   }
}
/* Function yjour */

func yjour( void )
/* DOCUMENT yjour

     (takes no argument)

   Accepts text input and stores in file: yjour_YYMMDD.txt
   in the current directory. Has the advantage in contrast
   to 'xjour' that quotes are copied literally.

   Exclamation mark '!' as the first character acts as 'escape' to
   enter commands:
      Termination           !quit
      Edit journal file     !ed  or !vi
      Show journal file     !cat
      Help                  !help  or   !?

   2013-04-17/NJW
*/
{
   extern Yjour;

   Yjour = "yjour_"+ndate(1)+".txt";
   write,"Updating file: "+Yjour+" (in extern variable 'Yjour')";
   write,"Get help by entering '!?' or '!help'";
   write,"Terminate by entering '!quit'";

   text = "none"; // simply initialization

   while( text != "!quit" ) {
      text = rdline(prompt=" yjour> ");
      if( strpart(text,1:1) == "!" ) {

         // editing

         part = strpart(text,1:3);
         if( part == "!ed" || part == "!vi" ) system,"vi "+Yjour;

         // showing journal file

         part = strpart(text,1:4);
         if( part == "!cat" ) system,"cat "+Yjour;

         // showing help

         part = strpart(text,1:5);
         p2 = strpart(text,1:2);
         if( part == "!help" || p2 == "!?" ) {
            write,"!ed      Envoke editing the journalfile";
            write,"!vi      Envoke editing the journalfile";
            write,"!cat     List the journalfile to screen";
            write,"!help    Show this help menu";
            write,"!?       Show this help menu";
            write,"!quit    Exit yjour";
         }

      } else { // simply enter the input line
         lu = open(Yjour,"a");
         write,lu,format="%s\n", text;
         close,lu;
         //+ write,format="%s\n", text;
      }
   }
}

/* Function gen_log_init */

func gen_log_init( logfilename, logflag=, erase=, cont=, reset= )
/* DOCUMENT id = gen_log_init( logfilename, logflag=, erase=, cont=, reset= )

   "Generic log keeping"

   Sets up a file: 'logfilename' where input can be given with 'gen_log_entry'.
                   The file is closed after entry (suited for debugging purposes).
   The external variables 'Gen_logflag' and 'Gen_logfilename' are defined.

   If the logfile exists beforehand the user will be inquired writing is to be
   continued or starting over (erasing).

   Several logfiles can be used simultaneously by using the identification number
   that is returned from the call of this function.

   Returns the allocated idenfication number.

   Default name : gen_log_nnnn.txt
       where 'nnnn' is a serial number
   Keyword   stop   :  Stop logging to file
            logflag :  The higher the more output (default value is 5)
                       (sets Gen_logflag and gen_log_entry is only
                        operational if level <= Gen_logflag)
            erase   :  Will erase the log file if it exists already (skips user interaction)
             cont   :  Will continue writing into an existing file (skips user interaction)
            reset   :  Removes externals (Gen_logfilename, Gen_logflag, and Gen_id)
                       before moving on to generate the log file.

   SEE ALSO: gen_log_entry

*/
{

// Example 1: > cat_id = gen_log_init("catalog.log",logflag=5)
//            > gen_log_entry, 2, "First line of log", id=cat_id
//            > point_id = gen_log_init("pointing.log",logflag=4)
//            > gen_log_entry, 3, "First line of pointing log",id=point_id
//              .
//              .

   extern Gen_logflag, Gen_logfilename, Gen_id;

   if( is_void(logflag) ) logflag = 5; // default value

   // Start over if 'reset' has been given

   if(reset) Gen_logflag = Gen_logfilename = Gen_id = [];

   // Define the name of the log file if none is given

   if( is_void(logfilename) ) {
      logfilename = get_next_filename("gen_log_????.txt");
   }
   if( file_test(logfilename) ) {
      if( erase ) {
         remove,logfilename;
         if( !is_void(Gen_logfilename) ) { // Name already in list of log file names
            w = where( logfilename == Gen_logfilename );
            if( numberof(w) ) { // found, get the id number
               id = w(1);
               Gen_logflag(id) = logflag; // update the logflag
               return id;
            }
         }
      } else {
         if( !cont ) { // interact if question is not settled already
            ans = rdline(prompt="Log file exists already - continue anyway ? ");
            if( strlowcase(strpart(ans,1:1)) != "y" ) {
               write,"Quitting ..."; return [];
            } else write,"Continuing ...";
         } else { // a continuation has been asked for, but then the filename
                  // must exist in the Gen_logfilename array
            if( is_void(Gen_logfilename) ) {
               Gen_logfilename = logfilename;
               Gen_logflag = logflag;
               Gen_id = 1;
               //+ write,"No logfiles have been recorded!";
               //+ return [];
            }
            w = where( logfilename == Gen_logfilename );
            if( numberof(w) ) {
               id = w(1);
            } else {
               write,"The requested logfile is not in the list of existing logfiles.";
               return [];
            }
         }
      }
   }
   if( !cont ) { // only create a new logfilename if not a continuation
      if( !is_void(Gen_logfilename) ) {
         id = Gen_id(0)+1;
         grow, Gen_logfilename, logfilename;
         grow, Gen_logflag, logflag;
         grow, Gen_id, id;
      } else {
         Gen_logfilename = logfilename;
         Gen_logflag = logflag;
         Gen_id = id = 1; // first one will have id==1
      }
   }
   write,format="Logging in: %s\n", logfilename;
   gen_log_entry,1,"Logfile started "+ndate(3), \
        swrite(format="Logflag = %i", logflag), "",id=id;

   return id;

}

/* Function gen_log_entry */

func gen_log_entry( level, str1, .., id=, mir= )
/* DOCUMENT gen_log_entry, level, str1, .., id=, mir=

   "Generic log file entry"

   Writes str1 and following strings or string arrays
   to Gen_logfilename(id) in append mode if Gen_logflag(id) >= level

   Keyword
      id     The identifier as returned by 'gen_log_init'.
             Need not be given if there is only a single logfile
             defined in the current yorick sesseion or the
             first (id==1) logfile is to be updated.

     mir     (bool) If set, mirror to terminal.

   Externals:
       Gen_logfilename, Gen_logflag, Gen_id

   SEE ALSO: gen_log_init, gen_log_info
*/
{
   extern Gen_logflag, Gen_logfilename, Gen_id;

   // Only output to log file if Gen_logflag is true and
   // Gen_logflag >= level

   if( numberof(Gen_logflag) > 1 ) { // identify the proper one to use
      if( is_void(id) ) id = 1;
      w = where(id == Gen_id);
      nw = numberof(w);
      if( nw == 0 ) error,"##33## no logfiles with this id";
      if( nw > 1 ) error,"##34## more than one logfile with this id";
      i = w(1);
      logflag = Gen_logflag(i);
      logfilename = Gen_logfilename(i);
   } else {
      logflag = Gen_logflag;
      logfilename = Gen_logfilename;
   }

   if( !logflag ) return; // i.e. if logflag is zero
   if( logflag < level ) return; // if level is too high

   logstream = open( logfilename, "a" );
   write,logstream,format="%s\n", str1;
   if( mir ) write,format="%s\n", str1;
   while( more_args() ) {
      text = next_arg();
      for( i = 1; i <= numberof(text); i++ ) {
         write,logstream,format="   %s\n", text(i);
         if( mir ) write,format="   %s\n", text(i);
      }
   }
   close, logstream;
}

/* Function gen_log_info */

func gen_log_info
/* DOCUMENT gen_log_info (takes no argument)

    SEE ALSO: gen_log_init, gen_log_entry
*/
{
   extern Gen_logfilename, Gen_logflag, Gen_id

   if( is_void(Gen_logfilename) ) {
      write,"No generic log files have been defined.";
   } else {
      n = numberof(Gen_logfilename);
      write," id logflag  filename";
      for( i = 1; i <= n; i++ ) {
         write,format="%4i  %4i    %s\n", Gen_id(i), Gen_logflag(i), Gen_logfilename(i);
      }
   }
}

/* Function groupn */

func groupn( map, src, tol ) // src is 1D index vector
/* DOCUMENT group_numbers = groupn( map, src, tol )
     'map' is an array of numbers.
     'src' is a 1D index vector of selected pixels in the map.
              May be omitted, then all array elements are considered.
     'tol' is the maximal width of a group.

     Returns group numbers for the map(src) selection
       in an array of same size as 'src'.

   2013-05-28/NJW

*/
{
   if( is_void(src) ) src = indgen(numberof(map));

   nsrc = numberof(src);
   v = map(src);
   w = sort(v);
   vs = v(w);
   gn = array(int, nsrc);
   upper_limit = vs(1) + tol;
   group = 1;

   // For each element ask if it belongs to the current group.
   // If not, then increase the ceiling until the element
   // is below the ceiling. In this way no empty groups are produced.

   for( j = 1; j <= nsrc; j++ ) {
      if( vs(j) < upper_limit ) {
         gn(w(j)) = group;
      } else {
         while( upper_limit <= vs(j) ) upper_limit += tol;
         gn(w(j)) = ++group;
      }
   }
   return gn;
}

/* Function SiGn */

func SiGn( a, b )
/* DOCUMENT res = SiGn( a, b)

   Returns 'a' with the sign of 'b'

If 'b' is scalar:
     Result of:
       return b > 0.0 ? abs(a) : -abs(a);
If 'b' is array
     Returns array of 'a' according to the rule above
        applied to each element
 
*/
{
   if( is_vector(b) ) {
      res = array(structof(a),numberof(b));
      w = where( b >= 0.0 );
      m = where( b <  0.0 );
      if( numberof(w) ) res(w) = abs(a(w));
      if( numberof(m) ) res(m) = -abs(a(m));
      return res;
   } else return b > 0.0 ? abs(a) : -abs(a);
}

/* Function random_i */

func random_i( i1, i2, n )
/* DOCUMENT   res = random_i( i1, i2, n )
        or    res = random_i( i2, n )
        or    res = random_i( i2 )

    return 'n' random long integers between i1 and i2 (included).
    In the second form i1 is set to 1 (one). In the third form
    a single long integer between 1 and i2 (i2 must be >= 1) is returned.

    2012-12-07/NJW
*/
{
   if( is_void(i1) ) error,"At least one argument must be given";
   if( is_void(i2) ) { // Third form
      i1 = long(i1);
      if( i1 < 1 ) error,"Argument must be >= 1";
      i2 = i1;
      i1 = 1;
      ioff = 0;
   } else {
      if( is_void(n) ) { // Second form
         i2 = long(i2);
         i1 = long(i1);
         if( i1 < 1 ) error,"1. argument must be >= 1";
         n = i2;
         if( n < 1 ) error,"2. arguments must be >= 1";
         i2 = i1;
         i1 = 1;
         ioff = 0;
      } else { // First form
         i1 = long(i1);
         i2 = long(i2);
         n  = long(n);
         ioff = i1;
      }
   }
   if( !is_void(n) && n == 1 ) n = [];
   return long(random(n)*(i2 - i1 + 1) + i1 - ioff) + ioff;

}

/* Function diag_elems */

func diag_elems( n )
/* DOCUMENT idx = diag_elems( n )
            idx = diag_elems( matrix )

     Returns the 1D indices of the diagonal elements in
     a square matrix of dimension 'n'.
*/
{
   if( is_scalar(n) ) {
      return (n+1)*indgen(n) - n;
   } else {
      dms = dimsof(n);
      if( dms(1) != 2 || dms(2) != dms(3) ) error,"No square 2D matrix";
      m = dms(2);
      return (m+1)*indgen(m) - m;
   }
}

/* Function deci2hex */

func deci2hex( n )
/* DOCUMENT hexstring = deci2hex( n )

    Converts decimal number to hexadecimal string
*/
{
   c = ["0","1","2","3","4","5","6","7","8","9","a","b","c","d","e","f"];
   str = "";
   if( n == 0 ) return "0";
   if( n < 0 ) { str = "-"; n = -n; }

   i = n/16;
   r = n%16;
   while( i > 0 ) {
      str = c(r+1) + str;
      n = i;
      i = n/16;
      r = n%16;
   }
   str = c(r+1) + str;
   return str;
}

/* Function hex2deci */

func hex2deci( s )
/* DOCUMENT n = hex2deci( hexstring )

    Converts a hexadecimal string to a decimal number
*/
{
   c = ["0","1","2","3","4","5","6","7","8","9","a","b","c","d","e","f"];
   len = strlen(s);
   s = strlowcase(s);
   res = 0;
   mul = 1;
   sg = 1;
   for( i = len; i > 0; i-- ) {
      ch = strpart(s,i:i);
      if( ch == "-" ) {
         sg = -1;
      } else {
         j = where(ch == c)(1) - 1;
         res += j*mul;
         mul *= 16;
      }
   }

   return res*sg;
}

/* Function deci2octal */

func deci2octal( n )
/* DOCUMENT octalstring = deci2octal( n )

    Converts decimal number to octal string
*/
{
   c = ["0","1","2","3","4","5","6","7"];
   str = "";
   if( n == 0 ) return "0";
   if( n < 0 ) { str = "-"; n = -n; }

   i = n/8;
   r = n%8;
   while( i > 0 ) {
      str = c(r+1) + str;
      n = i;
      i = n/8;
      r = n%8;
   }
   str = c(r+1) + str;
   return str;
}

/* Function octal2deci */

func octal2deci( s )
/* DOCUMENT n = octal2deci( octalstring )

    Converts an octal string to a decimal number
*/
{
   c = ["0","1","2","3","4","5","6","7"];
   len = strlen(s);
   res = 0;
   mul = 1;
   sg = 1;
   for( i = len; i > 0; i-- ) {
      ch = strpart(s,i:i);
      if( ch == "-" ) {
         sg = -1;
      } else {
         j = where(ch == c)(1) - 1;
         res += j*mul;
         mul *= 8;
      }
   }

   return res*sg;
}

/* Function dwhere */

func dwhere( var, selstring )
/* DOCUMENT sel = dwhere( var, selstring )

    "Dynamical where" function

    The expression is put into 'selstring' and
    the variable name used must be 'var'.
    Only a single variable is allowed for the selection.

    Example code:

       func myfunc( a, b, c, selec= ) {
          d = a + b;
          sel = dwhere( d, selec );
          return c(sel);
       }

    that can be called as:
      > i = myfunc( a, b, c, selec="abs(var) < 17.34 & var > -10.")

   2013-07-01/NJW
*/
{
   name = "tmp_"+ranstr(20)+".tmp";
   //+ write," - debug: name = "+name;
   f = open(name, "w");
   write,f,format="func ffsel( %s ) {\n", "var";
   write,f,format="return where(%s);}\n", selstring;
   close,f;
   require, name;
   return ffsel( var );
}

/* Function rdbinfil */

func rdbinfil( option, filename, outfile= )
/* DOCUMENT rdbinfil, "[aisx]", filename, outfile=

     Shows any file byte by byte as regular ASCII text.

       Options: i    interactive mode meaning: show first 256 bytes and then wait
                a    give ASCII symbols whenever human readable
                x    give only hexadecimal symbols
                s    (silent) only output to file

  The origin dates back to ~1985
   2005-12-22/NJW
   2013-10-17/NJW translated into Yorick
*/
{

   if( is_void(filename) ) {
      write,"Missing argument! Filename is not given.";
      return;
   }
   if( is_void(option) ) {
      a_flag = x_flag = i_flag = s_flag = 0;
   } else {
      a_flag = (strpos( option, "a" ) > 0);
      x_flag = (strpos( option, "x" ) > 0);
      i_flag = (strpos( option, "i" ) > 0);
      s_flag = (strpos( option, "s" ) > 0);
   }
   //+ if( a_flag + x_flag + i_flag == 0 ) {
   //+    write,"Illegal option! One of 'i', 'a' and 'x' must be present.";
   //+    return;
   //+ }
   if( a_flag + x_flag == 2 ) {
      write,"Illegal option! 'a' and 'x' cannot both be present.";
      return;
   }

   if( !(fp = open( filename,"rb") ) ) {
      write,format="Can't open %s\n",filename;
      return;
   }
   // test for output file
   if( structof(outfile) == string ) {
      wf = 1;
      fh = open( outfile, "w" );
   } else wf = 0;
   // If 's' is given as an option and a filename for output ('outfile')
   // then dumping to the screen is prohibited.
   vb = !(s_flag && wf);

   c = char();
   address = 0;
   while( _read( fp, address, c ) == 1 ) {
      if( address%256 == 0 && address != 0 && i_flag ) {
         write,"";
         ans = rdline(prompt="Command: (x to terminate, +n to advance n blocks ) ... ");
         if( ans == "x" ) {
            close,fp;
            if(wf)close, fh;
            return;
         }
         if( strpart(ans,1:1) == "+" ) {
            nblocks = atoi(ans);
            for( jloop = 0; jloop < nblocks; ++jloop ) {
               for( iloop = 0; iloop < 256; ++iloop ) {
                  if( _read( fp, address, c ) != 1 ) {
                     write,format=,"No more data! nbytes = %i\n", address;
                     close,fp;
                     if(wf)close, fh;
                     return;
                  }
                  ++address;
               }
            }
         }
      }
      if( address%16 == 0 ) {
         if(vb)write,format="\n%7i >%8i:",address,address+15;
         if(wf)write,fh,format="\n%7i >%8i:",address,address+15;
      }
      if( a_flag ) {
         if( c >= 33 && c <= 126 ) {
            if(vb)write,format="   %c",c;
            if(wf)write,fh,format="   %c",c;
         } else {
            if(vb)write,format="%4d", c;
            if(wf)write,fh,format="%4d", c;
         }
      } else if( x_flag ) {
         if(vb)write,format="  %02x", c;
         if(wf)write,fh,format="  %02x", c;
      } else {
         if(vb)write,format="%4d", c;
         if(wf)write,fh,format="%4d", c;
      }
      address++;
   }
   write,"\n --------- End of file -------------\n";
   if(wf)close, fh;
   close,fp;

}

/* Function stch2num */

func stch2num( a, &astr, basechar )
/* DOCUMENT number = stch2num( a, >astr, basechar )

    Converts a string or a char to a number, and, as a side-effect,
    returns the string representation in 'astr'.

    Example:  > num = stch2num( "d", astr, 'a' )
              > print, num, astr
              4  "d"
              > num = stch2num( 7, astr, 'a' )
              > print, num, astr
              7  "g"

   2013-11-22/NJW
*/
{
   if( structof(basechar) != char ) error,"Bad argument: basechar";
   if( structof(a) == string ) {
      ia = int((*pointer(a))(1)-basechar+1);
   } else if( structof(a) == char ) {
      ia = int(a - basechar + 1);
   } else {
      ia = a;
   }
   c = array(char,2);
   c(1) = ia + basechar - 1;
   c(2) = '\0';
   astr = string(&c);
   return ia;
}

/* Function selection */

func selection( style_or_palette )
/* DOCUMENT selection, style_or_palette

    Displays the available:
       styles   if style_or_palette == "sty" or style_or_palette == 0
       palettes if style_or_palette == "pal" or style_or_palette == 1

   based on the value of 'GISTDIR' that is set in the 'custom.i' file.

   2014-03-12/NJW
*/
{
   if( structof( style_or_palette ) == string ) {
      if( strpart(style_or_palette,1:3) == "pal" ) {
         c = 1;
      } else if( strpart(style_or_palette,1:3) == "sty" ) {
         c = 0;
      } else error,"SELECTION: Invalid string argument.";
   } else {
      c = long(style_or_palette);
      if( c < 0 || c > 1 ) error,"SELECTION: Argument is outside range.";
   }
   if( c == 0 ) {
      list = basename(file_search("*.gs",GISTDIR));
      write,"The selection of styles:";
      prstrarr, list;
   } else if( c == 1 ) {
      list = basename(file_search("*.gp",GISTDIR));
      write,"The selection of palettes:";
      prstrarr, list;
   } else error,"SELECTION Should never come here";
}

/* Function wherenone */

func wherenone( arr, takeaway )
/* DOCUMENT sub_arr = wherenone( arr, takeaway )

    Returns an array with the elements of 'arr' that do not appear in
    'takeaway'

   Similar to 'filter_done' but faster for large arrays (however note, that
   the order of the arguments is opposite).

   SEE ALSO: filter_done, whereany

*/
{
   if( is_void(takeaway) ) return arr;
   arr = arr(sort(arr));
   arr = arr(uniq(arr));
   takeaway = takeaway(sort(takeaway));
   takeaway = takeaway(uniq(takeaway));
   narr = numberof(arr);
   ntake = numberof(takeaway);

   i = j = 1;
   flag = array(1,narr);
   while( i <= narr && j <= ntake ) {
     if( arr(i) > takeaway(j) ) {
        j++;
     } else if( arr(i) < takeaway(j) ) {
        i++;
     } else {
        flag(i) = 0;
        i++; j++;
     }
   }
   return arr(where(flag));
}

/* Function filter_done1 */

func filter_done1( done, list )
/* DOCUMENT res = filter_done1( done, list )

     Returns an array 'res' with all elements of 'list' that
     are not present in 'done'.
     Both 'list' and 'done' must be arrays of the same type.

      SEE ALSO: filter_done, wherenone

     2014-07-22/NJW
*/
{
   done = done(sort(done));
   done = done(uniq(done));
   list = list(sort(list));
   list = list(uniq(list));
   nlist = numberof(list);
   ndone = numberof(done);
   res = list;

   k = 1;
   j = 1;
   for( i = 1; i <= nlist; i++ ) {
      if( list(i) < done(j) ) { // list(i) is not in done
         res(k++) = list(i);
         continue; // test the next 'i'
      }
      if( list(i) == done(j) ) { // list(i) is found in done
         if( ++j > ndone ) { // all testing is over, add remaining
                             // list to result except this list(i)
            if( i < nlist ) {
               res(k:k+nlist-i-1) = list(i+1:nlist);
               k += nlist-i;
            }
            break;
         }
         continue; // test the next 'i'
      }
      if( list(i) > done(j) ) { // list(i) might still be found in done
         if( j == ndone ) { // all testing is over, add remaining
                            // list to result
            res(k:k+nlist-i) =  list(i:nlist);
            k += nlist-i+1; // one higher than last one to be filled
            break;
         }
         while( j < ndone && done(j) < list(i) ) j++;
         i--; // test this one again
      }
   }
   return k == 1 ? [] : res(1:k-1);
}

/* Function uniq2d */

func uniq2d( &x, &y, eps )
/* DOCUMENT uniq2d, >x, >y, eps

     Returns pairs (x,y) that are not duplicated with a
     precision of 'eps'. 'x' and 'y' are also input.
     The order of the resulting pairs usually differs from the input.

*/
{
   n = numberof(x);
   xo = yo = [];
   is = msort(x,y);
   x = x(is);
   y = y(is);

   for( l = 1; l < n; l++ ) {
      if( !near( x(l), x(l+1), eps ) || !near( y(l), y(l+1), eps ) ) {
         grow, xo, x(l);
         grow, yo, y(l);
      }
   }
   // check last value
   if( !near(x(n-1),x(n), eps ) || !near(y(n-1),y(n), eps) ) {
      grow, xo, x(n);
      grow, yo, y(n);
   }
   x = xo;
   y = yo;
}

/* Function gt */

func gt( x, a )
/* DOCUMENT xp = gt( x, a )

     Works as IDL '>' :
        Returns 'xp' as x, but with 'a' where x <= a

    SEE ALSO: lt
*/
{
   xp = x;
   w = where( x <= a );
   if( numberof(w) ) xp(w) = a;
   return xp;
}

/* Function lt */

func lt( x, a )
/* DOCUMENT xp = lt( x, a )

     Works as IDL '<' :
        Returns 'xp' as x, but with 'a' where x >= a

    SEE ALSO: gt
*/
{
   xp = x;
   w = where( x >= a );
   if( numberof(w) ) xp(w) = a;
   return xp;
}

/* Function sexa2deci */

func sexa2deci( rah, ram, ras, decd, decm, decs )
/* DOCUMENT radec = sexa2deci( rah, ram, ras, decd, decm, decs )

*/
{
   if( is_void(rah) ) rah = 0.0;
   if( is_void(ram) ) ram = 0.0;
   if( is_void(ras) ) ras = 0.0;
   if( is_void(decd) ) decd = 0.0;
   if( is_void(decm) ) decm = 0.0;
   if( is_void(decs) ) decs = 0.0;

   return [15.*ten(rah, ram, ras), ten(decd, decm, decs)];

}

/* Function wstyles */

func wstyles( void )
/* DOCUMENT wstyles
            list = wstyles()

    Version one, called as a subroutine:
    Report the available window style files (*.gs) from the directory
    '/r9/njw/yorick/yorick-2.1.06/yorick-2.1/g'

    Version two, called as a function:
    Return the list of available window style files.

    2018-05-07/NJW
*/
{
   dir = "/r9/njw/yorick/yorick-2.1.06/yorick-2.1/g";
   write,"Searching directory: "+dir;
   list = basename(file_search("*.gs", dir ));
   if( am_subroutine() ) {
      arrange_words_in_columns, list, 3, 30;
   } else return list;
}

/* Function is_linux */

func is_linux( dum )
/* DOCUMENT  r = is_linux()

    Determines whether we are running on a Linux type system
    (r == 1) or e.g. a Windows type (r == 0)

*/
{
   if( strlen(get_env("OSTYPE")) == 0 ) { // Windows OS
      return 0;
   } else return 1;
}
