/***************************************
   Library for FFT convolution

   fconvol
   fkernel
   gfconvol

   2007-02-06/NJW
****************************************/


/* Function fkernel */

func fkernel( kernel, dim1, dim2)
/* DOCUMENT fk = fkernel( kernel, dim1, dim2 )

   Returns the FFT transform of the kernel in dimensions
   dim1 and dim2 that must be the same as the array to
   convolve with the kernel

   2005-03-30/NJW
*/

{
   require, "idlx.i";

   dms = dimsof(kernel);
   m1 = (dms(2)-1)/2;
   m2 = (dms(3)-1)/2;
   ks = array(double,dim1,dim2);
   n1 = (dim1 + 1)/2;
   n2 = (dim2 + 1)/2;
   ks(n1-m1:n1+m1, n2-m2:n2+m2) = kernel;
   ks = shift(ks,-n1,-n2);
   return fft(ks,1);
}

/* Function fconvol */

func fconvol( arr, fkern, dofft= )
/* DOCUMENT res = fconvol( arr, fkern, dofft= )

   Arguments:
        arr  :  An array to be convolved
      fkern  :  The convolution kernel, either already FFT'ed
                to same dimension as 'arr' or a 2D array to act as kernel.
                The FFT will happen here.

   Keyword:   dofft : (bool) to be set if 'fkern' has not been FFT'ed.

   Returns the convolution of double array 'arr'

   Remark: Works best with odd dimensions both for array and kernel.

   Original version: 2005-03-30/NJW, updated 2015-07-07/NJW
*/

{
   dms = dimsof(arr);
   if( dofft ) {
      fkern = fkernel( fkern, dms(2), dms(3) );
   } else {
      dmsk = dimsof(fkern);
      if( dms(2) != dmsk(2) || dms(3) != dmsk(3) ) {
         print,"fconvol: mismatching array dimensions";
         return [];
      }
   }
   norm = (1.0*dms(2))*dms(3);
   farr = fft(arr,1);
   p = fft(farr*fkern,-1);
   return p.re / norm;
}

/* Function gfconvol */

func gfconvol( image, sigma, widen= )
/* DOCUMENT folded_image = gfconvol( image, sigma, widen= )

   returns image folded with Gauss kernel with sigma 'sigma'
   in unit of pixels.

   Keyword 'widen' will enlarge image with this amount
                   before convolution to avoid spilling
                   overflow

   2007-02-06/NJW
   2009-11-04/NJW added keyword 'widen' and functionality
*/
{
   require, "image.i";

   wimage = image;
   dms = dimsof(wimage);
   nx = dms(2);
   ny = dms(3);
   if( dms(2)%2 == 0 ) {
      // repeat last row to make dimension odd
      we = array(structof(image),dms(2)+1,dms(3));
      we(1:dms(2),) = wimage;
      we(0,) = wimage(0,);
      wimage = we;
   }
   dms = dimsof(wimage);
   if( dms(3)%2 == 0 ) {
      // repeat last column to make dimension odd
      we = array(structof(image),dms(2),dms(3)+1);
      we(,1:dms(3)) = wimage;
      we(,0) = wimage(,0);
      wimage = we;
   }
   if( widen ) wimage = widen_image( wimage, widen );

   kdim = int(5.*sigma)*2 + 1;
   ker = gkernel(kdim,kdim,sigma);

   dms = dimsof(wimage);
   fker = fkernel( ker, dms(2), dms(3) );
   res = fconvol( wimage, fker );
   if( widen ) res = res(widen+1:-widen,widen+1:-widen);
   return res(1:nx,1:ny);

}
