/************************************************************
 *
 *  A visual or graphical selection of a file
 *
 *  Package with functions:
 *        ia_file_selection
 *        directory_selection
 *        file_selection
 *
 **************************************************************/

/* Function ia_file_selection */

func ia_file_selection( start_dir, &res_dir, sel=, prompt= )
/* DOCUMENT file = ia_file_selection( start_dir, >res_dir, sel=, prompt= )

    Argument: start_dir tells where to start the search
                if omitted the current directory will be used.
              res_dir is returned as the last directory searched.
    Keywords:
           sel     is used as an argument for 'strglob' to present
                   only a selection of files.
        prompt     Text on screen

    2011-12-16/NJW, updated 2017-05-24/NJW
*/
{
   if( structof(prompt) == string ) write, prompt;

   cwind = window();
   cwd = get_cwd();
   if( !is_void(start_dir) ) cd, start_dir;

   do {
      status = directory_selection(sel=sel);
      if( status ) {
         window,cwind;
         return [];
      }
      res_dir = get_cwd();
      file = file_selection(sel=sel);
   } while ( structof(file) == int );

   cd,cwd;
   write,"Now back in "+get_cwd();
   window,cwind;
   write,"  - and using window "+itoa(cwind);
   return file;
}

func directory_selection( sel= )
{
   local subdirs;

   // List directories

   window,7,style="nobox.gs";
   files = lsdir(".",subdirs);
   nfiles = numberof(files);
   if( nfiles ) files = files(sort(files));
   cwd = get_cwd();
   if( cwd != "/" ) grow,subdirs,"..";

   nsubdirs = numberof(subdirs);

   if( nsubdirs ) {
      subdirs = subdirs(sort(subdirs));
      nbatches = (nsubdirs-1)/60 + 1;
      ibatch = 1;

      showbatch:

      //+ write,"Showing batch#"+itoa(ibatch)+" of "+itoa(nbatches);
      nbstart = (ibatch - 1)*60 + 1;
      nbstop  = ibatch == nbatches ? nsubdirs : ibatch*60;

      plot,[0,0,1,1,0],[0,1,1,0,0];
      xyouts,0.16,0.894,esc_underscore(get_cwd()),charsize=0.7,font="schoolbook",ndc=1;
      y = 0.95;
      x = 0.1;
      xarr = yarr = [];
      for( i = nbstart; i <= nbstop; i++ ) {
         oplot,[x-0.05],[y+0.010],ps=19,symsize=0.35,color="blue";
         xyouts,x,y,esc_underscore(subdirs(i)),charsize=0.7,font="schoolbook",color="blue";
         grow, xarr, x - 0.05;
         grow, yarr, y + 0.01;
         y -= 0.031;
         if( y < 0.02 ) {
            x = 0.6;
            y = 0.95;
         }
      }
      // Draw box for shift to plain files but only if there are some
      if( nfiles ) {
         xyouts,0.5,0.015,"Go to files",charsize=0.8,font="schoolbook",color="green",align=0.5;
         oplot,0.3*([0,0,1,1,0]-0.5)+0.5,0.04*([0,1,1,0,0]-0.5)+0.025,color="green";
         grow, xarr, 0.5;
         grow, yarr, 0.025;
      } else {
         grow, xarr, 10.; // make values that can never be selected
         grow, yarr, 10.;
      }
      // When more directories exist then show arrow
      if( ibatch < nbatches ) {
         oplot,[0.98],[0.02],ps=21,fill=1,symsize=0.8,color="red";
         grow, xarr, 0.98;
         grow, yarr, 0.02;
         flagforw = 1;
      } else flagforw = 0;
      // When less directories exist then show arrow
      if( ibatch > 1 ) {
         oplot,[0.02],[0.98],ps=20,fill=1,symsize=0.8,color="red";
         grow, xarr, 0.02;
         grow, yarr, 0.98;
         flagbackw = 1;
      } else flagbackw = 0;

      r = curmark1(prompt="Select the directory ...",nomark=1);
      d = (xarr - r(1))^2 + (yarr - r(2))^2;
      idx = d(mnx);
      //+ write,"Min d = "+ftoa(d(idx),ndec=5);
      if( d(idx) > 0.02 ) return 1; // stop when too far from selection
      if( idx < numberof(xarr)-flagforw-flagbackw ) {
         //+ write,"You chose a subdirectory: "+subdirs(idx+nbstart-1)+", confirm ...";
         oplot,[xarr(idx)],[yarr(idx)],ps=19,symsize=0.35,color="red";
         r = curmark1(prompt="Confirm the selection ...",nomark=1);
         d = (xarr - r(1))^2 + (yarr - r(2))^2;
         idx2 = d(mnx);
         //+ write,"Min d = "+ftoa(d(idx2),ndec=5);
         if( d(idx2) > 0.02 ) return 1; // stop when too far from selection
         if( idx == idx2 ) {
            //+ write,"Yes, confirmed!";
            cd, subdirs(idx+nbstart-1);
         } //+ else write,"Try again!";
         status = directory_selection(sel=sel);
      } else {
         if( idx == numberof(xarr) - flagforw - flagbackw ) {
            //+ write,"You'll move to file selection.";
         } else {
            if( idx == numberof(xarr) - flagbackw ) {
               // move forwards
               //+ write," - forward move";
               ibatch++;
               goto showbatch;
            } else {
               // move backwards
               //+ write," - backward move";
               ibatch--;
               goto showbatch;
            }
         }
      }
   }
   return 0; // flag status OK at return
}


/* Function file_selection */


func file_selection( sel= )
{
   extern Previous_selections;

   local subdirs;

   // List files

   window,7,style="nobox.gs";
   files = lsdir(".",subdirs);
   if( !is_void(sel) ) {
      w = where(strglob(sel,files));
      files = numberof(w) ? files(w) : [];
   }

   nfiles = numberof(files);

   if( nfiles ) {
      files = files(sort(files));
      nbatches = (nfiles-1)/60 + 1;
      ibatch = 1;

      showbatch:

      //+ write,"Showing batch#"+itoa(ibatch)+" of "+itoa(nbatches);
      nbstart = (ibatch - 1)*60 + 1;
      nbstop  = ibatch == nbatches ? nfiles : ibatch*60;

      plot,[0,0,1,1,0],[0,1,1,0,0];
      xyouts,0.16,0.894,esc_underscore(get_cwd()),charsize=0.7,font="schoolbook",ndc=1;
      y = 0.95;
      x = 0.1;
      xarr = yarr = [];
      for( i = nbstart; i <= nbstop; i++ ) {
         // Check for previous selections
         if( numberof(Previous_selections) ) {
            if( numberof(where(Previous_selections == fullpath(files(i)))) ) {
               oplot,[x-0.05],[y+0.010],ps=19,symsize=0.3,color="yellow",fill=1;
            }
         }
         oplot,[x-0.05],[y+0.010],ps=19,symsize=0.35,color="green";
         xyouts,x,y,esc_underscore(files(i)),charsize=0.7,font="schoolbook",color="green";
         grow, xarr, x - 0.05;
         grow, yarr, y + 0.01;
         y -= 0.031;
         if( y < 0.02 ) {
            x = 0.6;
            y = 0.95;
         }
      }

      // Draw box for shifting back to directory
      xyouts,0.5,0.015,"Go back to directory",charsize=0.8,font="schoolbook",color="blue",align=0.5;
      oplot,0.3*([0,0,1,1,0]-0.5)+0.5,0.04*([0,1,1,0,0]-0.5)+0.025,color="blue";
      grow, xarr, 0.5;
      grow, yarr, 0.025;

      // When more files exist then show arrow
      if( ibatch < nbatches ) {
         oplot,[0.98],[0.02],ps=21,fill=1,symsize=0.8,color="red";
         grow, xarr, 0.98;
         grow, yarr, 0.02;
         flagforw = 1;
      } else flagforw = 0;
      // When less directories exist then show arrow
      if( ibatch > 1 ) {
         oplot,[0.02],[0.98],ps=20,fill=1,symsize=0.8,color="red";
         grow, xarr, 0.02;
         grow, yarr, 0.98;
         flagbackw = 1;
      } else flagbackw = 0;

      r = curmark1(prompt="Select a file ...",nomark=1);
      d = (xarr - r(1))^2 + (yarr - r(2))^2;
      idx = d(mnx);
      //+ write,"Min d = "+ftoa(d(idx),ndec=5);
      if( d(idx) > 0.02 ) return []; // stop when too far from selection
      if( idx == numberof(xarr)-flagforw-flagbackw ) return int(1);
      if( idx < numberof(xarr)-flagforw-flagbackw ) {
         //+ write,"You chose a file: "+files(idx+nbstart-1)+", confirm ...";
         oplot,[xarr(idx)],[yarr(idx)],ps=19,symsize=0.35,color="red";
         r = curmark1(prompt="Confirm selection ...",nomark=1);
         d = (xarr - r(1))^2 + (yarr - r(2))^2;
         idx2 = d(mnx);
         //+ write,"Min d = "+ftoa(d(idx2),ndec=5);
         if( d(idx2) > 0.02 ) return []; // stop when too far from selection
         if( idx == idx2 ) {
            //+ write,"Yes, confirmed!";
            grow, Previous_selections, fullpath(files(idx+nbstart-1));
            return fullpath(files(idx+nbstart-1));
         } else {
            //+ write,"Try again:";
            return file_selection(sel=sel);
         }
      } else {
         if( idx == numberof(xarr) - flagforw - flagbackw ) {
            //+ write,"You'll move to file selection.";
         } else {
            if( idx == numberof(xarr) - flagbackw ) {
               // move forwards
               //+ write," - forward move";
               ibatch++;
               goto showbatch;
            } else {
               // move backwards
               //+ write," - backward move";
               ibatch--;
               goto showbatch;
            }
         }
      }
   }
}

ifs = ia_file_selection;
write,"Shorthand: ifs for ia_file_selection.";
