extern idlxdoc;
/* DOCUMENT idlx package
 *
 *  Additional IDL type functions
 *
 *  Coming: file_test where type (dir or file) can be specified
 *          Also return value will indicate if gzipped version of
 *          file exists.
 *
 *  2004-08-23/Niels J. Westergaard
 *  2008-05-16/NJW, updated with get_boundaries
 *
       _arr_integral       fold_gaussx         rm_slashcom         
       _is_digit           fold_gaussy         set_par             
       app_slash           get_boundaries      shd_upd             
       arcdist             get_next_filename   shift               
       areapolygon         get_par             sigdig              
       areatriangle        histo               str2arr             
       arr_accum           histos              str_erase_between_symbols
       arr_fill_in         hpd                 str_get_words       
       arr_info            icoef_polyn         strcharrepl         
       arr_integral        insidepolygon       strcompare          
       array_gauss         interp2             strcompress         
       array_igauss        is_digit            strdelcom           
       arrcat              is_fnumber          strinsert           
       atof                is_number           strip_curly_br      
       atoi                itoa                strjoin             
       bin_overlap         linscale            strlowcase          
       cent2bds            local_rms           strpadd             
       centroid            maxarr              strpos              
       chi2                message_board       strput              
       circ_area1          monincr             strsplit            
       datagroups          most_freq_elem      strstrrepl          
       dcoef_polyn         ndate               strupcase           
       ddpolyn             outlier             systime             
       determinant         parpath             toint               
       distfix             pick_swid_str       tolong              
       dpolyn              polyn               wavg                
       draw_from_dist      put_box             write_slist         
       esc_underscore      qmcompare           wrms                
       extract_box         qmstrpos1           wvar                
       f2scienota          qmstrpos2           xjour               
       file_rsearch        ranstr              zero2pi             
       file_search         read_slist          zero360             
       file_test           rem_slash           
*/

/* Function arr_integral */

func arr_integral( x, y, a, b )
/* DOCUMENT arr_integral( x, y, a, b )

   Integrate the array 'y' : sum(y*dx) over the interval
   where a < x < b

*/
{
   nx = numberof(x);
   ny = numberof(y);
   if( nx < 1 ) error,"ARR_INTEGRAL ##1## too few array elements";
   if( nx != ny ) error,"ARR_INTEGRAL ##2## mismatching array sizes";
/*
 * Use the array elements with a < x < b (index range w)
 *  and expand arrays : x -> [a, x(w), b]
 *                      y -> [ya, y(w), yb]
 *  where ya and yb are adequately interpolated values
 */

   if( is_void(a) ) a = x(1);
   if( is_void(b) ) b = x(0);
   a = double(a);
   b = double(b);

   w = where( x >= a & x <= b );
   nw = numberof(w);
   yy = interp( y, x, a);
   yb = interp( y, x, b);
   xx = a;

   if( nw ) {

      grow, xx, x(w), b;
      grow, yy, y(w), yb;

   } else {

      grow, xx, b;
      grow, yy, yb;

   }

   return _arr_integral( xx, yy );

}

/* Function _arr_integral */

func _arr_integral( x, y )
/* DOCUMENT _arr_integral( x, y )

   Integrate the array 'y' : sum(y*dx) over all elements
   x limits are x(1) and x(0)

*/
{
   nx = numberof(x);
   ny = numberof(y);
   if( nx < 1 ) error,"_ARR_INTEGRAL ##1## too few array elements";
   if( nx != ny ) error,"_ARR_INTEGRAL ##2## mismatching array sizes";

   d = double(x(2:0) - x(1:-1));
   z = 0.5*(y(1:-1) + y(2:0));
   return sum(d*z);
}

/* Function get_boundaries */

func get_boundaries( x_arr, &x1_arr, &x2_arr, lg= )
/* DOCUMENT get_boundaries, x_arr, >x1_arr, >x2_arr, lg=

   Assume x_arr has increasing values,
   Keyword lg: If set produce boundaries in log(x_arr)

   x1_arr(i) < x_arr(i) < x2_arr(i)

*/
{
   if( lg ) {
      lx_arr = log(x_arr);
      d = lx_arr(dif);
      xcen = exp(lx_arr(1:-1) + 0.5*d);
      x1_arr = exp(lx_arr(1) - 0.5*d(1)); grow, x1_arr, xcen;
      x2_arr = xcen; grow, x2_arr, exp(lx_arr(0) + 0.5*d(0));
   } else {
      d = x_arr(dif);
      xcen = x_arr(1:-1) + 0.5*d;
      x1_arr = x_arr(1) - 0.5*d(1); grow, x1_arr, xcen;
      x2_arr = xcen; grow, x2_arr, x_arr(0) + 0.5*d(0);
   }
}

/* Function arr_accum */

func arr_accum( arr, norm= )
/* DOCUMENT accumulated_array = arr_accum( arr, norm= )

   Accumulate value of array 'arr'. Keyword 'norm' for
   normalization to 1. (largest value).

   2007-05-03/NJW
   2010-04-16/NJW Updated to use 'psum'
*/
{
   new_arr = arr(psum);
   if( norm ) {
      new_arr = double(new_arr)/new_arr(max);
   }

   return new_arr;

}

/* Function histos */

func histos( y, &h, &x, bmin=, bmax=, binsize= )
/* DOCUMENT histos, y, >h, >x, bmin=, bmax=, binsize=

   returns the histogram of the y (array) values in 'h'
   and the bin centers in 'x'. Optional keywords for
   bin-min, bin-max, and bin-size.

   SEE ALSO: histo - extended functionality but somewhat slower

   2007-04-25/NJW updated with test for number of bins

*/

{

   n_y = numberof(y);
   ymin = min(y);
   ymax = max(y);
   binmin = is_void(bmin) ? ymin - 0.01*(ymax - ymin) : bmin;
   binmax = is_void(bmax) ? ymax + 0.01*(ymax - ymin) : bmax;
   binsiz = is_void(binsize) ? 1.0 : binsize;
   nbins = long( (binmax - binmin) / binsiz - 1.e-5 ) + 1;
   if( nbins <= 1 ) {
      write,"HISTOS error: Number of bins less than 2!";
      return;
   }
   spanx = nbins * binsiz;
   overshoot = spanx - (binmax - binmin);
   binmin -= (overshoot/2);
   binmax += (overshoot/2);
   boundaries = span(binmin, binmax, nbins+1);
   x = span(binmin+0.5*binsiz, binmax-0.5*binsiz, nbins );
   h = array(double,nbins);
   idx = long((y(where(y>=binmin))-binmin)/binsiz) + 1;
   w = where( idx <= nbins );
   idx = idx(w);
   n_idx = numberof(idx);
   for( i = 1; i<= n_idx; i++ ) {
      h(idx(i))++;
   }
}

/* Function histo */

func histo( y, &h, &x, bmin=, bmax=, binsize=, bds=, weight= )
/* DOCUMENT histo, y, >h, >x, bmin=, bmax=, binsize=, bds=, weight=

   returns the histogram of the y (array) values in 'h'
   and the bin centers in 'x'. Optional keywords for
   bin-min, bin-max, and bin-size.
   Keyword 'bds' gives bin boundaries as an array. The
        number of bins is one less than the number of elements
        in this array.

   Keyword 'weight' can be an array conformable with 'y'.

   SEE ALSO: histos - faster but does not have the bds keyword

*/

{

   n_y = numberof(y);
   if( numberof(weight) == 0 ) {
      weight = array(1.,n_y);
   } else {
      if( numberof(weight) != n_y ) error,"weight not conformable with y";
   }

   if( is_void(bds) ) {

      ymin = min(y);
      ymax = max(y);
      binmin = is_void(bmin) ? ymin - 0.01*(ymax - ymin) : bmin;
      binmax = is_void(bmax) ? ymax + 0.01*(ymax - ymin) : bmax;
      binsiz = is_void(binsize) ? 1.0 : binsize;
      nbins = long( (binmax - binmin) / binsiz - 1.e-5 ) + 1;
      if( nbins <= 1 ) {
         write,"HISTO error: Number of bins less than 2!";
         return;
      }
      spanx = nbins * binsiz;
      overshoot = spanx - (binmax - binmin);
      binmin -= (overshoot/2);
      binmax += (overshoot/2);
      boundaries = span(binmin, binmax, nbins+1);
      x = span(binmin+0.5*binsiz, binmax-0.5*binsiz, nbins );

   } else {

      binmin = bds(1);
      binmax = bds(0);
      nbins = numberof(bds)-1;
      if( nbins <= 1 ) {
         write,"HISTO error: Number of bins less than 2!";
         return;
      }
      boundaries = bds;
      x = 0.5*(bds(2:0) + bds(1:-1));

   }
   h = array(double,nbins);
   for( i = 1; i<= n_y; i++ ) {
      w = where( boundaries < y(i) );
      if( numberof(w) > 0 ) if( w(0) <= nbins ) h(w(0)) += weight(i);
   }
}

/* Function histod */

func histod( y, &h, &x, bmin=, bmax=, binsize=, bds= )
/* DOCUMENT histod, y, >h, >x, bmin=, bmax=, binsize=, bds=

   returns the histogram of the y (array) values in 'h'
   and the bin centers in 'x'. Optional keywords for
   bin-min, bin-max, and bin-size.
   Keyword 'bds' gives bin boundaries as an array. The
        number of bins is one less than the number of elements
        in this array.

   Based on yorick functions 'digitize' and 'histogram'

   SEE ALSO: histos - faster but does not have the bds keyword
             histo  - earlier version of this function, has additional
                         'weight' keyword

*/

{

   n_y = numberof(y);

   if( is_void(bds) ) {

      ymin = min(y);
      ymax = max(y);
      binmin = is_void(bmin) ? ymin - 0.01*(ymax - ymin) : bmin;
      binmax = is_void(bmax) ? ymax + 0.01*(ymax - ymin) : bmax;
      binsiz = is_void(binsize) ? 1.0 : binsize;
      nbins = long( (binmax - binmin) / binsiz - 1.e-5 ) + 1;
      if( nbins <= 1 ) {
         write,"HISTOD error: Number of bins less than 2!";
         return;
      }
      spanx = nbins * binsiz;
      overshoot = spanx - (binmax - binmin);
      binmin -= (overshoot/2);
      binmax += (overshoot/2);
      boundaries = span(binmin, binmax, nbins+1);
      x = span(binmin+0.5*binsiz, binmax-0.5*binsiz, nbins );

   } else {

      binmin = bds(1);
      binmax = bds(0);
      nbins = numberof(bds)-1;
      if( nbins <= 1 ) {
         write,"HISTOD error: Number of bins less than 2!";
         return;
      }
      boundaries = bds;
      x = 0.5*(bds(2:0) + bds(1:-1));

   }
   h = array(double,nbins);
   //+ for( i = 1; i<= n_y; i++ ) {
   //+    w = where( boundaries < y(i) );
   //+    if( numberof(w) > 0 ) if( w(0) <= nbins ) h(w(0)) += weight(i);
   //+ }
   q = digitize( y, boundaries );
   // the first element contains the count of all y-values below bds(1)
   h = histogram(q)(2:0);
   //  the last element contains the count of all y-values above bds(0)
   //  however, it is only returned by 'histogram' if there are any
   if( numberof(h) > nbins ) h = h(1:-1);
}

/* Function maxarr */

func maxarr( arr, &maxval, &idxvec )
/* DOCUMENT maxarr, arr, >maxval, >idxvector

   returns the maximum of an array as well as the
   vector with indices.
*/
{
   maxval = max(arr);
   idx = where(arr == maxval)(1);
   idxvec = indices(arr,idx);
}

/* Function strcharrepl */

func strcharrepl( str, oldchar, newchar, all= )
/* DOCUMENT res = strcharrepl( str, oldchar, newchar, all= )

   Replaces every occurrence of character or string 'oldchar' in
   string or string array 'str' with character or string 'newchar'.
   Default operation will eliminate leading, trailing, and
   duplicated occurrences.

   The keyword 'all' will ensure that ALL occurrences are replaced.

*/
{
   n = numberof(str);
   if( n == 0 ) return [];
   if( n == 1 ) {
      tok = strsplit( str, oldchar, nocompress=all );
      return strjoin( tok, newchar );
   } else {
      res = array(string,n);
      for(i=1;i<=n;i++) {
         tok = strsplit( str(i), oldchar, nocompress=all );
         res(i) = strjoin( tok, newchar );
      }
      return res;
   }
}

/* Function strpadd */

func strpadd( str, length, paddchar, rj=, truncate= )
/* DOCUMENT res = strpadd( str, length, paddchar, rj=, truncate= )

   Return string 'str' (may be scalar or array) padded to length 'length' with
      character 'paddchar' (may be given as a string or a char).
   It not given 'paddchar' defaults to space.
   If keyword 'rj' is set then it is right justified (padding to the left).
   If keyword 'truncate' is set then the string is truncated if its length
       exceeds 'length'
*/
{
   if( is_void(paddchar) ) paddchar = ' ';

   if( is_scalar(str) ) {
      return _strpadd( str, length, paddchar,rj=rj,truncate=truncate);
   } else {
      n = numberof(str);
      res = array(string,n);
      for(i=1;i<=n;i++) res(i) = _strpadd( str(i), length, paddchar,rj=rj,truncate=truncate);
      return res;
   }
}

/* Function _strpadd */

func _strpadd( str, length, paddchar, rj=, truncate= )
/* DOCUMENT res = _strpadd( str, length, paddchar, rj=, truncate= )

   Work function for 'strpadd'.

   Return string 'str' (must be scalar) padded to length 'length' with
      character 'paddchar' (may be given as a string or a char).
   If keyword 'rj' is set then it is right justified (padding to the left).
   If keyword 'truncate' is set then the string is truncated if its length
       exceeds 'length'
*/
{
   // change to char representation if required (only first character)
   if( structof(paddchar) == string ) paddchar = (*pointer(paddchar))(1);
   if( numberof(paddchar) != 1 ) error,"STRPADD error, illegal paddchar";
   cstr = *pointer(str);
   len = numberof(cstr);
   if(len>1){
      cstr = cstr(1:-1); // get rid of trailing end-character
   }else{
      cstr = []; // if empty string input then remove completely
   }
   len--;
   if( length == len ) return str; // a perfect match
   if( length < len ) {
      if( truncate ) {
         l = len - length; // number of characters to remove
         if( rj ) {
            return strpart(str,l+1:0);
         } else {
            return strpart(str,1:length);
         }
      } else {
         return str;
      }
   }
   a = array(paddchar,length-len);
   res = [];
   if( rj ) {
      grow, res, a, cstr, '\0';
   } else {
      if( is_void(cstr) ) {
         grow, res, a, '\0';
      } else {
         grow, res, cstr, a, '\0';
      }
   }

   return string(&res);

}

/* Function strstrrepl */

func strstrrepl( str, oldstr, newstr )
/* DOCUMENT res = strstrrepl( str, oldstr, newstr )

   Replaces every occurrence of string 'oldstr' in 'str' with
   'newstr'.
   'str' can be an array of strings.

*/
{
   n = numberof( str );

   len_oldstr = strlen(oldstr);
   len_newstr = strlen(newstr);
   outstr = str;

   for( i = 1; i <= n; i++ ) {

      p1 = 1;
      wstr = str(i);

      while(1) {
         pos = strpos( wstr, oldstr, p1 );
         if( !pos ) break;
         len_str = strlen(wstr);

         part1 = pos > 1 ? strpart( wstr, 1:pos-1 ) : "";
         part2 = pos+len_oldstr <= len_str ? strpart( wstr, pos+len_oldstr:0) : "";
         wstr = part1 + newstr + part2;
         p1 = pos+len_newstr;
      }
      outstr(i) = wstr;

   }
   return outstr;
}

/* Function strjoin */

func strjoin( str, sep )
/* DOCUMENT res = strjoin( str, sep )

   The string array 'str' is returned as a single
   string glued together with the string 'sep'

*/
{
   n = numberof(str);
   if( n == 1 ) return str(1);
   newstr = str(1);
   for( i = 2; i <= n; i++ ) {
      newstr += sep + str(i);
   }
   return newstr;
}

/* Function toint */

func toint( argu )
/* DOCUMENT res = toint( argu )

   Returns an integer value of 'argu' that can be
   a string, a float, or a double.

*/
{
   if( "string" == typeof(argu) ) {
      return int(atoi(argu));
   } else {
      return int(argu);
   }
}

/* Function tolong */

func tolong( argu )
/* DOCUMENT res = tolong( argu )

   Returns an long integer value of 'argu' that can be
   a string, a float, or a double.

*/
{
   if( "string" == typeof(argu) ) {
      return atoi(argu);
   } else {
      return long(argu);
   }
}

/* Function atoi */

func atoi( str )
/* DOCUMENT res = atoi( str )

   Returns the decimal value of an array of
   strings as a long array

*/
{
   val = 1;
   n = numberof(str);
   narr = array( long, n );
   for(i=1;i<=n;i++) {
      m = sread( str(i), val );
      narr(i) = val;
   }
   return n == 1 ? narr(1) : narr;
}

/* Function atof */

func atof( str )
/* DOCUMENT res = atof( str )

   Returns the decimal value of an array of
   strings as a double array

   SEE ALSO: ftoa, f2scienota, wrdig (in basic.i)

*/
{
   val = 1.0;
   n = numberof( str );
   darr = array( double, n );
   for(i=1;i<=n;i++) {
      m = sread( str(i), val );
      darr(i) = val;
   }
   return n == 1 ? darr(1) : darr;
}

/* Function is_digit */

func is_digit( str, sgn= )
/* DOCUMENT res = is_digit( str, sgn= )

   Returns array of same dimension as 'str' indicating
   if only digits are present.
   With keyword 'sgn' initial signs (+-) will also be accepted.

   Argument str must be a string scalar or a string array.
*/
{
   n_str = numberof(str);
   if( n_str == 0 ) return [];
   if( structof(str) != string ) error,"IS_DIGIT bad argument";
   if( n_str > 1 ) {
      res = array(0,n_str);
      for( i = 1; i <= n_str; i++ ) {
         res(i) = _is_digit( str(i), sgn=sgn );
      }
      return res;
   } else return _is_digit( str(1), sgn=sgn );
}

func _is_digit( str, sgn= )
/* DOCUMENT res = _is_digit( str, sgn= )

   Returns 1 if string 'str' only contains digits
   zero otherwise.
   With keyword 'sgn' initial signs (+-) will also be accepted.

*/
{
   len = strlen(str);
   if( !len ) return 0;
   // check for initial + or - and remove if present
   // when keyword 'sgn' has been set
   if( sgn ) {
      c = strpart( str, 1:1 );
      if( c == "-" || c == "+" ) {
         str = strpart(str, 2:0);
         len--;
         if( !len ) return 0; // a "+" and "-" string is not accepted
      }
   }
   for( i = 1; i <= len; i++ ) {
      c = strpart( str, i:i );
      p = strpos( "0123456789", c, 1);
      if( p < 1 ) return 0;
   }
   return 1;
}

/* Function is_number */

func is_number( str )
/* DOCUMENT res = is_number( str )

   Returns 1 if string 'str' is a valid number,
   integer or floating point, zero otherwise.

*/
{
   if( is_digit(str) ) return 1; // simple integer
   // check for preceding sign
   c = strpart(str,1:1);
   if( c == "+" || c == "-" ) str = strpart( str,2:0);
   if( is_digit(str) ) return 1; // signed integer
   str = strupcase( str );
   pose = strpos( str, "E", 1 );
   if( pose < 1 ) { // 'E' was not found
      return is_fnumber( str, nosign=1 );
   } else {         // This is E-format number
      if( pose == strlen(str) ) return 0; // illegal to end with 'E'
      // check string before the 'E'
      sbefore = strpart(str,1:pose-1);
      if( !is_digit( sbefore ) ) {
         // if not an integer then it must be an f-number
         if( !is_fnumber( sbefore, nosign=1 ) ) return 0;
      }

      // check string after

      safter = strpart(str, pose+1:0 );
      // check for preceding sign
      c = strpart(safter,1:1);
      if( c == "+" || c == "-" ) safter = strpart( safter,2:0);
      return is_digit( safter );
   }
}


/* Function is_fnumber */

func is_fnumber( str, nosign= )
/* DOCUMENT res = is_fnumber( str, nosign= )

   Returns 1 if string 'str' is a valid f-format number,
   i.e. floating point, zero otherwise.
   If keyword 'nosign' is set then no sign can be present.

*/
{
   // check for preceding sign and remove if present
   if( !nosign ) {
      c = strpart(str,1:1);
      if( c == "+" || c == "-" ) str = strpart( str,2:0);
   }

   // check for decimal point
   posd = strpos( str, ".", 1 );
   if( posd < 1 ) return 0; // Must be present
   len = strlen( str );
   if( len == 1 ) return 0; // Only decimal point is present
   if( posd > 1 ) {
      if( !is_digit(strpart(str,1:posd-1)) ) return 0;
                        // only digits are accepted before the decimal point
   }
   if( posd < len ) {
      if( !is_digit(strpart(str,posd+1:len)) ) return 0;
                        // only digits are accepted after the decimal point
   }
   return 1; // Passed all tests

}


/* Function itoa */

func itoa( n, ndig, plus= )
/* DOCUMENT res = itoa( n, ndig, plus= )

   Returns a string with neither leading nor trailing
   blanks. The optional second argument gives the number of
   digits e.g.
     > itoa(25,4)
    "0025"

   Keyword 'plus' causes a '+' sign when the number is positive.

   SEE ALSO: swrite(...)      when a specific format is requested.

*/
{
   ss = "";
   if( plus && n > 0 ) ss = "+";
   if( is_void(ndig) ) {
      return swrite(long(n),format=ss+"%i");
   } else {
      return swrite(long(n),format=ss+"%0"+itoa(ndig)+"i");
   }
}

/* Function ftoa */

func ftoa( x, ndec=, lz=, plus=, sci= )
/* DOCUMENT res = ftoa( x, ndec=, lz=, plus=, sci= )

   Returns a string representing 'x' with neither leading nor trailing blanks.
   The number of decimals (i.e. 'd' in %w.df) can be defined by
   keyword 'ndec'.
   If keyword 'lz' is given (operates only together with 'ndec')
   the format is defined as '%0w.df' and 'w' takes the
   value of 'lz'+'ndec'+1 so 'lz' is the number of digits preceding
   the decimal point.

   If keyword 'plus' has been set then a plus-sign (+) will precede
   the number if zero or positive.
   The keyword 'sci' forces e-format (scientific format).

    SEE ALSO: atof, f2scienota, wrdig, fndeci (in basic.i)
*/
{
   dx = double(x);
   if( is_void(ndec) ) return swrite(dx,format="%g");
   adx = abs(dx);
   if( lz ) {
      fmt = "%0"+itoa(lz+ndec+1)+"."+itoa(ndec);
   } else {
      fmt = "%."+itoa(ndec);
   }
   if( plus && dx >= 0.0 ) fmt = "+"+fmt;
   if( sci || (adx != 0.0 && (adx > 1.e6 || adx < 1.e-6)) ) {
      fmt += "e";
   } else fmt += "f";
   return swrite(dx,format=fmt);
}

/* Function strupcase */

func strupcase( str )
/* DOCUMENT res = strupcase( str )

   Returns a string converted to upper case

*/
{
   strup = str;
   strcase,1,strup;
   return strup;
}

/* Function strlowcase */

func strlowcase( str )
/* DOCUMENT res = strlowcase( str )

   Returns a string converted to lower case

*/
{
   strlow = str;
   strcase,0,strlow;
   return strlow;
}


/* Function write_slist */

func write_slist( filename, list, app= )
/* DOCUMENT write_slist, filename, list, app=

    Writes the string array 'list' to the file given by 'filename'
    Keyword 'app' causes the list to be appended.
    2007-02-16/NJW
*/

{
   if( app ) {
      f = open( filename, "a" );
   } else {
      f = open( filename, "w" );
   }
   n = numberof(list);
   for( i = 1; i <= n; i++ ) {
      nw = write(f, format="%s\n", list(i));
   }
   close,f;
}

/* Function read_slist */

func read_slist( filename )
/* DOCUMENT list = read_slist( filename )

   Reads the text file content of file 'filename' into a string array
   Same as 'rdfile', but returns a void if the file is not found.
*/
{
   if( file_test( filename ) ) {
      return rdfile( filename );
   }
   return [];
}

/* Function polyn */

func polyn( x, c )
/* DOCUMENT res = polyn( x, c )

    Evaluates the polynomium given by coefficients in array 'c'
    for all x values given by 'x'

    res = c(1) + x*c(2) + x^2*c(3) + ...
*/
{

   n = numberof(c);
   t = c(0);
   for( i = 1; i < n; i++ ) t = c(-i) + t*x;
   return t;

}


/* Function dpolyn */

func dpolyn( x, c )
/* DOCUMENT res = dpolyn( x, c )

    Evaluates the first derivative of the polynomium given
    by coefficients in array 'c' for all x values given by 'x'
*/
{

   n = numberof(c);
   if( n == 1 ) return x*0;
   if( n == 2 ) return array(c(2),numberof(x));
   a = indgen(n) - 1;
   return polyn( x, a(2:0)*c(2:0));
}

/* Function ddpolyn */

func ddpolyn( x, c )
/* DOCUMENT res = ddpolyn( x, c )

    Evaluates the second derivative of the polynomium given
    by coefficients in array 'c' for all x values given by 'x'
*/
{

   n = numberof(c);
   if( n <= 2 ) return x*0;
   if( n == 3 ) return array(2*c(3),numberof(x));
   a = indgen(n) - 1;
   b = shift(a,-1);
   return polyn( x, b(3:0)*a(3:0)*c(3:0));
}

/* Function dcoef_polyn */

func dcoef_polyn( c )
/* DOCUMENT new_coefs = dcoef_polyn( c )

   Returns coefficents for derivative of a polynomium
*/
{
   n = numberof(c);
   if( n == 1 ) return [0.];
   a = indgen(n) - 1;
   return a(2:0)*c(2:0);
}


/* Function icoef_polyn */

func icoef_polyn( c )
/* DOCUMENT new_coefs = icoef_polyn( c )

   Returns coefficents for integral of a polynomium
   (Constant is set to zero).
*/
{
   n = numberof(c);
   cc = c/indgen(n);
   grow,cc,0.;
   cc = shift(cc,-1);
   return cc;
}


/* Function systime */

func systime(a) {
/* DOCUMENT systime
    Returns current time as YYYY-MM-DDTHH:MM:SS
 */

   getdate, date;
   gettime, time;
   s = "20"+strpart(date,7:8)+"-"+strpart(date,4:5)+"-"+strpart(date,1:2);
   s = s + "T"+time;
   return s;
}

/* Function ndate */

func ndate(n) {
/* DOCUMENT s = ndate(n)
    If n is not defined or n == 1 then
          returns today's date as YYMMDD
    If n == 2 then returns today's date as YYYY-MM-DD
    If n == 3 then returns today's date/time as YYYY-MM-DDThh:mm:ss
    If n == 4 then returns today's date as 3 element array [YY,MM,DD]
    If n == 5 then returns today's date/time as YYYY-MM-DD (hh:mm)
    If n == 6 then returns today's date/time as YYMMDDThhmm
       or higher
*/

   require,"string.i";
   local date;

   getdate, date;
   if( numberof(n) ) {
      if( n <= 1 ) {
         return strpart(date,7:8)+strpart(date,4:5)+strpart(date,1:2);
      } else if ( n == 2 ) {
         return "20"+strpart(date,7:8)+"-"+strpart(date,4:5)+"-"+strpart(date,1:2);
      } else if ( n == 3 ) {
         return "20"+strpart(date,7:8)+"-"+strpart(date,4:5)+"-"+strpart(date,1:2) \
           +"T"+gettime();
      } else if ( n == 4 ) {
        return [atoi(strpart(date,7:8)), atoi(strpart(date,4:5)), atoi(strpart(date,1:2)) ];
      } else if ( n == 5 ) {
         return "20"+strpart(date,7:8)+"-"+strpart(date,4:5)+"-"+strpart(date,1:2) \
             +" ("+strpart(gettime(),1:5)+")";
      } else {
         tstr = gettime();
         return strpart(date,7:8)+strpart(date,4:5)+strpart(date,1:2) \
             +"T"+strpart(tstr,1:2)+strpart(tstr,4:5);
      }
   } else {
      return strpart(date,7:8)+strpart(date,4:5)+strpart(date,1:2);
   }
}

/* Function wvar */

func wvar( arr )
/* DOCUMENT res = wvar(arr)
      Find Variance of array 'arr'
*/

{
   arr = double(arr);
   num = numberof(arr);
   aver = sum(arr)/num;
   return sum((arr-aver)^2)/num;
}

/* Function wrms */

func wrms( arr, weight )
/* DOCUMENT res = wrms(arr [, weight] )

      Find RMS of array 'arr'
      If weight is given it must have same dimensions as 'arr'
*/

{
   arr = double(arr);
   aver = wavg( arr, weight );
   if( is_void(weight) ) {
      return sqrt(sum((arr-aver)^2)/numberof(arr));
   } else {
      return sqrt(sum(weight*(arr-aver)^2)/sum(weight));
   }
}

/* Function wavg */

func wavg( arr, weight, confine= )
/* DOCUMENT res = wavg(arr [, weight][, confine=])

      Find average of array 'arr'

      If present 'weight' must conform with 'arr'.

      Keyword 'confine' is the number of RMS outside which data are excluded.

*/

{
   arr = double(arr);
   na = numberof(arr);
   dms = dimsof(arr);
   if( is_void(weight) ) weight = array( 1.0, dms);

   if( is_void(confine) ) {
      return sum(weight*arr)/sum(weight);
   } else {
      w = indgen(na);
      nw_previous = 0;
      nw = na;
      while( nw_previous != nw ) {
         arr = arr(w);
         weight = weight(w);
         nw_previous = nw;
         a = sum(weight*arr)/sum(weight);
         r = wrms(arr, weight);
         write,format="%7i elements, avg = %13.5e, rms = %13.5e\n", nw, a, r;
         w = where( arr > a - confine*r & arr < a + confine*r );
         nw = numberof(w);
      }
      return a;
   }
}

/* Function strpos */

func strpos( string_in, str, start, rev= )
/* DOCUMENT pos = strpos(string_in, str, start[, rev=])
      Find position of 'str' in 'string_in' from position 'start'.
      Returns 0 if no match.
      Keyword 'rev' will change the direction of the search.
*/

{
   n_string_in = numberof(string_in);
   res = array(long, n_string_in);
   if( is_void(start) ) start = rev ? 0 : 1;

   for( j = 1; j <= n_string_in; j++ ) {

      cur_string = string_in(j);
      ll = strlen(cur_string);
      lstart = start == 0 ? ll : start;
      stringx = rev ? strpart(cur_string,1:lstart) : strpart(cur_string,lstart:ll);
      if( !strmatch( stringx, str ) ) {
         res(j) = 0;
         continue;
      }
      l = strlen(str);
      if( rev ) {
         ll = lstart;
         for(i=ll-l+1;i>=1;i--) {
            if( strpart(stringx,i:i+l-1) == str ) break;
         }
         res(j) = i;
      } else {
         ll -= (lstart-1);
         for(i=1;i<=ll-l+1;i++) {
            if( strpart(stringx,i:i+l-1) == str ) break;
         }
         res(j) = i+lstart-1;
      }
   }
   if( n_string_in == 1 ) {
      return res(1);
   } else {
      return res;
   }

}


/* Function strcompress */

func strcompress( string_in, all=, st= )
/* DOCUMENT str = strcompress(string_in, all=, st=)
     Reduce all series blanks to single blanks
     Keyword 'all' causes removal of all blanks
     Keyword 'st' is a 1 char string to replace 'blank (or space)'
*/

{
   if( st ) {
      sep = strpart(st,1:1);
      dsep = sep + sep;
   } else {
      sep = " ";
      dsep = "  ";
   }

   result = string_in;
   nres = numberof(result);

   for( i = 1; i <= nres; i++ ) {
      if( all ) {
         while( strmatch(result(i),sep) ) {
            p = strpos(result(i),sep,1);
            if( p == 1 ) {
               result(i) = strpart(result(i),2:0);
            } else if( p == len ) {
               result(i) = strpart(result(i),1:-1);
            } else {
               result(i) = strpart(result(i),1:p-1)+strpart(result(i),p+1:0);
            }
         }
      } else {
         while( strmatch(result(i),dsep) ) {
            len = strlen(result(i));
            p = strpos(result(i),dsep,1);
            result(i) = strpart(result(i),1:p)+strpart(result(i),p+2:0);
         }

      }
   }
   return( result);
}

/* Function arrcat */

func arrcat( arr, ..)
/* DOCUMENT arr = arrcat( arr1, arr2, ..)
     Concatenate two or more 1D arrays
*/

{
   d = dimsof(arr);
   n = numberof(arr);
   if( d(1) > 1 ) {
      print,"arrcat error, bad dimension arg# 1";
      return -1;
   }
   n_arg = 1;
   while( more_args() ) {
      n_arg++;
      arr2 = next_arg();
      d2 = dimsof(arr2);
      n2 = numberof(arr2);
      if( d2(1) > 1 ) {
         print,"arrcat error, bad dimension arg#", n_arg;
         return -1;
      }
      if( typeof(arr) != typeof(arr2) ) {
         print,"arrcat error, not the same type arg#", n_arg;
         return -1;
      }
      grow, arr, arr2;
   }

   return arr;
}

/* Function shift */

func shift(x,..)
/* DOCUMENT z = shift(x,nd1,nd2,..)
   Shift array x 'nd1' places first index
   'nd2'places second index etc.
   with wrap around
   nd positive: left shift
   nd negative: right shift
*/

{
   dms = dimsof(x);
   if( dms(1) == 0 ) {
      print,"shift of a scalar is not possible";
   } else if( dms(1) == 1 ) {
      nd = next_arg();
      n = dms(2);
      return x((indgen(n)-1+n+nd)%n+1);
   } else if( dms(1) == 2 ) {
      nd1 = next_arg();
      nd2 = next_arg();
      n1 = dms(2);
      n2 = dms(3);
      return x((indgen(n1)-1+n1+nd1)%n1+1,(indgen(n2)-1+n2+nd2)%n2+1);
   } else if( dms(1) == 3 ) {
      nd1 = next_arg();
      nd2 = next_arg();
      nd3 = next_arg();
      n1 = dms(2);
      n2 = dms(3);
      n3 = dms(4);
      return x((indgen(n1)-1+n1+nd1)%n1+1,(indgen(n2)-1+n2+nd2)%n2+1, \
                       (indgen(n3)-1+n3+nd3)%n3+1);
   } else {
      print,"shift: number of dimensions is too high";
   }
   return x;
}

/* Function strput */

func strput( line, piece, pos )
/* DOCUMENT line = strput(line, piece, pos)

   Replace part of 'line' with 'piece' starting at 'pos'
   If 'piece' does not fit in 'line' then 'piece' is truncated
   'line' will keep its length.

   2004-09-15/NJW

*/

{
   len_line = strlen(line);
   len_piece = strlen(piece);
   if( pos < 1 || pos > len_line ) return( line );

   len_possible = len_line - pos + 1;
   if( len_possible < len_piece ) {
      piece = strpart(piece,1:len_possible);
      len_piece = len_possible;
   }
   if( pos > 1 ) {
      newline = strpart(line,1:pos-1);
   } else {
      newline = "";
   }
   newline += piece;
   len_rest = len_line - pos - len_piece + 1;
   if( len_rest > 0 ) newline += strpart(line,pos+len_piece:len_line);

   return( newline );

}

/* Function strmsplit */

func strmsplit( string_in, str, nocompress= )
/* DOCUMENT str_array = strmsplit( string_in, str, nocompress= )

   Return string array with substrings separated by any of the
   characters in 'str' (must be a string) in the original string.
   Normal operation is done with string compression to avoid
   duplicated, leading and trailing search characters.
   This can be overridden by keyword 'nocompress'

   2011-06-16/NJW

*/
{
   len = strlen(str);
   sstr = strpart(str,1:1);
   outstr = [];
   if( len > 1 ) {
      for( i = 2; i <= len; i++ ) {
         string_in = strcharrepl( string_in, strpart(str,i:i), sstr, all=1 );
      }
   }
   return strsplit( string_in, sstr, nocompress=nocompress );
}

/* Function strinsert */

func strinsert( string_in, piece, position )
/* DOCUMENT new_string = strinsert( string_in, piece, position )

    Returns a string (or string array) with 'piece' inserted
    immediately BEFORE the position 'position'.

    This function cannot be used to append the 'piece' to 'string_in'.

    2012-12-13/NJW
*/
{
   nstr = numberof(string_in);
   npos = numberof(position);
   if( npos != 1 && npos != nstr ) error,"Mismatch between argument dimensions.";
   res = array(string,nstr);
   if( nstr > 1 && npos == 1 ) position = array(position, nstr);
   for( i = 1; i <= nstr; i++ ) {
      if( position(i) > 1 ) {
         part1 = strpart( string_in(i), 1:position(i)-1 );
         part2 = strpart( string_in(i), position(i):0 );
         res(i) = part1+piece+part2;
      } else if( position(i) == 1 ) {
         res(i) = piece+string_in(i);
      } else { // 'position' is zero or negative
         pos = strlen(string_in(i)) + position(i);
         part1 = strpart( string_in(i), 1:pos-1 );
         part2 = strpart( string_in(i), pos:0 );
         res(i) = part1+piece+part2;
      }
   }
   return is_scalar(string_in) ? res(1) : res;
}

/* Function _strsplit */

func _strsplit( string_in, str, nocompress= )
/* DOCUMENT res = _strsplit( string_in, str, nocompress= )

   An auxiliary function for 'strsplit' to handle the case where 'str'
   is a string with more than one character.
   It calls 'strsplit' as many times as required.

   2014-12-03/NJW

   SEE ALSO strsplit
*/
{
   if( is_void(str) ) return strsplit( string_in, nocompress=nocompress );
   len = strlen(str);
   if( len == 1 ) return strsplit( string_in, str, nocompress=nocompress );
   res = strsplit( string_in, strpart(str,1:1), nocompress=nocompress );
   for( i = 2; i <= len; i++ ) {
      ress = [];
      n = numberof(res);
      for( j = 1; j <= n; j++ ) {
         grow,ress, strsplit( res(j), strpart(str,i:i), nocompress=nocompress );
      }
      res = ress;
   }
   return res;
}

/* Function strsplit */

func strsplit( string_in, str, nocompress= )
/* DOCUMENT str_array = strsplit( string_in, str, nocompress= )

   Return string array with substrings from the original string separated by any
   of the characters in 'str'. Normal operation
   is done with string compression to avoid duplicated, leading and
   trailing search characters.
   This can be overridden by keyword 'nocompress'

   Omitting 'str' causes a space to be used as separator.

   SEE ALSO _strpslit

   2004-09-16/NJW

*/

{
   if( is_void(str) ) str = " "; // default separator is space (blank)
   if( strlen(str) > 1 ) return _strsplit( string_in, str, nocompress=nocompress );
   if( !nocompress ) string_in = strcompress( strtrim(string_in,blank=str), st=str);
   result = [];
   prev = 1;
   n = 0;
   while( pos = strpos(string_in,str,prev) ) {
      n++;
      if( n == 1 ) {
         if( 1 == pos ) {
            result = "";
         } else {
            result = strpart(string_in,prev:pos-1);
         }
      } else {
         grow, result, strpart(string_in,prev:pos-1);
      }
      prev = pos + 1;
   }
   if( numberof(result) ) {
      grow, result, strpart(string_in,prev:0);
   } else {
      result = string_in;
   }

   return( result );

}

/* Function arcdist */

func arcdist( ra1, dc1, ra2, dc2, deg=, rad= )
/* DOCUMENT res = arcdist( ra1, dc1, ra2, dc2, deg=, rad=)

  Return distance along great circle from one point to another
  Unit must be supplied if not degrees (default)

*/

{
   require, "string.i"; // where 'is_scalar' resides

   iu = 0;
   if( deg ) iu++;
   if( rad ) iu += 2;
   if( iu == 3 ) {
      print,"ARCDIST: only one unit can be defined deg xor rad"
      return -1;
   }
   if( iu < 2 ) iu = 1; // choose degrees as default

   // Avoid singularity when coordinates are very close to
   // being identical
   factor = double(1);
   if( iu == 1 ) factor = pi/double(180);
   r1 = ra1 * factor;
   d1 = dc1 * factor;
   r2 = ra2 * factor;
   d2 = dc2 * factor;
   arg = (sin(d1)*sin(d2) + cos(d1)*cos(d2)*cos(r2-r1));
   if( is_scalar(arg) ) {
      if( arg > 1.0 ) arg = 1.0;
      if( arg < -1.0 ) arg = -1.0;
   } else {
      w = where( arg > 1.0 );
      if( numberof(w) > 0 ) arg(w) = 1.0;
      w = where( arg < -1.0 );
      if( numberof(w) > 0 ) arg(w) = -1.0;
   }
   return acos(arg)/factor;
}

/* Function file_search */

func file_search( name, directory, dir=, silent= )
/* DOCUMENT list = file_search(name, directory, dir=, silent=)

   Returns a list of regular files (i.e. not directories)
   by name of 'name' which may contain wild cards: * and ?.
   The directory must be given by 'directory'. If 'directory' is omitted
   then the search will be in the current directory.

   Keyword dir: When set only directories under the
   named one will be returned.

   SEE ALSO: file_rsearch

*/

{
   if( is_void(directory) ) directory = ".";
   // remove trailing '/' if found:
   if( strpart(directory,0:0) == "/" ) directory = strpart(directory,1:-1);
   if( !file_test(directory) ) {
      write,format="FILE_SEARCH: No such directory: %s\n", directory;
      return [];
   }
   list = lsdir( directory, subdirs );
   if( structof(list) == long ) {
      write,"No such directory: "+directory;
      return [];
   }
   if( dir ) list = subdirs;
   if( is_void(list) ) {
      if( !silent ) write,"There are no files in the directory";
      return [];
   }
   if( typeof(list) != "string" ) {
      if( !silent ) write,"There is no such directory";
      return [];
   }
   nlist = numberof(list);
   if( nlist == 0 ) return [];
   rlist = list;
   k = 0;
   nnames = numberof(name);
   for( j = 1; j <= nnames; j++ ) {
      for( i = 1; i <= nlist; i++ ) {
         if( strcompare( list(i), name(j) ) ) rlist(++k) = directory+"/"+list(i);
      }
   }
   if( k == 0 ) {
      return [];
   } else {
      rlist = rlist(1:k);
      return rlist(sort(rlist));
   }

}

/* Function strcompare */

func strcompare( obj, pattern ) {
/* DOCUMENT res = strcompare( obj, pattern)

   pattern may consist of wildcards such as * and ?
   Returns 1 if obj matches pattern else 0

*/

   if( obj == pattern ) return 1;
   if( pattern == "*" ) return 1;

   tok = strsplit(pattern,"*");
   ntok = numberof(tok);
   if( strpart(pattern,1:1) == "*" ) {ast1 = 1;} else {ast1 = 0;}
   if( strpart(pattern,0:0) == "*" ) {ast2 = 1;} else {ast2 = 0;}
   len = strlen(obj);
   len2 = strlen(tok(0));

   start = 1;
   for( i = 1; i <= ntok; i++ ) {
      p = qmstrpos2( obj, tok(i), start );
      if( p == 0 ) return 0;
      if( i == 1 && ast1 == 0 && p != 1 ) return 0;
     // if( i == ntok && ast2 == 0 && tok(i) != strpart( obj, -len2+1:0) ) return 0;
      if( i == ntok && ast2 == 0 &&
             qmcompare(strpart( obj, -len2+1:0),tok(i))==0 ) return 0;
      start = p + strlen(tok(i));
   }
   return 1;
}


/* Function qmcompare */

func qmcompare( obj, pattern ) {
/* DOCUMENT res = qmcompare( obj, pattern )

   pattern may contain ? wildcards

*/


   len = strlen(obj);
   if( len != strlen(pattern) ) return 0;
   // must have equal length

   for( i = 1; i <= len; i++ ) {
      c = strpart(pattern,i:i);
      if( c == "?" ) continue;
      if( strpart(obj,i:i) != c ) return 0;
   }
   return 1;
}

/* Function qmstrpos1 */

func qmstrpos1( str, piece, start ) {
/* DOCUMENT res = qmstrpos1( str, piece, start )

   Return position of 'piece' in 'str' considering
   that 'str' may contain ? wildcards

*/


   len = strlen(str);
   lenp = strlen(piece);
   if( lenp > len - start + 1 ) return 0;
   // must be room enough for 'piece'

   for( i = start; i <= len-lenp+1; i++ ) {
      c = strpart(str,i:i+lenp-1);
      if( qmcompare(piece,c) ) return i;
   }
   return 0;
}

/* Function qmstrpos2 */

func qmstrpos2( str, piece, start ) {
/* DOCUMENT res = qmstrpos2( str, piece, start )

   Return position of 'piece' in 'str' considering
   that 'piece' may contain ? wildcards

*/


   len = strlen(str);
   lenp = strlen(piece);
   if( lenp > len - start + 1 ) return 0;
   // must be room enough for 'piece'

   for( i = start; i <= len-lenp+1; i++ ) {
      c = strpart(str,i:i+lenp-1);
      if( qmcompare(c,piece) ) return i;
   }
   return 0;
}

/* Function get_next_filename */

func get_next_filename( pattern, &ser_str, dir=, latest= ) {
/* DOCUMENT name = get_next_filename( pattern, >ser_str, dir=, latest= )

    2003-02-05/NJW
    2006-06-15/NJW changed directory argument to keyword
    2007-08-15/NJW updated with argument 'ser_str'

   For a pattern abcNNNNdefg.hijk where NNNN is a serial number
   given as 'abc????defg.hijk' the name will be returned with
   NNNN+1. The string NNNN+1 by itself is returned in 'ser_str'.

   Keyword 'dir': If given search in that directory and return the
                  full path to it; else search local directory
           'latest': Returns latest rather than next
*/


   //  Special case of pattern: ./abcd?????efgh

   if( strpart(pattern,1:2) == "./") {
      patt = strpart(pattern,3:999);
   } else patt = pattern;

   if( numberof(dir) ) {
      //  Strip trailing '/' from dir
      if( strpart(dir,0:0) == "/" ) dir = strpart(dir,1:-1);
   } else dir = ".";

   //  Analyze the pattern

   qq = "?";
   // find first occurrence of '?'
   pos = strpos( patt, qq, 1);
   len = 0;
   p = pos;
   // count number of consequtive '?'s
   while( strpos( patt, qq, ++p ) );
   len = p - pos;

   lenstr = swrite(len,format="%i");
   fmt = "%"+lenstr+"."+lenstr+"i";
   sermax = 10^len - 1;


   lst = file_search(patt, dir);
   nlst = numberof(lst);

   if( nlst ) {
      // screen for non-digit patterns
      w = array(1,nlst);
      for( i = 1; i <= nlst; i++ ) {
         bname = basename(lst(i));
         check = strpart(bname,pos:pos+len-1);
         if( !is_digit(check) ) w(i) = 0;
      }
      lst = lst(where(w));
      nlst = numberof(lst);
   }

   if( nlst == 0 ) {
      ser_str = swrite(0,format=fmt);
      if( dir == "." ) {
         return strput(patt,ser_str,pos);
      } else {
         return dir+"/"+strput(patt,ser_str,pos);
      }
   }

   sh = strlen(dir) + 1;
   s = sort(lst); // sorting list of existing files
   high = lst(s(0));
   sernum = 1;
   // extract number of highest serial number file as a string
   numstr = strpart(high,sh+pos:sh+pos+len-1);
   n = sread(numstr,format="%d",sernum);
   if( sernum == sermax ) {
      print,"Warning from GET_NEXT_FILENAME, max serial number reached";
      --sernum;
   }
   if( latest ) --sernum;
   ser_str = swrite(++sernum,format=fmt);
   if( dir == "." ) {
      return strput(patt,ser_str,pos);
   } else {
      return dir+"/"+strput(patt,ser_str,pos);
   }

}


/* Function file_test */

func file_test( filename, gz=, nz= )
/* DOCUMENT flag = file_test(filename, gz=, nz= )

   Returns 1 if the file by name of 'filename' (including
   the path) is found. Else return 0.

   Keyword: gz   if set, accept filename.gz as well
                 If 'filename.gz' is found then 2 is returned
            nz   if set, accept only files of non-zero size

*/
{

   local dir, fname;

   if( typeof(filename) != "string" ) return 0;

   if( filename == "." ) return 1;
   if( filename == ".." ) return 1;
   if( filename == "./" ) return 1;
   if( filename == "../" ) return 1;

   if( strlen(get_env("HOST"))) {

      if( open(filename,"r",1) ) {
         if( nz ) if( filesize(filename)==0 ) return 0;
         return 1;
      } else {
         if( gz ) {
            if( open(filename+".gz","r",1) ) {
               if( nz ) if( filesize(filename+".gz")==0 ) return 0;
               return 2;
            }
         }
         return 0;
      }

   } else { // you are on a Windows(like) system where the 'open' trick
            // does not work for directories
      splitfname, filename, dir, fname;
      list = lsdir( dir );
      if( structof(list) == long ) return 0;
      if( numberof(list) == 0 ) return 0;
      n = numberof( where( fname == list ) );
      if( n ) {
         if(nz) if( filesize(filename)==0 ) return 0;
         return 1;
      }
      if( gz ) {
         return 2*numberof( where( fname+".gz" == list ) );
      } else return 0;
   }

}

/* Function pick_swid_str */

func pick_swid_str( arr_in, &arr_out, plus= )
/* DOCUMENT n = pick_swid_str( arr_in, >arr_out[, plus= ] )

    Update arr_out as string array with only the 12 digit SWID strings.
    If keyword 'plus' (int) is larger than zero then this additional
    number of characters following the SWID will be returned.
    Return number of SWIDs actually found.

    2004-02-16/NJW
    2006-08-16/NJW Converted to Yorick
    2006-10-04/NJW Added keyword 'plus'
*/
{
   nlist = numberof(arr_in);

   arr_out = [];
   count = 0;
   if( is_void(plus) ) plus = 0;

   for( i = 1; i <= nlist; i++ ) {

      // locate string with 12 consequtive digits

      len = strlen(arr_in(i));
      if( len < 12 ) continue;

      for( j = 1; j <= len-11; j++ ) {
         piece = strpart(arr_in(i),j:j+11);
         piecep = strpart(arr_in(i),j:j+11+plus);
         if( is_digit(piece) ) {
            if( plus ) {
               grow, arr_out, piecep;
            } else {
               grow, arr_out, piece;
            }
            count++;
            break;
         }
      }
   }
   return count;
}

/* Function zero2pi */

func zero2pi( x )
/* DOCUMENT newx = zero2pi( x )

    An adequate number of (2*pi) is added to or subtracted from
    x so that the result ('newx') is between 0 (zero) and 2*pi

    2006-12-28/NJW
*/
{
   xx = x;
   twopi = 2.0 * pi;
   if( dimsof(xx)(1) == 0 ) {
      while( xx >= twopi ) xx -= twopi
      while( xx < 0.0 ) xx += twopi
   } else {
      while( anyof(xx >= twopi) ) {
         w = where( xx >= twopi);
         xx(w) -= twopi;
      }
      while( anyof(xx < 0.0) ) {
         w = where( xx < 0.0 );
         xx(w) += twopi;
      }
   }

   return xx;
}

/* Function zero360 */

func zero360( x )
/* DOCUMENT newx = zero360( x )

    An adequate number of 360 is added to or subtracted from
    x so that the result ('newx') is between 0 (zero) and 360 degrees.

    2011-01-18/NJW, cloned from zero2pi
*/
{
   xx = x; // avoid altering 'x'
   if( dimsof(xx)(1) == 0 ) {
      while( xx >= 360.0 ) xx -= 360.0;
      while( xx < 0.0 ) xx += 360.0;
   } else {
      while( anyof(xx >= 360.0) ) {
         w = where( xx >= 360.0);
         xx(w) -= 360.0;
      }
      while( anyof(xx < 0.0) ) {
         w = where( xx < 0.0 );
         xx(w) += 360.0;
      }
   }

   return xx;
}

/* Function extract_box */

func extract_box( image, &pix1, &pix2, &w_1, &w_2, ll=, cen= )
/* DOCUMENT box = extract_box( image, (>)pix1, (>)pix2, (>)w_1, (>)w_2, ll=, cen= )

   image:  Input 2D image
   pix1 :  Lower left corner (ll=1) or center (cen=1) of box 1. coord.
   pix2 :  Lower left corner (ll=1) or center (cen=1) of box 2. coord.
   w_1  :  Width of box 1. coord.
   w_2  :  Width of box 2. coord.

   The box returned is always completely inside image and
   lower left coordinates are shifted to match this requirement.
   The center is defined as (w_? + 1)/2.

   At return the 'pix1' and 'pix2' values gives the lower-left (center)
   pixels of the extracted box (useful if position adjustment was
   necessary).

   If 'pix1' at input is a four element array then it is interpreted
   as (x1,y1,x2,y2) of pixel indices and the box is extracted accordingly.
   As a side-effect the contents of 'pix1' will be updated!

   2006-11-20/NJW
   2006-12-29/NJW updated with keywords and cen=1 option

   Example  > newimage = extract_box( image, [10,30,20,10] )
          is equivalent (except for the dimensional test) to
            > newimage = image(10:20,10:30)
*/
{

   if( numberof(pix1) == 4 ) { // interpret as (x1,y1,x2,y2)
      x1 = long(floor(pix1(1)+0.5));
      y1 = long(floor(pix1(2)+0.5));
      x2 = long(floor(pix1(3)+0.5));
      y2 = long(floor(pix1(4)+0.5));
      w_1 = abs(x2 - x1) + 1;
      ll = 1; cen = 0;
      pix1 = min(x1,x2);
      w_2 = abs(y2 - y1) + 1;
      pix2 = min(y1,y2);
   } else {
      ll = ll ? 1 : 0;
      cen = cen ? 1 : 0;
      if( ll+cen != 1 ) {
         write,"Exactly one of keywords 'll' or 'cen' must be set!";
         return [];
      }
   }
   h_1 = w_1 / 2;
   h_2 = w_2 / 2;
   if( cen ) {
      ll_1 = pix1 - h_1;
      ll_2 = pix2 - h_2;
   } else {
      ll_1 = pix1;
      ll_2 = pix2;
   }
   ll_1in = ll_1;
   ll_2in = ll_2;
   dms = dimsof(image);
   if( w_1 > dms(2) ) { // return entire array in this direction
      w_1 = dms(2);
      ll_1 = 1;
      pix1 = cen ? dms(2)/2 + 1 : 1;
   } else {
      if( ll_1 < 1 ) ll_1 = 1;
      if( ll_1 + w_1 - 1 > dms(2) ) ll_1 = dms(2) - w_1 + 1;
      pix1 += (ll_1 - ll_1in);
   }
   if( w_2 > dms(3) ) { // return entire array in this direction
      w_2 = dms(3);
      ll_2 = 1;
      pix2 = cen ? dms(3)/2 + 1 : 1;
   } else {
      if( ll_2 < 1 ) ll_2 = 1;
      if( ll_2 + w_2 - 1 > dms(3) ) ll_2 = dms(3) - w_2 + 1;
      pix2 += (ll_2 - ll_2in);
   }
   return image(ll_1:ll_1+w_1-1,ll_2:ll_2+w_2-1);
}

/* Function put_box */

func put_box( image, box, pix1, pix2, ll=, cen= )
/* DOCUMENT newimage = put_box( image, box, pix1, pix2, ll=, cen= )

   image:  Input 2D image
   box  :  Box, 2D image with dimensions <= image
   pix1 :  Lower left corner (ll=1) or center (cen=1) of box 1. coord.
   pix2 :  Lower left corner (ll=1) or center (cen=1) of box 2. coord.

   The image is returned with the box section updated to the box values.
   If the box does not fit completely inside the image no operation is performed.

*/
{

   imwork = image; // make working copy to avoid overwriting
   ll = ll ? 1 : 0;
   cen = cen ? 1 : 0;
   if( ll+cen != 1 ) {
      write,"Exactly one of keywords 'll' or 'cen' must be set!";
      return image;
   }
   dms_box = dimsof(box);
   if( dms_box(1) != 2 ) {
      write,"box is not a 2D array";
      return image;
   }
   h_1 = dms_box(2) / 2;
   h_2 = dms_box(3) / 2;
   if( cen ) {
      ll_1 = pix1 - h_1;
      ll_2 = pix2 - h_2;
   } else {
      ll_1 = pix1;
      ll_2 = pix2;
   }
   dms = dimsof(image);
   if( ll_1 < 1 || ll_1 + dms_box(2) - 1 > dms(2) ) {
      write,"Box does not fit in X-dimension";
      return image;
   }
   if( ll_2 < 1 || ll_2 + dms_box(3) - 1 > dms(3) ) {
      write,"Box does not fit in Y-dimension";
      return image;
   }
   imwork(ll_1:ll_1+dms_box(2)-1,ll_2:ll_2+dms_box(3)-1) = box;
   return imwork;
}

/* Function local_rms */

func local_rms( image, box_width )
/* DOCUMENT rms_map = local_rms( image, box_width )

   Returns the map of RMS derived in a box of size box_width * box_width
   centered around the pixel in question

   2006-11-20/NJW
*/
{

   dms = dimsof( image );
   res = array( 0.0, dms(2), dms(3) );

   h = long(box_width)/2;
   for( i = 1; i <= dms(2); i++ ) {
      for( j = 1; j <= dms(3); j++ ) {
         ipos = i; jpos = j;
         b = extract_box( image, ipos, jpos, box_width, box_width, cen=1 );
         res(i,j) = wrms(b);
      }
   }

   return res;

}

/* Function circ_area1 */

func circ_area1( d, capr, r )
/* DOCUMENT res = circ_area1( d, capr, r )

    A function to calculate the overlap area between two circles

     d is the distance between the two centers
     capr is the radius of the (larger) circle
     r    is the radius of the (smaller) circle

     xcapr is the distance from the center of the larger circle
           to the line between the two intersection points of the
           two circles
     xr    is equivalent to xcapr but referring to the smaller circle

     acapr is the area of the larger circle inside the smaller circle
     ar    is the area of the smaller circle inside the larger circle

    JEM-X parameters:

   * det_mask = 3385. ; mm
   * capr = 535./2
   * r    = 250./2

   2007-01-02/NJW, translated to Yorick

*/
{
   // define area common to both circles

   if( d  >  capr+r ) return 0.0;
   if( d  <  capr-r || d == 0.0 ) return pi * r^2;
   if( r > capr ) {
      write,"Bad arguments";
      return 0.0;
   }

   xcapr = (capr^2 + d^2 - r^2)/(2.*d);
   t1 = pi*capr^2/2;
   t2 = capr^2 * asin(xcapr/capr);
   t3 = xcapr*sqrt(capr^2 - xcapr^2);
   xr = d - xcapr;
   u1 = pi*r^2/2;
   u2 = r^2*asin(xr/r);
   u3 = xr * sqrt(r^2-xr^2);
   acapr = t1 - t2 - t3;
   ar = u1 - u2 - u3;
   a = ar+acapr;
   return a;
}

/* Function app_slash */

func app_slash( s )
/* DOCUMENT str = app_slash( s )

   Append a slash to the string 's' if there is no trailing slash

   SEE ALSO: rem_slash

    NJW/020305
    2006-10-03/NJW translated to Yorick

*/
{

   if( typeof(s) != "string" ) {
      write,"Bad argument for app_slash";
      return "/";
   }
   if( s == "" ) return "/";

   return (strpart(s,0:0) == "/") ? s : s+"/";

}

/* Function rem_slash */

func rem_slash( s )
/* DOCUMENT str = rem_slash( s )

   Remove a trailing slash from the string 's' if present

   SEE ALSO: app_slash

    2006-10-04/NJW

*/
{

   if( typeof(s) != "string" ) {
      write,"Bad argument for rem_slash";
      return "";
   }
   if( s == "" ) return "";

   return (strpart(s,0:0) == "/") ? strpart(s,1:-1) : s;

}

/* Function set_par */

func set_par( parfile, parname, new_value, lolim=, hilim=, arr=, chat= )
/* DOCUMENT status = set_par( parfile, parname, new_value, lolim=, hilim=, arr=, chat= )

    This is a tool to change a parameter in a parameter
    file in IRAF and DAL style. The main difference
    between this one and the FTOOL: pset is that pset
    will erase all comments.
    When keyword 'arr' is true then 'new_value' is given as a real array
    and should be converted to a string.

    status == 0 means running OK, a value < 0 implies an error of a kind
    and no change to the parameter file.

    This version will only change one parameter per call.

    2007-02-01/NJW reworked to Yorick
    2011-05-04/NJW updated to use PFILES

    SEE ALSO: get_par

*/
{

   if( !file_test(parfile) ) {
      ostype = get_env("OSTYPE");
      if( strlen(ostype) == 0 ) { // neither Linux nor Unix
         pfiles = "."; // *.par file must be in current directory
      } else pfiles = strsplit(get_env("PFILES"),";");
      // for writing: only in front of the ';' sign
      subpfiles = strsplit(pfiles(1),":");
      for( j=1;j<=numberof(subpfiles);j++ ) {
         if( file_test(subpfiles(j)+"/"+parfile) ) {
            parfile = subpfiles(j)+"/"+parfile;
            goto station;
         }
      }
      write,"SET_PAR: did not find any instance of "+parfile;
      return;
   }
   if(chat)write,"GET_PAR: got parfile as "+parfile;
   station:
   parlines = strtrim(rdfile( parfile ));
   numlines = numberof(parlines);

  /*
   * Look for 'arr' keyword and translate 'new_value' into string if true
   */

   if( arr ) {
      if( typeof(new_value) != "double" && typeof(new_value) != "float" ) {
         write,"Illegal combination of keyword and data type";
         return -12;
      }
      n = numberof(new_value);
      s = "";
      for( i = 1; i <= n; i++ ) {
         s += swrite(format="%f ",new_value(i));
      }
      new_value = s;
   }

  /*
   *  Analyze buffer line by line in search for the given parameter
   */

   found = 0;
   for( loop = 1; loop <= numlines; ++loop ) {
      line = parlines(loop);
      if( strpart(line,1:1) == "#" ) continue;
      // save parts in quotes, they may contain commas that are not to be counted
      qsave = [];
      psave = [];
      // locate first "
      p = strpos(line,"\"",1);
      while( p > 0 ) {
        // locate following "
        q = strpos(line,"\"",p+1);
        if( q > p ) {
           if( q > p+1 ) { // there are characters between the two "s
              grow, qsave, strpart(line, p+1:q-1 );
              // form the replacement string of only Xs
              repl_str = strjoin(array("X",q-p-1),"");
              // do the replacement
              line = strput(line, repl_str, p+1);
              grow, psave, p; // save start position
           }
           // locate next "
           p = strpos(line,"\"",q+1);
        } else {
           write,format="SET_PAR: Unpaired quotes in line %i\n", loop;
           return -1;
        }
      }
      // Locate the unquoted commas
      comma = 0; // assume a pseudo-comma in the very beginning
      pc = strpos(line,",",1);
      while( pc > 0 ) {
         grow, comma, pc;
         pc = strpos(line,",",pc+1);
      }
      grow, comma, strlen(line)+1; // assume a pseudo-comma in the very end
      // and with 6 ordinary commas this makes a total of 8
      if( numberof(comma) != 8 ) {
         write,format="SET_PAR: Format error in line %i\n", loop;
         return -2;
      }
      // replace the substituted pieces
      for(i=1;i<=numberof(qsave);i++) {
         line = strput( line, qsave(i), psave(i)+1);
      }
      // do the string splitting by the non-quoted commas
      tok = array(string,7);
      for(i=1;i<=7;i++) {
         tok(i) = strpart(line,comma(i)+1:comma(i+1)-1);
      }
      // Now we are ready to make a match
      if( tok(1) == parname ) {
         found++;
         if( tok(2) == "r" ) {  // new_value must be a number
            if( "string" == typeof(new_value) ) {
               write,format="SET_PAR: Got a string but a number was expected in line %i\n", loop;
               return -3;
            }
            new_value = double(new_value);
            // Check boundaries
	    lo_bound = is_void(lolim) ? atof(tok(5)) : lolim;
	    hi_bound = is_void(hilim) ? atof(tok(6)) : hilim;
            if( new_value < lo_bound || new_value > hi_bound ) {
               write,format="SET_PAR: Parameter %s is outside limits\n", parname;
               return -4;
            }
            tok(4) = swrite(new_value,format="%g");
            if( !is_void(lolim) ) tok(5) = swrite(lolim,format="%g");
            if( !is_void(hilim) ) tok(6) = swrite(hilim,format="%g");
         } else if( tok(2) == "i" ) { // new_value must be an integer
            t = typeof(new_value);
            if( t != "int" && t != "long" ) {
               write,format="SET_PAR: Got a non-integer but an integer was expected in line %i\n", loop;
               return -6;
            }
            // Check boundaries
	    lo_bound = is_void(lolim) ? atoi(tok(5)) : lolim;
	    hi_bound = is_void(hilim) ? atoi(tok(6)) : hilim;
            if( new_value < lo_bound || new_value > hi_bound ) {
               write,format="Parameter %s is outside limits\n", parname;
               return -6;
            }
            tok(4) = swrite(new_value,format="%i");
            if( !is_void(lolim) ) {
               t = typeof(lolim);
               if( t != "int" && t != "long" ) {
                  write,format="SET_PAR: Got a non-integer but an integer was expected in line %i\n", loop;
                  return -7;
               }
               tok(5) = swrite(lolim,format="%i");
            }
            if( !is_void(hilim) ) {
               t = typeof(hilim);
               if( t != "int" && t != "long" ) {
                  write,format="SET_PAR: Got a non-integer but an integer was expected in line %i\n", loop;
                  return -8;
               }
               tok(6) = swrite(hilim,format="%g");
            }
         } else if( tok(2) == "s" ) {
            if( "string" != typeof(new_value) ) {
               write,format="SET_PAR: Got a number but a string was expected in line %i\n", loop;
               return -9;
            }
            tok(4) = "\""+new_value+"\"";
            if( !is_void(lolim) ) tok(5) = "\""+lolim+"\"";
            if( !is_void(hilim) ) tok(6) = "\""+hilim+"\"";
         } else {
            write,format="Unsupported type definition in line %i\n", loop;
            return -10;
         }

         parlines(loop) = strjoin(tok,",");
      }

   }

   if( found != 1 ) {
      write,format="SET_PAR: Parameter %s was not found in %s\n", parname, parfile;
      return -11;
   }

   write_slist, parfile, parlines;

   return 0;

}

/* Function get_par */

func get_par( parfile, parname, arr=, chat= )
/* DOCUMENT value = get_par( parfile, parname, arr=, chat= )
    This is a tool to read the value of a parameter in a parameter
    file in IRAF and DAL style. The parameter file is first searched
    in the current directory then according to the PFILES
    environment variable.
    If keyword 'arr' is true then the parameter is expected to be
    a string that should be split up into a number of double's.

    A nil value ([]) is returned if an error occurred.

    2007-02-02/NJW
    2011-05-04/NJW, updated to use PFILES

    SEE ALSO: set_par

*/
{

   if( !file_test(parfile) ) {
      ostype = get_env("OSTYPE");
      if( strlen(ostype) == 0 ) { // neither Linux nor Unix
         pfiles = "."; // *.par file must be in current directory
      } else pfiles = strsplit(get_env("PFILES"),";");
      for( i=1;i<=numberof(pfiles);i++ ) {
         subpfiles = strsplit(pfiles(i),":");
         for( j=1;j<=numberof(subpfiles);j++ ) {
            if( file_test(subpfiles(j)+"/"+parfile) ) {
               parfile = subpfiles(j)+"/"+parfile;
               goto station;
            }
         }
      }
      write,"GET_PAR: did not find any instance of "+parfile;
      return [];
   }
   station:
   if(chat)write,"GET_PAR: got parfile as "+parfile;
   parlines = strtrim(rdfile( parfile ));
   numlines = numberof(parlines);

  /*
   *  Analyze buffer line by line in search for the given parameter
   */

   found = 0;
   for( loop = 1; loop <= numlines; ++loop ) {
      line = parlines(loop);
      if( strlen(line) == 0 ) continue;
      if( strpart(line,1:1) == "#" ) continue;
      // save parts in quotes
      qsave = [];
      psave = [];
      // locate first "
      p = strpos(line,"\"",1);
      while( p > 0 ) {
        // locate following "
        q = strpos(line,"\"",p+1);
        if( q > p ) {
           if( q > p+1 ) { // there is text between "s
              grow, qsave, strpart(line, p+1:q-1 ); // save the text
              repl_str = strjoin(array("X",q-p-1),"");
              // replace text with "XXXX..XXX"
              line = strput(line, repl_str, p+1);
              grow, psave, p; // save start position of text
           }
           // locate next "
           p = strpos(line,"\"",q+1);
        } else {
           write,format="Unpaired quotes in line %i\n", loop;
           return [];
        }
      }
      // Locate the unquoted commas
      comma = 0; // assume a pseudo comma in the very beginning
      pc = strpos(line,",",1);
      while( pc > 0 ) {
         grow, comma, pc;
         pc = strpos(line,",",pc+1);
      }
      grow, comma, strlen(line)+1; // assume a pseudo comma in the very end
      // and with 6 ordinary commas this makes a total of 8
      if( numberof(comma) != 8 ) {
         write,format="GET_PAR: Format error in line %i\n", loop;
         return [];
      }
      // replace the substituted pieces
      for(i=1;i<=numberof(qsave);i++) {
         line = strput( line, qsave(i), psave(i)+1);
      }
      // do the string splitting by the non-quoted commas
      tok = array(string,7);
      for(i=1;i<=7;i++) {
         tok(i) = strpart(line,comma(i)+1:comma(i+1)-1);
      }
      // Now we are ready to make a match
      if( tok(1) == parname ) {
         found++;
         if( tok(2) == "r" ) {  // returned value will be a double
            return atof(tok(4));
         } else if( tok(2) == "i" ) { // returned value will be a long
            return atoi(tok(4));
         } else if( tok(2) == "s" ) {
            s = tok(4);
            if( strpart(s,1:1) == "\"" ) s = strpart(s,2:0);
            if( strpart(s,0:0) == "\"" ) s = strpart(s,1:-1);
            if( arr ) {
               tok = strsplit(s," ");
               ntok = numberof(tok);
               vec = array(double,ntok);
               for(i=1;i<=ntok;i++) vec(i) = atof(tok(i));
               return vec;
            } else {
               return s;
            }
         } else {
            write,format="GET_PAR: Unsupported type definition in line %i\n", loop;
            return [];
         }

      }

   }

   if( found != 1 ) {
      write,format="GET_PAR: Parameter %s was not found in %s\n", parname, parfile;
      return [];
   }

   return [];

}

/* Function esc_underscore */

func esc_underscore( str_array )
/* DOCUMENT newstring = esc_underscore(  str_array  )

   The underscore character is prefixed with an exclamation mark (!)
   to avoid the special meaning for formatting (starting subscript).

   2007-03-09/NJW
*/
{
   newstring = str_array ;
   n = numberof(str_array);

   // Identify the occurrences of '_'

   for( i = 1; i <= n; i++ ) {

      spos = 1;
      pos = strpos(str_array(i),"_",1);
      if( pos > 0 ) {
         newstring(i) = "";
         while( pos > 0 ) {
            if( pos == 1 ) {
               newstring(i) += "!_";
            } else {
               newstring(i) += strpart(str_array(i),spos:pos-1)+"!_";
            }
            spos = pos + 1;
            pos = strpos(str_array(i),"_",spos);
         }
         newstring(i) += strpart(str_array(i),spos:0);
      }
   }

   return newstring;

}

/* Function array_gauss */

func array_gauss( x, coef)
/* DOCUMENT curve = array_gauss( x, coef )

   Return a gauss curve when
    coef = [amplitude, mean, sigma[,a1[,a2[,a3]]]]

   curve = amplitude*exp(-0.5*((x-mean)/sigma)^2) + a1 + x*a2 + x*x*a3

    2006-02-20/NJW (Yorick version)
    2007-05-03/NJW Expanded to include parabolic bkg term
*/
{
   ncoef = numberof( coef );
   if( ncoef < 3 || ncoef > 6 ) {
     write,"Array 'coef' does not have proper length";
     return [];
   }
   arg = abs((x - coef(2)) / coef(3));
   w = where( arg < 12.0 );
   y = x * 0.0;
   y(w) = coef(1)*exp(-0.5*arg(w)^2);
   n = numberof(coef);
   if( n > 3 ) y += coef(4);
   if( n > 4 ) y += x*coef(5);
   if( n > 5 ) y += x*x*coef(6);

   return y;

}

/* Function array_igauss */

func array_igauss( x, coef)
/* DOCUMENT curve = array_igauss( x, coef )

   Return a integrated, averaged gauss curve when
    coef = [amplitude, mean, sigma[,a1[,a2[,a3]]]]
   'x' is interpreted as boundaries and 'curve' has
   one element fewer than 'x'.

   curve = amplitude*Integ(exp(-0.5*((x-mean)/sigma)^2))/(x2-x1) + a1 + x*a2 + x*x*a3

    2010-04-02/NJW
*/
{
   ncoef = numberof( coef );
   if( ncoef < 3 || ncoef > 6 ) {
     write,"Array 'coef' does not have proper length";
     return [];
   }
   s2 = coef(3)*sqrt(2);
   x1 = x(1:-1);
   x2 = x(2:0);
   tb = erf((x2-coef(2))/s2);
   ta = erf((x1-coef(2))/s2);
   res = sqrt(pi/2)*coef(3)*coef(1)*(tb - ta)/(x2-x1);
   if( ncoef > 3 ) res += coef(4);
   if( ncoef > 4 ) res += x(zcen)*coef(5);
   if( ncoef > 5 ) res += x(zcen)*x(zcen)*coef(6);
   return res;
}

func draw_from_dist( x_arr, y_arr, n )
/* DOCUMENT values = draw_from_dist( x_arr, y_arr, n )
            values = draw_from_dist( y_arr, n)

 First version:
   returns 'n' values of 'x' when the distribution is
   described by y_arr and corresponding x_arr.
 Second version:
   returns 'n' integers, namely the indices of the
   distribution array 'y_arr'.

   2007-08-02/NJW
   2010-04-16/NJW Updated to use 'integ'
   2019-01-10/NJW Updated with second version (returning integers)
*/
{

   if( is_void(n) ) { // We have the second version
      n = y_arr;
      y_arr = x_arr;
      ac = y_arr(psum); // accumulated array
      ac /= ac(0); // normalize
      r = random(n);
      return digitize( r, ac );
   }

   // coming here means that the first version has been chosen

   x = double(x_arr);
   y = double(y_arr);
   //+ y(1) = 0;

   if( numberof(x) < 100 ) {
      xx = span(x(1),x(0),100);
      y = interp( y, x, xx );
      x = xx;
   }

   //+ if( nonlin ) {

   //+    get_boundaries, x, x1, x2;
   //+    dx = x2 - x1;
   //+    y *= dx;

   //+ }

   //+ ycum = y(psum);
   //+ ycum /= ycum(0);
   //+ xnew = (x_arr + shift(x_arr,1))/2.;
   //+ xnew(1) = x_arr(1);
   //+ xnew(0) = x_arr(0);
   //+ return interp(xnew, ycum, random(n));

   yint = integ(y, x, x);
   yint /= yint(0);

   return interp( x, yint, random(n));

}

/* Function file_rsearch */

func file_rsearch( name, dir )
/* DOCUMENT list = file_rsearch( name, dir )

   Searches for all files by name of 'name' in directory 'dir' and
   all its subdirectories (recursive search).

   Omitting 'dir' makes file_rsearch start in the current directory.

   SEE ALSO: file_search

   2007-08-19/NJW
*/
{
   local nsubdirs, list, nlist, res, sres, subdirs;

   if( is_void(dir) ) dir = ".";
   if( !file_test(dir) ) {
      write,format="FILE_RSEARCH: No such directory: %s\n", dir;
      return [];
   }
   res = [];
   list = lsdir( dir, subdirs );
   if( structof(list) == long ) { // directory does not exist
      write,"No such directory: "+dir;
      return [];
   }
   nlist = numberof(list);
   if ( nlist ) {
      for(i=1;i<=nlist;i++) {
         if( strcompare( list(i), name)) grow,res,dir+"/"+list(i);
      }
   }
   nsubdirs = numberof( subdirs );
   if( nsubdirs ) {
      for( i = 1; i <= nsubdirs; i++) {
         sres = file_rsearch( name, dir+"/"+subdirs(i) );
         if( !is_void(sres) ) grow, res, sres;
      }
   }
   if( is_void(res) ) {
      return res;
   } else {
      return res(sort(res));
   }
}

/* Function strip_curly_br */

func strip_curly_br( str )
/* DOCUMENT newstr = strip_curly_br( str )

  Strip part of string inside curly bracketts e.g. jmx1_srcl_arf.fits{1}
  write, strip_curly_br("jmx1_srcl_arf.fits{1}")
  > jmx1_srcl_arf.fits

     2005-05-09/NJW
     2008-01-17/NJW translated from IDL to Yorick
*/
{
   pl = strpos( str, "{", 1 );
   pr = strpos( str, "}", 0, rev=1 );
   if( pl  >  1 && pr  >=  1 && pr  >  pl ) {
      return strpart(str,1:pl-1)+strpart(str,pr+1:9999);
   } else if( pl  ==  1 && pr  >  1 ) {
      return strpart(str,pr+1:9999);
   } else return str;
}

/* Function most_freq_elem */

func most_freq_elem( arr, &nmax, fuzzy= )
/* DOCUMENT val = most_freq_elem( arr, >num, fuzzy= )

   Returns the most frequent element of array 'arr'
   If all elements are different arr(1) is returned. If arr(n) and
   arr(m) appear an equal number of times larger than any other then
   the one with the lowest index will be returned.

   Arguments:
       arr     Input array
       num     (Returned) Number of occurrences

   Keyword
     fuzzy    a limit for considering elements equal:
                       x == x + eps when |eps| < fuzzy

    2008-01-25/NJW
    2014-11-18/NJW updated with 'fuzzy' keyword
*/
{

   narr = numberof( arr );
   narr2 = narr/2;
   nmax = 1;
   imax = 1;
   for( i = 1; i <= narr; i++ ) {
      if( fuzzy ) {
         n = sum( near( arr, arr(i), fuzzy)  );
      } else {
         n = sum( arr(i) == arr );
      }
      if( n > nmax ) {
         nmax = n;
         imax = i;
         if( nmax > narr2 ) break;
      }
   }

   return arr(imax);

}

/* Function monincr */

func monincr( arr, num )
/* DOCUMENT idx_piece = monincr( arr, num )

   returns the index numbers of the 'num'th piece of
   'arr' where it is monotoneously increasing.

   2008-01-28/NJW
*/
{

   w = where( arr(dif) < 0 );
   nw = numberof(w);
   if( num > nw + 1 ) {
      write,format="MONINCR error: only %i sections\n", nw+1;
      return arr;
   }
   if( nw == 0 ) return arr;
   narr = numberof(arr);
   if( num == 1 ) {
      return indgen(w(1));
   } else if( num == nw+1 ) {
      return indgen(narr-w(0))+w(0);
   } else {
      return indgen(w(num)-w(num-1))+w(num-1);
   }
}

/* Function chi2 */

func chi2( model, y, dy, red=, warn= )
/* DOCUMENT value = chi2( model, y, dy, red=, warn= )

 Evaluate chi-square as sum(((y-model)/dy)^2)

  Keyword
    red   when given the reduced chi-square is returned
          with DOF = numberof(y) - red
   warn   Give a warning if there are strongly deviating
          values:
          With   d2 = ((y-model)/dy)^2
          Give warning if anyof( d2 > warn * RMS(d2) )

  2008-03-14/NJW Translated from IDL
  2013-12-10/NJW updated with 'warn' keyword

*/
{

   if( is_void(dy) ) {
      write,"chi2 syntax: value = chi2(model, y, dy)";
      write,"or reduced_chi2 = chi2(model, y, dy, red=n )";
      return [];
   }

   nm = numberof(model);
   if( nm  !=  numberof(y) || nm  !=  numberof(dy) ) {
      write,"The arrays have different dimensions !";
      return [];
   }

   notz = where( dy  >  0.0 );
   n_notz = numberof(notz);
   if( n_notz  >  0 ) {
      d = (y(notz) - model(notz)) / dy(notz);
   } else {
      write,"chi2 error! dy has only illegal values";
      d = 0.;
   }
   if( !is_void(warn) ) {
      d2 = d^2;
      r = wrms(d2);
      w = where( d2 > warn*r );
      if( (nw = numberof(w)) ) {
         for(i=1;i<=nw;i++) {
            write,format="Warning: index %i has large excess\n", notz(w(i));
         }
      }
   }
   if( is_void(red) ) {
      return sum(d^2);
   } else {
      if( red < n_notz ) {
         return sum(d^2) / (n_notz - red);
      } else {
         error,"chi2: too few data points";
      }
   }
}

/* Function hpd */

func hpd( x_in, y_in )
/* DOCUMENT res = hpd( x_in, y_in )

   returns HPD (Half Power Diameter) from event arrays
   x and y

   SEE ALSO image2events in 'image.i'

   2008-04-07/NJW
*/
{

   if( is_void(x_in) ) {
      write,"Entered hpd with void x_in";
      return [];
   }
   if( is_void(y_in) ) {
      write,"Entered hpd with void y_in";
      return [];
   }
   x = double(x_in);
   y = double(y_in);
   n_events = numberof(x);
   if( n_events < 3 ) {
      write,"HPD: Too few events";
      return [];
   }
   // get best position
   xp = avg(x);
   yp = avg(y);
   write,format="Center with all %i events: %7.3f %7.3f\n", n_events, xp, yp;
   xrms = wrms(x - xp);
   yrms = wrms(y - xp);
   w = where( abs(x - xp) < 3. * xrms );
   if( numberof(w) < 2 ) {
      write,"HPDx: Too few events";
      return [];
   }
   xp = avg(x(w));
   nwx = numberof(w);
   w = where( abs(y - yp) < 3. * yrms );
   if( numberof(w) < 2 ) {
      write,"HPDy: Too few events";
      return [];
   }
   yp = avg(y(w));
   nwy = numberof(w);
   write,format="Improved center with %i, %i events: %7.3f %7.3f\n", nwx, nwy, xp, yp;

   r = sqrt((x-xp)^2 + (y-yp)^2);
   is = sort(r);
   r = r(is);
   whpd = 2. * r(n_events/2);

   return whpd;

}

/* Function fold_gaussx */

func fold_gaussx( e, f, sigma, silent= )
/* DOCUMENT fnew = fold_gaussx( [e,] f, sigma, silent= )

  Fold a spectrum with a Gauss function with a non-constant sigma

   Call -  e : energy array
                If 'e' is absent it is assumed to be indgen(numberof(f))
           f : array with spectrum, same size as 'e'
         sigma  : scalar with standard deviation,
                  or array with same size as 'e'

   Return : folded spectrum

      When ? /NJW
       2003-08-29/NJW   updated (IDL version)
       2006-02-16/NJW   Yorick version
       2006-02-20/NJW   Allow scalar sigma

    SEE ALSO: fold_gaussy
              to fold several spectra simultaneously based on the same
              energy array.
*/

{
   time_keeper = elapsed_time = array(double,3);
   timer, time_keeper;

   if( is_void(sigma) ) {
      // There are at most two arguments
      sigma = f;
      f = e;
      e = double(indgen(numberof(f)));
   }

   n = numberof(e);
   ee = double( e );
   ns = numberof(sigma);
   if( numberof(f) != n ) {
      write,"mismatching dimensions of 'e' and 'f'\n";
      return -1;
   }
   if( ns != 1 && ns != n ) {
      write,"'sigma' has illegal dimension\n";
      return -1;
   }
   sigma2 = sigma * sqrt(2.);
   if( ns == 1 ) sigma2 = sigma2(-:1:n);

   // Define channel boundaries

   eb = 0.5*(ee + shift(ee,-1));
   eb(1) = (3*ee(1) - ee(2))/2;
   grow,eb,(3*ee(0) - ee(-1))/2;
   g = ee * 0.;
   for( i = 1; i <= n; i++ ) {
      j = where((e > ee(i)-3.5*sigma2(i)) + (e < ee(i)+3.5*sigma2(i)) == 2);
      if( numberof(j) ) {
         x2 = (eb(j+1) - ee(i))/sigma2(i);
         x1 = (eb(j) - ee(i))/sigma2(i);
         w = erf(x2) - erf(x1);
         zum = f(j) * w;
         g(i) = sum(zum) / sum(w);
      }
   }

   timer, time_keeper, elapsed_time;
   if(!silent)write,"CPU time used: "+ftoa(elapsed_time(1),ndec=3)+" s";

   return g;

}

/* Function str2arr */

func str2arr( str )
/* DOCUMENT  arr = str2arr( str )

    Convert an expression such as '3,5,6,9-14,16-20'
    to array: [3,5,6,9,10,11,12,13,14,16,17,18,19,20]

    2004-02-11/NJW
    2006-10-01/NJW translated to Yorick
*/

{

   if( typeof( str ) != "string" ) {
      write,"Argument must be a string";
      return -1;
   }

   tok = strsplit(str,",");
   ntok = numberof(tok);

   first = 1;
   for( i = 1; i <= ntok; i++ ) {

      pos = strpos(tok(i),"-");
      if( pos == 1 || pos == strlen(tok(i)) ) {
         print,"Illegal format";
         return -1;
      }
      if( pos > 1 ) {
         upl = strsplit(tok(i),"-");
         nupl = numberof(upl);
         if( nupl != 2 ) {
            print,"Illegal format";
            return -1;
         }
         k1 = toint(upl(1));
         k2 = toint(upl(2));
         if( k2 <= k1 ) {
            print,"Illegal format";
            return -1;
         }
         if( first ) {
            arr = indgen(k2-k1+1) + k1 - 1;
            first = 0;
         } else {
            // arr = [arr,indgen(k2-k1+1) + k1];
            grow, arr, indgen(k2-k1+1) + k1 - 1;
         }
      } else {
         if( first ) {
            arr = toint(tok(i));
            first = 0;
         } else {
            // arr = [arr,toint(tok(i))];
            grow, arr, toint(tok(i));
         }
      }

   }

   return arr;

}

/* Function outlier */

func outlier( arr, qlimit )
/* DOCUMENT is = outlier( arr, qlimit)

   returns the indices of the elements in array 'arr'
   where abs(arr(i)) > qlimit * abs(avg(arr(except i)))

   2009-06-05/NJW
*/
{
   n = numberof(arr);
   is = [];
   for(i=1;i<=n;i++) {
      if( abs(arr(i)) > qlimit * abs(avg(rem_elem(arr,i))) ) grow, is, i;
   }
   return is;
}

/* Function strdelcom */

func strdelcom( str, repl=, mode= )
/* DOCUMENT newstr = strdelcom( str, repl=, mode= )

   Returns string of same length as 'str' but with all
   contents between pair of "" or '' reset to spaces.

   abc"def"gh'ijkl'mn  -> abc"   "gh'    'mn

   Keywords:  repl   (char) replacement character (default is space)
              mode   if set then quotes will be replaced as well

   2009-06-25/NJW
*/
{
   len = strlen( str );
   newstr = str;
   flag = array(int,len);
   mode = mode ? 0 : 1;
   if( is_void(repl) ) repl = ' ';

   // locate double quotes
   pdqarr = [];
   pdq = 0;
   while( (pdq = strpos(str,"\"",pdq+1))) grow,pdqarr,pdq;
   num_pdq = numberof(pdqarr);

   // locate single quotes
   psqarr = [];
   psq = 0;
   while( (psq = strpos(str,"'",psq+1))) grow,psqarr,psq;
   num_psq = numberof(psqarr);

   if( num_pdq+num_pdq == 0 ) return str; // no quotes, so no action
   if( num_pdq%2 == 1 ) {
      write,"STRDELCOM Warning: Unpaired double quotes - no action";
      return str;
   }
   if( num_psq%2 == 1 ) {
      write,"STRDELCOM Warning: Unpaired single quotes - no action";
      return str;
   }

   for( i = 1; i <= num_pdq; i += 2 ) {
      if( pdqarr(i+1)-mode >= pdqarr(i)+mode ) \
         flag(indgen(pdqarr(i)+mode:pdqarr(i+1)-mode)) += 2;
   }
   for( i = 1; i <= num_psq; i += 2 ) {
      if( psqarr(i+1)-mode >= psqarr(i)+mode ) \
         flag(indgen(psqarr(i)+mode:psqarr(i+1)-mode)) += 1;
   }

   if( anyof(flag==3) ) {
      write,"STRDELCOM Warning: Intertwined quotes - no action";
      return str;
   }

   cstr = *pointer(str);

   w = where(flag);
   if( numberof(w) ) cstr(w) = repl;

   return string(&cstr);
}

/* Function cent2bds */

func cent2bds( arr, &bdslo, &bdshi, lg= )
/* DOCUMENT bds = cent2bds( arr, >bdslo, >bdshi, lg= )

 Returns array with boundaries where the "arr" values are
 centers in the bins.

 Keyword 'lg' for logarithmical spacing

  2009-06-29/NJW  based on cent2bds.pro

*/
{
   n = numberof(arr);
   res = array(double,n+1);

   if( lg ) {
      res(2:-1) = sqrt(arr(1:-1) * arr(2:0));
      res(1) = res(2) / arr(2) * arr(1);
      res(0) = res(-1) * arr(0) / arr(-1);
   } else {
      res(2:-1) = 0.5*(arr(1:-1) + arr(2:0));
      res(1) = res(2) - arr(2) + arr(1);
      res(0) = res(-1) + arr(0) - arr(-1);
   }

   bdslo = res(1:-1);
   bdshi = res(2:0);

   return res;

}

/* Function linscale */

func linscale( a, b, anew, bnew, &p, &q )
/* DOCUMENT linscale, a, b, anew, bnew, >p, >q

    Linear rescaling so that a -> anew, b -> bnew
      by (anew =) a*p + q etc.
    (linear transformation)

   2009-10-30/NJW
*/
{
   if( a == b ) error,"LINSCALE bad input";
   ba = double(b-a);
   p = (bnew - anew)/ba;
   q = (b*anew - a*bnew)/ba;
}

/* Function shd_upd */

func shd_upd( shd, idx )
/* DOCUMENT new_shd = shd_upd( shd, idx )

     where 'shd' is an array to be updated (increased by 1) at
     positions (indices) as given by 'idx'.

     Takes into account that indices may be repeated.

     2009-11-03/NJW
*/
{

   shdw = shd;
   idxw = idx(sort(idx));

   do {
      dxi = shift(idxw,1);
      d = shift(dxi - idxw,-1);
      d(1) = 1;
      // unique elements are now: where(d)
      shdw(idxw)++; // operates only once on each array element
      w = where(d==0); // repeated elements
      nw = numberof(w);
      if( nw ) idxw = idxw(w);
   } while( nw );

   return shdw;

}

/* Function str_erase_between_symbols */

func str_erase_between_symbols( line, symbol )
/* DOCUMENT new_line = str_erase_between_symbols( line, symbol )

    'symbol' can be a character or a string (only the first
    character of the string is considered).
    Only erasure between matching pairs.

   2010-06-09/NJW
*/
{
   lenline = strlen(line);
   cline = *pointer(line);
   if( structof(symbol) == string ) symbol = (*pointer(symbol))(1);
   n = 0;
   p = [];
   for( i = 1; i <= lenline; i++ ) {
      if( cline(i) == symbol ) {
         grow,p,i;
         n++;
      }
   }
   if( n == 0 || n%2 == 1 ) return line;
   for(j = 1; j < n; j += 2 ) {
      for( i = p(j); i <= p(j+1); i++ ) cline(i) = ' ';
   }
   return string(&cline);
}

/* Function str_get_words */

func str_get_words( str, skip= )
/* DOCUMENT res = str_get_words( str, skip= )

   Returns an array of strings holding each word from the
   input string 'str' (must be a scalar string).

   Keyword 'skip' can be set to a character and everything
                  between such a pair will be erased before
                  words are extracted.
   2010-06-09/NJW
*/
{
   if( skip ) str = str_erase_between_symbols( str, skip );
   //+ sel = strword( str,"- \n\t.,:;|<>()%@$?[]{}'\\!=#\"+&*/" ,100);
   sel = strword( str, "^_A-Za-z0-9", 100);
/*
 *   You might consider using
 *     delim = "^_A-Za-z0-9"
 */
   if( sel(2) < 0 ) return []; // indicating no words found
   words = strpart( str, sel );
   words = words(where(words));
   // eliminate words that consist of only digits
   dig = is_digit(words);
   words = words(where(!dig));
   return words;
}

/* Function rm_slashcom */

func rm_slashcom( filename )
/* DOCUMENT new_text = rm_slashcom( filename )

   Returns the file content but cleaned for
   double slash comments.

  2010-06-09/NJW
*/
{
   text = rdfile( filename );
   ntext = numberof(text);

   for( i = 1; i <= ntext; i++ ) {
      line = str_erase_between_symbols( text(i), '"' );
      if( is_void(line) ) continue;
      if( (pos = strpos(line,"//")) ) {
         text(i) = pos == 1 ? "" : strpart(text(i),1:pos-1);
      } else continue;
   }
   return text;
}

/* Function parpath */

func parpath( rootname )
/* DOCUMENT par_file_name = parpath( rootname )

   Returns name of .par file from PFILES
*/
{
   ostype = get_env("OSTYPE");
   if( strlen(ostype) == 0 ) { // neither Linux nor Unix
      pfiles = "."; // *.par file must be in current directory
   } else pfiles = strsplit(get_env("PFILES"),";")(1);
   ss = strsplit(pfiles,":");
   nss = numberof(ss);
   for( i = 1; i <= nss; i++ ) {
      fname = ss(i)+"/"+rootname+".par";
      if( file_test(fname) ) return fname;
   }
   return [];
}

/* Function sigdig */

func sigdig( x, n )
/* DOCUMENT xd = sigdig( x, n )

    returns 'x' with 'n' digits.

    Ex: > sigdig( 12.345, 2)
        12
        > sigdig(0.012345, 3)
        0.0123
*/
{
   if( double(x) == 0.0 ) return 0.0;
   ax = abs(double(x));
   sGn = x < 0 ? -1 : 1;
   f = 10.^(-floor(log10(ax)));
   g = 10^(n-1);
   y = long(ax*f*g+0.5)/(f*g);
   return sGn * y;
}

/* Function arr_info */

struct s_Arr_info {
   string type;
   long n;
   long dms(10);
   double mn;
   double mx;
   double av;
   double rs;
}
func arr_info( arr )
/* DOCUMENT arr_info, arr

   Function: returns struct
   Subroutine: prints results

   Type                 s.type
   Number of elements   s.n
   Dimensions           s.dms(10)
   Minimum, maximum     s.mn, s.mx
   Average, RMS         s.av, s.rs
*/
{
   t = typeof(arr);
   dms = dimsof(arr);
   if( am_subroutine() ) {
      write,format="%s: ","Dimensions";
      for(i=1; i <= dms(1); i++) {
         write,format="%i", dms(i+1);
         if( i < dms(1) ) write,format="%s"," x ";
         else write,"";
      }
      write,format="%i elements of type %s\n", numberof(arr), t;
      if( t == "long" || t == "int" || t == "short" ) {
         write,format="Min: %i, Max: %i\n", min(arr), max(arr);
      } else if( t == "double" || t == "float" ) {
         write,format="Min: %.6g, Max: %.6g\n", min(arr), max(arr);
      }
      write,format="Avg: %.6g, Rms: %.6g\n", wavg(arr), wrms(arr);
   } else {
      s = s_Arr_info();
      s.type = t;
      s.dms(1:dms(1)+1) = dms;
      s.n = numberof(arr);
      s.mn = double(min(arr));
      s.mx = double(max(arr));
      s.av = double(wavg(arr));
      s.rs = double(wrms(arr));
   }
   return s;
}

/* Function interp2 */

func interp2( zorig, xorig, yorig, x, y )
/* DOCUMENT z = interp2( zorig, xorig, yorig, x, y )

    Returns the bi-linearly interpolated value where
       zorig  is a MxN array of z-values
       xorig  is a M dim array of x values
       yorig  is a N dim array of y values
       x,y    is where the z value is wanted
   Outside of the area spanned by xorig,yorig the edge values of zorig
   are returned.

   SEE ALSO: interp (Yorick function), interpl (log interpolation, this file).

   2011-01-02/NJW
*/
{
   if( !is_scalar(x) ) error,"x must be a scalar";
   if( !is_scalar(y) ) error,"y must be a scalar";

   dmsz = dimsof( zorig );
   dmsx = dimsof( xorig );
   dmsy = dimsof( yorig );

   M = dmsx(2); // number of elements in xorig array
   N = dmsy(2); // number of elements in yorig array

   if( dmsz(1) != 2 ) error,"zorig must be a 2D array";
   if( dmsx(1) != 1 ) error,"xorig must be a 1D array";
   if( dmsy(1) != 1 ) error,"yorig must be a 1D array";
   if( dmsz(2) != M ) error,"Dim of xorig does no match first dim of zorig";
   if( dmsz(3) != N ) error,"Dim of yorig does no match second dim of zorig";

   wx = where( x > xorig );
   if( numberof(wx) == 0 ) { // x is 'too small'
      return interp( zorig(1,), yorig, y );
   } else if( numberof(wx) == M ) { // x is 'too large'
      return interp( zorig(0,), yorig, y );
   } else { // x is inside covered interval
      i = wx(0);
      wy = where( y > yorig );
      if( numberof(wy) == 0 ) { // y is 'too small'
         return interp( zorig(,1), xorig, x );
      } else if( numberof(wy) == N ) { // y is 'too large'
         return interp( zorig(,0), xorig, x );
      } else { // y is inside covered interval
         j = wy(0);
         f = (x - xorig(i))/(xorig(i+1)-xorig(i));
         za = zorig(i,j) + f*(zorig(i+1,j) - zorig(i,j));
         zb = zorig(i,j+1) + f*(zorig(i+1,j+1) - zorig(i,j+1));
         return za + (y - yorig(j))*(zb - za)/(yorig(j+1)-yorig(j));
      }
   }
}

/* Function interpl */

func interpl( v, vx, x )
/* DOCUMENT vals = interpl( v, vx, x )

    Returns an array of logarithmically interpolated values
    hence none of the argument arrays may contain neither zero nor negative values.

    SEE ALSO: interp (Yorick function).

*/
{
   if( anyof(v <= 0.0) ) error,"Unexpected value (v) <= 0";
   if( anyof(vx <= 0.0) ) error,"Unexpected value (vx) <= 0";
   if( anyof(x <= 0.0) ) error,"Unexpected value (x) <= 0";

   return exp(interp(log(v),log(vx), log(x)));
}

/* Function arr_fill_in */

func arr_fill_in( x, val=, epsi= )
/* DOCUMENT res = arr_fill_in( x, val=, epsi= )

   Returns an array (1D like 'x') filled in with interpolated
   values around elements with value near 'val' (default -99.)
   with maximal deviation 'epsi' (default 0.1). It wraps around.

   2011-02-01/NJW
*/
{
   n = numberof(x);
   xout = x;
   if( is_void(val) ) val = -99.0;
   if( is_void(epsi) ) epsi = 0.1;

   for( i = 1; i <= n; i++ ) {

      if( !near( xout(i), val, epsi ) ) continue;
      // locate left anchor
      i1 = i;
      while( near( xout(i1), val, epsi) ) i1--;
      // i1 may be <= 0

      // locate right anchor
      i2 = i;
      while( near( xout(i2%n), val, epsi) ) i2++;
      // i2 may be > n

      // Fill in
      range = indgen(i1:i2)%n;
      nrange = numberof(range);
      w = where( range <= 0 );
      if( numberof(w) ) range(w) += n;
      xout(range(2:nrange-1)) = interp([xout(i1),xout(i2%n)],[i1,i2],indgen(i1+1:i2-1));
   }
   return xout;
}

/* Function xjour */

func xjour( strngs, see=, ed= )
/* DOCUMENT xjour, strngs, see=, ed=

    Will write the content of 'strngs' (scalar or string
    array) to a file named 'journal_YYMMDD' while appending.
    This is intended for saving Yorick commands etc. for future
    reference and perhaps macro making.

    Since '"' is an annoying character that should be escaped
    if can be replaced with '@' that is translated to '"'.

    Sorry, no way to save '@' in the string.

   Keywords:
      see    for inspecting the journal file so far
      ed     for editing the journal file
*/
{
   jname = "journal_"+ndate(1);
   if( typeof(strngs) == "string" ) {
      write_slist,jname,strstrrepl(strngs,"@","\""),app=1;
   }
   if( see ) {
      write,"Today's journal: "+jname;
      prstrarr, rdfile(jname);
   }
   if( ed ) vi,jname;
}

/* Function f2scienota */

func f2scienota( x, ndec= )
/* DOCUMENT  strngs = f2scienota( x, ndec= )

    Returns strings with the numbers 'x' in scientific notation
    suited for yorick plot annotations.

    Keyword: ndec   Number of decimals (defaults to 2)

    SEE ALSO: ftoa, atof, wrdig (in basic.i)
*/
{
   if( is_scalar(x) ) return _f2scienota( x, ndec=ndec );
   n = numberof(x);
   scistr = array(string,n);
   for( i = 1; i <= n; i++ ) scistr(i) = _f2scienota( x(i), ndec=ndec );
   return scistr;
}

func _f2scienota( x, ndec= )
{
   if( x == 0 ) return "0";
   if( is_void(ndec) ) ndec = 3;
   s = " ";
   if( x < 0 ) {
      s = "-";
      x = -x;
   }
   lx = log10(x);
   exponent = long(floor(lx));
   mantissa = x * 10.^(-exponent);
   fmt = swrite(format="%%s%%%i.%if 10^%%i^", ndec+2, ndec);
   //+ write,"fmt : >"+fmt+"<";
   scistr = swrite(format=fmt, s, mantissa, exponent);
   //+ write,"scistr : >"+p1+"<";
   return scistr;
}

/* Function ranstr */

func ranstr( len, uc=, lc=, n= )
/* DOCUMENT str = ranstr( len, uc=, lc=, n= )

    Returns a random string of length 'len' using
    characters A-Z,a-z,0-9

    Keywords:
       uc      Only upper letters are used (i.e. no digits)
       lc      Only lower letters are used (i.e. no digits)
       n       Only numbers are used
*/
{
   if( uc ) {
      return string(&char('A'+indgen(0:25))(long(floor(random(len)*26)+1)));
   } else if( lc ) {
      return string(&char('a'+indgen(0:25))(long(floor(random(len)*26)+1)));
   } else if( n ) {
      return string(&char('0'+indgen(0:9))(long(floor(random(len)*10)+1)));
   } else {
      basket = array( char, 62);
      basket(1:26) = 'A'-1+indgen(26);
      basket(27:52) = 'a'-1+indgen(26);
      basket(53:62) = '0'-1+indgen(10);
      return string(&basket(long(floor(random(len)*62)+1)));
   }
}

/* Function datagroups */

func datagroups( arr, spacelimit )
/* DOCUMENT idx = datagroups( t, spacelimit )

   A function to analyze typically a timeseries (t) and
   return indices of the groups of elements.

   't' is assumed to be sorted in increasing order.

   The groups are identified by a certain amount of
   empty space between them.

   An example:
      t = [1.,1.2,1.5,1.7,4.,4.2,4.5, 5.1,5.2]
      datagroups(t, 2.) will return [1, 5]
      datagroups(t, 0.5 ) will return [1, 5, 8]

      Obtain the lengths of the groups by:
        g = datagroups(t, 0.5);
        gg = g;
        grow, gg, numberof(t)+1;
        len = gg(dif);

      Indices of group N are:
        idx = indgen(gg(N):gg(N+1)-1);

    Note the resemblance with the function 'uniq'

    2011-05-27/NJW
*/
{
   if( numberof( (w = where(arr(dif) > spacelimit) ) ) ) {
      return grow(1, w+1);
   } else return [1];
}

/* Function fold_gaussy */

func fold_gaussy( e, f, sigma )
/* DOCUMENT fnew = fold_gaussy( [e,] f, sigma )

  Fold a spectrum or a number of spectra with a Gauss function with a non-constant sigma
  The folding is done along the first dimension.

   Call -  e :  1D energy array
                If 'e' is absent it is assumed to be indgen(dimsof(f)(2))
           f : array with spectrum, first dimension must have same size as 'e'
         sigma  : scalar with standard deviation,
                  or array with same size as 'e'

   Return : folded spectrum, conformable with 'f'.

    2012-11-06/NJW based on 'fold_gaussx'

   SEE ALSO: fold_gaussx

   fold_gaussy is slightly slower than fold_gaussx when operating on a
   1D spectrum.

*/

{
   time_keeper = elapsed_time = array(double,3);
   timer, time_keeper;

   if( is_void(sigma) ) {
      // There are at most two arguments
      sigma = f;
      f = e;
      fdms = dimsof(f);
      e = double(indgen(fdms(2)));
   }
   fdms = dimsof(f);

   n = numberof(e); // == fdms(2)
   ee = double( e );
   ns = numberof(sigma);
   if( fdms(2) != n ) {
      write,"mismatching dimensions of 'e' and 'f'\n";
      return -1;
   }
   if( ns != 1 && ns != n ) {
      write,"'sigma' has illegal dimension\n";
      return -1;
   }
   sigma2 = sigma * sqrt(2.);
   if( ns == 1 ) sigma2 = sigma2(-:1:n);

   // Define channel boundaries

   eb = 0.5*(ee + shift(ee,-1));
   eb(1) = (3*ee(1) - ee(2))/2;
   grow,eb,(3*ee(0) - ee(-1))/2;

   w = array(double,n,n); // 'redistribution matrix'
   for( i = 1; i <= n; i++ ) {
      j = where((e > ee(i)-3.5*sigma2(i)) + (e < ee(i)+3.5*sigma2(i)) == 2);
      if( numberof(j) ) {
         x2 = (eb(j+1) - ee(i))/sigma2(i);
         x1 = (eb(j) - ee(i))/sigma2(i);
         wei = erf(x2) - erf(x1);
         w(i,j) = wei / wei(sum);
      }
   }

   g = w(,+) * f(+,..);

   timer, time_keeper, elapsed_time;
   write,"CPU time used: "+ftoa(elapsed_time(1),ndec=3)+" s";

   return g;

}
 /* Function message_board */

func message_board( text, pane= )
/* DOCUMENT message_board, text, pane=

     Displays the text 'text' (scalar string) in a
     plot window define by 'pane' (defaults to 7).

*/
{
   if( is_void(pane) ) pane = 7;
   prev_win = window();
   window,pane,style="nobox.gs";
   xyouts,0.5,0.58,ndate(3),charsize=1.0,vir=1,align=0.5;
   xyouts,0.5,0.42,text,charsize=3.,vir=1,align=0.5,font="times",color=blue;
}

/* Function distfix */

func distfix( ncounts, npix )
/* DOCUMENT res = distfix( ncounts, npix )

    Returns a (pseudo-) poisson distribution of
    precisely 'ncounts' in 'npix' pixels.

   2013-06-11/NJW
*/
{
   // two methods are offered, choose the faster

   res = array(long,npix);
   // choose indices at random; the 'idx'
   // array has ncounts elements
   idx = random_i(npix,ncounts);

   if( ncounts < 10*npix ) { // add one by one
      //+ tic,1;
      for( i = 1; i <= ncounts; i++ ) res(idx(i))++;
      //+ write,"Time spent: ",tac(1);
   } else { // add up in bunches
      //+ tic,1;
      idx = idx(sort(idx));
      u = uniq(idx);
      uu = grow( u, ncounts+1);
      res(idx(u)) = uu(dif);
      //+ write,"Time spent (2): ",tac(1);
   }
   return res;
}

/* Function bin_overlap */

func bin_overlap( a, b, e1, e2, &index, &fraction )
/* DOCUMENT num = bin_overlap( a, b, e1, e2, >index, >fraction )

    The interval under consideration is [a,b[ and
    e1 is an array of lower boundaries and e2 is the
    equivalent array of upper boundaries.

    Assumes a < b

    Returns the number of bins that overlap with [a,b[ and
    their indices and the fractions.
*/
{
   // There are various possibilities: 
   //  e1(i) < a && e2(i) >= b      interval [a,b[ is completely inside [e1(i),e2(i)]
   //  a > max(e2)     no interaction
   //  b < max(e1)     no interaction
   //  one or more of e1 is > a && <= b
   //  one or more of e2 is > a && <= b

   if( a > max(e2) ) return 0;
   if( b < min(e1) ) return 0;

   // check for the case where [a,b[ is completely embedded in a e1e2 interval

   n = numberof( (w = where( e1 <= a & e2 >= b )));
   if( n ) {
      index = w(1);
      fraction = (b - a) / (e2(index) - e1(index));
      return n;
   }

   i = where( e1 >= a & e1 <= b ); // all intervals that begin inside [a,b[
   j = where( e2 >= a & e2 <= b ); // all intervals that end inside [a,b[
   if( numberof(i)+numberof(j) == 0 ) return 0; // can happen if e1e2 are disjoint
   if( !numberof(i) ) i = [];
   if( !numberof(j) ) j = [];
   k = grow(i,j);
   k = k(sort(k));
   k = k(uniq(k));
   nk = numberof(k);
   index = fraction = [];
   for( kk = 1; kk <= nk; kk++ ) {
      nwi = numberof((wi = where( k(kk) == i )));
      nwj = numberof((wj = where( k(kk) == j )));
      if( nwi && nwj ) { // both begin and end of e-interval is inside [a,b[
         grow, index, k(kk);
         grow, fraction, 1.;
      } else if( nwi && !nwj ) { // e-interval begins inside [a,b[
         grow, index, k(kk);
         grow, fraction, (b - e1(k(kk)))/(e2(k(kk)) - e1(k(kk)));
      } else if( !nwi && nwj ) { // e_interval ends inside [a,b[
         grow, index, k(kk);
         grow, fraction, (e2(k(kk)) - a)/(e2(k(kk)) - e1(k(kk)));
      } else {
         error,"Ups, should not come here";
      }
   }
   return numberof(index);
}

/* Function centroid */

func centroid( arr )
/* DOCUMENT res = centroid( arr )

     Returns the centroid of 'arr' that may be a scalar (kind of stupid),
     a 1D array or a 2D array. Higher dimensions are not supported yet.
     For a 2D array a two_element vector is returned.

*/
{
   dms = dimsof(arr);
   if( dms(1) == 0 ) return arr;
   if( dms(1) == 1 ) return sum(double(arr)*indgen(dms(2))) / sum(arr);
   if( dms(1) == 2 ) return [sum(arr(,sum)*double(indgen(dms(2)))), \
                             sum(arr(sum,)*double(indgen(dms(3))))] / sum(arr);
   if( dms(1) > 2 ) error,"Centroide: not implemented yet";
}

/* Function determinant */

func determinant( arr )
/* DOCUMENT res = determinant( arr )

    Returns the determinant of the square matrix 'arr'

    2011-03-02/NJW
*/
{
   dms = dimsof(arr);
   if( dms(1) != 2 ) {
      write,"Must be a 2D array";
      return [];
   }
   if( dms(2) != dms(3) ) {
      write,"Must be a square array";
      return [];
   }
   n = dms(2);
   if( n == 1 ) return arr(1);
   if( n == 2 ) return arr(1,1)*arr(2,2) - arr(1,2)*arr(2,1);
   if( n > 2 ) {
      // resolve according to first row:
      dsum = 0.;
      ssign = 1;
      for( j = 1; j <= n; j++ ) {
         jarr = rem_elem(indgen(n),j);
         darr = arr(2:n,jarr);
         dsum += ssign*arr(1,j)*determinant(darr);
         ssign = -ssign;
      }
      return dsum;
   }
}

/* Function insidepolygon */

func insidepolygon( x, y, px, py )
/* DOCUMENT res = insidepolygon( x, y, px, py )

  Is a point inside polygon or not?

  From Coyote's guide to IDL programming.

   x - The x coordinate of the point.
   y - The y coordinate of the point.
  px - The x coordinates of the unclosed polygon.
  py - The y coordinates of the unclosed polygon.
 
  The return value of the function is 1 if the point is inside the
  polygon and 0 if it is outside the polygon.
*/
{
   sx = dimsof(px);
   sy = dimsof(py);
   if( !is_vector(px) ) error,"##1## px must be a vector";
   if( !is_vector(py) ) error,"##2## py must be a vector";
   if( anyof(sx-sy) ) error,"##3## incompatible dimensions";
   n = numberof(px);

   tmp_px = grow(px,px(1)); // Close Polygon in x
   tmp_py = grow(py,py(1)); // Close Polygon in y

  /* ---
   * ---  Computes all angles between apex(i+1) and apex(i)
   * ---  as measured from (x,y).
   * ---  When (x,y) is inside the polygon the sum must
   * ---  exceed PI.
   * ---  */
   i = indgen(n); // Counter (1:n)
   ip = i+1; // Counter (2:n+1)

   x1 = tmp_px(i)  - x;
   y1 = tmp_py(i)  - y;
   x2 = tmp_px(ip) - x;
   y2 = tmp_py(ip) - y;

   dp = x1*x2 + y1*y2; // Dot-product    |a| |b| cos(theta)
   cp = x1*y2 - y1*x2; // Cross-product  |a| |b| sin(theta)
   theta = atan(cp,dp);

   return abs(sum(theta)) > pi ? 1 : 0;
}

/* Function areatriangle */

func areatriangle( px, py )
{
   ppx = (px-px(1))(2:3);
   ppy = (py-py(1))(2:3);
   return 0.5 * ( ppx(1)*ppy(2) - ppx(2)*ppy(1) );
}

/* Function areapolygon */

func areapolygon( px, py )
{
   // Choose 1. apex as origin for triangles
   // Implies 2 triangles fewer than in 'areapolygon1' but
   //  the numerical instability might be worse.

   n = numberof(px); // consists of n-2 triangles
   if( n < 3 ) return 0.0;
   if( n == 3 ) return areatriangle(px,py);
   area = 0.0;
   for( i = 2; i < n; i++ ) {
      area += areatriangle( px([1,i,i+1]), py([1,i,i+1]) );
   }
   return area;
}

/* Function areapolygon1 */

func areapolygon1( px, py )
{
   n = numberof(px); // consists of n-2 triangles
   if( n < 3 ) return 0.0;
   if( n == 3 ) return areatriangle(px,py);
   // choose CM as origin
   cmx = avg(px);
   cmy = avg(py);
   // Close the polygon
   pxx = grow(px,px(1));
   pyy = grow(py,py(1));
   area = 0.0;
   for( i = 1; i <= n; i++ ) {
      area += areatriangle( grow(cmx,pxx(i:i+1)), grow(cmy,pyy(i:i+1)) );
   }
   return area;
}
