extern mem_storagedoc;
/* DOCUMENT  Package for data storage in memory

    e.g. when repeated data reading is performed
    Operates on 1D arrays of datatypes: int, long, float, double, string.

      mem_save    Saving an array
      mem_restore Restoring an array
      mem_reset   Erase allocated memory
      mem_info    Show info on memory contents

*/

/* Function mem_save */

func mem_save( data_id, arr, silent= )
/* DOCUMENT status = mem_save( data_id, arr, silent= )

   Saves the 1D array 'arr' to memory with the data identification
   name 'data_id'. It can be a filename plus a column id or a DOL etc.
   Accepted data types: int, long, float, double, string.

   Retrieval by function mem_restore.

   Returned status value:

          0  : operation went well
         -1  : duplicated 'data_id', skip saving
         -2  : wrong dimensionality of 'arr'

   See also mem_reset, mem_restore, mem_info

   2007-01-08/NJW
*/
{
   extern Didimsav, Idximsav, Numimsav, Datimsav;
   extern Didjmsav, Idxjmsav, Numjmsav, Datjmsav;
   extern Didfmsav, Idxfmsav, Numfmsav, Datfmsav;
   extern Diddmsav, Idxdmsav, Numdmsav, Datdmsav;
   extern Didsmsav, Idxsmsav, Numsmsav, Datsmsav;

   if( is_void(silent) ) silent = 0;

   dms = dimsof( arr );
   if( dms(1) != 1 ) {
      write,"Array must be 1D - skip";
      return -2;
   }
   dtype = typeof( arr );

// Normal Integers

   if( dtype == "int" ) {
      w = where( Didimsav == data_id );
      if( numberof(w) >= 1 ) {
         if( !silent ) write,data_id+" (int) exists already - skip";
         return -1;
      }
      grow, Datimsav, arr;
      grow, Didimsav, data_id;
      if( is_void(Idximsav) ) {
         Idximsav = [1];
      } else {
         grow, Idximsav, Idximsav(0)+Numimsav(0);
      }
      grow, Numimsav, dms(2);
      return 0;

// Long Integers

   } else if( dtype == "long" ) {
      w = where( Didjmsav == data_id );
      if( numberof(w) >= 1 ) {
         if( !silent ) write,data_id+" (long) exists already - skip";
         return -1;
      }
      grow, Datjmsav, arr;
      grow, Didjmsav, data_id;
      if( is_void(Idxjmsav) ) {
         Idxjmsav = [1];
      } else {
         grow, Idxjmsav, Idxjmsav(0)+Numjmsav(0);
      }
      grow, Numjmsav, dms(2);
      return 0;

// Floating point

   } else if( dtype == "float" ) {
      w = where( Didfmsav == data_id );
      if( numberof(w) >= 1 ) {
         if( !silent ) write,data_id+" (float) exists already - skip";
         return -1;
      }
      grow, Datfmsav, arr;
      grow, Didfmsav, data_id;
      if( is_void(Idxfmsav) ) {
         Idxfmsav = [1];
      } else {
         grow, Idxfmsav, Idxfmsav(0)+Numfmsav(0);
      }
      grow, Numfmsav, dms(2);
      return 0;

// Double precision

   } else if( dtype == "double" ) {
      w = where( Diddmsav == data_id );
      if( numberof(w) >= 1 ) {
         if( !silent ) write,data_id+" (double) exists already - skip";
         return -1;
      }
      grow, Datdmsav, arr;
      grow, Diddmsav, data_id;
      if( is_void(Idxdmsav) ) {
         Idxdmsav = [1];
      } else {
         grow, Idxdmsav, Idxdmsav(0)+Numdmsav(0);
      }
      grow, Numdmsav, dms(2);
      return 0;

// Strings

   } else if( dtype == "string" ) {
      w = where( Didsmsav == data_id );
      if( numberof(w) >= 1 ) {
         if( !silent ) write,data_id+" (string) exists already - skip";
         return -1;
      }
      grow, Datsmsav, arr;
      grow, Didsmsav, data_id;
      if( is_void(Idxsmsav) ) {
         Idxsmsav = [1];
      } else {
         grow, Idxsmsav, Idxsmsav(0)+Numsmsav(0);
      }
      grow, Numsmsav, dms(2);
      return 0;


   } else {
      write,"Unsupported data type";
      return -3;
   }
}

/* Function mem_reset */

func mem_reset( void )
/* DOCUMENT status = mem_reset();
      or    mem_reset;

   Erases all the variables used for data storage in momory.

   See also mem_save, mem_restore, mem_info

   2007-01-08/NJW
*/
{
   extern Didimsav, Idximsav, Numimsav, Datimsav;
   extern Didjmsav, Idxjmsav, Numjmsav, Datjmsav;
   extern Didfmsav, Idxfmsav, Numfmsav, Datfmsav;
   extern Diddmsav, Idxdmsav, Numdmsav, Datdmsav;
   extern Didsmsav, Idxsmsav, Numsmsav, Datsmsav;
   Didimsav = []; Idximsav = []; Numimsav = []; Datimsav = [];
   Didjmsav = []; Idxjmsav = []; Numjmsav = []; Datjmsav = [];
   Didfmsav = []; Idxfmsav = []; Numfmsav = []; Datfmsav = [];
   Diddmsav = []; Idxdmsav = []; Numdmsav = []; Datdmsav = [];
   Didsmsav = []; Idxsmsav = []; Numsmsav = []; Datsmsav = [];
   return 0;
}

/* Function mem_restore */

func mem_restore( data_id, &arr )
/* DOCUMENT status = mem_restore( data_id, arr )

   Restores the 1D array 'arr' previously saved in memory under the
   name 'data_id' by the function 'mem_save'.
   Accepted data types: int, long, float, double, string.

   Returned status value:

          0  : operation went well
         -1  : found no match with 'data_id', skip restoring

   See also mem_reset, mem_save, mem_info

   2007-01-08/NJW
*/
{
   extern Didimsav, Idximsav, Numimsav, Datimsav;
   extern Didjmsav, Idxjmsav, Numjmsav, Datjmsav;
   extern Didfmsav, Idxfmsav, Numfmsav, Datfmsav;
   extern Diddmsav, Idxdmsav, Numdmsav, Datdmsav;
   extern Didsmsav, Idxsmsav, Numsmsav, Datsmsav;

   if( !is_void(Didimsav) ) {
      w = where( data_id == Didimsav );
      if( numberof(w) > 0 ) {
         i = w(0);
         arr = Datimsav(Idximsav(i):Idximsav(i)+Numimsav(i)-1);
         return 0;
      }
   }

   if( !is_void(Didjmsav) ) {
      w = where( data_id == Didjmsav );
      if( numberof(w) > 0 ) {
         i = w(0);
         arr = Datjmsav(Idxjmsav(i):Idxjmsav(i)+Numjmsav(i)-1);
         return 0;
      }
   }
   
   if( !is_void(Didfmsav) ) {
      w = where( data_id == Didfmsav );
      if( numberof(w) > 0 ) {
         i = w(0);
         arr = Datfmsav(Idxfmsav(i):Idxfmsav(i)+Numfmsav(i)-1);
         return 0;
      }
   }
   
   if( !is_void(Diddmsav) ) {
      w = where( data_id == Diddmsav );
      if( numberof(w) > 0 ) {
         i = w(0);
         arr = Datdmsav(Idxdmsav(i):Idxdmsav(i)+Numdmsav(i)-1);
         return 0;
      }
   }
   
   if( !is_void(Didsmsav) ) {
      w = where( data_id == Didsmsav );
      if( numberof(w) > 0 ) {
         i = w(0);
         arr = Datsmsav(Idxsmsav(i):Idxsmsav(i)+Numsmsav(i)-1);
         return 0;
      }
   }

   return -1;

}

/* Function mem_info */

func mem_info(a,chat=)
/* DOCUMENT mem_info,chat=

   Gives an overview of memory contents

   See also mem_save, mem_restore, mem_reset

   2007-01-08/NJW
*/
{

   extern Didimsav, Idximsav, Numimsav, Datimsav;
   extern Didjmsav, Idxjmsav, Numjmsav, Datjmsav;
   extern Didfmsav, Idxfmsav, Numfmsav, Datfmsav;
   extern Diddmsav, Idxdmsav, Numdmsav, Datdmsav;
   extern Didsmsav, Idxsmsav, Numsmsav, Datsmsav;

   if( is_void(chat) ) chat = 0;
   if( !is_void(Didimsav) ) {
      num = numberof(Didimsav);
      plu = num > 1 ? "s" : "";
      write,format="%i int item%s\n", num, plu;
      if( chat == 1 ) {
         for(i=1;i<=num;i++) {
            write,format="%i elements in %s\n", Numimsav(i), Didimsav(i);
         }
      }
      if( chat > 1 ) {
         n = chat > num ? num : chat;
         for(i=1;i<=n;i++) {
            write,format="%s(%i): ", Didimsav(i), Numimsav(i);
            m = chat > Numimsav(i) ? Numimsav(i) : chat;
            for(j=1;j<=m;j++) write,format="%i ", Datimsav(Idximsav(i)+j-1);
            write,"";
         }
      }
   }
   if( !is_void(Didjmsav) ) {
      num = numberof(Didjmsav);
      plu = num > 1 ? "s" : "";
      write,format="%i long item%s\n", num, plu;
      if( chat == 1 ) {
         for(i=1;i<=num;i++) {
            write,format="%i elements in %s\n", Numjmsav(i), Didjmsav(i);
         }
      }
      if( chat > 1 ) {
         n = chat > num ? num : chat;
         for(i=1;i<=n;i++) {
            write,format="%s(%i): ", Didjmsav(i), Numjmsav(i);
            m = chat > Numjmsav(i) ? Numjmsav(i) : chat;
            for(j=1;j<=m;j++) write,format="%i ", Datjmsav(Idxjmsav(i)+j-1);
            write,"";
         }
      }
   }
   if( !is_void(Didfmsav) ) {
      num = numberof(Didfmsav);
      plu = num > 1 ? "s" : "";
      write,format="%i float item%s\n", num, plu;
      if( chat == 1 ) {
         for(i=1;i<=num;i++) {
            write,format="%i elements in %s\n", Numfmsav(i), Didfmsav(i);
         }
      }
      if( chat > 1 ) {
         n = chat > num ? num : chat;
         for(i=1;i<=n;i++) {
            write,format="%s(%i): ", Didfmsav(i), Numfmsav(i);
            m = chat > Numfmsav(i) ? Numfmsav(i) : chat;
            for(j=1;j<=m;j++) write,format="%f ", Datfmsav(Idxfmsav(i)+j-1);
            write,"";
         }
      }
   }
   if( !is_void(Diddmsav) ) {
      num = numberof(Diddmsav);
      plu = num > 1 ? "s" : "";
      write,format="%i double item%s\n", num, plu;
      if( chat == 1 ) {
         for(i=1;i<=num;i++) {
            write,format="%i elements in %s\n", Numdmsav(i), Diddmsav(i);
         }
      }
      if( chat > 0 ) {
         n = chat > num ? num : chat;
         for(i=1;i<=n;i++) {
            write,format="%s(%i): ", Diddmsav(i), Numdmsav(i);
            m = chat > Numdmsav(i) ? Numdmsav(i) : chat;
            for(j=1;j<=m;j++) write,format="%f ", Datdmsav(Idxdmsav(i)+j-1);
            write,"";
         }
      }
   }
   if( !is_void(Didsmsav) ) {
      num = numberof(Didsmsav);
      plu = num > 1 ? "s" : "";
      write,format="%i string item%s\n", num, plu;
      if( chat == 1 ) {
         for(i=1;i<=num;i++) {
            write,format="%i elements in %s\n", Numsmsav(i), Didsmsav(i);
         }
      }
      if( chat > 0 ) {
         n = chat > num ? num : chat;
         for(i=1;i<=n;i++) {
            write,format="%s(%i): ", Didsmsav(i), Numsmsav(i);
            m = chat > Numsmsav(i) ? Numsmsav(i) : chat;
            for(j=1;j<=m;j++) write,format="%s ", Datsmsav(Idxsmsav(i)+j-1);
            write,"";
         }
      }
   }
}
