/* Function mfitsdoc */

extern mfitsdoc;
/* DOCUMENT mfits ***************************************

   A set of funtions to facilitate the FITS interface with
   same names as the IDL private functions

   Note that the counting of the HDUs has origin in 1 for the
   Yorick functions but 0 for IDL like functions

   2006-02-07/NJW, latest update 2014-12-16/NJW

      addgrid           Adds a coordinate grid to an image
      check_colnames    Check column names in existing FITS bintable header
      fits_add_cols     Add columns to an existing bintable FITS extension
      fits_add_rows     Add rows to an existing bintable FITS extension
      fits_binsplit     Split a binary table into several files
      fits_bintable_peek  Extract value(s) from single table bin
      fits_bintable_poke  Update content of binary FITS table
      fits_btdes        Unpack FITS binary table description
      fits_ch_keyw      Change keyword value in place
      fits_coldump      Nice printing of column data
      fits_colname      Return column name(s) for given column number(s)
      fits_colnum       Return column number for given column name
      fits_copy_keys    Copy non-FITS specific keywords
      fits_copy_row     Copy one row to another one in a binary table
      fits_fh_verify    Updates keywords identifiers and removes empty ones
      fits_hdr2bintab   Save FITS header to a binary table
      fits_info_col     Display the column information
      fits_numcols      Get number of columns in binary table
      fselect           An emulation of the corresponding ftool
      fstrselect        fselect based on the 'strmatch' function
      fxpar             Get keyword value
      get_crd           Get coordinate definitions from header
      get_exten_no      Get filename and extension number from DOL
      get_image_dol     Locate first IMAGE extension in a FITS file
      get_wcs           Get WCS in a struct 's_Coords' from a FITS header
      gz_proxy_file     Define uncompressed file for reading
      headfits          Get FITS header for DOL
      im2skyfits        Save a 2D array as a sky image in a FITS file with WCS info.
      kwds_del          Delete a keyword from  KW-memory
      kwds_get          Get a keyword value from  KW-memory
      kwds_init         Erase all keywords from KW-memory
      kwds_put          Update FITS header with keywords from KW-memory
      kwds_set          Define a keyword in KW-memory
      kwds_setlongstr   Define a keyword with a long string in a sequence in KW-memory
      listhead          Display all header keywords
      nfits_extens      Get number of highest extension
      parse_expres      Resolve an expression
      pixsize           Find pixel sizes in image extension
      put_wcs           Update keywords with WCS information
      rdfitsbin         Return pointers to complete FITS binary table
      rdfitscol         Return a column from binary FITS table
      readfits          Returns an array from FITS IMAGE extension
      simple_peak       Finds number of counts around position (RA, Dec)
      skypos_fits       Finds celestial sky position from FITS header
      ttypes            Prints all column names in binary table
      wrfitsbin         Write binary table when pointer array is given
      writefits         Write FITS IMAGE extension
      wrmfitscols       Write multiple FITS binary table columns
      xnice             Find suitable floating point number for coordinate system

      _fits_var_length_info  Retrieve lengths and positions of
                             variable length records
      _tdims            Find existing TDIM keywords

      Version 1.1 as of 2012-09-27

***************************************************************
*/

//  Call randomize function because 'ranstr' is used
//  both in fits_ch_keyw and fits_bintable_poke

randomize;
write,"Function 'randomize' has been called from 'mfits.i'";

/* Function listhead */

func listhead( fh )
/* DOCUMENT listhead, fh
   or       list = listhead( fh )

   makes a list of all FITS header cards

   SEE ALSO:  fits_info

   2008-04-04/NJW
*/
{
   local cards; eq_nocopy, cards, _car(fh,1);

   if( am_subroutine() ) {
      for(i=1; i<=numberof(cards); i++) {
         write,cards(i);
      }
   } else {
      return cards;
   }
}

/* Function fits_hdr2bintab */

func fits_hdr2bintab( fh_or_dol, outfile, clobber= )
/* DOCUMENT fits_hdr2bintab, fh_or_dol, outfile, clobber=

    Converts a FITS header to a single column (80A)
    FITS binary table.

*/
{
   if( typeof(outfile) != "string" ) error,"Bad 2. argument";
   if( typeof(fh_or_dol) == "string" ) {
      local filename, extno;
      get_exten_no, fh_or_dol, filename, extno;
      if( !file_test(filename) ) error,filename+" was not found";
      fh = headfits( fh_or_dol );
      is_fh = 0;
   } else if( typeof(fh_or_dol) == "list" ) {
      fh = fh_or_dol;
      is_fh = 1;
   } else error,"First argument is neither string nor file handle";

   cards = listhead( fh );
   kwds_init;
   if(is_fh)kwds_set,"EXTNAME", "FITS_HDR", "Name of this extension";
   else kwds_set,"EXTNAME", basename(filename), "Name of this extension";
   kwds_set,"DATE", ndate(3), "Date/time of file creation";
   kwds_set,"SOFTWARE", "fits_hdr2bintab","Software that produced this file";
   if(!is_fh)kwds_set,"IN_DOL", fh_or_dol, "Input DOL";

   wrmfitscols, outfile, "HDR_CARDS", cards, clobber=clobber;

}

/* Function nfits_extens */

func nfits_extens( filename )
/* DOCUMENT number = nfits_extens( filename )

   Returns the number of last extension in the file.

   2006-02-07/NJW

*/
{
   local fh, list;
   fh = fits_open( filename );
   list = fits_list(fh);
   fits_close, fh;
   return numberof(list)-1;
}


/* Function fxpar */

func fxpar( fh, keyname )
/* DOCUMENT value = fxpar( fh, keyname )

   For reading a keyword value from a FITS header

   2006-02-07/NJW

*/
{
   return fits_get( fh, keyname );
}


/* Function writefits */

func writefits( filename_or_fh, data_array, cont=, clobber=, xax=, yax=, zax= )
/* DOCUMENT writefits, filename_or_fh, data_array, clobber=, xax=, yax=, zax=
         fh = writefits( filename_or_fh, data_array, cont=1, clobber=, xax=, yax=, zax= )

   For writing an array as an IMAGE extension in FITS.
   If the first argument is a string scalar then it is assumed to be a filename
   and if it exists beforehand then the 'clobber' keyword must be set to overwrite
   it. Else it will be created.

   First version - as a subroutine:
   If the first argument is a file handle, then it is assumed to be an open file
   and the extension will be appended to the file. There is no way to append
   an extension to a file that has been closed.

   Second version - as a function:
   Called with keyword 'cont' set : returns the filehandle and leaves
   the file open in order to allow for more extensions to be added.

   Keywords xax, yax, and zax: arrays with coordinates
                first and last elements are used

   2006-02-10/NJW

*/
{

   if( typeof(filename_or_fh) == "string" ) {
      filename = filename_or_fh;
      fh = [];
   } else if( typeof(filename_or_fh) == "list" ) {
      fh = filename_or_fh;
      filename = "";
   } else {
      write,"writefits: invalid first argument";
      return [];
   }
   if( is_void(fh) ) {

      if( file_test( filename ) ) {
         if( is_void(clobber) ) {
            write,format="writefits: %s exists already, no action!\n", filename;
            return;
         } else {
            remove, filename;
         }
      }
   }

   if( is_void(fh) ) {
      handle = fits_create( filename );
      fits_set, handle,"EXTEND",'T',"There may be extensions";
      fits_write_header, handle;
   } else handle = fh;
   fits_new_hdu, handle, "IMAGE";
   fits_set, handle, "BITPIX", fits_bitpix_of(data_array),"Datatype";
   sz = dimsof(data_array);
   fits_set_dims, handle, sz;
   if( numberof(xax) && sz(1) >= 1 ) {
      fits_set, handle, "CRPIX1", 1, "Reference pixel for axis 1";
      fits_set, handle, "CRVAL1", double(xax(1)), "Reference coordinate for axis 1";
      fits_set, handle, "CDELT1", (xax(0)-xax(1))/double(sz(2)), "Pixel size unit for axis 1";
   }
   if( numberof(yax) && sz(1) >= 2 ) {
      fits_set, handle, "CRPIX2", 1, "Reference pixel for axis 2";
      fits_set, handle, "CRVAL2", double(yax(1)), "Reference coordinate for axis 2";
      fits_set, handle, "CDELT2", (yax(0)-yax(1))/double(sz(3)), "Pixel size unit for axis 2";
   }
   if( numberof(zax) && sz(1) >= 3 ) {
      fits_set, handle, "CRPIX3", 1, "Reference pixel for axis 3";
      fits_set, handle, "CRVAL3", double(zax(1)), "Reference coordinate for axis 3";
      fits_set, handle, "CDELT3", (zax(0)-zax(1))/double(sz(4)), "Pixel size unit for axis 3";
   }

   kwds_put, handle;

   fits_write_header, handle;
   fits_write_array, handle, data_array;

   if( cont ) {
      return handle;
   } else {
      fits_close, handle;
      return [];
   }

}

/* Function readfits */

func readfits( dol, nocheck= )
/* DOCUMENT image = readfits( dol, nocheck= )

   For reading an array from an IMAGE extension in FITS
   The argumenent 'dol' is the filename combined with
   extension designation such as:
    fitsfile.fits+12, fitsfile.fits[3], fitsfile.fits[EARTHMAP]
   Note that extension number counting starts with 0 for the
   primary header (unlike original yorick fits functions, where
   counting begins with 1 (one)).

   Keyword  nocheck:  Causes skipping the file existence test

   2006-02-10/NJW

*/
{

   local filename, extno;

   get_exten_no, dol, filename, extno;
   if( !nocheck ) {
      if( !file_test(filename) ) {
         write,"READFITS: Cannot find "+filename;
         return [];
      }
   }
   fh = fits_open( filename );
   fits_goto_hdu, fh, extno+1;
   image = fits_read_array(fh);
   fits_close, fh;
   return image;

}

/* Function headfits */

func headfits( dol, nocheck= )
/* DOCUMENT hdr = headfits( dol, nocheck= )

   For reading a HDU header from a FITS extension
   Keyword nocheck to avoid checking for extension number
   exists in file [this has caused problems in certain
   swg.fits files].

   2006-02-03/NJW
   2007-04-19/NJW, Added nocheck keyword

*/
{

   local filename, extno;

   get_exten_no, dol, filename, extno;
   if( !file_test(filename) ) {
      write,"HEADFITS: File is not found: "+filename;
      return [];
   }
   if( extno < 0 ) {
      write,"HEADFITS: Extension is not found"+dol;
      return [];
   }
   fh = fits_open(filename);
   if( nocheck ) {
      num_exts = extno+1;
   } else {
      num_exts = numberof(fits_list(fh));
   }
   if( extno < num_exts ) {
      fh = fits_goto_hdu( fh, extno+1);
   } else {
      write,"Extension number too large";
      fits_close, fh;
      return [];
   }
   fits_close, fh;
   return fh;
}

/* Function rdfitscol */

func rdfitscol( dol, col_iden, silent=, ci= )
/* DOCUMENT coldata = rdfitscol( dol, col_iden, silent=, ci= )

   For reading a column from a binary table FITS
   extension. The column identifier 'col_iden' can
   be the column number or the name of the column
   (case insensitive) except if keyword 'ci' (case insist) is
   set.

   2006-02-03/NJW
   2006-11-15/NJW Updated to handle vector columns as in IDL:
                  array n x m produces m rows where each bin
                  contains n values
   2008-10-12/NJW Value returned is corrected for TZERO
   2014-12-16/NJW Handles the TDIM keyword

*/
{

   local fitsname, extno;
   chat = !silent;

   get_exten_no, dol, fitsname, extno;
   if( extno > 0 ) {
      fh = fits_open( fitsname );
      fits_goto_hdu, fh, extno+1;
      naxis2 = fxpar( fh, "NAXIS2");
      if( naxis2 == 0 ) {
         if( chat ) write,format="RDFITSCOL: No rows in table %s\n",dol;
         return [];
      }
      if( "string" == typeof(col_iden) ) {
         col_iden = strtrim(col_iden);
         if(!ci) col_iden = strupcase(col_iden);
         k = 1;
         colnum = 0;
         keynam = "TTYPE"+swrite(format="%i",k);
         while( colname = fits_get(fh, keynam) ) {
            if( colname == col_iden ) {
               colnum = k;
               break;
            }
            keynam = "TTYPE"+swrite(format="%i",++k);
         }
         if( !colnum ) {
            if( chat ) write,"RDFITSCOL: Column "+col_iden+" was not found";
            return [];
         }
      } else {
         colnum = col_iden;
      }
   } else { // 'extno' must be greater than zero for a binary table extension
      if( chat ) write,"RDFITSCOL: Invalid extension number";
      return [];
   }

   // Check for TDIM keyword

   tdims = _tdims( fh, colnum );

   p = fits_read_bintable(fh, select=colnum);
   tzero = fxpar( fh, swrite(format="TZERO%i", colnum));
   if( is_void(tzero) ) tzero = 0;
   fits_close, fh;
   // return *p(colnum);
   dms = dimsof(*p);
   if( structof(*p) == string ) {
      // TDIM is not implemented for strings
      return dms(1) == 2 ? transpose(*p) : *p ;
   } else {
      if( dms(1) < 2 ) {
         return *p + tzero;
      } else {
         if( numberof(tdims) ) {
            return transpose(*p,shift(indgen(tdims(1)):1:-1),-1);
         } else return transpose(*p) + tzero;
      }
      return dms(1) == 2 ? transpose(*p) + tzero : *p + tzero;
   }

}

/* Function get_exten_no */

func get_exten_no( fits_name_extno_in, &fits_name, &extno, tub=, nores=, casein= )
/* DOCUMENT get_exten_no, filename_extno, >filename, >extno[, tub=][, nores=][, casein=]

   For getting the filename and extension number from a DOL
   such as asdf.fits[12], asdf.fits+13, asdf.fits[EXTNAME],
   or asdf.fits[EXTNAME,EXTVER,XTENSION]

   Keyword "tub"  implies that {} are used in stead of []
   Keyword "nores" implies that if the DOL has the format ...[EXTNAME]
      then no attempt will be made to open the file to dig out the 'extno'
      Instead the inner string will be returned in 'extno'
   Keyword "casein" to force case insensitive

   2006-02-03/NJW

*/
{

   fits_name_extno = strtrim( fits_name_extno_in );

   len = strlen(fits_name_extno);
   pos = len;
   str_exten_no = "";
   extno = -1; // make sure it is always defined
   begchar = is_void(tub) ? "[" : "{";
   endchar = is_void(tub) ? "]" : "}";

   last = strpart(fits_name_extno, pos:);

   if( last == endchar || is_digit(last) ) { /* Still possible */

      if( last == endchar ) {

          /* Expects   [123] or [EXTNAME] */
          /* look for [ or x character */
          p = strpos(fits_name_extno,begchar,0,rev=1);
          if( p < 1 ) {
             extno = 0;
             fits_name = fits_name_extno;
             str_exten = "";
          } else {
             str_exten = strpart(fits_name_extno,p+1:pos-1);
             fits_name = strpart(fits_name_extno,:p-1);
             if( is_digit(str_exten) ) {
                extno = toint(str_exten);
             } else {
                extno = -1;
                if( is_void(nores) ) {
                   nlastext = nfits_extens(fits_name);
                   // check for EXTNAME,EXTVER,XTENSION convention
                   // force comma convention
                   str_exten = strstrrepl( str_exten, ":", ",");
                   if( casein ) str_exten = strupcase( str_exten );
                   tok = strsplit( str_exten,",");
                   ntok = numberof(tok);
                   if( ntok >= 2 ) {
                      in_extver = toint(tok(2));
                      str_exten = tok(1);
                   } else { // EXTVER is absent - set to 1
                      in_extver = 1;
                      // str_exten can be carried over, no delimiters are found
                   }
                   for( i = 0; i <= nlastext; ++i ) {
                      hdr = headfits(fits_name+"+"+itoa(i));
                      extname = fxpar(hdr,"EXTNAME");
                      if( casein ) extname = strupcase( extname );
                      extver = fxpar(hdr,"EXTVER");
                      if( !numberof(extver) ) extver = 1;
                      if( numberof(extname) ) {
                         extname = strtrim(extname);
                         if( extname == strtrim(str_exten) \
                                 && in_extver == extver ) {
                            extno = i;
                            break;
                         }
                      }
                   }
                }
             }
          }


       } else {  /* Expect something like +123 */

          flag = 1;
          while( flag ) {
             stp = strpart(fits_name_extno,pos:pos);
             if( is_digit(stp) )  {
                 str_exten_no = stp+str_exten_no;
             } else {
                 flag = 0;
             }
             pos = pos - 1;
          }
          if( stp == "+" ) {
            extno = toint(str_exten_no);
            fits_name = strpart(fits_name_extno,:pos);
            str_exten = strpart(fits_name_extno,pos+2:999);
         } else {
            extno = 0;
            fits_name = fits_name_extno;
            str_exten = "";
         }


       }

   } else {
      extno = 0;
      fits_name = fits_name_extno;
      str_exten = "";
   }

   if( nores ) extno = str_exten;

}

/* Function wrmfitscols */

func wrmfitscols( filename_or_fh, colname, coldata, .., var=, cont=, clobber=, extname= )
/* DOCUMENT  wrmfitscols, filename_or_fh, colname, coldata, .., var=, cont=, clobber=, extname=

   Write multiple FITS columns to a file.

   Keyword 'var' (scalar or array) can be set to give the column number(s) with variable length
   records i.e. the 1PX(max) convention. The coldata variable for such a column is disregarded
   and only used as a placeholder.

   After the filename the arguments are alternating "colname, coldata"
   'colname' will be converted to upper case.
   Setting keyword 'clobber' will cause an overwrite of the file
   if it exists beforehand. Keyword 'extname' will add the EXTNAME
   to the header.
   If the first argument is a string scalar then it is assumed to be a filename
   and if it exists beforehand then the 'clobber' keyword must be set to overwrite
   it. Else it will be created.
   If the first argument is a file handle, then it is assumed to be an open file
   and the extension will be appended to the file. There is no way to append
   an extension to a file that has been closed.

   Called as a functon with keyword 'cont' set : returns the filehandle and leaves
   the file open in order to allow for more extensions to be added.

   Limitations for this version:
      The first column cannot be of variable length i.e. 'var=1' is not possible
      since the input data give the number of rows.

   2006-02-08/NJW
   2006-08-17/NJW Updated to handle vector columns as in IDL:
                  array n x m produces m rows where each bin
                  contains n values
   2007-02-22/NJW Updated with continue option
   2008-10-02/NJW Updated with variable length option
   2010-11-19/NJW Updated with variable length option for several columns
   2014-12-12/NJW Updated to handle array k x l x m to produce m rows
                  with k x l images by setting TDIM keyword to '(k,l)'

   Example: Combine three images - each 2D - to put in a row:
             imout = grow(im1(,,-),im2(,,-),im3(,,-));
             wrmfitscols,<file>,..,"IMAGES", imout, .. ;

*/
{
   require,"idlx.i";
   require,"kfits.i"; // that contains a correction relative to fits.i

   if( numberof(coldata) < 1 ) {
      write,format="No data present in first argument, skip writing%s\n","";
      return;
   }

// Determine whether 'filename_or_fh' is a file or a FITS handle

   if( typeof(filename_or_fh) == "string" ) {
      filename = filename_or_fh;
      fh = [];
   } else if( typeof(filename_or_fh) == "list" ) {
      fh = filename_or_fh;
      filename = "";
   } else {
      write,"writefits: invalid first argument";
      return [];
   }

// Test for file existence - remove if it is to be overwritten

   if( is_void(fh) ) {
      if( file_test( filename ) ) {
         if( is_void(clobber) ) {
            write,format="wrmfitscols: %s exists already, no action!\n", filename;
            return;
         } else {
            remove, filename;
         }
      }
   }

// If new file then create file handle and write the primary data header unit

   if( is_void(fh) ) {
      fh = fits_create(filename);
      fh = fits_set(fh,"EXTEND",'T',"There may be extensions");
      fits_write_header, fh;
   }
   fits_new_bintable, fh;
   p = array(pointer, 1);

   if( !is_void(extname) ) fits_set,fh,"EXTNAME",extname,"Name of extension";

// Add data for column 1 to the array of pointers 'p'

   numcol = 1;
   ttype = "TTYPE"+swrite(format="%i",numcol);
   dm = dimsof(coldata);
   if( dm(1) == 1 ) {
      nrows = dm(2);
      p(1) = &coldata;
   } else if( dm(1) == 2 ) {
      // Case of vector bins. Assume that each column goes into a single bin
      // e.g. array 4x3 will produce 3 rows and 4 numbers in each bin
      // Rearranging is required
      tmp = transpose(coldata);
      nrows = dm(3);
      p(1) = &tmp;
   } else if( dm(1) > 2 ) {
      // Case of image bins. Assume that two first dimensions go into a single bin
      // e.g. array 5x4x3 will produce 3 rows and a 5x4 image in each bin
      // Rearranging is required
      tmp = transpose(coldata,shift(indgen(dm(1)),-1));
      nrows = dm(0);
      p(1) = &tmp;
   } else {
      write,format="Column %d has illegal format\n", 1;
      return;
   }
   fits_set, fh, ttype, strupcase(colname), "Name of column 1";

// Prepare for variable length column(s)

   stype = [];
   if( is_void(var) ) kvar = 0;
   else kvar = var;

   while( more_args() ) {
      ++numcol;
      scol = swrite(format="%i",numcol);
      ttype = "TTYPE"+scol;
      cname = next_arg();
      cdata = next_arg();
      if( nallof(kvar - numcol) ) { // check for match with 'kvar' (array or scalar)
         if( structof(cdata) == short ) {
            grow, stype, 'I';
         } else if( structof(cdata) == long || structof(cdata) == int ) {
            grow, stype, 'J';
         } else if( structof(cdata) == float ) {
            grow, stype, 'E';
         } else if( structof(cdata) == double ) {
            grow, stype, 'D';
         } else error,"Data type for variable length not supported";
         cdata = array( long, 2, nrows );
      }
      dm = dimsof(cdata);
      if( dm(1) == 1 ) {
         nrws = dm(2);
         grow, p, &cdata;
      } else if( dm(1) == 2 ) {
         // Case of vector bins. Assume that each column goes into a single bin
         // e.g. array 4x3 will produce 3 rows and 4 numbers in each bin
         // Rearranging seems to be required
         nrws = dm(3);
         tmp = transpose(cdata);
         grow, p, &tmp;
      } else if( dm(1) > 2 ) {
         // Case of image bins. Assume that two first dimensions go into a single bin
         // e.g. array 5x4x3 will produce 3 rows and a 5x4 image in each bin
         // Rearranging is required
         tmp = transpose(cdata,shift(indgen(dm(1)),-1));
         nrws = dm(0);
         grow, p, &tmp;
      } else {
         write,format="Column %d has illegal format\n", numcol;
         return;
      }
      if( nrws != nrows ) {
         write,format="Error in number of rows in column #%i\n", numcol;
         fits_close, fh;
         return;
      }
      fits_set, fh, ttype, strupcase(cname), "Name of column "+scol;
   }

   kwds_put, fh;

   fits_write_bintable, fh, p, var=var, stype=stype;
   if( cont ) {
      return fh;
   } else {
      fits_close, fh;
      return [];
    }
}

/* Function skypos_fits */

struct s_Coords {
   int flag;
   string ctype1, ctype2;
   double cdelt1, cdelt2, crota2;
   double crval1, crval2, crpix1, crpix2;
   double cd1_1, cd1_2, cd2_1, cd2_2;
   double scale; // degrees per pixel regardless of rotation for quadratic pixels
}

func skypos_fits( arg1, &pix1, &pix2, &coord1, &coord2, to_sky=, to_pix=, r1=,r2=, silent=)
/* DOCUMENT skypos_fits, arg1, (>)pix1, (>)pix2, (>)coord1, (>)coord2, to_sky=, to_pix=, r1=,r2=, silent=

   Get sky position for a given position in FITS image
   pix1 and pix2 are pixel numbers starting count with 1 (to_sky=1)
   i.e. 1 larger than in C and IDL
   or pixel numbers from coordinates (to_pix=1)

   First argument is either of:
             1) a FITS DOL
             2) a FITS header
             3) an instance of the struct 's_Coords'

    2003-09-16/NJW
    2003-12-17/NJW Updated to derive pixel numbers from coords
    2014-05-08/NJW Returns pixel numbers as floating point.
*/

{
   require, "matrix.i";

   vb = silent ? 0 : 1;
   if( !to_sky ) to_sky = 0;
   if( !to_pix ) to_pix = 0;
   if( to_sky + to_pix != 1 ) {
      write,"Exactly one of the keywords 'to_sky' or 'to_pix' must be set";
      return;
   }

   s_coords = [];
   if( typeof(arg1) == "string" ) { // 1. argument is a DOL

      file_ext = arg1;

      local file, extno;
      get_exten_no, file_ext, file, extno;

      if( !file_test(file) ) {
         write,"file "+file+" not found!";
         return;
      }

      hdr = headfits(file_ext);
      if( is_void(hdr) ) return;
      if( is_void(fits_get(hdr,"BITPIX") ) ) {
         write,"Requested extension ("+itoa(extno)+") is not found!";
         return;
      }
      s_coords = get_wcs( arg1 );

   } else if( typeof(arg1) == "struct_instance" ) {
      s_coords = arg1;
   } else s_coords = get_wcs( arg1 );

   dr = pi / 180;

   crval1 = s_coords.crval1;
   crval2 = s_coords.crval2;
   crpix1 = s_coords.crpix1;
   crpix2 = s_coords.crpix2;
   cd1_1 = s_coords.cd1_1;
   cd1_2 = s_coords.cd1_2;
   cd2_1 = s_coords.cd2_1;
   cd2_2 = s_coords.cd2_2;

 // Replace crpix values if keywords are given
   if( !is_void(r1) ) crpix1 = r1;
   if( !is_void(r2) ) crpix2 = r2;

   det = cd1_1 * cd2_2 - cd1_2 * cd2_1;

   posangle = atan(cd1_2,cd1_1) / dr;

   cdmat = [[cd1_1,cd2_1],[cd1_2,cd2_2]];

   if( vb ) {
      write,format="CRVAL1 = %f, CRVAL2 = %f\n", crval1, crval2;
      write,format="Derived POSANGLE = %f\n", posangle;
      write,format="CRPIX1 = %f, CRPIX2 = %f\n", crpix1, crpix2;
      write,format="CD1_1 = %f, CD1_2 = %f\n", cd1_1, cd1_2;
      write,format="CD2_1 = %f, CD2_2 = %f\n", cd2_1, cd2_2;
   }


   a0 = crval1 * dr;
   ca0 = cos(a0);
   sa0 = sin(a0);
   d0 = crval2 * dr;
   cd0 = cos(d0);
   sd0 = sin(d0);

// transformation matrices:

   t1 = [[ca0,sa0,0],[-sa0,ca0,0],[0,0,1]];
   t2 = [[cd0,0,sd0],[0,1,0],[-sd0,0,cd0]];

   t = t1(,+)*t2(+,);

   if( to_sky ) {

      // star position in image (y,z) expressed in degrees;
      yz = cdmat(,+) * [pix1-crpix1,pix2-crpix2](+);
      y = atan(yz(1) * dr);
      z = atan(yz(2) * dr);
     // unit vector expressed in image system;
      vi = [0,sin(y),sin(z)];
      vi(1) = sqrt(1. - vi(2)^2 - vi(3)^2);

     // star position in eq system

      ve = t(,+) * vi(+);

     // derive RA and Dec;

      d = asin(ve(3));
      a = zero2pi(atan(ve(2),ve(1)));
      ddeg = d / dr;
      adeg = a / dr;

      coord1 = adeg;
      coord2 = ddeg;

      if( vb ) write,format="RA: %.4f, Dec: %.4f\n", adeg, ddeg;

   } else {

      arad = coord1 * dr;
      drad = coord2 * dr;

      ve = [cos(arad)*cos(drad),sin(arad)*cos(drad),sin(drad)];
      vi = LUsolve( t )(,+) *  ve(+,);

      yz = tan(asin(vi(2:3))) / dr;

      p = LUsolve(cdmat)(,+) * yz(+) + [crpix1,crpix2];

      //+ pix1 = long(p(1)+0.5);
      //+ pix2 = long(p(2)+0.5);
      pix1 = p(1);
      pix2 = p(2);

      if( vb ) write,format="Xpix: %.4f, Ypix: %.4f\n", pix1, pix2;

   }

   return;

}

/* Function simple_peak */

func simple_peak( dol, ra_obj, dec_obj, peak_size, bkg_size, chat= )
/* DOCUMENT peaksize = simple_peak( dol, ra_obj, dec_obj, peak_size, bkg_size, chat= )

   Calculates the number of counts in an area peak_size x peak_size around
   the source placed at (ra_obj, dec_obj) in a FITS image (NAN values are
   replaced with zeros). The background is evaluated in a region bkg_size x bkg_size
   around the source position.

   If bkg_size <= peak_size then the background is ignored (set to zero).

   2006-12-28/NJW
*/
{
   require, "basic.i";
   require, "idlx.i";

   im = readfits(dol);
   if( is_void(im) ) {
      write,"No such DOL: "+dol;
      return [];
   } else im = nan2zero(double(im));

   sz = dimsof(im);
   skypos_fits, dol, pix1, pix2, ra_obj, dec_obj, to_pix=1, silent=1;
   pix1 = long(pix1 + 0.5);
   pix2 = long(pix2 + 0.5);
   if( pix1 < 1 || pix2 < 1 || pix1 > sz(2) || pix2 > sz(3) ) {
      write,"Warning - object outside the image";
   }
   sbox = extract_box(im,pix1,pix2,peak_size,peak_size,cen=1);
   source = sum(sbox);
   nsource = numberof(sbox);
   if( chat ) {
      write,format="Total in source box: %f\n", source;
      write,format="Num pixels in source box: %i\n", nsource;
   }
   if( bkg_size > peak_size ) {
      bbox = extract_box(im,pix1,pix2,bkg_size,bkg_size,cen=1);
      totbkg = sum(bbox);
      ntotbkg = numberof(bbox);
      if( chat ) {
         write,format="Total in bkg box: %f\n", totbkg;
         write,format="Num pixels in bkg box: %i\n", ntotbkg;
      }
      bkg = totbkg - source;
      nbkg = ntotbkg - nsource;
      if( chat ) {
         write,format="Bkg in bkg ring: %f\n", bkg;
         write,format="Num pixels in bkg ring: %i\n", nbkg;
         write,format="Bkg counts/pixel: %f\n", bkg / nbkg;
      }
      correction = (nsource * bkg) / nbkg;
   } else {
      correction = 0.0;
   }

   return source - correction;

}

/* Function gz_proxy_file */

func gz_proxy_file( filename, dir=, silent= )
/* DOCUMENT non_gzfilename =  gz_proxy_file( filename, dir=, silent= )

   The filename may be given with or without the trailing '.gz'.
   Returns the name of the proxy file for a gzipped file
   The original file is copied to /scratch/njw/proxies and unzipped

      2006-02-02/NJW
      2007-01-29/NJW, updated to use derivative of file name
           in order to avoid clashes. Directory can be specified.
*/
{

   require, "idlx.i";
   require, "basic.i";

   if( is_void(filename) ) {
      write,"Syntax: newfile = gz_proxy_file( filename )";
      return [];
   }
   if( is_void(silent) ) silent = 0;
   vb = silent == 0 ? 1 : 0;

// Define directory where to put the proxy file

   if( is_void(dir) ) {
      dir = "/r12/scratch/njw/proxies/";
   }

   // Make sure the 'filename' hasn't a '.gz' postfix
   if( strlen(filename) > 3 ) {
      if( strpart(filename, -2:0) == ".gz" ) filename = strpart(filename, 1:-3);
   }

   if( ! file_test(filename) ) {
      //+ write,"#2# did not find "+filename
      if( file_test(filename+".gz") ) { // found in compressed format
         sname = fullpath(filename);
         sname = strcharrepl( sname,"/","_");
         if( strlen(sname) > 150 ) sname = strpart(sname, -149:0);
         proxy_name = app_slash(dir)+sname;
         if( !file_test( proxy_name ) ) { // Only make a copy if not found
            if(vb) write,format="Do: > %s\n", "cp "+filename+".gz "+proxy_name+".gz";
            system,"cp "+filename+".gz "+proxy_name+".gz";
            system,"gunzip "+proxy_name+".gz";
         } else {
            if(vb) write,format="Exists already: %s\n", proxy_name;
         }
         return proxy_name;
         //+ write,"#3# New filename = "+filename
      } else {
         if(vb) {
            write,"GZ_PROXY_FILE: Neither "+filename;
            write," nor "+filename+".gz was found";
         }
         return [];
      }
   } else return filename;  // no change if file exists

}

/* Function kwds_init */

func kwds_init( value )
/* DOCUMENT kwds_init[, value]

    Initialize the in-memory keywords of the same type as 'value'.
    Omitting 'value' causes a complete initialization.

    SEE ALSO: kwds_set, kwds_get, kwds_put, kwds_init, kwds_del

    2007-02-04/NJW
*/
{

   require, "idlx.i";

   extern kwds_cnames, kwds_rnames, kwds_inames, kwds_snames;
   extern kwds_cvalues, kwds_rvalues, kwds_ivalues, kwds_svalues;
   extern kwds_ccomments, kwds_rcomments, kwds_icomments, kwds_scomments;
   extern kwds_csequence, kwds_rsequence, kwds_isequence, kwds_ssequence, kwds_count;

   if( is_void(value) ) {
      //+ write,"Initializing all keywords";
      kwds_cnames = kwds_cvalues = kwds_ccomments = kwds_csequence = [];
      kwds_rnames = kwds_rvalues = kwds_rcomments = kwds_rsequence = [];
      kwds_inames = kwds_ivalues = kwds_icomments = kwds_isequence = [];
      kwds_snames = kwds_svalues = kwds_scomments = kwds_ssequence = [];
      kwds_count = 0;
      return;
   }

   if( "double" == typeof(value) || "float" == typeof(value) ) {
      //+ write,"Initializing double keywords"
      kwds_rnames = kwds_rvalues = kwds_rcomments = kwds_rsequence = [];
   } else if( "int" == typeof(value) || "long" == typeof(value) ) {
      //+ write,"Initializing long keywords"
      kwds_inames = kwds_ivalues = kwds_icomments = kwds_isequence = [];
   } else if( "string" == typeof(value) ) {
      //+ write,"Initializing string keywords"
      kwds_snames = kwds_svalues = kwds_scomments = kwds_ssequence = [];
   } else if( "char" == typeof(value) ) {
      kwds_cnames = kwds_cvalues = kwds_ccomments = kwds_csequence = [];
   } else {
      write,"KWDS_INIT: No action since data type \""+typeof(value)+"\" is not supported";
   }

}

/* Function kwds_put */

func kwds_put( &fh )
/* DOCUMENT fh = kwds_put( fh )

    Update the filehandle 'fh' with all keyword information
    from database in memory.

    The LONGSTRN convention is implemented here.

    SEE ALSO: kwds_set, kwds_get, kwds_init, kwds_del

    2007-02-04/NJW
*/
{

   require, "idlx.i";

   extern kwds_cnames, kwds_rnames, kwds_inames, kwds_snames;
   extern kwds_cvalues, kwds_rvalues, kwds_ivalues, kwds_svalues;
   extern kwds_ccomments, kwds_rcomments, kwds_icomments, kwds_scomments;
   extern kwds_csequence, kwds_rsequence, kwds_isequence, kwds_ssequence, kwds_count;

   if( is_void(kwds_count) ) kwds_count = 0;

   /* The value of 'kwds_count' might be larger than the actual number of
      keywords since it does not keep track of deleted keywords. So the
      following loop does not necessarily call 'fits_set' every time.
   */
   for( i = 1; i <= kwds_count; i++ ) {
      if( (nw = numberof((w=where(kwds_csequence == i)))) ) { // Bring the char keyword
         fits_set, fh, kwds_cnames(w(1)),kwds_cvalues(w(1)),kwds_ccomments(w(1));
      } else if( (nw = numberof((w=where(kwds_rsequence == i)))) ) { // Bring the real keyword
         fits_set, fh, kwds_rnames(w(1)),kwds_rvalues(w(1)),kwds_rcomments(w(1));
      } else if( (nw = numberof((w=where(kwds_isequence == i)))) ) { // Bring the integer keyword
         fits_set, fh, kwds_inames(w(1)),kwds_ivalues(w(1)),kwds_icomments(w(1));
      } else if( (nw = numberof((w=where(kwds_ssequence == i)))) ) { // Bring the string keyword
         if( kwds_snames(w(1)) == "HISTORY" || kwds_snames(w(1)) == "COMMENT" ) {
            fits_set, fh, kwds_snames(w(1)),kwds_svalues(w(1));
         } else {
            fits_set, fh, kwds_snames(w(1)),kwds_svalues(w(1)),kwds_scomments(w(1));
         }
      }
   }

   return fh;

}

/* Function kwds_set */

func kwds_set( name, value, comment )
/* DOCUMENT kwds_set, keyword_name, value, comment

    Update keyword database with keyword information

    Uses the LONGSTRN convention for a string keyword where
    its length exceeds 70 characters.

    SEE ALSO: kwds_get, kwds_put, kwds_init, kwds_del

    2007-02-04/NJW
*/
{

   require, "idlx.i";

   name = strupcase(name);
   if( is_void(comment) ) comment = "";

   extern kwds_cnames, kwds_rnames, kwds_inames, kwds_snames;
   extern kwds_cvalues, kwds_rvalues, kwds_ivalues, kwds_svalues;
   extern kwds_ccomments, kwds_rcomments, kwds_icomments, kwds_scomments;
   extern kwds_csequence, kwds_rsequence, kwds_isequence, kwds_ssequence, kwds_count;

   if( is_void(kwds_count) ) kwds_count = 0;

   if( "double" == typeof(value) || "float" == typeof(value) ) {
      // see if keyword is already there
      w = where( name == kwds_rnames );
      if( numberof(w) > 0 ) {
         kwds_rvalues(w(1)) = value;
         kwds_rcomments(w(1)) = comment;
      } else {
         grow, kwds_rnames, name;
         grow, kwds_rvalues, double(value);
         grow, kwds_rcomments, comment;
         grow, kwds_rsequence, ++kwds_count;
      }
   } else if( "int" == typeof(value) || "long" == typeof(value) ) {
      // see if keyword is already there
      w = where( name == kwds_inames );
      if( numberof(w) > 0 ) {
         kwds_ivalues(w(1)) = value;
         kwds_icomments(w(1)) = comment;
      } else {
         grow, kwds_inames, name;
         grow, kwds_ivalues, long(value);
         grow, kwds_icomments, comment;
         grow, kwds_isequence, ++kwds_count;
      }
   } else if( "string" == typeof(value) ) {
      // avoid overwriting for special keywords
      if( name == "HISTORY" || name == "COMMENT" ) {
         grow, kwds_snames, name;
         grow, kwds_svalues, value;
         grow, kwds_scomments, ""; // to maintain accounting
         grow, kwds_ssequence, ++kwds_count;
      } else {

         // see if keyword is already there
         w = where( name == kwds_snames );
         if( numberof(w) > 0 ) {
            kwds_svalues(w(1)) = value;
            kwds_scomments(w(1)) = comment;
         } else {
            grow, kwds_snames, strupcase(name);
            grow, kwds_svalues, value;
            grow, kwds_scomments, comment;
            grow, kwds_ssequence, ++kwds_count;
         }
         // Do we have a long string ? If yes then
         // set the LONGSTRN keyword unless is already exists.
         if( strlen(value) > 68 ) {
            w = where( "LONGSTRN" == kwds_snames );
            if( numberof(w) == 0 ) {
               grow, kwds_snames, "LONGSTRN";
               grow, kwds_svalues, "OGIP 1.0";
               grow, kwds_scomments, "Uses the long string convention";
               grow, kwds_ssequence, ++kwds_count;
            }
         }
      }
   } else if( "char" == typeof(value) ) {
      // see if keyword is already there
      w = where( name == kwds_cnames );
      if( numberof(w) > 0 ) {
         kwds_cvalues(w(1)) = value;
         kwds_ccomments(w(1)) = comment;
      } else {
         grow, kwds_cnames, name;
         grow, kwds_cvalues, value;
         grow, kwds_ccomments, comment;
         grow, kwds_csequence, ++kwds_count;
      }
   } else {
      write,"KWDS_SET: No action for "+name+" since data type \""+typeof(value)+"\" is not supported";
   }

}

/* Function kwds_setlongstr */

func kwds_setlongstr( keyword, long_string, comment )
/* DOCUMENT kwds_setlongstr, keyword, long_string, comment

    Use kwds_set several times to split up 'long_string'.
    Use 'keyword' as it is for the first part, and then
    truncate to 6 letters (if necessary) and count from
    01 thru 99.

    SEE ALSO: kwds_init, kwds_set, kwds_del, kwds_put.

    2012-10-26/NJW

*/
{
   if( typeof(long_string) != "string" ) error,"Called with non-string 2. argument";

   lenk = strlen(keyword);
   lenl = strlen(long_string);
   if( lenl < 30 ) { // No real use for this function
      kwds_set, keyword, long_string, comment;
      return;
   }

   keyw6 = strpart(keyword,1:6);
   count = 0;

   while( lenl > 0 ) {
      body = strpart( long_string, 1:60 );
      long_string = strpart( long_string, 61: );
      lenl = strlen(long_string);
      cont = lenl == 0 ? "" : " &";
      if( count == 0 ) {
         kwds_set, keyword, body+cont;
      } else {
         kwds_set, keyw6+itoa(count,2), body+cont;
      }
      count++;
   }
   if( strlen(comment) ) kwds_set, keyw6+itoa(count,2), "COMMENT", comment;
}

/* Function kwds_del */

func kwds_del( name )
/* DOCUMENT kwds_del, keyword_name

    Delete a keyword

    Will not delete BYTPIX and HISTORY.

    SEE ALSO: kwds_set, kwds_get, kwds_put, kwds_init

    2007-02-20/NJW
    2008-06-17/NJW, may now delete "COMMENT" keywords.
*/
{

   require, "idlx.i";

   name = strupcase(name);

   extern kwds_cnames, kwds_rnames, kwds_inames, kwds_snames;
   extern kwds_cvalues, kwds_rvalues, kwds_ivalues, kwds_svalues;
   extern kwds_ccomments, kwds_rcomments, kwds_icomments, kwds_scomments;
   extern kwds_csequence, kwds_rsequence, kwds_isequence, kwds_ssequence, kwds_count;

   // Check if deletion is prohibited by keyword name

   if( name == "BYTPIX" ) { write,"Cannot delete BYTPIX"; return; }
   // if( name == "COMMENT" ) { write,"Cannot delete COMMENT"; return; }
   if( name == "HISTORY" ) { write,"Cannot delete HISTORY"; return; }

   // see if keyword is from 'char' group
   w = where( name == kwds_cnames );
   if( numberof(w) > 0 ) {
      m = where( name != kwds_cnames );
      if( numberof(m) > 0 ) {
         kwds_cnames = kwds_cnames(m);
         kwds_cvalues = kwds_cvalues(m);
         kwds_ccomments = kwds_ccomments(m);
         kwds_csequence = kwds_csequence(m);
      } else {
         kwds_cnames = kwds_cvalues = kwds_ccomments = kwds_csequence = [];
      }
      return;
   }

   // see if keyword is from 'real' group
   w = where( name == kwds_rnames );
   if( numberof(w) > 0 ) {
      m = where( name != kwds_rnames );
      if( numberof(m) > 0 ) {
         kwds_rnames = kwds_rnames(m);
         kwds_rvalues = kwds_rvalues(m);
         kwds_rcomments = kwds_rcomments(m);
         kwds_rsequence = kwds_rsequence(m);
      } else {
         kwds_rnames = kwds_rvalues = kwds_rcomments = kwds_rsequence = [];
      }
      return;
   }

   // see if keyword is from 'integer' group
   w = where( name == kwds_inames );
   if( numberof(w) > 0 ) {
      if( numberof(m) > 0 ) {
         m = where( name != kwds_inames );
         kwds_inames = kwds_inames(m);
         kwds_ivalues = kwds_ivalues(m);
         kwds_icomments = kwds_icomments(m);
         kwds_isequence = kwds_isequence(m);
         return;
      } else {
         kwds_inames = kwds_ivalues = kwds_icomments = kwds_isequence = [];
      }
   }

   // see if keyword is from 'string' group
   w = where( name == kwds_snames );
   if( numberof(w) > 0 ) {
      m = where( name != kwds_snames );
      if( numberof(m) > 0 ) {
         kwds_snames = kwds_snames(m);
         kwds_svalues = kwds_svalues(m);
         kwds_scomments = kwds_scomments(m);
         kwds_ssequence = kwds_ssequence(m);
      } else {
         kwds_snames = kwds_svalues = kwds_scomments = kwds_ssequence = [];
      }
      return;
   }

}

/* Function kwds_get */

func kwds_get( name, &comment )
/* DOCUMENT value = kwds_get( keyword_name, >comment )

    Get the value of a keyword previously stored in memory
    by 'kwds_set'

    SEE ALSO: kwds_set, kwds_put, kwds_init, kwds_del

    2008-06-27/NJW
*/
{

   require, "idlx.i";

   name = strupcase(name);

   extern kwds_cnames, kwds_rnames, kwds_inames, kwds_snames;
   extern kwds_cvalues, kwds_rvalues, kwds_ivalues, kwds_svalues;
   extern kwds_ccomments, kwds_rcomments, kwds_icomments, kwds_scomments;
   extern kwds_csequence, kwds_rsequence, kwds_isequence, kwds_ssequence, kwds_count;

   // see if keyword is from 'char' group
   w = where( name == kwds_cnames );
   if( numberof(w) > 0 ) {
      comment = kwds_ccomments(w);
      return kwds_cvalues(w);
   }

   // see if keyword is from 'real' group
   w = where( name == kwds_rnames );
   if( numberof(w) > 0 ) {
      comment = kwds_rcomments(w);
      return kwds_rvalues(w);
   }

   // see if keyword is from 'integer' group
   w = where( name == kwds_inames );
   if( numberof(w) > 0 ) {
      comment = kwds_icomments(w);
      return kwds_ivalues(w);
   }

   // see if keyword is from 'string' group
   // Special treatment for "COMMENT" and "HISTORY"
   //    i.e. they have no comment fields
   w = where( name == kwds_snames );
   if( numberof(w) > 0 ) {
      if( name == "COMMENT" || name == "HISTORY" ) {
         comment = [];
      } else {
         comment = kwds_scomments(w);
      }
      return kwds_svalues(w);
   }

   // The keyword was not found
   comment = [];
   return [];

}


/* Function fits_add_rows */

func fits_add_rows( dol, num_extra_rows, outfile= )
/* DOCUMENT fits_add_rows, dol, num_extra_rows, outfile=

   Add a number of (empty) rows to an existing FITS binary table extension.
   Copy all header keywords.
   If no keyword 'outfile' is given the original file will be
   overwritten.

   2008-08-22/NJW
   2019-05-14/NJW updated to cope with optional extension number.

*/
{
   local fh, nrows, filename, extno;

  /*
   *  Find number of extensions in the input file.
   *  If there is more than one, then copying must be performed.
   */

   get_exten_no, dol, filename, extno;
   n_extensions = nfits_extens( filename ); // extension number of last extension
       // remember that Yorick counting starts with 1 (one) whereas
       // the extension number in 'dol' begins with 0 (zero).

  /*
   *  if extno > n_extensions then 'we have a problem'
   *  if n_extensions == 1 then just write new file
   *  if n_extensions > 1 then
   *     copy all extensions lower than extno
   *     write new extension
   *     copy all extensions higher than extno
   *
   */

   if( extno > n_extensions ) error,"extno > n_extensions , illegal request";

  /*
   *  Make sure that the requested extension is really a binary table
   */

   fh = fits_open( filename, "r" ); // open in read-only mode
   fh = fits_goto_hdu( fh, extno+1 ); // move to requested extension
   if( "BINTABLE" != fits_get_xtension( fh ) ) {
      write,"The extension is not of type BINTABLE - quit ...";
      fits_close, fh;
      return;
   } else fits_close, fh;

  /*
   *   Add the requested number of rows to the data unit in question
   */

   ptr = _fits_add_rows( dol, num_extra_rows );

   if( n_extensions == 1 ) {
      if( numberof(outfile) ) filename = outfile;
      wrfitsbin, filename, fh, ptr, clobber=1;
   } else {
      workfile = get_next_filename("je4b5t_???.fits");
      fh = fits_copy_file( filename, workfile, hdumax=extno-1 );
      fh = fits_write_bintable( fh, ptr );
      if( extno < n_extensions ) {
         fits_copy_file, filename, fh, hdumin=extno+1;
      } else fits_close, fh;
      if( numberof(outfile) ) filename = outfile;
      mv, workfile, filename;
   }

}

func _fits_add_rows( dol, num_extra_rows )
{
   local fh, nrows;

   ptr = rdfitsbin( dol, fh, nrows );

   numcols = numberof(ptr);

  /*
   *  Loop over columns to grow each of them individually
   *  All other data types than 'string' may be two-dimensional
   */

   for( col = 1; col <= numcols; col++ ) {

      tmp = *ptr(col);
      type = typeof(tmp);
      dms = dimsof(tmp);

      if( type == "string" ) {
         grow, tmp, array( string, num_extra_rows );
      } else {
         if( dms(1) == 2 ) {     // in case of 2D array transposition
            tmp = transpose(tmp); // is required since internal
            dms = dimsof(tmp);    // representation is different
         }
         dms(0) = num_extra_rows;
         if( type == "char" ) {
            grow, tmp, array( char, dms );
         } else if( type == "int" ) {
            grow, tmp, array( int, dms );
         } else if( type == "long" ) {
            grow, tmp, array( long, dms );
         } else if( type == "float" ) {
            grow, tmp, array( float, dms );
         } else if( type == "double" ) {
            grow, tmp, array( double, dms );
         } else error,"FITS_ADD_ROWS: Data type not supported";
      }

      if( dms(1) == 2 ) tmp = transpose(tmp);
      ptr(col) = &tmp;

   }

   return ptr;

}
/* Function fits_copy_keys */

func fits_copy_keys( fhfrom, &fhto, list=, skip=, tokwds= )
/* DOCUMENT fits_copy_keys, fhfrom, >fhto, list=, skip=, tokwds=

   Copy keywords from one FITS header ('fhfrom') to another one ('fhto'), or from
   a string array (must conform with FITS syntax).

   Keywords
     list     can be a comma-separated list (scalar string)
              of FITS keywords to be copied.
              If keyword 'list' is not given then
              all non-FITS specific keywords will be copied from 'fhfrom'
              to 'fhto'.
     skip     A comma-separated list (scalar string) of FITS keywords
              to be skipped in the copying process. Overrides 'list'.
    tokwds    (bool)  When set then the 'kwds' functions
              are activated in stead.
              The keywords will be reset prior to the copying unless
              the value of 'tokwds' is larger than 1.

   2007-09-12/NJW
   2010-08-18/NJW, list keyword added
   2014-12-05/NJW, special handling of long strings
   2015-01-13/NJW, keyword 'skip' included.
*/
{
   cards = typeof(fhfrom) == "list" ? _car(fhfrom,1) : fhfrom;
   ncards = numberof(cards);
   if( tokwds ) {
      if( tokwds <= 1 ) kwds_init;
   }

   // check for the existence of the keyword 'LONGSTRN'
   use_longstrn = numberof(where(strmatch(strpart(cards,1:8),"LONGSTRN")));

   // The variable 'ql' is set to 1 if only a selected list of keywords
   // should be copied

   if( typeof(list) == "string" ) {
      selkeys = strsplit(list,",");
      nselkeys = numberof(selkeys);
      ql = 1;
   } else ql = 0;

   // The variable 'rl' is set to 1 if a selected list of keywords
   // should be skipped

   if( typeof(skip) == "string" ) {
      vetokeys = strsplit(skip,",");
      nvetokeys = numberof(vetokeys);
      rl = 1;
   } else rl = 0;

   extra_cont = "";

   for( i = 1; i <= ncards; i++ ) {

      key = strtrim(strpart(cards(i),1:8));
      if( key == "XTENSION" ) continue;
      if( key == "BITPIX" ) continue;
      if( key == "NAXIS" ) continue;
      if( key == "CONTINUE" ) { // should not appear unless LONGSTRN
         if( !use_longstrn ) write,"CONTINUE keyword appeared unexpectedly";
         continue;
      }
      key5 = strpart(key,1:5);
      if( key5 == "NAXIS" && is_digit(strpart(key,6:8)) ) continue;
      if( key == "PCOUNT" ) continue;
      if( key == "TFIELDS" ) continue;
      if( key5 == "TTYPE" && is_digit(strpart(key,6:8)) ) continue;
      if( key5 == "TFORM" && is_digit(strpart(key,6:8)) ) continue;
      if( key5 == "TZERO" && is_digit(strpart(key,6:8)) ) continue;
      if( key5 == "TSCAL" && is_digit(strpart(key,6:8)) ) continue;
      if( key5 == "TUNIT" && is_digit(strpart(key,6:8)) ) continue;

      // Conditional check for a copy based on inclusion list 'list'
      if( ql ) {
         accept = 0;
         for( j = 1; j <= nselkeys; j++ ) {
            if( selkeys(j) == key ) accept = 1;
         }
         if( !accept ) continue;
      }

      // Conditional check for a copy based on rejection list 'skip'
      if( rl ) {
         accept = 1;
         for( j = 1; j <= nvetokeys; j++ ) {
            if( vetokeys(j) == key ) accept = 0;
         }
         if( !accept ) continue;
      }

      value = fits_parse( cards(i) );
      if( use_longstrn && structof(value) == string ) {
         // check for a continuation
         lvalue = value;
         while( strpart(value,0:0) == "&" ) {
            lvalue = strpart(lvalue,1:-1); // chop off the '&' symbol
            // next card must be a continuation card
            if( strpart(cards(i+1),1:8) == "CONTINUE" ) { // all is well
               ++i;
               value = fits_parse( cards(i) );
               lvalue += value;
            } else {
               write,"A CONTINUE cards was expected at position "+itoa(i+1);
            }
         }
         value = lvalue;
      }
            
      if( tokwds ) {
         if( is_void(_fits_parse_comment) ) {
            kwds_set, key, value;
         } else {
            kwds_set, key, value, strtrim(_fits_parse_comment);
         }
      } else {
         if( is_void(_fits_parse_comment) ) {
            fits_set, fhto, key, value;
         } else {
            fits_set, fhto, key, value, _fits_parse_comment;
         }
      }

   }

}

/* Function rdfitsbin */

func rdfitsbin( dol, &fh, &nrows, silent= )
/* DOCUMENT ptr = rdfitsbin( dol, >fh, >nrows, silent= )

   Read entire FITS binary table and return the pointer array.

   If the file or the extension is not found then a 'void' is returned.

   2007-09-12/NJW, updated 2011-09-13/NJW
*/
{
   local filename, extno;

   get_exten_no, dol, filename, extno;
   if( !file_test(filename) ) {
      if( !silent ) write,"File not found: "+filename;
      return [];
   }
   if( extno < 1 ) {
      if( !silent ) write,"Extension not found: "+dol;
      return [];
   }
   fh = fits_open( filename, 'r');
   fits_goto_hdu, fh, extno+1;
   ptr = fits_read_bintable(fh);
   nrows = fits_get(fh, "NAXIS2");
   fits_close, fh;
   return ptr;
}

/* Function wrfitsbin */

func wrfitsbin( filename, fh, ptr, clobber= )
/* DOCUMENT wrfitsbin, filename, fh, ptr, clobber=

   Write entire FITS binary table to the first extension
   into the file 'filename' and copy the essential keywords
   from 'fh' that also holds the column name and unit info.
   The input is the array of pointers for the binary table.

   2007-09-12/NJW
*/
{
   require, "idlx.i";
   require, "kfits.i";

   if( file_test( filename ) ) {
      if( clobber ) {
         remove, filename;
      } else {
         write,format="No action, %s exists already\n", filename;
         return;
      }
   }
   newfh = fits_create(filename);
   newfh = fits_set( newfh, "EXTEND",'T',"There may be extensions");
   fits_write_header, newfh;
   fits_new_bintable, newfh;

   /* copy all non FITS specific keywords */

   fits_copy_keys, fh, newfh;
   fits_set, newfh, "DATE", ndate(3), "Time of file creation";

   /* copy all column names */

   colname = "";
   ncols = 0;
   while( 1 ) {
      ttype = swrite(format="TTYPE%i",++ncols);
      colname = fits_get(fh,ttype);
      if( is_void(colname) ) {
         break;
      } else {
         fits_set,newfh, ttype, colname,"Column name";
      }
   }

   if( --ncols != numberof(ptr) ) {
      write,format="Inconsistency: %i columns in fh but %i pointers\n", \
                          ncols, numberof(ptr);
      fits_close, newfh;
      return;
   }

   /* copy all unit names */

   unit = "";
   for( i = 1; i <= ncols; i++ ) {
      tunit = swrite(format="TUNIT%i",i);
      unit = fits_get(fh,tunit);
      if( !is_void(unit) ) fits_set,newfh, tunit, unit,"Unit for this column";
   }

   /* Check columns of type 'string' for consistency */

   for( i = 1; i <= ncols; i++ ) {
      if( typeof(*ptr(i)) == "string" ) {
         tform = swrite(format="TFORM%i", i);
         tval = fits_get(fh, tform);
         if( strpart(tval,0:0) != "A" ) {
            write,format="%s problem, value = %s\n", tform, tval;
            fits_close, newfh;
            return;
         }
         len = atoi(strpart(tval, 1:-1));
         ss = s = (*ptr(i))(1);
         while( strlen(s) < len ) s += " ";
         if( strlen(ss) != len ) (*ptr(i))(1) = s;
      }
   }

   fits_write_bintable, newfh, ptr;
   fits_close, newfh;
}

/* Function fits_colnum */

func fits_colnum( fh_or_dol, colname, silent= )
/* DOCUMENT number = fits_colnum( fh_or_dol, colname, silent= )

   Returns the column number of column 'colname' (may be lower case)
   when DOL or header information is in 'fh_or_dol'.

   Keyword silent:  To suppress screen message

   2007-09-12/NJW
*/
{

   if( typeof(fh_or_dol) == "string" ) {
      fh = headfits( fh_or_dol );
   } else if( typeof(fh_or_dol) == "list" ) {
      fh = fh_or_dol;
   } else error,"First argument is neither string nor file handle";

   colname = strupcase(colname);
   k = 1;
   colnum = 0;
   keynam = "TTYPE"+swrite(format="%i",k);
   while( coln = fits_get(fh, keynam) ) {
      if( coln == colname ) {
         colnum = k;
         break;
      }
      keynam = "TTYPE"+swrite(format="%i",++k);
   }
   if( !colnum ) {
      if(!silent)write,format="Column %s was not found\n", colname;
      return [];
   }

   return colnum;

}

/* fits_numcols */

func fits_numcols( fh_or_dol )
/* DOCUMENT number = fits_numcols( fh_or_dol )

   Returns the number of columns
   Argument is either a FITS file handle (header) or a DOL

   2007-09-25/NJW
*/
{

   if( typeof(fh_or_dol) == "string" ) {
      fh = headfits( fh_or_dol );
   } else if( typeof(fh_or_dol) == "list" ) {
      fh = fh_or_dol;
   } else error,"First argument is neither string nor file handle";

   numcols = 0;
   while( fits_get(fh, swrite(format="TTYPE%i",numcols+1) ) ) numcols++;

   return numcols;

}

/* Function fits_bintable_poke */

func fits_bintable_poke( fh_or_DOL, row_number, col_desig, value, keeppcount=, chat= )
/* DOCUMENT fits_bintable_poke, fh_or_DOL, row_number, col_desig, value, keeppcount=, chat=

   Substitutes the table bin with 'value' that can be an array if the
   bin is a vector bin.

   If row_number <= 0 then the entire column will be replaced
   provided 'value' is an array with as many elements as there are rows.
   For a fixed length vector column 'value' must be a 2D array (mxn)
   with 'm' as the number of values in each bin and 'n' equal to the
   number of rows in the table.
   Variable length vector columns must be updated one row at a time.

   'col_desig' is either the column number or its name.

   Keyword  keeppcount  If set, no changes are made to the FITS PCOUNT keyword.

   2007-09-26/NJW
   2008-01-18/NJW, updated with vector bin option
   2008-10-02/NJW, updated with variable length vector bin option
   2010-11-19/NJW, updated with more than one variable length vector bin
   2014-12-16/NJW, updated with TDIM keyword handling
*/
{

/*
   - overwrite single bin (row_number > 0) fixed length
      Accepts only DOL

   - overwrite entire column (row_number <= 0) fixed length
      Accepts only DOL

   - write single bin (row_number > 0) to variable length column
      When col_number points to variable length column

*/
   if( is_void(value) ) error,"Entered void value";
   if( is_void(chat) ) chat = 0;

   nvalue = numberof( value );
   dms = dimsof( value );
   if( dms(1) == 0 ) dms = [1,1]; // a scalar was input
  /*
   *    if row_number <= 0 then entire column is to be replaced
   *                       and in that case the last dimension must match
   *                       the number of rows already present in the table
   *    if row_number > 0  then a single bin is to be replaced and if the
   *                       input 'value' has more than a single dimension then
   *                       it must conform with the TDIM keyword
   */
   if( row_number <= 0 ) { // replacement of entire column
      nvrows = dms(0); // Last dimension
      nvector = 1;
      for(i=1;i<dms(1);i++) nvector *= dms(i+1);
   } else { // replacement of a single bin
      nvrows = 1;
      nvector = nvalue;
   }
   pcount = 0;

  /*
   * There are cases such a Linux on 64 bit machines where
   *  sizeof(long) == 8. This is not compatible with FITS J-type
   *  where 4 bytes are expected. Hence if 'value' is of type 'long'
   *  it must be converted to 'int'.
   */

   if( typeof(value) == "long" && sizeof(long) == 8 ) {
      value = int(value);
   }

   // Analyze first argument

   if( structof(fh_or_DOL) == list ) {
      is_dol = 0;
      fh = fh_or_DOL;
   } else if( structof(fh_or_DOL) == string ) {
      is_dol = 1;
      local filename, extno;
      get_exten_no, fh_or_DOL, filename, extno;
      // for a binary table extno must be > 0
      if( extno == 0 ) error,"No extension number found";
      if( !file_test(filename) ) error,filename+" not found!";
      fh = fits_open( filename, 'r');
      fits_goto_hdu, fh, extno+1; // Yorick counting starts with 1
   } else error,"First argument of illegal type";

   //  starting file

   btdes = fits_btdes( fh );

   // Number of columns in the extension
   ncols = _car( btdes, 2);
   if(chat)write,format="Number of columns in extension: %i\n", ncols;
   nrows = _car( btdes, 1);
   if(chat)write,format="Number of rows in extension: %i\n", nrows;

   // Identify the variable vector columns

   nreps = _car( btdes, 4);
   varcol = where( nreps == 0 );

   col_number = structof(col_desig) == string ? \
         fits_colnum( fh, col_desig ) : col_desig;
   if( col_number > ncols ) error,"Column number too high";
   if( row_number > nrows ) error,"Row number too high";

   //  Initial check for consistency on input data and TDIM

   tdims = _tdims( fh, col_number );
   nt = numberof(tdims);
   if( row_number <= 0 ) { // replacement of entire column
      if( nvrows != nrows ) {
         write,itoa(nvrows)+" rows in input data";
         write,itoa(nrows)+" rows in existing binary table";
         error,"Mismatching numbers of rows";
      }
      if( nt ) {
         // tdims( 2:0 ) must match dms( 2:-1 )
         if( tdims(1)+1 != dms(1) ) error,"##43##";
         if( anyof( tdims(2:0) - dms(2:-1) ) ) error,"##44##";
         nd = dms(2);
         for(i=3;i<=dms(1);i++) nd *= dms(i);
         value = reform(value, nd, nrows );
         dms = dimsof(value);
      }
   } else { // replacement of a single bin
      if( nt ) {
         // tdims must match dms
         if( tdims(1) != dms(1) ) error,"##45##";
         if( anyof( tdims - dms ) ) error,"##46##";
         value = reform( value, nvalue );
         dms = dimsof(value);
      }
   }

   //  Reorganize basic table information

   local sizecol; eq_nocopy, sizecol, _car(btdes,5);
   local repeats; eq_nocopy, repeats, _car(btdes,4);
   local dtypes; eq_nocopy, dtypes, _car(btdes,3);
   fixvmode = (repeats(col_number) > 1)&&(dtypes(col_number)!="A");
   if( chat > 1 ) {
      write,format="fixvmode (fixed vector bin mode) = %i\n", fixvmode;
   }

   // Check if requested column is variable length

   varvmode = !repeats(col_number); // repeats has value zero for var.len.
   if( chat > 1 ) {
      write,format="varvmode (variable length vector bin mode) = %i\n", varvmode;
   }

   if( row_number <= 0 ) row_number = 1;

   // new file (temporary)

   //+ filenameo = get_next_filename("_tmp_???.fits");
   filenameo = ranstr(10)+".fits";
   fho = fits_open( filenameo, 'w', overwrite=1 );

   // Test dimensional compatibility

   if( varvmode ) { // A scalar or vector is accepted

      if( dms(1) > 1 ) error,"Data dimensionality is illegal";
      if( dms(1) == 0 ) dms = [1,1];
      local lengths, positions;

      // find the highest address of the heap by searching all variable length
      //  columns

      position_of_insertion = 0;
      for( i = 1; i <= numberof(varcol); i++ ) {
         dtyp = dtypes(varcol(i));
         if( dtyp == "I" ) {
            colw = 2;
         } else if( dtyp == "J" ) {
            colw = 4;
         } else if( dtyp == "B" ) {
            colw = 1;
         } else if( dtyp == "E" ) {
            colw = 4;
         } else if( dtyp == "D" ) {
            colw = 8;
         } else {
            write,"Datatype "+dtyp+" is not supported";
            return;
         }
         _fits_var_length_info, fh, varcol(i), lengths, positions;
         p = max(colw*lengths + positions);
         if( p > position_of_insertion ) position_of_insertion = p;
      }

      // PCOUNT should reflect the size of the heap
      pcount = position_of_insertion;

        // relative to start of heap after ordinary table
      if( chat > 1 ) {
         write,format="Position of insertion of new data in heap: %i\n", position_of_insertion;
      }
      n_repl_rows = 1;

   } else {

      if( fixvmode ) { // This is a vector column hence there must be enough
                    // data values to fill it
         // first investigate if a format change is adequate
         if( dms(1) == 1 ) {
            if( dms(2) == repeats(col_number) ) {
               value = reform(value, dms(2),1);
               dms = dimsof(value);
            }
         }
         if( dms(1) != 2 ) error, "##25##";
         if( dms(2) != repeats(col_number) ) error,"##26##";
         if( dms(3) + row_number - 1 > nrows ) error,"##27##";
         n_repl_rows = dms(3);

      } else {  // We have a scalar column

         if( dms(2) + row_number - 1 > nrows ) error,"##28##";
         n_repl_rows = dms(2);

      }
   }

   // Calculate address where replacement begins


  //        begin of data section
   //+ address_row = _car(btdes,6)(3);
   rowsize = sizecol(sum);
   nbytes_before = col_number > 1 ? sum(sizecol(1:col_number-1)) : 0;
   nbytes_after = col_number < ncols ? sum(sizecol(col_number+1:0)) : 0;
   if( chat > 1 ) {
      write,format="nbytes_before : %10i\n", nbytes_before;
      write,format="nbytes_after  : %10i\n", nbytes_after;
   }

   address = _car(btdes,6)(3) \
           + (row_number-1) * sizecol(sum); // adding preceding rows
   if( chat > 2 ) {
      write,format="##100## row_number = %i, address = %i\n", row_number, address;
   }
   //+ if( col_number > 1 ) address += sum(sizecol(1:col_number-1));
   //+                    // adding first part of row

   sizec  = sizecol(col_number);
   dtype  = typeof( value );
   if( dtype == "string" ) {
      if( repeats(col_number) < max(strlen(value)) ) error,"String too long";
   }
   if(chat)write,format="Data type needs to be %s\n", dtypes(col_number);

   // Checking data type compatibility

   if( dtypes(col_number) == "B" ) {
      // A single byte is expected
      if( dtype != "char" ) error,"Data type not 'char'";
   } else if( dtypes(col_number) == "I" ) {
      // A size of two bytes is expected i.e. short
      if( dtype != "short" ) error,"Data type not 'short'";
      if( sizeof(short) != 2 ) error,"Short size incompatibility";
   } else if( dtypes(col_number) == "J" ) {
      if( dtype == "long" ) {
         if( sizeof(long) != 4 ) error,"Long size incompatibility - use 'int' instead";
      } else if( dtype == "int" ) {
         if( sizeof(int) != 4 ) error,"Int size incompatibility";
      } else error,"Data type neither int nor long";
   } else if( dtypes(col_number) == "E" ) {
      if( dtype != "float" ) error,"Data type not 'float'";
   } else if( dtypes(col_number) == "D" ) {
      if( dtype != "double" ) error,"Data type not 'double'";
   } else if( dtypes(col_number) == "A" ) {
      if( dtype != "string" ) error,"Data type not 'string'";
   } else error,"Unrecognized data type";

   vc = array(char,1); // defining byte for file copying

   //  Copy byte by byte until start address of first row
   //  where replacement must take place

   for( addr=0; addr < address; addr++ ) {
      _read, _car(fh,4), addr, vc;
      _write, _car(fho,4), addr, vc;
   }

   for( row = 1; row <= n_repl_rows; row++ ) {
      if( chat > 2 ) write,format="##101## row %i, address = %i\n", row, address;
      // Copy first part of row until replacement begins
      if( nbytes_before > 0 ) {
         for(i=0;i<nbytes_before;i++) {
            _read, _car(fh,4), address, vc;
            _write, _car(fho,4), address++, vc;
         }
      }
      if( chat > 2 ) write,format="##102## row %i, address = %i\n", row, address;

      if( varvmode ) {

         _write, _car(fho,4), address, dms(2); // Length of array
         address += 4;
         _write, _car(fho,4), address, position_of_insertion; // Position in heap
         address += 4;

      } else {

         // do the replacement

         if( fixvmode ) {
            for( i = 1; i <= repeats(col_number); i++ ) {
               _write, _car(fho,4), address, value(i,row);
               address += sizeof(value(i,row));
            }
         } else {
            if( dtypes(col_number) == "A" ) {
               stra = value(row);
               while( strlen(stra) < repeats(col_number) ) stra += " ";
               cstra = strchar( stra );
               for( i = 1; i <= repeats(col_number); i++ ) {
                  _write, _car(fho,4), address++, cstra(i);
               }
            } else {
               _write, _car(fho,4), address, value(row);
               address += sizec;
            }
         }
      }
      if( chat > 2 ) write,format="##103## row %i, address = %i\n", row, address;
      if( nbytes_after > 0 ) {
         for(i=0;i<nbytes_after;i++) {
            _read, _car(fh,4), address, vc;
            _write, _car(fho,4), address++, vc;
         }
      }
      if( chat > 2 ) write,format="##104## row %i, address = %i\n", row, address;
   }

   // Copy remainder of ordinary table data

   last_data_address = _car(fh,3)(3) + rowsize*nrows;

   while( address < last_data_address ) {
      _read, _car(fh,4), address, vc;
      _write, _car(fho,4), address++, vc;
   }

   // If data needs to be added to heap

   if( varvmode ) {

      // Copy already existing part of the heap

      for( i = 1; i <= position_of_insertion; i++ ) {
         _read, _car(fh,4), address, vc;
         _write, _car(fho,4), address++, vc;
      }

      // Add the new data values

      sizev = sizeof(value(1));
      for( i = 1; i <= dms(2); i++ ) {
         _write, _car(fho,4), address, value(i);
         address += sizev;
      }
   }

   if( varvmode ) {
      fits_close, fho, last=address;
   } else {
      // Copy remainder of the file
      while( _read( _car(fh,4), address, vc) == 1 ) _write, _car(fho,4), address++, vc;
      fits_close, fho, nopad=1;
   }

   fits_close, fh;

   // Update the PCOUNT keyword

   if( keeppcount ) {
      cp, filenameo, filename;
   } else {
      fits_ch_keyw,filenameo,"PCOUNT",pcount,outfile=filename;
   }
   remove, filenameo;

}

/* Function fits_btdes */

func fits_btdes( fh_or_DOL, chat= )
/* DOCUMENT btdes = fits_btdes( fh_or_DOL, chat= )

   returns Binary Table DEScription.

     num_rows = _car(btdes,1)
     num_cols = _car(btdes,2)
     data_types = _car(btdes,3) // string array of length num_cols
     repeats    = _car(btdes,4) // long array of repeat counts
     sizecols   = _car(btdes,5) // long array of length num_cols
     offset     = _car(btdes,6) // 5 element array as _car(fh,3)

    2007-09-25/NJW
*/
{
   local filename, extno;
   if( typeof(fh_or_DOL) == "string" ) {
      if(chat)write,"fits_btdes - arg is string";
      get_exten_no, fh_or_DOL, filename, extno;
      fh = fits_open( filename );
      fits_goto_hdu, fh, extno+1;
      is_dol = 1;
   } else if( typeof(fh_or_DOL) == "list" ) {
      if(chat)write,"fits_btdes - arg is list";
      fh = fh_or_DOL;
      is_dol = 0;
   } else error,"Argument is neither string nor list";

   local cards; eq_nocopy, cards, _car(fh,1);
   //+local offset; eq_nocopy, offset, _car(fh,3);
   offset = _car(fh,3);

   // current HDU
   if(chat)write,format="Current HDU: %i\n", offset(1);
   // Address in file of CHDU (Current HDU)
   if(chat)write,format="Address CHDU: %i\n", offset(2);
   // Address in file of data of CHDU
   if(chat)write,format="Address data CHDU: %i\n", offset(3);
   // Address in file of next HDU
   if(chat)write,format="Address next HDU: %i\n", offset(4);
   // Filemode
   if(chat)write,format="Filemode: %c\n", offset(5);
   // Number of columns
   ncols = fits_numcols(fh);
   if(chat)write,format="Number of columns: %i\n", ncols;

   naxis1 = fits_get(fh,"NAXIS1");
   naxis2 = fits_get(fh,"NAXIS2");
   if(chat)write,format="Number of rows: %i\n", naxis2;

   // Setup arrays for column size information

   sizecol = array(0,ncols);
   dtypes = array(string,ncols);
   repeats = array(0,ncols);

   // Walk through all columns to extract information

   for( col = 1; col <= ncols; col++ ) {
      d = fits_get( fh, swrite(format="TFORM%i",col) );
      // include possibility of variable length vector column
      if( strpart(d,1:2) == "1P" ) {
         repeat = 0; // signals variable length
         dtype = strpart(d,3:3); // Type of data in the heap (not in the column)
         sizecol(col) = 8; // Namely 2 long's (J)
      } else {
         // Format of TFORM:  NT where N is a number (can have more than one digit)
         //   and T is the type indicator: B, I, J, E, or D
         repeat = atoi(strpart( d, 1:-1 ));
         // if 'N' is left out then the result will be zero - it should be one
         if( repeat == 0 ) repeat = 1;
         dtype = strpart( d, 0:0 );
         if( dtype == "B" ) {
            sizecol(col) = repeat * 1;
         } else if ( dtype == "I" ) {
            sizecol(col) = repeat * 2;
         } else if ( dtype == "J" ) {
            sizecol(col) = repeat * 4;
         } else if ( dtype == "E" ) {
            sizecol(col) = repeat * 4;
         } else if ( dtype == "D" ) {
            sizecol(col) = repeat * 8;
         } else if ( dtype == "A" ) {
            sizecol(col) = repeat * 1;
         } else error,"Unrecognized data type: "+dtype;
      }
      dtypes(col) = dtype;
      repeats(col) = repeat;
   }
   if(chat)write,format="Summed size: %i, NAXIS1 = %i\n", \
           sizecol(sum), naxis1;

   btdes = _lst( naxis2, ncols, dtypes, repeats, sizecol, offset );

   if( is_dol ) fits_close, fh;

   return btdes;

}

/* Function get_crd */

func get_crd( dol )
/* DOCUMENT crd = get_crd( dol )

   Returns array crd(3,2)
      crd(1,1) = crpix1     crd(1,2) = crpix2
      crd(2,1) = crval1     crd(2,2) = crval2
      crd(3,1) = cdelt1     crd(3,2) = cdelt2

    2008-04-07/NJW
*/
{
   fh = headfits( dol );
   crpix1 = fxpar( fh, "crpix1" );
   crval1 = fxpar( fh, "crval1" );
   cdelt1 = fxpar( fh, "cdelt1" );
   crpix2 = fxpar( fh, "crpix2" );
   crval2 = fxpar( fh, "crval2" );
   cdelt2 = fxpar( fh, "cdelt2" );
   n_miss = 0;
   if( is_void(crpix1) ) n_miss++;
   if( is_void(crval1) ) n_miss++;
   if( is_void(cdelt1) ) n_miss++;
   if( is_void(crpix2) ) n_miss++;
   if( is_void(crval2) ) n_miss++;
   if( is_void(cdelt2) ) n_miss++;
   if( n_miss ) {
      write,format="%i missing CRPIX etc. in %s\n", n_miss, dol;
      write,format="Returning default %s\n","values";
      crpix1 = crpix2 = crval1 = crval2 = 0.0;
      cdelt1 = cdelt2 = 1.0;
   }
   crd = array(double, 3, 2 );
   crd(1,1) = crpix1;     crd(1,2) = crpix2;
   crd(2,1) = crval1;     crd(2,2) = crval2;
   crd(3,1) = cdelt1;     crd(3,2) = cdelt2;

   return crd;

}

/* Function fits_bintable_peek */

func fits_bintable_peek( DOL, row_number, col_desig, chat= )
/* DOCUMENT value = fits_bintable_peek( DOL, row_number, col_desig, chat= )

   'col_desig' is either the column number or its name.

   Returns the content of FITS binary table bin (row_number, col_number)

   2007-09-26/NJW
   2008-05-28/NJW updated to handle strings properly
   2008-10-03/NJW updated to handle variable length columns
   2014-12-16/NJW updated to handle the TDIM keyword

*/
{

   if( is_void(chat) ) chat = 0;
   local filename, extno;
   if( typeof(DOL) == "string" ) {
      isfile = 1;
      get_exten_no, DOL, filename, extno;
      if( extno == 0 ) error,"No extension number found";
      fh = fits_open( filename );
      fits_goto_hdu, fh, extno+1;
      btdes = fits_btdes( fh );
   } else error,"DOL is not a string";


   // Address in file of CHDU (Current HDU)
   if( chat > 2 )write,format="Address CHDU: %i\n", _car(fh,3)(2);
   // Address in file of data of CHDU
   if( chat > 2 )write,format="Address data CHDU: %i\n", _car(fh,3)(3);
   // Address in file of next HDU
   if( chat > 2 )write,format="Address next HDU: %i\n", _car(fh,3)(4);
   // Filemode
   if( chat > 3 )write,format="Filemode: %c\n", _car(fh,3)(5);
   // Number of columns
   ncols = _car( btdes, 2);
   if( chat > 0 ) write,format="Number of columns: %i\n", ncols;
   nrows = _car( btdes, 1);
   if( chat > 0 ) write,format="Number of rows: %i\n", nrows;

   col_number = structof(col_desig) == string ? \
         fits_colnum( fh, col_desig ) : col_desig;

   if( col_number > ncols ) error,"Column number too high";
   if( row_number > nrows ) error,"Row number too high";

   // Check for TDIM keyword

   tdims = _tdims( fh, col_number );

   // Calculate address

   local sizecol; eq_nocopy, sizecol, _car(btdes,5);
   local repeats; eq_nocopy, repeats, _car(btdes,4);
   local dtypes; eq_nocopy, dtypes, _car(btdes,3);

  //        begin of data section
   address = _car(fh,3)(3) \
           + (row_number-1) * sizecol(sum); // preceding rows
   if( col_number > 1 ) address += sum(sizecol(1:col_number-1));

   if( chat > 0) write,format="Reading %i%s at address %i\n", \
          repeats(col_number), dtypes(col_number), address;

   if( repeats(col_number) ) { // Fixed length column
      if( dtypes(col_number) == "I" ) {
         value = array( short, repeats(col_number));
         _read, _car(fh,4), address, value;
      } else if( dtypes(col_number) == "J" ) {
         if( sizeof(long) == 4 ) {
            value = array( long, repeats(col_number) );
         } else if( sizeof(int) == 4 ) {
            value = array( int, repeats(col_number) );
         } else error,"Incompatible datatype";
         _read, _car(fh,4), address, value;
      } else if( dtypes(col_number) == "B" ) {
         value = array( char, repeats(col_number) );
         _read, _car(fh,4), address, value;
      } else if( dtypes(col_number) == "E" ) {
         value = array( float, repeats(col_number) );
         _read, _car(fh,4), address, value;
      } else if( dtypes(col_number) == "D" ) {
         value = array( double, repeats(col_number) );
         _read, _car(fh,4), address, value;
      } else if( dtypes(col_number) == "A" ) {
         c = array(char,repeats(col_number));
         value = "";
         _read, _car(fh,4), address, c;
         for(i=1;i<=repeats(col_number);i++) {
            value += swrite(format="%c",c(i));
         }
      } else error,"Unrecognized data type";
      if( numberof(tdims) ) value = reform(value,tdims);
   } else { // variable length column
     // Get number of values in bin
     num_values = int(0);
     if( sizeof(num_values) != 4 ) error,"Bad size of int variable";
     _read, _car(fh,4), address, num_values;
     // Check for existence for values at all
     if( num_values == 0 ) return [];
     address += 4;
     // Get byte address in heap
     val_address = long(0);
     _read, _car(fh,4), address, val_address;

     // Read the actual values in the heap

     address = _car(fh,3)(3) + nrows * sizecol(sum) + val_address;
     if( dtypes(col_number) == "I" ) {
        value = array( short, num_values );
     } else if( dtypes(col_number) == "J" ) {
        if( sizeof(long) == 4 ) {
           value = array( long, num_values );
        } else if( sizeof(int) == 4 ) {
           value = array( int, num_values );
        } else error,"Incompatible datatype z";
     } else if( dtypes(col_number) == "B" ) {
        value = array( char, num_values );
     } else if( dtypes(col_number) == "E" ) {
        value = array( float, num_values );
     } else if( dtypes(col_number) == "D" ) {
        value = array( double, num_values );
     } else error,"Datatype not supported";

     _read, _car(fh,4), address, value;

   }

   if( isfile ) fits_close, fh;

   return value;

}

/* Function _fits_var_length_info */

func _fits_var_length_info( fh_or_DOL, colnum, &lengths, &positions, chat= )
/* DOCUMENT _fits_var_length_info, fh_or_DOL, colnum, >lengths, >positions, chat=

   Reads the vector lengths and heap addresses of a variable vector column
   and returns the values in the arguments.
   If first argument is a file handle then it must be opened for reading

    2008-01-02/NJW, 2010-11-19/NJW update with colnum for handling more than one
                                   variable length vector column
*/
{

   if( is_void(chat) ) chat = 0;

   if( structof(fh_or_DOL) == list ) {
      // Probably file handle
      fh = fh_or_DOL;
      if( _car(fh,3)(5) != 'r' ) error,"Not opened for reading";
      is_dol = 0;
   } else if( structof(fh_or_DOL) == string ) {
      // probably a DOL
      local filename, extno;
      get_exten_no, fh_or_DOL, filename, extno;

      fh = fits_open( filename, 'r');
      fits_goto_hdu, fh, extno+1;
      is_dol = 1;
   } else error,"Illegal first argument";

   btdes = fits_btdes( fh );

   // Identify position of variable length column information

   varcol = where( _car(btdes,4) == 0 );
   if( numberof(varcol) == 0 ) {
      write,"No variable length column found";
      if( is_dol ) fits_close, fh;
      return;
   }
   if( allof( varcol - colnum ) ) {
      write,"Requested colnum is not of variable length";
      if( is_dol ) fits_close, fh;
      return;
   }
   varcol = colnum;
   nrows = _car(btdes,1);
   ncols = _car(btdes,2);
   address_data = _car(fh,3)(3);
   lengths = positions = array( long, nrows );
   j = array( long, 1);

   nbytes_before = varcol == 1 ? 0 : long(sum(_car(btdes,5)(1:varcol-1)));
   nbytes_after = varcol == ncols ? 0 : long(sum(_car(btdes,5)(varcol+1:ncols)));

   address = address_data;
   for( row = 1; row <= nrows; row++ ) {
      address += nbytes_before;
      _read, _car(fh,4), address, j;
      lengths(row) = j;
      address += 4; // increase with 4 since datatype is J
      _read, _car(fh,4), address, j;
      positions(row) = j;
      address += 4;
      address += nbytes_after;
   }

   if( is_dol ) fits_close, fh;

}

/* Function pixsize */

func pixsize( list, outfile, silent= )
/* DOCUMENT pixsize, list, outfile, silent=

   Get the pixel size from the CDx_x keywords in FITS images
   as given in input string array 'list' of file names.
   For each files the first extension with valid CDx_x keywords is
   used and the remaining extensions are skipped.

   The result is written into file 'outfile'.

   2008-12-09/NJW
*/
{
   fout = open( outfile, "w" );
   n = numberof(list);
   for( i = 1; i <= n; i++ ) {
      next = nfits_extens( list(i) );
      for(j = 1; j <= next; j++ ) {
         fh = headfits( list(i)+"+"+itoa(j));
         cd1_1 = fxpar(fh,"cd1_1");
         cd1_2 = fxpar(fh,"cd1_2");
         cd2_1 = fxpar(fh,"cd2_1");
         cd2_2 = fxpar(fh,"cd2_2");
         if( is_void(cd1_1) || is_void(cd1_2) ) continue;
         pix1 = sqrt(cd1_1^2 + cd1_2^2);
         pix2 = sqrt(cd2_1^2 + cd2_2^2);
         if( abs(pix1-pix2) > 1.e-5 ) {
            write,format="Warning pixsizes in file: %s\n", list(i)+"+"+itoa(j);
         }
         if(!silent)write,format="%12.7f %s[%i]\n", pix1, list(i), j;
         write,fout,format="%12.7f %s[%i]\n", pix1, list(i), j;
         break; // assume all other extensions have same pixel size
      }
   }
   close, fout;
}

/* Function fits_add_cols */

func fits_add_cols( filename, colname, coldata, .. , nob= )
/* DOCUMENT  fits_add_cols, filename, colname, coldata, .., nob=

   Add columns to an existing FITS binary table file in
   extension number 1. The number of rows is kept.
   After the filename the arguments are alternating "colname, coldata"
   'colname' will be converted to upper case.

   Keyword 'nob' (no back up) prevents the automatic backup

   SEE ALSO: fits_add_rows

   2009-05-07/NJW

*/
{
   if( numberof(coldata) < 1 ) {
      write,format="No data present in first argument, skip writing%s\n","";
      return;
   }
   if( !file_test( filename ) ) {
      write,format="fits_add_cols: %s does not exist, quit!\n", filename;
      return;
   }
   if( is_void(nob) ) nob = 0;

//  read all data in existing file and find column names; close file afterwards

   old_fh = fits_open( filename );
   fits_goto_hdu, old_fh, 2; //  Extension 1 (Yorick counting starts at 1)
   ptr = fits_read_bintable( old_fh );
   fits_close, old_fh;
   olddms = dimsof(*ptr(1));
   nrows = olddms(2); // '2' because of internal representation if
                      // column 1 has vector bins

// Make a backup copy of old file and delete it

   if( !nob ) back, filename;
   remove, filename;

// Initiate the new version of the file

   fh = fits_create(filename);
   fh = fits_set(fh,"EXTEND",'T',"There may be extensions");
   fits_write_header, fh;

   fits_new_bintable, fh;

// copy all existing keywords

   fits_copy_keys, old_fh, fh;

//  get existing column names to avoid duplicating a name
//  and update 'fh'

   ncols = dimsof( ptr )(2);
   old_colname = array(string,ncols);
   for( i = 1; i <= ncols; i++ ) {
      keyname = swrite(format="TTYPE%i",i);
      cname = fxpar( old_fh, keyname );
      if( is_void(cname) ) {
         write,format="##2## problem with column name #%i\n", i;
         return;
      } else old_colname(i) = cname;
      fits_set,fh,keyname,cname,"Name of column";
   }

//   Check consistency between the given arguments and the existing file

   colname = strupcase(colname);
   if( anyof(old_colname == colname) ) {
      write,format="Colname %s exists already in %s\n", colname, filename;
      return;
   }
   grow, old_colname, colname;

   dms = dimsof(coldata);
   dms1 = dms(1);
   if( nrows != dms(1+dms1) ) {
      write,"Column %s does not match existing columns\n", colname;
      return;
   }

   if( dms(1) == 2 ) coldata = transpose(coldata);
   grow, ptr, &coldata;
   keyname = swrite(format="TTYPE%i",++ncols);
   fits_set, fh, keyname, colname,"Name of column";

   while( more_args() ) {
      cname = strupcase(next_arg());
      cdata = next_arg();

//   Check consistency between the given arguments and the existing file

      if( anyof(old_colname == cname) ) {
         write,format="Colname %s exists already in %s\n", cname, filename;
         return;
      }
      grow, old_colname, cname;

      dms = dimsof(cdata);
      dms1 = dms(1);
      if( nrows != dms(1+dms1) ) {
         write,"Column %s does not match existing columns\n", cname;
         return;
      }

      if( dms(1) == 2 ) cdata = transpose(cdata);
      grow, ptr, &cdata;
      keyname = swrite(format="TTYPE%i",++ncols);
      fits_set, fh, keyname, cname,"Name of column";

   }

   kwds_put, fh;

   fits_write_bintable, fh, ptr;
   fits_close, fh;

}

/* Function ttype */

func ttypes( dol_or_fh )
/* DOCUMENT ttypes, dol_or_fh
      list = ttypes( dol_or_fh )

   prints or returns the list of column names
   2009-05-25/NJW
*/
{
   if( typeof(dol_or_fh) == "string" ) {
      fh = headfits(dol_or_fh);
   } else fh = dol_or_fh;
   xtension = fxpar(fh,"xtension");
   if( xtension != "BINTABLE" ) error,"Not a binary table";

   i = 0;
   res = [];
   while( !is_void((colname = fxpar(fh,swrite(format="TTYPE%i",++i))))) {
      if( am_subroutine() ) {
         write,colname;
      } else grow, res, colname;
   }
   return res;
}

/* Function fits_info_col */

func fits_info_col( fh_or_dol, outfile= )
/* DOCUMENT fits_info_col, fh_or_dol, outfile=

    Displays column information.
    Keyword 'out' is an output file name.
*/
{
   fout =  structof(outfile) == string ? open(outfile,"w") : [];

   if( structof(fh_or_dol) == string ) {
      fh = headfits( fh_or_dol );
   } else fh = fh_or_dol;

   hdu = listhead( fh );
   ncols = fits_numcols( fh );

   for( i = 1; i <= ncols; i++ ) {
      ttype = fits_get( fh, swrite(format="TTYPE%i",i));
      tform = fits_get( fh, swrite(format="TFORM%i",i));
      tunit = fits_get( fh, swrite(format="TUNIT%i",i));
      write,fout,format="%3i  %16s %10s", i, ttype, tform;
      if( is_void(tunit) ) {
         write,fout,"";
      } else write,fout,format="%10s\n",tunit;
   }
   close,fout;

}

/* Function fits_ch_keyw */

func fits_ch_keyw( filename, keyword, new_value, outfile=, chat= )
/* DOCUMENT fits_ch_keyw, filename, keyword, new_value, outfile=, chat=

    Will change the FITS keyword 'keyword' to the 'new_value'
    in all extensions.
    ('keyword' will be converted to upper case)

    Keyword outfile : If given the new file will be written to this,
                      else the original file will be overwritten.

    2010-11-29/NJW

*/
{
   extern fits_lng_str_lim;

   // Get all keywords in all extensions as a first pass

   all_cards = [];
   num_extensions = nfits_extens( filename );
   for( i = 0; i <= num_extensions; i++ ) {
      hdr = headfits( filename+"+"+itoa(i) );
      grow, all_cards, _car(hdr,1);
   }
   // test for the use of the LONGSTRN convention
   use_longstrn = anyof( strmatch(strpart(all_cards,1:8), "LONGSTRN"));

   // test for the existence of the keyword in question
   keyword = strupcase(keyword);
   len = strlen(keyword);
   if( noneof( (m = strmatch(strpart(all_cards,1:len), keyword ))) ) {
      write,"FITS_CH_KEYW: Keyword "+keyword+" was not found - skip further action";
      return;
   }

   // Test for CONTINUE keywords following matches
   w = where(m); nw = numberof(w);
   mc = strmatch(strpart(all_cards,1:8), "CONTINUE" );
   nmin_cont = 9999;
   for( i = 1; i <= nw; i++ ) {
      count = 0;
      while( mc(w(i)+1+count) ) count++;
      if( count < nmin_cont ) nmin_cont = count;
   }
   we_have_cont = 0; // initialize

   /*
    *    nmin_cont      smallest number of contiguous CONTINUE cards
    *                   appearing right after the matches with the
    *                   actual keyword in the list of all keywords
    *    n_cont         number of required CONTINUE cards following the
    *                   actual keyword
    *    cstr_cont      char representation of "CONTINUE"
    *    we_have_cont   a flag which is set when the 'keyword' has been
    *                   encountered AND the 'new_value' is a long string
    */

   // Convert keyword to char array for the search

   ckeyword = *pointer(keyword);
   ckeyword = ckeyword(1:-1); // remove trailing '\0'
   nckeyword = numberof(ckeyword);
   cstr_cont = (*pointer("CONTINUE"))(1:-1);

   // Convert new_value to a string

   n_cont = 0; // only changed if a long string is requested
   if( structof(new_value) == string ) { // Already a string - test the length
      len = strlen(new_value);
      ncards_required = (len -1)/fits_lng_str_lim + 1;
      if( len == fits_lng_str_lim+1 ) ncards_required = 1; // if just above limit one card is enough
      n_cont = ncards_required - 1;
      if( n_cont ) { // test if the LONGSTRN convention has been adapted
         if( !use_longstrn ) {
            write,"FITS_CH_KEYW: The new string value is too long for a keyword - skip further action";
            return;
         } else {
            if( n_cont > nmin_cont ) {
               write,"FITS_CH_KEYW: Too few CONTINUE keywords follow the keyword - skip further action";
               return;
            }
         }
         new_value_str = "'"+strpart(new_value, 1:fits_lng_str_lim)+"&'";
         cont_remainder = strpart(new_value, fits_lng_str_lim+1: );
         cont_str = array(string,n_cont);
         for( i = 1; i <= n_cont; i++ ) {
            cont_str(i) = "'"+strpart(cont_remainder, 1:fits_lng_str_lim);
            if(i == n_cont ) cont_str(i) += "'";
            else cont_str(i) += "&'";
            cont_remainder = strpart(cont_remainder, fits_lng_str_lim+1: );
         }
      } else {
         new_value_str = "'"+new_value+"'";
      }
   } else if( structof(new_value) == int || structof(new_value) == long ) {
      new_value_str = swrite(format="%i",new_value);
   } else if( structof(new_value) == float || structof(new_value) == double ) {
      new_value_str = swrite(format="%g",new_value);
   } else error,"Type of new value: "+typeof(new_value)+" is not supported";
   cnew_value_str = *pointer(new_value_str);
   cnew_value_str = cnew_value_str(1:-1);
   ncnew_value_str = numberof(cnew_value_str);

   block = array(char,2880);
   stream = open(filename,"rb");
   // choose a random string as output filename to avoid name conflicts
   // [well, they will happen, eventually, but the probability is VERY low]
   filenameo = ranstr(10)+".fits";
   ostream = open( filenameo, "wb" );

   size = sizeof(stream);
   nblocks = size/2880;
   //+ write,"size = ", size;
   //+ write,"nblocks = ", nblocks;

/*
 *  Go through the input file block by block searching for the keyword
 */

   address = 0;
   for( iblock = 1; iblock <= nblocks; iblock++ ) {
     /*
      *  Read the block
      */
      _read, stream, address, block;
      //+ write,"Has read block#", iblock,"  at address ", address;
      for( j = 1; j <= 36; j++ ) {
         card = block((j-1)*80+1:j*80);
         // Do we have a string for a CONTINUE keyword?
         if( we_have_cont > 0) {
            if( noneof( cstr_cont - card(1:8) ) ) { // yes, this is CONTINUE
               newcard = grow( cstr_cont, ' ', ' ', (*pointer(cont_str(n_cont - (--we_have_cont))))(1:-1));
               n_newcard = numberof(newcard);
               if( n_newcard != 80 ) {
                  if( n_newcard > 80 ) {
                     write,"FITS_CH_KEYW: Trouble A n_newcard = "+itoa(n_newcard);
                     newcard = newcard(1:80);
                  } else {
                     newcard = grow( newcard, array(' ',80-n_newcard));
                  }
               }
               block((j-1)*80+1:j*80) = newcard;
            } else {
               write,"FITS_CH_KEYW: Trouble B mess up with we_have_cont";
               we_have_cont = 0;
               n_cont = 0;
            }
            continue; // skip the rest of the j-loop since we have dealt with it
         }
         
        /*
         *   Check for occurrence of string "keyword" which is
         *   unlikely unless in header block
         */
         if( noneof( card(1:nckeyword) - ckeyword) ) {  // Match!!!!
            // report original header card
            card0 = card;
            grow,card0,'\0';
            sline = string(&card0);
            if( chat ) write,"Original card with a match :";
            if( chat ) write,sline;
            newcard = card;
            // clear the new card by replacing with blanks
            for( i = 10; i <= 80; i++ ) newcard(i) = 32;

            // find where the comment starts
            for( islash = 80; islash > 0; islash-- ) {
               if( card(islash) == 47 ) break;
            }
            if( islash ) {
               comment = card(islash:80); // including the slash
               ncomment = numberof(comment);
            } else comment = []; // no comment found

            rmost = ncnew_value_str <= 20 ? 30 : 30 + (ncnew_value_str - 20);
            fpos = rmost - ncnew_value_str + 1;

            // Insert the string representation of the keyword
            for( i = fpos; i <= rmost; i++ ) newcard(i) = cnew_value_str(i-fpos+1);

            // Insert the comment if it has been found
            if( ncomment ) {
               i = rmost+2;
               k = 1;
               while( i <= 80 && k <= ncomment ) {
                  newcard(i++) = comment(k++);
               }
            }
            newcard0 = newcard;
            grow, newcard0, '\0';
            snewcard = string(&newcard0);
            if( chat ) write,"            New card :";
            if( chat ) write,snewcard;
            block((j-1)*80+1:j*80) = newcard;
            we_have_cont = n_cont;
         } else {                         // end of dealing with a match
            we_have_cont = 0; // no match, hence no continuation is required
         }
      } // end of search in block
      _write,ostream,address,block;
      address += 2880;
   } // end of loop over all blocks

   close, stream;
   close,ostream;

   if( !is_void( outfile ) ) filename = outfile;
   cp, filenameo, filename;
   remove, filenameo;
   remove, filenameo+"L";

}

/* Function fits_coldump */

func fits_coldump( dol, colname, .., rowsel=, outfile= )
/* DOCUMENT fits_coldump, dol, colname, .., rowsel=, outfile=

    Makes a nice print of the data in the chosen columns and
    selected rows.

    Argument: colname  Either a column name or a column number. Several such arguments
                       can be given.

    Keyword rowsel : Row selection by row number either as integer array
                     or as a text string e.g. "1-10,20-25".
            outfile: (string) If not given, then only terminal output.
                     If given the output stream is only to the file
                     unless preceded by a "+" sign e.g. "+mycols.txt"
                     in which case the columns will be written to the
                     terminal and to the file "mycols.txt".

*/
{
   p = 0; ps = 1;
   if( typeof(outfile) == "string" ) {
      if( strpart(outfile,1:1) == "+" ) {
         outfile = strpart( outfile, 2:0 );
      } else ps = 0;
      p = 1;
      fp = open(outfile,"w");
      write,fp,format="// Dump from %s\n", dol;
   }

   hdr = headfits(dol); // just in case
   ptr = &rdfitscol(dol,colname);
   if( structof(colname) != string ) colname = fxpar( hdr, "TTYPE"+itoa(colname));
   colname = strupcase(colname);
   ncols = 1;
   nrows_orig = numberof(*ptr);
   headline = strpadd(colname,12," ",rj=1,truncate=1);

   if( is_void(rowsel) ) {
      rowsel = indgen(nrows_orig);
   } else {
      if( structof(rowsel) == string ) rowsel = str2arr(rowsel);
   }
   nrows = numberof(rowsel);

   while( more_args() ) {
      coln = next_arg();
      if( structof(coln) != string ) coln = fxpar( hdr, "TTYPE"+itoa(coln));
      coln = strupcase(coln);
      headline += strpadd(coln,12," ",rj=1,truncate=1);
      grow,ptr,&rdfitscol(dol,coln);
      ncols++;
   }

   if(ps)write,format="   %s\n", headline;
   if(p)write,fp,format="   %s\n", headline;
   for( j = 1; j <= nrows; j++ ) {
      row = rowsel(j);
      if( row > nrows_orig ) continue;
      if(ps)write,format="%3i",row; // row number
      if(p)write,fp,format="%3i",row; // row number
      for( i = 1; i <= ncols; i++ ) {
         v = (*ptr(i))(row);
         sv = structof(v);
         if( sv == int || sv == long ) {
            if(ps)write,format="%12i", v;
            if(p)write,fp,format="%12i", v;
         }
         if( sv == float || sv == double ) {
            if(ps)write,format="%12.6f", v;
            if(p)write,fp,format="%12.6f", v;
         }
         if( sv == string ) {
            if(ps)write,format="%s", strpadd(strtrim(v),12," ",rj=1,truncate=1);
            if(p)write,fp,format="%s", strpadd(strtrim(v),12," ",rj=1,truncate=1);
         }
      }
      if(ps)write,"";
      if(p)write,fp,"";
   }

   if( p ) {
      write,"Results in "+outfile;
      close,fp;
   }
}

/* Function fits_binsplit */

func fits_binsplit( dol, nparts, stem=, colname= )
/* DOCUMENT filenames = fits_binsplit( dol, nparts, stem=, colname= )

    Splits the binary FITS table pointed to by 'dol' into
    'nparts' new files with names <stem>_nnnn.fits.
    If 'colname' has been given then the rows are split into
    files with equal sum of values in the 'colname' column,
    otherwise the splitting is on an equal size basis i.e.
    saving approx. NAXIS2/nparts rows into each.

    The default value for 'stem' is "split".

    Returns the list of the filenames.

    Variable length records are not supported.

    2012-02-10/NJW
*/
{
   local fname, extno;

   get_exten_no, dol, fname, extno;
   if( !file_test(fname) ) error,"File not found: "+fname;

   if( typeof(stem) != "string" ) stem = "split";
   hdr = headfits(dol);
   nrows = fxpar(hdr,"naxis2");
   if( nrows < nparts ) { // adjust requested nparts
      nparts = nrows;
      colname = []; // suppress this option since it makes no sense
   }

   // read all data
   ptr = rdfitsbin( dol, hdr, nrows );
   nptr = numberof(ptr);
   newptr = array(pointer, nptr);
   names = stem+"_"+itoa(indgen(nparts),4)+".fits";

   kept = []; // used to keep track of which selections produced a file
   if( typeof(colname) == "string" ) {
      colnum = fits_colnum( hdr, colname );
      if( is_void(colnum) ) error,"No such column: "+colname;
      accu =  double((*ptr(colnum))(psum));
      accu /= accu(0); // Normalize
      fract = double(indgen(nparts)) / nparts;
      for( i = 1; i <= nparts; i++ ) {
         if( i == 1) {
            w = where( accu <= fract(1) );
         } else {
            w = where( accu > fract(i-1) & accu <= fract(i) );
         }
         if( numberof(w) == 0 ) continue; // skip file writing
         write,itoa(i)+" rows "+itoa(w(1))+" - "+itoa(w(0));
         for( j = 1; j <= nptr; j++ ) {
            x = (*ptr(j))(w,..);
            newptr(j) = &x;
         }
         wrfitsbin, names(i), hdr, newptr, clobber=1;
         grow, kept, i;
      }

   } else { // splitting by numbers

      // the files must at least have nshare rows each
      nshare = nrows / nparts;
      r = nrows - nshare*nparts; // remainder
      // the remainder is 'given' to the first r files, one each
      row2 = 0;
      for( i = 0; i < nparts; i++ ) {
         row1 = row2 + 1;
         row2 = row1 + nshare - 1;
         if( i < r ) row2++;
         write,itoa(i)+" rows "+itoa(row1)+" - "+itoa(row2);
         m = indgen(row1:row2);
         for( j = 1; j <= nptr; j++ ) {
            x = (*ptr(j))(m,..);
            newptr(j) = &x;
         }
         wrfitsbin, names(i+1), hdr, newptr, clobber=1;
      }
      kept = indgen(nparts);
   }
   return names(kept);
}

/* Function fstrselect */

func fstrselect( dol, outfile, colid, str )
/* DOCUMENT fstrselect, dol, outfile, colid, str

   Selection of a FITS column that is of type 'string'.

   Makes a 'strmatch( colid, str)' and outputs to outfile

   'colid' can be a column number or a column name.

*/
{
   local fh, nrows;

   ptr = rdfitsbin( dol, fh, nrows );
   if( structof(colid) == string ) colid = fits_colnum(fh,colid);
   w = where( strmatch( *ptr(colid) , str ) );
   if( numberof(w) == 0 ) {
      write,"No action, no rows were selected!";
      return;
   }
   n = numberof(ptr);
   newptr = array(pointer,n);
   for( i = 1; i <= n; i++ ) newptr(i) = &((*ptr(i))(w));
   wrfitsbin, outfile, fh, newptr, clobber=clobber;
   write,outfile+" has been written.";

}

/* Function fselect */

func fselect( dol, outfile, expres )
/* DOCUMENT fselect, dol, outfile, expres
    or      idx = fselect( dol, "-", expres )

    The expression 'expres' can be a scalar string that is valid
    as an argument for 'where' with column names as variables
    ('#row' can be used as rownumber (as in the ftool 'fselect')).
    In this case a hyphen in a column name is not allowed.

    The other option for 'expres' is a two element array [row1, row2]
    (including both) with start and end row numbers
    and the selection will be between those.

    If 'outfile' is given as "-" then an index array is returned.

    2012-02-13/NJW

*/
{
   local hdr, nrows, fitsname, extno;

   get_exten_no, dol, fitsname, extno;
   if( !file_test(fitsname) ) error,"Did not find "+fitsname;
   hdr = headfits( dol );

   if( typeof(expres) == "string" ) { // expects a valid logical expression
      // replace occurrences of #row with 'r_n_u_m' (an unlikely column name)
      // since #row cannot be a variable name
      expres = strstrrepl( expres, "#row", "r_n_u_m" );
      // get column names embedded in the expression, including e.g. numbers
      names = parse_expres( expres );
      // match column names against those that actually are found in the header
      colnames = check_colnames( hdr, names );
      ncolnames = numberof(colnames);
      // Use column names to define variables in new function that is written
      // to a file for including by 'require' function.
      // A random file name is used in order to force 'require' to update for
      // every call.
      tmpname = ranstr(10,lc=1)+".tmp";
      ff = open(tmpname,"w");
      write,ff,format="%s\n","func ffsel( ffdol )";
      write,ff,format="%s\n","{";
      write,ff,format="%s\n","   local ffhdr;";
      //+ write,ff,format="%s\n","   write,\"ffsel DOL: \"+ffdol";
      write,ff,format="%s\n","   ptr = rdfitsbin( ffdol, ffhdr );"
      write,ff,format="%s\n","   r_n_u_m = indgen(dimsof(*ptr(1))(2));"
      for( i = 1; i <= ncolnames; i++ ) {
         write,ff,format="   if( \"%s\" != \"r_n_u_m\" ) %s = *ptr(fits_colnum( ffhdr, \"%s\"));\n",colnames(i),colnames(i),colnames(i);
      }
      write,ff,format="   w = where(%s);\n", expres;
      write,ff,format="%s\n","   return w;";
      write,ff,format="%s\n","}";
      close, ff;
      require,tmpname;
      //+ write,"About to call ffsel with DOL : "+dol;
      m = ffsel( dol );
      // Clean up after use.
      remove, tmpname;

   } else {   // expect selection by row numbers : from, to (included)

      if( numberof(expres) != 2 ) error,"Bad expression";
      expres = long(expres);
      m = indgen(expres(1):expres(2));

   }

   if( numberof(m) ) {

      if( outfile == "-" ) {
         return m;
      } else {
         ptr = rdfitsbin( dol, hdr, nrows );
         newptr = array(pointer, numberof(ptr));

         for( i = 1; i <= numberof(ptr); i++ ) {
            x = (*ptr(i))(m,..);
            newptr(i) = &x;
         }

         wrfitsbin, outfile, hdr, newptr, clobber=1;
      }

   } else {

      write,"No action since the selection is empty.";
      if( outfile == "-" ) return [];

   }

}

/* Function parse_expres */

func parse_expres( expres )
{
   syms = "<>=!&|*+-/";
   s = strcompress( expres, all=1 ); // remove all blanks
   names = [];
   t = strtok( s, syms );

   while( strlen(t(2)) ) {
      //+ grow, names, strlowcase(t(1));
      //+   why a 'strlowcase' here??
      grow, names, t(1);
      t = strtok( t(2), syms );
   }
   return grow(names,t(1));
}

/* Function check_colnames */

func check_colnames( hdr, names )
/* DOCUMENT colnames = check_colnames( hdr, names )

     Returns an array of column names where '#row' has been
     replaced with 'r_n_u_m' for the 'fselect' function.

    SEE ALSO: fselect
*/
{
   n = numberof(names);
   colnames = [];
   for( i = 1; i <= n; i++ ) {
      colnum = fits_colnum( hdr, names(i), silent=1 );
      if( colnum ) grow, colnames, names(i);
      if( names(i) == "r_n_u_m" ) grow, colnames, "r_n_u_m";
   }
   return colnames;
}

/* Function get_wcs */

func get_wcs( dol_or_hdu, nocheck=, t_opt= )
/* DOCUMENT coords = get_wcs( dol_or_hdu, nocheck=, t_opt= )

    Returns an instance of the struct 's_Coords' as defined in mfits.i
    from a FITS file header unit, either by DOL or by header directly

    coords.flag is set to 1 if all information is available, 0 otherwise.

    Keyword 'nocheck' is mirrored to 'headfits'; can be set if unexpected
    error message appears.

    Keyword 't_opt' can be set to the column number for translation
       TCDLT<t_opt>   equiv    CDELT1
       TCDLT<t_opt>+1 equiv    CDELT2
       etc. for TCRVL, TCTYP, TCRPX, 

    SEE ALSO: put_wcs

    2012-08-17/NJW
*/
{
   if( typeof(dol_or_hdu) == "string" ) { // we have a DOL
      hdr = headfits(dol_or_hdu,nocheck=nocheck);
   } else hdr = dol_or_hdu;

   coords = s_Coords();
   coords.flag = 0;

   CTYPE = "ctype"; ONE = "1"; TWO = "2";
   CRVAL = "crval"; CRPIX = "crpix"; CROTA2 = "crota2";
   CDELT = "cdelt";
   if( t_opt ) {
      CTYPE = "tctyp"; ONE = itoa(t_opt); TWO = itoa(t_opt+1);
      CRVAL = "tcrvl"; CRPIX = "tcrpx"; CROTA2 = "crota2";
      CDELT = "tcdlt";
   }

   if( is_void((ctype1 = fxpar(hdr,CTYPE+ONE))) ) {write,"CTYPE1 keyword missing"; return coords;}
   if( is_void((ctype2 = fxpar(hdr,CTYPE+TWO))) ) {write,"CTYPE2 keyword missing"; return coords;}
   if( is_void((crval1 = fxpar(hdr,CRVAL+ONE))) ) {write,"CRVAL1 keyword missing"; return coords;}
   if( is_void((crval2 = fxpar(hdr,CRVAL+TWO))) ) {write,"CRVAL2 keyword missing"; return coords;}
   if( is_void((crpix1 = fxpar(hdr,CRPIX+ONE))) ) {write,"CRPIX1 keyword missing"; return coords;}
   if( is_void((crpix2 = fxpar(hdr,CRPIX+TWO))) ) {write,"CRPIX2 keyword missing"; return coords;}
   cd1_1  = fxpar(hdr,"cd1_1");
   if( is_void(cd1_1) ) { // This may be oldfashioned CDELT etc.
      if( is_void((cdelt1 = fxpar(hdr,CDELT+ONE))) ) {write,"CDELT1 keyword missing"; return coords;}
      if( is_void((cdelt2 = fxpar(hdr,CDELT+TWO))) ) {write,"CDELT2 keyword missing"; return coords;}
      //+ if( is_void((crota2 = fxpar(hdr,"crota2"))) ) {write,"CROTA2 keyword missing, reset!"; crota2 = 0.0;}
      if( is_void((crota2 = fxpar(hdr,CROTA2))) ) { crota2 = 0.0;}
      cd1_1 =  cdelt1 * cos(crota2*pi/180.);
      cd2_1 =  cdelt1 * sin(crota2*pi/180.);
      cd1_2 = -cdelt2 * sin(crota2*pi/180.);
      cd2_2 =  cdelt2 * cos(crota2*pi/180.);
   } else {
      if( is_void((cd1_2  = fxpar(hdr,"cd1_2"))) ) {write,"CD1_2 keyword missing"; return coords;}
      if( is_void((cd2_1  = fxpar(hdr,"cd2_1"))) ) {write,"CD2_1 keyword missing"; return coords;}
      if( is_void((cd2_2  = fxpar(hdr,"cd2_2"))) ) {write,"CD2_2 keyword missing"; return coords;}
      cdelt1 = -sqrt(cd1_1^2 + cd2_1^2);
      cdelt2 =  sqrt(cd2_2^2 + cd1_2^2);
      crota2 = atan(cd2_1, cd2_2) * 180./pi;
   }

   coords.flag = 1; // signal OK values returned
   coords.ctype1 = ctype1;
   coords.ctype2 = ctype2;
   coords.crval1 = crval1;
   coords.crval2 = crval2;
   coords.crpix1 = crpix1;
   coords.crpix2 = crpix2;
   coords.cd1_1 = cd1_1;
   coords.cd1_2 = cd1_2;
   coords.cd2_1 = cd2_1;
   coords.cd2_2 = cd2_2;
   coords.cdelt1 = cdelt1;
   coords.cdelt2 = cdelt2;
   coords.crota2 = crota2;
   coords.scale = sqrt(cd1_1^2 + cd1_2^2);

   return coords;
}

/* Function put_wcs */

func put_wcs( coords, old= )
/* DOCUMENT put_wcs, coords, old=

     Adds the appropriate keywords CRVAL, CRPIX, CTYPE, CD1_1, .. to the
     keywords as accessed by functions kwds_set etc.

     'coords' must be an instance of the struct 's_Coords'.

     Keyword
      old      (bool) When set then CDELT1, CDELT2, and CROTA2 will be
                      written.

     SEE ALSO: get_wcs

*/
{
   kwds_set,"CTYPE1", coords.ctype1,"Coordinate system 1. axis";
   kwds_set,"CTYPE2", coords.ctype2,"Coordinate system 2. axis";
   kwds_set,"CRVAL1", coords.crval1,"Reference value coordinate 1";
   kwds_set,"CRVAL2", coords.crval2,"Reference value coordinate 2";
   kwds_set,"CRPIX1", coords.crpix1,"Reference pixel coordinate 1";
   kwds_set,"CRPIX2", coords.crpix2,"Reference pixel coordinate 2";
   kwds_set,"CUNIT1","deg","Unit of reference value coordinate 1";
   kwds_set,"CUNIT2","deg","Unit of reference value coordinate 2";
   if( old ) {
      kwds_set,"CDELT1", coords.cdelt1,"[deg] Pixel size in coordinate 1";
      kwds_set,"CDELT2", coords.cdelt2,"[deg] Pixel size in coordinate 2";
      kwds_set,"CROTA2", coords.crota2,"[deg] Rotation angle";
   } else {
      kwds_set,"CD1_1", coords.cd1_1,"[deg] Part of conversion matrix";
      kwds_set,"CD1_2", coords.cd1_2,"[deg] Part of conversion matrix";
      kwds_set,"CD2_1", coords.cd2_1,"[deg] Part of conversion matrix";
      kwds_set,"CD2_2", coords.cd2_2,"[deg] Part of conversion matrix";
   }
}

/* Function im2skyfits */

func im2skyfits( im, outfile, coords=, ra=, dec=, rot=, scale=, crpix1=, crpix2=, old=, clobber= )
/* DOCUMENT im2skyfits, im, outfile, coords=, ra=, dec=, rot=, scale=, crpix1=, crpix2=, old=, clobber=
      or    coords = im2skyfits( imdim, ra=, dec=, rot=, scale=, crpix1=, crpix2= )

     RA, dec, rot, and scale in degrees

    Subroutine version: Writes image to disk in FITS format in extension #1
    Function   version: Returns the struct s_Coords() with appropriate values
                        'imdim' is either a 2D array or a 2-element 1D integer array

    Coordinate type is "RA---TAN" and "DEC--TAN"

    Argument:
        im     (in subroutine version) A 2D image to be written to a FITS file
     imdim     (in function version) A 2-element array of integers with image dimensions

    Keywords
      coords   An instance of the struct 's_Coords'; if given then the other coordinate
               keywords will be ignored.
      ra       Right Ascension in degrees (coords.crval1)
      dec      Declination     in degrees (coords.crval2)
      rot      Rotation in degrees (coords.crota2)
      scale    pixel scale in degrees (coords.scale)
      crpix1   Reference pixel number (coords.crpix1)
      crpix2   Reference pixel number (coords.crpix2)
     clobber   Remove old file before (re-)writing
         old   Write coordinate keywords to the FITS file in CDELT+CROTA2 style

    2012-09-12/NJW

*/
{
   dr = pi / 180.;
   dms = dimsof(im);
   if( am_subroutine() ) {
      if( dms(1) != 2 ) error,"IM2SKYFITS: ##1##";
   } else {
      /*
       *  dms(1) == 2 : Image, no change
       *  dms(1) == 1 && dms(2) == 2 : 1D dimension array with 2 elements
       */
      if( dms(1) != 2 ) { // check for 1D 2 element array and create 'dms'
         if( dms(1) != 1 ) error,"IM2SKYFITS: ##2##";
         if( dms(2) != 2 ) error,"IM2SKYFITS: ##3##";
         dms = grow(2,im);
      }
   }
   if( typeof(coords) != "struct_instance" ) {

      if( is_void(ra) ) ra = 0.;
      if( is_void(dec) ) dec = 0.;
      if( is_void(rot) ) rot = 0.;
      if( is_void(scale) ) scale = 1./60.; // defaults to 1 arcmin
      if( is_void(crpix1) ) crpix1 = double(dms(2))/2;
      if( is_void(crpix2) ) crpix2 = double(dms(3))/2;

      coords = s_Coords();

      coords.ctype1 = "RA---TAN";
      coords.ctype2 = "DEC--TAN";
      coords.crpix1 = crpix1;
      coords.crpix2 = crpix2;
      coords.crval1 = ra;
      coords.crval2 = dec;
//    kwds_set,"CTYPE1","RA---TAN","Coordinate system 1. axis";
//    kwds_set,"CRPIX1",Num_pixels1/2.,"Reference pixel coordinate 1";
//    crval1 = is_void(Ra_scx) ? 0.0 : Ra_scx;
//    kwds_set,"CRVAL1",crval1,"Reference value coordinate 1";
//    kwds_set,"CUNIT1","deg","Unit of reference value coordinate 1";
//    kwds_set,"CTYPE2","DEC--TAN","Coordinate system 2. axis";
//    kwds_set,"CRPIX2",Num_pixels2/2.,"Reference pixel coordinate 2";
//    crval2 = is_void(Dec_scx) ? 0.0 : Dec_scx;
//    kwds_set,"CRVAL2",crval2,"Reference value coordinate 2";
//    kwds_set,"CUNIT2","deg","Unit of reference value coordinate 2";
//    pix1 = (Pixel_size1/Focal_length)*(180./pi);
//    rot = is_void(Posang) ? 0.0 : Posang * (pi/180.);
//    kwds_set,"CD1_1", -pix1*cos(rot), "[deg] -pix1*cos(rot)";
//    kwds_set,"CD1_2", -pix1*sin(rot), "[deg] -pix1*sin(rot)";
//    kwds_set,"CD2_1", -pix1*sin(rot), "[deg] -pix1*sin(rot)";
//    kwds_set,"CD2_2",  pix1*cos(rot), "[deg]  pix1*cos(rot)";
      rot_rad = rot * dr;
      coords.cd1_1 = -scale*cos(rot_rad);
      coords.cd1_2 = -scale*sin(rot_rad);
      coords.cd2_1 = -scale*sin(rot_rad);
      coords.cd2_2 =  scale*cos(rot_rad);
      coords.cdelt1 = -scale;
      coords.cdelt2 =  scale;
      coords.crota2 = rot;
      coords.scale = scale;
   }

   if( am_subroutine() ) {
      put_wcs, coords, old=old;
      writefits, outfile, im, clobber=clobber;
   } else return coords;

}

/* Function fits_copy_row */

func fits_copy_row( dol, rownum_from, rownum_to, except=, mode= )
/* DOCUMENT fits_copy_row, dol, rownum_from, rownum_to, except=, mode=

    Copies a row in a FITS binary table to another row in the
    same file. NB overwrites the file.

    The 'except' keyword can be given as a list of indices of
    column numbers to except from the copying.
    Examples:

       fits_copy_row, "test.fits+1", 23, 35, except="1-3"
       fits_copy_row, "test.fits+1", 23, 35, except=indgen(3)

   2014-01-06/NJW

*/
{
   if( is_void(mode) ) mode = 1;
   local hdr, nrows, filename, extno;
   get_exten_no, dol, filename, extno;
   ptr = rdfitsbin( dol, hdr, nrows );
   ncols = fits_numcols( hdr );
   if( rownum_from == rownum_to ) error,"from == to, no action";
   if( rownum_from < 1 || rownum_from > nrows ) error,"from is outside range";
   if( rownum_to < 1 || rownum_to > nrows ) error,"to is outside range";
   // ready to go
   if( is_void(except) ) {
      ignore = 0;
   } else {
      if( structof(except) == string ) {
         ignore = str2arr(except);
      } else ignore = except;
   }
   if( mode < 2 ) {
      write,"Version with rdfitsbin ...";
      for( icol = 1; icol <= ncols; icol++ ) {
         if( !anyof( ignore == icol ) ) (*ptr(icol))(rownum_to) = (*ptr(icol))(rownum_from);
      }
      wrfitsbin, filename, hdr, ptr, clobber=1;
   } else {
      write,"Version with fits_bintable_peek/poke ...";
      for( icol = 1; icol <= ncols; icol++ ) {
         if( !anyof( ignore == icol ) ) {
            value = fits_bintable_peek( dol, rownum_from, icol );
            fits_bintable_poke, dol, rownum_to, icol, value;
         }
      }
   }

/* Alternative - but slower - version
 *
 * for( icol = 1; icol <= ncols; icol++ ) {
 *    value = fits_bintable_peek( dol, rownum_from, icol );
 *    fits_bintable_poke, dol, rownum_to, icol, value;
 * }
 */
}

/* Function get_image_dol */

func get_image_dol( fitsname )
/* DOCUMENT dol = get_image_dol( fitsname )

     A FITS utility function.

     Returns a DOL for the first image found in the file.

     Even if 'fitsname' is a DOL this is not regarded.

    2014-11-19/NJW
*/
{
   local filename, extno;
   get_exten_no, fitsname, filename, extno;
   n = nfits_extens( filename );
   // First check primary array has any content
   hdr = headfits( filename+"+0" );
   naxis = fxpar( hdr, "naxis" );
   if( !is_void(naxis) ) {
      if( naxis >= 2 ) {
         naxis1 = fxpar( hdr, "naxis1" );
         naxis2 = fxpar( hdr, "naxis2" );
         if( naxis >= 2 && naxis1 > 0 && naxis2 > 0 ) return filename+"+0";
      }
   }

   // then look to the other extensions that must have XTENSION='IMAGE'

   for( i = 1; i <= n; i++ ) {
      hdr = headfits( filename+"+"+itoa(i) );
      xten = fxpar( hdr, "xtension" );
      if( strtrim(strupcase(xten)) == "IMAGE" ) return filename+"+"+itoa(i);
   }
   write,"No IMAGE extensions were found in "+filename;
   return [];
}

func _tdims( fh, col_number )
{
   // Check for TDIM keyword

   tdimstr = "TDIM"+itoa(col_number);
   tdim = fxpar( fh, tdimstr );
   if( is_void(tdim) ) {
      tdims = []; // signal for no dimensions given for data bin
   } else { // Now analyze the TDIM content
      tdim = strtrim(tdim);
      if( strpart(tdim,1:1) != "(" || strpart(tdim,0:0) != ")" ) error, "Bad TDIM keyword: "+tdim;
      tdim = strpart(tdim,2:-1);
      tdims = atoi(strsplit( tdim, "," ));
      tdims = grow( numberof(tdims), tdims );
   }
   return tdims;
}

/* Function addgrid */

func addgrid( im, coords, color=, thick= )
/* DOCUMENT addgrid, im, coords, color=, thick=

   Adds a sky coordinate grid to an image display, e.g. by 'disp'
   in the simple case where RA == 0 and at a suitable distance from the poles.
   No curvature is considered so representation may be in-exact.

   Arguments:
        im      the image, only for dimensionality
    coords      the WCS coordinates e.g. coords = get_wcs( DOL )

   Keywords:
     color, thick    Same meaning as for 'plot'.

   Necessary function: 'xnice' (see bottom of this file).

   2015-01-29/NJW
*/
{
   extern Inf_px, Inf_mx, Inf_py, Inf_my;

   local ra1, ra2, ra3, ra4, rac, dec1, dec2, dec3, dec4, decc;
   local p1, p2, q1, q2;

   dms = dimsof(im);

   // Setting infinity values for the 'geometry' package
   Inf_mx = -dms(2);
   Inf_px = 2*dms(2);
   Inf_my = -dms(3);
   Inf_py = 2*dms(3);

   // The sky coordinates for the corners are considered to be extreme
   // and the center is close to reference value
   skypos_fits, coords, 1, 1, ra1, dec1, to_sky=1, silent=1;
   skypos_fits, coords, 1, dms(3), ra2, dec2, to_sky=1, silent=1;
   skypos_fits, coords, dms(2), 1, ra3, dec3, to_sky=1, silent=1;
   skypos_fits, coords, dms(2), dms(3), ra4, dec4, to_sky=1, silent=1;
   skypos_fits, coords, dms(2)/2, dms(3)/2, rac, decc, to_sky=1, silent=1;

   ra_max = max( ra1, ra2, ra3, ra4 );
   ra_min = min( ra1, ra2, ra3, ra4 );
   dec_max = max( dec1, dec2, dec3, dec4 );
   dec_min = min( dec1, dec2, dec3, dec4 );

   delta_ra = 0.2*(ra_max - ra_min);
   delta_dec = 0.2*(dec_max - dec_min);

   rra = xnice( rac, delta_ra );
   rdec = xnice( decc, delta_dec );

   P = Q = s_Point();
   // left vertical axis: V1
   P.x = 1.; P.y = 1.; Q.x = 1.; Q.y = dms(3);
   V1 = line_p_p(P,Q);
   // right vertical axis: V2
   P.x = dms(2); P.y = 1.; Q.x = dms(2); Q.y = dms(3);
   V2 = line_p_p(P,Q);
   // lower horizontal axis: H1
   P.x = 1.; P.y = 1.; Q.x = dms(2); Q.y = 1.;
   H1 = line_p_p(P,Q);
   // upper horizontal axis: H2
   P.x = 1.; P.y = dms(3); Q.x = dms(2); Q.y = dms(3);
   H2 = line_p_p(P,Q);

   // plot the meridians

   for( i = -3; i <= 3; i++ ) {
      ra = rra(1) + i*rra(2);
      skypos_fits, coords, p1, q1, ra, dec_max, to_pix=1, silent=1;
      skypos_fits, coords, p2, q2, ra, dec_min, to_pix=1, silent=1;
      oplot,[p1,p2],[q1,q2],color=color,thick=thick;

      P.x = p1; P.y = q1; Q.x = p2; Q.y = q2;
      L = line_p_p(P,Q);
      PH1 = inter_l_l( L, H1 );
      PH2 = inter_l_l( L, H2 );
      PV1 = inter_l_l( L, V1 );
      PV2 = inter_l_l( L, V2 );
      if(!is_void(PH1))if( PH1.x > 1 && PH1.x < dms(2) ) xyouts,PH1.x,PH1.y,ftoa(ra),align=0.5,color=color,charsize=0.7;
      if(!is_void(PH2))if( PH2.x > 1 && PH2.x < dms(2) ) xyouts,PH2.x,PH2.y-dms(3)/25,ftoa(ra),align=0.5,color=color,charsize=0.7;
      if(!is_void(PV1))if( PV1.y > 1 && PV1.y < dms(3) ) xyouts,PV1.x,PV1.y,ftoa(ra),color=color,charsize=0.7;
      if(!is_void(PV2))if( PV2.y > 1 && PV2.y < dms(3) ) xyouts,PV2.x,PV2.y,ftoa(ra),align=1.,color=color,charsize=0.7;
   }
   for( j = -3; j <= 3; j++ ) {
      dec = rdec(1) + j*rdec(2);
      skypos_fits, coords, p1, q1, ra_max, dec, to_pix=1, silent=1;
      skypos_fits, coords, p2, q2, ra_min, dec, to_pix=1, silent=1;
      oplot,[p1,p2],[q1,q2],color=color,thick=thick; //+ ,thick=3;

      P.x = p1; P.y = q1; Q.x = p2; Q.y = q2;
      L = line_p_p(P,Q);
      PH1 = inter_l_l( L, H1 );
      PH2 = inter_l_l( L, H2 );
      PV1 = inter_l_l( L, V1 );
      PV2 = inter_l_l( L, V2 );
      if(!is_void(PH1))if( PH1.x > 1 && PH1.x < dms(2) ) xyouts,PH1.x,PH1.y,ftoa(dec),align=0.5,color=color,charsize=0.7;
      if(!is_void(PH2))if( PH2.x > 1 && PH2.x < dms(2) ) xyouts,PH2.x,PH2.y-dms(3)/25,ftoa(dec),align=0.5,color=color,charsize=0.7;
      if(!is_void(PV1))if( PV1.y > 1 && PV1.y < dms(3) ) xyouts,PV1.x,PV1.y,ftoa(dec),color=color,charsize=0.7;
      if(!is_void(PV2))if( PV2.y > 1 && PV2.y < dms(3) ) xyouts,PV2.x,PV2.y,ftoa(dec),align=1.,color=color,charsize=0.7;
   }
}

func xnice( x, delta_x )
/* DOCUMENT [rx, rdelta] = xnice( x, delta )

    Return suitable values for a grid start value and step
    size for a nice presentation.

*/
{
   ndeci = 0;
   ss = SiGn(1.,x);
   x = abs(x);
   // Find smallest number of decimals to represent 'delta_x'
   while( (rdelta_x = fndeci(delta_x,ndeci)) == 0 ) ndeci++;
   // Find equivalent value of 'x'
   rx = fndeci( x, ndeci );
   return [ss*rx, rdelta_x];
} 

/* Function fits_colname */

func fits_colname( fh_or_dol, colnum )
/* DOCUMENT column_name = fits_colname( fh_or_dol, colnum )

    Returns a string (or string array) with the column name(s)
    corresponding to the column number(s) in 'colnum'.

*/
{
   ncols = fits_numcols( fh_or_dol );
   if( structof( fh_or_dol ) == string ) { // we have a DOL
      hdr = headfits( fh_or_dol );
   } else hdr = fh_or_dol;
   colname = array( string, ncols );
   for( i = 1; i <= ncols; i++ ) colname(i) = fxpar( hdr, "TTYPE"+itoa(i) );

   if( is_void(colnum) ) { // return all names
      return colname;
   } else return colname(colnum);

}

/* Function fits_fh_verify */

func fits_fh_verify( &fh )
/* DOCUMENT fh = fits_fh_verify( >fh )

   Removes empty cards and updates the identifiers

   2009-06-25/NJW
*/
{
   cards = _car(fh,1);
   ncards = numberof(cards);
   ids = fits_ids(cards);
   w = where(ids >= 0.0 );
   if( numberof(w) == 0 ) error, "##5## fh is completely corrupted";
   cards = cards(w);
   ids = ids(w);
   _car, fh, 1, cards;
   _car, fh, 2, ids;
   return fh;
}
