extern scomdoc;
/* DOCUMENT ********************************************************
 *
 *    S file format package
 *
 *  2004-08-23/Niels J. Westergaard
 *  2009-07-21/NJW added function 'wstab'
 *
      comblock           rsmat              
      comget             rstab              
      comgets            scomblock          
      comput             scomget            
      computs            scomgets           
      comreplace         scomput            
      comreplaces        scomputs           
      get_spar           scomreplace        
      get_spar_dec       scomreplaces       
      get_spar_ra        sn_columns         
      get_spars          srstab             
      n_columns          wsmat              
      rd_qdp             wstab              
      rscol              

 ******************************************************************/

/* Function comget */

func comget( file, keyword_str, lng= )
/* DOCUMENT res = comget( file, keyword_str, lng=)
   Returns an array of values from lines:
      // keyword_str = value; comment field
   Format: Initial '//' (MUST be in position 1 and 2)
           keyword string CANNOT contain spaces
           Equal sign '=' MUST be present
   Everything following and including the ';' sign is disregarded

   Keyword   lng:   Converts to long integer

   2004-08-24/NJW
*/

{

   return scomget( rdfile(file), keyword_str, lng=lng );

}

/* Function comgets */

func comgets( file, keyword_str, unquote= )
/* DOCUMENT res = comgets( file, keyword_str, unquote=)
   Returns an array of string values from lines:
       // keyword_str = string

    Keyword  unquote  :  Removes leading and trailing '"' if present

   2004-09-12/NJW

*/

{

   return scomgets( rdfile(file), keyword_str, unquote=unquote );

}

/* Function scomget */

func scomget( string_arr, keyword_str, lng= )
/* DOCUMENT res = scomget(string_arr, keyword_str, lng=)
   Returns an array of values read from an S-format string array.
   If keyword 'lng ' is set then long integers are returned.

*/

{

   values = [];
   nval = 0;
   for ( i = 1; i <= numberof(string_arr); i++ ) {
      line = string_arr(i);
      if( strpart(line,1:2) == "//" ) {
         // remove comment part
         cpos = strpos(line,";",3);
         if( cpos > 3 ) line = strpart(line,1:cpos-1);
         pe = strpos(line,"=",3);
         comp = strtrim(strpart(line,3:pe-1));

         if( keyword_str == comp ) {

            if( lng ) {
               value = 1;
               n = sread(strpart(line,pe+1:),format="%d",value);
            } else {
               value = 1.0;
               n = sread(strpart(line,pe+1:),format="%f",value);
            }
            if( n == 1 ) grow, values, value;
         }
      }
   }

   n = numberof(values);
   return n == 1 ? values(1) : values;

}

/* Function scomgets */

func scomgets( string_arr, keyword_str, unquote= )
/* DOCUMENT res = scomgets(string_arr, keyword_str, unquote=)
   Returns an array of string values from an array in
   S-format.

   Keyword  unquote  :  Remove leading and trailing '"' if present
*/

{

   values = [];
   nval = 0;
   n = 0;
   for ( i = 1; i <= numberof(string_arr); i++ ) {
      line = string_arr(i);
      if( strpart(line,1:2) == "//" ) {
         // remove comment part
         cpos = strpos(line,";",3);
         if( cpos > 3 ) line = strpart(line,1:cpos-1);
         pe = strpos(line,"=",3);
         comp = strtrim(strpart(line,3:pe-1));

         if( keyword_str == comp ) {
            value = strtrim(strpart(line,pe+1:));
            if( unquote ) {
               if( strpart(value,1:1) == "\"" && strpart(value,0:0) == "\"" ) value = strpart(value,2:-1);
            }
            grow, values, value;
         }
      }
   }

   n = numberof(values);
   return n == 1 ? values(1) : values;

}

/* Function n_columns */

func n_columns( filename, sep= )
/* DOCUMENT n = n_columns(filename, sep=)

   2004-09-15/NJW

   Part of 'scom' package

    Return the number of columns in the first line that starts with
    a number i.e. not a comment.

    Keyword  sep  :  Separator string (single character)
                     Default is space (blank).
*/
{
   if( !file_test(filename) ) {
      write,"File not found: "+filename;
      return [];
   }

   lines = rdfile(filename);
   return sn_columns( lines, sep=sep );
}

/* Function sn_columns */

func sn_columns( lines, sep= )
/* DOCUMENT n = sn_columns(lines, sep=)

   2004-09-15/NJW

   Part of 'scom' package

    Return the number of columns in the first line that starts with
    a number i.e. not a comment.

    Keyword  sep  :  Separator string (single character)
                     Default is space (blank).
*/
{

   // Define separator

   if( is_void(sep) ) sep = " ";

   ncol = -1; // to be returned if no numbers found
   nlines = numberof(lines);
   for( i = 1; i <= nlines; i++ ) {

      line = lines(i);
      pos = strpos(line,";",1);
      if( pos > 0 ) line = strpart(line,1:pos-1);
                               // remove ';' and what follows (comments)
      line = strcompress(line,st=sep); // multiple separators and tabs to single occurrences
      line = strtrim(line); // remove leading and trailing blanks
      len = strlen(line);
      if( len > 0 ) {     // skip if no characters on line
         cc = strpart(line,1:1);
         if( cc == "." && len > 1 ) cc = strpart(line,2:2);
         n = strpos("+-0123456789",cc,1);
         if( n > 0 ) {  // here we have a number
            ncol = numberof(strsplit(line,sep));
            break;
         }
      }
   }

   return ncol;

}

/* Function comblock */

func comblock( filename, nblock )
/* DOCUMENT  [lnum1,lnum2] = comblock( filename, nblock )

    Member of scom package

    Returns 2 element line_number begin, line_number end.

    Reads an ASCII file 'file' searching for lines beginning with '//'
    followed by 'block = nnnnn'  e.g. :

      // block = 123

   SEE ALSO: comget, comgets, rscol, rstab, rsmat.

    2004-09-12/NJW
    2013-04-11/NJW updated to use scomblock
*/
{
   if( !file_test(filename) ) {
      write,"File not found: "+filename;
      return [];
   }
   lines = rdfile(filename);
   return scomblock( lines, nblock);

}

/* Function scomblock */

func scomblock( lines, nblock )
/* DOCUMENT  [lnum1,lnum2] = scomblock( lines, nblock )

    Member of scom package

    Returns 2 element line_number begin, line_number end.

    Reads an ASCII file 'file' searching for lines beginning with '//'
    followed by 'block = nnnnn'  e.g. :

      // block = 123

   SEE ALSO: comget, comgets, rscol, rstab, rsmat.

    2004-09-12/NJW
*/
{

   nlines = numberof(lines);
   lns = array(0,1000);
   bns = array(0,1000);
   bn = 0;
   linenumber = 1;
   k = 1;
   for( i = 1; i <= nlines; i++ ) {
      if( strpart(lines(i),1:2) == "//" ) {
         pos = strpos(lines(i),"block",3);
         if( pos >= 1 ) {    // string has been found
            pe = strpos(lines(i),"=",pos+5);
            if( pe >= 1 ) {   // an equal sign has been found
               sb = strpart(lines(i),3:pe-1);
               psemi = strpos(lines(i),";",3);
               if( psemi == 0 ) psemi = strlen(lines(i))+1;
               // Make sure that the only the search string is present
               if( strtrim(sb) == "block") {
                  sread,strpart(lines(i),pe+1:psemi-1),bn;
                  bns(k) = bn;
                  lns(k) = linenumber;
                  k++;
               }
            }
         }
      }
      linenumber++;
   }
   close,file;

   if( k > 1 ) {
      lns = lns(1:k);
      bns = bns(1:k);
      lns(k) = linenumber - 1;
      bns(k) = 99999;
      k_ok = where( nblock == bns );
      nk_ok = numberof(k_ok);
      if( nk_ok > 0 ) {
         ln_begin = lns(k_ok(1))+1;
         ln_end   = lns(k_ok(1)+1) - (bns(k_ok(1)+1) != 99999);
          // bns == 99999 signals end of file and ln_end is not to be reduced by one
      } else {
         ln_begin = -1;
         ln_end   = -1;
      }
   } else {
      ln_begin = -1;
      ln_end   = -1;
   }
   return([ln_begin,ln_end]);
}

/* Function rscol */

func rscol(filename, col_def, str=,nt=,lng=,dble=,silent=,block=,typ=,sep=,nomem=)
/* DOCUMENT col = rscol( filename, col_def, str=, nt=,lng=,
                      dble=, silent=, block=, typ=, sep=, nomem=)

     Read a column from an ASCII file where numbers are organized as a table
     If the column should be read as a string then the keyword 'str' must be set.
     Default type is float. Keywords exist for int, long and double.
     Alternatively 'typ' can be given "i":int, "j":long, "d":double, "f":float, "s":string.
     The columns can be numbers or character strings separated with
     blanks or the character as given in keyword 'sep'. The first column (number 1) must
     always start with one of +-.0123456789.

     The lines that start with other characters are interpreted as
     comment lines and will be displayed but skipped in
     the reading of the data.

     If the 'block' keyword is given with a block number
     then the file will be scanned for existence of lines:
         // block = nnn
     and only the table rows between where nnn matches the
     given block number and the next '// block = xxx'

     Keyword: nomem    Disselects the use of memory i.e. when files are expected to change
                       between readings

     Other procedures required: n_columns.pro
                                comblock.pro

     2004-09-12/NJW
     2009-10-22/NJW, updated to use mem_storage if not disselected

*/

{

   vb = silent ? 0 : 1;

   // Define type of returned array

   type = 4 ; // float
   if (str) type = 7 ; // string
   if (lng) type = 3 ; // long
   if (dble) type = 5 ; // double
   if (nt)  type = 2 ; // integer
   if( typeof(typ) == "string" ) {
      if( strlen(typ) == 1 ) {
         if( typ == "i" ) type = 2;
         if( typ == "j" ) type = 3;
         if( typ == "f" ) type = 4;
         if( typ == "d" ) type = 5;
         if( typ == "s" ) type = 7;
      }
   }

   //  --- memory usage: define storage name
   //      as filename+column-number+block-number

   // Column separator

   if( is_void(sep) ) sep = " "; // Default separator is space (blank)

   // If column designator is a string then find the matching
   // number

   vartype = typeof(col_def);
   if( vartype == "string" ) {
      // get the column names
      colnames = comgets(filename,"colname");
      if( numberof(colnames) == 0 ) {
         write,"Error - no keyword colname found";
         return [];
      }
      w = where( colnames == col_def );
      if( numberof(w) ) {
         col_number = w(1);
      } else {
         write,"Error: no such column name has been found";
         return([]);
      }
   } else {
      col_number = col_def;
   }
   col_num = col_number;

   // Initiate use of block (i.e. only take data from a
   // specified block

   ln_begin = 1;
   ln_end = 0; // means no upper limit
   useblock = 0;
   if (block) {
      useblock = 1;
      // nblock = long(block);
      line_nums = comblock(filename, block);
      if(line_nums(1) == -1 ) {
         write,"RSCOL Warning: Requested block not found";
         write,"                 Proceed reading all data";
         useblock = 0;
      } else {
         ln_begin = line_nums(1);
         ln_end   = line_nums(2);
      }
   }

   // --- set memory name

   mem_name = filename+itoa(col_num);
   if( useblock ) mem_name += "_"+itoa(block);

   if( 0 == mem_restore( mem_name, col_dat ) && !nomem ) return col_dat;

   // See if requested column number is less than or equal to
   // the number of columns

   if( vb ) write,format="RSCOL: at least %d columns expected in %s\n",col_number, filename;
   nc = n_columns(filename,sep=sep);
   if( vb ) write,format="%d columns found in file: %s\n", nc, filename;

   if( col_number > nc ) {
      write,format="RSCOL: only %d columns found in file\n",nc;
      return([]);
   }

   // Prepare reading the file in chunks of 'nmax' lines

   nmax = 300;
   cur_size = nmax;

   file = open(filename,"r");

   if( type == 2 ) {
      col_dat = array(int,nmax);
   } else if( type == 3 ) {
      col_dat = array(long,nmax);
   } else if( type == 4 ) {
      col_dat = array(float,nmax);
   } else if( type == 5 ) {
      col_dat = array(double,nmax);
   } else if( type == 7 ) {
      col_dat = array("",nmax);
   }

   x = array(float,nc);
   s = " ";
   kount = 1;
   linenumber = 0;

   while( s = rdline(file) ) {

     linenumber++;
     pos = strpos(s,";",1);
     if( pos > 0 ) {
        ss = strpart(s,1:pos-1);
     } else ss = s;
     st = strtrim(ss,blank=" "+sep); // remove leading and trailing blanks
     if( strlen(st) > 0 ) {
        if( !strmatch("+-.0123456789",strpart(st,1:1))  ) {
              if( vb ) write,format="%s\n",s;
        } else {
           su = strcompress(st,st=sep); // reduce all separators to single occurrences
           elmnts = strsplit(su,sep); // separate into columns

           if( linenumber >= ln_begin &&
                  (linenumber <= ln_end || ln_end == 0) ) {
              if( type == 7 ) {
                   col_dat(kount) = elmnts(col_number);
              } else {
                   n = sread(elmnts(col_number),col_dat(kount));
              }
              kount++;
           }


           // Check if expansion is needed

           if( kount == cur_size + 1) {

              grow,col_dat,array(col_dat(1),nmax);
              cur_size += nmax;

           }
        }
      }
   }
   if( vb ) write,format="RSCOL Size of table = %d\n",kount-1;
   col_dat = kount == 1 ? [] : col_dat(1:kount-1);
   close,file;

   if( !nomem ) mem_save, mem_name, col_dat;

   return( col_dat );

}

/* Function rstab */

func rstab( filename, ncols, &c1, &c2, &c3, &c4, &c5, &c6, &c7, &c8, &c9, &c10, &c11, \
            &c12, &c13, &c14, &c15, block=, typ=, sep=, silent= )
/* DOCUMENT rstab, filename, ncols, >c1, >c2, ..., block=, typ=, sep=, silent=

   Reads S-format text file and returns columns in c1, c2, ... up to and including c15
   The second argument must give the requested number of columns.

   Keyword 'typ' can be given as a string with as many characters as there are columns.
   i: int, j: long, f: float, d: double, s: string
   (defaults to 'double'). It must be given if some column values are incompatible with
   numbers. A dash causes skipping the column entirely.
   Keyword 'sep' indicates the column separator if different from space.

   e.g. file "myfile.scm":
     // here are some data
        1 1.2345e2 background 11.23
        2 5.5142e3 foreground 12.85

   can be read with: > rstab,"myfile.scm",4,n,y,denom,calib,typ="idsf"

   or, if only column 1 and 4 are to be read:

   > rstab,"myfile.scm",2,n,calib,typ="i--f"

   2008-08-13/NJW
   2009-07-20/NJW updated for more efficient reading
   2015-08-28/NJW included c14 and c15
*/
{
   if( !file_test(filename) ) {
      write,"File not found: "+filename;
      return;
   }

   lines = rdfile(filename);
   srstab, lines, ncols, c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, c11, \
            c12, c13, c14, c15, block=block, typ=typ, sep=sep, silent=silent;

}

/* Function srstab */

func srstab( lines, ncols, &c1, &c2, &c3, &c4, &c5, &c6, &c7, &c8, &c9, &c10, &c11, \
            &c12, &c13, &c14, &c15, block=, typ=, sep=, silent= )
/* DOCUMENT srstab, lines, ncols, >c1, >c2, ..., block=, typ=, sep=, silent=

   Interprets S-format string array and returns columns in c1, c2, ... up to and including c13
   The second argument must give the requested number of columns.

   Keyword 'typ' can be given as a string with as many characters as there are columns.
   i: int, j: long, f: float, d: double, s: string
   (defaults to 'double'). It must be given if some column values are incompatible with
   numbers. A dash causes skipping the column entirely.
   Keyword 'sep' indicates the column separator if different from space.

   e.g. content of 'lines':
     // here are some data
        1 1.2345e2 background 11.23
        2 5.5142e3 foreground 12.85

   can be read with: > srstab,"myfile.scm",4,n,y,denom,calib,typ="idsf"

   or, if only column 1 and 4 are to be read:

   > rstab,"myfile.scm",2,n,calib,typ="i--f"

   2008-08-13/NJW
   2009-07-20/NJW updated for more efficient reading
   2015-08-28/NJW included c14 and c15
*/
{
   vb = !silent;
   if(is_void(sep) ) sep = " ";
   nlines = numberof(lines);

   // 'ncols' is the number of requested columns to return

   if(vb) write,format="SRSTAB: %i first columns were requested\n", ncols;

   if( typeof(typ) == "string" ) {
      types = (*pointer(typ))(1:-1); // drop trailing null
      wtypes = where(types != '-');
      typw = types(wtypes);
      len = numberof(wtypes);
      if( len != ncols ) {
         write,"Mismatching typ and number of columns - default to double";
         typw = array('d',ncols);
         wtypes = indgen(ncols);
      }
   } else {
      typw = array('d',ncols);
      wtypes = indgen(ncols);
   }

   // Initiate use of block (i.e. only take data from a
   // specified block

   ln_begin = 1;
   ln_end = 0; // means no upper limit
   useblock = 0;
   if (block) {
      useblock = 1;
      // nblock = long(block);
      line_nums = scomblock(lines, block);
      if(line_nums(1) == -1 ) {
         write,"SRSTAB Warning: Requested block not found";
         write,"                 Proceed reading all data";
         useblock = 0;
      } else {
         ln_begin = line_nums(1);
         ln_end   = line_nums(2);
      }
   }

   // See if requested number of columns is less than or equal to
   // the existing number of columns

   if( vb ) write,format="SRSTAB: at least %i columns expected\n", ncols;
   // 'nc' is number of existing columns in the file
   nc = sn_columns(lines,sep=sep);
   if( vb ) write,format="%d columns found\n", nc;

   if( ncols > nc ) {
      write,format="SRSTAB: only %d columns found\n",nc;
      return([]);
   }

   // Prepare reading the file

   nmax = nlines;
   cur_size = nmax;

   c1 = array(struc(typw(1)),nmax);
   if( ncols > 1 ) c2 = array(struc(typw(2)),nmax);
   if( ncols > 2 ) c3 = array(struc(typw(3)),nmax);
   if( ncols > 3 ) c4 = array(struc(typw(4)),nmax);
   if( ncols > 4 ) c5 = array(struc(typw(5)),nmax);
   if( ncols > 5 ) c6 = array(struc(typw(6)),nmax);
   if( ncols > 6 ) c7 = array(struc(typw(7)),nmax);
   if( ncols > 7 ) c8 = array(struc(typw(8)),nmax);
   if( ncols > 8 ) c9 = array(struc(typw(9)),nmax);
   if( ncols > 9 ) c10 = array(struc(typw(10)),nmax);
   if( ncols > 10 ) c11 = array(struc(typw(11)),nmax);
   if( ncols > 11 ) c12 = array(struc(typw(12)),nmax);
   if( ncols > 12 ) c13 = array(struc(typw(13)),nmax);
   if( ncols > 13 ) c14 = array(struc(typw(14)),nmax);
   if( ncols > 14 ) c15 = array(struc(typw(15)),nmax);

   kount = 1;
   linenumber = 0;

   for( i = 1; i <= nlines; i++ ) {

     s = lines(i);
     linenumber++;
     pos = strpos(s,";",1);
     if( pos > 0 ) {
        ss = strpart(s,1:pos-1);
     } else ss = s;
     cst = strtrim(ss,blank=sep+" "); // remove leading and trailing separators
     if( strlen(cst) > 0 ) {
        if( !strmatch("+-.0123456789",strpart(cst,1:1))  ) {
              if( vb ) write,format="%s\n",s;
        } else {
           su = strcompress(cst,st=sep); // reduce all separators to single separators
           elmnts = strsplit(su,sep); // separate into columns

           if( linenumber >= ln_begin &&
                  (linenumber <= ln_end || ln_end == 0) ) {
              c1(kount) = _str2x( elmnts(wtypes(1)), typw(1) );
              if( ncols > 1 ) c2(kount) = _str2x( elmnts(wtypes(2)), typw(2) );
              if( ncols > 2 ) c3(kount) = _str2x( elmnts(wtypes(3)), typw(3) );
              if( ncols > 3 ) c4(kount) = _str2x( elmnts(wtypes(4)), typw(4) );
              if( ncols > 4 ) c5(kount) = _str2x( elmnts(wtypes(5)), typw(5) );
              if( ncols > 5 ) c6(kount) = _str2x( elmnts(wtypes(6)), typw(6) );
              if( ncols > 6 ) c7(kount) = _str2x( elmnts(wtypes(7)), typw(7) );
              if( ncols > 7 ) c8(kount) = _str2x( elmnts(wtypes(8)), typw(8) );
              if( ncols > 8 ) c9(kount) = _str2x( elmnts(wtypes(9)), typw(9) );
              if( ncols > 9 ) c10(kount) = _str2x( elmnts(wtypes(10)), typw(10) );
              if( ncols > 10 ) c11(kount) = _str2x( elmnts(wtypes(11)), typw(11) );
              if( ncols > 11 ) c12(kount) = _str2x( elmnts(wtypes(12)), typw(12) );
              if( ncols > 12 ) c13(kount) = _str2x( elmnts(wtypes(13)), typw(13) );
              if( ncols > 13 ) c14(kount) = _str2x( elmnts(wtypes(14)), typw(14) );
              if( ncols > 14 ) c15(kount) = _str2x( elmnts(wtypes(15)), typw(15) );
              kount++;
           }
        }
      }
   }
   if( vb ) write,format="SRSTAB Size of table = %d\n",kount-1;
   c1 = kount == 1 ? [] : c1(1:kount-1);
   if( ncols > 1 ) c2 = kount == 1 ? [] : c2(1:kount-1);
   if( ncols > 2 ) c3 = kount == 1 ? [] : c3(1:kount-1);
   if( ncols > 3 ) c4 = kount == 1 ? [] : c4(1:kount-1);
   if( ncols > 4 ) c5 = kount == 1 ? [] : c5(1:kount-1);
   if( ncols > 5 ) c6 = kount == 1 ? [] : c6(1:kount-1);
   if( ncols > 6 ) c7 = kount == 1 ? [] : c7(1:kount-1);
   if( ncols > 7 ) c8 = kount == 1 ? [] : c8(1:kount-1);
   if( ncols > 8 ) c9 = kount == 1 ? [] : c9(1:kount-1);
   if( ncols > 9 ) c10 = kount == 1 ? [] : c10(1:kount-1);
   if( ncols > 10 ) c11 = kount == 1 ? [] : c11(1:kount-1);
   if( ncols > 11 ) c12 = kount == 1 ? [] : c12(1:kount-1);
   if( ncols > 12 ) c13 = kount == 1 ? [] : c13(1:kount-1);
   if( ncols > 13 ) c14 = kount == 1 ? [] : c14(1:kount-1);
   if( ncols > 14 ) c15 = kount == 1 ? [] : c15(1:kount-1);

   if(vb)write,format="SSRSTAB: %i columns with %i rows read.\n", ncols, numberof(c1);

}

/* Function scomreplaces */

func scomreplaces( string_arr, keyword_str, new_string )
/* DOCUMENT res = scomreplaces(string_arr, keyword_str, new_string)

   Replaces the string value(s) given with keyword 'keyword_str'
   with 'new_string'. 'new_string' can be a single string or an
   array of same number of elements as there are keywords of the
   chosen name in the input string array.

    2009-03-27/NJW

*/

{

   nnew = numberof(new_string);
   res = []; // string array to be returned

   values = [];
   nval = 0;
   n = 0;
   for ( i = 1; i <= numberof(string_arr); i++ ) {
      line = string_arr(i);
      if( strpart(line,1:2) == "//" ) {
         // save comment part and strip
         cpos = strpos(line,";",3);
         csave = [];
         wline = line;
         if( cpos > 3 ) {
            csave = strpart(line,cpos:);
            wline = strpart(wline,1:cpos-1);
         }
         pe = strpos(wline,"=",3);
         if( pe <= 3 ) { // not a keyword anyway
            grow, res, line;
            continue;
         }
         // trim to find keyword candidate
         comp = strtrim(strpart(wline,3:pe-1));

         if( keyword_str == comp ) {
            // found the keyword, do replacement
            nval++;
            if( nnew != 1 && nval > nnew ) {
               write,"scomreplaces error, too few replacement values";
               return string_arr;
            }
            j = nnew == 1 ? 1 : nval;
            line = strpart(wline,1:pe)+" "+new_string(j);
            if( !is_void(csave) ) line += csave;
            grow, res, line;
         } else { // wrong keyword
            grow, res, line;
         }
      } else {
         grow, res, line;
      }
   }

   return res;

}

/* Function comreplaces */

func comreplaces( file, keyword_str, new_string )
/* DOCUMENT complaces, file, keyword_str, new_string


*/
{
   str = rdfile( file );
   str = scomreplaces( str, keyword_str, new_string );
   write_slist, file, str;
}

/* Function scomreplace */

func scomreplace( string_arr, keyword_str, new_values )
/* DOCUMENT res = scomplace( string_arr, keyword_str, new_values )

*/
{
   tty = typeof(new_values);
   if( tty == "double" ) {
      new_string = swrite(format="%.14e", new_values);
   } else if( tty == "float" ) {
      new_string = swrite(format="%.8e", new_values);
   } else if( tty == "long" || tty == "int" ) {
      new_string = swrite(format="%i", new_values);
   } else {
      write,"data format not supported";
      return string_arr;
   }
   return scomreplaces( string_arr, keyword_str, new_string );
}

/* Function comreplace */

func comreplace( file, keyword_str, new_values )
/* DOCUMENT complace, file, keyword_str, new_values


*/
{
   str = rdfile( file );
   str = scomreplace( str, keyword_str, new_values );
   write_slist, file, str;
}

/* Function get_spar */

func get_spar( filename, parname, lng= )
/* DOCUMENT parval = get_spar( filename, parname, lng= )

 Return the new parameter : double [default]
                           long     -  keyword "lng"

*/
{
   par_new = comget(filename, parname,lng=lng);
   if( is_void(par_new) ) {
      write,"No such parameter found: "+parname;
      return [];
   }

   if( lng ) write,format="Default %s [%i]\n",parname,par_new;
   else write,format="Default %s [%.4g]\n",parname,par_new;

   answer = rdline(prompt="Accept with Rtrn or enter new "+parname+" (string) : ... ");
   if( strlen(answer) ) {
      if( lng ) {
         par_new = tolong(answer);
      } else {
         par_new = atof(answer)
      }
      comreplace, filename, parname, par_new;
   }
   if(lng)write,format="The new %s is %i\n", parname, par_new;
   else write,format="The new %s is %.4g\n", parname, par_new;

   return par_new;
}

/* Function get_spar_dec */

func get_spar_dec( filename, parname )
/* DOCUMENT par = get_spar_dec( filename, parname )

    Return an s-format real Dec parameter by dialogue
        where DMS format is possible
*/
{

   par_new = comget(filename, parname);
   if( is_void(par_new) ) {
      write,"No such parameter found: "+parname;
      return [];
   }
   
   write,"Default "+parname+" [", par_new," or ", adstring(0,par_new),"]";
   answer = rdline( prompt="Accept with Rtrn or enter new "+parname+" (string) : ... ");
   
   if( strlen(answer) ) {
      answer = strcompress(strtrim(answer));
      tok = strsplit(answer," ")
      if( numberof(tok)  ==  1 ) par_new = atof(answer);
      if( numberof(tok)  ==  2 ) par_new = ten(atof(tok(1)),atof(tok(2)));
      if( numberof(tok)  >  2 ) par_new = ten(atof(tok(1)),atof(tok(2)),atof(tok(3)));
      comreplace,filename, parname, par_new;
   }
   write,"The new "+parname+" is: ", par_new;
   return par_new;
}

/* Function get_spar_ra */

func get_spar_ra( filename, parname )
/* DOCUMENT par = get_spar_ra( filename, parname )

    Return an s-format real RA parameter by dialogue
      where HMS format is possible
*/
{

// Get the new parameter; Special for RA

   par_new = comget(filename, parname);
   if( is_void(par_new) ) {
      write,"No such parameter found: "+parname;
      return [];
   }
   
   write,"Default "+parname+" [", par_new," or ", adstring(par_new,0),"]";
   answer = rdline( prompt="Accept with Rtrn or enter new "+parname+" (string) : ... ");
   
   if( strlen(answer) ) {
      answer = strcompress(strtrim(answer));
      tok = strsplit(answer," ");
      if( numberof(tok)  ==  1 ) par_new = atof(answer);
      if( numberof(tok)  ==  2 ) par_new = 15.*ten(atof(tok(1)),atof(tok(2)));
      if( numberof(tok)  >  2 ) par_new = 15.*ten(atof(tok(1)),atof(tok(2)),atof(tok(3)));
      ierr = comreplace(filename, parname, par_new);
   }
   write,"The new "+parname+" is: ", par_new;
   return par_new;
}

/* Function get_spars */

func get_spars( filename, parname )
/* DOCUMENT par = get_spars( filename, parname )

    Return an s-format string parameter by dialogue
*/
{

   // Get the new string valued parameter
   
   par_new = comgets(filename, parname);
   if( is_void(par_new) ) {
      write,"No such parameter found: "+parname;
      return [];
   }

   write,"Default "+parname+" [", par_new,"]";
   answer = rdline( prompt="Accept with Rtrn or enter new "+parname+" (string) : ... ");
   if( strlen(answer) ) {
      par_new = answer;
      comreplaces,filename, parname, par_new;
   }
   write,"The new "+parname+" is: ", par_new;
   return par_new;
}

/* Function wstab */

func wstab( filename, .., fmt=, hdr=, app= )
/* DOCUMENT wstab, filename, arr1, arr2, ..., fmt=, hdr=, app=

   Write a number, basically unlimited, of columns to a file as specified by a format string

   KEYWORD: hdr     a string array that will be written
                    in the start of the file.
            app     Append to an existing file
            fmt     Format string (like "%5.0f%13s%6.2f")
                       Default format is  (string)     : %s
                                     (char, int, long) : %i
                                          (float)      : %13.6e
                                          (double)     : %17.10e

   2005-01-03/NJW, update 2005-04-14/NJW
   2007-09-13/NJW, update with append option
   2008-08-13/NJW, update with format option

*/

{
   local coldata;

   if( app ) {
      f = open( filename,"a");
   } else {
      f = open( filename,"w");
   }
   if( (num = numberof(hdr)) ) {
      for( i = 1; i <= num; i++ ) {
         write,f,format="%s\n",hdr(i);
      }
   }

   first = 1;
   ncols = 0;
   while( more_args() ) {
      ncols++;
      buffer = next_arg();
      //+ write,format="%i elements in col#%i\n", numberof(buffer), ncols;
      if( first ) {
         list = _lst(buffer);
         nrows = numberof(buffer);
         //+ write,format="nrows = %i (ncols = %i)\n", nrows, ncols;
         first = 0;
      } else {
         list = _cat(list,_lst(buffer));
         nr = numberof(buffer);
         //+ write,format="nr = %i (ncols = %i)\n", nr, ncols;
         if( nr != nrows ) {
            write,format="Column #%i has wrong number of elements\n", ncols;
            return;
         }
      }
   }

   //  assign standard formats if not given

   fmtflag = typeof( fmt ) == "string" ? 1 : 0;
   if( fmtflag ) {
      fmtpart = strsplit( fmt, "%" );
      nfmtpart = numberof(fmtpart);
      for(i=1;i<=nfmtpart;i++) {
         fmtpart(i) = "%"+fmtpart(i);
      }
      if( nfmtpart != ncols ) {
         write,"Bad number of format specifications - use defaults";
         fmtflag = 0;
      }
   }
   if( !fmtflag) {
      fmtpart = array(string,ncols);
      for(i=1;i<=ncols;i++) {
         typ = typeof(_car(list,i));
         if( typ == "string" ) {
            fmtpart(i) = " %s ";
         } else if( typ == "int" || typ == "long" || typ == "char" ) {
            fmtpart(i) = " %i ";
         } else if( typ == "float" ) {
            fmtpart(i) = " %13.6e ";
         } else if( typ == "double" ) {
            fmtpart(i) = " %17.10e ";
         } else {
            write,"Data type not supported";
            return;
         }
      }
   }

   //  do the writing to file

   for( i = 1; i <= nrows; i++ ) {
      for( j = 1; j <= ncols; j++ ) {
         eq_nocopy, coldata, _car(list,j);
         write,f,format=fmtpart(j), coldata(i);
      }
      write,f,format="%s\n","";
   }

   close,f;

}

/* Function rsmat */

func rsmat( filename, nt=, lng=, dble=, flt=, block=, skip=, silent= )
/* DOCUMENT arr = rsmat( filename, nt=, lng=, dble=, flt=, block=, skip=, silent= )

   Returns s-format array of type specified by keyword nt, lng, dble, or flt
   (default is double).

   Keyword 'skip' is the number of columns to skip counting from the first
           'silent' will suppress additional screen output

   2009-07-23/NJW
*/
{
   vb = !silent;
   if(is_void(skip)) skip = 0;

   typ = 'd'; // default
   if( nt ) typ = 'i';
   if( lng ) typ = 'j';
   if( flt ) typ = 'f';

   if( !file_test(filename) ) {
      write,"RSMAT: File not found: "+filename;
      return [];
   }
   file = open(filename,"r");

   nc = n_columns(filename);
   if( vb ) write,format="%d columns found in file: %s\n", nc, filename;

   // Initiate use of block (i.e. only take data from a
   // specified block

   ln_begin = 1;
   ln_end = 0; // means no upper limit
   useblock = 0;
   if (block) {
      useblock = 1;
      // nblock = long(block);
      line_nums = comblock(filename, block);
      if(line_nums(1) == -1 ) {
         write,"RSMAT Warning: Requested block not found";
         write,"                 Proceed reading all data";
         useblock = 0;
      } else {
         ln_begin = line_nums(1);
         ln_end   = line_nums(2);
      }
   }

   linenumber = 0;
   kount = 1;
   nmax = cur_size = 300;
   arr = array( struc(typ), nc-skip, nmax );

   while( s = rdline(file) ) {

     linenumber++;
     pos = strpos(s,";",1);
     if( pos > 0 ) {
        ss = strpart(s,1:pos-1);
     } else ss = s;
     st = strtrim(ss); // remove leading and trailing blanks
     if( strlen(st) > 0 ) {
        if( !strmatch("+-.0123456789",strpart(st,1:1))  ) {
              if( vb ) write,format="%s\n",s;
        } else {
           su = strcompress(st); // reduce all space to single spaces
           elmnts = strsplit(su," "); // separate into columns

           if( linenumber >= ln_begin &&
                  (linenumber <= ln_end || ln_end == 0) ) {
              for(i = 1+skip; i <= nc; i++ ) {
                 arr(i-skip,kount) = _str2x( elmnts(i),typ);
              }
              kount++;
           }


           // Check if expansion is needed

           if( kount == cur_size + 1) {

              grow,arr,array(struc(typ),nc-skip,nmax);
              cur_size += nmax;

           }
        }
      }
   }

   close, file;

   return transpose(arr(,1:kount-1));

}

/* Function wsmat */

func wsmat( filename, arr, fmt=, hdr=, app= )
/* DOCUMENT wsmat, filename, arr, fmt=, hdr=, app=

   Write an array in ASCII format to a file as specified by
   a format string

   KEYWORD: hdr     a string array that will be written
                    in the start of the file.
            app     Append to an existing file
            fmt     Format string (like "%5.1f")

   2009-07-24/NJW, cloned from wstab

*/

{

   if( app ) {
      f = open( filename,"a");
   } else {
      f = open( filename,"w");
   }
   if( (num = numberof(hdr)) ) {
      for( i = 1; i <= num; i++ ) {
         write,f,format="%s\n",hdr(i);
      }
   }

   typ = typeof(arr);
   dms = dimsof(arr);
   if( dms(1) != 2 ) {
      write,"WSMAT: array must be 2D, no action";
      return;
   }
   nrows = dms(2);
   ncols = dms(3);

   //  assign standard formats if not given

   fmtflag = typeof( fmt ) == "string" ? 1 : 0;
   if( !fmtflag) {
      if( typ == "string" ) {
         fmt = " %s ";
      } else if( typ == "int" || typ == "long" || typ == "char" ) {
         fmt = " %i ";
      } else if( typ == "float" ) {
         fmt = " %13.6e ";
      } else if( typ == "double" ) {
         fmt = " %17.10e ";
      } else {
         write,"Data type not supported";
         return;
      }
   }

   //  do the writing to file

   for( i = 1; i <= nrows; i++ ) {
      for( j = 1; j <= ncols; j++ ) {
         write,f,format=fmt, arr(i,j);
      }
      write,f,format="%s\n","";
   }

   close,f;

}

/* Function rd_qdp */

func rd_qdp( namestem_or_name, &x, &dx, &y, &dy, &m, part=, silent= )
/* DOCUMENT rd_qdp, namestem_or_name, >x, >dx, >y, >dy, >m, part=, silent=
    or    num_parts = rd_qdp( namestem_or_name, >x, >dx, >y, >dy, >m, part=, silent= )

    Reads a 3, 4, or 5 column QDP ASCII file that may consist of several parts
    separated by lines: "NO ..."

    Second version returns the number of parts in the qdp file.

    Filename must have postfix '.qdp'

*/
{
   filename = namestem_or_name;
   if( strpart( namestem_or_name, -3:0) != ".qdp" ) \
                  filename = namestem_or_name+".qdp";
   if( !file_test(filename) ) {
      write,"No such file: "+filename;
      return;
   }
   local c1,c2,c3,c4,c5;

   if( is_void(part) ) part = 1;

   // read to memory

   text = strtrim(rdfile(filename));
   nno = where(strpart(text,1:2) == "NO")
   if(numberof(nno)==0) nno = [];
   grow,nno,numberof(text)+1;
   n_nno = numberof(nno);
   if( !silent ) {
      if( n_nno == 1 ) {
         write,"Holds a single set of data";
      } else {
         write,"Holds "+itoa(n_nno)+" sets of data";
      }
   }
   if( part > n_nno ) {
      write,"'part' exceeds the present number of parts";
      return;
   }
   lin1 = part > 1 ? nno(part-1)+1 : 1;
   lines= text(lin1:nno(part)-1);
   ncols = sn_columns(lines);
   if(!silent) write,"Found "+itoa(numberof(lines))+" rows and "+itoa(ncols)+" columns";
   if( ncols < 3 ) {
      write,"Too few columns";
      return;
   }
   if( ncols > 5 ) {
      write,"Too many columns";
      return;
   }
   if( ncols == 3 ) {
      srstab,lines,3,c1,c2,c3,typ="ddd",silent=1;
      x = c1; dx = c2; y = c3;
   } else if( ncols == 4 ) {
      srstab,lines,4,c1,c2,c3,c4,typ="dddd",silent=1;
      x = c1; dx = c2; y = c3; dy = c4;
   } else { // ncols must be 5 here
      // see, if last column consists of "NO"
      test_no = strpart( lines, -1:0);
      //  -----------
      // Well, here is a simple test. It might be wiser to test all
      // endings and report an error if an inconsistency is found.
      //  -----------
      if( test_no(1) == "NO" ) {
         srstab,lines,4,c1,c2,c3,c4,typ="dddd",silent=1;
         c5 = array(1., numberof(lines));
      } else {
         srstab,lines,5,c1,c2,c3,c4,c5,typ="ddddd",silent=1;
      }
      x = c1; dx = c2; y = c3; dy = c4; m = c5;
   }

   if( !am_subroutine() ) return n_nno;
}

/* Function comput */

func comput( file, keyword, value, atline=, comment= )
/* DOCUMENT comput, file, keyword, value, atline=, comment=

    Insert one or more numerical values under keyword 'keyword'
    immediately after first bunch of '//' comment lines.

    Keyword 'atline' forces insertion immediately before this
    line number.

    2013-06-04/NJW
*/
{
   lines = rdfile(file);
   nlines = numberof(lines);
   nval = numberof(value);
   tty = typeof(value);
   if( tty != "long" && tty != "int" && tty != "double" && tty != "float" ) {
      write,"Datatype: "+tty+" is not supported!";
      return;
   }

   // Compose the extra lines

   com = "";
   if( typeof(comment) == "string" ) com = " ; "+comment;
   nt = (tty == "long" || tty == "int");
   extra_lines = array("// ",nval);
   for( i = 1; i <= nval; i++ ) {
      if( nt ) {
         extra_lines(i) += keyword+" = "+itoa(value(i))+com;
      } else {
         extra_lines(i) += keyword+" = "+swrite(value(i),format="%.5g")+com;
      }
   }
   if( is_void(atline) ) {
      // locate first bunch of comment lines
      w = where(strpart(lines,1:2) == "//" );
      if( numberof(w) > 1 ) {
         v = where(w(dif) > 1);
         atline = numberof(v) ? w(v(1))+1 : w(0)+1;
      } else if( numberof(w) == 1 ) {
         atline = w(1)+1;
      } else atline = 1;
   }
   newlines = insert( extra_lines, lines, atline );

   write_slist,file,newlines;

}

/* Function computs */

func computs( file, keyword, value, atline=, comment= )
/* DOCUMENT computs, file, keyword, value, atline=, comment=

    Insert one or more string values under keyword 'keyword'
    immediately after first bunch of '//' comment lines.

    Keyword 'atline' forces insertion immediately before this
    line number.

    2013-06-04/NJW
*/
{
   lines = rdfile(file);
   nlines = numberof(lines);
   nval = numberof(value);
   tty = typeof(value);
   if( tty != "string" ) {
      write,"Datatype: "+tty+" is not supported!";
      return;
   }

   // Compose the extra lines

   com = "";
   if( typeof(comment) == "string" ) com = " ; "+comment;
   extra_lines = array("// ",nval);
   for( i = 1; i <= nval; i++ ) {
      extra_lines(i) += keyword+" = "+value(i)+com;
   }
   if( is_void(atline) ) {
      // locate first bunch of comment lines
      w = where(strpart(lines,1:2) == "//" );
      if( numberof(w) > 1 ) {
         v = where(w(dif) > 1);
         atline = numberof(v) ? w(v(1))+1 : w(0)+1;
      } else if( numberof(w) == 1 ) {
         atline = w(1)+1;
      } else atline = 1;
   }
   newlines = insert( extra_lines, lines, atline );

   write_slist,file,newlines;

}

/* Function scomput */

func scomput( lines, keyword, value, atline=, comment= )
/* DOCUMENT new_lines = scomput( lines, keyword, value, atline=, comment= )

    Insert one or more numerical values under keyword 'keyword'
    immediately after first bunch of '//' comment lines.

    Keyword 'atline' forces insertion immediately before this
    line number.

    2013-06-04/NJW
*/
{
   nlines = numberof(lines);
   nval = numberof(value);
   tty = typeof(value);
   if( tty != "long" && tty != "int" && tty != "double" && tty != "float" ) {
      write,"Datatype: "+tty+" is not supported!";
      return;
   }

   // Compose the extra lines

   com = "";
   if( typeof(comment) == "string" ) com = " ; "+comment;
   nt = (tty == "long" || tty == "int");
   extra_lines = array("// ",nval);
   for( i = 1; i <= nval; i++ ) {
      if( nt ) {
         extra_lines(i) += keyword+" = "+itoa(value(i))+com;
      } else {
         extra_lines(i) += keyword+" = "+swrite(value(i),format="%.5g")+com;
      }
   }
   if( is_void(atline) ) {
      // locate first bunch of comment lines
      w = where(strpart(lines,1:2) == "//" );
      if( numberof(w) > 1 ) {
         v = where(w(dif) > 1);
         atline = numberof(v) ? w(v(1))+1 : w(0)+1;
      } else if( numberof(w) == 1 ) {
         atline = w(1)+1;
      } else atline = 1;
   }
   newlines = insert( extra_lines, lines, atline );

   return newlines;

}

/* Function scomputs */

func scomputs( lines, keyword, value, atline=, comment= )
/* DOCUMENT new_lines = scomputs( file, keyword, value, atline=, comment= )

    Insert one or more string values under keyword 'keyword'
    immediately after first bunch of '//' comment lines.

    Keyword 'atline' forces insertion immediately before this
    line number.

    2013-06-04/NJW
*/
{
   nlines = numberof(lines);
   nval = numberof(value);
   tty = typeof(value);
   if( tty != "string" ) {
      write,"Datatype: "+tty+" is not supported!";
      return;
   }

   // Compose the extra lines

   com = "";
   if( typeof(comment) == "string" ) com = " ; "+comment;
   extra_lines = array("// ",nval);
   for( i = 1; i <= nval; i++ ) {
      extra_lines(i) += keyword+" = "+value(i)+com;
   }
   if( is_void(atline) ) {
      // locate first bunch of comment lines
      w = where(strpart(lines,1:2) == "//" );
      if( numberof(w) > 1 ) {
         v = where(w(dif) > 1);
         atline = numberof(v) ? w(v(1))+1 : w(0)+1;
      } else if( numberof(w) == 1 ) {
         atline = w(1)+1;
      } else atline = 1;
   }
   newlines = insert( extra_lines, lines, atline );

   return newlines;

}
