extern varfuncsdoc;
/* DOCUMENT varfuncs.i *******************************************

    Various functions - some intended for wavelet example

    2005-04-01/NJW

        badiomgis The inverse function of 'sigmoidab' (see below)
        diomgis   The inverse function of 'sigmoid' (see below)
        e0        Integral -inf -> x [exp(-0.5*t*t)] / sqrt(2pi)
        e1        Integral -x -> x   [exp(-0.5*t*t)] / sqrt(2pi)
        expo      Shorthand for exponential function: exp(-0.5*(x/sigma)^2)
        h_int     Returns x*exp(-x*x/2)
        hxexp     Returns integral[0 to x] t*exp(-t*t/2)*dt
        king1     Returns p(1)*(1 + ((x-p(2))/p(3))^2)^(-3*p(4) + 0.5)
                     p(1): amplitude, p(2): offset (x0), p(3): core (r0), p(4): beta
        sigmoid   Returns 1/(1 + exp(-(x-x0)/lambda))
        sigmoidab Returns (b-a)/(1 + exp(-(x-x0)/lambda)) + a
        sost      Returns 0.5*exp((x-x0)/lambda)        x < x0
                     or 1. - 0.5*exp((x0-x)/lambda)   x > x0

*********************************************/

/* Function king1 */

func king1( x, p )
/* DOCUMENT y = king1( x, p )

        Returns p(1)*(1 + ((x-p(2))/p(3))^2)^(-3*p(4) + 0.5)

        p(1): amplitude, p(2): offset (x0), p(3): core (r0), p(4): beta


   If 'p' is not given then the external variable King1_p is used.

    2016-07-11/NJW
*/
{
   extern King1_p;

   if( is_void(p) ) {
      if( is_void(King1_p) ) error,"Either argument 'p' or external 'King1_p' must be defined";
      pwork = King1;
   } else pwork = p;
   if( numberof(pwork) != 4 ) error,"pwork must be a 4-element vector";

   return pwork(1)*(1 + ((x-pwork(2))/pwork(3))^2)^(-3*pwork(4) + 0.5)
}

/* Function expo */

func expo(x,sigma)
/* DOCUMENT res = expo(x,sigma)

   Shorthand for exponential function: exp(-0.5*(x/sigma)^2)
*/

{
   return exp(-0.5*(x/sigma)^2);
}

/* Function h_int */

func h_int(x)
/* DOCUMENT res = h_int(x)

   returns x*exp(-x*x/2)
*/

{
   return x * exp(-x*x/2);
}

/* Function hxexp */

func hxexp(x)
/* DOCUMENT res = hxexp(x)

   returns integral[0 to x] t*exp(-t*t/2)*dt
*/

{
   return 1.0 - exp(-0.5*x*x);
}

/* Function e0 */

func e0(x)
/* DOCUMENT res = e0(x)

   Integral -inf -> x[exp(-0.5*t*t)] / sqrt(2pi)
*/
{
   xnorm = 0.3989423;
   p     = 0.231642;
   b1    = 0.319382;
   b2    = -0.356564;
   b3    = 1.78148;
   b4    = -1.82126;
   b5    = 1.33027;

   if( is_scalar(x) ) x = [x];
   res = array(double,numberof(x));
   w = where( x < -5.0 );
   if( numberof(w) ) res(w) = 0.0;
   w = where( x >  5.0 );
   if( numberof(w) ) res(w) = 1.0;
   w = where( x >= -5.0 & x <= 5.0 );
   if( numberof(w) ) {
      a = x(w);
      v = where(a < 0.0 );
      u = where(a >= 0.0 );
      if( numberof(v) ) a(v) = -a(v);
      t = 1./(1. + p*a);
      z = xnorm * exp( -0.5 * a * a );
      t2 = t * t;
      t3 = t2 * t;
      y = z * (b1*t + b2*t2 + b3*t3 + b4*t2*t2 + b5*t2*t3);
      if( numberof(u) ) y(u) = 1.0 - y(u);
      res(w) = y;
   }
   return numberof(x) == 1 ? res(1) : res;
}

/* Function e1 */

func e1( x )
/* DOCUMENT res = e1( x )

   Integral -x -> x   [exp(-0.5*t*t)] / sqrt(2pi)
*/
{
   return e0(x) - e0(-x);
}

/* Function sigmoid */

func sigmoid( x, x0, lambda )
/* DOCUMENT y = sigmoid( x, x0, lambda )

    returns 1/(1 + exp(-(x-x0)/lambda))

    2008-04-25/NJW
*/
{
   if( is_void(lambda) ) lambda = 1.0;
   if( is_void(x0) ) x0 = 0.0;
   lambda = double(lambda);
   x0 = double(x0);
   arg = -(x-x0)/lambda;
   if( anyof(arg > 400.) ) {
      w = where(arg > 400.);
      m = where(arg <= 400.);
      res = double(x);
      s = is_scalar(x);
      if(s) res=[res];
      res(w) = 0.0;
      if( numberof(m) ) res(m) = 1.0/(1.0 + exp(-(x(m)-x0)/lambda));
      if(s) res=res(1);
      return res;
   }

   return 1.0/(1.0 + exp(-(x-x0)/lambda));

}

/* Function diomgis */

func diomgis( y, x0, lambda )
/* DOCUMENT x = diomgis( y, x0, lambda )

   returns  x0 - lambda*log((1-y)/y)

   i.e. the inverse of 'sigmoid'
*/
{
   if( anyof(y >= 1) || anyof(y <= 0) ) error,"DIOMGIS: y is outside limits";
   if( is_void(x0) ) x0 = 0.0;
   if( is_void(lambda) ) lambda = 1.0;
   x0 = double(x0);
   lambda = double(lambda);
   y = double(y);

   return x0 - lambda * log( (1-y) / y );

}

/* Function sigmoidab */

func sigmoidab( x, x0, lambda, a, b )
/* DOCUMENT y = sigmoidab(  x, x0, lambda, a, b )

   returns y = a + (b-a)*sigmoid(x,x0,lambda)
*/
{
   return a + (b-a)*sigmoid(x,x0,lambda);
}

/* Function badiomgis */

func badiomgis( y, x0, lambda, a, b )
/* DOCUMENT x = badiomgis( y, x0, lambda, a, b )

    Inverse function of sigmoidab
*/
{
   yhat = (y - a)/(b - a);
   return diomgis( yhat, x0, lambda );
}

/* Function guard */

func guard( x, a, b, c, d )
/* DOCUMENT xhat = guard( x, a, b, c, d )

    Returns scalar or array where:
       xhat = x for a <= x <= b
       xhat > c for x < a
       xhat < d for x > b

    such that xhat outside [a,b] is nicely connected to inside.

    Requirements:   c < a   AND   d > b
            AND     b > a   AND   d > c

    If 'c' or 'd' are given as void (omitted) then no action will
    be taken for the equivalent part.

    Working tool is 'sigmoid'
*/
{
   s = is_scalar(x);
   res = double(x);
   if(s) res = [res];

   // Deal with part between a and b

   if( is_void(a) ) a = min(res);
   if( is_void(b) ) b = max(res);

   w = where( x >= a & x <= b );
   nw = numberof(w);

   if( !is_void(c) && !is_void(d) ) {
      if( d <= c ) error," d <= c";
   }
   if( !is_void(c) ) {
      if( c >= a ) error," c >= a";
      w = where( x < a );
      nw = numberof(w);
      if( nw ) {
         dd = 2.*a - c;
         lambda = 0.5*(a - c);
         res(w) = sigmoidab( x(w), a, lambda, c, dd );
      }
   }
   if( !is_void(d) ) {
      if( d <= b ) error," d <= b";
      w = where( x > b );
      nw = numberof(w);
      if( nw ) {
         cc = 2.*b - d;
         lambda = 0.5*(d - b);
         res(w) = sigmoidab( x(w), b, lambda, cc, d );
      }
   }

   if(s) res = res(1);
   return res;

}

/* Function sost */

func sost( x, x0, lambda )
/* DOCUMENT res = sost( x, x0, lambda )

      2003-04-07/NJW (IDL version)
      2006-08-23/NJW Yorick version

   Return exponential low value cutting function

*/
{
   sz = dimsof(x);
   if( sz(1) == 0 ) {
      if( x <= x0 ) {
         return 0.5*exp((x-x0)/lambda);
      } else {
        return 1.-0.5*exp((x0-x)/lambda);
      }
   } else {
      res = array( double, sz);
      w = where( x <= x0 );
      if( (nw = numberof(w)) > 0 ) res(w) = 0.5*exp((x(w)-x0)/lambda);
      w = where( x > x0 );
      if( (nw = numberof(w)) > 0 ) res(w) = 1. - 0.5*exp((x0 - x(w))/lambda);
   }
   return res;
}
