extern xfitdoc;
/* DOCUMENT ************************************************************
 *
 *    Functions for spectral modelling and fitting. PHA II spectral files
 *    are assumed.
 *
 *      xfit_anamin  analyze range of parameters for display
 *      xfit_bins    change detector bins to use in fit
 *      xfit_chi2    return Chi2 for at given parameter set
 *      xfit_chi2_1d_1
 *      xfit_chi2_1d_2
 *      xfit_chi2_1d_angle
 *      xfit_data    read spectral data from PHAII file + RMF
 *      xfit_ebase   set the powerlaw energy base (Ebase)
 *      xfit_fakeit  produces a simulated count spectrum
 *      xfit_fit     do the fitting using 'amoeba'
 *      xfit_freeze  freeze a given parameter to a value
 *      xfit_init    needs input from an observation
 *      xfit_lmfit   do the fitting using 'lmfit'
 *      xfit_log     define the log file
 *      xfit_model   sets the spectral model (Spectrcode)
 *      xfit_nfit    get the normalization
 *      xfit_pargrid find chi2 in a 2D grid
 *      xfit_photspec  define source spectrum from existing parameters
 *      xfit_plot    plotting results
 *      xfit_read    load basic information: RMF, ARF (for xfit_fakeit)
 *      xfit_show    Show variables in memory
 *      xfit_thaw    Thaw a given parameter (previously frozen)
 *      xfit_title   set the title for plots
 *      xfit_tune    Adjust statistical errors to chi2red = 1
 *
 *
 *      xfit_nustar_fakeit   fake a NuStar observation
 *      xfit_nustar_init  set up NuStar fitting
 *      xfit_nustar_plot  plotting NuStar results
 *
 *  Version 2.1, initiated 2012-03-28/NJW to rebin to sensible number of counts in each bin
 *      Version date: 2012-04-13/NJW
 *
 ***********************************************************************/

// setting the name of the parameter file (mainly for xfit_fakeit)

Parfilename = "xfit.par";

// setting the standard energy limits for the sflux calculation

Emin_std = 2.; // keV
Emax_std = 10.; // keV

Freeze = array(0,3);
Frozen_value = array(double,3);

// setting the default Ebase value

//+ Ebase = 5.; // [keV] Adequate value for LOFT WFC

Ebase = 1.0; // Commonly accepted value

/* Function xfit_pset */

func xfit_pset
/* DOCUMENT  xfit_pset

   is an interactive function for setting various
   external variables such as
      Emin_std, Emax_std, Ebase

*/
{
   extern Emin_std, Emax_std, Ebase;

   write,format="(1) Emin_std : %10.3f keV\n", Emin_std;
   write,format="(2) Emax_std : %10.3f keV\n", Emax_std;
   write,format="(3) Ebase    : %10.3f keV\n", Ebase;

   ans = strsplit(rdline(prompt="Enter number and new value : ... ")," ");
   n = atoi(ans(1));
   v = atof(ans(2));
   if( n == 1 ) {
      Emin_std = v;
      write,format="Emin_std : %10.3f keV\n", Emin_std;
   } else if( n == 2 ) {
      Emax_std = v;
      write,format="Emax_std : %10.3f keV\n", Emax_std;
   } else if( n == 3 ) {
      Ebase = v;
      write,format="Ebase    : %10.3f keV\n", Ebase;
   } else write,"Bad number, no action ...";

}

/* Function xfit_log */

func xfit_log( logfilename, reset= )
/* DOCUMENT xfit_log, logfilename, reset=
     or     xfit_log

    First version: A logging to 'logfilename' will be continued unless
                      the keyword 'reset' has been set.
    Second version: Terminates the logging

*/
{
// ynclude = xfit_package.externs_log
  extern  Logfilename,  Logflag;
// yxclude =

   if( is_void(logfilename) ) {
      Logflag = 0;
   } else {
      Logfilename = logfilename;
      Logflag = 1;
      cm = reset ? "w" : "a";
      f = open(logfilename,cm);
      write,f,format="//  Log file for xfit_package started %s\n", ndate(3);
      write,f,format="//    Errors are 3sigma i.e. 99%% confidenc%s\n","e";
      write,f,format="// %s\n", "colname = const";
      write,f,format="// %s\n", "colname = d_const";
      write,f,format="// %s\n", "colname = param1";
      write,f,format="// %s\n", "colname = d_param1";
      write,f,format="// %s\n", "colname = nh";
      write,f,format="// %s\n", "colname = d_nh";
      write,f,format="// %s\n", "colname = chi2red";
      write,f,format="// %s\n", "colname = specfile";
      close,f;
   }
}

/* Function xfit_init */

func xfit_init( dol_of_spec, dol_of_bkg, num_spec=, num_spec_bkg=, bins=, title=, model=, ebase= )
/* DOCUMENT xfit_init( dol_of_spec, dol_of_bkg, num_spec=, num_spec_bkg=, bins=, title=, model=, ebase= )

  Prepare fitting an observed X-ray spectrum from j_src_spectra or similar.
    Keywords:
         num_spec        row number of the spectrum to use (default: 1)
         num_spec_bkg    row number of the bkg spectrum to use (default: 1)
         bins            array with spectral bins to use in plots and fits
         title           plot title
         model           spectral model to use (PL, TB, BB, DF)
         ebase           Energy base for powerlaw spectrum

   Steps: xfit_init defines the energy channels, reads the spectrum,
                    defines the bins to use, reads the line energies,
                    reads and rebins the redistribution matrix, sets
                    the spectral model.
          xfit_fit  executes the fitting (no arguments)

          xfit_plot plots the data with the best fit

   NJW/2002-06-28
   2008-03-14/NJW translated from IDL to Yorick
   2011-05-10/NJW updated to get SPECRESP + ARF from header in spectral file

*/
{
   local specfile, extno, exposure, bexposure;
   local rate, stat_err, brate, bstat_err, rdm, brdm;

   //   Setup the common areas

// ynclude = xfit_package.externs_init
  extern  Arf,            E_max_orig,     Ehi,            Matrix_orig,    Spectrcode, \
          Dol_of_bkg,     E_min,          Eline,          Plottitle,      Stat_err, \
          Dol_of_spec,    E_min_orig,     Elo,            Rate,           Stat_err_orig, \
          E_max,          Ebase,          Matrix,         Rate_orig, Exposure;
// yxclude =

   if( !is_void(dol_of_spec) ) {
      get_exten_no, dol_of_spec, specfile, extno;
      if( !file_test(specfile) ) {
         write,format="Sorry, %s cannot be found\n", specfile;
         return;
      }
      Dol_of_spec = dol_of_spec;
   }
   if( !is_void(dol_of_bkg) ) {
      get_exten_no, dol_of_bkg, specfile, extno;
      if( !file_test(specfile) ) {
         write,format="Sorry, %s cannot be found\n", specfile;
         return;
      }
      Dol_of_bkg = dol_of_bkg;
   }
   Ebase = is_void(ebase) ? 1.0 : double(ebase);
   if( is_void(num_spec) ) num_spec = 1;
   if( is_void(num_spec_bkg) ) num_spec_bkg = 1;

   if( !is_void(dol_of_spec) ) { // Assume already known if not given

      if( is_void(title) ) {
         Plottitle = "";
         read,prompt="Enter the plot title: ... ",format="%s", Plottitle;
      } else {
         Plottitle = title;
      }

      //   Get the spectral data

      write,"Getting spectral data ...";
      type = j_get_src_spectrum( dol_of_spec, num_spec, rate, stat_err, rdm, exposure);
      Exposure = exposure;
      if( type == 1 ) write,"Warning! Spectrum flagged as BKG";

      //  Get the background spectrum if given as argument
      //    and do the subtraction with proper error assignment

      if( !is_void(dol_of_bkg) ) {
         write,"Getting background data ...";
         btype = j_get_src_spectrum( dol_of_bkg, num_spec_bkg, brate, bstat_err, brdm, bexposure);
         if( !near( bexposure, exposure, 0.99 ) ) {
            write,"Warning! BKG exposure differs from net exposure";
         }
         if( type != 1 ) write,"Warning! Bkg spectrum flagged as NET or TOTAL";
         rate -= brate;
         stat_err = sqrt(stat_err^2 + bstat_err^2);
      }

      Rate_orig = Rate = rate;
      Stat_err_orig = Stat_err = stat_err;

   }

   Matrix_orig = Matrix = *rdm.rdm;
   Elo = *rdm.energ_lo;
   Ehi = *rdm.energ_hi;
   Eline = sqrt(Elo * Ehi);
   Arf = *rdm.arf;
   E_min_orig = E_min = *rdm.e_min;
   E_max_orig = E_max = *rdm.e_max;

   //   Limit the spectral bins if requested

   if( !is_void(bins) ) xfit_bins, bins;

   //   Define the spectral model

   if( !is_void(model) ) {
      Spectrcode = strupcase(strtrim(model));
   } else {
      Spectrcode = "";
      while( Spectrcode != "PL" && Spectrcode != "TB" && Spectrcode != "BB" && Spectrcode != "DF" ) {
         read,prompt="What is the spectral model? (PL,TB,BB,DF) : ... ", \
               format="%s",Spectrcode;
         Spectrcode = strupcase(strtrim(Spectrcode));
      }
   }


}

/* Function xfit_data */

func xfit_data( dol_of_spec, dol_of_bkg, num_spec=, num_spec_bkg= )
/* DOCUMENT xfit_data, dol_of_spec, dol_of_bkg, num_spec=, num_spec_bkg=

  Prepare fitting an observed X-ray spectrum from j_src_spectra or similar.
    Keywords:
         num_spec        row number of the spectrum to use (default: 1)
         num_spec_bkg    row number of the bkg spectrum to use (default: 1)

   The detector energy bin selection is reset to all (can be updated with
                         'xfit_bins').

   Steps: xfit_data reads the spectrum and the associated anchor file and
                    RMF data.
          xfit_fit  executes the fitting (no arguments)

          xfit_plot plots the data with the best fit

   2011-05-13/NJW

*/
{
   local specfile, bkgfile, extno;
   local rate, stat_err, brate, bstat_err, rdm, brdm, exposure, bexposure;

   //   Setup the externals

// ynclude = xfit_package.externs_data
  extern  Arf,            E_max,          Ebase,          Matrix,         Rate_orig, \
          Bins,           E_max_orig,     Ehi,            Matrix_orig,    Specfile, \
          Dol_of_bkg,     E_min,          Eline,          Plottitle,      Stat_err, \
          Dol_of_spec,    E_min_orig,     Elo,            Rate,           Stat_err_orig, \
          Exposure;
// yxclude =

   if( !is_void(dol_of_spec) ) {
      get_exten_no, dol_of_spec, specfile, extno;
      if( !file_test(specfile) ) {
         write,format="Sorry, %s cannot be found\n", specfile;
         return;
      }
      Dol_of_spec = dol_of_spec;
   }
   if( !is_void(dol_of_bkg) ) {
      get_exten_no, dol_of_bkg, bkgfile, extno;
      if( !file_test(bkgfile) ) {
         write,format="Sorry, %s cannot be found\n", bkgfile;
         return;
      }
      Dol_of_bkg = dol_of_bkg;
   }
   Specfile = fullpath(specfile);
   Ebase = 1.0; // default setting
   Plottitle = specfile;
   if( is_void(num_spec) ) num_spec = 1;
   if( is_void(num_spec_bkg) ) num_spec_bkg = 1;

   //   Get the spectral data

   write,"Getting spectral data ...";
   type = j_get_src_spectrum( dol_of_spec, num_spec, rate, stat_err, rdm, exposure);
   Exposure = exposure;
   if( type == 1 ) write,"Warning! Spectrum flagged as BKG";

   //  Get the background spectrum if given as argument
   //    and do the subtraction with proper error assignment
   //    But only if spectral data are designated "TOTAL" (type == 2)

   if( !is_void(dol_of_bkg) && type == 2 ) {
      write,"Getting background data ...";
      btype = j_get_src_spectrum( dol_of_bkg, num_spec_bkg, brate, bstat_err, brdm, bexposure);
      if( !near( bexposure, exposure, 0.99 ) ) {
         write,"Warning! BKG exposure differs from net exposure";
      }
      if( type != 1 ) write,"Warning! Bkg spectrum flagged as NET or TOTAL";
      rate -= brate;
      stat_err = sqrt(stat_err^2 + bstat_err^2);
   }

   Rate_orig = Rate = rate;
   Stat_err_orig = Stat_err = stat_err;
   Bins = indgen(numberof(rate));

   Matrix_orig = Matrix = *rdm.rdm;
   Elo = *rdm.energ_lo;
   Ehi = *rdm.energ_hi;
   Eline = sqrt(Elo * Ehi);
   Arf = *rdm.arf;
   E_min_orig = E_min = *rdm.e_min;
   E_max_orig = E_max = *rdm.e_max;

}

/* Function xfit_model */

func xfit_model( model )
/* DOCUMENT xfit_model, spectrcode
    or      xfit_model

      Can be PL, TB, BB, or DF
*/
{
// ynclude = xfit_package.externs_model
  extern  Spectrcode;
// yxclude =

   //   Define the spectral model

   if( !is_void(model) ) {
      Spectrcode = strupcase(strtrim(model));
   } else {
      Spectrcode = "";
      while( Spectrcode != "PL" && Spectrcode != "TB" && Spectrcode != "BB" && Spectrcode != "DF" ) {
         read,prompt="What is the spectral model? (PL,TB,BB,DF) : ... ", \
               format="%s",Spectrcode;
         Spectrcode = strupcase(strtrim(Spectrcode));
      }
   }
}

/* Function xfit_fit */

func xfit_fit( &p_res, cont= )
/* DOCUMENT chi2_red = xfit_fit( >p_res, cont= )

  Fitting a model to an observed spectrum.
  If keyword 'cont' is set then p_res are the initial values
  for the fitting process.
  Freezing/thawing of a variable can be done with functions
  xfit_freeze and xfit_thaw resp.

    NJW/2002-06-28
    2008-03-14/NJW Translated to Yorick
    2011-05-11/NJW updated to be more general

*/
{

// ynclude = xfit_package.externs_fit
  extern  Chi2min,       Rate_model,     Frozen_value,  Nh_bf,         Param1_bf, \
          Chi2red,       Ehi,           N_freedom,     Normfit,       Spectrcode, \
          Const,         Elo,           Nh,            Param1,        Sx, \
          Const_bf,      Freeze,        Emin_std,      Emax_std, Logflag, Logfilename, \
          Rate, Stat_err;
// yxclude =

   Normfit = 0;

/*
 *********     BEGIN  FITTING  PROCESS    *******************
 */

   // Define weights

   m = where( Stat_err <= 0.0 );
   if( numberof(m) ) {
      write,"Warning - "+itoa(numberof(m))+" stat errs are zero or negative";
      Stat_err(m) = 1.;
   }
   weights = 1./Stat_err^2;

   p = [1.,1.,1.];
   if( cont && numberof(p_res) == 3 ) p = p_res;

   p_to_fit = [];
   if( Freeze(1) ) p(1) = Frozen_value(1);
   else grow,p_to_fit,1;
   if( Freeze(2) ) p(2) = Frozen_value(2);
   else grow,p_to_fit,2;
   if( Freeze(3) ) p(3) = Frozen_value(3);
   else grow,p_to_fit,3;

   resfit = lmfit( xfit_rate, dummy, p, Rate, weights, fit=p_to_fit );
   p_res = p;

   write,format="resfit.neval = %i\n", resfit.neval;
   write,format="resfit.niter = %i\n", resfit.niter;
   write,format="resfit.chi2_last = %f\n", resfit.chi2_last;

   //  this call will silently update the 'Rate_model' external
   //  in 'xfit_chi2'

/*
 *********     END  OF  FITTING   PROCESS    *******************
 */

   //   Present the results

   if( Logflag ) {
      fh = open( Logfilename, "a" );
      write,fh,format="// Fitting ...%s","\n";
   }

   N_freedom = numberof(Rate_model) - 3 + Freeze(sum);

   Chi2min = xfit_chi2(p_res);
   Chi2red = Chi2min/N_freedom;
   write,format="Chi2,red =   %10.4f\n", Chi2red;
   if( Logflag ) write,fh,format="// Chi2,red =   %10.4f\n", Chi2red;

   wstr = swrite(format="Constant =    %12.3e",p_res(1));
   if( Freeze(1) ) wstr += " (frozen)";
   write,format="%s\n", wstr;
   if( Logflag ) write, fh, format="// %s\n", wstr;

   wstr = swrite(format="Param1   =    %10.4f",p_res(2));
   if( Freeze(2) ) wstr += " (frozen)";
   write,format="%s\n", wstr;
   if( Logflag ) write, fh, format="// %s\n", wstr;

   wstr = swrite(format="Nh/1e22  =    %10.4f",p_res(3));
   if( Freeze(3) ) wstr += " (frozen)";
   write,format="%s\n", wstr;
   if( Logflag ) write, fh, format="// %s\n", wstr;

   // Setting externals with best fit parameters

   Const_bf = Const = p_res(1);
   Param1_bf = Param1 = p_res(2);
   Nh_bf = Nh = p_res(3)*1.e22;

   //   Get the photon spectrum and source intensity

   if( Spectrcode  ==  "DF" ) {
      flux = xfit_photspec(p_res);
      dum = max(flux);
      index = where(dum==flux)(1);
      Sx = dum * (Ehi(index)-Elo(index)) * 1.6e-9;
      write,format="Sx          = %13.4e erg/cm2/s\n", Sx;
      if( Logflag ) write,fh,format="// Sx          = %13.4e erg/cm2/s\n", Sx;
   } else {
      sx_eline = spanl(Emin_std,Emax_std,200);
      flux = xfit_photspec(p_res, eline=sx_eline);
      Sx = sflux( Emin_std, Emax_std, sx_eline, flux);
      write,format="Sx(%.0f-%.0f keV) = %13.4e erg/cm2/s\n", Emin_std, Emax_std, Sx;
      if( Logflag ) write,fh,format="// Sx(%.0f-%.0f keV) = %13.4e erg/cm2/s\n", Emin_std, Emax_std, Sx;
   }

   if( Logflag ) close, fh;

   return Chi2red;

}

/* Function xfit_nfit */

func xfit_nfit( &p_res )
/* DOCUMENT chi2_red = xfit_nfit( >p_res )

  Fitting normalization for a model to an observed spectrum
  i.e. only one free parameter.

  p_res is both input and output
  p(1) will be calculated
  p(2) is Photon Index
  p(3) is NH - absorption column in 1e22 cm-2

    2008-08-20/NJW

*/
{

// ynclude = xfit_package.externs_nfit
  extern  Chi2min,     Const_bf,    Elo,         Nh_bf,       Param1_bf,   Stat_err, \
          Chi2red,     Rate_model,   N_freedom,   Normfit,     Rate,        Sx, \
          Const,       Ehi,         Nh,          Param1,      Spectrcode, Emin_std, Emax_std;
// yxclude =


   p_res(1) = 1.0;
   Normfit = 1;

   c2 = xfit_chi2( p_res );

   p_res(1) = sum((Rate/Stat_err)^2) / sum(Rate_model*Rate/Stat_err^2);

   //   Present the results

   N_freedom = numberof(Rate_model) - 1;

   Chi2min = xfit_chi2(p_res);
   // This call updates the external variable 'Rate_model'

   Chi2red = Chi2min/N_freedom;
   write,format="Chi2,red =   %10.4f\n", Chi2red;

   write,format="Constant =    %10.4f\n",p_res(1);
   write,format="Param1   =    %10.4f (frozen)\n",p_res(2);
   write,format="Nh/1e22  =    %10.4f (frozen)\n",p_res(3);

   // Defining the best fit parameters

   Const_bf = Const = p_res(1);
   Param1_bf = Param1 = p_res(2);
   Nh_bf = Nh = p_res(3)*1.e22;

   //   Get the photon spectrum and source intensity

   if( Spectrcode  ==  "DF" ) {
      flux = xfit_photspec(p_res);
      dum = max(flux);
      index = where(dum==flux)(1);
      Sx = dum * (Ehi(index)-Elo(index)) * 1.6e-9;
      write,format="Sx          = %13.4e erg/cm2/s\n", Sx;
   } else {
      sx_eline = spanl(Emin_std,Emax_std,200);
      flux = xfit_photspec(p_res, eline=sx_eline);
      Sx = sflux( Emin_std, Emax_std, sx_eline, flux);
      write,format="Sx(%.0f-%.0f keV) = %13.4e erg/cm2/s\n", Emin_std, Emax_std, Sx;
   }

   return Chi2red;

}

/* Function xfit_chi2 */

func xfit_chi2( p )
/* DOCUMENT res = xfit_chi2( p )

  Fit function that returns Chi2 for( a given set of parameters
  in array "p":  (norm, param1, nh/1.e22)

  Side effect:  Updates the external 'Rate_model' !

   NJW/2002-06-28
   2008-03-14/NJW Translated to Yorick
*/
{

   if( is_void(p) ) {
      write,"Syntax: chi2 = xfit_chi2(p)";
      return [];
   }
   if( numberof(p) !=  3 ) {
      write,"xfit_chi2 must be called with 3 element array";
      return [];
   }

   //   Define the externals

// ynclude = xfit_package.externs_chi2
  extern  Arf,        Ehi,        Elo,        Matrix,     Rate,       Stat_err, \
          Rate_model;
// yxclude =

   //   Generate the input spectrum

   flux = xfit_photspec(p);

   line_counts = flux * (Ehi - Elo) * Arf;
   Rate_model = Matrix(,+)*line_counts(+);

   c2 = chi2( Rate_model, Rate, Stat_err );

   return c2;

}

/* Function xfit_rate */

func xfit_rate( dummy, p, outfile= )
/* DOCUMENT res = xfit_rate( dummy, p, outfile= )

  Fit function that returns the model rate for a given set of parameters
  in array "p":  (norm, param1, nh/1.e22)
  'dummy' is an argument required by lmfit - not used here.

  Used as fitting function in 'lmfit' (xfit_lmfit)

  Side effect:  Updates the external 'Rate_model' !

   A FITS output file may be produced by setting keywork 'outfile' to a valid
   filename string.

   2011-09-15/NJW
*/
{

   //   Define the externals

// ynclude = xfit_package.externs_chi2
  extern  Arf,        Ehi,        Elo,        Matrix,    Rate_model, \
          E_min, E_max;
// yxclude =

   //   Generate the input spectrum

   line_counts = xfit_photspec(p) * (Ehi - Elo) * Arf;
   Rate_model = Matrix(,+)*line_counts(+);

   if( typeof(outfile) == "string" ) {
      kwds_init;
      kwds_set,"DATE",ndate(3),"Date/time of file production";
      kwds_set,"ORIGIN","xfit_package-2.1{xfit_rate}","Software";
      for( i = 1; i <= numberof(p); i++ ) {
         kwds_set,"PARAM"+itoa(i),p(i),"Parameter value";
      }
      wrmfitscols, outfile, "E_MIN", E_min, "E_max", E_max, "RATE", Rate_model;
   }

   return Rate_model;

}

/* Function xfit_nustar_fakeit */

func xfit_nustar_fakeit(a)
/* DOCUMENT xfit_nustar_fakeit


   NJW/2002-06-28
   2008-03-14/NJW Translated to Yorick
   2008-05-22/NJW
*/
{

   //   Define the externals

// ynclude = xfit_package.externs_nustar_fakeit
  extern  Arf,            Countspec_err,  Elo,            Nh,             Rate, \
          Const,          Ehi,            Exposure,       Param1,         Stat_err, \
          Countspec,      Eline,          Matrix, \
            Emin_std, Emax_std, Photonspec;
// yxclude =

   p = array(double,3);

   Const = 1.0; Param1 = 1.0; nh = 1.0;
   read,prompt="Enter constant : ... ",format="%f", Const;
   read,prompt="Enter param1   : ... ",format="%f", Param1;
   read,prompt="Enter NH/1.e22 : ... ",format="%f", nh;
   Exposure = 1.0;
   read,prompt="Enter exposure : ... ",format="%f", Exposure;


   p(1) = Const;
   p(2) = Param1;
   p(3) = nh;
   Nh = p(3) * 1.e22;

   //   Generate the input spectrum

   Photonspec = xfit_photspec(p);
   write,format="Sflux %.0f-%.0f keV = %12.3e erg/cm2/s\n", \
              Emin_std, Emax_std, sflux(Emin_std,Emax_std,Eline, Photonspec);

   // Note the difference between 'Countspec' that is the observed (or
   // simulated) number of counts, whereas 'Rate_model' is the result of
   // the current spectral parameter values.

   line_counts = Photonspec * (Ehi - Elo) * Arf;
   Countspec = (Matrix(,+)*line_counts(+))*Exposure;
   countspec_err = sqrt(Countspec);
   Countspec += random_n(numberof(Countspec))*countspec_err;
   Rate = Countspec / Exposure;
   Stat_err = countspec_err / Exposure;

   write,"Rate and Stat_err are now prepared for e.g. plotting";

}

/* Function xfit_nustar_init */

func xfit_nustar_init( dol_of_spec, num_spec )
/* DOCUMENT xfit_nustar_init( dol_of_spec, num_spec )

  prepare fitting an observed X-ray spectrum from j_src_spectra

   2008-05-22/NJW cloned from xfit_init.i

*/
{

   //   Setup the common areas

// ynclude = xfit_package.externs_nustar_init
  extern  Arf,         E_min,       Eline,       Matrix,      Spectrcode,  Stat_err, \
          E_max,       Ehi,         Elo,         Rate;
// yxclude =

   if( is_void(num_spec) ) num_spec = 1;
   number = num_spec;

   //   Set the rebinning i.e. avoid first 30 channels
   //   and last 56

   //+ rebin = indgen(200-30) + 30;

   //   Get the spectral data

   
   if( !is_void( dol_of_spec ) ) {
      j_get_src_spectrum, dol_of_spec, number, rate, stat_err;
      Rate = rate(rebin);
      Stat_err = stat_err(rebin);
   }


   //   Get the RMF data

   //+ j_get_pi_ebds, e_min, e_max;
   rmfname = "/home/njw/nustar/rmf.fits";
   arfname = "/home/njw/nustar/ARF_onaxis.fits";

   e_min = rdfitscol(rmfname+"+2","e_min");
   e_max = rdfitscol(rmfname+"+2","e_max");
   rebin = indgen(numberof(e_min));

   E_min = e_min(rebin);
   E_max = e_max(rebin);

   Elo = rdfitscol(rmfname+"[MATRIX]","ENERG_LO");
   Ehi = rdfitscol(rmfname+"[MATRIX]","ENERG_HI");
   Eline = sqrt(Elo*Ehi);
   matrix = rdfitscol(rmfname+"[MATRIX]","MATRIX");
   Arf = rdfitscol(arfname+"+1","SPECRESP");
   elo_Arf = rdfitscol(arfname+"+1","ENERG_LO");
   ehi_Arf = rdfitscol(arfname+"+1","ENERG_HI");
   e_Arf = 0.5*(elo_Arf + ehi_Arf);
   Arf = interp( Arf, e_Arf, Eline);

   Matrix = matrix(rebin,*);

   //   Define the spectral model

   Spectrcode = "";
   while( Spectrcode != "PL" && Spectrcode != "TB" && Spectrcode != "DF" ) {
      read,prompt="What is the spectral model? (PL,TB,DF) : ... ", \
            format="%s",Spectrcode;
      Spectrcode = strupcase(strtrim(Spectrcode));
   }


}

/* Function xfit_nustar_plot */

func xfit_nustar_plot( a )
/* DOCUMENT xfit_nustar_plot

  plots the result of rfitting
    Default is energy plot.

   NJW/2002-06-28
   2008-05-16/NJW translated to Yorick
*/
{

   //   Define the common areas

// ynclude = xfit_package.externs_nustar_plot
  extern  E_max,      E_min,      Plottitle,  Rate,       Stat_err;
// yxclude =

   window,0,style="boxed.gs";

   REBIN = 75;
   z = array(3,15); grow, REBIN, z;
   z = array(6,40); grow, REBIN, z;
   z = array(10,20); grow, REBIN, z;
   z = array(20,20); grow, REBIN, z;
   z = array(30,20); grow, REBIN, z;

   local eb1, eb2, orate, ostat_err;
   specrebinning, E_min, E_max, Rate, Stat_err, REBIN, eb1, eb2, orate, ostat_err;
   //+ w = where( Rate > Stat_err & Stat_err > 0.0 );
   //+ ymin = min((Rate(w)-Stat_err(w))/(E_max(w)-E_min(w)));
   //+ ymax = max((Rate(w)+Stat_err(w))/(E_max(w)-E_min(w)));
   //+ dataplot, 0.5*(E_min+E_max), Rate/(E_max-E_min), Stat_err/(E_max-E_min), \
   //+     yr=[ymin,ymax], xr=[5,80],  xbar=1,itype=3, \
   //+     xtitle="Energy [keV]",ytitle="Cts /s /keV",title=esc_underscore(Plottitle);
   w = where( orate > ostat_err & ostat_err > 0.0 );
   ymin = min((orate(w)-ostat_err(w))/(eb2(w)-eb1(w)));
   ymax = max((orate(w)+ostat_err(w))/(eb2(w)-eb1(w)));
   dataplot, 0.5*(eb1+eb2), orate/(eb2-eb1), ostat_err/(eb2-eb1), \
       yr=[ymin,ymax*2], xr=[3,90],  xbar=1,itype=3, \
       xtitle="Energy [keV]",ytitle="Cts /s /keV",title=esc_underscore(Plottitle);


}

/* Function xfit_photspec */

func xfit_photspec( p, eline=, spectrcode= )
/* DOCUMENT phot_spec = xfit_photspec( p, eline=, spectrcode= )


  Function that returns the photon spectrum for a
  given set of spectral parameters
  in array "p":  (norm, param1, nh/1.e22)

   Default operation uses the externals Eline and Spectrcode
   These can be overridden by the keywords.

   NJW/2002-06-28
   2011-05-11/NJW  updated with keywords

*/
{
   if( is_void(p) ) {
      write,"Syntax: flux = xfit_photspec(p)";
      return [];
   }
   if( numberof(p)  !=  3 ) {
      write,"xfit_photspec must be called with 3 element array";
      return [];
   }

   //   Define the common areas

// ynclude = xfit_package.externs_photspec
  extern  Const,       Ebase,       Eline,       Nh,          Param1,      Spectrcode;
// yxclude =

   Const = p(1);
   Param1 = p(2);
   Nh = p(3) * 1.e22;

   aEline = is_void(eline) ? Eline : eline;
   aSpectrcode = is_void(spectrcode) ? Spectrcode : strupcase(spectrcode);

   //   Generate the input spectrum


   if( aSpectrcode == "PL" ) {
                flux = Const * (aEline/Ebase)^(-Param1) * absorp(Nh, aEline);
   } else if( aSpectrcode == "TB" ) {
                // 2012-03-29  Param1 should allways > 0.1
                //+ proxy_param1 = guard( Param1, 0.1, , 0.01 );
                flux = Const * (exp(-aEline/Param1) / aEline) * absorp(Nh, aEline);
   } else if( aSpectrcode == "BB" ) {
                // 2012-03-29  Param1 should allways > 0.1
                //+ proxy_param1 = guard( Param1, 0.1, , 0.01 );
                flux = Const * (aEline^2 / (exp(aEline/Param1) - 1 )) * absorp(Nh, aEline);
   } else if( aSpectrcode == "DF" ) {
                dum = min(abs(aEline-Param1));
                index = where(dum==abs(aEline-Param1))(1);
                flux = array(double, numberof(aEline));
                flux(index) = Const * absorp(Nh, Param1);
   } else {
                write,format="No such spectral model allowed: %s\n", aSpectrcode;
                return [];
   }

   return flux;

}

/* Function xfit_plot */

func xfit_plot( frebin=, outfile=, pname=, win=, nofit= )
/* DOCUMENT xfit_plot, frebin=, outfile=, pname=, win=, nofit=

  Plots the current Rate, Stat_err

   outfile    (string) name of output text file with spectra
                <set>  the output file becomes xfit_nnnn.scm
   frebin     sets a rebinning before the plotting where the value
              is the maximal sigma/signal
   pname      is an optional plot name (string)
   win        defines plotting window (default is current)
   nofit      for plotting only data and skip the fit result

   NJW/2002-06-28
   2008-05-16/NJW translated to Yorick
*/
{

// ynclude = xfit_package.externs_plot
  extern  Const,         Dol_of_spec,   Freeze,        Nh,            Plottitle, \
          Rate_model,     E_max,         Frozen_value,  Normfit,       Rate, \
          Dol_of_bkg,    E_min,         N_freedom,     Param1,        Stat_err, \
           Emin_std, Emax_std, Spectrcode;
// yxclude =

   local oemin, oemax, orate, ostat_err, orate_model, xdum, ydum, zdum;

   if( is_void(win) ) win = 3;
   // Define two plot frames in the same window
   mplot_setup, 21, hideticklabels=1, vrelsize=[3,1], spacing=0.005, \
          vport=[0.15,0.7,0.4,0.85], pane=win;

   fstr = ["","",""];
   if( Freeze(sum) ) fstr(where(Freeze)) = " (frozen)";

   if( numberof(outfile) ) {
      if( typeof(outfile) == "string" ) {
         dump = 1;
      } else {
         outfile = get_next_filename("xfit_????.scm");
         dump = 1;
      }
   } else dump = 0;
 
   // Make sure that Rate_model is for the current parameters
   // but only if the variables exist

   if( !is_void(Const) && !is_void(Param1) && !is_void(Nh) && !nofit ) {
      chi2red = xfit_chi2( [Const, Param1, Nh*1.e-22] )/N_freedom;
      // this call updates 'Rate_model'
      sx_eline = span(Emin_std,Emax_std,200);
      flux = xfit_photspec( [Const, Param1, Nh*1.e-22], eline=sx_eline );
      sx = sflux( Emin_std, Emax_std, sx_eline, flux);
   }

   if( frebin ) {
      REBIN = specrebinninga( E_min, E_max, Rate, Stat_err, frebin, oemin, oemax, orate, ostat_err );
      specrebinning, E_min, E_max, Rate_model, Stat_err, REBIN, xdum, ydum, orate_model, zdum;
   } else {
      oemin = E_min;
      oemax = E_max;
      orate = Rate;
      ostat_err = Stat_err;
      orate_model = Rate_model;
   }

  /*
   *   Note that the Yorick peculiarity of plotting negative values on a log scale
   *   actually shows the absolute values. The negative rate values must therefore
   *   be changed to a resonable (positive) value before plotting.
   */

   wneg = where( orate <= 0.0 );
   nwneg = numberof( wneg );
   or = orate / (oemax - oemin);
   if( nwneg ) {
      wpos = where( orate >  0.0 );
      ymin = min(or(wpos));
      ymax = max(or(wpos));
      r = (ymax / ymin)^0.05;
      ymax *= r;
      ymin /= r;
      or(wneg) = ymin;
   } else {
      ymin = or(min);
      ymax = or(max);
      r = (ymax / ymin)^0.05;
      ymax *= r;
      ymin /= r;
   }

   //+ ymin = min([(orate-ostat_err)/(oemax-oemin),Countspec/(oemax-oemin)]);
   //+ ymax = max([(orate+ostat_err)/(oemax-oemin),Countspec/(oemax-oemin)]);
   //+ xmin = (0.5*(oemin+oemax))(1) * 0.98;
   //+ xmax = (0.5*(oemin+oemax))(0) * 1.02;
   xmin = oemin(1) * 0.95;
   xmax = oemax(0) * 1.05;

   // fix lower y boundaries < 0 to something positive

   ylower = orate/(oemax-oemin) - ostat_err/(oemax-oemin);
   wneg = where( ylower <= 0.0 );
   nwneg = numberof(wneg);
   if( nwneg ) ylower(wneg) = ymin;

   // fix upper y boundaries < 0 to something positive

   yupper = orate/(oemax-oemin) + ostat_err/(oemax-oemin);
   wneg = where( yupper <= 0.0 );
   nwneg = numberof(wneg);
   if( nwneg ) yupper(wneg) = ymin;

   dataplotf, oemin,sqrt(oemin*oemax), oemax,  \
       ylower, \
       //+ orate/(oemax-oemin) - ostat_err/(oemax-oemin), \
       or, \
       //+ orate/(oemax-oemin), \
       yupper, \
       //+ orate/(oemax-oemin) + ostat_err/(oemax-oemin), \
       xr=[xmin,xmax], yr=[ymin,ymax], itype=3, mpl=1, \
       ytitle="Cts /s /keV",title=esc_underscore(Plottitle), \
       xyadjust=[0.02,0.02],titlefont="times",xytitleheight=14;
   if( !is_void(orate_model) && !nofit ) oplthis,oemin, oemax, orate_model/(oemax-oemin), color="red";
   plotsign;
   if( structof(pname) == string ) plotname, pname;

   if( !nofit ) {

      ytxt = 0.55;
      dy   = 0.020;
      csize = 0.7;
      if( !is_void(chi2red) ) \
      xyouts,0.2,ytxt,swrite(format="!c^2^_red_  =     %9.3f",chi2red),ndc=1,charsize=csize;
      ytxt += dy;
      if( !is_void(sx) ) \
      xyouts,0.2,ytxt,swrite(format="S_x_ (%.0f-%.0f keV) =  %s erg/s/cm^2^", \
               Emin_std, Emax_std, f2scienota(sx,ndec=2)),ndc=1,charsize=csize;
      ytxt += dy;
      if( !is_void(Param1) ) \
      xyouts,0.2,ytxt,swrite(format="Param1   =     %9.3f%s",Param1,fstr(2)),ndc=1,charsize=csize;
      ytxt += dy;
      if( !is_void(Nh) ) \
      xyouts,0.2,ytxt,swrite(format="Nh/10^22^  =     %9.3f%s",Nh*1.e-22,fstr(3)),ndc=1,charsize=csize;
      ytxt += dy;
      if( !is_void(Const) ) {
         if( Const < 1.e-2 ) {
            xyouts,0.2,ytxt,swrite(format="Constant =     %s%s",f2scienota(Const,ndec=2),fstr(1)),ndc=1,charsize=csize;
         } else {
            xyouts,0.2,ytxt,swrite(format="Constant =     %9.3f%s",Const,fstr(1)),ndc=1,charsize=csize;
         }
      }
      ytxt += dy;
      if( !is_void(Spectrcode) ) \
      xyouts,0.2,ytxt,swrite(format="Spec model %s",Spectrcode),ndc=1,charsize=csize;

      // Plotting ratio

      if( !is_void(orate_model) ) {
         w = where( orate_model > 0.0 & orate > 0.0 );

         y = orate(w)/orate_model(w);
         ylower = orate(w)/orate_model(w) - ostat_err(w)/orate_model(w);
         yupper = orate(w)/orate_model(w) + ostat_err(w)/orate_model(w);
         ymin = y(min);
         ymax = y(max);
         r = (ymax / ymin)^0.15;
         ymax *= r;
         ymin /= r;

         // fix lower y boundaries < 0 to something positive

         wneg = where( ylower <= 0.0 );
         nwneg = numberof(wneg);
         if( nwneg ) ylower(wneg) = ymin;

         dataplotf,oemin(w), sqrt(oemin(w)*oemax(w)), oemax(w), \
            ylower, \
            //+ orate(w)/orate_model(w) - ostat_err(w)/orate_model(w), \
            y, \
            //+ orate(w)/orate_model(w), \
            yupper, \
            //+ orate(w)/orate_model(w) + ostat_err(w)/orate_model(w), \
            xr=[xmin,xmax],yr=[ymin,ymax],mpl=2,itype=3,xtitle="Energy [keV]", ytitle="Ratio", color="blue", \
             xyadjust=[0.02,0.02],titlefont="times",xytitleheight=14;
         oplot,sqrt(oemin*oemax),array(1.,numberof(oemin)),li=2,color="blue";
      }
   }

   // Dump results to file if required

   if( dump && !nofit ) {
      write,format="Saving curves in %s\n", outfile;
      hdr = [];
      grow, hdr, "// xfit results  "+ndate(3);
      grow, hdr, "// DOL_SPE = "+Dol_of_spec;
      if( numberof(Dol_of_bkg) ) grow, hdr, "// DOL_BKG = "+Dol_of_bkg;
      for( i = 1; i <= 3; i++ ) {
         if( Freeze(i) ) {
            grow, hdr,swrite(format="// freeze = %i", i);
            grow, hdr,swrite(format="// value = %f", Frozen_value(i));
         }
      }
      if( !is_void(chi2red) ) \
      grow, hdr, swrite(format="// Chi2red = %f", chi2red);
      if( !is_void(sx) ) \
      grow, hdr, swrite(format="// Sx = %e; erg/cm2/s", sx);
      if( !is_void(Const) ) \
      grow, hdr, swrite(format="// Const = %f", Const);
      if( !is_void(Param1) ) \
      grow, hdr, swrite(format="// Param1 = %f", Param1);
      if( !is_void(Nh) ) \
      grow, hdr, swrite(format="// Nh = %f; 1e22 /cm2", Nh*1.e-22);
      grow, hdr, "//";
      grow, hdr, "// colname = Energy; keV";
      grow, hdr, "// colname = Rate; cts/s";
      grow, hdr, "// colname = Stat_err; cts/s";
      grow, hdr, "// colname = Rate_model; cts/s the model";
      grow, hdr, "//";
      wstab,outfile,sqrt(E_min*E_max),Rate,Stat_err,Rate_model,hdr=hdr;
   }

}

/* Function xfit_show */

func xfit_show
/* DOCUMENT xfit_show

    For a dump to screen of the externals.
*/
{
// ynclude = xfit_package.externs_show
  extern  Dol_of_bkg,    E_max,         Ebase,         Elo,           Frozen_value, \
          Dol_of_spec,   E_min,         Ehi,           Freeze,        Spectrcode;
// yxclude =

   names = ["Const","Param1","Nh"];

   write,format="Dol_of_spec: %s\n", Dol_of_spec;
   if( is_void(Dol_of_bkg) ) {
      write,format="Dol_of_bkg: %s\n", "<not given>";
   } else {
      write,format="Dol_of_bkg: %s\n", Dol_of_bkg;
   }
   write,format="Ebase = %.3f keV, energy base for powerlaw\n", Ebase;
   write,format="E_min and E_max has %i bins from %.3f - %.3f keV\n", numberof(E_min), E_min(1), E_max(0);
   write,format="Elo and Ehi has %i elements from %.3f - %.3f keV\n", numberof(Elo), Elo(1), Ehi(0);
   write,format="Matrix, Arf, Eline exist, current spectral code: %s\n", Spectrcode;
   if( Freeze(sum) ) {
      for( i = 1; i <= 3; i++ ) {
         if( Freeze(i) ) \
         write,format="Parameter %i (%s) is frozen to %.3f\n",i, names(i), Frozen_value(i);
      }
   } else {
      write,"No frozen parameters";
   }
   write,"  Other variables:";
   write,"Arf Bins Chi2red Const Rate_model **_orig Freeze Frozen_value Nh Normfit";
   write,"Param1 Plottitle Rate Spectrcode Stat_err Sx";
   if( !is_void(Spectrcode) ) write,format="\nCurrent spectral model: %s\n", Spectrcode;

}

/* Function xfit_bins */

func xfit_bins( bins, bins2, cont= )
/* DOCUMENT xfit_bins, bins[, bins2][, cont=]

   Use a new subset of spectral bins for fitting etc.
   'bins' is void -> reset to include all bins
   'bins' is a positive scalar -> indgen(bins)
   'bins' is a negative scalar -> rebin to at lest -nbins counts in each bin
   'bins' is array -> no change
   'bins2' given and bins2 > bins (both scalar) -> indgen(bins:bins2)

    Keyword 'cont' is for 'continue' i.e. not setting it will rebin from start,
    but combined rebinning can be done by e.g.

    > xfit_bins, 100;
    > xfit_bins, -10, cont=1;

    that will use first 100 bins and then combine to have at least 10 counts in each.
*/
{
// ynclude = xfit_package.externs_bins
  extern  Bins,           E_min,          Matrix,         Rate,           Stat_err, \
          E_max,          E_min_orig,     Matrix_orig,    Rate_orig,      Stat_err_orig, \
          E_max_orig, Exposure;
// yxclude =

   if( cont ) {
      rate_start = Rate;
      stat_err_start = Stat_err;
      e_min_start = E_min;
      e_max_start = E_max;
      matrix_start = Matrix;
   } else {
      rate_start = Rate_orig;
      stat_err_start = Stat_err_orig;
      e_min_start = E_min_orig;
      e_max_start = E_max_orig;
      matrix_start = Matrix_orig;
   }

   if( is_void(bins) ) bins = indgen(numberof(rate_start));

   //   First see if 'bins' is scalar and negative

   if( is_scalar(bins) ) {
      if( bins < 0 ) {    // binning to at least -bins counts in each bin
         REBIN = specrebinningb( e_min_start, e_max_start, rate_start, stat_err_start, \
                      Exposure, double(-bins), E_min, E_max, Rate, Stat_err );
         Matrix = rebin_rdm( matrix_start, REBIN );
         write,format="New E_min/max: %.3f %.3f keV with %i bins\n", E_min(1), E_max(0), numberof(Rate);
         return;
      }
   }

   //  else apply bin specifications

   if( is_scalar(bins) ) {
      if( is_void(bins2) ) {
         bins = indgen(bins);
      } else {
         if( is_scalar(bins2) ) {
            if( bins2 > bins ) {
               bins = indgen(bins:bins2);
            } else error,"bins2 <= bins";
         } else error,"bins2 is not scalar";
      }
   }
   if( bins(0) > numberof(rate_start) ) error,"bins exceed range ("+itoa(numberof(rate_start))+")";
   Bins = bins;
   E_min = e_min_start(bins);
   E_max = e_max_start(bins);
   Matrix = matrix_start(bins,);
   Rate = rate_start(bins);
   Stat_err = stat_err_start(bins);
   write,format="New E_min/max: %.3f %.3f keV with %i bins\n", E_min(1), E_max(0), numberof(Bins);
}

/* Function xfit_ebase */

func xfit_ebase( ebase )
/* DOCUMENT xfit_ebase, new_ebase
     or     xfit_ebase

    Second version will reset to 1.0
*/
{
// ynclude = xfit_package.externs_ebase
  extern  Ebase;
// yxclude =

   Ebase = is_void(ebase) ? 1.0 : double(ebase);
}

/* Function xfit_title */

func xfit_title( plottitle )
/* DOCUMENT xfit_title, plottitle
      or    xfit_title

     Second version sets the title to an empty string
*/
{
// ynclude = xfit_package.externs_title
  extern  Plottitle;
// yxclude =

   Plottitle = is_void(plottitle) ? "" : plottitle;
}

/* Function xfit_freeze */

func xfit_freeze( parnum, val )
/* DOCUMENT  xfit_freeze, parnum, val

    'val' must be the parameter value that goes into the p array
    For Nh it should be Nh/1e22
*/
{

// ynclude = xfit_package.externs_freeze
  extern  Freeze,        Frozen_value;
// yxclude =

   Freeze(parnum) = 1;
   Frozen_value(parnum) = val;
}

/* Function xfit_thaw */

func xfit_thaw( parnum )
/* DOCUMENT  xfit_thaw, parnum
*/
{

// ynclude = xfit_package.externs_thaw
  extern  Freeze;
// yxclude =

   Freeze(parnum) = 0;
}

/* Function xfit_tune */

func xfit_tune
/* DOCUMENT xfit_tune

    Adjust the errors of observation to give a reduced Chi2 of 1.0
    for the current selection of bins.
*/
{
// ynclude = xfit_package.externs_tune
  extern  Chi2min,    Chi2red,    N_freedom,  Stat_err;
// yxclude =

   Stat_err *= sqrt(Chi2red);
   Chi2min = double(N_freedom);
}

/* Function xfit_pargrid */

func xfit_pargrid( dim1, dim2, range1, range2, win=  )
/* DOCUMENT xfit_pargrid, dim1, dim2[, range1, range2], win=

    Produces a 2D plot in window #2 or as defined by keyword 'win'

    'range1' and 'range2' are scalars that give the range:
     i.e. from bestfit value - range/2 to bestfit value + range/2

*/
{
// ynclude = xfit_package.externs_pargrid
  extern  Chi2min,       Delta_const,   Nh_bf,         Pargrid_arr,   Plottitle, \
          Const_bf,      Delta_param1,  Param1_bf;
// yxclude =
 
   if( is_void(win) ) win = 2;

   Pargrid_arr = array(double,dim1, dim2);
   p = [Const_bf, Param1_bf, Nh_bf*1.e-22];

   if( is_void(range1) ) range1 = Delta_const;
   if( is_void(range2) ) range2 = Delta_param1;
   vals1 = span(  Const_bf - 0.5*range1,  Const_bf + 0.5*range1, dim1 );
   vals2 = span(  Param1_bf - 0.5*range2,  Param1_bf + 0.5*range2, dim2 );
   for( i = 1; i <= dim1; i++ ) {
      for( j = 1; j <= dim2; j++ ) {
         p(1) = vals1(i);
         p(2) = vals2(j);
         Pargrid_arr(i,j) = xfit_chi2( p ) - Chi2min;
      }
   }
   window,win;
   dispc,Pargrid_arr,xax=vals1,yax=vals2,levels=[2.28,4.6,9.21],over=1, \
        title=esc_underscore(Plottitle),xtitle="Const",ytitle="Param1";
   oplot,[Const_bf],[Param1_bf],ps=12,thick=5,color="white",symsize=1.2;
}

/* Function xfit_chi2_1d_1 */

func xfit_chi2_1d_1( x )
{
// ynclude = xfit_package.externs_chi2_1d_1
  extern  Chi2min,  Level,    Pcur;
// yxclude =

   // x is p(1)

   Pcur(1) = x;
   return xfit_chi2( Pcur ) - Chi2min - Level;

}

/* Function xfit_chi2_1d_2 */

func xfit_chi2_1d_2( x )
{
// ynclude = xfit_package.externs_chi2_1d_2
  extern  Chi2min,  Level,    Pcur;
// yxclude =

   // x is p(2)

   Pcur(2) = x;
   return xfit_chi2( Pcur ) - Chi2min - Level;

}

/* Function xfit_chi2_1d_angle */

func xfit_chi2_1d_angle( xi )
{
// ynclude = xfit_package.externs_chi2_1d_angle
  extern  Angle,         Const_bf,      Delta_param1,  Param1_bf,     Pcur, \
          Chi2min,       Delta_const,   Level;
// yxclude =

   // x is p(2)

   Pcur(1) = Const_bf + Delta_const * xi * cos(Angle);
   Pcur(2) = Param1_bf + Delta_param1 * xi * sin(Angle);

   return xfit_chi2( Pcur ) - Chi2min - Level;

}

/* Function xfit_anamin */

func xfit_anamin( astep= )
/* DOCUMENT xfit_anamin[, astep=]

     Scans the const vs. param1 space for 99% confidence interval.

     Keyword 'astep' is angular step (in radians, defaults to 0.1*pi).
*/
{
// ynclude = xfit_package.externs_anamin
  extern  Angle,         Delta_const,   Logfilename,   Nh_bf,         Pcur, \
          Chi2red,       Delta_param1,  Logflag,       Param1_bf,     Specfile, \
          Const_bf,      Level,         Nh;
// yxclude =

   if( is_void(astep) ) astep = 0.1*pi;

  /*
   *  Locate the 99% confidence limits for Pcur(1) == Const
   *  corresponding to chi2 = Chi2min + 9.21
   */

   Pcur = [Const_bf, Param1_bf, Nh_bf*1.e-22];
   Level = 9.21;
   //  Search to the right
   tol = 1.e-5;
   ax = bx = Const_bf;
   do {
      bx *= 1.01;
      fb = xfit_chi2_1d_1(bx);
   } while ( fb <= 0. );
   const_max = zbrent( xfit_chi2_1d_1, ax, bx, tol );

   //  Search to the left
   ax = bx = Const_bf;
   do {
      bx *= 0.99;
      fb = xfit_chi2_1d_1(bx);
   } while ( fb <= 0. );
   const_min = zbrent( xfit_chi2_1d_1, ax, bx, tol );

   Delta_const = const_max - const_min;
   write,format="99%% confid. interval for Const: %.4f - %.4f (range %.4f)\n", \
                                      const_min, const_max, Delta_const;
   write,format="     for Param1 = %.4f and Nh/1e22 = %.4f\n", Param1_bf, Nh_bf*1.e-22;

  /*
   *  Locate the 99% confidence limits for Pcur(2) == Param1
   *  corresponding to chi2 = Chi2min + 9.21
   */

   Pcur = [Const_bf, Param1_bf, Nh_bf*1.e-22];
   //  Search to the right
   tol = 1.e-5;
   ax = bx = Param1_bf;
   do {
      bx *= 1.01;
      fb = xfit_chi2_1d_2(bx);
   } while ( fb <= 0. );
   param1_max = zbrent( xfit_chi2_1d_2, ax, bx, tol );

   //  Search to the left
   ax = bx = Param1_bf;
   do {
      bx *= 0.99;
      fb = xfit_chi2_1d_2(bx);
   } while ( fb <= 0. );
   param1_min = zbrent( xfit_chi2_1d_2, ax, bx, tol );

   Delta_param1 = param1_max - param1_min;
   write,format="99%% confid. interval for Param1: %.4f - %.4f (range %.4f)\n", \
                                      param1_min, param1_max, Delta_param1;
   write,format="     for Const = %.4f and Nh/1e22 = %.4f\n", Const_bf, Nh_bf*1.e-22;

   // Search various angles

   for( Angle = pi*0.2; Angle < 2*pi; Angle += astep ) {
      xi = 0.;
      do {
         xi += 0.1;
         fb = xfit_chi2_1d_angle( xi );
      } while( fb <= 0.0 );
      ximin = zbrent( xfit_chi2_1d_angle, 0., xi, tol );
      c = Const_bf + Delta_const * ximin * cos(Angle);
      if( c > const_max ) const_max = c;
      if( c < const_min ) const_min = c;
      c = Param1_bf + Delta_param1 * ximin * sin(Angle);
      if( c > param1_max ) param1_max = c;
      if( c < param1_min ) param1_min = c;
   }
   Delta_const = const_max-const_min;
   Delta_param1 = param1_max-param1_min;
   write,format="99%% confid. interval for Const: %.4f - %.4f (range %.4f)\n", \
                                      const_min, const_max, const_max-const_min;
   write,format="99%% confid. interval for Param1: %.4f - %.4f (range %.4f)\n", \
                                      param1_min, param1_max, param1_max-param1_min;

   if( Logflag ) {
      flog = open(Logfilename,"a");
      write,flog,format="// called xfit_anamin ...%s","\n";
      s = structof(Specfile) == string ? Specfile : "from \"xfit_fakeit\"";
      write,flog,format=" %12.3e %12.3e %.4f %.4f %.4f %.4f %.4f %s\n", Const_bf, 0.5*Delta_const, \
               Param1_bf, 0.5*Delta_param1, Nh*1.e-22, 0., Chi2red, s;
      close, flog;
   }
}

/* Function xfit_read */

func xfit_read( rmffile, arffile )
/* DOCUMENT xfit_read, rmffile, arffile

   Load the Matrix, Elo, Ehi, Eline, E_min, E_max for the
   relevant instrument.

   2011-09-07/NJW cloned from xfit_nustar_init.i

*/
{

// ynclude = xfit_package.externs_read
  extern  Arf,     E_max,   E_min,   Ehi,     Eline,   Elo,     Matrix, \
          E_min_orig, E_max_orig, Matrix_orig;
// yxclude =

   local elo, ehi, emin, emax;

   //   Get the RMF data

   Matrix_orig = Matrix = read_ogip_rmf( rmffile, elo, ehi, emin, emax );

   E_min_orig = E_min = emin;
   E_max_orig = E_max = emax;

   Elo = elo;
   Ehi = ehi;
   Eline = sqrt(Elo*Ehi);

   //    Get the ARF

   Arf = rdfitscol(arffile+"+1","SPECRESP");
   elo_Arf = rdfitscol(arffile+"+1","ENERG_LO");
   //+ ehi_Arf = rdfitscol(arffile+"+1","ENERG_HI");
   //  --- NB temporary fix for LOFT WFM data
   ehi_Arf = rdfitscol(arffile+"+1","ENERGY_HI");
   e_Arf = 0.5*(elo_Arf + ehi_Arf);
   Arf = interp( Arf, e_Arf, Eline);

   write,"Elo, Ehi, Matrix, Arf, E_min, and E_max have been loaded ...";
}

/* Function xfit_fakeit */

func xfit_fakeit( bkgdol, fbkg=, openfrac=, sc=, cst=, p1=, nh=, expo=, ia= )
/* DOCUMENT xfit_fakeit, bkgdol, fbkg=, openfrac=, sc=, cst=, p1=, nh=, expo=, ia=

    Simulate an observation using the ARF and RMF previously loaded by
    xfit_read.

    Giving the DOL of a background will include it in the simulation and fitting.
    The background table is assumed to be number of counts per detector channel
    in the time given by the EXPOSURE keyword on the entire detector.

    The keyword 'openfrac' will attenuate the background actually used with this
    factor (applies for a coded mask instrument).

    It is assumed that the open fraction for the source is covered by the Arf.

    The keyword 'fbkg' will distinguish between the two cases:
       1) The background is determined in the same observation as the source
          so it is determined with a statistical error with a possibility
          to 'overobserve' with a factor, namely fbkg. (fbkg > 0, usually 1,
          but can be 3 as for a 0.25 coded mask).
       2) The background is determined in a long observation so that
          a perfect subtraction can be done (fbkg is not set)
       This is inspired by the discussions around 2011-09-04 with Soren Brandt and
       Jerome Chenevez.

    keywords:
        sc:    Spectral code, "pl", "tb", "bb", or "df"
       cst:    Normalization
        p1:    Spectral parameter
        nh:    Column density in units of 1.e22 cm-2
      expo:    Exposure time in seconds
        ia:    Sets interactive mode

    2011-09-07/NJW cloned from xfit_nustar_fakeit
*/
{

   //   Define the externals

// ynclude = xfit_package.externs_fakeit
  extern  Arf,          Const,        Eline,        Nh,           Spectrcode, \
          Bkgexposure,  Countspec,    Elo,          Param1,       Stat_err, \
          Bkgspec,      Bkgrate, Ebase,        Matrix,       Rate,         Use_bkg, \
          Bkgspec_err,  Ehi,  Photonspec, Emin_std, Emax_std, N_freedom,  \
          Parfilename, Exposure, Fbkg, Logflag, Logfilename, Rate_orig, Stat_err_orig, \
          Matrix_orig;
// yxclude =

   if( is_void(openfrac) ) openfrac = 1.0;
   if( is_void(Ebase) ) Ebase = 1.; // [keV] pivot point for spectra
   if( is_void(bkgdol) ) {
      Use_bkg = 0;
   } else {
      local bkgfile, bkgext;
      get_exten_no, bkgdol, bkgfile, bkgext;

      if( !file_test(bkgfile) ) error,"Bkg file was not found";
      hdr = headfits( bkgdol );
      // reduce the background counts with the open fraction
      Bkgspec = openfrac*rdfitscol( bkgdol, "COUNTS" );
      Bkgspec_err = sqrt(Bkgspec);
      Bkgexposure = fxpar( hdr, "exposure" );
      Use_bkg = 1;
   }

   p = array(double,3);

   if( ia ) {
      Spectrcode = "";
      read,prompt="Enter spectrcode : ... ",format="%s", Spectrcode;
      Spectrcode = strupcase(Spectrcode);
      Const = 1.0; Param1 = 1.0; nhcol = 1.0;
      read,prompt="Enter constant : ... ",format="%f", Const;
      read,prompt="Enter param1   : ... ",format="%f", Param1;
      read,prompt="Enter NH/1.e22 : ... ",format="%f", nhcol;
      Exposure = 1.0;
      read,prompt="Enter exposure : ... ",format="%f", Exposure;
   } else {
      Const = numberof(cst) ? double(cst) : get_par( Parfilename, "const" );
      Spectrcode =  numberof(sc) ? strupcase(sc) : get_par( Parfilename, "spectrcode" );
      Param1 = numberof(p1) ? double(p1) : get_par( Parfilename, "param1" );
      nhcol = numberof(nh) ? double(nh) : get_par( Parfilename, "nh" ); // in units of 1.e22
      Exposure = numberof(expo) ? double(expo) : get_par( Parfilename, "exposure" );
   }
   set_par, Parfilename, "spectrcode", Spectrcode;
   set_par, Parfilename, "const", Const;
   set_par, Parfilename, "param1", Param1;
   set_par, Parfilename, "nh", nhcol;
   set_par, Parfilename, "exposure", Exposure;

   p(1) = Const;
   p(2) = Param1;
   p(3) = nhcol;
   Nh = p(3) * 1.e22;

   //   Generate the input photon spectrum

   Photonspec = xfit_photspec(p);
   write,format="Sflux %.0f-%.0f keV = %12.3e erg/cm2/s\n", \
              Emin_std, Emax_std, sflux(Emin_std,Emax_std,Eline, Photonspec);

   if( Logflag ) {
      fh = open( Logfilename, "a" );
      write,fh,format="// Call xfit_fakeit with ...%s","\n";
      write,fh,format="// Spectrcode = %s\n", Spectrcode;
      write,fh,format="// Const = %10.4f\n", Const;
      write,fh,format="// Param1 = %10.4f\n", Param1;
      write,fh,format="// Nh = %13.2e\n", nhcol*1.e22;
      write,fh,format="// Exposure = %13.2e\n", Exposure;
      write,fh,format="// Sflux %.0f-%.0f keV = %12.3e erg/cm2/s\n", \
              Emin_std, Emax_std, sflux(Emin_std,Emax_std,Eline, Photonspec);
      if( Use_bkg ) {
         write,fh,format="// dol_bkg = %s\n", bkgdol;
         if( fbkg ) {
            write,fh,format="// fbkg = %10.2f\n", fbkg;
         } else write, fh, format="// inf bkg determination%s","\n";
         write,fh,format="// openfraction = %10.5f\n", openfrac;
      } else write,fh,format="// no bkg used%s","\n";
      close, fh;
   }


   line_counts = Photonspec * (Ehi - Elo) * Arf;
   Countspec = (Matrix_orig(,+)*line_counts(+))*Exposure; // expected count spectrum
   //+ Countspec += random_n(numberof(Countspec))*countspec_err;

   // Note the difference between 'Countspec' that is the observed (or
   // simulated) number of counts, whereas 'Rate_model' is the result of
   // the current spectral parameter values.

   //  Add counting statistics for realistic simulation
   Countspec = poisson(Countspec);

   // Add background if supplied

   if( Use_bkg ) {
      Bkgspec *= (Exposure/Bkgexposure); // rescale to observation
      Countspec += poisson(Bkgspec);
      Fbkg = fbkg;
      if( fbkg ) Bkgspec = poisson(fbkg*Bkgspec)/fbkg;
      Bkgrate = Bkgspec / Exposure;
   }
   countspec_err = sqrt(Countspec); // usual expression
   Rate = Countspec / Exposure;
   if( Use_bkg ) Rate -= Bkgrate;
   Stat_err = countspec_err / Exposure;
   Rate_orig = Rate;
   Stat_err_orig = Stat_err;

   N_freedom = numberof(Rate) - 3; // to update the external for e.g. xfit_plot

   write,"Rate(_orig) and Stat_err(_orig) are now prepared for e.g. fitting and plotting.";
   if( Use_bkg ) write,"Bkgspec and Bkgrate have been loaded as well.";

}
