extern xraydoc;
/* DOCUMENT **********************************
    A collection of useful functions in connection
    with X-ray astronomy

     2007-06-14/NJW
     2010-12-01/NJW updated with spec_info and spec2phaii from jemx.i

        absorp          Interstellar absorption coefficient
        add_peak_1d     Add gaussian peak to spectrum
        add_phaii_spectra   Add the PHAII format spectra from given X-ray source
        add_p2spectra   Alias for add_phaii_spectra
        avg_spectra     Make weighted average of spectra
        comb_spectra    Make sum or average of spectra
        cross_sect      Interstellar absorption cross section
        dsexit          Emulation of script 'dsexit' for ds9 control
        dsinit          Emulation of script 'dsinit' for ds9 control
        dsplay          Emulation of scripts 'dsplay'  and 'dsplay0' for ds9 control
        dsreg           Emulation of script 'dsreg' for ds9 control
        get_xspec_fit_params  Retrieve the result of an XSPEC fit from the log file
        mk_photflux     Produce FITS file with photon flux (ph/cm2/s/keV)
        oplot_effa      Overplot eff. area curve
        phatype         Return type of spectrum (PHAI or PHAII)
        plot_effa       Plot effective area
        plot_2effas     Plot two effective area curves and their ratio
        predict_cr      Predict countrate given an ARF and Eline + Flux
        rd_arf          Read an ARF, PHAI or PHAII
        rd_photflux     Read the output file from 'mk_photflux' to create Eline + Flux
        read_effa       Read energy and eff. area columns
        rd_phai_spectrum  Read a PHAI type spectrum
        read_spectrum   Read a spectrum regardless of type
        sflux           Photon flux integration
        spec_info       Report from PHAII spectral file
        spec2phai       Write spectra to PHAI format file
        spec2phaii      Write spectra to PHAII format file
        specbinning     Binning an energy array into a spectrum
        specrebinning   Rebinning of a spectrum
        specrebinninga  Rebinning of a spectrum to a given fraction drate/rate
        specrebinningb  Rebinning of a spectrum to a given number of counts
        specunpack      Copy struct 's_Spectrum' elements to arrays
        sub_spectra     Subtract one spectrum from another one
        write_effa      Write eff. area table to FITS file

******************************************************/

struct s_Effarea{
        pointer ener; // energy array
        pointer area; // effective area array
        pointer hdr;  // FITS header card list
}

struct s_Spectrum {
   long n; // number of channels
   long specnum; // Number of spectrum (only relevant for PHAII format)
   double exposure; // [s] Exposure time
   double backscal; // Relative region area
   pointer rate; // RATE i.e. countrate per channel
   pointer stat_err; // STAT_ERR i.e. 1 sigma error of RATE
   pointer e_min; // E_MIN from EBOUNDS
   pointer e_max; // E_MAX from EBOUNDS
   pointer energ_lo; // ENERG_LO for ARF
   pointer energ_hi; // ENERG_HI for ARF
   string backfile; // Background file BACKFILE
   string rmf; // RMF file
   string arf; // ARF (Anchor) file
   string specdol; // DOL for spectrum
}

// Three following structs are for 'get_xspec_fit_params'

struct s_Parval {
   string model;
   int param_num;
   int model_num;
   string component;
   string name;
   string unit;
   double value;
   double value_err;
   string status;
}
struct s_Flux {
   double elims(2);
   double flux;
}
struct s_Fitqual {
   int cstat;
   int dof;
   double red_chi2;
}


extern presetdoc;
/* DOCUMENT
     Extit    "Energy [keV]"
     Fytit    "Flux [ph cm^-2^s^-1^keV^-1^]"
*/
Extit = "Energy [keV]";
Fytit = "Flux [ph cm^-2^s^-1^keV^-1^]";

/* Function absorp */

func absorp( nh, earr )
/* DOCUMENT res = absorp( nh, earr )

  Returns the interstellar cold absorption
  given the column density value, nh.

     NJW, some time in the past ~1998
*/
{

   if( is_void(earr) ) {
      write,"Syntax: res = absorp( nh, earr)";
      return [];
   }

   n = numberof(earr);
   res = array(double,n);
   for( i = 1; i <= n; i++) res(i) = exp(-nh*cross_sect(earr(i)));
   return res;
}

/* Function cross_sect */

func cross_sect( energy )
/* DOCUMENT res = cross_sect( energy )

 Morrison && McCammon interstellar matter cross sections
 ApJ 270 (1983) 119

 The energy is in keV && returned cross section in cm-2

  ~1998/NJW
*/
{
   ebound=[ 0.100,0.284,0.400,0.532,0.707, \
          0.867,1.303,1.840,2.471,3.210,4.038, \
          7.111,8.331,10.00];
   coeff = array(double,3,14);
   coeff(,1) = [ 17.3,  608.1, -2150.0 ];
   coeff(,2) = [ 34.6,  267.9,  -476.1 ];
   coeff(,3) = [ 78.1,   18.8,     4.3 ];
   coeff(,4) = [ 71.4,   66.8,   -51.4 ];
   coeff(,5) = [   95.5,  145.8,   -61.1 ];
   coeff(,6) = [ 308.9, -380.6,   294.0 ];
   coeff(,7) = [  120.6,  169.3,   -47.7 ];
   coeff(,8) = [ 141.3,  146.8,   -31.5 ];
   coeff(,9) = [  202.7,  104.7,   -17.0 ];
   coeff(,10) = [ 342.7,   18.7,     0.0 ];
   coeff(,11) = [ 352.2,   18.7,     0.0 ];
   coeff(,12) = [ 433.9,   -2.4,    0.75 ];
   coeff(,13) = [  629.0,   30.9,     0.0 ];
   coeff(,14) = [ 701.2,   25.2,     0.0];
   i = 1;
   while( energy  >=  ebound(i) && i  <  14 ) i++;
   csect = coeff(1,i) + energy * (coeff(2,i) + energy * coeff(3,i));
   return 1.e-24 * csect / (energy^3);
}

/* Function sflux */

func sflux( e1, e2, earr, spectrum, kev= )
/* DOCUMENT sx = sflux( e1, e2, earr, spectrum, kev= )

  Calculate source strength (S) in erg cm-2 s-1
  from e1 to e2 of "spectrum" where "earr" is the
  energy scale

  Keyword  kev : When set the conversion to erg is skipped

*/
{
   if( is_void(spectrum) ) {
      write,"SFLUX syntax: sx = sflux( e1, e2, energy_array, spectrum )";
      return [];
   }
   n = numberof(earr);
   if( n  !=  numberof(spectrum) ) {
     write,"Arrays do not match";
     return [];
   }
   delta_e = earr - shift(earr, -1);
   delta_e(1) = delta_e(2);
   index = where( (e1  <  earr) & (e2  >=  earr) );

   res_kev = sum(delta_e(index)*earr(index)*spectrum(index));
   return kev ? res_kev : res_kev * 1.6e-9;

}

/* Function specrebinning */

func specrebinning( eb1, eb2, rate, rate_err, rebin, &ob1, &ob2, &orate, &orate_err )
/* DOCUMENT specrebinning, eb1, eb2, rate, rate_err, rebin, >ob1, >ob2, >orate, >orate_err

   Combine an integer number of input spectral bins to an output spectral bin
   by addition (errors are added quadratically) i.e. 'rate' is NOT per keV but
   per bin.

   eb1, eb2 are the input bin boundaries, rate and rate_err the spectrum.
   rebin is an array (integer) where each element is the number of input
   bins to combine into the next output bin.

   Negative values in 'rebin(i)' means skipping -rebin(i) values in the original spectrum.

   The resulting bin boundaries are ob1 and ob2 and the resulting spectrum
   is given by orate and orate_err.

   SEE ALSO: specrebinninga, specrebinningb

   2008-01-01(?)/NJW
*/
{
   nreb = numberof(rebin);
   nrate = numberof(rate);
   if( sum(abs(rebin)) < nrate ) {
      write,"Insufficient rebinning information";
      return;
   }

   ob1 = [];
   ob2 = [];
   orate = [];
   orate_err = [];

   k = 1;
   for( i = 1; i <= nreb; i++ ) {
      if( k > nrate ) break;
      if( rebin(i) < 0 ) {
         k += -rebin(i);
      } else {
         rsum = 0.0;
         errsum = 0.0;
         grow, ob1, eb1(k);
         for( j = 1; j <= rebin(i); j++ ) {
            if( k > nrate ) break;
            rsum += rate(k);
            errsum += rate_err(k)^2;
            k++;
         }
         grow, orate, rsum;
         grow, orate_err, sqrt(errsum);
         grow, ob2, eb2(k-1);
      }
   }
}


/* Function specrebinninga */

func specrebinninga( eb1, eb2, rate, rate_err, frac, &ob1, &ob2, &orate, &orate_err )
/* DOCUMENT specrebinninga, eb1, eb2, rate, rate_err, frac, >ob1, >ob2, >orate, >orate_err
            rebinarr = specrebinninga( eb1, eb2, rate, rate_err, frac, >ob1, >ob2, >orate, >orate_err )

   Combine an integer number of input spectral bins to an output spectral bin
   in such a way that orate_err/orate < frac if at all possible.
   Bin values are added (errors are added quadratically) i.e. 'rate' is NOT per keV but
   per bin.

   eb1, eb2 are the input bin boundaries, rate and rate_err the spectrum.

   The resulting bin boundaries are ob1 and ob2 and the resulting spectrum
   is given by orate and orate_err.

   The function returns rebinarr which is an array (integer) where each element
   is the number of input bins to combine into the next output bin such as
   required for 'specrebinning'.

   2010-11-23/NJW
*/
{
   ob1 = [];
   ob2 = [];
   orate = [];
   orate_err = [];
   ob1 = eb1;
   ob2 = eb2;

   n_in = numberof(rate);
   rebinarr = array(1,n_in);
   k_in = 1;
   k_out = 1;

   while( k_in <= n_in ) {
      ob1(k_out) = eb1(k_in);
      rate_out = rate(k_in);
      rate_out_err2 = rate_err(k_in)^2;
      rate_out_err = sqrt(rate_out_err2);
      test_frac = rate_out <= 0.0 ? frac + 1. : rate_out_err/rate_out;
      while( test_frac > frac ) {
         rebinarr(k_out)++;
         k_in++;
         if( k_in > n_in ) {
            k_in = n_in;
            break;
         }
         rate_out += rate(k_in);
         rate_out_err2 += rate_err(k_in)^2;
         rate_out_err = sqrt(rate_out_err2);
         test_frac = rate_out <= 0.0 ? frac + 1. : rate_out_err/rate_out;
      }
      ob2(k_out) = eb2(k_in);
      grow, orate, rate_out;
      grow, orate_err, rate_out_err;
      k_out++;
      k_in++;
   }
   // see if last value passes the test
   test_frac = orate(0) <= 0.0 ? frac + 1. : orate_err(0)/orate(0);
   n_out = numberof(orate);
   if( n_out > 1 && test_frac > frac ) {
      n_out--;
      orate = orate(1:-1);
      orate_err = orate_err(1:-1);
   }
   ob1 = ob1(1:n_out);
   ob2 = ob2(1:n_out);

   rebinarr = rebinarr(1:n_out);
   if( rebinarr(sum) < n_in ) grow, rebinarr, rebinarr(sum) - n_in - 1;
                  // A negative value will skip the rest

   return rebinarr;
}


/* Function specrebinningb */

func specrebinningb( eb1, eb2, rate, rate_err, exposure, mincts, &ob1, &ob2, &orate, &orate_err )
/* DOCUMENT specrebinningb, eb1, eb2, rate, rate_err, exposure, mincts, >ob1, >ob2, >orate, >orate_err
            rebinarr = specrebinningb( eb1, eb2, rate, rate_err, exposure, mincts, >ob1, >ob2, >orate, >orate_err )

   Combine an integer number of input spectral bins to an output spectral bin
   in such a way that the number of counts equal or exceed 'mincts' if at all possible.
   Bin values are added (errors are added quadratically) i.e. 'rate' is NOT per keV but
   per bin.

   eb1, eb2 are the input bin boundaries, rate and rate_err the spectrum.

   The resulting bin boundaries are ob1 and ob2 and the resulting spectrum
   is given by orate and orate_err.

   The function returns rebinarr which is an array (integer) where each element
   is the number of input bins to combine into the next output bin such as
   required for 'specrebinning'.

   SEE ALSO:  specrebinning, specrebinninga

   2012-03-29/NJW, based on specrebinninga
*/
{
   ob1 = [];
   ob2 = [];
   orate = [];
   orate_err = [];
   ob1 = eb1;
   ob2 = eb2;

   n_in = numberof(rate);
   rebinarr = array(1,n_in);
   k_in = 1;
   k_out = 1;

   while( k_in <= n_in ) {
      ob1(k_out) = eb1(k_in);
      rate_out = rate(k_in);

      rate_out_err2 = rate_err(k_in)^2;

      while( rate_out*exposure < mincts ) {
         rebinarr(k_out)++;
         k_in++;
         if( k_in > n_in ) {
            k_in = n_in;
            break;
         }
         rate_out += rate(k_in);
         rate_out_err2 += rate_err(k_in)^2;
      }
      ob2(k_out) = eb2(k_in);
      grow, orate, rate_out;
      grow, orate_err, sqrt(rate_out_err2);
      k_out++;
      k_in++;
   }
   // see if last value passes the test
   n_out = numberof(orate);
   if( n_out > 1 && orate(n_out)*exposure < mincts ) {
      n_out--;
      orate = orate(1:-1);
      orate_err = orate_err(1:-1);
   }
   ob1 = ob1(1:n_out);
   ob2 = ob2(1:n_out);

   rebinarr = rebinarr(1:n_out);
   if( rebinarr(sum) < n_in ) grow, rebinarr, rebinarr(sum) - n_in - 1;
                  // A negative value will skip the rest

   return rebinarr;
}


/* Function specbinning */

func specbinning( e_evts, eb1, eb2, &rate, &rate_err, exposure=, perkev=, silent= )
/* DOCUMENT specbinning, e_evts, eb1, eb2, >rate, >rate_err, exposure=, perkev=, silent=

   Produce an energy spectrum from an array of event energies

   Calls:
      e_evts   :  array of energies
      eb1 and eb2: Energy bin boundaries (they need not be contiguous but no overlap
                         is accepted).
   Returns:
      rate     : number of counts per bin
      rate_err : statistical error (Poisson)
   Keywords:
      exposure : Time interval; if given then rate and rate_err will be divided
                     with this value
      perkev   : If set, then rate and rate_err will be per energy unit (i.s.o. per bin)
      silent   : To suppress screen messages

   2009-03-06/NJW
*/
{

   nevs = numberof( e_evts );
   nbins = numberof(eb1);
   if( nbins != numberof(eb2) ) error,"SPECBINNING Error ##1## mismatching numbers of boundaries";
   d = eb2(1:nbins-1) - eb1(2:nbins);
   if( anyof( d < 0.0 ) ) error,"SPECBINNING Error ##2## overlapping energy boundaries";
   wb = eb2 - eb1;
   if( anyof(wb <= 0.0) ) error,"SPECBINNING Error ##3## zero size energy bins";

// populate the energy bins

   rate = rate_err = array(double, nbins);
   ntot = 0;
   for( i = 1; i <= nbins; i++ ) {
      w = where( e_evts >= eb1(i) & e_evts < eb2(i) );
      nw = numberof(w);
      rate(i) = double(nw);
      ntot += nw;
   }
   rate_err = sqrt(rate);
   if( !is_void(exposure) ) {
      if(!silent)write,"SPECBINNING exposure corrected";
      rate /= exposure;
      rate_err /= exposure;
   }
   if( perkev ) {
      if(!silent)write,"SPECBINNING spectrum per energy unit";
      rate /= wb;
      rate_err /= wb;
   }

}

/* Function add_peak_1d */

func add_peak_1d( ener, flux, eline, sigma, ampl, rel= )
/* DOCUMENT new_flux = add_peak_1d( ener, flux, eline, sigma, ampl, rel= )

   Add a Gaussian peak to a 1D array - return the new flux.

   2009-03-06/NJW cloned from IDL version 'add_peak_1d.pro'
*/
{

   if( rel ) {
      fluxline = flux * ampl*exp(-0.5*(((ener-eline)/sigma)^2));
   } else {
      fluxline = ampl*exp(-0.5*(((ener-eline)/sigma)^2));
   }
   orig_flux_at_line = interp( flux, ener, eline );
   equi_width = sum(fluxline(zcen)*ener(dif)) / orig_flux_at_line;
   write,format="Equivalent width = %10.3e keV\n", equi_width;
   return flux + fluxline;

}

/* Function mk_photflux */

func mk_photflux(xspecmodel=,xspecparms=,sc=,nh=,norm=,p1=,e1=,e2=,nchan=,nof=,outfile=,mem=,silent=)
/* DOCUMENT mk_photflux,xspecmodel=, xspecparms=, sc=,nh=,norm=,p1=,e1=,e2=,nchan=,nof=,outfile=,mem=,silent=

   Takes no argument, just keywords. It will ask for the
   missing parameters ('nh' in /cm2).
   A mk_photflux.par file is used as (suggestive) input.
   If it does not exist it is created in the current directory.

   External variables defined: Eline, Flux, and Sx_photflux

   There are two main modes of operation:
       1) The quick and simple that uses keywords: sc, nh, norm, and p1
       2) By invoking 'xspec' with two keywords: xspecmodel (string)
                                  and            xspecparms (array of double)
          The presence of these two keywords will override any of the
          sc, nh, norm, and p1 keywords.

   Keyword 'nof' will prevent file writing
   Keyword 'mem' will save the result to memory using 'mem_save'
                 from package 'mem_storage.i' with 'outfile' plus
                 column name as identifier.
                 If 'mem' is 1 then the ENERG_LO and ENERG_HI will be saved
                 with these identifiers else the 'outfile' will be used for
                 the 'Flux' array.

   2010-04-16/NJW Updated, 2013-01-24/NJW updated with memory option
   2013-05-23/NJW Updated with XSPEC functionality

*/
{
   extern Elo, Ehi, Eline, Spectrcode;
   extern Const, Param1, Nh;
   extern Flux, Sx_photflux;
   extern Ebase;

   // Communication with 'mt_rayor'
   extern Energy_start, Energy_stop, NumChans;

   in_xspecmode = ( structof(xspecmodel) == string && numberof(xspecparms) );

   if( !file_test("mk_photflux.par") ) {
      fh = open("mk_photflux.par","w");
      write,fh,format="// Spectrcode = %s\n","PL";
      write,fh,format="// Const = %f\n",10.;
      write,fh,format="// Param1 = %f\n",1.5;
      write,fh,format="// Nh = %e\n",1.5e22;
      write,fh,format="// Energy_start = %f\n",5.;
      write,fh,format="// Energy_stop = %f\n",80.;
      write,fh,format="// NumChans = %i\n",80;
      close, fh;
   }

   if( !in_xspecmode ) {

      if( is_void(sc) ) {
         Spectrcode = strupcase(get_spars("mk_photflux.par","Spectrcode"));
      } else {
         if( typeof(sc) != "string" ) error,"sc must be a string";
         sc = strupcase(sc);
         if( noneof(sc == ["PL","BB","TB","DF"]) ) error,"sc must be PL, BB, TB, or DF";
         Spectrcode = sc;
      }
      if( is_void(norm) ) {
         Const = get_spar("mk_photflux.par","Const");
      } else Const = double(norm);
      if( is_void(p1) ) {
         Param1 = get_spar("mk_photflux.par","Param1");
      } else Param1 = double(p1);
      if( is_void(nh) ) {
         Nh = get_spar("mk_photflux.par","Nh");
      } else Nh = double(nh);

   }

   //  following keywords are common for the two modes

   if( is_void(e1) ) {
      Energy_start = get_spar("mk_photflux.par","Energy_start");
   } else Energy_start =  double(e1);
   if( is_void(e2) ) {
      Energy_stop = get_spar("mk_photflux.par","Energy_stop");
   } else Energy_stop =  double(e2);
   if( is_void(nchan) ) {
      NumChans = get_spar("mk_photflux.par","NumChans",lng=1);
   } else NumChans = long(nchan);
   
   b = spanl(Energy_start, Energy_stop, NumChans+1);
   Elo = energ_lo = b(1:-1);
   Ehi = energ_hi = b(2:0);
   Eline = sqrt(energ_lo*energ_hi);

   if( in_xspecmode ) {

      // Create the script file
      rstr = "r_"+ranstr(10);
      qstr = "q_"+ranstr(10);
      script = open(rstr+".xcm","w");
      write,script,format="dummyrsp %.3f %.3f %i\n", Energy_start, Energy_stop, NumChans;
      write,script,format="%s\n","setp com wdata "+rstr;
      s = "model "+xspecmodel;
      n = numberof(xspecparms);
      for(i = 1; i <= n; i++ ) s += swrite(format=" & %g", xspecparms(i));
      write,script,format="%s\n",s;
      write,script,format="%s\n","pl model";
      write,script,format="%s\n","syscall mv "+rstr+".qdp "+qstr+".qdp";
      write,script,format="%s\n","exit";
      close, script;
      system,"xspec - "+rstr;
      local qener, dqener;
      rd_qdp,qstr,qener, dqener, Flux; 
      // Clean up the temporary files
      remove, rstr+".xcm";
      remove, qstr+".qdp";
   } else {
      Flux = xfit_photspec( [Const, Param1, Nh/1.e22] );
   }

   Sx_photflux = sflux(Energy_start,Energy_stop, Eline, Flux);

   if( mem ) { // save result to memory
      if( mem == 1 ) { // save the energy boundaries
         mem_save,"ENERG_LO", energ_lo;
         mem_save,"ENERG_HI", energ_hi;
      }
      savname = is_void(outfile) ? "flux" : outfile;
      mem_save,savname,Flux;
   }

   if( !nof ) {
      kwds_init;
      kwds_set,"EXTNAME","PHOTON_FLUX","Name of extension";
      kwds_set,"DATE",ndate(3),"Date and time of creation";
      kwds_set,"ORIGIN","mk_photflux.i","Software used";
      kwds_set,"EBASE",Ebase,"Pivot energy for spectrum";
      if( in_xspecmode ) {
         kwds_set,"ORIGIN1","XSPEC model","Using an XSPEC generated flux";
         kwds_set,"MODEL", xspecmodel, "The actual spectral model";
         for( i = 1; i <= numberof(xspecparms); i++ ) \
            kwds_set,"PARAM"+itoa(i), xspecparms(i),"Value of spectral parameter "+itoa(i);
      } else { // the 'primitive' way
         kwds_set,"ORIGIN1","Simple model","Using an xfit_photspec generated flux";
         kwds_set,"CODE",Spectrcode,"Spectral type";
         kwds_set,"CONST",Const,"Normalization constant";
         kwds_set,"PARAM1",Param1,"Parameter 1 photindex or kT";
         kwds_set,"NH",Nh,"[/cm2] Column density";
      }
      kwds_set,"E_MIN",Energy_start,"[keV] Lower energy limit";
      kwds_set,"E_MAX",Energy_stop,"[keV] Upper energy limit";
      kwds_set,"SX",Sx_photflux,"[erg/cm2/s] Integrated flux from E_MIN to E_MAX";

      if( is_void(outfile) ) {
         outfile = get_next_filename("photflux_????.fits");
         if(!silent) write,format="Has defined new output filename: %s\n", outfile;
      } else {
         if( !silent ) write,format="Keeps the keyword output filename: %s\n", outfile;
      }
      wrmfitscols, outfile, "ENERG_LO", energ_lo, "ENERG_HI", energ_hi, \
                "PHOTFLUX", Flux, clobber=1;
      if( !silent) write,format="Has written file: %s\n", outfile;
   }
   if( !silent ) {
      write,format="Sx (%4.1f-%4.1f keV) = %9.2e erg/cm2/s\n", Energy_start, Energy_stop, Sx_photflux;
      if( !is_void(Exposure) ) {
         write,format="Photon density: %.2f /mm2\n", sum(Flux*(Ehi-Elo))*1.e-2*Exposure;
         write,format="      with an Exposure of %.1f s\n", Exposure;
      }
   }
 
}

/* Function rd_arf */

func rd_arf( filename, &e_arf, &arf, num=, extname= )
/* DOCUMENT rd_arf, filename, >e_arf, >arf, num=, extname=

   Returns an energy  and arf array from extension "SPECRESP"
   in the FITS file: filename. The columns ENERG_LO, ENERG_HI,
   and SPECRESP must exist.

   For PHAII format 'num' may be given (defaults to 1 (one)).

   Arguments:
      filename      name of file with ARF information
         e_arf      (returned) Energy array
           arf      (returned) ARF array

   Keywords:
           num      Row number (relevant for typeII file)
       extname      Extension name if different from "SPECRESP"

   SEE ALSO: read_arf (rmf_funcs.i)

*/
{
   if( typeof(extname) == "string" ) extname = strupcase(extname);
   else extname = "SPECRESP";

   dol = filename+"["+extname+"]";
   arf = rdfitscol(dol,"SPECRESP");
   elo = rdfitscol(dol,"ENERG_LO");
   ehi = rdfitscol(dol,"ENERG_HI");

   // Reformatting is required for PHAII ARF

   dms = dimsof( arf );
   if( dms(1) > 1 ) {
      if( is_void(num) ) num = 1;
      arf = arf(,num);
      elo = elo(,num);
      ehi = ehi(,num);
   }

   e_arf = 0.5*(elo + ehi);
}

/* Function predict_cr */

func predict_cr( file_with_arf, emin=, emax= )
/* DOCUMENT cr = predict_cr( file_with_arf, emin=, emax= )

   Returns the predicted count rate of a source with flux
   as produced by 'mk_photflux' and stored in the external
   variables: Eline and Flux.

   The energy min/max of 'Eline' defines the energy interval

   It reads the energy and arf array from extension "SPECRESP"
   in the FITS file: file_with_arf. The columns ENERG_LO, ENERG_HI,
   and SPECRESP must exist.

   SEE ALSO: rd_arf
*/
{
   if( is_void(emin) ) emin = Eline(1);
   if( is_void(emax) ) emax = Eline(0);

   local e_arf, arf;
   rd_arf, file_with_arf, e_arf, arf;

   // Check consistency
   if( emin < e_arf(1) ) {
      write,"Warning: Requested emin ("+ftoa(emin,ndec=2)+" keV) < Earf,min ("+ftoa(e_arf(1),ndec=2)+" keV)";
   }
   if( emin < Eline(1) ) {
      write,"Warning: Requested emin ("+ftoa(emin,ndec=2)+" keV) < Eline,min ("+ftoa(Eline(1),ndec=2)+" keV)";
   }
   if( emax > e_arf(0) ) {
      write,"Warning: Requested emax ("+ftoa(emax,ndec=2)+" keV) > Earf,max ("+ftoa(e_arf(0),ndec=2)+" keV)";
   }
   if( emax > Eline(0) ) {
      write,"Warning: Requested emax ("+ftoa(emax,ndec=2)+" keV) > Eline,max ("+ftoa(Eline(0),ndec=2)+" keV)";
   }

   is = where( emin <= Eline & emax >= Eline );

   if( !numberof(is) ) error,"No Eline elements in interval";

   qarf = interp(arf, e_arf, Eline(is));
   return sum(qarf(zcen)*(Flux(is))(zcen)*(Eline(is))(dif));

}

/* Function rd_photflux */

func rd_photflux( filename, chat= )
/* DOCUMENT rd_photflux, filename, chat=

    Defines external variables 'Eline' and 'Flux'
    from the contents of the photon flux file
    given in the first extension of 'filename'.

    Keyword 'chat' will trigger the display of the
    FITS header information.
*/
{
   extern Eline, Flux, Sx_photflux;

   dol = filename+"+1";
   elo = rdfitscol( dol, "ENERG_LO" );
   ehi = rdfitscol( dol, "ENERG_HI" );
   Flux = rdfitscol( dol, "PHOTFLUX" );
   Eline = sqrt(elo*ehi);
   Sx_photflux = sflux(elo(1),ehi(0),Eline,Flux);

   write,"Externals 'Eline', 'Flux', and 'Sx_photflux' have now been defined.";

   if( chat ) {
      hdr = headfits(dol);
      const = fxpar(hdr,"const");
      code = fxpar(hdr,"code");
      param1 = fxpar(hdr,"param1");
      nh = fxpar(hdr,"nh");
      e_min = fxpar(hdr,"e_min");
      e_max = fxpar(hdr,"e_max");
      sx = fxpar(hdr,"sx");
      nchans = fxpar(hdr,"naxis2");
      write,format="%s : const = %9.2e, par = %5.2f, NH = %9.2e\n", \
                code, const, param1, nh;
      write,format="  between %.2f and %.2f keV in %i channels\n", \
                e_min, e_max, nchans;
      write,format="  and Sx = %9.2e erg/cm2/s in that interval\n", sx;
   }
}

/* Function add_phaii_spectra */

/*************************************************

 Add two or more PHAII spectra with exposure as
 weighting factor

  add_p2spectra alias add_phaii_spectra
*************************************************/

func add_phaii_spectra( list, outfile, sourceid=, name=, rdr= )
/* DOCUMENT add_phaii_spectra, list_of_filenames, outfilename, \
                           sourceid=, name=, rdr=

   Assumes extension #1

  Keywords (only one of these can be given):
     sourceid     SOURCE_ID  requires exact match
     name         NAME       requires matching lowercase, no spaces
     rdr          3-element array [RA, dec, radius] all in degrees

*/
{
   local ancrfile, ndum;

   sel = 0;
   if( !is_void(sourceid) ) sel += 1;
   if( !is_void(name) ) sel += 2;
   if( !is_void(rdr) ) sel += 4;
   if( sel != 1 && sel != 2 && sel != 4 ) error,"Bad keywords";

   kwds_init;
   if( sel == 1 ) {
      sourceid = strtrim(sourceid);
      if( strtolower(sourceid)  ==  "crab" ) sourceid = "J053432.0+220052";
      kwds_set,"COMMENT","Selecting SOURCE_ID == "+sourceid;
   }
   if( sel == 2 ) {
      name = strlowcase(strcompress(name,all=1));
      kwds_set,"COMMENT","Selecting source NAME == "+name;
   }
   if( sel == 4 ) kwds_set,"COMMENT","Selecting source at "+\
              swrite(format="RA=%7.3f dec=%7.3f rad=%5.3f",rdr(1),rdr(2),rdr(3));
   kwds_set,"COMMENT","Including these files:";
   n = numberof(list);

   rate_sum = [];
   rate_sum_err2 = [];
   expo_sum = 0.0;
   arf_sum = [];
   for( i = 1; i <= n; i++ ) {
      write,"Extracting spectrum from "+list(i)+" ...";
      if( !file_test(list(i)) ) {
         write,"Did not find "+list(i)+", skip and continue ...";
         continue;
      }
      dol = list(i)+"+1";
      hdr = headfits( dol );
      exposure = fxpar( hdr, "exposure" );
      if( is_void(exposure) ) exposure = 1.0;
      write,"           with EXPOSURE = ",exposure;

      rate = rdfitscol( dol, "rate" );
      rate_err = rdfitscol( dol, "stat_err" );
      ancrfiles = rdfitscol( dol, "ancrfile" );
      respfile = fxpar( hdr, "respfile" );
      if( sel == 1 ) {
         source_id = strtrim(rdfitscol( dol, "source_id" ));
         w = where( source_id == sourceid );
         if( numberof(w) == 0 ) {
            write,"Source "+sourceid+" not found in "+dol;
            continue;
         }
      } else if( sel == 2 ) {
         names = strlowcase(strcompress(rdfitscol( dol, "name" ),all=1));
         w = where( names == name );
         if( numberof(w) == 0 ) {
            write,"Source "+name+" not found in "+dol;
            continue;
         }
      } else {
         ra = rdfitscol( dol, "ra_obj" );
         dec = rdfitscol( dol, "dec_obj");
         r = arcdist( ra, dec, rdr(1), rdr(2) );
         rmin = min(r);
         if( rmin > rdr(3) ) {
            write,"No sources close to selection";
            continue;
         }
         w = where( r == rmin );
      }
      src_num = w(1);
      rate = rate(,src_num);
      rate_err = rate_err(,src_num);
      // remove the {1} on the ancrfile name
      get_exten_no, ancrfiles(src_num), ancrfile, ndum, tub=1;

      if( !file_test(ancrfile) ) {
         write,"Did not find "+ancrfile+", skip and continue ...";
         continue;
      }

      if( !file_test(respfile) ) {
         write,"Did not find response file: "+respfile;
         return;
      }

      kwds_set,"COMMENT",list(i);
      arf = rdfitscol( ancrfile+"[SPECRESP]","specresp");
      elo = rdfitscol( ancrfile+"[SPECRESP]","energ_lo");
      ehi = rdfitscol( ancrfile+"[SPECRESP]","energ_hi");
      arf = arf(,src_num);
      elo = elo(,src_num);
      ehi = ehi(,src_num);
      if( is_void(arf_sum) ) {
         arf_sum = arf*exposure;
      } else {
         arf_sum += arf*exposure;
      }
      expo_sum += exposure;
      if( is_void(rate_sum) ) {
         rate_sum = rate;
         rate_sum_err2 = rate_err^2;
      } else {
         rate_sum += rate;
         rate_sum_err2 += rate_err^2;
      }

      //+ e_min = rdfitscol( respfile+"[EBOUNDS]", "e_min" );
      //+ e_max = rdfitscol( respfile+"[EBOUNDS]", "e_max" );

   }

   arf = arf_sum / expo_sum;
   ancrfile = strpart(outfile,1:-5)+"_ancr.fits";
   wrmfitscols,ancrfile,"ARF_NUM", [1], "ENERG_LO", elo(,-), "ENERG_HI", ehi(,-), \
               "SPECRESP", arf(,-), extname="SPECRESP",clobber=1;

   spec2phaii, outfile, rate_sum, sqrt(rate_sum_err2), exposure=expo_sum, \
               ancrfile=ancrfile, respfile=respfile,telescop=telescop, \
               instrume=instrume, no_kwds_init=1;

}
add_p2spectra = add_phaii_spectra;

/* Function spec2phaii */

func spec2phaii( filename, rate, stat_err, type=, name=, rowid=, ra_obj=, dec_obj=, exposure=, \
                 poisserr=, ancrfile=, respfile=, backfile=, backscal=, telescop=, instrume=, no_kwds_init= )
/* DOCUMENT spec2phaii, filename, rate, stat_err

     Writes spectrum or spectra in counts/s to PHA II format with clobber=1
     Dimension of rate and stat_err:    nbins x nrows

     Keywords:
        The following keywords must have as many elements as there are
        spectra (rows):
             ra_obj    [deg]
             dec_obj   [deg]
             exposure  [s]
             name      (string)  synonymous with 'rowid'
             rowid     (string)  synonymous with 'name'
             ancrfile  (string)  file containing pertinent ARF
             backfile  (string)  file with background spectrum with same number of spectra
                                 as in 'rate'
             backscal  (double)  ratio of region area to total detector area

        These keywords are single valued (becomes part of FITS header)

             type      (string)  One of: "net" (default), "total", or "bkg"
             respfile  (string)  name of response (RMF) file
             telescop  (string)  Name of the mission or telescope
             instrume  (string)  Name of the instrument
             poisserr  (char)    Logical for Poisson error, defaults to 'T'
                                 Must be set to zero or 'F' for 'F'

        Keyword to cancel keyword initialization (kwds_init):  no_kwds_init

    2009-06-29/NJW
*/
{
   if( numberof(stat_err) != numberof(rate) ) error,"Argument size incompatibility";

   dms = dimsof(rate);
   if( anyof(dms-dimsof(stat_err)) ) error,"Argument dimension incompatibility";
   if( dms(1) == 2 ) {
      nrows = dms(3);
      nbins = dms(2);
   } else if( dms(1) == 1 ) {
      nrows = 1;
      nbins = dms(2);
      rate = reform(rate,nbins,nrows);
      stat_err = reform(stat_err,nbins,nrows);
   } else error,"Does not handle single value spectrum";

   if( is_void(type) ) {
      type = "net";
   } else { // check of legality
      if( typeof(type) != "string" ) error,"Keyword 'type' is not a string";
      type = strtrim(strlowcase(type));
      if( type != "net" && type != "total" && type != "bkg" ) error,"Illegal keyword 'type': "+type;
   }

   if( !is_void(ra_obj) ) {
      if( numberof(ra_obj) != nrows ) {
         write,"Wrong number of values in keyword ra_obj - forget";
         ra_obj = [];
      }
      ra_obj = reform(ra_obj,nrows);
   }

   if( !is_void(dec_obj) ) {
      if( numberof(dec_obj) != nrows ) {
         write,"Wrong number of values in keyword dec_obj - forget";
         dec_obj = [];
      }
      dec_obj = reform(dec_obj,nrows);
   }

   if( is_void(poisserr) ) {
      poisserr = 'T'; // default value
   } else {
      if( poisserr == 0 || poisserr == 'F' ) {
         poisserr = 'F';
      } else poisserr = 'T';
   }

   if( !is_void(exposure) ) {
      if( numberof(exposure) != nrows ) {
         write,"Wrong number of values in keyword exposure - expand first value";
         exposure = array(double(exposure(1)),nrows);
      }
      exposure = reform(exposure,nrows);
   } else {  // set default exposure of 1.0 s
      exposure = array( 1.0, nrows );
   }

   if( !is_void(rowid) ) name = rowid;
   if( !is_void(name) ) {
      if( numberof(name) != nrows ) {
         write,"Wrong number of values in keyword name (rowid) - forget";
         name = [];
      }
      name = reform(name,nrows);
   }

   if( !is_void(ancrfile) ) {
      if( numberof(ancrfile) != nrows ) {
         write,"Wrong number of values in keyword ancrfile - forget";
         ancrfile = [];
      }
      ancrfile = reform(ancrfile,nrows);
   }

   if( !is_void(respfile) ) {
      if( numberof(respfile) != 1 ) {
         write,"Wrong number of elements in keyword respfile - forget";
         respfile = [];
      }
      respfile = respfile(1);
   } else respfile = "unknown";

   ttype1 = "spec_num";
   data1 = short(indgen(nrows));
   ttype2 = "rowid";
   if( is_void(name) ) {
      data2 = array(strpadd(" ",32," "),nrows);
   } else {
      data2 = array(string,nrows);
      for(i=1;i<=nrows;i++) data2(i) = strpadd(name(i),32," ",truncate=1);
   }
   ttype3 = "ra_obj";
   data3 = is_void(ra_obj) ? array(float(-99.),nrows) : float(ra_obj);
   ttype4 = "dec_obj";
   data4 = is_void(dec_obj) ? array(float(-99.),nrows) : float(dec_obj);
   ttype5 = "channel";
   data5 = short(indgen(nbins))(,-:1:nrows);
   ttype6 = "rate";
   data6 = float(rate);
   ttype7 = "stat_err";
   data7 = float(stat_err);
   ttype8 = "sys_err";
   data8 = float(stat_err*0);
   ttype9 = "quality";
   data9 = array(short,nbins,nrows);
   ttype10 = "backfile";
   if( is_void(backfile) ) {
      data10 = array(strpadd(" ",160," "),nrows);
   } else {
      data10 = array(string,nrows);
      for(i=1;i<=nrows;i++) data10(i) = strpadd(backfile(i)+"{"+itoa(i)+"}",160," ",truncate=1);
   }
   ttype11 = "backscal";
   data11 = array(1.0,nrows);
   if( !is_void(backscal) ) {
      if( numberof(backscal == nrows) ) {
         data11 = backscal;
      } else write,"Wrong number of 'backscal' values - ignored";
   }
   ttype12 = "ancrfile";
   if( is_void(ancrfile) ) {
      data12 = array(strpadd(" ",160," "),nrows);
   } else {
      data12 = array(string,nrows);
      for(i=1;i<=nrows;i++) data12(i) = strpadd(ancrfile(i)+"{"+itoa(i)+"}",160," ",truncate=1);
   }
   ttype13 = "exposure";
   data13 = exposure;
   ttype14 = "telapse";
   data14 = array(1.0,nrows);
   ttype15 = "ontime";
   data15 = array(1.0,nrows);
   ttype16 = "deadc";
   data16 = array(1.0,nrows);
   ttype17 = "tfirst";
   data17 = array(1.0,nrows);
   ttype18 = "tlast";
   data18 = array(1.0,nrows);
   ttype19 = "evt_type";
   data19 = array(short,nrows);

   if( !no_kwds_init ) kwds_init;

   kwds_set,"extname","SPECTRA","Extension name";
   if( !is_void(telescop) ) kwds_set,"telescop",telescop,"Mission name";
   if( !is_void(instrume) ) kwds_set,"instrume",instrume,"Instrument";
   kwds_set,"date",ndate(3),"CET Central European time";
   kwds_set,"detchans", nbins,"Total number of detector channels";
   kwds_set,"hduclass","OGIP","Format conforms mostly to OGIP standards";
   kwds_set,"hduclas1","SPECTRUM","Dataset contains a spectrum";
   if( type == "net" ) commt = "Background subtracted spectra";
   if( type == "total" ) commt = "Total spectra";
   if( type == "bkg" ) commt = "Background spectra";
   kwds_set,"hduclas2",strupcase(type),commt;
   kwds_set,"hduclas3","RATE","Data is stored in counts/s";
   kwds_set,"hduclas4","TYPE:II","Format is PHAII";
   kwds_set,"hduvers","1.2.0","Version of format";
   kwds_set,"grouping",0,"No channel grouping";
   kwds_set,"poisserr",poisserr,"Logical for Poisson statistics";
   kwds_set,"corrscal",1.0,"Correction scaling factor";
   kwds_set,"areascal",1.0,"Nominal effective area";
   kwds_set,"respfile",respfile,"Pertaining RESPFILE";
   kwds_set,"tunit3","deg","Unit of column RA_OBJ";
   kwds_set,"tunit4","deg","Unit of column DEC_OBJ";
   kwds_set,"tlmin",1,"Minimum legal value of column CHANNEL";
   kwds_set,"tlmax",nbins,"Maximum legal value of column CHANNEL";
   kwds_set,"tunit6","counts/s","Unit of column RATE";
   kwds_set,"tunit7","counts/s","Unit of column STAT_ERR";
   kwds_set,"tunit8"," ","Unit of column SYS_ERR";
   kwds_set,"tunit13","s","Unit of column EXPOSURE";
   kwds_set,"tunit14","s","Unit of column TELAPSE";
   kwds_set,"tunit15","s","Unit of column ONTIME";
   kwds_set,"tunit17","s","Unit of column TFIRST";
   kwds_set,"tunit18","s","Unit of column TLAST";

   wrmfitscols,filename, ttype1, data1, ttype2, data2, ttype3, data3, \
                         ttype4, data4, ttype5, data5, ttype6, data6, \
                         ttype7, data7, ttype8, data8, ttype9, data9, \
                         ttype10, data10, ttype11, data11, ttype12, data12, \
                         ttype13, data13, ttype14, data14, ttype15, data15, \
                         ttype16, data16, ttype17, data17, ttype18, data18, \
                         ttype19, data19, clobber=1;

}

/* Function spec2phai */

func spec2phai( outfile, rate, stat_err, type=, name=, ra_obj=, dec_obj=, exposure=, \
                 poisserr=, ancrfile=, respfile=, backfile=, backscal=, telescop=, instrume=, no_kwds_init= )
/* DOCUMENT spec2phai, outfile, rate, stat_err

     Writes spectrum or spectra in counts/s to PHA I format with clobber=1
     Dimension of rate and stat_err:    nbins

     Arguments:
          outfile      Name of output file
             rate      [counts/s] or [counts]
         stat_err      [counts/s] If this argument is omitted then 'rate' is
                       interpreted as COUNTS.

     Keywords:
             ra_obj    [deg]
             dec_obj   [deg]
             exposure  [s]
             name      (string)  name of object
             ancrfile  (string)  file containing pertinent ARF
             backfile  (string)  file with background spectrum with same number of spectra
                                 as in 'rate'
             backscal  (double)  ratio of region area to total detector area
             type      (string)  One of: "net" (default), "total", or "bkg"
             respfile  (string)  name of response (RMF) file
             telescop  (string)  Name of the mission or telescope
             instrume  (string)  Name of the instrument
             poisserr  (char)    Logical for Poisson error, defaults to 'T'
                                 Must be set to zero or 'F' for 'F'

        Keyword to cancel keyword initialization (kwds_init):  no_kwds_init

    2015-02-27/NJW based on spec2phaii
*/
{
   nbins = numberof(rate);
   if( !is_void(stat_err) ) if( numberof(stat_err) != nbins ) error,"Argument size incompatibility";

   if( is_void(type) ) {
      type = "net";
   } else { // check of legality
      if( typeof(type) != "string" ) error,"Keyword 'type' is not a string";
      type = strtrim(strlowcase(type));
      if( type != "net" && type != "total" && type != "bkg" ) error,"Illegal keyword 'type': "+type;
   }

   if( is_void(poisserr) ) {
      poisserr = 'T'; // default value
   } else {
      if( poisserr == 0 || poisserr == 'F' ) {
         poisserr = 'F';
      } else poisserr = 'T';
   }

   if( is_void(exposure) ) exposure = 1.0; // default
   if( is_void(ancrfile) ) ancrfile = "none";
   if( is_void(respfile) ) respfile = "none";
   if( is_void(backscal) ) backscal = 1.0;

   ttype1 = "CHANNEL";
   data1 = short(indgen(nbins));
   if( is_void(stat_err) ) {
      ttype2 = "COUNTS";
      data2 = int(rate);
   } else {
      ttype2 = "RATE";
      data2 = float(rate);
      ttype3 = "STAT_ERR";
      data3 = float(stat_err);
   }

   if( !no_kwds_init ) kwds_init;
   kwds_set,"extname","SPECTRUM","Extension name";
   kwds_set,"ORIGIN","spec2phai","Software used";
   kwds_set,"DATE", ndate(3), "Date/time of file creation";
   if( structof(name) == string ) kwds_set,"OBJECT", name, "Name of object";
   if( !is_void(ra_obj) ) kwds_set,"RA_OBJ", ra_obj, "[deg] RA of object";
   if( !is_void(dec_obj) ) kwds_set,"DEC_OBJ", dec_obj, "[deg] DEC of object";
   kwds_set,"ANCRFILE", ancrfile, "Anchor file (ARF) for spectrum";
   kwds_set,"RESPFILE", respfile, "Response file (RMF) for spectrum";
   backfile = structof(backfile) == string ? backfile : "None";
   kwds_set,"BACKFILE", backfile, "Background file";

   telescop = structof(telescop) == string ? telescop : "NuSTAR";
   kwds_set,"telescop",telescop,"Mission name";
   instrume = structof(instrume) == string ? instrume : "FPMA";
   kwds_set,"instrume",instrume,"Instrument";
   kwds_set,"detchans", nbins,"Total number of detector channels";
   kwds_set,"hduclass","OGIP","Format conforms mostly to OGIP standards";
   kwds_set,"hduclas1","SPECTRUM","Dataset contains a spectrum";
   if( type == "net" ) commt = "Background subtracted spectra";
   if( type == "total" ) commt = "Total spectra";
   if( type == "bkg" ) commt = "Background spectra";
   kwds_set,"hduclas2",strupcase(type),commt;
   if( is_void(stat_err) ) {
      kwds_set,"hduclas3", "COUNTS","Data is stored in counts";
   } else {
      kwds_set,"hduclas3", "RATE","Data is stored in counts/s";
      kwds_set,"tunit2","counts/s","Unit of column RATE";
      kwds_set,"tunit3","counts/s","Unit of column STAT_ERR";
   }
   kwds_set,"hduclas4","TYPE:I","Format is PHAI";
   kwds_set,"hduvers","1.2.0","Version of format";
   kwds_set,"grouping",0,"No channel grouping";
   kwds_set,"poisserr",poisserr,"Logical for Poisson statistics";
   kwds_set,"corrscal",1.0,"Correction scaling factor";
   kwds_set,"areascal",1.0,"Nominal effective area";
   kwds_set,"backscal",backscal,"Ratio of region area to detector area";
   kwds_set,"exposure",exposure,"[s] Exposure time";
   kwds_set,"tlmin",1,"Minimum legal value of column CHANNEL";
   kwds_set,"tlmax",nbins,"Maximum legal value of column CHANNEL";

   if( is_void(stat_err) ) { // COUNTS
      wrmfitscols,outfile, ttype1, data1, ttype2, data2,  clobber=1;
   } else {
      wrmfitscols,outfile, ttype1, data1, ttype2, data2, \
                            ttype3, data3, clobber=1;
   }

}

/* Function spec_info */

func spec_info( specfile )
/* DOCUMENT spec_info, specfile

*/
{
   if( !file_test(specfile) ) {
      write,"File not found";
      return;
   }

   if( (pha_type = phatype( specfile )) == 0 ) {
      write,"Not a spectral file";
      return;
   }

   fh = headfits( specfile+"+1");

   if( pha_type == 2 ) {

      rate = rdfitscol(specfile+"+1","rate");
      dms = dimsof(rate);
      if( dms(1) == 2 ) {
         if( dms(3) == 1 ) {
            write,format="PHAII format with a single spectrum%s\n", "";
         } else {
            write,format="PHAII format with %i spectra", dms(3);
         }
         ancrfile = rdfitscol(specfile+"+1","ancrfile");
         write,format="First ANCRFILE: %s\n", strtrim(ancrfile(1));
      } else {
         write,"PHAI format";
         write,format="ANCRFILE: %s\n",fxpar(fh,"ancrfile");
      }
      write,format="%i energy bins\n", dms(2);
      write,format="RESPFILE = %s\n", fxpar(fh,"RESPFILE");

   } else {

      write,format="PHAI format%s\n","";
      ancrfile = fxpar( fh, "ANCRFILE");
      write,format="ARF: %s\n", ancrfile;
      respfile = fxpar( fh, "RESPFILE");
      write,format="RMF: %s\n", respfile;

   }

}

/* Function read_effa */

func read_effa( filename )
/* DOCUMENT effa = read_effa( filename )

    Returns a struct with pointers to energy array and 
    effective area array provided
      1) the file is a binary FITS extension containing
         the columns "ENERGY" and "EFF_AREA".
    or
      2) the file is a text table conforming to .scm format
         with columns "energy" and "eff_area"

    e_arr    = *effa.ener
    effa_arr = *effa.area
    cards    = *effa.hdr
*/
{
   s = s_Effarea();
   if( fits_check_file(filename) ) {
      hdr = headfits(filename+"+1");
      c = fits_colnum(hdr,"eff_area");
      u = fxpar(hdr,"TUNIT"+itoa(c));
      f = 1.0;
      if( typeof(u) == "string" ) {
         if( strlowcase(u) == "mm2" ) f = 1.e-2;
      }
      s.ener = &rdfitscol(filename+"+1","energy");
      //+ s.area = &(0.01*rdfitscol(filename+"+1","eff_area"));
      area = f*rdfitscol(filename+"+1","eff_area");
      s.area = &area;
      s.hdr = &listhead(hdr);
   } else {
      s.ener = &rscol(filename, "energy", nomem=1,silent=1);
      s.area = &rscol(filename, "eff_area", nomem=1,silent=1);
      null = [];
      s.hdr = &null;
   }
   return s;
}

/* Function write_effa */

func write_effa( filename, eastruct )
/* DOCUMENT write_effa, filename, eastruct
*/
{
   kwds_init;
   if( !is_void(*eastruct.hdr) ) fits_copy_keys, *eastruct.hdr, tokwds=2;
            // tokwds is set to 2 to avoid re-initialization
   kwds_set,"ORIGIN","write_effa in xray.i","Software used";
   kwds_set,"DATE", ndate(3),"Date/time of creation";
   kwds_set,"TUNIT1","keV","Unit of column ENERGY";
   kwds_set,"TUNIT2","cm2","Unit of column EFF_AREA";
   wrmfitscols,filename, "ENERGY", *eastruct.ener, "EFF_AREA", *eastruct.area;
}

/* Function plot_effa */

func plot_effa( filename, color=, itype=, li=, ps=, xr=, yr=, \
                thick=, title=, xtitle=, ytitle=, pane= )
/* DOCUMENT plot_effa, filename, color=, itype=, li=, ps=, xr=, yr=,
                thick=, title=, xtitle=, ytitle=, pane=

      'filename' is either a genuine filename or a struct (s_Effarea) as
      returned by 'read_effa'.
*/
{
   if( structof(filename) == string ) {
      effa = read_effa( filename );
   } else effa = filename;

   if( !is_void(pane) ) window,pane;

   if( is_void(xtitle) ) xtitle = "Energy [keV]";
   if( is_void(ytitle) ) ytitle = "Effective Area [cm^2^]";

   plot,*effa.ener, *effa.area, color=color, itype=itype, \
         li=li, ps=ps, thick=thick, xr=xr, yr=yr, title=title, \
         xtitle="Energy [keV]", ytitle="Eff. Area [cm^2^]";
}

/* Function oplot_effa */

func oplot_effa( filename, color=, li=, thick=, ps= )
/* DOCUMENT oplot_effa, filename, color=, li=, ps=, thick=

      'filename' is either a genuine filename or a struct (s_Effarea) as
      returned by 'read_effa'.
*/
{
   if( structof(filename) == string ) {
      effa = read_effa( filename );
   } else effa = filename;

   oplot,*effa.ener, *effa.area, color=color, \
         li=li, ps=ps, thick=thick;
}

/* Function plot_2effas */

func plot_2effas( filename1, filename2, color=, itype=, li=, ps=, xr=, yr=, \
                  thick=, title=, xtitle=, ytitle=, frsize=, ndcymin=, pane= )
/* DOCUMENT plot_2effas, filename1, filename2, color=, itype=, li=, ps=, xr=, yr=,
                thick=, title=, xtitle=, ytitle=, frsize=, ndcymin=, pane=

    Filenames may also be structs (s_Effarea) as returned by 'read_effa'.

    Keyword color may be an array with two elements.

    Keyword
     frsize    Relative frame size will be frsize:1 (default 3:1)
     ndcymin   Lower edge of plot frame in NDC (default 0.1)
*/
{
   if( structof(filename1) == string ) {
      effa1 = read_effa( filename1 );
   } else effa1 = filename1;

   if( structof(filename2) == string ) {
      effa2 = read_effa( filename2 );
   } else effa2 = filename2;

   if( !is_void(color) ) {
      if( numberof(color) == 1 ) color = [color,color];
   }
   if( is_void(xtitle) ) xtitle = "Energy [keV]";
   if( is_void(ytitle) ) ytitle = "Effective Area [cm^2^]";
   if( is_void(frsize) ) frsize = 3.;
   if( is_void(ndcymin) ) ndcymin = 0.1;

   mplot_setup, 21, spacing=0.001, hideticklabels=1, vrelsize=[frsize,1], \
              vport=[0.1,0.7,ndcymin,0.95], pane=pane;

   c = is_void(color) ? [] : color(1);
   plot,*effa1.ener, *effa1.area, color=c, itype=itype, \
         li=li, ps=ps, thick=thick, xr=xr, yr=yr, mpl=1;
   c = is_void(color) ? [] : color(2);
   oplot,*effa2.ener, *effa2.area, color=c, \
         li=li, ps=ps, thick=thick;

   // determine overlap in energy ranges
   e1 = *effa1.ener;
   e2 = *effa2.ener;
   a1 = *effa1.area;
   a2 = *effa2.area;
   emin = max(min(e1),min(e2));
   emax = min(max(e1),max(e2));
   if( emin < emax ) {
      w1 = where( e1 >= emin & e1 <= emax );
      //+ w2 = where( e2 >= emin & e2 <= emax );
      a2_adj = interp( a2, e2, e1(w1) );
      w = where(a2_adj > 0. );
      r = a1(w1(w))/a2_adj(w);
      rmax = r(max);
      rmin = 2. - rmax;
      rmn = r(min);
      if( rmn < rmin ) {
         rmin = rmn;
         rmax = 2. - rmin;
      }
      dr = (rmax - rmin)*0.07;
      plot, [e1(1),e1(0)], [1,1], li=2, xr=xr, yr=[rmin-dr,rmax+dr], mpl=2;
      oplot, e1(w1(w)), r;
   }

   dy = (0.95 - ndcymin)/(2*(frsize+1));
   ypos = dy * (frsize + 2) + ndcymin;
   devicetitles,title, [0.4,0.97],xtitle,[0.4,ndcymin-0.05],ytitle,[0.05,ypos];
   ypos = dy + ndcymin;
   devicetitles,,,,,"Ratio",[0.05,ypos];
}

/* Function specunpack */

func specunpack( res, &eb1, &eb2, &rate, &stat_err )
/* DOCUMENT specunpack, res_struct, >eb1, >eb2, >rate, >stat_err

      Unpack the struct 's_Spectrum' for the basic information
*/
{
   eb1 = *res.e_min;
   eb2 = *res.e_max;
   rate = *res.rate;
   stat_err = *res.stat_err;
}

/* Function read_spectrum */

func read_spectrum( dol_or_file, num=, adhoc=, respfile=, silent= )
/* DOCUMENT res = read_spectrum( dol_or_file, num=, adhoc=, respfile=, silent= )

     Returns a spectrum that may be either PHAI or PHAII in 'dol_or_file'
     as a struct (s_Spectrum).
     The struct elements are:
        n       : (long) number of channels
        specnum : (long) number of spectrum (for PHAII format)
        exposure: (double) [s] exposure time
        backscal: (double) Relative region area
        rate    : (pointer) The RATE array
        stat_err: (pointer) the STAT_ERR array
        backfile: (string) BACKFILE file
        rmf     : (string) RMF file
        arf     : (string) ARF
        specdol : (string) DOL of spectrum
    In case of missing file or a non-spectral extension a void is returned.

   Keywords:
      num      number of spectrum in PHAII file, ignored in PHAI file
    adhoc      A temporary work-around of FITS format not compliant with
               current version of mfits.i and kfits.i
 respfile      sets the RMF file, overrides what is given in spec file
   silent      Suppress terminal messages (but not errors)

   2013-07-18/NJW
*/
{

   local filename, extno;

   if( dol_or_file == "ia" ) {
      dol_or_file = ia_file_selection();
   }

   if( (ipha = phatype( dol_or_file )) == 0 ) {
      write,"Not a spectrum";
      return [];
   }

   v = !silent;

   if( ipha == 1 ) return rd_phai_spectrum( dol_or_file, respfile=respfile, adhoc=adhoc, silent=silent );

   // When you come here it must be a PHAII spectrum

   // Check for the xxxx.pha{num} notation
   if( strpart(dol_or_file,0:0) == "}" ) {
      get_exten_no, dol_or_file, filename, extno, tub=1;
      num = extno;
      dol_or_file = filename;
   }

   if( is_void(num) ) num = 1;

   dol = dol_or_file;
   get_exten_no, dol, filename, extno;
   if( extno == 0 ) dol += "+1"; // defaults to extension 1

   hdr = headfits( dol );
   nrows = fxpar( hdr, "naxis2" );
   if(v)write,itoa(nrows)+" spectra in "+dol;
   if( num > nrows ) {
      write,"You asked for a too high spectrum number";
      return [];
   }

   res = s_Spectrum();
   res.specdol = dol;
   res.specnum = num;

   if( is_void( (expo = rdfitscol( dol, "exposure" )) ) ) error,"READ_SPECTRUM: no EXPOSURE column";
   if( is_void( (rate = rdfitscol( dol, "rate" )) ) ) error,"READ_SPECTRUM: no RATE column";
   if( is_void( (backscal = rdfitscol( dol, "backscal" )) ) ) error,"READ_SPECTRUM: no BACKSCAL column";
   if( is_void( (backfile = rdfitscol( dol, "backfile" )) ) ) error,"READ_SPECTRUM: no BACKFILE column";
   rate_err = rdfitscol( dol, "stat_err" );
   if( is_void(respfile) ) respfile = fxpar( hdr, "respfile" );
   dms = dimsof(rate);

   res.n = dms(2);
   res.exposure = expo(num);
   res.backscal = backscal(num);
   res.backfile = strtrim(backfile(num));
   res.rmf = respfile;
   res.arf = strtrim(rdfitscol( dol, "ancrfile")(num));
   if(v) {
      write,"EXPOSURE: "+ftoa(res.exposure,ndec=1)+" s";
      write,"BACKSCAL: "+ftoa(res.backscal,ndec=6);
      write,"BACKFILE: "+res.backfile;
      write,"RESPFILE: "+respfile;
      write,"ANCRFILE: "+res.arf;
   }

   // select the requested spectrum

   res.rate = &rate(,num);
   res.stat_err = &rate_err(,num);

   if( !file_test(respfile) ) {
      write,"Did not find response file: "+respfile;
      return;
   }

   // Analyze respfile to find EXTNAMEs

   n_extens = nfits_extens( respfile );
   extnum = 0;
   for( i = 1; i <= n_extens; i++ ) {
      hdr = headfits( respfile+"+"+itoa(i) );
      extnm = fxpar( hdr, "extname" );
      if( extnm == "EBOUNDS" ) extnum = i;
      if( strmatch( extnm , "FBDS" ) ) extnum = i;
   }
   if( extnum == 0 ) {
      write,"Warning! Neither EBOUNDS nor JMXi-FBDS-MOD extension found in "+respfile;
      e_min = e_max = [];
   } else {
      e_min = rdfitscol( respfile+"+"+itoa(extnum), "e_min" );
      e_max = rdfitscol( respfile+"+"+itoa(extnum), "e_max" );
   }

   res.e_min = &e_min;
   res.e_max = &e_max;

   return res;

}

/* Function rd_phai_spectrum */

func rd_phai_spectrum( dol_or_filename, respfile=, adhoc=, silent= )
/* DOCUMENT res = rd_phai_spectrum( dol_or_filename, adhoc=, silent= )

     Returns a spectrum in the struct 'res' with elements:
        n       : (long) number of channels
        specnum : (long) number of spectrum (for PHAII format)
        exposure: (double) [s] exposure time
        backscal: (double) Relative region size
        rate    : (pointer) The RATE array
        stat_err: (pointer) the STAT_ERR array
        backfile: (string) BACKFILE file
        rmf     : (string) RMF file
        arf     : (string) ARF
        specdol : (string) DOL of spectrum

     Keywords
     respfile   :  Override RESPFILE in FITS header
        adhoc   :  Set when a fix in dataformat is required
       silent   :  Suppress terminal messages (but not errors)
*/
{
   local dol, extno;

   v = !silent;
   get_exten_no, dol_or_filename, dol, extno;
   if( !file_test(dol) ) error,dol+" was not found.";
   res = s_Spectrum();
   if( extno == 0 ) {
      dol += "+1"; // default extension number for spectrum
   } else dol = dol_or_filename;
   res.specdol = dol;
   res.specnum = 1; // only option for PHAI format
   hdr = headfits( dol );
   res.exposure = exposure = fxpar( hdr, "exposure" );
   res.backscal = fxpar( hdr, "backscal" );
   rb = fxpar( hdr, "backfile" );
   res.backfile = is_void(rb) ? "" : rb;
   if( is_void(respfile) ) respfile = fxpar( hdr, "RESPFILE" );
   if(v)write,"RESPFILE: "+respfile;
   if( is_void(respfile) ) {
      res.rmf = "";
   } else {
      if( file_test(respfile) ) {
         res.rmf = respfile;
         a = rdfitscol( respfile+"[EBOUNDS]", "e_min" );
         res.e_min = &a;
         a = rdfitscol( respfile+"[EBOUNDS]", "e_max" );
         res.e_max = &a;
      } else {
         write,"RESPFILE: '"+respfile+"' was not found";
         res.rmf = "";
      }
   }
   ancrfile = fxpar( hdr, "ANCRFILE" );
   if(v)write,"ANCRFILE: "+ancrfile;
   if( is_void(ancrfile) ) {
      res.arf = "";
   } else {
      if( file_test(ancrfile) ) {
         res.arf = ancrfile;
         if( ! adhoc ) {
            a = rdfitscol( respfile+"[MATRIX]", "energ_lo" );
            res.energ_lo = &a;
            a = rdfitscol( respfile+"[MATRIX]", "energ_hi" );
            res.energ_hi = &a;
         }
      } else {
         write,"ANCRFILE: '"+ancrfile+"' was not found";
         res.arf = "";
      }
   }

   naxis2 = fxpar(hdr, "NAXIS2");
   extname = fxpar(hdr,"EXTNAME");
   ogip = fxpar(hdr,"HDUCLASS");
   if( strtrim(ogip) == "OGIP" ) {
      if(v)write,dol_or_filename+" follows OGIP conventions";
      clas1 = fxpar(hdr,"HDUCLAS1");
      clas2 = fxpar(hdr,"HDUCLAS2");
      clas3 = fxpar(hdr,"HDUCLAS3");
      if( is_void(clas1) ) clas1 = "<not found>";
      if( is_void(clas2) ) clas2 = "<not found>";
      if( is_void(clas3) ) clas3 = "<not found>";
      if(v)write,"HDUCLAS1 = "+clas1;
      if(v)write,"HDUCLAS2 = "+clas2;
      if(v)write,"HDUCLAS3 = "+clas3;
   }

   // There must be a column by name of "RATE" or "COUNTS"

   rate = rdfitscol(dol,"RATE",silent=silent);
   if( is_void(rate) ) {
      is_rate = 0;
      if(v)write,"No RATE column found, search for COUNTS ...";
      counts = rdfitscol(dol,"COUNTS");
      if( is_void(counts) ) {
         if(v)write,"No COUNTS column found, not a spectral file";
         return [];
      } else { // found the COUNTS column, check exposure
         if(v)write,"                                    - found!";
         if( is_void(exposure) ) {
            write,"RD_PHAI_SPECTRUM: No EXPOSURE keyword - strange, set to 1.0 s";
            exposure = 1.0;
         }
         rate = float(counts) / exposure;
         rate_err = sqrt(counts) / exposure;
      }
   } else is_rate = 1;


   res.n = numberof(rate);
   if( is_rate ) {
      rate_err = rdfitscol( dol, "stat_err" );
   }
   res.rate = &rate;
   res.stat_err = &rate_err;

   return res;
}

/* Function phatype */

func phatype( dol_or_file, silent= )
/* DOCUMENT type = phatype( dol_or_file, silent= )

     Returns 1 for a PHAI type spectral file and
             2 for a PHAII spectral file.
             0 for indetermined type (or non-spectrum)

     If a filename is given then the spectrum is assumed to be in
     extension one.

     Keywords
       silent  :   Skip messages on terminal
*/
{
   local filename, extno;

   // check for filename of type :   xxxxx.pha{2}
   // indicating a PHAII file with spectrum number

   if( strpart(dol_or_file,0:0) == "}" ) {
      // strip the {N} part
      get_exten_no, dol_or_file, filename, extno, tub=1;
      // now extno contains 'N' and 'filename' is the stripped version
      dol_or_file = filename;
   }
   get_exten_no, dol_or_file, filename, extno;

   ires = 0;
   if( extno == 0 ) {
      if( strlen(dol_or_file) > strlen(filename) ) {
         write,"Ext#0 is illegal for PHA file";
         return ires;
      }
      // We have a simple filename, assume ext# is 1
      extno = 1;
   }
   if( !file_test(filename) ) {
      write,"File not found: "+filename;
      return ires;
   }
   dol = filename+"+"+itoa(extno);
   hdr = headfits(dol);
   naxis2 = fxpar(hdr, "NAXIS2");
   if(v)write,"NAXIS2 = "+itoa(naxis2);
   extname = fxpar(hdr,"EXTNAME");
   if(v)write,"EXTNAME = "+extname;
   ogip = fxpar(hdr,"HDUCLASS");
   if( strtrim(ogip) == "OGIP" ) {
      if(v)write,"Follows OGIP conventions";
      clas1 = fxpar(hdr,"HDUCLAS1");
      clas2 = fxpar(hdr,"HDUCLAS2");
      clas3 = fxpar(hdr,"HDUCLAS3");
      if( is_void(clas1) ) clas1 = "<not found>";
      if( is_void(clas2) ) clas2 = "<not found>";
      if( is_void(clas3) ) clas3 = "<not found>";
      if(v)write,"HDUCLAS1 = "+clas1;
      if(v)write,"HDUCLAS2 = "+clas2;
      if(v)write,"HDUCLAS3 = "+clas3;
   }

   // There must be a column by name of "RATE" or "COUNTS"

   colname = "RATE";
   rate = rdfitscol(dol,colname);
   if( is_void(rate) ) {
      is_rate = 0;
      if(v)write,"No RATE column found, search for COUNTS ...";
      colname = "COUNTS";
      counts = rdfitscol(dol,colname);
      if( is_void(counts) ) {
         if(v)write,"No COUNTS column found, not a spectral file";
         return ires;
      } else { // found the COUNTS column, check exposure
         if(v)write,"Found COUNTS column";
         exposure = fxpar(hdr,"EXPOSURE");
         if( is_void(exposure) ) {
            write,"No EXPOSURE keyword - strange, set to 1.0 s";
            exposure = 1.0;
         }
         rate = float(counts) / exposure;
      }
   } else {
      if(v)write,"Found RATE column";
      is_rate = 1;
   }

   dms = dimsof(rate);
   if( dms(1) == 0 ) {
      write,"Column "+colname+" has bad dimensions, not a spectral file";
      return ires;
   } else if( dms(1) == 1 ) {
      if(v)write,"Column "+colname+" has a single dimension, assume PHAI format";
      if(v)write,"with "+itoa(numberof(rate))+" channels";
      ires = 1;
   } else if( dms(1) == 2 ) {
      if(v)write,"Column "+colname+" has two dimensions, assume PHAII format";
      s = dms(3) == 1 ? "with one spectrum" : "with "+itoa(dms(3))+" spectra, each";
      if(v)write,s+" with "+itoa(dms(2))+" channels";
      ires = 2;
   } else {
      if(v)write,"Column "+colname+" has bad dimensions, not a spectral file";
   }
   return ires;
}

// ==============================================================================================

/* Function get_xspec_fit_params */

func get_xspec_fit_params( xspec_log_file, &fitqual, &flux, nofit=, outfile=, table=, silent= )
/* DOCUMENT get_xspec_fit_params, xspec_log_file, >fitqual, nofit=, outfile=, table=, silent=
    or      res = get_xspec_fit_params( xspec_log_file, >fitqual, nofit=, silent= )

  Extract fitting parameters from XSPEC logfile

   Assumes that the fitting parameters are found between a line with
     =================
   and
     _________________

  Returns an array of structs 's_Parval' with elements:
     model, param_num, model_num, component, name, unit, value, value_err, status

  and a struct 'fitqual' with elements red_chi2 and dof.

    Keyword
    outfile   Name of output file (works only when called as a subroutine)
    table     (bool) For output as a table, one line per log file:
                val1 err_val1 val2 err_val2 ... xspec_log_file
    silent    Quenches the monitoring screen prints
    nofit     (bool) set when no fit has been done, only given parameters are returned

 2004-06-03/NJW, original IDL version
 2008-08-12/NJW translated to Yorick
 2011-03-31/NJW updated with keyword 'flux'
 2013-01-18/NJW More general version
 2017-05-24/NJW Accepts fitting the response parameters
*/
{


   logfile = xspec_log_file;

   flag = typeof( outfile ) == "string";
   if( flag ) lun = open( outfile,"a");
   
   
   lines = rdfile(logfile);
   nlines = numberof(lines);

   // Search for the presence of 'cstat'. It is a prerequisite for
   // looking for a 'C-statistic' value

   cstat = where(strmatch(lines,"cstat"));
   cstat = numberof(cstat);

   // Search for the presence of '#Response Parameters:' to tell if they have
   // been fitted as well

   iresp = where(strmatch(lines,"#Response Parameters:"));
   niresp = numberof(iresp);

  /*
   *  The valid parameters are between a line with ===========
   *  and a line with ______________
   *
   *  The sections between ==========  and ----------
   *  do not contain interesting information parameter information
   *
   */
   
   // Search for the presence of "============", "___________", and "------------"

   idx1 = where(strmatch(lines,"================================"));
   idx2 = where(strmatch(lines,"________________________________"));
   idx3 = where(strmatch(lines,"--------------------------------"));

   // Discard all 'idx1' values that are followed by 'idx3' values
   if( numberof(idx3) ) {
      for( k = 1; k <= numberof(idx3); k++ ) {
         w = where( idx1 < idx3(k) );
         if( !numberof(w) ) error,"##1##";
         i = w(0); // latest before idx3(k) - is there any idx2 in between?
         v = where( idx2 < idx3(k) );
         if( numberof(v) ) if( idx2(v(0)) > idx1(i) ) error,"##1##";
         idx1 = rem_elem( idx1, i);
      }
   }
   nidx1 = numberof(idx1);
   if( nidx1 != numberof(idx2) ) error,"##1##";

   // 'k' is the index of idx1 to source model parameters
   k = niresp ? nidx1-1 : nidx1;

   // idx1(k) and idx2(k) now points to the last appearance of the
   //  source model parameters

   nparams = idx2(k) - idx1(k) - 4;
   if( niresp ) {   // i.e. when a gain fitting has been performed
      ngparams = idx2(k+1) - idx1(k+1) - 4; // appears after model parameters
      idx1b = idx1(k+1) + 4;
      idx2b = idx2(k+1) - 1;
   } else ngparams = 0;

   idx1a = idx1(k) + 4;
   idx2a = idx2(k) - 1;
   if(!silent) {
      last = idx2a+10;
      if( last > nlines ) last = nlines;
      prstrarr, lines(idx1a:last);
   }

   // Extract model name from line idx1a-3
   line = lines(idx1a-3);
   p = strpos(line,"Model"); // model component starts after this
   q = strpos(line,"Source");
   mdescr = strtrim(strpart( line, p+5:q-1 ));
   mdescr = _rm_brack( mdescr );

   // Locate 'Reduced chi-squared' following the parameter list
   // which means that the line number must be higher than idx2a

   fitqual = s_Fitqual();
   if( cstat ) {
      idx3a = where(strmatch(lines(idx2a+2:0),"C-Statistic ="));
      if( numberof(idx3a) > 0 ) {
         line = lines(idx3a(1)+idx2a+1);
         p = strpos(line,"C-Statistic")+13; // begin of value
         pf = strpos(line,"using",p+13); // end of value
         cstatis = atof(strpart(line,p:pf-1));
         p = strpos(line,"and",pf);
         pf = strpos(line,"degrees",p+3);
         fitqual.dof = atoi(strpart(line,p+3:pf-1));
         fitqual.red_chi2 = cstatis / fitqual.dof;
      } else {
         fitqual.red_chi2 = 0.0;
         fitqual.dof = 0;
      }
      fitqual.cstat = int(1);
   } else { // Genuine Chi-squared
      idx3a = where(strmatch(lines(idx2a+2:0),"Reduced chi-squared"));
      if( numberof(idx3a) > 0 ) {
         line = lines(idx3a(1)+idx2a+1);
         p = strpos(line,"squared")+9;
         pf = strpos(line,"for",p);
         pd = strpos(line,"degrees",pf);
         fitqual.red_chi2 = atof(strpart(line,p:pf-1));
         fitqual.dof = atoi(strpart(line,pf+3:pd-1));
      } else {
         fitqual.red_chi2 = 0.0;
         fitqual.dof = 0;
      }
      fitqual.cstat = int(0);
   }

   // Locate 'Model Flux' following the parameter list
   // which means that the line number must be higher than idx2a

   flux = s_Flux();
   idx3a = where(strmatch(lines(idx2a+2:0),"Model Flux"));
   if( numberof(idx3a) > 0 ) {
      line = lines(idx3a(1)+idx2a+1);
      p  = strpos(line,"("); // look for '('
      pf = strpos(line,"erg",p); // look for 'erg'
      flux.flux = atof( strpart(line,p+1:pf-1) ); // then number is found in between
      pd = strpos(line,"range",pf);
      p  = strpos(line,"(",pd); // look for '(' after 'range'
      pf = strpos(line,"-",p); // look for '-' after '('
      flux.elims(1) = atof(strpart(line,p+1:pf-1)); // number is in between
      p  = strpos(line,")",pf); // look for ')' after '-'
      flux.elims(2) = atof(strpart(line,pf+1:p-1)); // number is in between
   } else {
      flux.flux = 0.0;
      flux.elims = [-1.,-1.];
   }

   res = array( s_Parval, nparams+ngparams );

   for( i = 1; i <= nparams+ngparams; i++ ) {
      if( i <= nparams ) {
         line = lines(idx1a + i - 1);
         ot = 1;
      } else {
         line = lines(idx1b + i - 1 - nparams);
         ot = 0;
      }
      res(i) = _parse_par_line( line );
      res(i).model = mdescr;

      if( am_subroutine() ) {
         if( res(i).status == "active" ) {
            if(!silent)write,format="%6i%6i%10s%10s%10s%14.6f%14.6f\n", i, res(i).model_num, \
               res(i).component, res(i).name, res(i).unit, res(i).value, res(i).value_err;
            if( flag && !table ) { // 'flag' is for outfile and 'lun' is the file pointer
               write,lun,format="%6i%6i%10s%10s%10s%14.6f%14.6f\n", i, res(i).model_num, \
                  res(i).component, res(i).name, res(i).unit, res(i).value, res(i).value_err;
            }
         } else {
            if(!silent)write,format="%6i%6i%10s%10s%10s%14.6f%10s\n", i, res(i).model_num, \
               res(i).component, res(i).name, res(i).unit, res(i).value, res(i).status;
            if( flag && !table ) { // 'flag' is for outfile and 'lun' is the file pointer
               write,lun,format="%6i%6i%10s%10s%10s%14.6f%10s\n", i, res(i).model_num, \
                  res(i).component, res(i).name, res(i).unit, res(i).value, res(i).status;
            }
         }
      }
   }
   if( flag ) { // if 'an outfile has been defined with lun as file pointer'
      if( table ) {
         for( i = 1; i <= nparams; i++ ) {
            write,lun,format=" %10.4f +-%10.4f", res(i).value, res(i).value_err;
         }
         write,lun,format=" %s\n", logfile;
      }
      close, lun;
   }
   return res;
}

gx = get_xspec_fit_params;
write,"Now you have shorthand: gx  for get_xspec_fit_params";

func _parse_par_line( line )
{
   s = strsplit( line );
   r = s_Parval();
   if( s(1) == "#" ) s = s(2:);
   ns = numberof(s);
   if( ns < 8 ) s = grow(s(1:4),"",s(5:));
   ns = numberof(s);
   r.param_num = atoi(s(1));
   r.model_num = atoi(s(2));
   r.component = s(3);
   r.name = s(4);
   r.unit = s(5);
   r.value = atof(s(6));
   if( s(7) == "+/-" ) {
      r.status = "active";
      r.value_err = atof(s(8));
   } else if( s(7) == "frozen" ) {
      r.status = "frozen";
      r.value_err = 0.0;
   } else write,"Unexpected +/- or frozen indicator";
   return r;
}

func _rm_brack( line )
{
   cline = *pointer(line);
   v = where(cline == '<');
   nv = numberof(v);
   w = where(cline == '>');
   if( nv != numberof(w) ) {
      write,"_rm_brack  mismatching number of <>";
      return line;
   }
   m = array(long,2*nv);
   m(1::2) = v;
   m(2::2) = w;
   if( anyof( m(dif) <= 0 ) ) {
      write,"_rm_brack  <> appear in wrong order";
      return line;
   }
   idx = [];
   for( i = 1; i <= nv; i++ ) {
      grow, idx, indgen(v(i):w(i));
   }
   ccline = rem_elem(cline, idx);
   newline = string(&ccline);
   return newline;
}

// ==============================================================================================

/* Function comb_spectra */

func comb_spectra( list, oper, outfilename, respfile=, silent=, src_num= )
/* DOCUMENT comb_spectra, list_of_filenames, oper[, outfilename], respfile=, silent=, src_num=
    or  res = comb_spectra( list_of_filenames, oper[, outfilename], respfile=, silent=, src_num= )

    Evaluates the sum of two or more spectra.

    First version (subroutine) will write to 'outfilename' if defined.
    Second version (function) will return an instance of struct 's_Spectrum'
            and write to 'outfilename' if defined.

    Arguments
      list_of_filenames   A spectrum will be read in from each by 'read_spectrum'
                   oper   Operation: "sum" (or "add") or "avg"
            outfilename   (Optional string) when given then the resulting spectrum will be writte to this file

    Keywords
      respfile      RMF file that will be used for all spectra if given
        silent      (bool) suppress screen messages
       src_num      source number for the PHAII spectral files
                    Ignored for PHAI spectra

     2014-09-17/NJW

*/
{

   // If 'respfile' is given then it must be used whatever the
   // respfile is in the individual files (the other approach, namely that
   // the respfile by keyword is only used for the files where it is missing.

   if( typeof(oper) != "string" ) error,"Second argument is invalid";
   oper = strlowcase(oper);
   if( oper == "add" || oper == "sum" ) {
      oper = 1;
      kwds_set,"COMMENT", "Summing the spectra";
   } else if( oper == "avg" ) {
      oper = 2;
      kwds_set,"COMMENT", "Averaging the spectra";
   } else error,"Invalid string in second argument: "+oper;

   if( is_void(respfile) ) {
      use_kwfile = 0;
   } else {
      use_kwfile = 1;
      kwfile = respfile;
   }
   n = numberof(list);
   if( is_void(src_num) ) src_num = 1;

   rate_sum = [];
   rate_sum_err2 = [];
   expo_sum = 0.0;
   arf_sum = [];
   for( i = 1; i <= n; i++ ) {
      write,"Extracting spectrum from "+list(i)+" ...";
      if( !file_test(list(i)) ) {
         write,"Did not find "+list(i)+", skip and continue ...";
         continue;
      }
      phtype = phatype( list(i), silent=silent );
      rs = read_spectrum( list(i), adhoc=1, num=src_num, respfile=respfile, silent=silent );
      exposure = rs.exposure;
      if( is_void(exposure) ) exposure = 1.0;
      write,"           with EXPOSURE = ",exposure;

      ancrfilex = rs.arf;
      respfile = use_kwfile ? kwfile : rs.rmf;
      // remove the {1} on the ancrfile name
      get_exten_no, ancrfilex, ancrfile, ndum, tub=1;

      if( is_void(ancrfile) || strlen(ancrfile) == 0 ) {
         write,"ARF (or ancrfile) is not defined for this spectrum";
      } else {
         if( !file_test(ancrfile) ) {
            write,"Did not find "+ancrfile+", skip and continue ...";
            continue;
         } else {
            arf = rdfitscol( ancrfile+"[SPECRESP]","specresp");
            elo = rdfitscol( ancrfile+"[SPECRESP]","energ_lo");
            ehi = rdfitscol( ancrfile+"[SPECRESP]","energ_hi");
            if( phtype == 2 ) {
               arf = arf(,src_num);
               elo = elo(,src_num);
               ehi = ehi(,src_num);
            }
            if( is_void(arf_sum) ) {
               arf_sum = arf*exposure;
            } else {
               arf_sum += arf*exposure;
            }
         }
      }

      rate = *rs.rate;
      if( oper == 2 ) rate *= exposure;
      if( is_void(rate_sum) ) rate_sum = rate;
      else rate_sum += rate;
      rate_err = *rs.stat_err;
      if( oper == 2 ) rate_err *= sqrt(exposure);
      if( is_void(rate_sum_err2) ) rate_sum_err2 = rate_err^2;
      else rate_sum_err2 += rate_err^2;

      if( !file_test(respfile) ) {
         write,"Did not find response file: "+respfile;
         return;
      }

      kwds_set,"COMMENT",list(i);
      expo_sum += exposure;

   }

   if( typeof(outfilename) == "string" ) {
      if( !is_void(arf_sum) ) {
         arf = arf_sum / expo_sum;
         p = strpos( outfilename, ".", rev=1 );
         l = strlen( outfilename );
         ancrfile = strpart(outfilename,1:p-l-1)+".arf";
         wrmfitscols,ancrfile,"ARF_NUM", [1], "ENERG_LO", elo(,-), "ENERG_HI", ehi(,-), \
                     "SPECRESP", arf, extname="SPECRESP",clobber=1;
      }

      div = oper == 2 ? expo_sum : 1.00;
      spec2phaii, outfilename, rate_sum/div, sqrt(rate_sum_err2/div), exposure=expo_sum, \
                  ancrfile=ancrfile, respfile=respfile, no_kwds_init=1;
      write,"Has produced: "+outfilename;
   } else write,"No spectral file written.";

   if( !am_subroutine() ) {
      res = s_Spectrum();
      res.n = numberof(rate_sum);
      res.rate = &rate_sum;
      res.stat_err = &(sqrt(rate_sum_err2));
      res.e_min = rs.e_min;
      res.e_max = rs.e_max;
      res.energ_lo = rs.energ_lo;
      res.energ_hi = rs.energ_hi;
      res.exposure = expo_sum;
      res.rmf = respfile;

      return res;
   }

}

/* Function sub_spectra */

func sub_spectra( specfile1, specfile2, outfile, respfile=, silent=, src_num1=, src_num2=, \
                  clobber= )
/* DOCUMENT sub_spectra, specfile1, specfile2, outfile
    or  res = sub_spectra( specfile1, specfile2[, outfile] )

    Evaluates the difference spectrum1 - spectrum2 if they have the same number of channels.
    Otherwise it does not work.

    First version (subroutine) will write to 'outfile' if defined.
    Second version (function) will return an instance of struct 's_Spectrum'
            and write to 'outfile' if defined.

    Arguments
      specfile1   A FITS file with a spectrum
      specfile2   A FITS file with a spectrum to subtract from the first one
        outfile   (Optional string) when given then the resulting spectrum will be written to this file
                  with same PHAI or PHAII as 'specfile1'

    Keywords
      respfile      RMF file that will be written to the outfile
        silent      (bool) suppress screen messages
      src_num1      source number if 'specfile1' is a PHAII type file (defaults to 1)
      src_num2      source number if 'specfile2' is a PHAII type file (defaults to 1)
                    Both are ignored for PHAI spectra

     2015-02-27/NJW

*/
{

   if( !file_test(specfile1) ) {
      write,"Did not find "+specfile1;
      return [];
   }
   if( !file_test(specfile2) ) {
      write,"Did not find "+specfile2;
      return [];
   }

   if( !(phat = phatype( specfile1 )) ) {
      write, specfile1+" is not a spectral file";
      return [];
   }
   if( !phatype( specfile2 ) ) {
      write, specfile2+" is not a spectral file";
      return [];
   }
   num = is_void(src_num1) ? 1 : src_num1;
   s1 = read_spectrum( specfile1, num=num, adhoc=1 );
   num = is_void(src_num2) ? 1 : src_num2;
   s2 = read_spectrum( specfile2, num=num, adhoc=1 );

   if( s1.n != s2.n ) {
      write,"Found "+itoa(s1.n)+" channels in "+specfile1;
      write,"  but "+itoa(s2.n)+" channels in "+specfile2;
      return [];
   }

   if( is_void(respfile) ) respfile = s1.rmf;

   rate1 = *s1.rate;
   rate2 = *s2.rate;
   rate_err1 = *s1.stat_err;
   rate_err2 = *s2.stat_err;

   //  ***************** Gaussian error propagation  *****************

   net_rate = rate1 - rate2;
   net_rate_err = sqrt(rate_err1^2 + rate_err2^2);



   if( typeof(outfile) == "string" ) {
      if( phat == 1 ) {
         spec2phai, outfile, net_rate, net_rate_err, exposure=s1.exposure, \
                     ancrfile=s1.arf, respfile=respfile, no_kwds_init=1;
         write,"Has produced: "+outfile+" as PHAI";
      } else {
         spec2phaii, outfile, net_rate, net_rate_err, exposure=s1.exposure, \
                     ancrfile=s1.arf, respfile=respfile, no_kwds_init=1;
         write,"Has produced: "+outfile+" as PHAII";
      }
   } else write,"No spectral file written.";

   if( !am_subroutine() ) {
      res = s_Spectrum();
      res.n = numberof(net_rate);
      res.rate = &net_rate;
      res.stat_err = &net_rate_err;
      res.e_min = s1.e_min;
      res.e_max = s1.e_max;
      res.energ_lo = s1.energ_lo;
      res.energ_hi = s1.energ_hi;
      res.exposure = s1.exposure;
      res.backfile = s1.backfile;
      res.backscal = s1.backscal;
      if(  typeof(outfile) == "string" ) res.specdol = outfile+"+1";
      res.rmf = respfile;
      res.arf = s1.arf;

      return res;
   }

}

/* Function ds */

func dsinit( delay= )
{
   if( !is_void(delay) ) pause, long(1000*delay);
   system,"dsinit";
}

func dsexit( void )
{ system,"dsexit"; }

func dsreg( regfilename )
{ system,"dsreg "+regfilename; }

func dsplay( filename, extno=, opt=, delay= )
{
   if( is_void(extno) ) extno = 1;
   if( is_void(opt) ) opt = "a";
   if( !is_void(delay) ) pause, long(1000*delay);
   if( extno ) {
      system,"dsplay \""+filename+"\["+itoa(extno)+"\]\" "+opt;
   } else system, "dsplay0 "+filename;
}
