/* Function jemx */

extern jemxdoc;
/* DOCUMENT jemx package 2.6

    Specific JEMX related Yorick functions

     _read_catalog                 j_get_scz                     
     best_skypos                   j_get_sloc_res_sources        
     bti_info                      j_get_src_spectrum            
     build_swid_time_table         j_get_srcl_res_sources        
     cat_compare                   j_get_srcl_res_sources_a      
     cbj_image_name                j_get_srcl_res_spectra        
     crab_count_rate               j_get_srcl_res_spectrum       
     dattim2ijd                    j_get_used_gain               
     find_src_by_name              j_mk_vignet                   
     find_swid_galac               j_src_on_shadowgram           
     find_swid_radec               j_std_rebin                   
     find_swid_slice               j_thruput                     
     get_exposure_for_swid         jconv_axes2posang             
     get_gain_for_swid             jconv_posang2axes             
     get_pointing_for_swid         jmx1_offaxis_azim             
     get_revol_no                  misalign                      
     get_revol_start_time
     get_time_for_swid             mk_swid_lists                 
     gti_print                     mk_unique_cat                 
     ijd2dattim                    nearby_sources                
     j_add_spectra                 nearest_source                
     j_collect_spectra             pointing_plot                 
     j_comb_dsp_spectra            qidsrcs                       
     j_cor2raw                     read_catalog                  
     j_gain_values                 srcl2txt                      
     j_get_arc_events              timeconv                      
     j_get_events                  upd_ancrfile                  
     j_get_pi_ebds                 validate_swid_list            
     j_get_point_info_for_swid     sunposition
     ijd2utc                       ijd2tt
     utc2ijd                       tt2ijd

************************************************************/

write,"You are using jemx.i-2.6";

//+ #include "kfits.i"
//+ #include "mfits.i"
//+ #include "idlx.i"
//+ #include "basic.i"

/* Function crab_count_rate */

func crab_count_rate( ener1, ener2, gain=, arf=, extname= )
/* DOCUMENT rate = crab_count_rate( ener1, ener2, gain=, arf=, extname= )

    where the energies are given in keV.

   Keywords:
       gain      The gain should be given in PHA/keV - the default value is 20.0
        arf      ARF filename - if given then this ARF is used.
    extname      To be supplied if the extension name is not "SPECRESP"

    Only one keyword can be given.

   2007-06-14/NJW, 2015-08-05/NJW updated with the 'arf' keyword

*/
{

   eb = spanl( ener1, ener2, 301 );
   de = eb(2:0) - eb(1:-1);
   e = eb(1:-1);

   if( typeof(arf) == "string" ) {
      local e_arf, a_arf;
      specresp = rd_arf( arf, e_arf, a_arf, extname=extname );
      q_area = interp( a_arf, e_arf, e );
   } else {
      if( is_void(gain) ) gain = 20.0;
      basdir = get_env("IC_BASE")+"/imod_grp";
      imodfiles = file_search("jmx1_imod_grp_????.fits", basdir );
      imodfiles = imodfiles(sort(imodfiles));
      imodfile = imodfiles(0); // Choose latest

      //   Get the quantum efficiency

      eqeff = rdfitscol(imodfile+"[JMX1-QEFF-MOD]","ENERGY");
      qeff  = rdfitscol(imodfile+"[JMX1-QEFF-MOD]","QUANTEFF");

      //   Get the thermal foil transmission

      efoil = rdfitscol(imodfile+"[JMX1-FOIL-MOD]","ENERGY");
      foil  = rdfitscol(imodfile+"[JMX1-FOIL-MOD]","TRANSCOEF");

      //   Get the electronic efficiency

      peeff = rdfitscol(imodfile+"[JMX1-EEFF-MOD]","LINPHA");
      eeff  = rdfitscol(imodfile+"[JMX1-EEFF-MOD]","ELEC_EFF");

      q_area = 80. * interp( qeff, eqeff, e ) * interp( foil, efoil, e ) \
                   * interp( eeff, peeff, e*gain );
   }

   flux = 9.7 * e^(-2.1); // photons/cm2/s/keV

   return sum(flux*q_area*de);

}

/* Function get_gain_for_swid */

func get_gain_for_swid( jemxNum, swid )
/* DOCUMENT gain = get_gain_for_swid( jemxNum, swid )

   looks up the gain value from /r6/jemx/pointings/pointings_RRRR.fits

   2008-08-13/NJW
*/
{
   if( jemxNum != 1 && jemxNum != 2 ) {
      write,"GET_GAIN_FOR_SWID: Bad JEMX instrument number";
      return [];
   }
   colname = jemxNum == 1 ? "gain_j1" : "gain_j2";

   if( typeof(swid) != "string" ) {
      write,"GET_GAIN_FOR_SWID: Bad SWID indication";
      return [];
   }

   nswids = numberof(swid);
   len = strlen(swid) - 12;
   if( anyof(len) ) {
      write,"GET_GAIN_FOR_SWID: Bad SWID string length";
      return [];
   }

   if( nswids > 1 ) {
      isort = sort(swid);
      swid = swid(isort);
   }

   revol = strpart(swid,1:4);
   u = uniq(revol);
   nu = numberof(u);
   up = nswids == 1 ? u : shift(u-1,1);
   up(0) = nswids;

   res = array(double, nswids );

   for(i = 1; i <= nu; i++ ) {
      rev = revol(u(i));
      filename = "pointings_"+rev+".fits";
      pswids = rdfitscol("/r6/jemx/pointings/"+filename+"+1", "swid");
      gain = rdfitscol("/r6/jemx/pointings/"+filename+"+1", colname);
      for(j = u(i); j <= up(i); j++ ) {
         w = where( swid(j) == pswids );
         if( numberof(w) != 1 ) {
            write,format="Failed to find gain for SWID: %s\n", swid(j);
            gainx = -1.;
         } else {
            gainx = gain(w(1));
         }
         res(j) = gainx;
      }
   }

   if( nswids > 1 ) {
 // Returns values in same order as the input
      return res(sort(isort));
   } else {
      return res(1);
   }

}

/* Function get_exposure_for_swid */

func get_exposure_for_swid( jemxNum, swid )
/* DOCUMENT exposure = get_exposure_for_swid( jemxNum, swid )

   looks up the exposure value from /r6/jemx/pointings/pointings_RRRR.fits

   2009-11-23/NJW, cloned from get_gain_for_swid
*/
{
   if( jemxNum != 1 && jemxNum != 2 ) {
      write,"GET_EXPOSURE_FOR_SWID: Bad JEMX instrument number";
      return [];
   }
   colname = jemxNum == 1 ? "exposure_j1" : "exposure_j2";

   if( typeof(swid) != "string" ) {
      write,"GET_EXPOSURE_FOR_SWID: Bad SWID indication";
      return [];
   }

   nswids = numberof(swid);
   len = strlen(swid) - 12;
   if( anyof(len) ) {
      write,"GET_EXPOSURE_FOR_SWID: Bad SWID string length";
      return [];
   }

   if( nswids > 1 ) {
      isort = sort(swid);
      swid = swid(isort);
   }

   revol = strpart(swid,1:4);
   u = uniq(revol);
   nu = numberof(u);
   up = nswids == 1 ? u : shift(u-1,1);
   up(0) = nswids;

   res = array(double, nswids );

   for(i = 1; i <= nu; i++ ) {
      rev = revol(u(i));
      filename = "pointings_"+rev+".fits";
      pswids = rdfitscol("/r6/jemx/pointings/"+filename+"+1", "swid");
      exposure = rdfitscol("/r6/jemx/pointings/"+filename+"+1", colname);
      for(j = u(i); j <= up(i); j++ ) {
         w = where( swid(j) == pswids );
         if( numberof(w) != 1 ) {
            write,format="Failed to find exposure for SWID: %s\n", swid(j);
            exposurex = -1.;
         } else {
            exposurex = exposure(w(1));
         }
         res(j) = exposurex;
      }
   }

   if( nswids > 1 ) {
 // Returns values in same order as the input
      return res(sort(isort));
   } else {
      return res(1);
   }

}

/* Function find_swid_radec */

func find_swid_radec( ra0, dec0, rad1, .., offax=, rev=, tstart=, \
                      tstop=, list=, nof=, silent=, lst=, gal= )
/* DOCUMENT list = find_swid_radec( ra0, dec0, rad1[, rad2][, offax=]
                       [,rev=][, tstart=][, tstop=][, list=]
                       [,nof=][, silent=][, lst=][, gal=] )

   Find the INTEGRAL SWIDs where the pointing is less than 'rad1'
   away from (ra0,dec0) or - if 'rad2' is greater than rad1 - greater than 'rad1'
   and less than 'rad2' away

    Ca. 2003-05/NJW
    2006-10-02/NJW  translated to Yorick language.
    2007-11-16/NJW  updated with 'gal' keyword.
    2018-03-20/NJW  updated to read '/r6/jemx/jemx_history/overview_revols.txt'
                    to eliminate messages about missing data when this is
                    already known.

     If rad2 (in degrees) is not given or less than rad1 then the
     pointings with distance < rad1 (in degrees) will be presented
     else the pointings with distance rad1 < dist < rad2
     will be presented.

   Keywords:
             offax:  When set the result will be sorted
                     by increasing off-axis angle
             rev:    1) Integer: Either scalar integer or 2 element integer
                        array with min and max revolution number
                     2) String: example "100-120,200-201,900"
                     3) Void: All pointing files as 'pointings_RRRRp.dat'
                        in the standard directory '$J_POINTINGS' are taken
                        as revolution numbers.
             tstart: Time in IJD for start of interval
             tstop:  Time in IJD for end of interval
             list :  Makes returned list a list of SWIDs (default)
             lst  :  Makes returned list a list of SWIDs in jemx.lst format
                         (overrides setting of 'list')
             nof  :  Flag to avoid file writing
             gal  :  Flag to interpret input coordinates as galactic
*/
{

   local swid, ra, dec, posangle, ut, ijd, telapse, mode1, mode2;

   if( is_void(rad1) ) {
      write,"Syntax: list = find_swid_radec(ra,dec,rad1[,rad2], keywords... )";
      return [];
   }
   //
   //  Initialize
   //
   if( is_void(silent) ) silent = 0;
   first = 1;
   tbegin = is_void( tstart ) ? 0.0 : tstart;
   // if( tstart ) { tbegin = tstart; } else { tbegin = 0.0; }
   tend = is_void( tstop ) ? 1.0e9 : tstop;
   // if( tstop ) { tend = tstop; } else { tend = 1.0e9; }
   ra0 = double(ra0);
   dec0 = double(dec0);
   rad1 = double(rad1);
   if( gal ) {
      lon0 = ra0;
      lat0 = dec0;
      radec0 = equatorial( ra0, dec0 );
      ra0 = radec0(1);
      dec0 = radec0(2);
   } else {
      galac0 = galactic( ra0, dec0 );
      lon0 = galac0(1);
      lat0 = galac0(2);
   }

   rad2 = 0.0;
   if( more_args() ) {
      rad2 = double(next_arg());
   }
   //
   //   Load data about missing event files
   //
   mis_revols = rscol( "/r6/jemx/jemx_history/overview_revols.txt",\
                     1, nt=1, silent=1, block=2);
   //
   // Get list of pointing files
   //
   base = get_env("J_POINTINGS");

   if( numberof(rev) ) {
      if( typeof(rev) == "string" ) {
         revarr = str2arr( rev );
      } else {
         if( numberof( rev ) == 2 ) {
            revarr = indgen(long(rev(2)-rev(1))+1) + long(rev(1)) - 1;
         } else {
            revarr = long(rev)
         }
      }
      // check against missing data
      take_out = whereany( revarr, mis_revols );
      if( numberof(take_out) ) revarr = rem_elem( revarr, take_out );
      // -------
      nrev = numberof(revarr);
      for( i = 1; i <= nrev; i++ ) {
         revstr = swrite(revarr(i),format="%4.4i");
         tmp = base+"/pointings_"+revstr+"p.dat";
         if( !silent ) write,tmp;
         ybase = base+"/ysav"+swrite(revarr(i)/100,format="%02i/");
         ytmp = ybase+"pointings_"+revstr+"p.ysav";
         if( !silent ) write,ytmp;
         if( i == 1 ) {
            pf_list = tmp;
            ypf_list = ytmp;
         } else {
            grow,pf_list,tmp;
            grow,ypf_list,ytmp;
         }
      }
   } else {
      pf_list = file_search("pointings_*p.dat", base);
      npf_list = numberof(pf_list);
      // check against missing data
      revarr = array( int, npf_list );
      p = strpos(pf_list,"p.dat");
      for( i = 1; i <= npf_list; i++ ) revarr(i) = atoi( strpart(pf_list(i), p(i)-4:p(i)-1 ) );
      take_out = whereany( revarr, mis_revols );
      if( numberof(take_out) ) pf_list = rem_elem( pf_list, take_out );
      // --------
      ypf_list = pf_list;
      npf_list = numberof(pf_list);
      for( i = 1; i <= npf_list; i++ ) {
         pos = strpos( pf_list(i),"_" );
         revstra = strpart(pf_list(i), pos+1:pos+4 );
         drevstra = strpart(pf_list(i),pos+1:pos+2);
         ybase = base+"/ysav"+drevstra;
         ytmp = ybase+"/pointings_"+revstra+"p.ysav";
         ypf_list(i) = ytmp;
      }
   }
   npf_list = numberof(pf_list);
   if( npf_list == 0 ) {
      if(!silent) write,"find_swid_radec error: no pointing files of required kind found";
      return [];
   }

   if( !nof ) {

      // Open output file

      outfilename = get_next_filename("find_swid_radec_???.dat");
      fout = open(outfilename,"w");
      n = write(fout,format="%s\n","// Results from find_swid_radec");
      n = write(fout,format="// created %s\n", ndate(3));
      n = write(fout,format="// ra0  = %7.3f; deg\n", ra0);
      n = write(fout,format="// dec0 = %7.3f; deg\n", dec0);
      n = write(fout,format="// lon0 = %7.3f; deg\n", lon0);
      n = write(fout,format="// lat0 = %7.3f; deg\n", lat0);
      if( rad2 < rad1 ) {
         n = write(fout,format="// radius = %7.3f; deg\n", rad1);
      } else {
         n = write(fout,format="%s\n","// selection in annulus");
         n = write(fout,format="// rad1 = %7.3f; deg\n", rad1);
         n = write(fout,format="// rad2 = %7.3f; deg\n", rad2);
      }
      if( tstart ) {
         n = write(fout,format="// tstart = %14.8f; IJD\n", tstart);
      }
      if( tstop ) {
         n = write(fout,format="// tstop = %14.8f; IJD\n", tstop);
      }
      //+ if( numberof(rev) ) {
      //+    if( nrev == 1 ) {
      //+       n = write(fout,format="// rev_begin = %4i\n", revarr(1));
      //+       n = write(fout,format="// rev_end   = %4i\n", revarr(1));
      //+    } else {
      //+       n = write(fout,format="// rev_begin = %4i\n", revarr(1));
      //+       n = write(fout,format="// rev_end   = %4i\n", revarr(0));
      //+    }
      //+ }

   }

   ntotal = 0;
   for( ipf_list = 1; ipf_list <= npf_list; ipf_list++ ) {

      if( file_test(ypf_list(ipf_list)) ) {
         if( !silent ) write,"Reading binary file "+ypf_list(ipf_list);
         bfile = openb( ypf_list(ipf_list) );
         restore,bfile, swid, ra, dec, posangle, ut, ijd, telapse, \
                     mode1, mode2;
         close, bfile;
      } else {
         if( !silent ) write,"Reading usual text file "+pf_list(ipf_list);
         if( ! file_test(pf_list(ipf_list)) ) {
            if( !silent ) write,pf_list(ipf_list)+" was not found";
            continue;
         }
      /*
       * swid = rscol( pf_list(ipf_list),1, str=1, silent=1);
       * ra = rscol( pf_list(ipf_list), 2, silent=1);
       * dec = rscol( pf_list(ipf_list), 3, silent=1);
       * posangle = rscol( pf_list(ipf_list), 4, silent=1);
       * ut = rscol( pf_list(ipf_list), 5, str=1, silent=1);
       * ijd = rscol( pf_list(ipf_list), 6, silent=1);
       * telapse = rscol( pf_list(ipf_list),7, silent=1);
       * mode1 = rscol( pf_list(ipf_list), 8, str=1, silent=1);
       * mode2 = rscol( pf_list(ipf_list), 9, str=1, silent=1);
       */

         rstab,pf_list(ipf_list),9,swid,ra,dec,posangle,ut,ijd,telapse, \
              mode1, mode2,typ="sfffsffss",silent=1;

        /*
         * Prepare for the next time you need these data. Crate a Yorick
         * binary save file for much faster reading. Create the subdirectory
         * if not already existing.
         */

         subdir = pathname( ypf_list(ipf_list) );
         if( !file_test( subdir ) ) mkdir, subdir;
         system,"chmod g+w "+subdir; // Update the write permission

         bfile = createb( ypf_list(ipf_list) );
         save, bfile,  swid, ra, dec, posangle, ut, ijd, telapse, \
                 mode1, mode2;
         close, bfile;
         //  Make sure other users in your group get write permission
         system,"chmod g+w "+ypf_list(ipf_list);
      }

      dist = arcdist(ra0, dec0, ra, dec);

      if( abs(rad2) < 1.e-10 ) {

         w = where( dist < rad1);
         nw = numberof(w);

      } else {

         w = where( dist > rad1 & dist < rad2);
         nw = numberof(w);

      }

      if( nw > 0 ) {

         if( first ) {
            first = 0;
            swid_all = swid(w);
            ra_all = ra(w);
            dec_all = dec(w);
            posangle_all = posangle(w);
            ut_all = ut(w);
            ijd_all = ijd(w);
            telapse_all = telapse(w);
            mode1_all = mode1(w);
            mode2_all = mode2(w);
            dist_all = dist(w);
         } else {
            grow,swid_all,swid(w);
            grow,ra_all,ra(w);
            grow,dec_all,dec(w);
            grow,posangle_all,posangle(w);
            grow,ut_all,ut(w);
            grow,ijd_all,ijd(w);
            grow,telapse_all,telapse(w);
            grow,mode1_all,mode1(w);
            grow,mode2_all,mode2(w);
            grow,dist_all,dist(w);
         }

         ntotal += nw;

      }

   }
   if( ntotal == 0 ) {
      if(!silent) write,"Sorry, no pointings found!";
      return [];
   }

   if( offax ) {
      is = sort(dist_all);
      swid_all = swid_all(is);
      ra_all = ra_all(is);
      dec_all = dec_all(is);
      posangle_all = posangle_all(is);
      ut_all = ut_all(is);
      ijd_all = ijd_all(is);
      telapse_all = telapse_all(is);
      mode1_all = mode1_all(is);
      mode2_all = mode2_all(is);
      dist_all = dist_all(is);
   }

   noutput = 0;
   list = "";
   list_list = [];
   expo_tot = 0.0;
   for( i = 1; i <= ntotal; i++ ) {
      // apply the time selection
      if( ijd_all(i) > tbegin && ijd_all(i) < tend ) {
         //+ eff_factor = 1. - dist_all(i)/6.4;
         eff_factor = 1. - dist_all(i)/5.0;
         expo_contrib = eff_factor > 0.0 ? telapse_all(i) * eff_factor : 0.0;
         expo_tot += expo_contrib;
         grow,list_list,swid_all(i);
         if( !nof) {
            write,fout,format="%s%10.4f%9.4f%9.3f  %s%10.3f%6.0f  %s  %s %8.2f %7.0f\n",
            swid_all(i),ra_all(i),dec_all(i),posangle_all(i),
            ut_all(i),ijd_all(i),telapse_all(i),
            mode1_all(i),mode2_all(i),dist_all(i), expo_contrib;
         }
         noutput++;
      }
   }

   lst_list = array("",ntotal);
   for( i = 1; i <= ntotal; i++ ) {
      lst_list(i) = "./scw/"+strpart(swid_all(i),1:4)+"/"+swid_all(i)+".001/swg.fits[1]";
   }

   if( !nof ) {
      write,fout,format="//\n// Total vignetting corrected exposure for selected position is %14.5e s\n", \
                expo_tot;
      close,fout;
   }

   if( !silent ) {
      write,format="Found %i SWIDs in selected revolutions\n", ntotal;
      if( noutput == 1 ) {
         write,format="%i has been selected by time\n", noutput;
      } else {
         write,format="%i have been selected by time\n", noutput;
      }
      write,format="Total vignetting corrected exposure for selected position is %14.4e s\n", expo_tot;
      if( !nof ) write,"Output is directed to "+outfilename;

   }

   if( !is_void(lst) ) { return lst_list; } else {return list_list;}

}


/* Function j_get_sloc_res_sources */

func j_get_sloc_res_sources( dol_list, outfile=, comb=, silent= )
/* DOCUMENT j_get_sloc_res_sources, dol_list, outfile=, comb=, silent=

   Accepts a list of sloc_res DOLs and will extract the source information
   Keyword 'outfile' is the output FITS filename (default "resfile.fits").
   Keyword 'comb' will choose between all, combined or exclude combined:
            comb = 0 or not given : include all
            comb = 1   :  include only combined results
            comb = -1  :  exclude combined results

   If dol_list is just a file list then 1. extension will be assumed.

   The SLOC-RES DS is a result from j_ima_src_locator

   2006-11-13/NJW  Cloned from j_get_srcl_res_sources

*/
{

   if( is_void(outfile) ) outfile = "resfile.fits";
   chat = is_void(silent);
   if( is_void(comb) ) comb = 0;
   if( comb > 0 ) comb = 1;
   if( comb < 0 ) comb = -1;
   kwds_init;
   if( comb == 0 ) {
      kwds_set,"COMMENT","Both individual results and combinations of sources";
   } else if( comb == 1 ) {
      kwds_set,"COMMENT","Only combined source results are included";
   } else {
      kwds_set,"COMMENT","Combined source results are excluded";
   }


   name_arr = [];
   // sourceid_arr = [];
   swid_arr = [];
   instr_arr = [];
   ra_arr = [];
   dec_arr = [];
   err_rad_arr = [];
   ra_cat_arr = [];
   dec_cat_arr = [];
   detsig_arr = [];
   sigma_arr = [];
   kval_arr = [];
   chi2_arr = [];
   // offaxis_arr = [];
   // cosy_arr = [];
   // cosz_arr = [];
   // flux_arr = [];
   peaksize_arr = [];
   e_min_arr = [];
   e_max_arr = [];
   // ra_scx_arr = [];
   // dec_scx_arr = [];
   // posang_arr = [];
   dol_arr = [];
   swid_nosrcs = [];
   tot_nrows = 0;

   nfiles = numberof(dol_list);

   for( i = 1; i <= nfiles; i++ ) {

      dolres = dol_list(i);
      get_exten_no, dolres, fileres, extres;
     /*  apply default extension if filename is given */
      if( extres == 0 ) {
         extres = 1;
         dolres = dolres+"+1";
      }
      hdr = headfits( dolres );
      if( chat ) write,format="dolres: %s\n", dolres;

     /*
      * First try to get SWID from header, then
      * from file name - else give up
      */

      swid = fxpar(hdr,"SWID");
      if( typeof(swid) != "string" ) {
         pick_swid_str,dolres,swid;
         if( typeof( swid ) != "string" ) swid = "000000000000";
      } else if( swid(1) == "String" ) {
         pick_swid_str,dolres,swid;
         if( typeof( swid ) != "string" ) swid = "000000000000";
      }
      swid = swid(1);
     /*
      * First try to get INSTRUME from header, then
      * from file name - else give up
      */
      instrin = fxpar(hdr,"INSTRUME");
      if( typeof(instrin) != "string" ) {
         if( strpos(fileres,"jmx1",1) > 0 ) instrin = "JMX1";
         if( strpos(fileres,"JMX1",1) > 0 ) instrin = "JMX1";
         if( strpos(fileres,"jmx2",1) > 0 ) instrin = "JMX2";
         if( strpos(fileres,"JMX2",1) > 0 ) instrin = "JMX2";
         if( typeof(instrin) != "string" ) instrin = "XXXX";
      } else if( instrin(1) == "String" ) {
         instrin = "XXXX";
      }

      nebin = fxpar(hdr,"EBIN_NUM");
      if( is_void(nebin) ) nebin = 1;
      if( typeof(nebin) == "string" ) nebin = 1;

      rain = rdfitscol(dolres,"RA_OBJ",silent=silent);
      nrows = numberof(rain);
      //tot_nrows += nrows;
      if( chat ) write,format="%i rows found\n", nrows;
      if( nrows >= 1 ) {

         decin = rdfitscol(dolres,"DEC_OBJ");
         err_radin = rdfitscol(dolres,"ERR_RAD");
         ra_catin = rdfitscol(dolres,"RA_CAT");
         if( is_void(ra_catin) ) ra_catin = float(rain*0.0 - 99.0);
         dec_catin = rdfitscol(dolres,"DEC_CAT");
         if( is_void(dec_catin) ) dec_catin = float(decin*0.0 - 99.0);
         detsigin = rdfitscol(dolres,"DETSIG");
         namein = rdfitscol(dolres,"NAME");
         // sourceidin = rdfitscol(dolres,"SOURCE_ID");
         // fluxin = rdfitscol(dolres,"FLUX");
         peaksizein = rdfitscol(dolres,"PEAKSIZE");
         sigmain = rdfitscol(dolres,"SIGMA");
         selw = where( sigmain == 0.0 );
         selm = where( sigmain > 0.0 );
         kvalin = rdfitscol(dolres,"KVAL");
         chi2in = rdfitscol(dolres,"CHI2");
         // cosy = rdfitscol(dolres,"COSY_JMX");
         // cosz = rdfitscol(dolres,"COSZ_JMX");
         // fluxin = transpose(fluxin(,1:nebin));
         e_minin = rdfitscol(dolres,"E_MIN");
         e_maxin = rdfitscol(dolres,"E_MAX");

         if( comb == 1 ) {
            nr = numberof(selw);
            tot_nrows += nr;
            if( nr > 0 ) {
               grow, swid_arr, array(swid, nr);
               grow, instr_arr, array(instrin, nr);
               grow, name_arr, namein(selw);
               grow, ra_arr, rain(selw);
               grow, dec_arr, decin(selw);
               grow, err_rad_arr, err_radin(selw);
               grow, ra_cat_arr, ra_catin(selw);
               grow, dec_cat_arr, dec_catin(selw);
               grow, detsig_arr, detsigin(selw);
               grow, peaksize_arr, peaksizein(selw);
               grow, sigma_arr, sigmain(selw);
               grow, kval_arr, kvalin(selw);
               grow, chi2_arr, chi2in(selw);
               grow, e_min_arr, e_minin(1,selw);
               grow, e_max_arr, e_maxin(1,selw);
               grow, dol_arr, array(dolres, nr);
            }
         } else if( comb == -1 ) {
            nr = numberof(selm);
            tot_nrows += nr;
            if( nr > 0 ) {
               grow, swid_arr, array(swid, nr);
               grow, instr_arr, array(instrin, nr);
               grow, name_arr, namein(selm);
               grow, ra_arr, rain(selm);
               grow, dec_arr, decin(selm);
               grow, err_rad_arr, err_radin(selm);
               grow, ra_cat_arr, ra_catin(selm);
               grow, dec_cat_arr, dec_catin(selm);
               grow, detsig_arr, detsigin(selm);
               grow, peaksize_arr, peaksizein(selm);
               grow, sigma_arr, sigmain(selm);
               grow, kval_arr, kvalin(selm);
               grow, chi2_arr, chi2in(selm);
               grow, e_min_arr, e_minin(1,selm);
               grow, e_max_arr, e_maxin(1,selm);
               grow, dol_arr, array(dolres, nr);
            }
         } else {
            tot_nrows += nrows;
            grow, swid_arr, array(swid, nrows);
            grow, instr_arr, array(instrin, nrows);
            grow, name_arr, namein;
            grow, ra_arr, rain;
            grow, dec_arr, decin;
            grow, err_rad_arr, err_radin;
            grow, ra_cat_arr, ra_catin;
            grow, dec_cat_arr, dec_catin;
            grow, detsig_arr, detsigin;
            grow, peaksize_arr, peaksizein;
            grow, sigma_arr, sigmain;
            grow, kval_arr, kvalin;
            grow, chi2_arr, chi2in;
            grow, e_min_arr, e_minin(1,);
            grow, e_max_arr, e_maxin(1,);
            grow, dol_arr, array(dolres, nrows);
         }

      } else {
         if( chat ) write,format="No sources found in %s\n", dolres;
         grow, swid_nosrcs, swid;
      }
      if( chat ) write,format="----------------------------%s\n","";

   }

   if( tot_nrows > 0 ) {
      wrmfitscols,outfile,"SWID",swid_arr,"NAME",name_arr, \
             "RA_OBJ",ra_arr,"DEC_OBJ",dec_arr, "ERR_RAD", err_rad_arr, \
             "RA_CAT",ra_cat_arr,"DEC_CAT",dec_cat_arr, \
             "DETSIG",detsig_arr, \
             "PEAKSIZE",peaksize_arr,"E_MIN",e_min_arr, \
             "E_MAX",e_max_arr, "SIGMA", sigma_arr, "KVAL", kval_arr, \
             "CHI2", chi2_arr, \
             "INSTRUME",instr_arr,\
             "DOL_SRCL",dol_arr,clobber=1;
      if( chat ) write,format="Information on %i sources is found in: %s\n", tot_nrows, outfile;
   } else {
      if( chat ) write,format="No sources found at all%s\n","";
   }
   nsno = numberof(swid_nosrcs);
   if( nsno > 0 ) {
      wrmfitscols,"not_"+outfile,"SWID",swid_nosrcs,clobber=1;
      if( chat ) write,format="The %i SWIDs for empty fields are found in: %s\n", nsno, "not_"+outfile;
   } else {
      if( chat ) write,format="There were sources in all SLOC-RES files%s\n","";
   }

}




/* Function j_get_srcl_res_sources */

func j_get_srcl_res_sources( dol_list, outfile=, silent= )
/* DOCUMENT j_get_srcl_res_sources, dol_list, outfile=, silent=

   Accepts a list of srcl_res DOLs and will extract the source information
   Keyword 'outfile' is the output FITS filename (default "resfile.fits").
   If dol_list is just a file list then 1. extension will be assumed.

   This version produces a row for each energy interval which means that
   the same source from the same SWID is given several times if there are
   more than one energy interval.
   The advantage is that SRCL-RES files with different number of energy
   intervals can be combined.

   SEE ALSO: j_get_srcl_res_sources_a

   NJW/2004-01-09
   2006-08-16/NJW  Converted to yorick
   2006-08-22/NJW  Updated for better handling of SWID, INSTRUME
   2006-11-13/NJW  Updated for including E_MIN & E_MAX
   2006-04-27/NJW  Updated for proper handling of auxiliary 'not_' file

*/
{

   if( is_void(outfile) ) outfile = "resfile.fits";
   chat = is_void(silent); // chattiness


   name_arr = [];
   sourceid_arr = [];
   swid_arr = [];
   instr_arr = [];
   ra_arr = [];
   dec_arr = [];
   err_rad_arr = [];
   detsig_arr = [];
   offaxis_arr = [];
   cosy_arr = [];
   cosz_arr = [];
   flux_arr = [];
   flux_err_arr = [];
   e_min_arr = [];
   e_max_arr = [];
   ra_scx_arr = [];
   dec_scx_arr = [];
   posang_arr = [];
   dol_arr = [];
   swid_nosrcs = [];
   tot_nrows = 0;

   nfiles = numberof(dol_list);

   for( i = 1; i <= nfiles; i++ ) {

      dolres = dol_list(i);
      get_exten_no, dolres, fileres, extres;
     /*  apply default extension if filename is given */
      if( extres == 0 ) {
         extres = 1;
         dolres = dolres+"+1";
      }
      hdr = headfits( dolres );
      if( chat ) write,format="dolres: %s\n", dolres;

     /*
      * First try to get SWID from header, then
      * from file name - else give up
      */

      swidx = fxpar(hdr,"SWID");
      if( typeof(swidx) != "string" ) {
         pick_swid_str,dolres,swidx;
         if( typeof( swidx ) != "string" ) swidx = "000000000000";
      } else if( swidx(1) == "String" ) {
         pick_swid_str,dolres,swidx;
         if( typeof( swidx ) != "string" ) swidx = "000000000000";
      }
      swidx = swidx(1);
      if( chat ) write,format="swid: %s\n", swidx;
     /*
      * First try to get INSTRUME from header, then
      * from file name - else give up
      */
      instrin = fxpar(hdr,"INSTRUME");
      if( typeof(instrin) != "string" ) {
         if( strpos(fileres,"jmx1",1) > 0 ) instrin = "JMX1";
         if( strpos(fileres,"JMX1",1) > 0 ) instrin = "JMX1";
         if( strpos(fileres,"jmx2",1) > 0 ) instrin = "JMX2";
         if( strpos(fileres,"JMX2",1) > 0 ) instrin = "JMX2";
         if( typeof(instrin) != "string" ) instrin = "XXXX";
      }

      nebin = fxpar(hdr,"EBIN_NUM");
      if( is_void(nebin) || typeof(nebin) == "string" ) nebin = 1;

      rain = rdfitscol(dolres,"RA_OBJ",silent=silent);
      nrows = numberof(rain);
      tot_nrows += nrows;
      if( chat ) write,format="%i rows found\n", nrows;
      if( nrows >= 1 ) {

         decin = rdfitscol(dolres,"DEC_OBJ",silent=silent);

         // get offaxis angle
         status = get_pointing_for_swid(swidx,ra_scx,dec_scx,posang,silent=silent);
         if( status == 0 ) {
            offaxis = arcdist( rain, decin, ra_scx, dec_scx);
         } else {
            write,"Cannot get pointing for swid: "+swidx;
            offaxis = 7.0;
         }
         err_radin = rdfitscol(dolres,"ERR_RAD",silent=silent);
         detsigin = rdfitscol(dolres,"DETSIG",silent=silent);
         namein = rdfitscol(dolres,"NAME",silent=silent);
         sourceidin = rdfitscol(dolres,"SOURCE_ID",silent=silent);
         fluxin = rdfitscol(dolres,"FLUX",silent=silent);
         fluxerrin = rdfitscol(dolres,"FLUX_ERR",silent=silent);
         cosy = rdfitscol(dolres,"COSY_JMX",silent=silent);
         cosz = rdfitscol(dolres,"COSZ_JMX",silent=silent);
         e_minin = rdfitscol(dolres,"E_MIN",silent=silent);
         e_maxin = rdfitscol(dolres,"E_MAX",silent=silent);

         for( iebin = 1; iebin <= nebin; iebin++ ) {
            grow, swid_arr, array(swidx, nrows);
            grow, instr_arr, array(instrin, nrows);
            grow, name_arr, namein;
            grow, sourceid_arr, sourceidin;
            grow, ra_arr, rain;
            grow, dec_arr, decin;
            grow, err_rad_arr, err_radin;
            grow, detsig_arr, detsigin;
            grow, offaxis_arr, offaxis;
            grow, cosy_arr, cosy*100;
            grow, cosz_arr, cosz*100;
            grow, flux_arr, fluxin(iebin,);
            grow, flux_err_arr, fluxerrin(iebin,);
            grow, e_min_arr, e_minin(iebin,);
            grow, e_max_arr, e_maxin(iebin,);
            grow, ra_scx_arr, array(ra_scx, nrows);
            grow, dec_scx_arr, array(dec_scx, nrows);
            grow, posang_arr, array(posang, nrows);
            grow, dol_arr, array(dolres, nrows);
         }

      } else {
         if( chat ) write,format="No sources found in %s\n", dolres;
         if( chat ) write,format="grow swid_nosrcs with swid: %s\n", swidx;
         grow, swid_nosrcs, swidx;
      }
      if( chat ) write,format="----------------------------%s\n","";

   }

   kwds_init;
   kwds_set,"EXTNAME","COMB-SRCL-RES","Name of extension";
   kwds_set,"DATE", ndate(3),"Date/time of creation";
   kwds_set,"ORIGIN","j_get_srcl_res_sources","Yorick program";
   kwds_set,"RESPONSI","Niels J. Westergaard","Responsible for this file";
   kwds_set,"TELESCOP","INTEGRAL","Name of mission";
   kwds_set,"INSTRUME","JEMX","Recorded the data";
   kwds_set,"EBIN_NUM",nebin,"Number of energy bins";

   if( tot_nrows > 0 ) {
      wrmfitscols,outfile,"SWID",swid_arr,"SOURCE_ID",sourceid_arr,"NAME",name_arr, \
             "RA_OBJ",ra_arr,"DEC_OBJ",dec_arr, "ERR_RAD", err_rad_arr, \
             "DETSIG",detsig_arr,"OFFAXIS",offaxis_arr, "COSY_JMX",cosy_arr, \
             "COSZ_JMX",cosz_arr,"FLUX",flux_arr,"FLUX_ERR", flux_err_arr, "E_MIN",e_min_arr, \
             "E_MAX",e_max_arr,"RA_SCX",ra_scx_arr, \
             "DEC_SCX",dec_scx_arr,"ROLL",posang_arr,"INSTRUME",instr_arr,\
             "DOL_SRCL",dol_arr,clobber=1;
      if( chat ) write,format="Information on %i sources is found in: %s\n", tot_nrows, outfile;
   } else {
      if( chat ) write,format="No sources found at all%s\n","";
   }
   nsno = numberof(swid_nosrcs);
   if( nsno > 0 ) {
      splitfname, outfile, dir_outfile, name_outfile;
      wrmfitscols,dir_outfile+"/not_"+name_outfile,"SWID",swid_nosrcs,clobber=1;
      if( chat ) write,format="The %i SWIDs for empty fields are found in: %s\n", nsno, \
            dir_outfile+"/not_"+name_outfile;
   } else {
      if( chat ) write,format="There were sources in all SRCL-RES files%s\n","";
   }

}


/* Function j_get_srcl_res_sources_a */

func j_get_srcl_res_sources_a( dol_list, srcname=, outfile=, silent= )
/* DOCUMENT j_get_srcl_res_sources_a, dol_list, srcname=, outfile=, silent=

   Accepts a list of srcl_res DOLs and will extract the source information
   Keyword 'srcname' is the source name (exact match) to be given if results
   from a specific source are requested.
   Keyword 'outfile' is the output FITS filename (default "resfile.fits").
   If dol_list is just a file list then 1. extension will be assumed.

   This version reproduces one row per source per SWID. The FLUX etc.
   columns are vector columns. It might fail if the number of energy
   intervals is not the same for all SRCL-RES tables given in the list.

   NJW/2004-01-09
   2006-08-16/NJW  Converted to yorick
   2006-08-22/NJW  Updated for better handling of SWID, INSTRUME
   2006-11-13/NJW  Updated for including E_MIN & E_MAX
   2007-04-27/NJW  Changed to give one row per source
   2008-06-24/NJW  Updated to include RA_CAT and DEC_CAT
   2009-07-22/NJW  Updated to include CLASS

*/
{

   if( is_void(outfile) ) outfile = "resfile.fits";
   chat = is_void(silent);


   name_arr = [];
   sourceid_arr = [];
   swid_arr = [];
   instr_arr = [];
   ra_arr = [];
   dec_arr = [];
   racat_arr = [];
   deccat_arr = [];
   class_arr = [];
   err_rad_arr = [];
   detsig_arr = [];
   offaxis_arr = [];
   cosy_arr = [];
   cosz_arr = [];
   flux_arr = [];
   flux_err_arr = [];
   e_min_arr = [];
   e_max_arr = [];
   ra_scx_arr = [];
   dec_scx_arr = [];
   posang_arr = [];
   dol_arr = [];
   swid_nosrcs = [];
   tot_nrows = 0;

   nfiles = numberof(dol_list);

   for( i = 1; i <= nfiles; i++ ) {

      dolres = dol_list(i);
      get_exten_no, dolres, fileres, extres;
     /*  apply default extension if filename is given */
      if( extres == 0 ) {
         extres = 1;
         dolres = dolres+"+1";
      }
      hdr = headfits( dolres );
      if( chat ) write,format="dolres: %s\n", dolres;

     /*
      * First try to get SWID from header, then
      * from file name - else give up
      */

      swidx = fxpar(hdr,"SWID");
      if( typeof(swidx) != "string" ) {
         pick_swid_str,dolres,swidx;
         if( typeof( swidx ) != "string" ) swidx = "000000000000";
      } else if( swidx(1) == "String" ) {
         pick_swid_str,dolres,swidx;
         if( typeof( swidx ) != "string" ) swidx = "000000000000";
      }
      swidx = swidx(1);
      if( chat ) write,format="swid: %s\n", swidx;
     /*
      * First try to get INSTRUME from header, then
      * from file name - else give up
      */
      instrin = fxpar(hdr,"INSTRUME");
      if( typeof(instrin) != "string" ) {
         if( strpos(fileres,"jmx1",1) > 0 ) instrin = "JMX1";
         if( strpos(fileres,"JMX1",1) > 0 ) instrin = "JMX1";
         if( strpos(fileres,"jmx2",1) > 0 ) instrin = "JMX2";
         if( strpos(fileres,"JMX2",1) > 0 ) instrin = "JMX2";
         if( typeof(instrin) != "string" ) instrin = "XXXX";
      }

      nebin = fxpar(hdr,"EBIN_NUM");
      if( is_void(nebin) || typeof(nebin) == "string" ) nebin = 1;

      rain = rdfitscol(dolres,"RA_OBJ",silent=silent);
      nrows = numberof(rain);
      tot_nrows += nrows;
      if( chat ) write,format="%i rows found\n", nrows;
      if( nrows >= 1 ) {

         decin = rdfitscol(dolres,"DEC_OBJ",silent=silent);
         racatin = rdfitscol(dolres,"RA_CAT",silent=silent);
         deccatin = rdfitscol(dolres,"DEC_CAT",silent=silent);

         // get offaxis angle
         offaxis = array(7.0, nrows);
         if( swidx != "000000000000" ) { // try to get the offaxis angle
            status = get_pointing_for_swid(swidx,ra_scx,dec_scx,posang,silent=silent);
            if( status == 0 ) {
               offaxis = arcdist( rain, decin, ra_scx, dec_scx);
            } else {
               write,"Cannot get pointing for swid: "+swidx;
            }
         }
         err_radin = rdfitscol(dolres,"ERR_RAD",silent=silent);
         detsigin = rdfitscol(dolres,"DETSIG",silent=silent);
         namein = rdfitscol(dolres,"NAME",silent=silent);
         classin = rdfitscol(dolres,"CLASS",silent=silent);
         if( is_void(classin) ) classin = array(0n,nrows);
         w = where(strlen(namein)==0);
         nw = numberof(w); if( nw ) namein(w) = "INPUT ERR";
         sourceidin = rdfitscol(dolres,"SOURCE_ID",silent=silent);
         w = where(strlen(sourceidin)==0);
         nw = numberof(w); if( nw ) sourceidin(w) = "INPUT ERR";
         fluxin = rdfitscol(dolres,"FLUX",silent=silent);
         fluxerrin = rdfitscol(dolres,"FLUX_ERR",silent=silent);
         cosy = rdfitscol(dolres,"COSY_JMX",silent=silent);
         cosz = rdfitscol(dolres,"COSZ_JMX",silent=silent);
         e_minin = rdfitscol(dolres,"E_MIN",silent=silent);
         e_maxin = rdfitscol(dolres,"E_MAX",silent=silent);

         if( numberof(srcname) ) {
            w = where( strtrim(srcname) == strtrim(namein) );
            nrows = numberof(w);
            if( nrows ) {
               rain = rain(w);
               decin = decin(w);
               racatin = racatin(w);
               deccatin = deccatin(w);
               offaxis = offaxis(w);
               err_radin = err_radin(w);
               detsigin = detsigin(w);
               namein = namein(w);
               sourceidin = sourceidin(w);
               fluxin = fluxin(,w);
               fluxerrin = fluxerrin(,w);
               cosy = cosy(w);
               cosz = cosz(w);
               e_minin = e_minin(,w);
               e_maxin = e_maxin(,w);
            } else {
               if( chat ) write,format="%s not found in %s\n", dolres;
            }
         }
         if( nrows ) {
            grow, swid_arr, array(swidx, nrows);
            grow, instr_arr, array(instrin, nrows);
            grow, name_arr, namein;
            grow, sourceid_arr, sourceidin;
            grow, ra_arr, rain;
            grow, dec_arr, decin;
            grow, class_arr, classin;
            grow, racat_arr, racatin;
            grow, deccat_arr, deccatin;
            grow, err_rad_arr, err_radin;
            grow, detsig_arr, detsigin;
            grow, offaxis_arr, offaxis;
            grow, cosy_arr, cosy*100;
            grow, cosz_arr, cosz*100;
            grow, flux_arr, fluxin(1:nebin,);
            grow, flux_err_arr, fluxerrin(1:nebin,);
            grow, e_min_arr, e_minin(1:nebin,);
            grow, e_max_arr, e_maxin(1:nebin,);
            grow, ra_scx_arr, array(ra_scx, nrows);
            grow, dec_scx_arr, array(dec_scx, nrows);
            grow, posang_arr, array(posang, nrows);
            grow, dol_arr, array(dolres, nrows);
         }

      } else {

         if( chat ) write,format="No sources found in %s\n", dolres;
         if( chat ) write,format="grow swid_nosrcs with swid: %s\n", swidx;
         grow, swid_nosrcs, swidx;
      }
      if( chat ) write,format="----------------------------%s\n","";

   }

   kwds_init;
   kwds_set,"EXTNAME","COMB-SRCL-RES","Name of extension";
   kwds_set,"DATE", ndate(3),"Date/time of creation";
   kwds_set,"ORIGIN","j_get_srcl_res_sources_a","Yorick program";
   kwds_set,"RESPONSI","Niels J. Westergaard","Responsible for this file";
   kwds_set,"TELESCOP","INTEGRAL","Name of mission";
   kwds_set,"INSTRUME","JEMX","Recorded the data";
   kwds_set,"EBIN_NUM",nebin,"Number of energy bins";
   if( tot_nrows > 0 ) {
      name_arr = strpart(name_arr,1:20);
      name_arr(1) = strpadd(name_arr(1),20," "); // make sure that name has length 20
      sourceid_arr = strpart(sourceid_arr,1:16);
      sourceid_arr(1) = strpadd(sourceid_arr(1),16," "); // make sure that source_id has length 16
      dol_arr = strpart(dol_arr,1:128);
      dol_arr(1) = strpadd(dol_arr(1),128," "); // make sure that dol has length 128
      wrmfitscols,outfile,"SWID",swid_arr,"SOURCE_ID",sourceid_arr,"NAME",name_arr, \
             "CLASS", class_arr, \
             "RA_OBJ", ra_arr,"DEC_OBJ", dec_arr, "ERR_RAD", err_rad_arr, \
             "RA_CAT", racat_arr,"DEC_CAT", deccat_arr, \
             "DETSIG", detsig_arr, "OFFAXIS", offaxis_arr, "COSY_JMX", cosy_arr, \
             "COSZ_JMX",cosz_arr,"FLUX",flux_arr,"FLUX_ERR", flux_err_arr, \
             "E_MIN", e_min_arr, "E_MAX", e_max_arr, "RA_SCX", ra_scx_arr, \
             "DEC_SCX", dec_scx_arr,"ROLL",posang_arr,"INSTRUME",instr_arr,\
             "DOL_SRCL",dol_arr,clobber=1;
      if( chat ) write,format="Information on %i sources is found in: %s\n", tot_nrows, outfile;
   } else {
      if( chat ) write,format="No sources found at all%s\n","";
   }
   nsno = numberof(swid_nosrcs);
   if( nsno > 0 ) {
      splitfname, outfile, dir_outfile, name_outfile;
      wrmfitscols,dir_outfile+"/not_"+name_outfile,"SWID",swid_nosrcs,clobber=1;
      if( chat ) write,format="The %i SWIDs for empty fields are found in: %s\n", nsno, \
             dir_outfile+"/not_"+name_outfile;
   } else {
      if( chat ) write,format="There were sources in all SRCL-RES files%s\n","";
   }

}



/* Function get_pointing_for_swid */

func get_pointing_for_swid( swid_in, &ra_scx, &dec_scx, &posang, silent= )
/* DOCUMENT status = get_pointing_for_swid( swid, >ra_scx, >dec_scx, >posang, silent= )

    Get RA and Dec for an INTEGRAL pointing characterized by the SWID by
    scalar argument 'swid' (of type string)
    from the pointing file 'pointings_RRRRp.dat' in directory given by
    the environment variable 'J_POINTINGS'.

    Returns a value of zero for succesful running (a negative value otherwise)
    and ra_scx, dec_scx, posang in the arguments.

    Arguments:
         swid      Scalar string (otherwise first element is selected)
         ra_scx    (returned) Right Ascension (deg) of INTEGRAL pointing
         dec_scx   (returned) Declination (deg) of INTEGRAL pointing
         posang    (returned) Position Angle (deg) of INTEGRAL pointing

    Keyword:
         silent    (bool) Suppress process monitoring

    2003-02-05/NJW
    2006-12-28/NJW, updated to look for binary Yorick saved pointings files
*/
{

   extern GPFS_counter;

   local swid, ra, dec, posangle, posang_arr, swid_list;

   status = 0;
   swid_in = swid_in(1); // Make sure it is a scalar
   if( is_void(GPFS_counter) ) GPFS_counter = 0;

   ra_scx = 0.0;
   dec_scx = 0.0;
   posang = 0.0;

   if( is_void(silent) ) silent = 0;

   if( strlen( swid_in ) != 12 ) {
      write,format="Invalid SWID value%s\n","";
      return -1;
   }
   revstr = strpart( swid_in, 1:4 );
   str = strpart( revstr, 1:2 );
   //+ point_dir = "/r6/jemx/pointings/";
   //+ ypoint_dir = "/r6/jemx/pointings/ysav"+str+"/";
   point_dir = get_env("J_POINTINGS")+"/";
   ypoint_dir = point_dir + "ysav"+str+"/";

   filenam = "pointings_"+revstr+"p.";

// check if data already in memory

   if( 0 == mem_restore( filenam+"ra", ra ) ) {
      if( !silent ) write,"Restoring from memory "+itoa(++GPFS_counter);
      status = mem_restore( filenam+"dec", dec );
      status = mem_restore( filenam+"swid", swid_list );
      status = mem_restore( filenam+"posang", posang_arr );

// check for existence of binary file

   } else if( file_test( ypoint_dir+filenam+"ysav") ) {

      if( !silent ) write,"Reading binary file "+ypoint_dir+filenam+"ysav";
      bfile = openb( ypoint_dir+filenam+"ysav" );
      restore,bfile, swid, ra, dec, posangle;
      close, bfile;
      swid_list = swid;
      posang_arr = posangle;

   } else {

      filename = filenam + "dat";
      if( !silent ) write,"Reading text file: "+point_dir+filename;
      if( !file_test(point_dir+filename) ) {
         write,format="%s was not found\n", point_dir+filename;
         return -2;
      }

   /*
    * swid_list = rscol(point_dir+filename,1,str=1,silent=1);
    * ra = rscol(point_dir+filename,2,silent=1);
    * dec = rscol(point_dir+filename,3,silent=1);
    * posang_arr = rscol(point_dir+filename,4,silent=1);
    */

      rstab,point_dir+filename,4,swid_list,ra,dec,posang_arr,typ="sfff",silent=1;

   }

// Save to memory to save time in case of repeated calls with same revolution

   if( 0 == mem_save( filenam+"ra", ra, silent=1 ) ) {
      status = mem_save( filenam+"dec", dec, silent=1 );
      status = mem_save( filenam+"swid", swid_list, silent=1 );
      status = mem_save( filenam+"posang", posang_arr, silent=1 );
   }
   w = where( swid_list == swid_in );
   nw = numberof(w);
   if( nw == 0 ) {
      write,format="Pointing for SWID %s was not found\n", swid_in;
      return -3;
   }

   ra_scx = ra(w(1));
   dec_scx = dec(w(1));
   posang = posang_arr(w(1));

   return 0;

}



/* Function j_get_pi_ebds */

func j_get_pi_ebds( &eb1, &eb2 )
/* DOCUMENT j_get_pi_ebds, eb1, eb2

  2005-11-11/NJW
  2006-02-20/NJW Yorick version

   Return the standard PI energy boundaries [keV]
   and set the externals '_PI_Energy_bds_lo' and ' _PI_Energy_bds_hi'.

*/
{
   extern _PI_Energy_bds_lo, _PI_Energy_bds_hi;

   basdir = get_env("IC_BASE")+"/imod_grp";
   imodfiles = file_search("jmx2_imod_grp_0???.fits",basdir);
   imodfiles = imodfiles(sort(imodfiles));
   imod_grp = imodfiles(0);

   eb1 = rdfitscol( imod_grp+"[JMX2-FBDS-MOD]","E_MIN");
   eb2 = rdfitscol( imod_grp+"[JMX2-FBDS-MOD]","E_MAX");
   _PI_Energy_bds_lo = eb1;
   _PI_Energy_bds_hi = eb2;

}

/* Function j_thruput */

func j_thruput( off_axis_angle, radius_lim, omega )
/* DOCUMENT res = j_thruput( off_axis_angle, radius_lim, omega )

   Calculates the throughput of JEMX for a pointing at
   off_axis_angle [deg] with an angle "omega" [deg]
   with respect to a collimator grid direction

    NJW  970221
    2006-09-08/NJW Updated to handle detector radius limit
    2007-01-02/NJW Translated to Yorick

*/
{
   // --- off axis angle

   alpha = off_axis_angle * pi / 180.0; // converted to radians

   // --- area of the two circles

   d = tan(alpha) * 3401.0;   // Det-Mask distance is 3401 mm
   f = circ_area1(d, 267.5, radius_lim) / (pi * radius_lim^2);
   //+ write,"rel. circ area = ", f

   // --- through collimator

   q = (57./6.6) * tan(alpha);
   //+ write,"q = ", q
   t = (1 - q * cos(omega*pi/180.0)) * (1. - q * sin(omega*pi/180.0));
   //+ write," thru coll = ", t

   return t > 0.0 ? f*t : 0.0;

}

func cbj_image_name( swid )
/* DOCUMENT cbj_name = cbj_image_name( swid )

   Return the image file name following the CBJ convention
   from a given SWID (12 characters).

   2007-01-09/NJW
*/
{

   revno = atoi(strpart(swid,1:4));
   pidno = atoi(strpart(swid,5:8));
   return  "skyimage_1_5_"+swrite(format="%i",revno)+"_"+swrite(format="%i",pidno)+"_0.fits";
}


/* Function dattim2ijd */

func dattim2ijd( dattim )
/* DOCUMENT ijd = dattim2ijd ( dattim )

   Convert a (array of) date/time string in UTC (such as 2003-01-27T10:48:17) to
   IJD i.e. number of days since Jan 01, 00:00, 2000

   Including 64.184 s offset prior to year 2000 and leap seconds at
      2005-12-31, 2008-12-31, and 2012-06-30 at the end of the day.

   SEE ALSO: ijd2dattim, timeconv, ijd2utc, ijd2tt, utc2ijd, tt2ijd

   2003-01-27/NJW
   2006-10-03/NJW translated to Yorick
   2006-12-21/NJW updated to account for difference between TT and UTC

*/

{


// Initial data definition

   days_in_month = [31,28,31,30,31,30,31,31,30,31,30,31];
   days_in_month_leap = [31,29,31,30,31,30,31,31,30,31,30,31];
   ijd_tmp = -1.0;
   ijd = -1.0;

   acc_days = array(long,12);
   acc_days_leap = array(long,12);
   for( i = 1; i <= 11; i++ ) {
      acc_days(i+1) = acc_days(i) + days_in_month(i);
      acc_days_leap(i+1) = acc_days_leap(i) + days_in_month_leap(i);
   }

   first = 1;

   // Analyze input (array of) string(s) and check format

   if( typeof( dattim ) != "string" ) {
      write,"1. argument is not of type string";
      return;
   }
   nelem = numberof(dattim);

   for( i = 1; i <= nelem; i++ ) {

      istr = swrite(format="%i", i);
      curstr = strcompress(dattim(i),all=1);
      len = strlen(curstr);

      if( len  <  13 ) {
         write,format="##34## %s %s invalid format (too short)\n", istr, dattim(i);
         continue;
      }

   //  Requirements   2002-11-16T23:45:56.342
   //            or   2002-11-16T23:45:56
   //            or   2002-11-16T23:45
   //            or   2002-11-16T23
   //      One "T", two "-"s

      ch = strpart(curstr,11:11)

      if( ch  !=  "T" ) {
         write,format="##35## %s %s invalid format (no 'T' in pos 11)\n",\
                    istr, dattim(i);
         continue;
      }

      ch = strpart(curstr,5:5);

      if( ch  !=  "-" ) {
         write,format="##35## %s %s invalid format (no '-' in pos 5)\n",\
                    istr, dattim(i);
         continue;
      }

      ch = strpart(curstr,8:8);

      if( ch  !=  "-" ) {
         write,format="##36## %s %s invalid format (no '-' in pos 8)\n",\
                    istr, dattim(i);
         continue;
      }

      year = strpart(curstr,1:4);

      if( !is_digit(year) ) {
         write,format="##37## %s %s invalid format (year is not a number)\n",\
                    istr, dattim(i);
         continue;
      }

      iyear = atoi(year);
      if( iyear  <  2000 || iyear  >  2099 ) {
         write,format="##38## %s %s invalid format (requirement: 2000 <= year <= 2100)\n",\
                    istr, dattim(i);
         continue;
      }

      cur_year_is_leap = ( iyear%4 == 0 ) ? 1 : 0;

      month = strpart(curstr,6:7);

      if( !is_digit(month) ) {
         write,format="##39## %s %s invalid format (month is not a number)\n",\
                    istr, dattim(i);
         continue;
      }

      imonth = atoi(month);
      if( imonth  <  1 || imonth  >  12 ) {
         write,format="##40## %s %s invalid format (illegal month number)\n",\
                    istr, dattim(i);
         continue;
      }

      day = strpart(curstr,9:10);

      if( !is_digit(day) ) {
         write,format="##41## %s %s invalid format (day is not a number)\n",\
                    istr, dattim(i);
         continue;
      }
      iday = atoi(day);

      daymax = days_in_month(imonth);
      if( cur_year_is_leap && imonth == 2 ) ++daymax;

      if( iday  <  1 || iday  >  daymax ) {
         write,format="##42## %s %s invalid format (invalid day number)\n",\
                    istr, dattim(i);
         continue;
      }

      hour = strpart(curstr,12:13);

      if( !is_digit(hour) ) {
         write,format="##43## %s %s invalid format (hour is not a number)\n",\
                    istr, dattim(i);
         continue;
      }

      if( atoi(hour) < 0 || atoi(hour) > 23 ) {
         write,format="##44## %s %s invalid format (invalid hour number)\n",\
                    istr, dattim(i);
         continue;
      }

      minute = "0";
      second = "0";

      if( len >= 16 ) { // The minute is given as well

         ch = strpart(curstr,14:14);
         if( ch != ":" ) {
            write,format="##45## %s %s invalid format (no ':' in pos 14)\n",\
                    istr, dattim(i);
            continue;
         }

         minute = strpart(curstr,15:16);

         if( !is_digit(minute) ) {
            write,format="##46## %s %s invalid format (minute is not a number)\n",\
                    istr, dattim(i);
            continue;
         }

         if( atoi(minute) < 0 || atoi(minute) > 59 ) {
            write,format="##47## %s %s invalid format (invalid minute number)\n",\
                    istr, dattim(i);
            continue;
         }

         if( len >= 19 ) { // The second is given as well

            ch = strpart(curstr,17:17);
            if( ch  !=  ":" ) {
               write,format="##48## %s %s invalid format (no ':' in pos 17)\n",\
                    istr, dattim(i);
               continue;
            }

            //140708+ second = strpart(curstr,18:19);
            second = strpart(curstr,18:); // accept decimals

            if( !is_number(second) ) {
               write,format="##49## %s %s invalid format (second is not a number)\n",\
                    istr, dattim(i);
               continue;
            }

            if( atof(second) < 0 || atof(second) > 59.9999999 ) {
               write,format="##50## %s %s invalid format (invalid second number)\n",\
                    istr, dattim(i);
               continue;
            }

         }

      }

      num_years = long(iyear - 2000);
      num_leap_years = (num_years+3L)/4; // before current year

      num_days = num_years * 365 + num_leap_years;

      if( cur_year_is_leap ) {
         num_days = num_days + acc_days_leap(imonth) + iday - 1;
      } else {
         num_days = num_days + acc_days(imonth) + iday - 1;
      }
      num_seconds = atoi(hour)*3600.0 + atoi(minute)*60.0 + atof(second);

      // account for leap seconds
      t_dif_s = 64.184;
      if( iyear >= 2006 ) t_dif_s = 65.184;               // at end of 2005
      if( iyear >= 2009 ) t_dif_s = 66.184;               // at end of 2008
      if( iyear == 2012 && imonth > 6 ) t_dif_s = 67.184; // mid 2012
      if( iyear >= 2013 ) t_dif_s = 67.184;               // after mid 2012

      t_dif_ijd = t_dif_s / 86400.0;
      ijd_tmp = num_days + num_seconds/(24.0*3600.0) + t_dif_ijd;

      if( first ) {
         ijd = ijd_tmp;
         first = 0;
      } else {
          grow, ijd, ijd_tmp;
      }



   }

   return ijd;

}


/* Function ijd2dattim */

func ijd2dattim( ijd )
/* DOCUMENT dattim = ijd2dattim( ijd )

    Converts scalar or array of IJD values to 'Gregorian' strings:
     YYYY-MM-DDTHH:MM:SS.ddd UTC

   Including 64.184 s offset prior to year 2000 and leap seconds at
      2005-12-31, 2008-12-31, and 2012-06-30 at the end of the day.

   SEE ALSO: dattim2ijd, timeconv, ijd2utc, ijd2tt, utc2ijd, tt2ijd
*/
{
   ijd = double(ijd);
   if( is_scalar(ijd) ) return _ijd2dattim( ijd );
   n = numberof( ijd );
   res = array(string, n );
   for(i = 1; i <= n; i++ ) res(i) = _ijd2dattim( ijd(i) );
   return res;
}

func _ijd2dattim( ijd )
/* DOCUMENT dattim = _ijd2dattim( ijd )

   Work function for 'dattim2ijd', takes only a scalar.

   2006-10-03/NJW

*/
{

   if( ijd < 0.1 ) {
      write,format="Does not work for IJD = %f\n",ijd;
      return "";
   }

   days_in_month = [31,28,31,30,31,30,31,31,30,31,30,31];

// Including the difference between TT (Terrestrial Time) and UTC
// See FAQ of ISDC home page

   t_dif_s = 64.184;
   // leap second end of year 2005
   if( ijd > 2192.0007428704 ) t_dif_s = 65.184;
   // leap second end of year 2008
   if( ijd > 3288.0007544444 ) t_dif_s = 66.184;
   // leap second mid year 2012
   if( ijd > 4565.0007660185 ) t_dif_s = 67.184;
   t_dif_ijd = t_dif_s / 86400.0;
   ijd_utc = ijd - t_dif_ijd;

   dat = "2000-01-01T00:00:00.000";

   // Get year

   for( year = 2001; year <= 2099; year++ ) {

      str_year = swrite(format="%4.4i",year);
      dat = strput( dat, str_year, 1 );
      t = dattim2ijd(dat);
      if( t > ijd ) break;

   }

   year--;
   str_year = swrite(format="%4.4i", year);
   dat = strput(dat, str_year, 1 );
   if( year%4 == 0 ) days_in_month(2) = 29; // leap year

   // Get month

   for( month = 1; month <= 12; month++ ) {

      str_month = swrite(format="%2.2i", month);
      dat = strput(dat,str_month,6);
      t = dattim2ijd(dat);
      if( t > ijd ) break;

   }

   month--;
   str_month = swrite(format="%2.2i", month );
   dat = strput(dat,str_month,6);

   // Get day

   for( day = 1; day <= days_in_month(month); day++ ) {

      str_day = swrite(format="%2.2i", day );
      dat = strput(dat,str_day,9);
      t = dattim2ijd(dat);
      if( t > ijd ) break;

   }

   day--;
   str_day = swrite(format="%2.2i", day);
   dat = strput(dat,str_day,9);

   frac = ijd_utc - floor(ijd_utc);
   h = frac * 24;
   hour = toint(h);
   m = (h - hour) * 60;
   minute = toint(m);
   //140708+ s = (m - minute) * 60;
   //140708+ second = toint(s);
   second = (m - minute) * 60;

   str_hour = swrite(format="%2.2i", hour);
   dat = strput(dat,str_hour,12);

   str_minute = swrite(format="%2.2i", minute);
   dat = strput(dat,str_minute,15);

   //140708+ str_second = swrite(format="%2.2i", second);
   str_second = swrite(format="%06.3f", second);
   dat = strput(dat,str_second,18);

   return dat;

}



/* Function j_get_used_gain */

func j_get_used_gain( jemxNum, swid, events, &selection )
/* DOCUMENT gain = j_get_used_gain( jemxNum, swid, events, >selection )

   Returns an array with gain values for each event that fulfills
   STATUS < 33 (as reported in returned argument 'selection' with indices).

   The gain [linPHA per keV] is derived by comparing PHA and PI
   for processed data.

   events is a struct with members: detx, dety, rawx, rawy, piarr,
	pha, status, evtime

   2005-11-03/NJW. Based on "predict_arf"


   Define the local gain as the number of linPHA channels per keV:

      local_gain = linPHA / Energy

   The local gain is both a function of detector position and of time

   Define the global gain (global_gain) as

      global_gain = average[ linPHA / (Energy * f_spag) ]

   where f_spag is the local spatial gain factor and the average
   is taken over a suitable area of the detector

   The global gain is a function of time i.e. "global" refers to the
   detector surface

   2007-01-23/NJW translated from 0idl/proc/j_get_used_gain.pro
*/
{

   if( is_void(events) ) {
      write,"Syntax: gain = j_get_used_gain( jemxNum, swid, events )";
      write,"   'events' is a struct as returned by 'j_get_events'";
      return [];
   }

   jmxstr = swrite(format="jmx%1i",jemxNum);
   JMXstr = swrite(format="JMX%1i",jemxNum);
   jstr = swrite(format="%1i",jemxNum);

   icversion = get_env("ICVERSION");
   system,"get_insta_swijd -swid "+swid+" "+jstr+" imod "+icversion+" > tmpa";
   insta_imod = read_slist("tmpa")(1);

   //  Define "flag" array to make sure that each detector position is only
   //  counted once when making the average
   /*
    ***           This precaution seems not to have any effect, maybe a
    ***           forgotten or abandoned initiative (2015-04-16/NJW)
    ***           Also notice variable 'n_used' which is not used (sic!).
    */

   flag = array( int, 256,256);

   //  Get the PI energy boundaries

   imod = get_env("IC_BASE")+"/imod_grp/"+jmxstr+"_imod_grp_"+insta_imod+".fits";
   //+ dol_fbds = imod+"["+JMXstr+"-FBDS-MOD]";
   //+ write,"Reading e_min, e_max ...";
   //+ e_min = rdfitscol( dol_fbds, "e_min" );
   //+ e_max = rdfitscol( dol_fbds, "e_max" );

   //  Get the PHA rebinning table i.e. from linear PHA to PHA in telemetry

   dol_fulb = imod+"["+JMXstr+"-FULB-MOD]";
   fulb = rdfitscol( dol_fulb, "channel" );

   //  Get the spatial gain array with two components

   dol_spag = imod+"["+JMXstr+"-SPAG-MOD]";
   write,"Reading SPAG ...";
   spag = readfits( dol_spag );

   //  Make an event selection based on STATUS flag

   nrows = numberof(events);
   selection = where( events.status  <  33 );
   nw = numberof(selection);
   write,itoa(nw)+" were selected from "+itoa(nrows)+" events";
   if( nw == 0 ) {
      return [];
   }
   evs = events(selection);

   lin_pha = array( float, nw);

   // Since the global gain is a function of time an array
   // is made

   global_gain_arr = array( float, nw);

   n_used = 0;

   //   Loop over the selected events

   for( i = 1; i <= nw; i++ ) {

      //  assign the linPHA value randomly between given limits

      lin_pha_min = fulb(evs(i).pha) + 1;
      lin_pha_max = fulb(evs(i).pha+1);
      if( evs(i).pha  <=  128 ) {
         lin_pha(i) = lin_pha_max;
      } else {
         lin_pha(i) = lin_pha_min + \
            float(toint(random([])*(lin_pha_max-lin_pha_min)+0.5));
      }

      //  "lin_pha" is (close to) the linPHA signal as delivered from the DFEE

      //  assign an energy between PI limits

      f_spag = ((1. + 0.01*spag(evs(i).rawx+1,evs(i).rawy+1,1)) \
                *spag(evs(i).rawx+1,evs(i).rawy+1,2));

      global_gain_arr(i) = lin_pha(i) / (evs(i).energy * f_spag);

   }

   return global_gain_arr;

}

/* Function j_get_events */

struct Event {
       float detx;
       float dety;
       int   rawx;
       int   rawy;
       double evtime;
       double energy;
       int   pha;
       int piarr;
       int status; }

func j_get_events( jemxNum, swid, silent= )
/* DOCUMENT res = j_get_events( jemxNum, swid, silent= )

   Returns an array of struct 'Event' with events from the
   specified SWID read from 'jmxi_full_cor.fits.gz'

   Keyword 'silent' is transmitted to gz_proxy_file

    2007-01-25/NJW
    2008-02-11/NJW updated to use OSA7 and optional running of COR level
    2010-01-28/NJW removed the automatic jmxi_full_cor.fits production
*/
{

   local Events;
   local dummy, extno;

   jmxstr = swrite(format="JMX%1i",jemxNum);
   jstr = swrite(format="%1i",jemxNum);
   revol = strpart(swid,1:4);

// Set externals (_PI_Energy_bds_lo/hi) with PI energy boundaries

   if( is_void(_PI_Energy_bds_lo) ) j_get_pi_ebds;

//  prepare filenames for useful files

   savname = "/r12/scratch/jemx/njw/"+revol+"/j"+jstr+"_"+swid+".ysav";
   corname = "/jemx/arc/rev_3/scw/"+revol+"/"+swid+".001/jmx"+jstr+"_full_cor.fits";

//  Check for previously saved event file for faster reading

   if( file_test(savname) ) { // The easy way

      stream = openb(savname);
      restore, stream, Events;
      close, stream;

   } else { // Find and read the jmxi_full_cor.fits file

      cor = 0;
      if( file_test(corname) || file_test(corname+".gz") ) {

// Check for jmxi_full_cor.fits file in JEM-X archive

         full_cor = gz_proxy_file( corname, silent=silent, dir="/r12/scratch/njw/proxies" );
         cor = 1;

      } else {

         write,"full_cor file is not available, try using jmxi_events.fits";

      }

   //     The event file is to be found in the data archive:
      evts_file = "/jemx/arc/rev_3/scw/"+revol+"/"+swid+".001/jmx"+jstr+"_events.fits";
      evts_file_p = gz_proxy_file( evts_file, silent=silent, dir="/r12/scratch/njw/proxies" );
      if( is_void(evts_file_p) ) {
         write,"Failed to find: "+evts_file;
         return [];
      }

      rhdr = headfits( evts_file_p+"+1" );
      //+ rswid = itoa(fxpar(rhdr, "SWID"));
      rswid = fxpar(rhdr, "SWID"); // is returned as a string of length 12
      //+ rinstr = itoa(fxpar(rhdr, "INSTRUME"));
      rinstr = fxpar(rhdr, "INSTRUME"); // is returned as "JMX1" or "JMX2"

      if( cor ) {
         fhdr = headfits( full_cor+"+1" );
         fswid = fxpar(fhdr, "SWID");
         finstr = fxpar(fhdr, "INSTRUME");
         if( fswid  !=  rswid ) {
            write,"Mismatching SWIDs: ", fswid+"  "+rswid;
            return [];
         }
         if( finstr  !=  rinstr ) {
            write,"Mismatching INSTRUMEs: ", finstr+"  "+rinstr;
            return [];
         }
      }

      //  Get the events
      //  Since jmxi_full_cor.fits.gz with columns DETX, DETY, PI, ENERGY, STATUS
      //  may have better correction than jmxi_events.fits[JMXi-FULL-ALL]
      //  these values are loaded afterwards

      extnam = cor ? finstr+"-FULL-ALL" : rinstr+"-FULL-ALL";
      get_exten_no, evts_file_p+"["+extnam+"]", dummy, extno;
      if( extno == -1 ) {
         write,"No FULL-ALL extension found in jxmi_events.fits, quit!";
         return [];
      }
      dol = evts_file_p+"+"+itoa(extno);
      hdr = headfits( dol );
      ptr = rdfitsbin( dol );
      rawx   = *ptr(fits_colnum( hdr, "RAWX"));
      rawy   = *ptr(fits_colnum( hdr, "RAWY"));
      detx   = *ptr(fits_colnum( hdr, "DETX"));
      dety   = *ptr(fits_colnum( hdr, "DETY"));
      pha    = *ptr(fits_colnum( hdr, "PHA"));
      evtime = *ptr(fits_colnum( hdr, "TIME"));
      piarr  = *ptr(fits_colnum( hdr,"PI"));
      status = *ptr(fits_colnum( hdr,"STATUS"));
      cn = fits_colnum(hdr,"ENERGY",silent=silent);
      if( cn ) energy = *ptr(cn);
      nrows = numberof(rawx);

      if( cor ) { // if jmxi_full_cor.fits exists then override some values
         dol = full_cor+"+1";
         hdr = headfits( dol );
         ptr = rdfitsbin( dol );
         detx   = *ptr(fits_colnum( hdr, "DETX"));
         dety   = *ptr(fits_colnum( hdr, "DETY"));
         piarr  = *ptr(fits_colnum( hdr,"PI"));
         cn = fits_colnum(hdr,"ENERGY",silent=silent);
         if( cn ) energy = *ptr(cn);
         status = *ptr(fits_colnum(hdr,"STATUS"));
      }
      if( nrows == 0 ) {
         write,"No events in  file";
         return [];
      }

      Events = array(Event, nrows);
      Events.rawx = toint(rawx);
      Events.rawy = toint(rawy);
      Events.pha  = toint(pha);
      Events.evtime  = evtime;
      if( cn ) {
         Events.energy  = energy;
      } else {
          Events.energy = _PI_Energy_bds_lo(piarr+1) + \
            random(nrows)*(_PI_Energy_bds_hi(piarr+1) - _PI_Energy_bds_lo(piarr+1));
      }

      Events.detx = detx;
      Events.dety = dety;
      Events.piarr = toint(piarr);
      Events.status = toint(status);

//  Save this event array for quicker read next time it is requested

      //+ savname = "/r12/scratch/jemx/njw/"+revol+"/j"+jstr+"_"+swid+".ysav";
      //    defined earlier in the function
      setup_dir, savname, type='f';
      stream = createb(savname);
      save, stream, Events;
      close, stream;

   }

   return Events;

}


/* Function j_gain_values */

func j_gain_values( jemxNum, swid, diag= )
/* DOCUMENT res = j_gain_values( jemxNum, swid, diag= )


    2007-01-25/NJW
*/

{

   local selection;

   num_bins = 20;

   events = j_get_events( jemxNum, swid );
   gain = j_get_used_gain( jemxNum, swid, events, selection );
   if( is_void(gain) ) return [];
   events = events(selection);
   w = where(gain < 50.);
   ngain = numberof(w);
   if( ngain == 0 ) {
      write,"No gain values below 50 - skip";
      return [];
   }
   gain = gain(w);
   ev_time = events(w).evtime;
   if( ngain < num_bins ) num_bins = ngain;
   n_per_bin = ngain / num_bins;
   gain_bin = array(float, num_bins);
   time_bin = array(double, num_bins);
   for( i = 1; i <= num_bins; i++ ) {
      i1 = (i-1)*n_per_bin + 1;
      i2 = i * n_per_bin;
      gain_bin(i) = avg(gain(i1:i2));
      time_bin(i) = 0.5*(ev_time(i1)+ev_time(i2));
   }
   ev_time0 = 0.5*(ev_time(1)+ev_time(0));
   coefs = lin_regress( time_bin - ev_time0, gain_bin, 0*gain_bin+1 );
   fit = coefs(1) + coefs(2)*(time_bin - ev_time0);
   chi2 = sum((gain_bin - fit)^2);
   quality = chi2 > 20. ? 1.0 : 0.0;

   if( diag ) {
      jmxstr = swrite(format="JMX%1i", jemxNum);
      plot, (ev_time-ev_time0)*86400.0, gain, ps=1,yr=[0,50];
      xyouts, 0.5,0.8, swid, align=0.5,ndc=1;
      oplot, (time_bin-ev_time0)*86400.0, gain_bin+3;
      oplot, (time_bin-ev_time0)*86400.0, fit+3;
   }

   return [coefs(1),coefs(2),ev_time0,quality];

}

/* Function pointing_plot */

func pointing_plot( rev, zoom=, read=, nn=, flog=, pane= )
/* DOCUMENT pointing_plot, rev, zoom=, read=, nn=, flog=, pane=

   plot the INTEGRAL pointings from a revolution.

   Keywords:

   If 'zoom' is given as a 3 element array: (ra,dec,radius)
      then a close-up of that region is shown.
   If 'flog' is a filename (string) then various information is put there.
   Keyword 'pane' defines the window number (defaults to zero).
   Keyword 'read' opens the possibility to get a close-up of a region,
      whose center can be selected by a cursor click. The value of 'read'
      determines the radius of zoomed area.
   Keyword 'nn' prevents putting the pointing numbers on the plot

   2005-00-00/NJW
   2008-05-28/NJW, added keyword pane.
*/

{

   local swid, ra, dec;

   mon = 0;
   if( numberof(flog) )  {
      if( "string" == typeof( flog ) ) {
         lun = create(flog);
         mon = 1;
      } else {
         write,"Keyword 'flog' was not a string - skip logging";
      }
   }

   if( is_void(pane) ) pane = 0;

   revstr = swrite(format="%4.4i",rev);
   basdir = app_slash(get_env("J_POINTINGS"));
   filename = basdir+"pointings_"+revstr+"p.dat";
   if( mon ) write,lun,filename;

/*
 * swid = rscol( filename, 1, silent=1, str=1);
 * ra = rscol( filename, 2, silent=1);
 * dec = rscol( filename, 3, silent=1);
 */

   rstab, filename, 3, swid, ra, dec, typ="sff",silent=1;

   if( mon ) {
      for( imon = 1; imon <= numberof(swid); imon++ ) {
         write,lun,swid(imon),ra(imon),dec(imon);
      }
   }

   if( numberof(zoom) == 3 )  {

      window,pane,style="boxed.gs";

      r = arcdist( zoom(1), zoom(2), ra, dec );
      w = where( r < zoom(3) );
      nw = numberof(w);
      if( nw == 0 )  {
         write,"Sorry, no pointings were close enough";
         if( mon )  {
            write,lun,"Sorry, no pointings were close enough";
            close,lun;
         }
         return 0;
      }

      swid = swid(w);
      ra = ra(w);
      dec = dec(w);

      plot, ra, dec, ps=3, symsize=2, \
         xr=zoom(1)+zoom(3)*[1,-1]/cos(zoom(2)*pi/180), \
         yr=zoom(2)+zoom(3)*[-1,1], xtitle="R.A. [degrees]", \
         ytitle="Declination [degrees]";
      // res = convert_coord([0.12,0.88],/normal,/to_data);
      plt,"Revol: "+swrite(format="%3.3i",rev),0.4,0.89,justify="CA",height=20;
      oplot,[zoom(1)],[zoom(2)],ps=12, symsize=0.5;
      oplot,[zoom(1)],[zoom(2)],ps=13, symsize=0.5;

      // Add text for each point with PID except when inhibited
      if( !nn ) {
         for( i = 1; i <= nw; i++) {
            pid = strpart(swid(i),5:8);
            sub = strpart(swid(i),11:11);
            off = 0.06;
            for( j = 1; j <= i; j++ ) {
               if( arcdist(ra(j),dec(j),ra(i),dec(i)) < 0.1 ) off = off + 0.06;
            }
            plt,pid+"/"+sub,ra(i),dec(i)+zoom(3)*off,tosys=1,justify="CA",height=8;
         }
      }

   } else  {

      window,pane,style="nobox.gs";

      /*  Plot the grid */

      plot,[0],[0],xr=[180,-180],yr=1.4*[-90,90];

      /* lon grid */

      b = span(-90,90,100);
      for( lon = -179.99; lon < 180.1; lon += 89.99 ) {
         l = array(0,100) + lon;
         xy = aitoff(l,b);
         listy = 2;
         if( lon < -179 || lon > 179 ) listy = 1;
         oplot,xy(1,),xy(2,),li=listy;
      }

      /* lat grid */

      l = span(-179.999,180,100);
      for( lat = -60.; lat < 60.01; lat += 30. ) {
         b = array(0,100) + lat;
         xy = aitoff(l,b);
         oplot,xy(1,),xy(2,),li=2;
      }

      //+ plt,"Revol: "+swrite(rev),0,70,tosys=1,height=20,justify="CA";
      plt,"Revol: "+swrite(rev),0.39,0.85,height=20,justify="CA";

      // Overplot the pointings;

      glb = galactic(ra, dec);

      if( mon ) {
         for( imon = 1; imon <= numberof(swid); imon++ ) {
            write,lun,swid(imon),glb(imon,1),glb(imon,2);
         }
      }

      xy = aitoff(glb(,1),glb(,2));

      oplot, xy(1,), xy(2,), ps=3,symsize=1;

      xyouts,0.4,0.88,"INTEGRAL pointings",align=0.5,ndc=1,charsize=1.3;

      // Include selected sources

      seso_dol = "/r6/jemx/catalogs/selected_sources.fits+1";
      seso_name = strtrim(rdfitscol(seso_dol,"name"));
      seso_ra_obj = rdfitscol(seso_dol,"ra_obj");
      seso_dec_obj = rdfitscol(seso_dol,"dec_obj");
      nseso = numberof(seso_name);

      for( i = 1; i <= nseso; i++ ) {
         xysrc = galactic(seso_ra_obj(i),seso_dec_obj(i));
         xysrc = aitoff(xysrc(1), xysrc(2));
         oplot,xysrc(1,),xysrc(2,),ps=13,color="red",symsize=0.2,fill=1;
         dx = seso_name(i) == "Crab" ? 5.0 : 0.0;
         dy = 3.;
         xyouts,xysrc(1,1)+dx,xysrc(2,1)+dy,seso_name(i),align=0.5,charsize=0.5,color="red";
      }

      // Add the Sun position for the middle of the revolution

      supos = sunposition(get_revol_start_time(rev) + 1.5);
      xysrc = galactic(supos(1),supos(2));
      xysrc = aitoff(xysrc(1), xysrc(2));
      oplot, xysrc(1:1), xysrc(2:2), ps=13, color="yellow",symsize=0.6,fill=1;
      oplot, xysrc(1:1), xysrc(2:2), ps=13, symsize=0.6;
      oplot, xysrc(1:1), xysrc(2:2), ps=13, symsize=0.15,fill=1;

      if( read ) {
         res = curmark1( style=0, prompt="Click on the sky position : ..." );
         write,format="Coords in plot: %f %f\n", res(1), res(2);
         gala = rever_aitoff( res(1), res(2) );
         write,format="L,B: %8.3f, %8.3f\n", gala(1), gala(2);
         equa = equatorial( gala(1), gala(2) );
         write,format="RAdec: %8.3f, %8.3f\n", equa(1), equa(2);
         grow,equa,read;
         pointing_plot,rev,zoom=equa,nn=nn,pane=pane;
      }

   }

   if( mon ) close,lun;

}

/* Function validate_swid_list */

func validate_swid_list( swid_arr, jemxNum, ign_bti=, chat=, silent= )
/* DOCUMENT new_swid_arr = validate_swid_list( swid_arr, jemxNum,
                   ign_bti=, chat=, silent= )

   'swid_arr' is assumed to be an array of SWIDs. Here
   they will be checked against the FITS pointing file
   /r6/jemx/pointings/pointings_RRRR.fits

   Keyword 'ign_bti' will cause the function to ignore BTI information.

    2007-02-14/NJW
    2008-01-14/NJW, updated to use ../RRRR/jmxi_shd.list
    2008-05-29/NJW, updated to use /r6/jemx/pointings/pointings_RRRR.fits
    2012-04-19/NJW, updated to check new BTI columns as well
*/

{

   if( is_void(chat) ) chat = 0;

   if( is_void(jemxNum) ) {
      write,"Syntax: new_swid_arr = validate_swid_list( swid_arr, jemxNum )";
      return [];
   }
   jstr = itoa(jemxNum);
   nswids = numberof(swid_arr);
   new_swid_arr = [];
   revol_prev = "0000";
   prev_pointfile = "asdf";
   for( i = 1; i <= nswids; i++ ) {
      revol = strpart(swid_arr(i),1:4);
      if( revol != revol_prev ) {
         pointfile = "/r6/jemx/pointings/pointings_"+revol+".fits";
         if( !file_test( pointfile ) ) {
            if( prev_pointfile != pointfile ) write,format="%s not found\n", pointfile;
            prev_pointfile = pointfile;
            continue;
         }
         if( chat > 1 ) write,"Reading new pointdata "+revol;
         swids_tab = rdfitscol( pointfile+"+1", "swid" );
         shd_flag  = ( rdfitscol( pointfile+"+1", "shd_j"+jstr ) == 1 );
         bti_flag  = strmatch( rdfitscol( pointfile+"+1", "bti_j"+jstr ), "N" );
         if( !ign_bti ) shd_flag = ( shd_flag & bti_flag );
         revol_prev = revol;
      }
      w = where( swids_tab == swid_arr(i) );
      if( numberof(w) == 0 ) {
         if( !silent ) write,format="SWID %s not in %s\n", swid_arr(i), \
                         pointfile;
         continue;
      }
      if( shd_flag(w(1)) == 1 ) {
         grow, new_swid_arr, swid_arr(i);
         if( !bti_flag(w(1)) ) write,"BTI warning for SWID "+swid_arr(i);
      } else {
         if( !silent ) write,format="%s was rejected, shd_flag & bti_flag != 1\n", swid_arr(i);
      }
   }

   return new_swid_arr;

}

/* Function nearby_sources */

func nearby_sources( ra_in, dec_in, radius, force=, clog= )
/* DOCUMENT res = nearby_sources( ra_in, dec_in, radius, force=, clog= )
      or:   nearby_sources, ra_in, dec_in, radius, force=, clog=

  2007-08-15/NJW copied from nearest_source.i

   When called as a function it returns an array of struct 's_Src'
   with the source within the given radius, sorted according to distance.
   If there are no sources within the given radius then [] is returned

   If called as a subroutine output will be given on the terminal.

   SEE ALSO: nearest_source, find_src_by_name

  Keyword  force:   Force reading the catalog contents
           clog :   The DOL of the catalogue
*/

{

   extern Source_cat, Source_cat_dol, Source_cat_num;

   if( dimsof(ra_in)(1)  !=  0 || dimsof(dec_in)(1)  !=  0 ) {
      write,"Input coordinate must be scalar";
      return [];
   }

   read_catalog, force=force, clog=clog; // Will read catalog if required

   ra_in = double(ra_in);
   dec_in = double(dec_in);

   dista = arcdist(Source_cat.ra_obj,Source_cat.dec_obj,ra_in,dec_in);

   w = where( dista < radius );
   nw = numberof(w);
   if( nw == 0 ) return [];

   if( am_subroutine() ) {
      len_name = max(strlen(strtrim(Source_cat.name)));
      s1 = strpadd("",len_name," ");
      write,"Name"+s1+"  RA        DEC    Distance";
      write,"    "+s1+"  deg       deg      deg";
      fmt = swrite(format="%%%is  %%8.4f  %%8.4f  %%8.4f\n",len_name+3);
   }
   is = sort(dista(w));
   //+ res = array(string,4,nw);
   for( i = 1; i <= nw; i++ ) {
      //+ res(1,i) = Source_cat(w(is(i))).name;
      //+ res(2,i) = itoa(Source_cat(w(is(i))).ra_obj);
      //+ res(3,i) = itoa(Source_cat(w(is(i))).dec_obj);
      //+ res(4,i) = itoa(dista(w(is(i))));
      if( am_subroutine() ) {
         write,format=fmt, \
           strtrim(Source_cat(w(is(i))).name), Source_cat(w(is(i))).ra_obj, \
           Source_cat(w(is(i))).dec_obj, dista(w(is(i)));
      }
   }

   return Source_cat(w(is));

}

/* Function nearest_source */

func nearest_source( ra_in, dec_in, radius, &dist, force=, clog= )
/* DOCUMENT res = nearest_source( ra_in, dec_in, radius, >dist, force=, clog= )

   Find the nearest source to the given position inside radius
   'radius' (degrees).

   When called as a function it
   returns a struct (s_Src) with name, ra_obj, dec_obj, err_rad,
   source_id, and class. The distance to the given position is
   returned in the variable 'dist' (in degrees).

   If there is no source within the given radius a [] is returned

   If called as a subroutine the output appears on the terminal.

   Keyword  force:   Force reading the catalog contents
            clog :   The DOL of the catalogue

   SEE ALSO: nearby_sources, find_src_by_name

   2007-08-15/NJW translated to Yorick
   2008-11-25/NJW returns also source_id and class

*/

{

   extern Source_cat, Source_cat_dol, Source_cat_num;

   if( dimsof(ra_in)(1)  !=  0 || dimsof(dec_in)(1)  !=  0 ) {
      write,"Input coordinate must be scalar";
      return [];
   }

   read_catalog, force=force, clog=clog;

   ra_in = double(ra_in);
   dec_in = double(dec_in);

   dista = arcdist(Source_cat.ra_obj,Source_cat.dec_obj,ra_in,dec_in);
   is = sort(dista)(1);
   dist = dista(is);

   if( dist > radius ) {
      dist = 180.;
      return [];
   }

   name = Source_cat(is).name;
   ra = Source_cat(is).ra_obj;
   dec = Source_cat(is).dec_obj;
   src_id = Source_cat(is).source_id;
   class = Source_cat(is).class;

   if( am_subroutine() ) {
      write,format="%s %s RAdec: %9.4f %9.4f deg, dist = %7.2f arcmin, class: %s\n", \
                name, src_id, ra, dec, dist*60.0, itoa(class);
      return;
   } else {
      return Source_cat(is);
   }

}

/* Function _read_catalog */

struct s_Src {
   double ra_obj;
   double dec_obj;
   double err_rad;
   string name;
   string source_id;
   int class;
}

func _read_catalog( dol_of_cat, chat= )
/* DOCUMENT _read_catalog, dol_of_cat, chat=

  Read an X-ray source catalog and store it as external variables:
    Source_cat{ ra_obj, dec_obj, err_rad, name, source_id } (array)
    Source_cat_dol (scalar string)
    Source_cat_num (number of entries)

  NJW/010126  Read a catalog in format like DSRI_CAT.fits
  NJW/2005-10-27  Updated to version 2 for gnrl_refr_cat_latest.fits
  NJW/2006-01-10  Updated to also have the entry: catalogue
  2007-08-15/NJW Translated to yorick

      Some of the "not so important" columns may be missing
      in which case some dummy information is produced

*/

{

   extern Source_cat, Source_cat_dol, Source_cat_num;

   local file_of_cat, extno;

   get_exten_no, dol_of_cat, file_of_cat, extno;
   if( ! file_test( file_of_cat ) ) {
      write,format="_READ_CATALOG: Requested catalog: %s does not exist ...\n", \
               file_of_cat;
      Source_cat_num = -1;
      return;
   }
   Source_cat_dol = fullpath(dol_of_cat);

   // Read the contents of the database

   if( chat ) write,format="_READ_CATALOG: Reading %s ...\n", dol_of_cat;

   fh = headfits(Source_cat_dol,nocheck=1);

   //  RA(_OBJ) and DEC(_OBJ) are required columns

   colnum = fits_colnum( fh, "RA_OBJ" );
   if( is_void(colnum) ) { // Try RA
      colnum = fits_colnum( fh, "RA");
      if( is_void(colnum) ) {
         write,"_READ_CATALOG: Neither RA nor RA_OBJ column found, give up";
         return;
      }
   }
   ra = rdfitscol( dol_of_cat, colnum);

   colnum = fits_colnum( fh, "DEC_OBJ");
   if( is_void(colnum) ) { // Try DEC
      colnum = fits_colnum( fh, "DEC");
      if( is_void(colnum) ) {
         write,"_READ_CATALOG: Neither DEC nor DEC_OBJ column found, give up";
         return;
      }
   }
   dec = rdfitscol( dol_of_cat, colnum);

   Source_cat_num = numberof(ra);

   // The remaining columns may be omitted

   colnum = fits_colnum( fh, "NAME");
   if( is_void(colnum) ) { //  Try "SOURCENAME" in stead
      colnum = fits_colnum( fh, "SOURCENAME");
      if( is_void(colnum) ) { // Try "SOURCE_ID" in stead
         colnum = fits_colnum( fh, "SOURCE_ID" );
      }
   }
   if( is_void(colnum) ) {
      name = array("?NAME",Source_cat_num);
      if( chat ) write,"_READ_CATALOG:  substituted all names with ?NAME";
   } else name = rdfitscol( dol_of_cat, colnum );

   source_id = rdfitscol( dol_of_cat, "SOURCE_ID");
   if( is_void(source_id) ) {
      source_id = array("?SOURCE_ID",Source_cat_num);
      if( chat ) write,"_READ_CATALOG:  substituted all source_ids with ?SOURCE_ID";
   }

   err_rad = rdfitscol( dol_of_cat, "ERR_RAD");
   if( is_void(err_rad) ) {
      err_rad = array( 0.05,Source_cat_num);
      if( chat ) write,"_READ_CATALOG:  substituted all err_rads with 0.05";
   }

   class = rdfitscol( dol_of_cat, "CLASS");
   if( is_void(class) ) {
      class = array( int, Source_cat_num);
      if( chat ) write,"_READ_CATALOG:  substituted all class s with 0";
   }

   Source_cat = array( s_Src, Source_cat_num );

   for( i = 1; i <= Source_cat_num; i++ ) {
      Source_cat(i).ra_obj     = ra(i);
      Source_cat(i).dec_obj    = dec(i);
      Source_cat(i).err_rad    = err_rad(i);
      Source_cat(i).name       = name(i);
      Source_cat(i).source_id  = source_id(i);
      Source_cat(i).class      = class(i);
   }

}

/* Function read_catalog */

func read_catalog( catdol, force=, clog= )
/* DOCUMENT read_catalog, catdol, force=, clog=

    Loads a catalog of X-ray sources to memory in externals:
       Source_cat, Source_cat_dol, Source_cat_num

    Either the argument 'catdol' may be given or the keyword 'clog'.
    If both are given 'clog' will dominate.

    Logics:

    If 'clog' is a string then do: catdol = clog

    After that follow this table:

    catdol Source_cat_dol force
 (1) void      void         0       Read standard ISDC catalog
 (2) void      void         1       Read standard ISDC catalog
 (3) void      string       0       Do nothing
 (4) void      string       1       Read Source_cat_dol
 (5) string    void         0       Read catdol
 (6) string    void         1       Read catdol
 (7) string    string       0       if catdol != Source_cat_dol : read catdol, else do nothing
 (8) string    string       1       Read catdol



*/
{
   extern Source_cat, Source_cat_dol, Source_cat_num; 

   if( typeof(clog) == "string" ) catdol = clog; // let 'clog' prevail

   if( is_void(catdol) ) {
      write,"Called read_catalog with void catdol";
   } else {
      local fname, extno;
      get_exten_no, catdol, fname, extno;
      if( extno == 0 ) catdol += "+1"; // defaults to extension #1
      write,"Called read_catalog with catdol = "+catdol;
   }
   if( is_void(Source_cat_dol) ) {
      write,"Called read_catalog while Source_cat_dol is void";
   } else {
      write,"Called read_catalog while Source_cat_dol = "+Source_cat_dol;
   }

   if( is_void(catdol) ) {

      if( is_void(Source_cat_dol) ) {

         // Case (1) and (2) - no input name so use default
         write,"read_catalog, case 1 or 2 ...";
         Source_cat_dol = "/r6/jemx/catalogs/gnrl_refr_cat_latest.fits+1";
         _read_catalog, Source_cat_dol, chat=1;

      } else {

         if( force ) {
            // Case (4)
            write,"read_catalog, case 4 ...";
            _read_catalog, Source_cat_dol, chat=1;
         } else {
            write,"read_catalog, case 3, do nothing ...";
         }

      }

   } else {

      if( is_void(Source_cat_dol) ) {

         // Case (5) and (6)
         write,"read_catalog, case 5 or 6 ...";
         Source_cat_dol = catdol;
         _read_catalog, Source_cat_dol, chat=1;

      } else {

         if( force ) {
            // Case (8)
            write,"read_catalog, case 8 ...";
            Source_cat_dol = catdol;
            _read_catalog, Source_cat_dol, chat=1;
         } else {
            // Case (7) is really two cases
            local Fname, Extno, fname, extno;
            get_exten_no, Source_cat_dol, Fname, Extno;
            get_exten_no, catdol, fname, extno;
            Fname = fullpath(Fname);
            fname = fullpath(fname);
            if( Fname != fname || extno != Extno ) {
               write,"read_catalog, case 7a ...";
               Source_cat_dol = catdol;
               _read_catalog, Source_cat_dol, chat=1;
            } else {
               write,"read_catalog, case 7b, do nothing ...";
            }

         }

      }

   }

}

/* Function qidsrcs */

func qidsrcs( dol, check=, outfile=, rad=, clobber=, force=, clog=, unique=, chat= )
/* DOCUMENT qidsrcs, dol, check=, outfile=, rad=, clobber=, force=, clog=, unique=, chat=

   Identify sources in a FITS binary table (the input source list given
   by argument 'dol') with at least the columns 'NAME', 'RA_OBJ', and 'DEC_OBJ'.

   The header (HDU) of the updated file will have additional keywords to mark
   that 'qidsrcs' has been applied and to give the catalog name.

   Keywords:
   check     : is to identify sources where the name deviates
               from the one proposed by 'qidsrcs'. No update is done.
   outfile   : if given then the result (with NAME column
               replaced) will be written into that file; else the
               file defined by 'dol' will be overwritten.
               Note that the resulting file will have the binary
               table in the first extension and the rest of the
               file is lost.
   rad       : is a radius limit (in degrees, defaults to 0.1) outside which no
               identification will be done (NAME will be replaced by 'UNKNOWN').
   chat      : the chattiness level
   unique    : implies that the source list consists of unique sources
               so that each catalog source can only be assigned to a single
               source from the list (resembles standard running of
               q_identify_srcs).
   force     : Force re-reading of reference source catalog
   clog      : DOL of reference source catalog
   clobber   : Allow overwriting of output file

   2007-09-12/NJW
   2008-11-18/NJW updated with 'unique' keyword
   2008-11-25/NJW updated with CLASS information
*/
{

   extern Source_cat, Source_cat_dol, Source_cat_num;

   local filename, extno, fh, nrows, nam, src_id, ra, dec;
   local racat, deccat, class;

   if( is_void(chat) ) chat = 0;

   if( is_void(rad) ) rad = 0.1; // degrees
   get_exten_no, dol, filename, extno;
   if( extno == 0 ) dol = dol+"+1"; // extension #1 is the default
   if( !file_test(filename) ) {
      write,format="%s does not exist\n", filename;
      return [];
   }
   // read the contents before it might be destroyed
   ptr = rdfitsbin( dol, fh, nrows );
   if( is_void(ptr) ) {
      write,format="%s has no rows, skip ...\n", dol;
      return [];
   }

   if( !check ) {
      if( is_void(outfile) ) outfile = filename;
      if( file_test(outfile) ) {
         if( clobber ) {
            remove, outfile;
         } else {
            write,format="%s exists already, quit\n", outfile;
            return;
         }
      }
   }


   // get reference catalog if not already in memory

   read_catalog, force=force, clog=clog;

   // localize specific columns in the source list

   nnam = fits_colnum( fh, "name" );
   if( is_void(nnam) ) error,"Column NAME is missing";
   nra  = fits_colnum( fh, "ra_obj" );
   if( is_void(nra) ) error,"Column RA_OBJ is missing";
   ndec = fits_colnum( fh, "dec_obj" );
   if( is_void(ndec) ) error,"Column DEC_OBJ is missing";
   nsrc_id = fits_colnum( fh, "source_id" );
   if( is_void(nsrc_id) ) error,"Column SOURCE_ID is missing";
   nracat  = fits_colnum( fh, "ra_cat" );
   catflag = is_void(nracat) ? 0 : 1;
   if(catflag) ndeccat = fits_colnum( fh, "dec_cat" );
   nclass = fits_colnum( fh, "class" );
   classflag = is_void(nclass) ? 0 : 1;


   eq_nocopy, nam, *ptr(nnam);
   eq_nocopy, ra , *ptr(nra);
   eq_nocopy, dec, *ptr(ndec);
   if( catflag ) {
      eq_nocopy, racat, *ptr(nracat);
      eq_nocopy, deccat, *ptr(ndeccat);
   }
   if( classflag ) {
      eq_nocopy, class, *ptr(nclass);
   }

   // Get SOURCE_ID column with correct string length
   // for proper update of source list
   eq_nocopy, src_id, *ptr(nsrc_id);
   tform = fxpar( fh, swrite(format="tform%i", nsrc_id));
   if( strlen(tform) > 1 ) {
      srclen = atoi(strpart(tform,1:-1));
   } else error,"Bad SOURCE_ID TFORM format";

   // Get correct string length for NAME column
   // for proper update of source list
   tform = fxpar( fh, swrite(format="tform%i", nnam));
   if( strlen(tform) > 1 ) {
      namlen = atoi(strpart(tform,1:-1));
   } else error,"Bad NAME TFORM format";

 /*************
  *
  *   The default operation mode is to walk through the input
  *   source list in order to find the catalog source that
  *   is closest and then decide if it is close enough. If so,
  *   the identification is done.
  *   This approach is useful when the input source list can
  *   have multiple representations of the same physical source.
  *
  *   On the other hand, when it is known (or assumed) that the
  *   input list consists of unique sources then each catalog
  *   source can only be assigned once. In this case (indicated
  *   by the setting of the keyword 'unique') it is better to
  *   walk through the source catalog and for each source
  *   find the best matching source from the source list.
  *
  **************/

   if( unique ) {  // Walk through the reference source catalog
                 // to see if a source from the given list matches
                 // the known source position

      // reset all source names and source_ids

      snew = strpadd("NEW SOURCE", namlen, " ");
      for( i = 1; i <= nrows; i++ ) (*ptr(nnam))(i) = snew;
      srid = strpadd("UNKNOWN", srclen, " ");
      for( i = 1; i <= nrows; i++ ) (*ptr(nsrc_id))(i) = srid;

      // 'rflag' is used to mark the sources where an identification has
      // been found. Set to 180. for those and add to 'r' before sorting
      // making sure that it will not be close again

      rflag = array(0.0,nrows);

      for( i = 1; i <= Source_cat_num; i++ ) {
         r = arcdist( Source_cat(i).ra_obj, Source_cat(i).dec_obj, ra, dec );
         is = sort(r + rflag);
         if( r(is(1)) < rad ) { // an identification has been found
            // alright, this is a candidate, but there might be a better match
            rr = arcdist(Source_cat.ra_obj, Source_cat.dec_obj,ra(is(1)),dec(is(1)));
            if( rr(min) < r(is(1)) ) continue; // continue steeping through
               // refr cat sources; sooner or later the best one will be found
            rflag(is(1)) = 180.;
            name = strpadd(Source_cat(i).name, namlen, " ");
            s_id = strpadd(Source_cat(i).source_id, srclen, " ");
            (*ptr(nnam))(is(1)) = name;
            (*ptr(nsrc_id))(is(1)) = s_id;
            if( catflag ) {
               (*ptr(nracat))(is(1)) = Source_cat(i).ra_obj;
               (*ptr(ndeccat))(is(1)) = Source_cat(i).dec_obj;
            }
            if( classflag ) {
               (*ptr(nclass))(is(1)) = Source_cat(i).class;
            }
         }
      }

   } else { //  Walk through the input source list

      // make sure to keep original string length
      for( i = 1; i <= nrows; i++ ) {
         nearest = nearest_source( ra(i), dec(i), rad );
         if( is_void(nearest) ) { // No catalog object within given radius
            name = strpadd("NEW SOURCE", namlen, " ");
            s_id = strpadd("UNKNOWN", srclen, " ");
         } else { // A source has been found
            name = strpadd(strpart(nearest.name,1:namlen),namlen," ");
            s_id = strpadd(strpart(nearest.source_id,1:srclen),srclen," ");
         }
         if( chat > 0 ) {
            write,format="Row#%i ID: %s, NAME: %s\n", i, s_id, name;
         }
         if( check ) {
            if( strtrim(name) != strtrim(nam(i)) ) {
               write,format="Row#%i, %s to change to %s\n", \
                       i, strtrim(nam(i)), strtrim(name);
            }
         } else {
            //+ while(strlen(name) < namlen ) name += " ";
            //+ while(strlen(s_id) < srclen ) s_id += " ";
            (*ptr(nnam))(i) = name;
            (*ptr(nsrc_id))(i) = s_id;
            if( catflag ) {
               if( is_void(nearest) ) {
                  (*ptr(nracat))(i) = float(-99.0);
                  (*ptr(ndeccat))(i) = float(-99.0);
               } else {
                  (*ptr(nracat))(i) = float(nearest.ra_obj);
                  (*ptr(ndeccat))(i) = float(nearest.dec_obj);
               }
            }
            if( classflag ) {
               if( is_void(nearest) ) {
                  (*ptr(nclass))(i) = 0n;
               } else {
                  (*ptr(nclass))(i) = int(nearest.class);
               }
            }
         }
      }

   }

   kwds_init;
   if( typeof(outfile) == "string" ) kwds_set,"ORIGIN", dol, "Original source list";
   kwds_set,"QDATE",ndate(3),"Updated with qidsrcs";
   kwds_set,"Q_RAD", rad, "[deg] Acceptance radius for qidsrcs";
   kwds_set,"REFR_CAT", Source_cat_dol,"Reference catalog of X-ray sources";
   mode = unique ? "UNIQUE_MOD" : "SIMPLE_MOD";
   kwds_set,"Q_MODE", mode, "Mode for qidsrcs id process";

   fh = kwds_put( fh );
   if( !check ) wrfitsbin, outfile, fh, ptr;

}

/* Function find_src_by_name */

func find_src_by_name( name, force=, clog= )
/* DOCUMENT res = find_src_by_name( name, force=, clog= )

   Returns struct array with the elements:
       ra_obj, dec_obj, err_rad, name, source_id, class
   for each source with a name that matches the input name.

      [See external 'Source_cat' that is an array of struct 's_Src'
       created by _read_catalog.]

  Keyword  force:   Force reading the catalog contents
           clog :   The DOL of the catalogue

   SEE ALSO: nearest_source, nearby_sources

  2007-10-26/NJW
*/

{

   extern Source_cat, Source_cat_dol, Source_cat_num;

   if( dimsof(name)(1)  !=  0 ) {
      write,"Input name must be scalar";
      return [];
   }
   if( typeof(name) != "string" ) {
      write,"Input name must be string";
      return [];
   }

   read_catalog, force=force, clog=clog;

   w = where( strglob( strcompress(name,all=1),  \
               strtrim(strcompress(Source_cat.name,all=1)), case=0));

   nw = numberof(w);

   if( nw == 0 ) {
      write,"No matching source found";
      return [];
   }

   if( am_subroutine() ) {
      write,"  # Name                     RA        DEC";
      write,"                             deg       deg";
      for( i = 1; i <= nw; i++ ) {
         write,format="%4i %s   %8.4f %8.4f\n", \
            w(i),Source_cat(w(i)).name, Source_cat(w(i)).ra_obj, \
            Source_cat(w(i)).dec_obj;
      }
   } else {
      if( nw == 1 ) return Source_cat(w(1));
      else return Source_cat(w);
   }

}

/* Function find_swid_galac */

func find_swid_galac( lon0, lat0, lonw, latw, offax=, rev=, tstart=,
                      tstop=, list=, nof=, silent=, lst= )
/* DOCUMENT list = find_swid_galac( lon0, lat0, lonw, latw2[, offax=]
                       [,rev=][, tstart=][, tstop=][, list=]
                       [,nof=][, silent=][, lst=] )

   Find the INTEGRAL SWIDs where the angular distance of the pointings
   is less than lonw from lon0 in longitudinal direction and similarly
   for the latitude: A belt is selected e.g.
       lon0 = 0, lat0 = 0, lonw = 30., latw = 10
   will include pointings with longitude 0 - 30 deg and 330 - 360 deg
   in the latitude interval -10 -> 10 deg.

    2007-11-16/NJW  copied from find_swid_radec

   Keywords:
             offax:  When set the result will be sorted
                     by increasing off-axis angle
             rev:    Either scalar integer or 2 element integer
                     array with min and max revolution number
             tstart: Time in IJD for start of interval
             tstop:  Time in IJD for end of interval
             list :  Makes returned list a list of SWIDs (default)
             lst  :  Makes returned list a list of SWIDs in jemx.lst format
                         (overrides setting of 'list')
             nof  :  Flag to avoid file writing
*/
{


   local swid, ra, dec, posangle, ut, ijd, telapse, mode1, mode2;

   if( is_void(latw) ) {
      write,"Syntax: list = find_swid_galac(lon,lat,lonw,latw, keywords... )";
      return [];
   }
   //
   //  Initialize
   //
   if( is_void(silent) ) silent = 0;
   first = 1;
   tbegin = is_void( tstart ) ? 0.0 : tstart;
   // if( tstart ) { tbegin = tstart; } else { tbegin = 0.0; }
   tend = is_void( tstop ) ? 1.0e9 : tstop;
   // if( tstop ) { tend = tstop; } else { tend = 1.0e9; }
   lon0 = double(lon0);
   lat0 = double(lat0);
   lonw = double(lonw);
   latw = double(latw);
   radec = equatorial( lon0, lat0 );
   ra0 = radec(1);
   dec0 = radec(2);

   //
   // Get list of pointing files
   //
   base = get_env("J_POINTINGS");
   if( numberof(rev) ) {
      if( typeof(rev) == "string" ) {
         revarr = str2arr( rev );
      } else {
         if( numberof( rev ) == 2 ) {
            revarr = indgen(long(rev(2)-rev(1))+1) + long(rev(1)) - 1;
         } else {
            revarr = long(rev)
         }
      }
      nrev = numberof(revarr);
      for( i = 1; i <= nrev; i++ ) {
         revstr = swrite(revarr(i),format="%4.4i");
         tmp = base+"/pointings_"+revstr+"p.dat";
         if( !silent ) write,tmp;
         ybase = base+"/ysav"+swrite(revarr(i)/100,format="%02i/");
         ytmp = ybase+"pointings_"+revstr+"p.ysav";
         if( !silent ) write,ytmp;
         if( i == 1 ) {
            pf_list = tmp;
            ypf_list = ytmp;
         } else {
            grow,pf_list,tmp;
            grow,ypf_list,ytmp;
         }
      }
   } else {
      pf_list = file_search("pointings_*p.dat", base);
      ypf_list = pf_list;
      npf_list = numberof(pf_list);
      for( i = 1; i <= npf_list; i++ ) {
         pos = strpos( pf_list(i),"_" );
         revstra = strpart(pf_list(i), pos+1:pos+4 );
         drevstra = strpart(pf_list(i),pos+1:pos+2);
         ybase = base+"/ysav"+drevstra;
         ytmp = ybase+"/pointings_"+revstra+"p.ysav";
         ypf_list(i) = ytmp;
      }
   }
   npf_list = numberof(pf_list);
   if( npf_list == 0 ) {
      write,"find_swid_galac error: no pointing files of required kind found";
      return -1;
   }

   if( !nof ) {

      // Open output file

      outfilename = get_next_filename("find_swid_galac_???.dat");
      fout = open(outfilename,"w");
      n = write(fout,format="%s\n","//");
      n = write(fout,format="//        %s\n", ndate(2));
      n = write(fout,format="// lon0  = %7.3f; deg\n", lon0);
      n = write(fout,format="// lat0 = %7.3f; deg\n", lat0);
      n = write(fout,format="// ra0  = %7.3f; deg\n", ra0);
      n = write(fout,format="// dec0 = %7.3f; deg\n", dec0);
      n = write(fout,format="// lonw = %7.3f; deg\n", lonw);
      n = write(fout,format="// latw = %7.3f; deg\n", latw);
      if( tstart ) {
         n = write(fout,format="// tstart = %14.8f; IJD\n", tstart);
      }
      if( tstop ) {
         n = write(fout,format="// tstop = %14.8f; IJD\n", tstop);
      }
      if( numberof(rev) ) {
         if( nrev == 1 ) {
            n = write(fout,format="// rev_begin = %4i\n", revarr(1));
            n = write(fout,format="// rev_end   = %4i\n", revarr(1));
         } else {
            n = write(fout,format="// rev_begin = %4i\n", revarr(1));
            n = write(fout,format="// rev_end   = %4i\n", revarr(0));
         }
      }

   }

   ntotal = 0;
   for( ipf_list = 1; ipf_list <= npf_list; ipf_list++ ) {

      if( file_test(ypf_list(ipf_list)) ) {
         if( !silent ) write,"Reading binary file "+ypf_list(ipf_list);
         bfile = openb( ypf_list(ipf_list) );
         restore,bfile, swid, ra, dec, posangle, ut, ijd, telapse, \
                     mode1, mode2;
         close, bfile;
      } else {
         if( !silent ) write,"Reading usual text file "+pf_list(ipf_list);
         if( ! file_test(pf_list(ipf_list)) ) {
            if( !silent ) write,pf_list(ipf_list)+" was not found";
            continue;
         }

      /*
       * swid = rscol( pf_list(ipf_list),1, str=1, silent=1);
       * ra = rscol( pf_list(ipf_list), 2, silent=1);
       * dec = rscol( pf_list(ipf_list), 3, silent=1);
       * posangle = rscol( pf_list(ipf_list), 4, silent=1);
       * ut = rscol( pf_list(ipf_list), 5, str=1, silent=1);
       * ijd = rscol( pf_list(ipf_list), 6, silent=1);
       * telapse = rscol( pf_list(ipf_list),7, silent=1);
       * mode1 = rscol( pf_list(ipf_list), 8, str=1, silent=1);
       * mode2 = rscol( pf_list(ipf_list), 9, str=1, silent=1);
       */

         rstab,pf_list(ipf_list), 9, swid, ra, dec, posangle, ut, \
              ijd, telapse, mode1, mode2, typ="sfffsffss",silent=1;

         bfile = createb( ypf_list(ipf_list) );
         save, bfile,  swid, ra, dec, posangle, ut, ijd, telapse, \
                 mode1, mode2;
         close, bfile;
      }

      npointings = numberof( ra );

      galac = galactic( ra, dec );
      zero = array(0.0, npointings );
      dist_lon = arcdist( lon0, 0., galac(,1), zero );
      dist_lat = arcdist( 0., lat0, zero, galac(,2) );

      dist = arcdist( ra0, dec0, ra, dec );
      w = where( dist_lon < lonw & dist_lat < latw );
      nw = numberof(w);

      if( nw > 0 ) {

         if( first ) {
            first = 0;
            swid_all = swid(w);
            ra_all = ra(w);
            dec_all = dec(w);
            posangle_all = posangle(w);
            ut_all = ut(w);
            ijd_all = ijd(w);
            telapse_all = telapse(w);
            mode1_all = mode1(w);
            mode2_all = mode2(w);
            dist_all = dist(w);
         } else {
            grow,swid_all,swid(w);
            grow,ra_all,ra(w);
            grow,dec_all,dec(w);
            grow,posangle_all,posangle(w);
            grow,ut_all,ut(w);
            grow,ijd_all,ijd(w);
            grow,telapse_all,telapse(w);
            grow,mode1_all,mode1(w);
            grow,mode2_all,mode2(w);
            grow,dist_all,dist(w);
         }

         ntotal += nw;

      }

   }
   if( ntotal == 0 ) {
      write,"Sorry, no pointings found!";
      return [];
   }

   if( offax ) {
      is = sort(dist_all);
      swid_all = swid_all(is);
      ra_all = ra_all(is);
      dec_all = dec_all(is);
      posangle_all = posangle_all(is);
      ut_all = ut_all(is);
      ijd_all = ijd_all(is);
      telapse_all = telapse_all(is);
      mode1_all = mode1_all(is);
      mode2_all = mode2_all(is);
      dist_all = dist_all(is);
   }

   noutput = 0;
   list = "";
   lfirst = 1;
   for( i = 1; i <= ntotal; i++ ) {
      if( ijd_all(i) > tbegin && ijd_all(i) < tend ) {

         if( lfirst ) {
            list_list = [swid_all(i)];
            lfirst = 0;
         } else grow,list_list,swid_all(i);
         if( !nof) {
          n = write(fout,format="%s%10.4f%9.4f%9.3f  %s%10.3f%6.0f  %s  %s %8.2f\n",
            swid_all(i),ra_all(i),dec_all(i),posangle_all(i),
            ut_all(i),ijd_all(i),telapse_all(i),
            mode1_all(i),mode2_all(i),dist_all(i));
         }
         noutput++;
      }
   }

   lst_list = array("",ntotal);
   for( i = 1; i <= ntotal; i++ ) {
      lst_list(i) = "./scw/"+strpart(swid_all(i),1:4)+"/"+swid_all(i)+".001/swg.fits[1]";
   }

   if( !nof ) close,fout;

   if( !silent ) {
      write,format="Found %i SWIDs in selected revolutions\n", ntotal;
      if( noutput == 1 ) {
         write,format="%i has been selected by time\n", noutput;
      } else {
         write,format="%i have been selected by time\n", noutput;
      }
      if( !nof ) write,"Output is directed to "+outfilename;

   }

   if( !is_void(lst) ) { return lst_list; } else {return list_list;}

}

/* Function j_mk_vignet */

func j_mk_vignet( imDims, radiusLimit, im_cent_x, im_cent_y, scale= )
/* DOCUMENT vignet = j_mk_vignet( imDims, radiusLimit, im_cent_x, im_cent_y, scale= )

   Returns a vignetting array with dimensions 'imDims' e.g. [2,790,790]
   (result of dimsof(image)).
   'radiusLimit' is the radius of the (circular) detector area used, in mm.
   The image must have pixels of 1.00 mm. Otherwise the 'scale' keyword
   must be given as pixel size in mm.

   2007-11-23/NJW, copied from jic_mk_vignet.i
*/
{


   if( is_void(im_cent_y) ) {
      write,"Syntax: vignet = j_mk_vignet( imDims, radiusLimit, im_cent_x, im_cent_y )";
      return [];
   }

   if( numberof(imDims) != 3 ) {
      write,"j_mk_vignet: First argument must be 3-element array";
      return [];
   }

   if( is_void(scale) ) scale = 1.0;

   rad2deg = 180.0 / pi;
   im_cent_x = double(im_cent_x);
   im_cent_y = double(im_cent_y);
   vignet = array(double,imDims);
   // Make pixel distances in mm
   d_mm = distances(imDims(2),imDims(3),im_cent_x,im_cent_y)*scale;
   for( i = 1; i <= imDims(2); i++ ) {
   for( j = 1; j <= imDims(3); j++ ) {
      offaxis = atan( d_mm(i,j) / 3401. ) * rad2deg; // Detector Mask distance is 3401 mm
      if( i == im_cent_x && j == im_cent_y ) {
         vignet(i,j) = 1.0;
      } else {
         omega = atan(abs(j - im_cent_y), abs(i - im_cent_x)) * rad2deg;
         vignet(i,j) = j_thruput( offaxis, radiusLimit, omega );
      }
   }
   }

   return vignet;
}

/* Function get_revol_start_time */

func get_revol_start_time( revol, fmt= )
/* DOCUMENT res = get_revol_start_time( revol, fmt= )

   Keyword fmt:   "ijd" (default) or "ymd"

   2009-04-17/NJW
*/
{
   dol = "/r6/jemx/integral_history/revolution_latest.fits+1";

   revol_no = double(rdfitscol(dol, "REVOLUTION"));
   time_perigee = rdfitscol(dol, "TIME_PERIGEE");
   nrev = numberof(revol_no);

   w = where( revol_no == revol );
   nw = numberof(w);

   if( nw != 1 ) {
      write,"Exceeded revol information";
      return [];
   }
   if( is_void(fmt) ) fmt = "ijd";
   res =  fmt == "ymd" ? ijd2dattim(time_perigee(w(1))) : time_perigee(w(1));
   if( am_subroutine() ) {
      if( fmt == "ymd" ) {
         write,format="Time of perigee: %s\n", res;
      } else {
         write,format="Time of perigee (IJD) : %11.6f\n", res;
      }
   } else return res;
}
   

/* Function get_revol_no */

func get_revol_no( time, frac= )
/* DOCUMENT revol = get_revol_no( time, frac= )

   Use /r6/jemx/integral_history/revolution_latest.fits to return
   revolution number for a given time

   The time parameter can be given as IJD or as string 'YYYY-MM-DDTHH:MM:SS' (UTC)

   2004-06-04/NJW
   2005-08-16/NJW updated with fractional result
   2007-12-11/NJW translated to Yorick

*/
{
   if( is_void(time) ) {
      write,"Syntax: revol = get_revol_no( time, frac= )";
      write,"   'time' can be given as IJD or as string 'YYYY-MM-DDTHH:MM:SS'";
      write,"   If keyword 'frac' is set then it returns revolution with fraction";
      return [];
   }

   // Determine the type of input

   ijd = typeof(time) == "string" ? dattim2ijd(time) : time;

   dol = "/r6/jemx/integral_history/revolution_latest.fits+1";

   revol_no = double(rdfitscol(dol, "REVOLUTION"));
   time_perigee = rdfitscol(dol, "TIME_PERIGEE");
   nrev = numberof(revol_no);

   // The TIME_PERIGEE refers to the beginning of the revolution

   if( ijd < time_perigee(1) - 1.0 )  return -1;  // Definitely before INTEGRAL mission
   if( ijd > time_perigee(0) )  return -999;  // After highest revolution given

   if( ijd < time_perigee(1) ) { // Reset to time_perigee(1) plus epsilon
      ijd = time_perigee(1) + 0.001;
   }

   w = where( time_perigee < ijd );
   nw = numberof(w);

   if( frac ) {
      fraction = (ijd - time_perigee(w(0))) \
         / (time_perigee(w(0)+1) - time_perigee(w(0)));
      return  revol_no(w(0)) + fraction;
   } else {
      return  revol_no(w(0));
   }

}

/* Function j_cor2raw */

func j_cor2raw( jemxNum, corx, cory, &rawx, &rawy )
/* DOCUMENT j_cor2raw, jemxNum, corx, cory, >rawx, >rawy

    Procedure to get raw positions from corrected ones

    2003-04-17/NJW original version
    2004-08-19/NJW This "procedure" version
    2007-12-21/NJW Translated to Yorick

*/

{
   extern Njmx_com, Acorx, Acory;

   if( is_void( cory ) ) {
      write,"Syntax: j_cor2raw, jemxNum, corx, cory, >rawx, >rawy";
      return [];
   }

   reread = 0;
   if( is_void(Njmx_com) ) {
      reread = 1;
   } else {
      if( jemxNum  !=  Njmx_com ) reread = 1;
   }

   if( reread ) {
      write,"Reading correction matrices ...";
      if( jemxNum  ==  1 ) {
         Acorx = readfits("/r6/jemx/ic/imod_grp/jmx1_imod_grp_0117.fits+22");
         Acory = readfits("/r6/jemx/ic/imod_grp/jmx1_imod_grp_0117.fits+23");
      } else {
         Acorx = readfits("/r6/jemx/ic/imod_grp/jmx2_imod_grp_0117.fits+22");
         Acory = readfits("/r6/jemx/ic/imod_grp/jmx2_imod_grp_0117.fits+23");
      }
      Njmx_com = jemxNum;
   }

   //   Define outline of squares

   dx1 = [-1,0,1,1,1,0,-1,-1];
   dy1 = [-1,-1,-1,0,1,1,1,0];
   dx2 = [-2,-1,0,1,2,2,2,2,2,1,0,-1,-2,-2,-2,-2];
   dy2 = [-2,-2,-2,-2,-2,-1,0,1,2,2,2,2,2,1,0,-1];
   dx3 = [-3,-2,-1,0,1,2,3,3,3,3,3,3,3,2,1,0,-1,-2,-3,-3,-3,-3,-3,-3];
   dy3 = [-3,-3,-3,-3,-3,-3,-3,-2,-1,0,1,2,3,3,3,3,3,3,3,2,1,0,-1,-2];

   num = numberof(corx);
   rawx = array(int,num);
   rawy = array(int,num);

   for( j = 1; j <= num; j++ ) {

   //  approximate result

      rx0 = toint(corx(j) + 127);
      ry0 = toint(cory(j) + 124);
      rx = rx0; // This is index with start zero
      ry = ry0;
   
      found = 0;
      if( abs(corx(j) - Acorx(rx+1,ry+1))  <  1.e-4 \
             && abs(cory(j) - Acory(rx+1,ry+1))  <  1.e-4 ) {
         found = 1;
      } else { // search inner circle
         for( i = 1; i <= 8; i++ ) {
            rx = rx0 + dx1(i);
            ry = ry0 + dy1(i);
            if( 0  <=  rx && rx  <=  255 && 0  <=  ry && ry  <=  255 ) {
               if( abs(corx(j) - Acorx(rx+1,ry+1))  <  1.e-4 \
                        && abs(cory(j) - Acory(rx+1,ry+1))  <  1.e-4 ) {
                  found = 1;
                  break;
               }
            }
         }
      }
   
      if( ! found ) { // search next ring
         for( i = 1; i <= 16; i++ ) {
            rx = rx0 + dx2(i);
            ry = ry0 + dy2(i);
            if( 0  <=  rx && rx  <=  255 && 0  <=  ry && ry  <=  255 ) {
               if( abs(corx(j) - Acorx(rx+1,ry+1))  <  1.e-4 \
                           && abs(cory(j) - Acory(rx+1,ry+1))  <  1.e-4 ) {
                  found = 1;
                  break;
               }
            }
         }
      }
   
   
      if( ! found ) { // search next again
         for( i = 1; i <= 24; i++ ) {
            rx = rx0 + dx3(i);
            ry = ry0 + dy3(i);
            if( 0  <=  rx && rx  <=  255 && 0  <=  ry && ry  <=  255 ) {
               if( abs(corx(j) - Acorx(rx+1,ry+1))  <  1.e-4 \
                           && abs(cory(j) - Acory(rx+1,ry+1))  <  1.e-4 ) {
                  found = 1;
                  break;
               }
            }
         }
      }
   
      if( found ) {
         rawx(j) = rx;
         rawy(j) = ry;
      } else {
         rawx(j) = 255;
         rawy(j) = 255;
      }
   
   
   }

   if( num == 1 ) {
      rawx = rawx(1);
      rawy = rawy(1);
   }

}

/* Function mk_swid_lists */

func mk_swid_lists( ra, dec, rad1, rad2, rev= )
/* DOCUMENT mk_swid_lists, ra, dec, rad1[ , rad2], rev=

   Make SWID lists for both JMX1 and JMX2 where position
   (ra,dec) is inside radius 'rad1' (if 'rad2' is NOT given),
   and between 'rad1' and 'rad2' if 'rad2' has been given.
   See help for 'find_swid_radec'.

   Keyword
       rev     Same as for 'find_swid_radec'

   2019-07-02/NJW
*/
{
   list = find_swid_radec( ra, dec, rad1, rad2, rev=rev, nof=1, list=1,silent=1);
   list = list(sort(list));

   for( j = 1; j <= 2; j++ ) {
      jstr = "jmx"+itoa(j);
      listv = validate_swid_list( list, 1, silent=1 );
      if( numberof(listv) ) {
         write_slist,jstr+"_swid.list", listv;
         write,numberof(listv),jstr,format="%i SWIDs in %s_swid.list\n";
      } else write,"No SWIDs survived the validation for "+strupcase(jstr);
   }

}

/* Function build_swid_time_table */

func build_swid_time_table( topdir, outfile, clobber= )
/* DOCUMENT build_swid_time_table, topdir, outfile, clobber=

  Build a table of start and stop IJD times of pointings.

  'topdir' points to the directory under which the JEM-X data
           can be found for the given revolution.
  'outfile' is the path and name for the output table.

   If 'outfile' is a four character string then the filename will be set to
      /r6/jemx/integral_history/timetables/swid_timetable_RRRR.fits
      where RRRR is <outfile>

   If 'topdir' is a four character string ("RRRR") and 'outfile' is missing
   then standard values are assumed: /jemx/arc/rev_3/scw/RRRR and
   /r6/jemx/integral_history/timetables/swid_timetable_RRRR.fits

   In case the timetable file already exists the user will be asked if it
   should be overwritten. This is cancelled with keyword 'clobber'.

     2007-08-17/NJW copied from build_swid_time_table.pro
*/

{
   if( is_void(outfile) ) {
      if( typeof(topdir) != "string" ) topdir = itoa(topdir,4);
      if( strlen(topdir) == 4 ) {
         outfile = topdir;
         topdir = "/jemx/arc/rev_3/scw/"+topdir;
      } else {
         write,"Missing the second argument.";
         return;
      }
   }
   if( strlen(outfile) == 4 ) {
      outfilx = "/r6/jemx/integral_history/timetables/swid_timetable_"+outfile+".fits";
   } else outfilx = outfile;
  /*
   *  Check existence of previous version of table;
   */
   if( file_test(outfilx) ) {
      write,"Will update: "+outfilx;
      if( clobber ) {
         remove, outfilx;
      } else {
         //  Old version exists, ask what to do
         ans = rdline(prompt="File exists, stop (s) or remove (r) ? ... ");
         if( ans == "r" ) {
            remove, outfilx
         } else {
            write,"Skip further action ...";
            return;
         }
      }
   }
  /*
   * Get list of all SWGs below "topdir";
   */
   swg_name = "swg.fits";
   list = file_rsearch( swg_name, topdir );
   nlist = numberof( list );
   write,format="Found %i files by name of: %s\n", nlist, swg_name;

   if( nlist  ==  0 ) {
      write,format="Sorry, no SWGs in and below %s\n",topdir;
      return;
   }

   list = list(sort(list));

   tstart = array(double, nlist);
   tstop  = tstart;
   swid   = array( string, nlist);
   ra_scx = tstart;
   dec_scx= tstart;
   ra_scz = tstart;
   dec_scz= tstart;
   posangle = tstart;

   for( i = 1; i <= nlist; i++ ) {

      write,format="%s,  %i of %i\n", list(i), i, nlist;
      hdr = headfits(list(i)+"+1",nocheck=1);
      tstart(i) = fxpar(hdr,"TSTART");
      tstop(i) = fxpar(hdr,"TSTOP");
      swid(i) = fxpar(hdr,"SWID");
      ra_scx(i) = fxpar(hdr,"RA_SCX");
      dec_scx(i)= fxpar(hdr,"DEC_SCX");
      ra_scz(i) = fxpar(hdr,"RA_SCZ");
      dec_scz(i)= fxpar(hdr,"DEC_SCZ");
      posangle(i) = fxpar(hdr,"POSANGLE");

   }

   wrmfitscols, outfilx,"SWID",swid,"TSTART",tstart, \
                "TSTOP",tstop,"RA_SCX",ra_scx,"DEC_SCX",dec_scx, \
                "RA_SCZ",ra_scz,"DEC_SCZ",dec_scz, \
                "POSANGLE",posangle;

}

/* Function j_src_on_shadowgram */

func j_src_on_shadowgram( ra_scx, dec_scx, posangle, ra_src, dec_src )
/* DOCUMENT arr = j_src_on_shadowgram( ra_scx, dec_scx, posangle, ra_src, dec_src )

   Returns a 256x256 map with 1 (one) where the source can hit
   the detector. All other pixels are zero.

   NB: Currently only for JMX1

   2008-05-27/NJW

*/
{
   r = j_get_scz( ra_scx, dec_scx, posangle, deg=1 );
   ra_scz = r(1);
   dec_scz = r(2);

   srcvec = vector( ra_src, dec_src, deg=1 );
   xvec = vector( ra_scx, dec_scx, deg=1 );
   zvec = vector( ra_scz, dec_scz, deg=1 );
   yvec = crossprod( zvec, xvec );

   mask_center = [sum(yvec*srcvec),sum(zvec*srcvec)]*(-3401.);
   offset = [128.,124.];

   d = distances(256,256,mask_center(1)+offset(1), \
                         mask_center(2)+offset(2) );
   res = array(long,256,256);
   w = where( d < 267.5 );
   if( numberof(w) > 0 ) {
      res(w) = 1;
   }
   return res;

}

/* Function j_get_scz */

func j_get_scz( ra_scx, dec_scx, posangle, deg= )
/* DOCUMENT res = j_get_scz( ra_scx, dec_scx, posangle, deg= )

   Returns the ra_scz and dec_scz in a two-element array.
   By default the unit is radian, but setting the keyword 'deg'
   will change to degrees (both input and output).

   2008-05-27/NJW
*/
{
   dr = deg ? pi/180.0 : 1.0;
   ax = ra_scx * dr;
   dx = dec_scx * dr;
   Theta = posangle * dr;

   cT = cos(Theta);
   dz = asin(cT * cos(dx));
   B = atan( sin(Theta), -sin(dx)*cT );

   az = zero2pi(ax - B);
   return [ az, dz ]/dr;

}

/* Function j_collect_spectra */

func j_collect_spectra( ref_sourceid, spec_path=, srcid2=, jemxNum=, \
                     flist=, outfile=, outarf=, noplot=, proj= )
/* DOCUMENT j_collect_spectra, ref_sourceid, spec_path=, srcid2=, jemxNum=, \
                     flist=, outfile=, outarf=, noplot=, proj=

   Get several spectra combined from data produced with j_src_spectra i.e.
   including mandatory keywords, ARF etc. assuming the naming convention
   jmxi_srcl_spe_proj_*.fits, where 'proj' is defined by keyword 'proj'.
   If 'proj' is not given, then jmxi_srcl_spe.fits is assumed (and then
   in various sub-directories).

    2003-04-11/NJW
    2008-01-17/NJW Translated from IDL to Yorick

 Arguments:
   ref_sourceid = "crab"
       will translate to ref_sourceid = "J053432.0+220052";
   ref_sourceid = "J195821.7+351206" ; Cyg X-1
   ref_sourceid = "J171958.0-314648"  ; XTE1720-318
   ref_sourceid2 = "XTE1720-318"

 Keywords:
   spec_path    directory under which the spectra can be found
   srcid2       alternative source name
   jemxNum      JEM-X unit
   flist        list of files to override 'spec_path'
   outfile       output spectral file
   outarf       output ARF file
          both outfile and outarf must be defined to have an effect
          else the output file will be jmxi_0000_spe.fits etc.
   noplot       to avoid plotting

*/
{

   if( is_void(jemxNum) ) jemxNum = 1;
   if( jemxNum  !=  1 && jemxNum  !=  2 ) {
      write,"Illegal JEMX number";
      return [];
   }
   jstr = "jmx"+itoa(jemxNum);

   if( ! numberof(flist) && is_void(spec_path) ) {
      write,"There is no indication of where to find the spectra ...";
      return [];
   }

   // See if both keywords "outfile" and "outarf" are set

   default_out_names = 1;
   if( numberof(outfile) ) {
      if( numberof(outarf) ) {
         default_out_names = 0; // don't use default output names
      } else {
         write,"Warning: only 'outfile' has been set - use default output names";
      }
   }

   pflag = is_void(noplot);

   // Get list of spectral files to include

   if( numberof(flist) ) {

      list = flist;

   } else {

      write,"Start building list  (takes some time) ...";
      if( is_void(proj) ) {
         searchname = jstr+"_srcl_spe.fits";
      } else {
         searchname = jstr+"_srcl_spe_"+proj+"_*.fits";
      }
      list = file_search( searchname, spec_path );

      if( numberof(list) == 0 ) error,"No spectral files found of name: "+searchname;

   }

   // Apply special case for Crab

   if( strtolower(ref_sourceid)  ==  "crab" ) ref_sourceid = "J053432.0+220052";

   nlist = numberof(list);
   list = list(sort(list));
   list_used = list;
   nlist_used = 0;


   // Prepare the output files jmxi_nnnn_spe.fits and jmxi_nnnn_arf.fits;
   // in the current directory but only if keywords outfile and outarf are set;

   if( default_out_names ) {

      local strnum;
      fname_spe = get_next_filename(jstr+"_????_spe.fits",strnum);
      fname_arf = strput( fname_spe, "arf", 11 );
      logfile = "collect_spectra_"+strnum+".log";

   } else {

      fname_spe = outfile;
      fname_arf = outarf;
      logfile = get_next_filename("collect_spectra_????.log");

   }

   //+ cp,"/home/njw/jemx/spectra/"+jstr+"_coll_spe.fits",fname_spe;
   //+ cp,"/home/njw/jemx/spectra/"+jstr+"_coll_arf.fits",fname_arf;

   tot_arf  = array(float,500);
   tot_expo = 0.0;

   write,"Output logging data in "+logfile;
   lug = open(logfile,"w");

   write,lug,"  Logging of collect_spectra   "+ndate(3);
   write,lug;
   write,lug,"Resulting SPE file: "+fname_spe;
   write,lug,"Resulting ARF file: "+fname_arf;

   first = 1; // used for checking number of spectral bins (nchans)
   firstsum = 1; // used for making the sums of spectra

   for( i = 1; i <= nlist; i++ ) {

      write,"Analyzing "+itoa(i)+" of "+itoa(nlist);
      write,lug,"-------------------------------";
      write,lug,list(i);
      // Note if path is absolute
      first_char = strpart(list(i),1:1); // to be used for ARF name
      // Get header
      if( i  >  1 ) respfile_prev = respfile_ori;
      hdr = headfits(list(i)+"+1");
      nsources = fxpar(hdr,"NAXIS2");
      if( nsources  <=  0 ) {
         write,"Skip - no sources";
         write,lug,"Skip this one, there are no sources";
         continue;
      }
      respfile_ori = fxpar(hdr,"RESPFILE");
      tok = strsplit(respfile_ori,"/");
      ntok = numberof(tok);
      jj = -1;
      for( j = 1; j <= ntok; j++ ) {
         if( strlen(tok(j))  >=  18 ) {
            p1 = strpos(tok(j),"jmx",1);
            p2 = strpos(tok(j),"rmf",1);
            p3 = strpos(tok(j),"fits",1);
            if( p1  >  0 && p2  >  0 && p3  >  0 ) {
               jj = j;
               break;
            }
         }
      }
      if( jj  >  0 ) {
         respfile = tok(jj);
         write,lug,"RESPFILE = "+respfile;
      } else {
         write,lug,"RESPFILE is unidentified";
         write,"Problem finding the RESPFILE";
      }
      if( i  >  1 ) {
         if( respfile_ori  !=  respfile_prev ) {
            write,"Warning - change of response file";
            write,lug,"Warning - change of response file";
         }
      }
      // read the source identifications
      sourceid = rdfitscol(list(i)+"+1","ROWID");
      exposure = rdfitscol(list(i)+"+1","EXPOSURE");
      ancrfile = rdfitscol(list(i)+"+1","ANCRFILE");
      ancrfile = strtrim(ancrfile);
      sourceid = strtrim(sourceid);
      write,lug,"Source IDs: "+sourceid;
      // locate actual source;
      if( numberof(srcid2) ) {
         w = where( sourceid  ==  ref_sourceid | sourceid  ==  srcid2);
      } else {
         w = where( sourceid  ==  ref_sourceid);
      }
      n_w = numberof(w);

      if( n_w  ==  1 ) {

         // include this one in the list of used spectral files;
         list_used(nlist_used) = list(i);
         nlist_used = nlist_used + 1;

         write,lug,"Actual source is number "+itoa(w(1))+" (starting with 1)";

         pick_swid_str, list(i), swid;
         swid = numberof(swid) ? swid(1) : "Unknown SWID";

         //+ tok = strsplit(list(i),"/",/extract);
         //+ ntok = numberof(tok);
         write,lug,swid+"  "+itoa(nsources)+" sources";
         write,lug,format="Exposure time is %10.2f s\n",exposure(w(1));

         // Get the spectrum etc.

         ra_obj =   rdfitscol(list(i)+"+1","RA_OBJ");
         dec_obj =  rdfitscol(list(i)+"+1","DEC_OBJ");
         rate =     rdfitscol(list(i)+"+1","RATE");
         stat_err = rdfitscol(list(i)+"+1","STAT_ERR");
         nch = dimsof(rate)(2);
         if( first ) {
            nchans = nch;
            first = 0;
         } else {
            if( nch != nchans ) error,"##1## varying number of channels";
         }
         if( pflag ) {
   	    dataplot,indgen(nchans),rate(,w(1)),stat_err(,w(1));

            if( jj  >=  1 ) xyouts,0.25,0.82,swid,charsize=1.3,ndc=1;
            xyouts,0.25,0.80,swrite(format="Exposure %10.2f s",exposure(w(1))), \
                      charsize=1.3,ndc=1;

         }

         //  --    Get the corresponding ARF file;

         // First check if ARF file name is absolute or relative

         ancr = strip_curly_br(strtrim(ancrfile(w(1))));
         if( strpart( ancr, 1:1 ) == "/" ) { // absolute path name, no change
            arffile = ancr;
         } else {  // we have a relative name, append to path found in 'list(i)'

            //+ tok = strsplit( list(i), "/" );
            dirlist = dirname( list(i) );
            //+ ntok = numberof(tok);
            //+ tok(ntok) = strip_curly_br(ancrfile(w(1)));
            //+ arffile = strjoin(tok, "/");
            //+ if( first_char  ==  "/" ) arffile = "/"+arffile;

            arffile = dirlist + "/" + ancr;
         }

         if( !file_test(arffile) ) {
            write,format="ARF is missing: %s\n", arffile;
            close,lug;
            return [];
         }
         arf = rdfitscol(arffile+"+1","SPECRESP");
         energ_lo = rdfitscol(arffile+"+1","ENERG_LO");
         energ_hi = rdfitscol(arffile+"+1","ENERG_HI");
         tot_arf  += reform(arf(,w(1)),500)*exposure(w(1));
         if( firstsum ) {
            tot_spec = float(reform(rate(,w(1)),nchans)*exposure(w(1)));
            tot_err2 = float((reform(stat_err(,w(1)),nchans)*exposure(w(1)))^2);
            firstsum = 0;
         } else {
            tot_spec += float(reform(rate(,w(1)),nchans)*exposure(w(1)));
            tot_err2 += float((reform(stat_err(,w(1)),nchans)*exposure(w(1)))^2);
         }
         tot_expo += exposure(w(1));
      } else write,lug,"Source was not found";

   }

   write,lug,"// total_exposure = ", tot_expo," ; s";
   write,lug,"//  --- Normal exit ---";

   close,lug;

   tot_arf  = tot_arf  / tot_expo;
   tot_spec = tot_spec / tot_expo;
   tot_err  = sqrt(tot_err2) / tot_expo;

   if( pflag ) dataplot,indgen(nchans),tot_spec,tot_err;

   //  Update standard file with collected spectrum;

   write,"New spectrum file: "+fname_spe;

 /*  ----------  remove
  *
  *fh = headfits( fname_spe+"+1" );
  *colnum = get_colnum( fh, "rowid" );
  *fits_bintable_poke, fname_spe+"+1", -1, colnum, sourceid(w(1));
  *colnum = get_colnum( fh, "ra_obj" );
  *fits_bintable_poke, fname_spe+"+1", -1, colnum, ra_obj(w(1));
  *colnum = get_colnum( fh, "dec_obj" );
  *fits_bintable_poke, fname_spe+"+1", -1, colnum, dec_obj(w(1));
  *colnum = get_colnum( fh, "rate" );
  *fits_bintable_poke, fname_spe+"+1",  1, colnum, float(tot_spec);
  *colnum = get_colnum( fh, "stat_err" );
  *fits_bintable_poke, fname_spe+"+1",  1, colnum, float(tot_err);
  *colnum = get_colnum( fh, "exposure" );
  *fits_bintable_poke, fname_spe+"+1",  1, colnum, double(tot_expo);
  *colnum = get_colnum( fh, "ancrfile" );
  *fits_bintable_poke, fname_spe+"+1", -1, colnum, [fname_arf+"{1}"];
  ********/

/******************
   //  Update the RESPFILE keyword && add the list;
   //  of input files;

   lut = open("tmpfil","w");
   write,lut,format="RESPFILE = '%s'\n","/r6/jemx/ic/rmf_grp/"+respfile;
   write,lut,format="SOURCEID = '%s'\n",ref_sourceid;
   for( i = 1; i <= nlist_used; i++ ) {
      s = list_used(i);
      len = strlen(s);
      while( len  >  70 ) {
         scut = strpart(s,1:70)+" $";
         write,lut,format="COMMENT %s\n", scut;
         s = strpart(s,71:9999);
         len = strlen(s);
      }
      write,lut,format="COMMENT %s\n",s;
   }

   close,lut;

   system,"fmodhead "+fname_spe+"+1 tmpfil=tmpfil";

   //  Update the checksum;

   system,"fchecksum "+fname_spe+" update=yes";

   //  Update file with collected average ARF;


   write,"New ARF file: "+fname_arf;
   colnum = get_colnum(fname_arf+"+1", "SPECRESP" );
   fits_bintable_poke, fname_arf+"+1", 1, colnum, float(tot_arf);
***************************/

   kwds_init;

   for( i = 1; i <= nlist_used; i++ ) {
      s = list_used(i);
      len = strlen(s);
      while( len  >  70 ) {
         scut = strpart(s,1:70)+" $";
         kwds_set,"COMMENT", scut;
         s = strpart(s,71:9999);
         len = strlen(s);
      }
      kwds_set,"COMMENT", s;
   }
   kwds_set,"DATE",ndate(3),"Date/time of file creation";

   spec2phaii, fname_spe, tot_spec, tot_err, type="net", exposure=tot_expo, \
       ancrfile=fname_arf, respfile=respfile_ori, rowid=ref_sourceid, \
       ra_obj=ra_obj(w(1)), dec_obj=dec_obj(w(1)), \
       telescop="INTEGRAL", instrume=strupcase(jstr), no_kwds_init=1;

   system,"fchecksum "+fname_spe+" update=yes";

   arf2phaii, fname_arf, tot_arf, energ_lo(,1), energ_hi(,1), telescop="INTEGRAL", \
              instrume=strupcase(jstr);

   //  Update the checksum;

   system,"fchecksum "+fname_arf+" update=yes";

   write,"New ARF file: "+fname_arf;

}

/* Function upd_ancrfile */

func upd_ancrfile( spec_file_list, ancrfile=, chat= )
/* DOCUMENT upd_ancrfile, spec_file_list, ancrfile=, chat=

   No keyword 'ancrfile' given:
   Assuming that spectral file names are given as *_srcl_spe_*.fits
   then the ANCRFILE information is updated to *_srcl_arf_*.fits in
   the same directory as the spectral file.

   Else when keyword 'ancrfile' has been given:
   Update to the given file name.

   2008-08-08/NJW
   2011-07-02/NJW, updated with keyword ancrfile
*/
{
   local dirname, basename;

   nlist = numberof(spec_file_list);

   if( typeof(ancrfile) == "string" ) {

      for( i = 1; i <= nlist; i++ ) {
         specfile = spec_file_list(i);
         if( chat ) write,"Update ANCRFILE in "+specfile;
         arf = rdfitscol( specfile+"+1","ancrfile");
         narf = numberof(arf);
         arfs = array( ancrfile, narf );
         for( j = 1; j <= narf; j++ ) arfs(j) += "{"+itoa(j)+"}";
         fits_bintable_poke, specfile+"+1", 0, "ancrfile", arfs;
      }

   } else { // no ancrfile name given, compose one

      for( i = 1; i <= nlist; i++ ) {
         specfile = fullpath( spec_file_list(i) );
         if( chat ) write,"Update ANCRFILE in "+specfile;
         // define ANCRFILE is same directory as spectral file
         splitfname, specfile, dirname, basename;
         pos = strpos( basename, "_srcl_spe_", 1 );
         if( pos == 0 ) {
            write,format="Sorry, %s does not conform - skip\n", specfile;
            continue;
         }

         arffile = dirname+"/"+strpart( basename, 1:pos+5 )+"arf"+strpart(basename, pos+9:0);
         if( !file_test( arffile ) ) {
            write,format="Sorry, %s has no matching ARF file\n", specfile;
            continue;
         }
         ancrfile = arffile+"{1}";
         arf = rdfitscol( specfile+"+1","ancrfile");
         length = strlen(arf(1));
         len = strlen(ancrfile);
         for(j=1;j<=length-len;j++) ancrfile += " ";
         for( j = 1; j <= numberof(arf); j++ ) arf(j) = ancrfile;

         fits_bintable_poke, specfile+"+1", 0, "ancrfile", arf;

         write,format="Has now updated %s\n", specfile;
      }

   }

}

/* Function j_get_src_spectrum */

struct s_Rdm {
   pointer rdm;
   pointer e_min;
   pointer e_max;
   pointer energ_lo;
   pointer energ_hi;
   pointer arf;
}

func j_get_src_spectrum( dol_of_spec, num_spec, &rate, &stat_err, &rdm, &exposure, sourceid=, silent= )
/* DOCUMENT j_get_src_spectrum, dol_of_spec, num_spec, >rate, >stat_err,
                          >rdm, >exposure, sourceid=, silent=

   Simple function to retrieve a given spectrum in PHA II format.
   The row number can be given as 'num_spec' but a SOURCE_ID name
   given in keyword 'sourceid' will make a search for the specified source.

   Returns 'rate', 'stat_err', the struct 'rdm', and 'exposure' in the arguments.
   The function value returned is 0, 1, or 2 giving the spectrum type 'net', 'bkg', or 'total'
   as specified with the OGIP standard.

   'rdm' has elements (pointers): rdm, e_min, e_max, energ_lo, and energ_hi.

   This function has an alias:  read_phaii_spectrum

   2008-01-01(?)/NJW
   2010-10-28/NJW updated with e_min, e_max return.
   2012-30-29/NJW updated with exposure and type return.
*/
{
   local file_of_spec, ext_num;
   rdm = s_Rdm();

   get_exten_no, dol_of_spec, file_of_spec, ext_num;
   if( ext_num == 0 ) dol_of_spec += "+1"; // assumes first extension

   if( !file_test( file_of_spec ) ) error,"Cannot find: "+file_of_spec;

   if( is_void(num_spec) ) num_spec = 1; // defaults to first spectrum
   if( num_spec <= 0 ) error,"Illegal num_spec value";

   rate = stat_err = [];

   if( !silent ) write,format="Data: #%i in %s\n", num_spec, dol_of_spec;

   // Get the header and various keywords

   hdr = headfits( dol_of_spec );
   naxis2 = fxpar( hdr, "naxis2" );
   respfile = fxpar( hdr, "respfile" );
   type = 0; // the default value implying a net spectrum (i.e. background subtracted)
   ogip = fxpar( hdr, "hduclass" );
   if( typeof(ogip) == "string" ) {
      if( strtrim(strlowcase(ogip)) == "ogip" ) {
         clas1 = fxpar( hdr, "hduclas1" );
         if( is_void(clas1) ) {
            write,"Warning! HDUCLAS1 keyword is missing";
         } else {
            if( strtrim(clas1) != "SPECTRUM" ) write,"Warning! HDUCLAS1 is not SPECTRUM";
         }
         clas2 = fxpar( hdr, "hduclas2" );
         if( is_void(clas2) ) {
            write,"Warning! HDUCLAS2 keyword is missing";
         } else {
            if( strtrim(clas2) == "NET" ) {
               type = 0;
            } else if( strtrim(clas2) == "BKG" ) {
               type = 1;
            } else if( strtrim(clas2) == "TOTAL" ) {
               type = 2;
            } else write,"Warning! HDUCLAS2 is alien: "+clas2;
         }
         clas3 = fxpar( hdr, "hduclas3" );
         if( is_void(clas3) ) {
            write,"Warning! HDUCLAS3 keyword is missing";
         } else {
            if( strtrim(clas3) != "RATE" ) write,"Warning! HDUCLAS3 is not RATE ("+clas3+")";
         }
      }
   }

   if( !is_void(sourceid) ) {
      rowid = strtrim(rdfitscol( dol_of_spec, "rowid" ));
      w = where( sourceid == rowid );
      if( numberof(w) == 0 ) error,"Requested source not found";
      num_spec = w(1);
      if( !silent ) write,"Extracting spec no. "+itoa(num_spec)+": "+sourceid;
   }
   if( !silent ) {
      if( naxis2 == 1 ) { write,itoa(naxis2)+" spectrum found";
      } else write,itoa(naxis2)+" spectra found";
      rowid = rdfitscol( dol_of_spec, "rowid" );
      for( i = 1; i <= naxis2; i++ ) write,format=" %2i: %s\n", i, strtrim(rowid(i));
      if( !is_void(respfile) ) {
         write,format="RESPFILE = %s\n", respfile;
      } else write,"No RESPFILE keyword found";
   }
   e_min = e_max = [];
   if( !is_void(respfile) ) {
      if( file_test(respfile) ) {
         if( !silent ) write,format="Response: %s\n", respfile;
         e_min = rdfitscol(respfile+"[EBOUNDS]","e_min");
         e_max = rdfitscol(respfile+"[EBOUNDS]","e_max");
         rdm.e_min = &e_min;
         rdm.e_max = &e_max;
         energ_lo = rdfitscol(respfile+"[MATRIX]","energ_lo");
         energ_hi = rdfitscol(respfile+"[MATRIX]","energ_hi");
         matrix = rdfitscol(respfile+"[MATRIX]","matrix");
         rdm.energ_lo = &energ_lo;
         rdm.energ_hi = &energ_hi;
         rdm.rdm = &matrix;
      } else write,"j_get_src_spectrum: "+respfile+" was not found";
   } else write,"j_get_src_spectrum: keyword RESPFILE was not found";

   if( naxis2 ) {
   
      // Test the number of the requested spectrum
      if( num_spec > naxis2 ) {
         write,format="Only %i spectra in file\n", naxis2;
         return;
      }

      // Get the spectrum and errors and exposure

      rate = rdfitscol(dol_of_spec,"rate")(,num_spec);
      stat_err = rdfitscol(dol_of_spec,"stat_err")(,num_spec);
      exposure = rdfitscol(dol_of_spec,"exposure")(num_spec);

      // Get the ANCRFILE i.e. the ARF

      ancrfile = strip_curly_br(strtrim(rdfitscol( dol_of_spec, "ancrfile" )(num_spec)));
      if( !file_test(ancrfile) ) {
         // try same directory as spec file
         dddir = dirname(file_of_spec);
         ancrfile = dddir+"/"+basename(ancrfile);
         if( !file_test(ancrfile) ) write,"Search for "+ancrfile+" failed";
      }
      if( file_test(ancrfile) ) {
         arf = rdfitscol(ancrfile+"+1","specresp")(,num_spec);
         if( is_void(arf) ) {
            write,format="No ARF found in %s\n", ancrfile;
         } else {
            if( !silent ) write,format="ARF from: %s\n", ancrfile;
            rdm.arf = &arf;
         }
      }

   } else {

      write,format="No spectra in %s\n", file_of_spec;

   }

}
read_phaii_spectrum = j_get_src_spectrum;

/* Function j_get_point_info_for_swid */

struct s_All {
   double ra_scx;
   double dec_scx;
   double posangle;
   double exposure;
   double gain;
}

/*
 *  This is in a rudimentary form and should be expanded
 */

func j_get_point_info_for_swid( jemxNum, swid )
/* DOCUMENT res = j_get_point_info_for_swid( jemxNum, swid )

   Reads the FITS pointing file in the directory given by the
   environment variable 'J_POINTINGS' and returns the struct s_All
   with the information for the chosen instrument in the
   'pointings_RRRR.fits' file.

   Struct elements: ra_scx, dec_scx, posangle, exposure, gain.

   2008-05-19/NJW (get_all_for_swid)
   2008-08-19/NJW Updated to get_point_info_for_swid
*/
{
   revol = strpart(swid,1:4);
   fil = get_env("J_POINTINGS") + "/pointings_" + revol + ".fits";

   if( !file_test(fil) ) {
      write,format="File not found: %s\n", fil;
      return [];
   }

   dol = fil+"+1";

   local fh, nrows;
   col_ptrs = rdfitsbin( dol, fh, nrows );
   colnum = fits_colnum( fh, "swid" );
   w = where(swid == *col_ptrs(colnum));

   if( numberof(w) < 1 ) {
      write,format="SWID %s not found in pointing file\n", swid;
      return [];
   }

   res = s_All();

   suffix = jemxNum == 1 ? "_j1" : "_j2";

   res.exposure = (*col_ptrs(fits_colnum(fh,"exposure"+suffix)))(w(1));
   res.gain     = (*col_ptrs(fits_colnum(fh,"gain"+suffix)))(w(1));
   res.ra_scx   = (*col_ptrs(fits_colnum(fh,"ra_scx")))(w(1));
   res.dec_scx  = (*col_ptrs(fits_colnum(fh,"dec_scx")))(w(1));
   res.posangle  = (*col_ptrs(fits_colnum(fh,"posangle")))(w(1));

   return res;

}

/* Function j_std_rebin */

func j_std_rebin( dol_spec, &ob1, &ob2, &orate, &orate_err, degree= )
/* DOCUMENT REBIN = j_std_rebin( dol_spec, >ob1, >ob2, >orate, >orate_err, degree= )
        or  REBIN = j_std_rebin()

   The former version will read the spectrum and return the
         rebinned spectrum
   The latter version will simply return the rebinning array

   Keyword  degree: 1 is the default
                    2 is coarser rebinning (for fainter sources)
                    3 is very coarse

   2008-05-29/NJW
   2008-08-19/NJW, updated with option
*/

{

   extern Eb1, Eb2;

   if( is_void(degree) ) degree = 1;

   // Define the standard rebinning array

   if( degree == 1 ) {
      REBIN = -39;
      grow, REBIN, array(1,10);
      grow, REBIN, array(2,10);
      grow, REBIN, array(3,30);
      grow, REBIN, array(2,33);
      grow, REBIN, -31;
   } else if( degree == 2 ) {
      REBIN = -45;
      grow, REBIN, array(2,15);
      grow, REBIN, array(3,10);
      grow, REBIN, array(4,10);
      grow, REBIN, array(5,5);
      grow, REBIN, array(6,5);
      grow, REBIN, -56;
   } else {
      REBIN = -45;
      grow, REBIN, array(3,15);
      grow, REBIN, array(5,10);
      grow, REBIN, array(7,5);
      r = 256 - sum(abs(REBIN));
      grow, REBIN, -r;
   }

   if( is_void(dol_spec) ) return REBIN;

   local rate, rate_err;
   j_get_src_spectrum, dol_spec, 1, rate, rate_err;

   if( is_void(Eb1) ) {
      j_get_pi_ebds, Eb1, Eb2;
   }

   specrebinning, Eb1, Eb2, rate, rate_err, REBIN, ob1, ob2, orate, orate_err;

   return REBIN;

}

/* Function mk_unique_cat */

func mk_unique_cat( dol_in_cat, file_unique_cat, clog=, chat=, accrad=, idrad= )
/* DOCUMENT mk_unique_cat, dol_in_cat, file_unique_cat, clog=, chat=, accrad=, idrad=

  Make a "unique catalog" i.e. where a source only appears once
  from the catalog in dol_in_cat.

  Preliminary identifications are inserted but it is advised to follow up by running
  'qidsrcs' with keyword uniq=1 (implying that a given gnrl_src_cat source can only
  by used once).

  Required columns in the input catalog: NAME, RA_OBJ, DEC_OBJ, DETSIG, ERR_RAD

  Keywords:
     clog     Set catalog DOL if different from the default one:
                                  /r6/jemx/catalogs/gnrl_refr_cat_latest.fits+1
     accrad   Acceptance radius in degrees (default value: 0.07 deg = 4.2 arcmin)
                 e.g. only accepted as same source if the distance is less than accrad
     idrad    Identification radius in degrees (default value: 0.07 deg)
     chat     Chattiness parameter

  2005-06-08/NJW  Cloned from /home/njw/0idl/proc/mk_unique_issw_cat.pro
  2005-10-28/NJW  Updated with column "NAME" in stead of "SRCNAME" and
                  extra column "ERR_RAD"
  2006-01-13/NJW  Only assign the source name after final positions has been found
                  and introduce a proper definition of the error radius (err_rad)
                  based on err_rad from input catalog

  2008-10-29/NJW translated to Yorick
  2008-11-18/NJW changed to handle the source identification better
  2008-11-25/NJW includes now the CLASS (source category) column

*/
{

   version = 5;   // NB: remember to update this value
   verdate = "2008-11-18";

   if( is_void(accrad) ) accrad = 0.07;
   if( is_void(idrad) ) idrad = 0.07;

   if( is_void(chat) ) chat = 1;

   logfile = get_next_filename("mujc_???.txt");
   flog = open(logfile,"w");



   // Read the contents of the input catalog

   write,format="Reading %s\n",dol_in_cat;
   write,flog,format="Input catlog: %s\n",fullpath(dol_in_cat);
   name = rdfitscol(dol_in_cat,"NAME");
   if( is_void(name) ) error,"Column NAME is missing in catalog";
   ra_obj = rdfitscol(dol_in_cat,"RA_OBJ");
   if( is_void(ra_obj) ) error,"Column RA_OBJ is missing in catalog";
   dec_obj = rdfitscol(dol_in_cat,"DEC_OBJ");
   if( is_void(dec_obj) ) error,"Column DEC_OBJ is missing in catalog";
   detsig = rdfitscol(dol_in_cat,"DETSIG");
   if( is_void(detsig) ) error,"Column DETSIG is missing in catalog";
   err_rad = rdfitscol(dol_in_cat,"ERR_RAD");
   if( is_void(err_rad) ) error,"Column ERR_RAD is missing in catalog";

   // Avoid negative DETSIG

   w = where( detsig  <  0.01 );
   nw = numberof(w);
   if( nw  >  0 ) {
      write,format="Warning: %i DETSIG values smaller than 0.01 exist\n", nw;
      write,"  The value of those is reset to 0.01";
      write,flog,format="Warning: %i DETSIG values smaller than 0.01 exist\n", nw;
      write,flog,"  The value of those is reset to 0.01";
      detsig(w) = 0.01;
   }

   num_entries_cat = numberof(ra_obj);
   if( chat  >  2 ) write,format="##0## num_entries_cat = %i\n", num_entries_cat;
   write,flog,format="Number of rows: %i\n", num_entries_cat;
   ismarked = array( int, num_entries_cat);


   // Define DOL and filename for the unique catalog as well as for the auxiliary
   //  Yorick save file

   pos = strpos( file_unique_cat,  ".fits", 0, rev=1 );
   if( pos > 0 ) {
      file_ysave = strpart( file_unique_cat, 1:pos ) + "ysave";
   } else file_ysave = "unique_cat_aux_"+ndate(1)+".ysave";

   dol_unique_cat = file_unique_cat+"+1";

   write,flog,format="DOL of output Unique cat: %s\n", fullpath(dol_unique_cat);
   write,flog,format="Filename of output Unique cat: %s\n", fullpath(file_unique_cat);
   write,flog,format="Yorick save file for list_int: %s\n", fullpath(file_ysave);


   //  Prepare arrays for saving unique data and allocating room for catalog
   //  source positions

   ra_obj_uniq = [];
   dec_obj_uniq = [];
   srcname_uniq = [];
   detsig_uniq = [];
   err_rad_uniq = [];
   pos_list_int = [];
   len_list_int = [];
   ra_cat_uniq = [];
   dec_cat_uniq = [];
   source_id_uniq = [];
   class_uniq = [];

   list_int = array( long, num_entries_cat);
   list_uno = array( long, num_entries_cat);

   //  Walk through the input catalog sources (i-sources)
   //  Pick the first one as the first resulting unique source (u-source)
   //  then assign all i-sources at same position to this one
   //  After that pick the next un-assigned i-source etc.


   cur = 1;
   uniq_no = 0;
   ans = "";
   dum = where( ismarked  ==  0 ); // Get number of unmarked i-sources
   num = numberof( dum );

   while( num  >  0 ) {

      uniq_no++;

      if( chat > 2 ) {
         write,format="Next unique source: %s\n", name(cur);
      }

      // Find the i-sources that can be identified with current u-source

      r = arcdist(ra_obj(cur), dec_obj(cur), ra_obj, dec_obj );
      w = where( r <=  accrad & ismarked == 0 );
      m = where( r > accrad | ismarked != 0 );
      nw = numberof(w);
      nm = numberof(m);
      bestpos = [ra_obj(cur), dec_obj(cur)];
      weight = array(1./nw,nw); // initial definition

      if( chat  >  4 ) {
         write," ----------  Begin with uniq_no = "+itoa(uniq_no)+" ---------";
         write,"     Initial  nw = "+itoa(nw);
         write,"     Working with unmarked finding cur:  "+itoa(cur);
         write,flog," ----------  Begin with uniq_no = "+itoa(uniq_no)+" ---------";
         write,flog,"     Initial  nw = "+itoa(nw);
         write,flog,"     Working with unmarked finding cur:  "+itoa(cur);
      }

      if( nw  ==  0 ) {
         write,format="Error ##1##, nw == 0, cur = %i, uniq_no = %i\n",cur,uniq_no;
         write,format="nw = %i, nm = %i\n", nw, nm;
         write,flog,format="Error ##1##, nw == 0, cur = %i, uniq_no = %i\n",cur,uniq_no;
         write,flog,format="nw = %i, nm = %i\n", nw, nm;
         close, flog;
         return;
      }

      if( nw  >  1 ) { // Try to iterate if it makes sense

         do {  // See if a better position as an average of all initially assigned
               // i-sources will change the selection
               // (the first source may be an outlier in a group)

            w_previous = w; // points to selected i-sources

            weight = detsig(w);
            if( max(weight)  <=  0.0 ) {
               weight = replicate(1., nw);
               write,format="DETSIG problem with input source cur = %i", cur;
               write,format="                and  uniq_no = %i\n", uniq_no;
               write,flog,format="DETSIG problem with input source cur = %i", cur;
               write,flog,format="                and  uniq_no = %i\n", uniq_no;
            }

            // normalize "weight"
            weight = weight / sum(weight);

            bestpos = best_skypos( ra_obj(w), dec_obj(w), weight );

            r = arcdist( bestpos(1), bestpos(2), ra_obj, dec_obj );
            w = where( r <= accrad & ismarked == 0 );
            m = where( r  > accrad | ismarked != 0 );
            nw = numberof(w);
            nm = numberof(m);

            if( chat  >  4 ) write,format="uniq_no = %i 2. nw = %i\n", uniq_no, nw;

            if( nw  ==  0 ) {
               write,"Mark ##2##";
               write,format="cur = %i,  uniq_no = %i\n", cur, uniq_no;
               write,format="nw = %i,  nm = %i\n", nw, nm;
               close, flog;
               return;
            }

            diff = numberof(w) == numberof(w_previous) ? 0 : 1;
            if( !diff ) {
               diff = anyof( w - w_previous );
            }
         } while( diff );

      }

      // Now we have a new unique source

      if( chat  >  4 ) {
         write,flog,format="New unique source, uniq_no: %i\n", uniq_no;
         write,flog,format="Number of contributions (nw) is %i\n", nw;
         write,flog,format="Number of contributions (numberof(w)) is %i\n", numberof(w);
         write,flog,format="bestpos: %9.4f %9.4f\n", bestpos(1), bestpos(2);
      }

      list_uno(w) = uniq_no;

      // Save values to storage arrays

      grow, ra_obj_uniq, bestpos(1);
      grow, dec_obj_uniq, bestpos(2);
      grow, detsig_uniq, max(detsig(w));
      grow, len_list_int, nw;
      len_list_int(uniq_no) = nw;
      if( uniq_no  ==  1 ) {
         pos_list_int = 1;
      } else {
         grow, pos_list_int, pos_list_int(uniq_no-1) + len_list_int(uniq_no-1);
      }
      list_int(pos_list_int(uniq_no):pos_list_int(uniq_no)+nw-1) = w;

      // Get sourcename from standard X-ray source catalog or
      //     from the chosen one

      res = nearest_source( bestpos(1), bestpos(2), idrad, clog=clog );
      if( is_void(res) ) {
         name_tmp = "NEW SOURCE";
         s_id_tmp = "UNKNOWN";
         class_tmp = 0n;
         ra_cat_tmp = -99.0;
         dec_cat_tmp = -99.0;
      } else {
         name_tmp = res.name;
         s_id_tmp = res.source_id;
         class_tmp = int(res.class);
         ra_cat_tmp = res.ra_obj;
         dec_cat_tmp = res.dec_obj;
      }
      grow, srcname_uniq, name_tmp;
      grow, source_id_uniq, s_id_tmp;
      grow, class_uniq, class_tmp;
      if( chat  >  4 ) write,flog,format=" name: %s\n",name_tmp;
      grow, ra_cat_uniq, ra_cat_tmp;
      grow, dec_cat_uniq, dec_cat_tmp;

      //  Define the position error

      var_err_rad = sum( (weight*err_rad(w))^2 );
      syst_err_rad = 2.778e-3; // in degrees, corresponding to 10 arcsec
      grow, err_rad_uniq, sqrt( var_err_rad + syst_err_rad^2 );
      if( chat  >  4 ) {
         nweight = numberof(w);
         write,flog,format=" numberof(weight): %i\n", numberof(w);
         for( k = 1; k <= min([5,nweight]); k++ ) {
            write,flog,"         weight(%i) = %8.5f\n", k, weight(k);
         }
         write,flog,format=" var_err_rad: %8.5f\n", var_err_rad;
         write,flog,format=" err_rad_uniq: %8.5f\n", sqrt( var_err_rad + syst_err_rad^2 );
      }

      //  Mark the i-sources now assigned

      ismarked(w) = 1;

      //  Find next unmarked i-source

      dum = where( ismarked == 0 ); // Get number of unmarked findings
      num = numberof( dum );
      if( num  >  0 ) {
         //+ while ismarked(cur)  ==  1 do cur = cur + 1
         cur = dum(1);
      }

      if( chat  >  4 ) {
         write,format="num = %i,  cur for next loop = %i\n", num, cur;
      }

   }

   if( uniq_no != numberof(ra_obj_uniq) ) error,"Mismatch between uniq_no and num of ra_obj_uniq";

   if( chat > 1 ) write,format="Writing file %s with %i sources\n", \
                     file_unique_cat, uniq_no;
   s20 = "01234567890123456789"
   kwds_init;
   kwds_set,"EXTNAME","JEMX_UNIQUE_CAT","JEM-X catalog of unique sources";
   kwds_set,"DATE",ndate(3),"Date of creation";
   kwds_set,"ORIGIN","mk_unique_cat.i","made this";
   kwds_set,"VERSION",version,"Version of generating program";
   kwds_set,"VERDATE",verdate,"Date of this version";
   kwds_set,"RESPONSI","Niels J. Westergaard, DTU Space","Programmer and responsible";
   kwds_set,"INPUTCAT", fullpath(dol_in_cat),"Input catalog";
   kwds_set,"LIST_INT", fullpath(file_ysave),"Yorick save file of LIST_INT";
   kwds_set,"ACCRAD", accrad,"[deg] Setting of acceptance radius keyword";
   kwds_set,"IDRAD", idrad,"[deg] Setting of identification radius keyword";
   kwds_set,"REFR_CAT", Source_cat_dol, "Reference catalog of X-ray sources";
   kwds_set,"TUNIT4","deg","Unit of RA_OBJ";
   kwds_set,"TUNIT5","deg","Unit of DEC_OBJ";
   kwds_set,"TUNIT6","deg","Unit of RA_CAT";
   kwds_set,"TUNIT7","deg","Unit of DEC_CAT";
   kwds_set,"TUNIT9","deg","Unit of ERR_RAD";

   l = strlen(srcname_uniq(1));
   for( i=1; i<= 20-l; i++ ) srcname_uniq(1) += " ";
   l = strlen(source_id_uniq(1));
   for( i=1; i<= 16-l; i++ ) source_id_uniq(1) += " ";
   wrmfitscols, file_unique_cat,"SOURCE_ID", source_id_uniq, "NAME", srcname_uniq, \
          "CLASS", class_uniq, \
          "RA_OBJ", ra_obj_uniq, "DEC_OBJ", dec_obj_uniq, \
          "RA_CAT", ra_cat_uniq, "DEC_CAT", dec_cat_uniq, \
          "DETSIG", detsig_uniq, "ERR_RAD", err_rad_uniq, \
          "NUMFIND", len_list_int, clobber=1;

  /******  The following step seems unnecessary ***
   *
   * fsav = createb( file_ysave );
   * save, fsav, uniq_no, pos_list_int, len_list_int, list_int;
   * close, fsav;
   *
   ***      ***         ***       ***/

  /*** skip this step temporarily ***
   *  for( i = 1; i <= uniq_no; i++ ) {
   *     p = pos_list_int(i);
   *     l = len_list_int(i);
   *     fits_bintable_poke, file_unique_cat+"+1", i, 7, list_int(p:p+l-1);
   *  }
   ***      ***         ***       ***/


   write,format="\n  Log file : %s\n\n", logfile;

   close, flog;

}

/* Function best_skypos */

func best_skypos( ra, dec, weight )
/* DOCUMENT newpos = best_skypos( ra, dec, weight )

  Returns average sky position array: [ra,dec]
  when arrays of RA and Dec are given in degrees.

  The third argument is a weight array
  If it is not given then same weight is applied to
  all positions.

   2003-10-27/NJW
   2008-10-29/NJW translated to Yorick from IDL
*/
{

   nval = numberof(ra);
   nd = numberof(dec);
   if( nval  !=  nd ) {
      write,"BEST_SKYPOS error, bad dimensions of input (dec)";
      return [];
   }
   if( is_void(weight) ) {
      weight = array( 1., nval);
   } else {
      if( nval  !=  numberof(weight) ) {
         write,"BEST_SKYPOS error, bad dimensions of input(weight)";
         return [];
      }
   }

   dr = pi / 180;

   rsum = array( double,3);
   for( i = 1; i <= nval; i++ ) {
      r = [cos(ra(i)*dr)*cos(dec(i)*dr),sin(ra(i)*dr)*cos(dec(i)*dr), sin(dec(i)*dr)];
      rsum += r*weight(i);
   }

   // normalize

   norm = sqrt(sum(rsum^2));
   rsum = rsum / norm;
   ra_res = zero2pi(atan(rsum(2),rsum(1)))/dr;
   dec_res = asin(rsum(3))/dr;

   return [ra_res,dec_res];

}

/* Function cat_compare */

func cat_compare( dol_cat_1, dol_cat_2, outfile=, silent=, \
          names=, reci=, rad=, diff= )
/* DOCUMENT cat_compare, dol_cat_1, dol_cat_2, outfile=, silent=,
             names=, reci=, rad=, diff=

   Keywords:
      outfile  : will override default name 'cat_compare_0000.txt'
      silent   : to suppress output on terminal
      names    :  1  -> to force only a comparison on source names
                  2  -> to force also a comparison on source names
      reci     : to force output from the reciprocal test i.e.
                    when walking through CAT2 to find coincidences
                    in CAT1
      rad      : minimum error radius [degrees], defaults to 0.01
      diff     : causes only differences to be reported in outfile

   2009-08-26/NJW   from cat_compare.pro
*/
{

   vb = is_void(silent);
   if( is_void(outfile) ) outfile = get_next_filename("cat_compare_????.txt");
   lun = open(outfile, "w");

   if( is_void(rad) ) {min_err_rad = 0.01;} else {min_err_rad = rad;}
   if( is_void(names) ) names = 0;

   cat1 = dol_cat_1;
   cat2 = dol_cat_2;
   write,lun,"  Output file of \"cat_compare\",    "+ndate(3);

   write,lun,"CAT1: "+fullpath(cat1);
   write,"Reading "+cat1+" ...";

   fh1 = headfits( cat1 );
   nrows1 = fxpar( fh1, "NAXIS2" );

   // See if source names are present as 'NAME' or 'SOURCENAME'
   cname = "NAME";
   if( !is_void(fits_colnum( fh1, cname )) ) {
      name1 = rdfitscol(cat1,cname);
   } else {
      cname = "SOURCENAME";
      if( !is_void(fits_colnum( fh1, cname )) ) {
         name1 = rdfitscol(cat1,cname);
      } else { // I must give up - neither of NAME or SOURCENAME are found
               // Invent some names
         write,"No NAME or SOURCENAME column in cat1 - use standard names";
         name1 = swrite(format="CAT1_%06i",indgen(nrows1));
      }
   }
   // See if R.A. column is given as RA or as RA_OBJ
   cname = "RA_OBJ";
   if( !is_void(fits_colnum( fh1, cname )) ) {
      ra_obj1 = rdfitscol(cat1,cname);
   } else {
      cname = "RA";
      if( !is_void(fits_colnum( fh1, cname )) ) {
         ra_obj1 = rdfitscol(cat1,cname);
      } else error,"No R.A. information in cat1 ...";
   }
   // See if declination column is given as DEC or as DEC_OBJ
   cname = "DEC_OBJ";
   if( !is_void(fits_colnum( fh1, cname )) ) {
      dec_obj1 = rdfitscol(cat1,cname);
   } else {
      cname = "DEC";
      if( !is_void(fits_colnum( fh1, cname )) ) {
         dec_obj1 = rdfitscol(cat1,cname);
      } else error,"No declination information in cat1 ...";
   }
   // See if error radius column is present as ERR_RAD
   cname = "ERR_RAD";
   if( !is_void(fits_colnum( fh1, cname )) ) {
      err_rad1 = rdfitscol(cat1,cname);
   } else {
      // Use min_err_rad as a default
      err_rad1 = array(min_err_rad, nrows1);
   }

   write,lun,"           with "+itoa(nrows1)+" sources";
   is = sort(ra_obj1);
   name1 = strtrim(name1(is));
   ra_obj1 = ra_obj1(is);
   dec_obj1 = dec_obj1(is);
   err_rad1 = err_rad1(is);
   w = where( err_rad1 < min_err_rad );
   if(numberof(w)) err_rad1(w) = min_err_rad;

   write,lun, "CAT2: "+fullpath(cat2);
   write,"Reading "+cat2+" ...";

   fh2 = headfits( cat2 );
   nrows2 = fxpar( fh2, "NAXIS2" );

   // See if source names are present as 'NAME' or 'SOURCENAME'
   cname = "NAME";
   if( !is_void(fits_colnum( fh2, cname )) ) {
      name2 = rdfitscol(cat2,cname);
   } else {
      cname = "SOURCENAME";
      if( !is_void(fits_colnum( fh2, cname )) ) {
         name2 = rdfitscol(cat2,cname);
      } else { // I must give up - neither of NAME or SOURCENAME are found
               // Invent some names
         write,"No NAME or SOURCENAME column in cat2 - use standard names";
         name2 = swrite(format="CAT2_%06i",indgen(nrows2));
      }
   }
   // See if R.A. column is given as RA or as RA_OBJ
   cname = "RA_OBJ";
   if( !is_void(fits_colnum( fh2, cname )) ) {
      ra_obj2 = rdfitscol(cat2,cname);
   } else {
      cname = "RA";
      if( !is_void(fits_colnum( fh2, cname )) ) {
         ra_obj2 = rdfitscol(cat2,cname);
      } else error,"No R.A. information in cat2 ...";
   }
   // See if declination column is given as DEC or as DEC_OBJ
   cname = "DEC_OBJ";
   if( !is_void(fits_colnum( fh2, cname )) ) {
      dec_obj2 = rdfitscol(cat2,cname);
   } else {
      cname = "DEC";
      if( !is_void(fits_colnum( fh2, cname )) ) {
         dec_obj2 = rdfitscol(cat2,cname);
      } else error,"No declination information in cat2 ...";
   }
   // See if error radius column is present as ERR_RAD
   cname = "ERR_RAD";
   if( !is_void(fits_colnum( fh2, cname )) ) {
      err_rad2 = rdfitscol(cat2,cname);
   } else {
      // Use min_err_rad as a default
      err_rad2 = array(min_err_rad, nrows2);
   }

   write,lun,"           with "+itoa(nrows2)+" sources";
   is = sort(ra_obj2);
   name2 = strtrim(name2(is));
   ra_obj2 = ra_obj2(is);
   dec_obj2 = dec_obj2(is);
   err_rad2 = err_rad2(is);
   w = where( err_rad2 < min_err_rad );
   if(numberof(w)) err_rad2(w) = min_err_rad;

   // report keyword settings

   write,lun,"Keywords have been set as";
   write,lun,"   names = "+itoa(names);
   if( reci ) { write,lun,"   reci  = 1"; } else { write,lun,"   reci  = <not set>";}
   write,lun,"   rad   = "+swrite(format="%.4f",min_err_rad);
   if( diff ) { write,lun,"   diff  = 1"; } else { write,lun,"   diff  = <not set>";}

   if( names ) {

      write,"Checking names ...";
      if( !diff ) {
         write,lun,"";
         write,lun,"-----------------------------------------------------------";
         write,lun,"         MATCHING SOURCE NAMES";
         write,lun,"-----------------------------------------------------------";
      }

      tname1 = strtrim(name1);
      tname2 = strtrim(name2);
      n_coin = 0;
      coin_names = [];
      miss_names = [];
      coin_num = [];
      for( i = 1; i <= nrows1; i++ ) {

         w = where( tname2  ==  tname1(i) );
         nw = numberof(w);

         if( nw  >=  1 ) {
            if( vb ) write,"Both have: "+name1(i);
            //+ write,lun,"         "+name1(i);
            grow,coin_names, name1(i);
            grow,coin_num, nw;
            n_coin++;
         } else {
            grow, miss_names, name1(i);
         }

         if( nw  >  1 ) {
            if( vb ) write,name1(i)+" (CAT1) is found "+itoa(nw)+" times in CAT2";
            //+ write,lun,"         is found "+itoa(nw)+" times in CAT2";
         }

      }

      if( n_coin  ==  0 ) {
         write,"No coincident names found.";
         write,lun,"No coincident names found.";
      } else {

         if( numberof(coin_names)  !=  n_coin ) {
            write,"##3## Error!";
            close,lun;
            return;
         }
         is = sort(coin_names);
         coin_names = coin_names(is);
         coin_num   = coin_num(is);
         if( !diff ) {
            for( i = 1; i <= n_coin; i++ ) {
               write,lun,format="%20s   %i\n", coin_names(i), coin_num(i);
            }
            write,itoa(n_coin)+" coincidences of names.";
            write,lun,"";
            write,lun,itoa(n_coin)+" coincidences of names.";
         }
      }

      if( (nmiss = numberof(miss_names)) ) {
         write,lun,"";
         write,lun,"     ---------------------------------------";
         write,lun,"         Names in CAT1 not found in CAT2"
         write,lun,"     ---------------------------------------";
         for( i = 1; i <= nmiss; i++ ) {
            write,lun,"     "+miss_names(i);
         }
         write,lun,"     ---------------------------------------";
      } else {
         write,lun,"";
         write,lun,"---   All names in CAT1 also exist in CAT2 -----";
      }

      // repeat - if required - starting from CAT2

      n_coin = 0;
      coin_names = [];
      miss_names = [];
      coin_num = [];
      for( i = 1; i <= nrows2; i++ ) {

         w = where( tname1  ==  tname2(i) );
         nw = numberof(w);

         if( nw  >=  1 ) {
            if( vb && reci ) write,"Both have: "+name2(i);
            //+ write,lun,"         "+name2(i);
            grow,coin_names, name2(i);
            grow,coin_num, nw;
            n_coin++;
         } else {
            grow, miss_names, name2(i);
         }

         if( nw  >  1 ) {
            if( vb && reci ) write,name2(i)+" (CAT2) is found "+itoa(nw)+" times in CAT1";
            //+ write,lun,"         is found "+itoa(nw)+" times in CAT1";
         }

      }

      if( n_coin  ==  0 ) {
         write,"No coincident names found.";
         write,lun,"No coincident names found.";
      } else {

         if( numberof(coin_names)  !=  n_coin ) {
            write,"##4## Error!";
            close,lun;
            return;
         }
         is = sort(coin_names);
         coin_names = coin_names(is);
         coin_num   = coin_num(is);
         if( !diff ) {
            for( i = 1; i <= n_coin; i++ ) {
               write,lun,format="%20s   %i\n", coin_names(i), coin_num(i);
            }
            write,itoa(n_coin)+" coincidences of names.";
            write,lun,"";
            write,lun,itoa(n_coin)+" coincidences of names.";
         }
      }

      if( (nmiss = numberof(miss_names)) ) {
         write,lun,"";
         write,lun,"     ---------------------------------------";
         write,lun,"         Names in CAT2 not found in CAT1"
         write,lun,"     ---------------------------------------";
         for( i = 1; i <= nmiss; i++ ) {
            write,lun,"     "+miss_names(i);
         }
         write,lun,"     ---------------------------------------";
      } else {
         write,lun,"";
         write,lun,"---   All names in CAT2 also exist in CAT1 -----";
      }

   } // Terminate test for keyword "names";

   // only continue with matching positions if 'names' is 0 or 2

   if( names != 1 ) {

      if( !diff ) {
         write,lun,"";
         write,lun,"-----------------------------------------------------------";
         write,lun,"                MATCHING SOURCE POSITIONS";
         write,lun,"-----------------------------------------------------------";
         write,lun,"In CAT1                                  In CAT2";
      }

      write,"Checking source positions ...";

      n_coin = 0;
      aname = [];
      ara = [];
      adec = [];
      for( i = 1; i <= nrows1; i++ ) {

         // find cases where dist < err_rad1 + err_rad2;
         // or - equivalently : dist - err_rad2 < err_rad1;


         dist = arcdist(ra_obj1(i), dec_obj1(i), ra_obj2, dec_obj2 );
         dista = dist - err_rad2;
         w = where( dista  <  err_rad1(i) ) ;
         nw = numberof(w);

         if( nw  >  0 ) {
            n_coin++;
            if( vb ) {
               write,"";
               write,"Catalog 1 source: "+name1(i)+" "+ \
                  swrite(format="%9.3f%9.3f",ra_obj1(i),dec_obj1(i))+" matches the position of";
               for( j = 1; j <= nw; j++ ) {
                  write,"Catalog 2 source: "+name2(w(j))+" "+ \
                     swrite(format="%9.3f%9.3f",ra_obj2(w(j)),dec_obj2(w(j)));
               }
            }
            if( !diff ) {
               write,lun,format="%20s%9.3f%9.3f    %20s%9.3f%9.3f\n", \
                       name1(i), ra_obj1(i), dec_obj1(i), \
                       name2(w(1)), ra_obj2(w(1)), dec_obj2(w(1));
               if( nw  >  1 ) {
                  for( j = 2; j <= nw; j++ ) {
                     write,lun,format="                                          %20s%9.3f%9.3f\n", \
                        name2(w(j)), ra_obj2(w(j)), dec_obj2(w(j));
                  }
               }
            }
         } else {
            grow, aname, name1(i)
            grow, ara, ra_obj1(i)
            grow, adec, dec_obj1(i)
         }

      }

      if( n_coin  ==  0 ) {
         write,"No coincident positions found.";
      } else {
          write,itoa(n_coin)+" coincidences of positions.";
      }

      n_in1_not_in2 = numberof(aname);
      write,lun,"";
      write,lun,"-----------------------------------------------------------";
      write,lun,"                SOURCES IN CAT1 BUT NOT IN CAT2";
      write,lun,"-----------------------------------------------------------";
      write,lun,"                       RA       Dec       Glon    Glat";
      for( i = 1; i <= n_in1_not_in2; i++ ) {
         gcoor = galactic(ara(i),adec(i));
         if( gcoor(1) < 0. ) gcoor(1) += 360.;
         write,lun,format="%20s%9.3f%9.3f%12.3f%9.3f\n", \
                    aname(i),ara(i),adec(i),gcoor(1),gcoor(2);
      }

   /******************************************************************
    *
    *       Walk through catalog 2 to find matches in catalog 1
    *
    *********************************************************************/

      if( reci && !diff ) {
         write,lun,"-----------------------------------------------------------";
         write,lun,"                MATCHING SOURCE POSITIONS (2)";
         write,lun,"-----------------------------------------------------------";
         write,lun,"In CAT2                                  In CAT1";
      }

      aname = [];
      ara = [];
      adec = [];
      n_coin = 0;
      for( i = 1; i <= nrows2; i++ ) {

         // find cases where dist < err_rad1 + err_rad2;
         // or - equivalently : dist - err_rad1 < err_rad2;

         dist = arcdist(ra_obj2(i), dec_obj2(i), ra_obj1, dec_obj1 );
         dista = dist - err_rad1;
         w = where( dista  <  err_rad2(i));
         nw = numberof(w);

         if( nw  >  0 ) {
            n_coin++;

            if( reci ) {

               if( vb ) {
                  write,"";
                  write,"Catalog 2 source: "+name2(i)+" "+ \
                     swrite(format="%9.3f%9.3f",ra_obj2(i),dec_obj2(i))+" matches the position of";
                  for( j = 1; j <= nw; j++ ) {
                     write,"Catalog 1 source: "+name1(w(j))+" "+ \
                        swrite(format="%9.3f%9.3f",ra_obj1(w(j)),dec_obj1(w(j)));
                  }
               }
               if( !diff ) {
                  write,lun,format="%20s%9.3f%9.3f   %20s%9.3f%9.3f\n",name2(i), \
                          ra_obj2(i), dec_obj2(i), \
                          name1(w(1)), ra_obj1(w(1)), dec_obj1(w(1));
                  if( nw  >  1 ) {
                     for( j = 2; j <= nw; j++ ) {
                        write,lun, format="                                         %20s%9.3f%9.3f\n", \
                              name1(w(j)), ra_obj1(w(j)), dec_obj1(w(j));
                     }
                  }
               }

            }

         } else {
            grow, aname, name2(i);
            grow, ara, ra_obj2(i);
            grow, adec, dec_obj2(i);
         }


      }

      if( n_coin  ==  0 ) {
         write,"No coincident positions found.";
      } else {
          write,itoa(n_coin)+" coincidences of positions.";
      }

      n_in2_not_in1 = numberof(aname);
      write,lun,"";
      write,lun,"-----------------------------------------------------------";
      write,lun,"                SOURCES IN CAT2 BUT NOT IN CAT1";
      write,lun,"-----------------------------------------------------------";
      write,lun,"                       RA       Dec       Glon    Glat";
      for( i = 1; i <= n_in2_not_in1; i++ ) {
         gcoor = galactic(ara(i),adec(i));
         if( gcoor(1) < 0. ) gcoor(1) += 360.;
         write,lun,format="%20s%9.3f%9.3f%12.3f%9.3f\n", \
                    aname(i),ara(i),adec(i),gcoor(1),gcoor(2);
      }

   } // terminate test for names != 1

   close, lun;

   write,"The output has been written to: "+outfile;

}

/* Function jmx1_offaxis_azim */

func jmx1_offaxis_azim( ra_src, dec_src, ra_scx,  \
              dec_scx, posang, &offaxis, &azim )
/* DOCUMENT jmx1_offaxis_azim,  ra_src, dec_src, ra_scx, \
               dec_scx, posang, >offaxis, >azim


   Derive a source position in JEM-X 1 coordinates

     2009-02-09/NJW

   Let s_j1 be the vector pointing to the source in
   JMX1 coordinates (s_j1 = M Omega s_eq)
   then
            offaxis = acos(s_j1(1))
            azimuth = atan( s_j1(3), s_j1(2) )

    All angles in degrees
*/
{

   dr = pi/180.;
   alpha = -0.001045;
   beta  = -0.002205;
   gamma = -0.007520;

   s_eq = vector( ra_src, dec_src, deg=1 );
   omega = jconv_coord( ra_scx, dec_scx, posang, to_sc=1 );
   mis = misalign(alpha,beta,gamma);

   s_sc = omega(,+) * s_eq(+);
   s_j1 = mis(,+) * s_sc(+);

   offaxis = acos( s_j1(1) ) / dr;
   azim = zero2pi(atan( s_j1(3), s_j1(2) ) ) / dr;

}

/* Function misalign */

func misalign( alpha, beta, gamma )
/* DOCUMENT mis = misalign( alpha, beta, gamma )

   returns the JEM-X misalignment matrix

   Best values as of 2009-08-19:
        alpha = -0.001045
        beta  = -0.002205
        gamma = -0.007520
*/
{
   if( is_void(alpha) ) {
      alpha = -0.001045;
      beta  = -0.002205;
      gamma = -0.007520;
   }

   mis = array(double,3,3);
   if( alpha == 0.0 && beta == 0.0 && gamma == 0.0 ) {
      mis(1,1) = mis(2,2) = mis(3,3) = 1.0;
      return mis;
   }

   A = sqrt(1.-alpha^2-beta^2);
   C = sqrt(1.-beta^2-gamma^2);

   if( alpha == 0.0 && gamma == 0.0 ) {

      B = 1.0;
      x = 0.0;
      z = 0.0;
      y = -beta;

   } else {

      B_1 = (A*C - alpha*beta*gamma)/(gamma^2+C^2);
      B_2 = (-A*C - alpha*beta*gamma)/(gamma^2+C^2);

   // Equation for B

   /***    unnecessary
    *  AA = gamma^2 + C^2;
    *  BB = 2*alpha*beta*gamma;
    *  CC = alpha^2*C^2 + alpha^2*beta^2 - C^2;
    *
    *  nroots = quadr_eq( AA, BB, CC, B_x1, B_x2 );
    *  B = B_x1;
    ****/

      B = B_1; // the two systems are almost aligned

      x = -(alpha*beta + B*gamma)/C;
      z = -(gamma*beta + B*alpha)/A;
      if( abs(z) > abs(x) ) {
         y = -(B*x + C*gamma)/z;
      } else {
         y = -(B*z + A*alpha)/x;
      }

   }

   mis(1,1) = A;
   mis(1,2) = alpha;
   mis(1,3) = beta;
   mis(2,1) = z;
   mis(2,2) = B;
   mis(2,3) = gamma;
   mis(3,1) = y;
   mis(3,2) = x;
   mis(3,3) = C;

   return mis;

}

/* Function jconv_axes2posang */

func jconv_axes2posang( ra_scx, dec_scx, &posang, ra_scy, \
                  dec_scy, ra_scz, dec_scz )
/* DOCUMENT jconv_axes2posang, ra_scx, dec_scx, >posang, ra_scy,
                         dec_scy, ra_scz, dec_scz

 Procedure to get the posang angle from X, Y, and Z axis directions

 At input: all angles are given in degrees

   2009-02-09/NJW
*/
{

   dr = pi / 180; // degrees to radians

   ra_scx_rad = ra_scx * dr;
   dec_scx_rad = dec_scx * dr;
   ra_scy_rad = ra_scy * dr;
   dec_scy_rad = dec_scy * dr;
   ra_scz_rad = ra_scz * dr;
   dec_scz_rad = dec_scz * dr;

   posang_rad = atan( sin(dec_scy_rad), sin(dec_scz_rad) );
   posang = posang_rad / dr;

}

/* Function jconv_posang2axes */

func jconv_posang2axes( ra_scx, dec_scx, posang, &ra_scy, \
                  &dec_scy, &ra_scz, &dec_scz )
/* DOCUMENT jconv_posang2axes, ra_scx, dec_scx, posang, >ra_scy,
                         >dec_scy, >ra_scz, >dec_scz

 Procedure to convert posang to Y and Z axis directions

 At input: all angles are given in degrees

   2009-02-09/NJW
*/
{

   dr = pi / 180; // degrees to radians

   ra_scx_rad = ra_scx * dr;
   dec_scx_rad = dec_scx * dr;
   posang_rad = posang * dr;

   // Pointing direction: x-axis

   // RA && Dec of two other axes: y and z

   ra_scy_rad = zero2pi(ra_scx_rad - atan(-cos(posang_rad), -sin(dec_scx_rad)*sin(posang_rad)));

   dec_scy_rad = asin(cos(dec_scx_rad)*sin(posang_rad));

   ra_scz_rad = zero2pi(ra_scx_rad - atan(sin(posang_rad), -sin(dec_scx_rad)*cos(posang_rad)));
   dec_scz_rad = -asin(-cos(dec_scx_rad)*cos(posang_rad));

   ra_scy = ra_scy_rad / dr;
   dec_scy = dec_scy_rad / dr;
   ra_scz = ra_scz_rad / dr;
   dec_scz = dec_scz_rad / dr;

}

/* Function get_time_for_swid */

func get_time_for_swid( swid, &tstart, &tstop, chat= )
/* DOCUMENT status = get_time_for_swid( swid, >tstart, >tstop, chat= )

   returns the start and stop times (IJD, unit: day) obtained from
   /r6/jemx/integral_history/timetables/swid_timetable_RRRR.fits

   2009-03-12/NJW
*/
{
   local tswids, ttstart, ttstop;

   if( is_void(chat) ) chat = 0;

   basdir = "/r6/jemx/integral_history/timetables/";
   nswids = numberof(swid); // may be an array
   tstart = array(double, nswids);
   tstop = array(double, nswids);
   is = sort(swid);
   si = sort(is);
   swid = swid(is); // sort to minimize file reading

   revol = strpart(swid, 1:4);
   urevol = uniq( revol );
   nurevol = numberof(urevol);
   for( k = 1; k <= nurevol; k++ ) {
      rev = revol(urevol(k));
      if(chat > 1 ) write,format="Going to read timetable for revol %s\n", rev;
      filenam = basdir+"swid_timetable_"+rev+".fits";
      // check existence of arrays in memory storage
      if( 0 == mem_restore(filenam+"swid",tswids) ) {
         status = mem_restore(filenam+"tstart",ttstart);
         status = mem_restore(filenam+"tstop",ttstop);
      } else {
         tswids = rdfitscol(basdir+"swid_timetable_"+rev+".fits+1","swid");
         ttstart = rdfitscol(basdir+"swid_timetable_"+rev+".fits+1","tstart");
         ttstop = rdfitscol(basdir+"swid_timetable_"+rev+".fits+1","tstop");
         status = mem_save(filenam+"swid",tswids);
         status = mem_save(filenam+"tstart",ttstart);
         status = mem_save(filenam+"tstop",ttstop);
      }
      i2 = k == nurevol ? numberof(swid) : urevol(k+1)-1;
      for( i = urevol(k); i <= i2; i++ ) {
         w = where( swid(i) == tswids );
         nw = numberof(w);
         if( nw == 1 ) {
            tstart(i) = ttstart(w(1));
            tstop(i) = ttstop(w(1));
         } else if( nw == 0 ) {
            write,format="Warning! SWID %s not found in time table\n", swid(i);
         } else error,"##333## in get_time_for_swid (SWID found several times)";
      }
   }

   tstart = tstart(si); // reverse the previous sorting
   tstop = tstop(si); // so that times are matched with SWIDs

   return 0;

}

/* Function srcl2txt */

func srcl2txt( dol_srcl_res, outfile=, mospec=, size=, cat=, detsig= )
/* DOCUMENT srcl2txt, dol_srcl_res, outfile=, mospec=, size=, cat=, detsig=

   Reads a SRCL-RES file

   Keywords:
      outfile   Name of output (text) file
      mospec    If set, a mosaic_spec script is created
      size      is the 'size' parameter for mosaic_spec
      cat       If set, the catalog position is given
      detsig    If set, the detection significance is given

    2009-05-20/NJW
*/
{
   local filename, extno;

   get_exten_no, dol_srcl_res, filename, extno;
   if( extno == 0 ) dol_srcl_res += "+1";

   if( !file_test(filename) ) {
      write, filename+" was not found";
      return;
   }

   if( is_void(size) ) size = 10;

   fh = headfits( dol_srcl_res );
   naxis2 = fxpar(fh, "naxis2");
   instrume = fxpar(fh, "instrume" );
   jstr = strtrim(instrume)=="JMX1" ? "jmx1" : "jmx2";
   Jstr = strupcase(jstr);

   if( naxis2 == 0 ) {
      write,"No rows in "+dol_srcl_res;
      return;
   }
   ra = rdfitscol(dol_srcl_res,"ra_obj");
   if( is_void(ra) ) {
      write,"Problem reading RA_OBJ from "+dol_srcl_res;
      return;
   }
   nrows = numberof(ra);
   dec = rdfitscol(dol_srcl_res,"dec_obj");
   if( is_void(dec) ) {
      write,"Problem reading DEC_OBJ from "+dol_srcl_res;
      return;
   }
   name = rdfitscol(dol_srcl_res,"name");
   if( is_void(dec) ) {
      write,"Problem reading NAME from "+dol_srcl_res;
      return;
   }

   if( structof(outfile) == string ) {
      stream = open(outfile,"w");
      p = 1;
   } else p = 0;

   // fix the output formats and the headline

   name = strtrim(name);
   mlen = max(strlen(name));
   fmt = swrite(format="%%%is %%8.3f%%8.3f",mlen+1);
   headline = strpadd(" NAME",mlen+2," ");
   headline += "  RA_OBJ DEC_OBJ";
   if( cat ) {
      ra_cat = rdfitscol(dol_srcl_res,"ra_cat");
      dec_cat = rdfitscol(dol_srcl_res,"dec_cat");
      fmtc = " %8.3f%8.3f";
      headline += "   RA_CAT DEC_CAT";
   }
   if( detsig ) {
      dtsg = rdfitscol(dol_srcl_res,"detsig");
      fmtd = " %6.1f";
      headline += "  DETSIG";
   }

   write,format="%s\n\n", headline;
   if(p)write,stream,format="%s\n\n", headline;

   for( i = 1; i <= nrows; i++ ) {
      write,format=fmt, name(i), ra(i), dec(i);
      if(p)write,stream,format=fmt, name(i), ra(i), dec(i);
      if( cat ) {
         write,format=fmtc,ra_cat(i),dec_cat(i);
         if(p)write,stream,format=fmtc,ra_cat(i),dec_cat(i);
      }
      if( detsig ) {
         write,format=fmtd,dtsg(i);
         if(p)write,stream,format=fmtd,dtsg(i);
      }
      write,format="%s\n","";
      if(p)write,stream,format="%s\n","";
      if( mospec ) {
         fname = get_next_filename("mospec_???");
         f = open(fname,"w");
         write,f,format="mosaic_spec %s\n","\\";
         write,f,format="   DOL_idx=\"%s_sky_ima.fits[GROUPING]\" %s\n",jstr,"\\";
         write,f,format="   DOL_spec=\"%s_mosaic_spec.fits(%s-PHA1-SPE.tpl)\" %s\n",jstr,Jstr,"\\";
         write,f,format="   ra=%.3f \\\n", ra(i);
         write,f,format="   dec=%.3f \\\n", dec(i);
         write,f,format="   posmode=-1 %s\n","\\";
         write,f,format="   widthmode=-2 %s\n","\\";
         write,f,format="   size=%i %s\n",size,"\\";
         write,f,format="   allEnergies=yes %s\n","\\";
         write,f,format="   EXTNAME=%s-SKY.-IMA %s\n",Jstr,"\\";
         write,f,format="   Intensity=RECONSTRUCTED %s\n","\\";
         write,f,format="   chatty=3%s\n","";
         close, f;
      }
   }
   if(p)close, stream;
}

/* Function j_get_srcl_res_spectrum */

func j_get_srcl_res_spectrum( dol_srcl, srcid, &eb1, &eb2, &rate, &rate_err, &exposure )
/* DOCUMENT j_get_srcl_res_spectrum, dol_srcl, srcid, >eb1, >eb2, >rate, >rate_err, >exposure
      status = j_get_srcl_res_spectrum( dol_srcl, srcid, >eb1, >eb2, >rate, >rate_err, >exposure)

     Retrieve the E_MIN, E_MAX, FLUX, and FLUX_ERR values for a specified
     row or source name from a SRCL-RES file.
     'srcid' can be a string (source name) or a number (row number)

     Checks for first three energy intervals: 3.04 7.04 11.04 19.98 keV
     In that case they are disregarded (detection energy bands).

     Returns 0 for succesful running
             -1 if no rows
             -2 if file does not exist
             -3 source was not found

     2009-06-13/NJW
*/
{
   local filename, extno;

   get_exten_no, dol_srcl, filename, extno;
   if( extno == 0 ) {
      extno = 1;
      dol_srcl = filename+"+1";
   }

   if( !file_test(filename) ) {
      write,"Error ",filename," was not found";
      return -2;
   }
   fh = headfits(dol_srcl);
   nrows = fxpar( fh, "naxis2");
   if( nrows==0 ) {
      write,"Warning - no rows in ", filename;
      return -1;
   }
   exposure = fxpar( fh, "exposure" );

   e_min = rdfitscol( dol_srcl, "e_min" )(,1);
   e_max = rdfitscol( dol_srcl, "e_max" )(,1);
   rate = rdfitscol( dol_srcl, "flux" );
   rate_err = rdfitscol( dol_srcl, "flux_err" );
   name = strlowcase(strtrim(rdfitscol( dol_srcl, "name" )));

   if( typeof(srcid) == "string" ) { // check for match
      w = where(strlowcase(strtrim(srcid)) == name);
      if( numberof(w)==0 ) {
         write,"Source ",srcid," was not found";
         return -3;
      }
      srcid = w(1);
   }
   if( srcid > nrows ) {
      write,"Row number exceeded number of rows";
      return -3;
   }

   w = where( e_max(1:150) > 0. );
   eb1 = e_min(w);
   eb2 = e_max(w);
   rate = rate(w,srcid);
   rate_err = rate_err(w,srcid);
   nener = numberof(w);
   if( nener > 3 ) { // See if first bands are detection bands
      if( near(e_min(1),3.04,0.01) && near(e_min(2),7.04,0.01) \
       && near(e_min(3),11.04,0.01) && near(e_max(3),19.98,0.01) ) {
         eb1 = eb1(4:0);
         eb2 = eb2(4:0);
         rate = rate(4:0);
         rate_err = rate_err(4:0);
      }
   }

   return 0;

}

/* Function j_get_srcl_res_spectra */

func j_get_srcl_res_spectra( file_list, srcid, &eb1, &eb2, &rate, &rate_err )
/* DOCUMENT j_get_srcl_res_spectra, file_list, srcid, >eb1, >eb2, >rate, >rate_err

   Returns energy boundaries eb1 and eb2 under the assumption that they are
   identical for all srcl_res file in the list.

   Also returns 'rate' as a n x m array where n is the number of energy bins
   and m is the number of spectra. rate_err is defined similarly.

   2009-07-03/NJW
*/
{
   local loc_rate, loc_rate_err;

   n_file_list = numberof(file_list);

   rate = [];
   rate_err = [];
   for( i = 1; i <= n_file_list; i++ ) {
      write,i,file_list(i);
      status = j_get_srcl_res_spectrum( file_list(i)+"+1", srcid, eb1, eb2, loc_rate, loc_rate_err );
      if( status == 0 ) {
         nbins = numberof(loc_rate);
         grow,rate, reform(loc_rate,nbins,1);
         grow,rate_err, reform(loc_rate_err,nbins,1);
      }
   }
}

/* Function j_add_spectra */

func j_add_spectra( rate, rate_err, &avg_rate, &avg_rate_err )
/* DOCUMENT j_add_spectra, rate, rate_err, >avg_rate, >avg_rate_err

   Find the average spectrum (with errors) from several spectra.

   Add rate(nbins, n_spectra) weighted with 1/rate_err^2
   and divide with sum of weights.

   2009-07-03/NJW
*/
{
   dms = dimsof(rate);
   if( anyof(dms-dimsof(rate_err)) ) error,"Mismatching array dimensions";

   nbins = dms(2);
   nspectra = dms(3);

   w = where( rate_err == 0.0 );
   if( numberof(w) ) rate_err(w) = 1.;

   weight = 1./rate_err^2;
   rw = rate * weight;
   drw = rate_err * weight;

   avg_rate = rw(,sum)/weight(,sum);
   avg_rate_err = sqrt((drw^2)(,sum)) / weight(,sum);

}

/* Function timeconv */

func timeconv( t_in, in=, out= )
/* DOCUMENT t_out = timeconv( t_in, in=, out= )

   Time conversion between
         IJD            INTEGRAL Julian Day
         UTC            string like: 2013-10-25T14:54:17.342
         TT             string like: 2013-10-25T14:55:24.836
         JDTT           Julian Day (TT)
         MJDTT          Modified Julian Day (TT)
         JDUTC          Julian Day (UTC)
         MJDUTC         Modified Julian Day (UTC)
         revolution     INTEGRAL revolution
         MST            Microsoft time i.e. number of seconds since Jan 1, 1970
         NST            NuSTAR time, number of seconds since Jan 1, 00:00:00, 2010

   Keywords:
        in     code for input time value (see list below),
                    defaults to "UTC" if 't_in' is a string and to "IJD" if 't_in' is a number.
       out     code for output time value (see list below)
                    defaults to "IJD" if 't_in' is a string and to "UTC" if 't_in' is a number.

    Accepted codes: "TT", "UTC", "IJD", "REVOL", "MJDTT", "MJDUTC",
                                                  "JDTT",  "JDUTC", "MST", or "NST"
                 (case insensitive)

   "TT" and "UTC" must both be in the format
        YYYY-MM-DDTHH:MM:SS.
            ^  ^  ^  ^  ^    required symbols.

   Does not distinguish between lower and upper case.

   Example:  > timeconv("2009-10-21T09:35:44",in="utc",out="ijd")

   SEE ALSO: dattim2ijd, ijd2dattim, ijd2utc, ijd2tt, utc2ijd, tt2ijd

   2009-10-20/NJW

*/
{

   if( is_void(in) ) {
      in =  structof( t_in ) == string ? "UTC" : "IJD";
   }
   in = strupcase(in);
   if( is_void(out) ) {
      out =  structof( t_in ) == string ? "IJD" : "UTC";
   }
   out = strupcase(out);
   if( structof( t_in ) == string && strupcase(t_in) == "NOW" ) t_in = ndate(3);

   // Read revolution information if at all requested

   if( in == "REVOL" || out == "REVOL" ) {
      dol = "/r6/jemx/integral_history/revolution_latest.fits+1";
      revol_no = double(rdfitscol(dol, "REVOLUTION"));
      time_perigee = rdfitscol(dol, "TIME_PERIGEE");
   }

   t_ref_ijd = 946677533; //
   //+ t_ref_nst = -315618599-667;
   t_ref_nst = -315619266;

   //   Convert t_in to IJD

   if( in == "UTC" ) {
      ijd = utc2ijd( t_in );
   } else if( in == "TT" ) {
      ijd = tt2ijd( t_in );
   } else if( in == "IJD" ) {
      ijd = t_in;
   } else if( in == "REVOL" ) {
      ijd = interp( time_perigee, revol_no, t_in);
   } else if( in == "MJDTT" ) {
      ijd = t_in - 51544.0; // This difference applies when MJD is in TT
   } else if( in == "MJDUTC" ) {
      ijd_x = t_in - 51544.0; // approximation
      // There may be times around a leap second
      // when this procedure fails with 1 s
      corr = ttutcdif( ijd_x );
      ijd = ijd_x + corr / 86400.0;
      if( corr != ttutcdif( ijd ) ) write,"Warning: There may be a leap second problem";
   } else if( in == "JDTT" ) {
      ijd = t_in - 2451544.5; // This difference applies when JD is in TT
   } else if( in == "JDUTC" ) {
      ijd_x = t_in - 2451544.5; // This difference applies when JD is in TT
      corr = ttutcdif( ijd_x );
      ijd = ijd_x + corr / 86400.0;
      if( corr != ttutcdif( ijd ) ) write,"Warning: There may be a leap second problem";
   } else if( in == "MST" ) {
      ijd = (t_in - t_ref_ijd)/86400.0;
   } else if( in == "NST" ) {
      ijd = (t_in - t_ref_nst )/86400.0;
   } else {
      error,"Illegal value of keyword 'in'";
   }

   // Convert ijd to desired output type

   if( out == "UTC" ) {
      return ijd2utc(ijd);
   } else if( out == "TT" ) {
      return ijd2tt(ijd);
   } else if( out == "IJD" ) {
      return ijd;
   } else if( out == "REVOL" ) {
      return interp( revol_no, time_perigee, ijd);
   } else if( out == "MJDTT" ) {
      return ijd + 51544.0;
   } else if( out == "MJDUTC" ) {
      return ijd + 51544.0 - ttutcdif(ijd)/86400.0;
   } else if( out == "JDTT" ) {
      return ijd + 2451544.5;
   } else if( out == "JDUTC" ) {
      return ijd + 2451544.5 - ttutcdif(ijd)/86400.0;
   } else if( out == "MST" ) {
      return long(ijd * 86400.0+0.5) + t_ref_ijd;
   } else if( out == "NST" ) {
      return long(ijd * 86400.0+0.5) + t_ref_nst;
   } else {
      error,"Illegal value of keyword 'out'";
   }

}

/* Function j_get_arc_events */

func j_get_arc_events( jemxNum, swid )
/* DOCUMENT res = j_get_arc_events( jemxNum, swid )

   Returns array of 'struct Event' (defined in jemx.i)
   with all events from science window 'swid'.

    2010-01-21/NJW cloned from j_get_events in 'jemx.i'
*/

{

   local Events;
   local dummy, extno;

   jmxstr = swrite(format="JMX%1i",jemxNum);
   jstr = swrite(format="%1i",jemxNum);
   revol = strpart(swid,1:4);

// Set externals (_PI_Energy_bds_lo/hi) with PI energy boundaries

   if( is_void(_PI_Energy_bds_lo) ) j_get_pi_ebds;

//  prepare filenames for useful files

   savname = "/net/uhuru/pool/pool28/njw/"+revol+"/j"+jstr+"_"+swid+".ysav";
   corname = "/jemx/arc/rev_2/scw/"+revol+"/"+swid+".001/jmx"+jstr+"_full_cor.fits";

//  Check for previously saved event file for faster reading

   if( file_test(savname) ) { // The easy way

      stream = openb(savname);
      restore, stream, Events;
      close, stream;

   } else { // start the entire machinery

      cor = 0;
      if( file_test(corname) || file_test(corname+".gz") ) {

// Check for jmxi_full_cor.fits file in JEM-X archive

         full_cor = gz_proxy_file( corname, dir="/r12/scratch/njw/proxies" );
         cor = 1;

      } else {

         write,"full_cor file is not available, try using jmxi_events.fits";

      }

   //     The event file is to be found in the data archive:
      evts_file = "/jemx/arc/rev_2/scw/"+revol+"/"+swid+".001/jmx"+jstr+"_events.fits";
      evts_file_p = gz_proxy_file( evts_file, dir="/r12/scratch/njw/proxies" );
      if( is_void(evts_file_p) ) {
         write,"Failed to find: "+evts_file;
         return [];
      }

      rhdr = headfits( evts_file_p+"+1" );
      rswid = itoa(fxpar(rhdr, "SWID"));
      rinstr = itoa(fxpar(rhdr, "INSTRUME"));

      if( cor ) {
         fhdr = headfits( full_cor+"+1" );
         fswid = itoa(fxpar(fhdr, "SWID"));
         finstr = itoa(fxpar(fhdr, "INSTRUME"));
         if( fswid  !=  rswid ) {
            write,"Mismatching SWIDs: ", fswid+"  "+rswid;
            return [];
         }
         if( finstr  !=  rinstr ) {
            write,"Mismatching INSTRUMEs: ", finstr+"  "+rinstr;
            return [];
         }
      }

      //  Get the events
      //  Since jmxi_full_cor.fits.gz with columns DETX, DETY, PI, ENERGY, STATUS
      //  may have better correction than jmxi_events.fits[JMXi-FULL-ALL]
      //  these values are loaded afterwards

      extnam = finstr+"-FULL-ALL";
      get_exten_no, evts_file_p+"["+extnam+"]", dummy, extno;
      if( extno == -1 ) {
         write,"No FULL-ALL extension found in jxmi_events.fits, quit!";
         return [];
      }
      dol = evts_file_p+"+"+itoa(extno);
      hdr = headfits( dol );
      ptr = rdfitsbin( dol );
      rawx   = *ptr(fits_colnum( hdr, "RAWX"));
      rawy   = *ptr(fits_colnum( hdr, "RAWY"));
      detx   = *ptr(fits_colnum( hdr, "DETX"));
      dety   = *ptr(fits_colnum( hdr, "DETY"));
      pha    = *ptr(fits_colnum( hdr, "PHA"));
      evtime = *ptr(fits_colnum( hdr, "TIME"));
      piarr  = *ptr(fits_colnum(hdr,"PI"));
      status = *ptr(fits_colnum(hdr,"STATUS"));
      cn = fits_colnum(hdr,"ENERGY");
      if( cn ) energy = *ptr(cn);
      nrows = numberof(rawx);

      if( cor ) { // if jmxi_full_cor.fits exists then override some values
         dol = full_cor+"+1";
         hdr = headfits( dol );
         ptr = rdfitsbin( dol );
         detx   = *ptr(fits_colnum( hdr, "DETX"));
         dety   = *ptr(fits_colnum( hdr, "DETY"));
         piarr  = *ptr(fits_colnum(hdr,"PI"));
         cn = fits_colnum(hdr,"ENERGY");
         if( cn ) energy = *ptr(cn);
         status = *ptr(fits_colnum(hdr,"STATUS"));
      }
      if( nrows == 0 ) {
         write,"No events in  file";
         return [];
      }

      Events = array(Event, nrows);
      Events.rawx = toint(rawx);
      Events.rawy = toint(rawy);
      Events.pha  = toint(pha);
      Events.evtime  = evtime;
      if( cn ) {
         Events.energy  = energy;
      } else {
          Events.energy = _PI_Energy_bds_lo(piarr+1) + \
            random(nrows)*(_PI_Energy_bds_hi(piarr+1) - _PI_Energy_bds_lo(piarr+1));
      }

      Events.detx = detx;
      Events.dety = dety;
      Events.piarr = toint(piarr);
      Events.status = toint(status);

//  Save this event array for quicker read next time it is requested

      savname = "/net/uhuru/pool/pool28/njw/"+revol+"/j"+jstr+"_"+swid+".ysav";
      stream = createb(savname);
      save, stream, Events;
      close, stream;

   }

   return Events;

}

/* Function j_comb_dsp_spectra */

func j_comb_dsp_spectra( filelist, outspe=, noplot=, silent=, nolog= )
/* DOCUMENT j_comb_dsp_spectra, filelist, outspe=, noplot=, silent=, nolog=

   Get several detector spectra combined from data e.g. produced with j_bin_spec

   Assumes PHAII representation of spectra i.e. in a vector columns: RATE and
   STAT_ERR

    2008-01-30/NJW Cloned from collect_spectra.i
    2008-05-19/NJW Cloned from j_comb_spectra.i

*/
{
   local strnum;

   if( ! numberof(filelist) ) {
      write,"There is no indication of where to find the spectra ...";
      return [];
   }
   v = !silent;

   num_of_spec = 1; // Standard for detector spectra
   pflag = is_void(noplot);

   nlist = numberof(filelist);
   filelist = filelist(sort(filelist));
   filelist_used = filelist;
   nlist_used = 0;


   // Prepare the output files jmxi_nnnn_spe.fits
   // in the current directory but only if keyword outspe is set

   if( !numberof(outspe) ) {

      fname_spe = get_next_filename("jemx_????_spe.fits", strnum);
      logfile = "j_comb_dsp_spectra"+strnum+".log";

   } else {

      fname_spe = outspe;
      logfile = get_next_filename("j_comb_dsp_spectra????.log");

   }

   cp,"/home/njw/jemx/spectra/jmx1_full_dsp_sample.fits",fname_spe;

   tot_spec = array(float,256);
   tot_err2 = array(float,256);
   tot_expo = 0.0;

   if( !nolog ) {
      write,"Output logging data in "+logfile;
      lun = open(logfile,"w");

      write,lun,"  Logging of j_comb_dsp_spectra   "+ndate(3);
      write,lun;
      write,lun,"Resulting SPE file: "+fname_spe;
   }

   for( i = 1; i <= nlist; i++ ) {

      if(v)write,"Analyzing "+itoa(i)+" of "+itoa(nlist);
      if( !nolog ) {
         write,lun,"-------------------------------";
         write,lun,filelist(i);
      }
      // Get header
      hdr = headfits(filelist(i)+"+1");
      naxis2 = fxpar(hdr,"NAXIS2");
      if( naxis2  <=  0 ) {
         if(v)write,"Skip - no rows";
         if(!nolog)write,lun,"Skip this one, there are no rows";
         continue;
      }
      if( naxis2 < num_of_spec ) {
         if(v)write,"Too few rows in file";
         if(!nolog)write,lun,"Too few rows in file";
         continue;
      }

      pick_swid_str, filelist(i), swid;
      swid = numberof(swid) ? swid(1) : "Unknown SWID";

      // read the exposure times
      exposure = rdfitscol(filelist(i)+"+1","EXPOSURE");

      // Get the spectrum etc.

      rate =     rdfitscol(filelist(i)+"+1","RATE");
      stat_err = rdfitscol(filelist(i)+"+1","STAT_ERR");

      if( pflag ) {
         dataplot,indgen(256),rate(,num_of_spec),stat_err(,num_of_spec);
         xyouts,0.25,0.82,swid,charsize=1.3,ndc=1;
         xyouts,0.25,0.80,swrite(format="Exposure %10.2f s",exposure(num_of_spec)), \
                      charsize=1.3,ndc=1;
      }

      // include this one in the list of used spectral files;
      filelist_used(++nlist_used) = filelist(i);

      if(!nolog)write,lun,format="Exposure time is %10.2f s\n",exposure(num_of_spec);

      tot_spec += reform(rate(,num_of_spec),256)*exposure(num_of_spec);
      tot_err2 += (reform(stat_err(,num_of_spec),256)*exposure(num_of_spec))^2;
      tot_expo += exposure(num_of_spec);

   }

   if( !nolog ) {
      write,lun,"// total_exposure = ", tot_expo," ; s";
      write,lun,"//  --- Normal exit ---";

      close,lun;
   }

   tot_spec = tot_spec / tot_expo;
   tot_err  = sqrt(tot_err2) / tot_expo;

   if( pflag ) dataplot,indgen(256),tot_spec,tot_err;

   //  write file with collected spectrum;

   write,"New spectrum file: "+fname_spe;

   fh = headfits( fname_spe+"+1" );
   colnum = get_colnum( fh, "rate" );
   fits_bintable_poke, fname_spe+"+1",  1, colnum, float(tot_spec);
   colnum = get_colnum( fh, "stat_err" );
   fits_bintable_poke, fname_spe+"+1",  1, colnum, float(tot_err);
   colnum = get_colnum( fh, "exposure" );
   fits_bintable_poke, fname_spe+"+1",  1, colnum, double(tot_expo);


   //  Update the checksum;

   //+ system,"fchecksum "+fname_spe+" update=yes";


}

/* Function gti_print */

func gti_print( filename, outfile= )
/* DOCUMENT gti_print, filename, outfile=

   Nice print of GTI information in jmxi_gti.fits

   2010-04-16/NJW
*/
{
   if( strlen(filename) > 3 ) {
      postfix = strpart( filename, -2:0);
      if( postfix == ".gz" ) filename = strpart( filename, 1:-3 );
   }
   if( !file_test(filename) ) {
      if( !file_test(filename+".gz") ) {
         error,filename+" not found";
      } else {
         filename = gz_proxy_file( filename, dir="/r12/scratch/njw/proxies" );
      }
   }

   if( typeof(outfile) == "string" ) {
      fo = open(outfile,"a");
   } else fo = [];

   nexts = nfits_extens( filename );

// ---   PASS 1 --- get earliest GTISTART

   gti0 = [];
   for( iext = 1; iext <= nexts; iext++ ) {
      fh = headfits( filename+"+"+itoa(iext) );
      extname = fxpar( fh, "EXTNAME" );
      ending = strpart( extname, -2:0 );
      if( ending == "GTI" ) {
         if( numberof( (gtistart = rdfitscol(filename+"+"+itoa(iext),"start")))) {
            gtistop  = rdfitscol(filename+"+"+itoa(iext),"stop");
            if( is_void(gti0) ) {
               gti0 = min(gtistart);
            } else {
               if( min(gtistart) < gti0 ) gti0 = min(gtistart);
            }
         }
      }
   }

// ---   PASS 2 --- print out the GTI IJD values

   for( iext = 1; iext <= nexts; iext++ ) {
      fh = headfits( filename+"+"+itoa(iext) );
      extname = fxpar( fh, "EXTNAME" );
      ending = strpart( extname, -2:0 );
      write,fo,format="\nExt# %2i - EXTNAME = %s\n\n", iext, extname;
      if( ending == "GTI" ) {
         gtiname = fxpar( fh, "GTI_NAME" );
         if( is_void(gtiname) ) gtiname = "?????";
         write,fo,format="    GTI_NAME = %s\n", gtiname;
         if( numberof( (gtistart = rdfitscol(filename+"+"+itoa(iext),"start")))) {
            gtistop  = rdfitscol(filename+"+"+itoa(iext),"stop");
            ngti = numberof(gtistart);
            for( i = 1; i <= ngti; i++ ) {
               write,fo,format="   %13.8f %13.8f %10.2f %10.2f\n", \
                      gtistart(i), gtistop(i), (gtistart(i)-gti0)*86400., \
                                               (gtistop(i)-gti0)*86400.;
            }
         }
      }
   }

   if( !is_void(fo) ) close,fo;

}

/* Function find_swid_slice */

func find_swid_slice( p0, q0, dp, dq, rev=, tstart=, \
                      tstop=, nof=, silent=, lst=, gal=, constw= )
/* DOCUMENT list = find_swid_slice( p0, q0, dp, dq,
                       [,rev=][, tstart=][, tstop=]
                       [,nof=][, silent=][, lst=][, gal=][, constw=] )

    Find the INTEGRAL SWIDs where the pointings are inside the slice
    p0 - dp/2 <= p <= p0 + dp/2 AND q0 - dq/2 <= q <= q0 + dq/2

    (p,q) is either (ra,dec)- no keyword -  or (lon,lat) when 'gal' is set.

    The interval where ra0 is close to zero or 360 deg is treated as expected.
    No declination is outside [-90,90] degrees.

    2011-01-18/NJW  Cloned from find_swid_radec

   Keywords:
             rev   :  Either scalar integer or 2 element integer
                      array with min and max revolution number
             tstart:  Time in IJD for start of interval
             tstop :  Time in IJD for end of interval
             lst   :  Makes returned list a list of SWIDs in jemx.lst format
                         (overrides setting of 'list')
             nof   :  Flag to avoid file writing
             gal   :  Flag to interpret input coordinates as galactic
             constw:  Constant width in true angle - 'dp' applies for
                      q == 0
*/
{

   local swid, ra, dec, posangle, ut, ijd, telapse, mode1, mode2;

   if( is_void(dq) ) {
      write,"Syntax: list = find_swid_radec_slice(p0,q0, dp, dq, keywords... )";
      return [];
   }
   //
   //  Initialize
   //
   if( is_void(silent) ) silent = 0;
   first = 1;
   tbegin = is_void( tstart ) ? 0.0 : tstart;
   // if( tstart ) { tbegin = tstart; } else { tbegin = 0.0; }
   tend = is_void( tstop ) ? 1.0e9 : tstop;
   // if( tstop ) { tend = tstop; } else { tend = 1.0e9; }
   p0 = double(p0);
   q0 = double(q0);
   dp = double(dp);
   dq = double(dq);
   if( gal ) {
      lon0 = p0;
      lat0 = q0;
      radec0 = equatorial( p0, q0 );
      ra0 = radec0(1);
      dec0 = radec0(2);
   } else {
      ra0 = p0;
      dec0 = q0;
      galac0 = galactic( p0, q0 );
      lon0 = galac0(1);
      lat0 = galac0(2);
   }

   //
   // Get list of pointing files
   //
   base = get_env("J_POINTINGS");

   if( numberof(rev) ) {
      if( typeof(rev) == "string" ) {
         revarr = str2arr( rev );
      } else {
         if( numberof( rev ) == 2 ) {
            revarr = indgen(long(rev(2)-rev(1))+1) + long(rev(1)) - 1;
         } else {
            revarr = long(rev)
         }
      }
      nrev = numberof(revarr);
      for( i = 1; i <= nrev; i++ ) {
         revstr = swrite(revarr(i),format="%4.4i");
         tmp = base+"/pointings_"+revstr+"p.dat";
         if( !silent ) write,tmp;
         ybase = base+"/ysav"+swrite(revarr(i)/100,format="%02i/");
         ytmp = ybase+"pointings_"+revstr+"p.ysav";
         if( !silent ) write,ytmp;
         if( i == 1 ) {
            pf_list = tmp;
            ypf_list = ytmp;
         } else {
            grow,pf_list,tmp;
            grow,ypf_list,ytmp;
         }
      }
   } else {
      pf_list = file_search("pointings_*p.dat", base);
      ypf_list = pf_list;
      npf_list = numberof(pf_list);
      for( i = 1; i <= npf_list; i++ ) {
         pos = strpos( pf_list(i),"_" );
         revstra = strpart(pf_list(i), pos+1:pos+4 );
         drevstra = strpart(pf_list(i),pos+1:pos+2);
         ybase = base+"/ysav"+drevstra;
         ytmp = ybase+"/pointings_"+revstra+"p.ysav";
         ypf_list(i) = ytmp;
      }
   }
   npf_list = numberof(pf_list);
   if( npf_list == 0 ) {
      write,"find_swid_radec_slice error: no pointing files of required kind found";
      return -1;
   }

   if( !nof ) {

      // Open output file

      outfilename = get_next_filename("find_swid_radec_???.dat");
      fout = open(outfilename,"w");
      n = write(fout,format="%s\n","// Results from find_swid_radec_slice");
      n = write(fout,format="// created %s\n", ndate(3));
      n = write(fout,format="// ra0  = %7.3f; deg\n", ra0);
      n = write(fout,format="// dec0 = %7.3f; deg\n", dec0);
      n = write(fout,format="// lon0 = %7.3f; deg\n", lon0);
      n = write(fout,format="// lat0 = %7.3f; deg\n", lat0);
      n = write(fout,format="//   dp = %7.3f; deg\n",   dp);
      n = write(fout,format="//   dq = %7.3f; deg\n",   dq);
      if( tstart ) {
         n = write(fout,format="// tstart = %14.8f; IJD\n", tstart);
      }
      if( tstop ) {
         n = write(fout,format="// tstop = %14.8f; IJD\n", tstop);
      }
      if( numberof(rev) ) {
         if( nrev == 1 ) {
            n = write(fout,format="// rev_begin = %4i\n", revarr(1));
            n = write(fout,format="// rev_end   = %4i\n", revarr(1));
         } else {
            n = write(fout,format="// rev_begin = %4i\n", revarr(1));
            n = write(fout,format="// rev_end   = %4i\n", revarr(0));
         }
      }

   }

   ntotal = 0;
   for( ipf_list = 1; ipf_list <= npf_list; ipf_list++ ) {

      if( file_test(ypf_list(ipf_list)) ) {
         if( !silent ) write,"Reading binary file "+ypf_list(ipf_list);
         bfile = openb( ypf_list(ipf_list) );
         restore,bfile, swid, ra, dec, posangle, ut, ijd, telapse, \
                     mode1, mode2;
         close, bfile;
      } else {
         if( !silent ) write,"Reading usual text file "+pf_list(ipf_list);
         if( ! file_test(pf_list(ipf_list)) ) {
            if( !silent ) write,pf_list(ipf_list)+" was not found";
            continue;
         }
      /*
       * swid = rscol( pf_list(ipf_list),1, str=1, silent=1);
       * ra = rscol( pf_list(ipf_list), 2, silent=1);
       * dec = rscol( pf_list(ipf_list), 3, silent=1);
       * posangle = rscol( pf_list(ipf_list), 4, silent=1);
       * ut = rscol( pf_list(ipf_list), 5, str=1, silent=1);
       * ijd = rscol( pf_list(ipf_list), 6, silent=1);
       * telapse = rscol( pf_list(ipf_list),7, silent=1);
       * mode1 = rscol( pf_list(ipf_list), 8, str=1, silent=1);
       * mode2 = rscol( pf_list(ipf_list), 9, str=1, silent=1);
       */

         rstab,pf_list(ipf_list),9,swid,ra,dec,posangle,ut,ijd,telapse, \
              mode1, mode2,typ="sfffsffss",silent=1;

         bfile = createb( ypf_list(ipf_list) );
         save, bfile,  swid, ra, dec, posangle, ut, ijd, telapse, \
                 mode1, mode2;
         close, bfile;
      }

      lonlat = galactic(ra, dec);
      lon = lonlat(,1);
      lat = lonlat(,2);

      if( gal ) {
         p = lon;
         q = lat;
      } else {
         p = ra;
         q = dec;
      }

      if( constw ) dp = dp/cos(q*pi/180.);

      ip = near360( p0, p, 0.5*dp );
      iq = q >= q0 - 0.5*dq & q <= q0 + 0.5*dq;
      iboth = ip & iq;
      w = where( iboth );
      nw = numberof(w);

      if( nw > 0 ) {

         if( first ) {
            first = 0;
            swid_all = swid(w);
            ra_all = ra(w);
            dec_all = dec(w);
            posangle_all = posangle(w);
            ut_all = ut(w);
            ijd_all = ijd(w);
            telapse_all = telapse(w);
            mode1_all = mode1(w);
            mode2_all = mode2(w);
         } else {
            grow,swid_all,swid(w);
            grow,ra_all,ra(w);
            grow,dec_all,dec(w);
            grow,posangle_all,posangle(w);
            grow,ut_all,ut(w);
            grow,ijd_all,ijd(w);
            grow,telapse_all,telapse(w);
            grow,mode1_all,mode1(w);
            grow,mode2_all,mode2(w);
         }

         ntotal += nw;

      }

   }
   if( ntotal == 0 ) {
      write,"Sorry, no pointings found!";
      return [];
   }

   noutput = 0;
   list = "";
   list_list = [];
   for( i = 1; i <= ntotal; i++ ) {
      // apply the time selection
      if( ijd_all(i) > tbegin && ijd_all(i) < tend ) {
         grow,list_list,swid_all(i);
         if( !nof) {
            write,fout,format="%s%10.4f%9.4f%9.3f  %s%10.3f%6.0f  %s  %s %8.2f %7.0f\n",
            swid_all(i),ra_all(i),dec_all(i),posangle_all(i),
            ut_all(i),ijd_all(i),telapse_all(i),
            mode1_all(i),mode2_all(i),dist_all(i), expo_contrib;
         }
         noutput++;
      }
   }


   lst_list = array("",ntotal);
   for( i = 1; i <= ntotal; i++ ) {
      lst_list(i) = "./scw/"+strpart(swid_all(i),1:4)+"/"+swid_all(i)+".001/swg.fits[1]";
   }

   if( !silent ) {
      write,format="Found %i SWIDs in selected revolutions\n", ntotal;
      if( noutput == 1 ) {
         write,format="%i has been selected by time\n", noutput;
      } else {
         write,format="%i have been selected by time\n", noutput;
      }
      if( !nof ) write,"Output is directed to "+outfilename;

   }

   if(!nof) close,fout;

   if( !is_void(lst) ) { return lst_list; } else {return list_list;}

}

/* Function bti_info */

func bti_info( jemxNum, outfile= )
/* DOCUMENT bti_info, jemxNum, outfile=

     Make a table of most recent BTI information in OSA10.1.

     Outputs to a text file if 'outfile' contains a string (filename).

   2012-04-19/NJW, updated for OSA10.1  2015-04-15/NJW

*/
{
   if( typeof(outfile) == "string" ) {
      p = 1;
      f = open(outfile,"w");
   } else p = 0;

   jstr = itoa(jemxNum);
   dir = "/pack/osa_10.1/rep_base_prod/ic/jmx"+jstr+"/lim";
   name = "jmx"+jstr+"_gnrl_bti_*.fits";
   list = file_search(name,dir);
   write,format="Files available in %s:\n", dir;
   if(p)write,f,format="Files available in %s:\n", dir;
   for(i=1;i<=numberof(list);i++) {
      write,format="   %s\n", basename(list(i));
      if(p)write,f,format="   %s\n", basename(list(i));
   }
   dol = list(0)+"+1"; // Choose most recent
   tstart = rdfitscol(dol,"start");
   tstop = rdfitscol(dol,"stop");
   w = where( tstart > 1097. );
   tstart = tstart(w);
   tstop = tstop(w);
   n = numberof(tstart);

   for( i = 1; i <= n; i++ ) {
      write,format="%12.6f %12.6f  ", tstart(i), tstop(i);
      if(p)write,f,format="%12.6f %12.6f  ", tstart(i), tstop(i);
      revol_start = long(get_revol_no(tstart(i)+0.04));
      revol_stop  = long(get_revol_no(tstop(i)));
      for( revol = revol_start; revol <= revol_stop; revol++ ) {
         if( revol > revol_start ) {
            write,format="%s","                           ";
            if(p)write,f,format="%s","                           ";
      }
         write,format="%04i ", revol;
         if(p)write,f,format="%04i ", revol;
         pointfile = "/r6/jemx/pointings/pointings_"+itoa(revol,4)+".fits";
         if( file_test(pointfile) ) {
            pointdol = pointfile+"+1";
            swid = rdfitscol( pointdol, "swid" );
            tswid1 = rdfitscol( pointdol, "tstart" );
            tswid2 = rdfitscol( pointdol, "tstop" );
            w = where( tswid1 > tstart(i) ); // those later than BTI start
            m = where( tswid2 < tstop(i)  ); // those earlier than BTI stop
            swid_first = numberof(w) ? swid(w(1)) : "000000000000";
            swid_last  = numberof(m) ? swid(m(0)) : "000000000000";
            write,format="  %s  %s\n", swid_first, swid_last;
            if(p)write,f,format="  %s  %s\n", swid_first, swid_last;
         } else {
            write,"  <pointing file not available>";
            if(p)write,f,"  <pointing file not available>";
         }
      }
   }
   if(p) close,f;
}

/* Function bti_struck */

func bti_struck( jemxNum, t1, t2 )
/* DOCUMENT res = bti_struck( jemxNum, t1, t2 )

    Returns "Y" if there is a BTI set up somewhere after 't1' (IJD)
    and before 't2' (IJD). Typically t1 and t2 will be start and stop
    for a SWID or a revolution.

   2012-04-19/NJW, updated for OSA10.1  2015-04-15/NJW
*/
{
   extern BTI_JMX1_TSTART, BTI_JMX1_TSTOP, BTI_JMX2_TSTART, BTI_JMX2_TSTOP;

   jstr = itoa(jemxNum);
   if( t2 < t1 ) { x = t1; t1 = t2; t2 = x; } // swap so t1 <= t2


   if( ( jemxNum == 1 && is_void(BTI_JMX1_TSTART) ) || \
       ( jemxNum == 2 && is_void(BTI_JMX2_TSTART) ) ) { // data reading is required

      dir = "/pack/osa_10.1/rep_base_prod/ic/jmx"+jstr+"/lim";
      name = "jmx"+jstr+"_gnrl_bti_*.fits";
      list = file_search(name,dir);
      write,format="Files available in %s:\n", dir;
      for(i=1;i<=numberof(list);i++) {
         write,format="   %s\n", basename(list(i));
      }
      dol = list(0)+"+1"; // Choose most recent
      if( jemxNum == 1 ) {
         BTI_JMX1_TSTART = rdfitscol(dol,"start");
         BTI_JMX1_TSTOP  = rdfitscol(dol,"stop");
      } else {
         BTI_JMX2_TSTART = rdfitscol(dol,"start");
         BTI_JMX2_TSTOP  = rdfitscol(dol,"stop");
      }
   }
   if( jemxNum == 1 ) {
      bti_tstart = BTI_JMX1_TSTART;
      bti_tstop  = BTI_JMX1_TSTOP;
   } else {
      bti_tstart = BTI_JMX2_TSTART;
      bti_tstop  = BTI_JMX2_TSTOP;
   }

   w = where(bti_tstop > t1 );
   if( numberof(w) == 0 ) return "N";
   if( bti_tstart(w(1)) < t2 ) return "Y";
   return "N";

}

/* Function sunposition */

func sunposition( IJD_or_dattim )
/* DOCUMENT v = sunposition( IJD_or_dattim )
    where v(1) returns RA(Sun) and v(2) returns Dec(Sun)
    in degrees.

    Approximate Solar Coordinates from Naval Oceanography Portal
    http://www.usno.navy.mil/USNO/astronomical-applications/astr...

    The accuracy is reported to better than a couple of arcminutes.

    Translated to Yorick: 2009-10-20/NJW

*/
{
   if( typeof(IJD_or_dattim) == "string" ) {
      ijd = dattim2ijd(IJD_or_dattim);
   } else {
      ijd = IJD_or_dattim;
   }

   dr = pi/180;
   D = ijd - 0.5;

   g = 357.529 + 0.98560028*D;
   g_rad = zero2pi(g*dr);
   q = 280.459 + 0.98564736*D;
   q_rad = zero2pi(q*dr);
   L = q_rad/dr + 1.915*sin(g_rad) + 0.020*sin(2*g_rad);

   R = 1.00014 - 0.01671*cos(g_rad) - 0.00014*cos(2*g_rad);

   e = 23.439 - 0.00000036*D;

   RA = zero2pi(atan(cos(e*dr)*sin(L*dr),cos(L*dr)))/dr;
   d = asin(sin(e*dr)*sin(L*dr))/dr;

   return [RA, d];

}

/* Function utc2ijd */

func utc2ijd( utc )
/* DOCUMENT ijd = utc2ijd ( utc )

   Convert a (array of) date/time string in UTC (such as 2003-01-27T10:48:17) to
   IJD i.e. number of days since Jan 01, 00:00, 2000

   Including 64.184 s offset prior to year 2000 and leap seconds at
      2005-12-31, 2008-12-31, and 2012-06-30 at the end of the day.

   2003-01-27/NJW
   2006-10-03/NJW translated to Yorick
   2006-12-21/NJW updated to account for difference between TT and UTC

   SEE ALSO: dattim2ijd, ijd2dattim, timeconv, ijd2utc, ijd2tt, tt2ijd

*/

{


// Initial data definition

   days_in_month = [31,28,31,30,31,30,31,31,30,31,30,31];
   days_in_month_leap = [31,29,31,30,31,30,31,31,30,31,30,31];
   ijd_tmp = -1.0;
   ijd = -1.0;

   acc_days = array(long,12);
   acc_days_leap = array(long,12);
   for( i = 1; i <= 11; i++ ) {
      acc_days(i+1) = acc_days(i) + days_in_month(i);
      acc_days_leap(i+1) = acc_days_leap(i) + days_in_month_leap(i);
   }

   first = 1;

   // Analyze input (array of) string(s) and check format

   if( typeof( utc ) != "string" ) {
      write,"1. argument is not of type string";
      return;
   }
   nelem = numberof(utc);

   for( i = 1; i <= nelem; i++ ) {

      istr = swrite(format="%i", i);
      curstr = strcompress(utc(i),all=1);
      len = strlen(curstr);

      if( len  <  13 ) {
         write,format="##34## %s %s invalid format (too short)\n", istr, utc(i);
         continue;
      }

   //  Requirements   2002-11-16T23:45:56.342
   //            or   2002-11-16T23:45:56
   //            or   2002-11-16T23:45
   //            or   2002-11-16T23
   //      One "T", two "-"s

      ch = strpart(curstr,11:11)

      if( ch  !=  "T" ) {
         write,format="##35## %s %s invalid format (no 'T' in pos 11)\n",\
                    istr, utc(i);
         continue;
      }

      ch = strpart(curstr,5:5);

      if( ch  !=  "-" ) {
         write,format="##35## %s %s invalid format (no '-' in pos 5)\n",\
                    istr, utc(i);
         continue;
      }

      ch = strpart(curstr,8:8);

      if( ch  !=  "-" ) {
         write,format="##36## %s %s invalid format (no '-' in pos 8)\n",\
                    istr, utc(i);
         continue;
      }

      year = strpart(curstr,1:4);

      if( !is_digit(year) ) {
         write,format="##37## %s %s invalid format (year is not a number)\n",\
                    istr, utc(i);
         continue;
      }

      iyear = atoi(year);
      if( iyear  <  2000 || iyear  >  2099 ) {
         write,format="##38## %s %s invalid format (requirement: 2000 <= year <= 2100)\n",\
                    istr, utc(i);
         continue;
      }

      cur_year_is_leap = ( iyear%4 == 0 ) ? 1 : 0;

      month = strpart(curstr,6:7);

      if( !is_digit(month) ) {
         write,format="##39## %s %s invalid format (month is not a number)\n",\
                    istr, utc(i);
         continue;
      }

      imonth = atoi(month);
      if( imonth  <  1 || imonth  >  12 ) {
         write,format="##40## %s %s invalid format (illegal month number)\n",\
                    istr, utc(i);
         continue;
      }

      day = strpart(curstr,9:10);

      if( !is_digit(day) ) {
         write,format="##41## %s %s invalid format (day is not a number)\n",\
                    istr, utc(i);
         continue;
      }
      iday = atoi(day);

      daymax = days_in_month(imonth);
      if( cur_year_is_leap && imonth == 2 ) ++daymax;

      if( iday  <  1 || iday  >  daymax ) {
         write,format="##42## %s %s invalid format (invalid day number)\n",\
                    istr, utc(i);
         continue;
      }

      hour = strpart(curstr,12:13);

      if( !is_digit(hour) ) {
         write,format="##43## %s %s invalid format (hour is not a number)\n",\
                    istr, utc(i);
         continue;
      }

      if( atoi(hour) < 0 || atoi(hour) > 23 ) {
         write,format="##44## %s %s invalid format (invalid hour number)\n",\
                    istr, utc(i);
         continue;
      }

      minute = "0";
      second = "0";

      if( len >= 16 ) { // The minute is given as well

         ch = strpart(curstr,14:14);
         if( ch != ":" ) {
            write,format="##45## %s %s invalid format (no ':' in pos 14)\n",\
                    istr, utc(i);
            continue;
         }

         minute = strpart(curstr,15:16);

         if( !is_digit(minute) ) {
            write,format="##46## %s %s invalid format (minute is not a number)\n",\
                    istr, utc(i);
            continue;
         }

         if( atoi(minute) < 0 || atoi(minute) > 59 ) {
            write,format="##47## %s %s invalid format (invalid minute number)\n",\
                    istr, utc(i);
            continue;
         }

         if( len >= 19 ) { // The second is given as well

            ch = strpart(curstr,17:17);
            if( ch  !=  ":" ) {
               write,format="##48## %s %s invalid format (no ':' in pos 17)\n",\
                    istr, utc(i);
               continue;
            }

            //140708+ second = strpart(curstr,18:19);
            second = strpart(curstr,18:);

            if( !is_number(second) ) {
               write,format="##49## %s %s invalid format (second is not a number)\n",\
                    istr, utc(i);
               continue;
            }

            if( atof(second) < 0 || atof(second) > 59.9999999 ) {
               write,format="##50## %s %s invalid format (invalid second number)\n",\
                    istr, utc(i);
               continue;
            }

         }

      }

      num_years = long(iyear - 2000);
      num_leap_years = (num_years+3L)/4; // before current year

      num_days = num_years * 365 + num_leap_years;

      if( cur_year_is_leap ) {
         num_days = num_days + acc_days_leap(imonth) + iday - 1;
      } else {
         num_days = num_days + acc_days(imonth) + iday - 1;
      }
      num_seconds = atoi(hour)*3600.0 + atoi(minute)*60.0 + atof(second);

      // account for leap seconds
      t_dif_s = 64.184;
      if( iyear >= 2006 ) t_dif_s = 65.184;               // at end of 2005
      if( iyear >= 2009 ) t_dif_s = 66.184;               // at end of 2008
      if( iyear == 2012 && imonth > 6 ) t_dif_s = 67.184; // mid 2012
      if( iyear >= 2013 ) t_dif_s = 67.184;               // after mid 2012

      t_dif_ijd = t_dif_s / 86400.0;
      ijd_tmp = num_days + num_seconds/(24.0*3600.0) + t_dif_ijd;

      if( first ) {
         ijd = ijd_tmp;
         first = 0;
      } else {
          grow, ijd, ijd_tmp;
      }



   }

   return ijd;

}

/* Function ttutcdif */

func ttutcdif( x )
/* DOCUMENT d = ttutcdif( utc_or_ijd )

    Returns the difference (in seconds) between TT and UTC for a given
    UTC time string (YYYY-MM-DDTHH:MM:SS) or a given IJD.

*/
{
   if( structof( x ) != string ) x = ijd2utc( x );
   year = atoi(strpart(x, 1:4));
   month= atoi(strpart(x,6:7));
   // account for leap seconds
   t_dif_s = 64.184;
   if( year >= 2006 ) t_dif_s = 65.184;               // at end of 2005
   if( year >= 2009 ) t_dif_s = 66.184;               // at end of 2008
   if( year == 2012 && month > 6 ) t_dif_s = 67.184; // mid 2012
   if( year >= 2013 ) t_dif_s = 67.184;               // after mid 2012

   return t_dif_s;

}


/* Function tt2ijd */

func tt2ijd( tt )
/* DOCUMENT ijd = tt2ijd ( tt )

   Convert a (array of) date/time string in TT (such as 2003-01-27T10:48:17) to
   IJD i.e. number of days since Jan 01, 00:00, 2000

   SEE ALSO: dattim2ijd, ijd2dattim, timeconv, ijd2utc, ijd2tt, utc2ijd

*/

{


// Initial data definition

   days_in_month = [31,28,31,30,31,30,31,31,30,31,30,31];
   days_in_month_leap = [31,29,31,30,31,30,31,31,30,31,30,31];
   ijd_tmp = -1.0;
   ijd = -1.0;

   acc_days = array(long,12);
   acc_days_leap = array(long,12);
   for( i = 1; i <= 11; i++ ) {
      acc_days(i+1) = acc_days(i) + days_in_month(i);
      acc_days_leap(i+1) = acc_days_leap(i) + days_in_month_leap(i);
   }

   first = 1;

   // Analyze input (array of) string(s) and check format

   if( typeof( tt ) != "string" ) {
      write,"1. argument is not of type string";
      return;
   }
   nelem = numberof(tt);

   for( i = 1; i <= nelem; i++ ) {

      istr = swrite(format="%i", i);
      curstr = strcompress(tt(i),all=1);
      len = strlen(curstr);

      if( len  <  13 ) {
         write,format="##34## %s %s invalid format (too short)\n", istr, tt(i);
         continue;
      }

   //  Requirements   2002-11-16T23:45:56.342
   //            or   2002-11-16T23:45:56
   //            or   2002-11-16T23:45
   //            or   2002-11-16T23
   //      One "T", two "-"s

      ch = strpart(curstr,11:11)

      if( ch  !=  "T" ) {
         write,format="##35## %s %s invalid format (no 'T' in pos 11)\n",\
                    istr, tt(i);
         continue;
      }

      ch = strpart(curstr,5:5);

      if( ch  !=  "-" ) {
         write,format="##35## %s %s invalid format (no '-' in pos 5)\n",\
                    istr, tt(i);
         continue;
      }

      ch = strpart(curstr,8:8);

      if( ch  !=  "-" ) {
         write,format="##36## %s %s invalid format (no '-' in pos 8)\n",\
                    istr, tt(i);
         continue;
      }

      year = strpart(curstr,1:4);

      if( !is_digit(year) ) {
         write,format="##37## %s %s invalid format (year is not a number)\n",\
                    istr, tt(i);
         continue;
      }

      iyear = atoi(year);
      if( iyear  <  2000 || iyear  >  2099 ) {
         write,format="##38## %s %s invalid format (requirement: 2000 <= year <= 2100)\n",\
                    istr, tt(i);
         continue;
      }

      cur_year_is_leap = ( iyear%4 == 0 ) ? 1 : 0;

      month = strpart(curstr,6:7);

      if( !is_digit(month) ) {
         write,format="##39## %s %s invalid format (month is not a number)\n",\
                    istr, tt(i);
         continue;
      }

      imonth = atoi(month);
      if( imonth  <  1 || imonth  >  12 ) {
         write,format="##40## %s %s invalid format (illegal month number)\n",\
                    istr, tt(i);
         continue;
      }

      day = strpart(curstr,9:10);

      if( !is_digit(day) ) {
         write,format="##41## %s %s invalid format (day is not a number)\n",\
                    istr, tt(i);
         continue;
      }
      iday = atoi(day);

      daymax = days_in_month(imonth);
      if( cur_year_is_leap && imonth == 2 ) ++daymax;

      if( iday  <  1 || iday  >  daymax ) {
         write,format="##42## %s %s invalid format (invalid day number)\n",\
                    istr, tt(i);
         continue;
      }

      hour = strpart(curstr,12:13);

      if( !is_digit(hour) ) {
         write,format="##43## %s %s invalid format (hour is not a number)\n",\
                    istr, tt(i);
         continue;
      }

      if( atoi(hour) < 0 || atoi(hour) > 23 ) {
         write,format="##44## %s %s invalid format (invalid hour number)\n",\
                    istr, tt(i);
         continue;
      }

      minute = "0";
      second = "0";

      if( len >= 16 ) { // The minute is given as well

         ch = strpart(curstr,14:14);
         if( ch != ":" ) {
            write,format="##45## %s %s invalid format (no ':' in pos 14)\n",\
                    istr, tt(i);
            continue;
         }

         minute = strpart(curstr,15:16);

         if( !is_digit(minute) ) {
            write,format="##46## %s %s invalid format (minute is not a number)\n",\
                    istr, tt(i);
            continue;
         }

         if( atoi(minute) < 0 || atoi(minute) > 59 ) {
            write,format="##47## %s %s invalid format (invalid minute number)\n",\
                    istr, tt(i);
            continue;
         }

         if( len >= 19 ) { // The second is given as well

            ch = strpart(curstr,17:17);
            if( ch  !=  ":" ) {
               write,format="##48## %s %s invalid format (no ':' in pos 17)\n",\
                    istr, tt(i);
               continue;
            }

            //140708+ second = strpart(curstr,18:19);
            second = strpart(curstr,18:);

            if( !is_number(second) ) {
               write,format="##49## %s %s invalid format (second is not a number)\n",\
                    istr, tt(i);
               continue;
            }

            if( atof(second) < 0 || atof(second) > 59.9999999 ) {
               write,format="##50## %s %s invalid format (invalid second number)\n",\
                    istr, tt(i);
               continue;
            }

         }

      }

      num_years = long(iyear - 2000);
      num_leap_years = (num_years+3L)/4; // before current year

      num_days = num_years * 365 + num_leap_years;

      if( cur_year_is_leap ) {
         num_days = num_days + acc_days_leap(imonth) + iday - 1;
      } else {
         num_days = num_days + acc_days(imonth) + iday - 1;
      }
      num_seconds = atoi(hour)*3600.0 + atoi(minute)*60.0 + atof(second);

      ijd_tmp = num_days + num_seconds/(24.0*3600.0);

      if( first ) {
         ijd = ijd_tmp;
         first = 0;
      } else {
          grow, ijd, ijd_tmp;
      }



   }

   return ijd;

}

/* Function ijd2utc */

func ijd2utc( ijd )
/* DOCUMENT utc = ijd2utc( ijd )

    Converts scalar or array of IJD values to 'Gregorian' strings:
     YYYY-MM-DDTHH:MM:SS.ddd UTC

   Including 64.184 s offset prior to year 2000 and leap seconds at
      2005-12-31, 2008-12-31, and 2012-06-30 at the end of the day.

   SEE ALSO: dattim2ijd, ijd2dattim, timeconv, ijd2tt, utc2ijd, tt2ijd
*/
{
   ijd = double(ijd);
   if( is_scalar(ijd) ) return _ijd2utc( ijd );
   n = numberof( ijd );
   res = array(string, n );
   for(i = 1; i <= n; i++ ) res(i) = _ijd2utc( ijd(i) );
   return res;
}

func _ijd2utc( ijd )
/* DOCUMENT utc = _ijd2utc( ijd )

   Work function for 'utc2ijd', takes only a scalar.

   2006-10-03/NJW

*/
{

   if( ijd < 0.1 ) {
      write,format="Does not work for IJD = %f\n",ijd;
      return "";
   }

   days_in_month = [31,28,31,30,31,30,31,31,30,31,30,31];

// Including the difference between TT (Terrestrial Time) and UTC
// See FAQ of ISDC home page

   t_dif_s = 64.184;
   // leap second end of year 2005
   if( ijd > 2192.0007428704 ) t_dif_s = 65.184;
   // leap second end of year 2008
   if( ijd > 3288.0007544444 ) t_dif_s = 66.184;
   // leap second mid year 2012
   if( ijd > 4565.0007660185 ) t_dif_s = 67.184;
   t_dif_ijd = t_dif_s / 86400.0;
   ijd_utc = ijd - t_dif_ijd;

   dat = "2000-01-01T00:00:00.000";

   // Get year

   for( year = 2001; year <= 2099; year++ ) {

      str_year = swrite(format="%4.4i",year);
      dat = strput( dat, str_year, 1 );
      t = utc2ijd(dat);
      if( t > ijd ) break;

   }

   year--;
   str_year = swrite(format="%4.4i", year);
   dat = strput(dat, str_year, 1 );
   if( year%4 == 0 ) days_in_month(2) = 29; // leap year

   // Get month

   for( month = 1; month <= 12; month++ ) {

      str_month = swrite(format="%2.2i", month);
      dat = strput(dat,str_month,6);
      t = utc2ijd(dat);
      if( t > ijd ) break;

   }

   month--;
   str_month = swrite(format="%2.2i", month );
   dat = strput(dat,str_month,6);

   // Get day

   for( day = 1; day <= days_in_month(month); day++ ) {

      str_day = swrite(format="%2.2i", day );
      dat = strput(dat,str_day,9);
      t = utc2ijd(dat);
      if( t > ijd ) break;

   }

   day--;
   str_day = swrite(format="%2.2i", day);
   dat = strput(dat,str_day,9);

   frac = ijd_utc - floor(ijd_utc);
   h = frac * 24;
   hour = toint(h);
   m = (h - hour) * 60;
   minute = toint(m);
   //140708+ s = (m - minute) * 60;
   //140708+ second = toint(s);
   second = (m - minute) * 60;

   str_hour = swrite(format="%2.2i", hour);
   dat = strput(dat,str_hour,12);

   str_minute = swrite(format="%2.2i", minute);
   dat = strput(dat,str_minute,15);

   //140708+ str_second = swrite(format="%2.2i", second);
   str_second = swrite(format="%06.3f", second);
   dat = strput(dat,str_second,18);

   return dat;

}

/* Function ijd2tt */

func ijd2tt( ijd )
/* DOCUMENT tt = ijd2tt( ijd )

    Converts scalar or array of IJD values to 'Gregorian' strings:
     YYYY-MM-DDTHH:MM:SS.ddd TT

   SEE ALSO: dattim2ijd, ijd2dattim, timeconv, ijd2utc, utc2ijd, tt2ijd
*/
{
   ijd = double(ijd);
   if( is_scalar(ijd) ) return _ijd2tt( ijd );
   n = numberof( ijd );
   res = array(string, n );
   for(i = 1; i <= n; i++ ) res(i) = _ijd2tt( ijd(i) );
   return res;
}

func _ijd2tt( ijd )
/* DOCUMENT tt = _ijd2tt( ijd )

   Work function for 'tt', takes only a scalar.

   2014-07-07/NJW, based on _ijd2dattim

*/
{

   if( ijd < 0.1 ) {
      write,format="Does not work for IJD = %f\n",ijd;
      return "";
   }

   days_in_month = [31,28,31,30,31,30,31,31,30,31,30,31];

   dat = "2000-01-01T00:00:00.000";

   // Get year

   for( year = 2001; year <= 2099; year++ ) {

      str_year = swrite(format="%4.4i",year);
      dat = strput( dat, str_year, 1 );
      t = tt2ijd(dat);
      if( t > ijd ) break;

   }

   year--;
   str_year = swrite(format="%4.4i", year);
   dat = strput(dat, str_year, 1 );
   if( year%4 == 0 ) days_in_month(2) = 29; // leap year

   // Get month

   for( month = 1; month <= 12; month++ ) {

      str_month = swrite(format="%2.2i", month);
      dat = strput(dat,str_month,6);
      t = tt2ijd(dat);
      if( t > ijd ) break;

   }

   month--;
   str_month = swrite(format="%2.2i", month );
   dat = strput(dat,str_month,6);

   // Get day

   for( day = 1; day <= days_in_month(month); day++ ) {

      str_day = swrite(format="%2.2i", day );
      dat = strput(dat,str_day,9);
      t = tt2ijd(dat);
      if( t > ijd ) break;

   }

   day--;
   str_day = swrite(format="%2.2i", day);
   dat = strput(dat,str_day,9);

   frac = ijd - floor(ijd);
   h = frac * 24;
   hour = toint(h);
   m = (h - hour) * 60;
   minute = toint(m);
   //140708+ s = (m - minute) * 60;
   //140708+ second = toint(s);
   second = (m - minute) * 60;

   str_hour = swrite(format="%2.2i", hour);
   dat = strput(dat,str_hour,12);

   str_minute = swrite(format="%2.2i", minute);
   dat = strput(dat,str_minute,15);

   //140708+ str_second = swrite(format="%2.2i", second);
   str_second = swrite(format="%06.3f", second);
   dat = strput(dat,str_second,18);

   return dat;

}
