#include "xfit_package.i"
#include "nustar.i"

/* Function mt_rayor */

extern mt_rayor;
/* DOCUMENT Overview of raytracing procedures

Setup system:
  > mt_setup_system,"system_xxx.scm"
  > mt_log[, logfilename]    Initiate logging

Generate optical module file:
      Independent or master module
  > mt_create_om_par1,filename="om_xxx.fits" (variable mirror lengths)
  > mt_create_om_par2,filename="om_xxx.fits"
  > mt_create_om_con2,filename="om_xxx.fits"
  > mt_create_om_con4,filename="om_xxx.fits"
      Secondary (slave) module
  > mt_create_om_hyp2,filename="om_xxx2.fits"
  > mt_create_om_con3,filename="om_xxx2.fits"
  > mt_create_om_con5,filename="om_xxx2.fits"
      Make new one from previously loaded one:
  > mt_upd_om (interactive)
  > mt_write_om, filename

Help for designing Wolter-1 conical approx. gap-telescope
  > mt_tel_design, sysname

Update coating information in module file:
  > mt_upd_om_coating,"mircoat_xxx.scm","om_xxx.fits"

Create fake scattering table files:
  > mt_fake_scatter_data,"scat_xyz.fits",fwhm=,angle_max=,ener_max=,coat=

Create scattering table files from ASCII tables:
  > mt_scatter_data_file (see separate help)
  > mt_scatter_data_file_type2 (see separate help)

Update or find scattering width or distribution in scattering table file:
  > mt_upd_scatter_width (see separate help)
  > mt_get_scatter_hpd(energy, angle_in) [keV, radians]
  > distri = mt_sel_scatter( energy, angle_in, >rcoef )
  > mt_scatfrac, scatterfile, scatfrac

Create mirror deformation file
  > mt_mk_mdeform_file,"filename",mode,parameter,nz=,naz=
  > mt_gravity_bend,gbend,mlength=

Administrate blocking by spokes
  > mt_spoke_read, filename, pos=
  > mt_spoke_blocking, xyphot, pos=

Run a case
  > mt_run,energy,src_offaxis,src_azimuth, <keywords>
      or
  > mt_run,source_flux_file,src_offaxis,src_azimuth, <keywords>
      or
  > mt_run,sky_definition_file, <keywords>
      or
  > mt_run,labsource_definition_file,exposure=, <keywords>

  For several runs in the background (on Unix/linux systems with several CPUs)
  > mt_bg_run,system_file,energy,src_offaxis,src_azimuth, <keywords>
      or
  > mt_bg_run,system_file,sky_definition_file, <keywords>

Define the photons
  > mt_def_photons
  > mt_pre_def_photons

Apply aperture stop
  > mt_aperture_stop, z_position, radius

Apply 'telescope rocking'
  > mt_jitter, angle // in radians

Smooth input image
  > res = mt_blur( direc, angle, nphot)

Show setup and results
  > mt_qimage[,dz],size=          focal plane image
  > mt_e1image[,dz],size=         image at entrance plane
  > mt_mirdiag[,om_file],rr=,gl=  mirror diagram, configuration of shells
  > mt_rayplot, iphot, mod_num    photon path through optical module
  > mt_drayplot, iphot            photon path through two optical modules
  > mt_stat[, status][,w=]        print out statistics
  > mt_mirplot, mirnum, azimuth   plot mirror surface shape
  > mt_photpr, iphot              print tags for specified photon

Analyze result (finding HPD)
  > mt_analysis, phd, photon_file=, frac=, geom=

Save results
  > mt_save,mode="efps"

Calculate effective area
  > mt_eff_area_photons,earr=,samp=
  > mt_bg_run_eff_area,
  > mt_eff_area_quick,earr=,outfile=
  > mt_get_mirror_eff_factors

Plot reflectivity curves
  > mt_reflplot

Other utilities
  > mt_propagate, znew[, iphot]
  > mt_translate, [iphot, ], dx=, dy=
  > rcoef = mt_get_rcoef( energy, angle )
  > mt_substr_volume()

Observation simulator
  > mt_dxb2skydef
  > mt_skyima2skyspec
  > mt_skyspec2skydef
  > mt_catalog2skydef
  > mt_build_cat
  > mt_simulation
  > mt_evt_ds9
  > mt_trace_events

Observation simulation by detector
  > mt_detector,offset=, cont=
  > mt_det_add_bkg[, filename]
  > mt_det_add_instr_bkg[, filename]
  > mt_det_add_dxb_bkg[, filename]
  > mt_det_image,outfile=,emin=,emax=,bkglvl=


Version 4.6.7, 2020-03-09/NJW

*/

/* Function mt_init */

/********************************************************
 *   An initilization macro for the MT_RAYOR package
 *   for raytracing a variety of X-ray telescopes
 *
 *   Naming rule:
 *   External variables shall have a name with the first letter
 *   as a capital (as the only one)
 *
 ***********************************************************/

Version = "4.6.7"; // Setting Version external variable
write,"Loading MT_RAYOR Version "+Version;

/**********************************************************
     Changes

   4.2.2  2012-03-01 Added n_dist_angles keyword in mt_fake_scatter_data
                     and made it to give extension name SCATTER_TYPE1
   4.2.1  2012-02-21 Added Exposure keyword in mt_save,mode="e" files
   4.4.0  2012-06-04 Added function mt_gravity_bend
   4.4.1  2012-07-04 kkkk
   4.4.2  2012-09-14 Introduced dead pixel map in the detector description
                     and translation.
   4.4.5  2013-01-24 Introduced storage of flux arrays in the extended source
                     simulation (using mem_storage.i package).
   4.4.6  2013-05-03 Added the 'bounce' element to the Evlist struct
   4.5.0  2013-05-23 Introduced the XSPEC models in the extended source simulations
   4.5.1  2013-05-23 Economized in the number of XSPEC models to evaluate
   4.5.2  2013-06-13 Economized further in the number of XSPEC models to evaluate
   4.5.4  2013-07-31 Small changes
   4.5.5  2013-08-09 Added mt_catalog2skydef and mt_build_cat
   4.5.6  2013-08-16 Added mt_tel_design
   4.5.7  2013-12-16 Various changes to adapt to investigating an X-ray concentrator
                     or 'collector' (stop assuming mirror angle > 0)
   4.5.8  2014-02-20 Relaxed the requirement that parameter maps should have same
                     coordinates and dimensions as the sky-image in mt_skyima2skyspec
   4.5.9  2014-03-25 Dropped the use of 'kwds_setlongstr' since the long string
                     convention is now incorporated in the 'kwds_set' function.
   4.6.0  2014-04-24 Using bg running for sky simulations by splitting the
                     sky definition file.
   4.6.1  2014-10-08 Bug fix in mt_spoke_blocking to always return a 2-element array
   4.6.2  2014-12-05 Put 'dead_pixel_map' for detector into event file to let 'aspec'
                     be able to produce an exposure map from the event file.
   4.6.3  2015-01-15 Add the function 'mt_simulation' to simulate observations.
   4.6.4  2015-09-01 Add the function 'mt_scatfrac' to adjust ratio
                     scattered part/total distribution reflected
   4.6.5  2016-05-19 Used better accuracy for hitting mirror
   4.6.6  2016-09-15 Various small updates
   4.6.7  2020-03-09 Various small updates

************************************************************/

randomize; // Start a different randomization every time mt_rayor is included.

/*
 *  Include all required functions
 */

//  Define the struct to return the photons that result
//  from the raytracing process

struct s_Ray {
   double E(3);       // current position
   double R(3);       // current direction
   double angle_in1;  // ingoing grazing angle, first reflection
   double angle_out1; // outgoing grazing angle, first reflection
   double angle_in2;  // ingoing grazing angle, second reflection
   double angle_out2; // outgoing grazing angle, second reflection
   double rcoef;      // reflection coefficient
   double energy;     // photon energy
   long mirror;       // mirror number
   long status;       // status flag, zero for successful photon
   long bounce;       // bounce flag
// following are only for display purposes:
   double E1(3);      // position at entrance of first module
   double E2(3);      // position at entrance of second module
   double I1(3);      // position of first reflection
   double I2(3);      // position of second reflection
   double D1(3);      // direction before first reflection
   double D2(3);      // direction after first reflection
// for debugging:
   long flag;         // source number from mt_pre_def_photons
}

//  Define the struct to contain the events i.e. after
//  detection in a detector at the focal plane

struct s_MTEvent{
   int rawx;
   int rawy;
   float detx;
   float dety;
   int pha;
   float energy;
   int flag;
   int bounce;
   long phs_index;
}

//  Define the struct to link the coating with the
//  appropriate scatter definition file

struct s_Coat_list {
   long id;
   string file;
}

Logflag = 0;
STEPSIZE = 1.0; // Default value for 'infinitisimal' step in
                // function 'deriv'
TOLERANCE = 0.0005; // Distance to mirror to accept a hit in
                    // case no deformation is used

// Determine the OSTYPE
OSTYPE = get_env("OSTYPE");
if( strlen(OSTYPE) == 0 ) OSTYPE = "nonx";

/***************************************************************
 * External variables:
 *
 *    See list in PDB file : extvar.pdb
 *
 *
 *      Debugging
 *
 *   Logflag       :    Flag for saving logging information
 *   Logfilename   :    File to receive logging information
 ********************************************************************/

/* Function mt_log */

func mt_log( logfilename, logflag=, stop= )
/* DOCUMENT mt_log, logfilename, logflag=, stop=

   Initiates logging the session. Output in 'logfilename'
   Default name : mt_log_nnnn.txt
       where 'nnnn' is a serial number
   Keyword   stop   :  Stop logging to file
            logflag :  The higher the more output
                       (sets Logflag and mt_log_entry is only
                        operational if level <= Logflag)

   Version 1.7  2008-12-18/NJW
*/
{
   extern Logflag, Logfilename, Version;

  /*
   *  Every time an entry is made to the log file it is opened
   *  in append mode and closed afterwards so that the information
   *  is available even in case of a program crash
   */

   if( stop ) {
      Logflag = 0;
      write,format="Ended loggin%s\n","g";
      return;
   }

   // Define the name of the log file if none is given

   if( is_void(logfilename) ) {
      logfilename = get_next_filename("mt_log_????.txt");
   }
   Logfilename = logfilename;
   Logflag = 1; // Default value - minimal output to logfile
   if( !is_void(logflag) ) {
      // reject values <= 0
      if( logflag > 1 ) Logflag = logflag;
   }
   write,format="Logging in: %s\n", Logfilename;
   mt_log_entry,1,"MT_RAYOR-"+Version+" logfile started "+ndate(3), \
        swrite(format="level %i", Logflag), "";

}

/* Function mt_log_entry */

func mt_log_entry( level, str1, .. )
/* DOCUMENT mt_log_entry, level, str1, ..

   writes str1 and following strings to Logfilename
   in append mode if Logflag >= level

   Version 1.7   2008-12-18/NJW
*/
{
   extern Logflag, Logfilename;

   // Only output to log file if Logflag is true and
   // Logflag >= level

   if( !Logflag ) return;
   if( Logflag < level ) return;

   logstream = open( Logfilename, "a" );
   write,logstream,format="%s\n", str1;
   while( more_args() ) {
      text = next_arg();
      for( i = 1; i <= numberof(text); i++ ) {
         write,logstream,format="   %s\n", text(i);
      }
   }
   close, logstream;
}

/* Function mt_analysis */

func mt_analysis( phs, photon_file=, geom=, frac=, allbounce=, silent= )
/* DOCUMENT mt_analysis[,phs], photon_file=, geom=, frac=, allbounce=, silent=
        or  HPD (mm) = mt_analysis([,phs], photon_file=, geom=, frac=, allbounce=, \
                          silent= )

   Derives the HPD of the status==0  AND bounce==(2*Num_modules-1) photons
   and Rcoef accepted photons.
   If neither argument nor keyword 'photon_file' is given then the
   memory contents (array 'Phs') is used.

   If argument 'phs' is  given then data is taken from that.

   If keyword 'photon_file' is given (and no 'phs' argument )
   then this file is read and data used.
   If keyword 'geom' is set then the Rcoef will be set to one.

   Keyword 'frac' : The fraction of weighted counts inside the diameter
                    Defaults to 0.5 (HPD)
   Keyword 'allbounce' : Only Phs.status==0 condition is applied
                         (eliminates the selection on 'bounce' i.e. ghosts are included)
   Keyword 'silent'    : Only active if called as a function

   The result is based on the current position of the photons
   (z = 0 in the focal plane). Calling 'mt_propagate' prior to
   'mt_analysis' can be used to investigate other z-planes.

   If called as a subroutine the result will be printed on the screen.

     2007-11-06/NJW, version 1.1
   Version 2.1   2010-04-16/NJW
*/
{

// ynclude = zaa
  extern  Focal_length,  Num_modules,   Phs,           Zfocus;
// yxclude =

   local cE;

   if( is_void(frac) ) frac = 0.5;

   if( !is_void(phs) ) {
      photon_file = []; // overriding keyword
   } else {
      eq_nocopy, phs, Phs;
   }
   if( silent && am_subroutine() ) silent = 0;

   if( is_void(photon_file) ) {

      eq_nocopy, cE, phs.E;
      good_bounce = 2*Num_modules - 1; // 1 if a single module
                                       // 3 if two modules
      phs_bounce = phs.bounce;
      if( allbounce ) phs_bounce = phs.bounce * 0 + good_bounce;
      w = where( phs.status == 0 & phs_bounce == good_bounce );
      nw = numberof(w);
      if( nw == 0 ) {
         write,"No photons with OK status (status 0 and bounce "+itoa(good_bounce)+")";
         return [];
      }
      detx = cE(1,w);
      dety = cE(2,w);
      rcoef = phs(w).rcoef;
      s1=swrite(format="Analysis based on Phs with %i status==0 and bounce==%i photons", nw, good_bounce);
      s2=swrite(format="   The focal length is %.2f m\n", Focal_length/1000.0);
      if( !silent ) {
         write,format="%s\n", s1;
         write,format="%s\n", s2;
      }
      cur_focal_length = Focal_length;
      mt_log_entry,1,"mt_analysis",s1,s2;

   } else {

      if( !file_test(photon_file) ) {
         write,"Photon file: "+photon_file+" was not found - skip.";
         return [];
      }
      hdr = headfits(photon_file+"+1");
      detx = rdfitscol(photon_file+"+1","detx");
      dety = rdfitscol(photon_file+"+1","dety");
      rcoef = rdfitscol(photon_file+"+1","rcoef");
      status = rdfitscol(photon_file+"+1","status" );
      bounce = rdfitscol(photon_file+"+1","bounce" );
      cur_focal_length = fxpar(hdr,"FOCALLEN");
      num_modules = fxpar(hdr,"NUMMODLS");
      good_bounce = 2*num_modules - 1; // 1 if a single module
                                       // 3 if two modules
      if( allbounce ) bounce = bounce * 0 + good_bounce;
      w = where( status == 0 & bounce == good_bounce );
      nw = numberof(w);
      if( nw == 0 ) {
         write,"No photons with OK status and bounce";
         return [];
      }

      detx = detx(w);
      dety = dety(w);
      rcoef = rcoef(w);

      hdr = headfits( photon_file+"+1" );
     /*
      *  Get focal length from the Optical Module file
      *  If omfile2 keyword exists then use that
      */
      omfile = fxpar( hdr, "omfile2" );
      if( is_void(omfile) ) {
         omfile = fxpar( hdr, "omfile1" );
         if( is_void(omfile) ) error,"##16## OMFILE1/2 keyword does not exist";
      }
      thdr = headfits( omfile+"+1" );
      focal_length = fxpar( thdr, "fclength" );
      s1=swrite(format="Photon file: %s with %i status==0 and bounce==%i photons", \
             photon_file, nw, good_bounce);
      s1=swrite(format="   The focal length is %6.2f m", cur_focal_length/1000.0);
      if( am_subroutine() ) {
         write,format="%s\n", s1;
         write,format="%s\n", s2;
      }
      mt_log_entry,1,"mt_analysis",s1,s2;

   }

// Determine HPD = Half Power Diameter with Rcoef weighting

   if( geom ) rcoef() = 1.; // neutralize the Rcoef information
   n_photons = numberof(detx);
   // get best position
   xp = wavg(detx,rcoef);
   yp = wavg(dety,rcoef);
   if( am_subroutine() ) {
      write,format="Center with all %i photons: %7.3f %7.3f\n", n_photons, xp, yp;
   }
   xrms = wrms(detx,rcoef);
   yrms = wrms(dety,rcoef);
   w = where( abs(detx - xp) < 3. * xrms );
   xp = wavg(detx(w),rcoef(w));
   nwx = numberof(w);
   w = where( abs(dety - yp) < 3. * yrms );
   yp = wavg(dety(w),rcoef(w));
   nwy = numberof(w);
   s1=swrite(format="Improved center with %i, %i photons: %7.3f %7.3f", \
           nwx, nwy, xp, yp);

   r = sqrt((detx-xp)^2 + (dety-yp)^2);
   is = sort(r);
   r = r(is);
   rcoef = rcoef(is);
   hpd_mm = 2.*interp( r, arr_accum(rcoef,norm=1), frac )(1);
   hpd_arcsec = (hpd_mm / cur_focal_length) * 648000. / pi;

   if( frac == 0.5 ) {
      if( hpd_mm < 0.001 ) {
         s2=swrite(format="HPD : %10.3e mm <> %10.3e arcsec", hpd_mm, hpd_arcsec);
      } else {
         s2=swrite(format="HPD : %.3f mm <> %.2f arcsec", hpd_mm, hpd_arcsec);
      }
   } else {
      iff = long(frac*100 + 0.5);
      if( hpd_mm < 0.001 ) {
         s2=swrite(format="%0i%%D : %10.3e mm <> %10.3e arcsec", iff, hpd_mm, hpd_arcsec);
      } else {
         s2=swrite(format="%0i%%D : %.3f mm <> %.2f arcsec", iff, hpd_mm, hpd_arcsec);
      }
   }
   if( am_subroutine() ) {
      write,format="%s\n",s1;
      write,format="%s\n",s2;
   }
   mt_log_entry,1,"   "+s1,s2;
   mt_log_entry,1,"// HPD = "+ftoa(hpd_mm,ndec=3,sci=1)+"; mm";

  /*
   *   FWHM determination
   */

 /**********    omitted for the time being  ...
   n_annuli = n_photons/200;
   if( n_annuli < 5 ) n_annuli = 5;
   if( n_annuli > 100 ) n_annuli = 100;
   rlim = array(double,n_annuli);
   // let the first (real) radius be the 100 cts radius
   rlim(2) = r(100);
   rlim(2:0) = spanl(rlim(2), 2*hpd_mm, n_annuli-1);

   parr = array(double,n_annuli-1);
   for( i = 2; i <= n_annuli; i++ ) {
      w = where( r >= rlim(i-1) & r < rlim(i) );
      area = pi*(rlim(i)^2 - rlim(i-1)^2);
      nw = numberof(w);
      if( i == 2 ) {
         write,format="FWHM 1 : counts in inner circle : %i\n", nw;
         peak = nw / area;
      }
      parr(i-1) = nw / area;
   }
   parr /= peak;
****************************/

   return hpd_mm;
}

/* Function mt_create_om_hyp2 */

func mt_create_om_hyp2( filename=, master= )
/* DOCUMENT mt_create_om_hyp2, filename=, master=

   Create Optical Module of type hyperboloide 2
   - constant length mirrors
   The mirror lengths and thicknesses are defined
   in the system definition file and set as external variables by
   'mt_setup_system'.

   This is a 'slave'-module so the mirror radii at entrance
   are defined by the 'R2' array of the master module which
   by default is Om_files(1) (set by mt_setup_system). This
   can be overridden by keyword 'master' which may define
   the chosen master optical module.

   The photons move almost along the z-axis in the
   'negative direction' i.e. direction vector ~(0,0,-1).

   The reference surface (origin of z-axis) is at module
   entrance plane.

   2007-10-01/NJW
   2008-04-04/NJW updated e.g. with keyword 'master'
   2011-08-25/NJW updated to version 4.0
*/
{
// ynclude = zab
  extern  Acoef,                   Mirror_thickness_files,  Z_reference, \
          Fcoef,Telescop,Instrume, Num_modules,             Zfocus, \
          Mirror_lengths,          Om_files,                Zfocusarr;
// yxclude =

   if( is_void(filename) ) filename = get_next_filename("om_hyp_???.fits");

   mirror_length = Mirror_lengths(2);

  /*
   *   Consistency check
   */

   if( is_void(Num_modules) ) {
      write,"External Num_modules not set - did you run mt_setup_system ?";
      return [];
   }
   if( numberof(Z_reference) != Num_modules ) {
      write,"Inconsistency between externals Z_reference and Num_modules";
      return [];
   }
   if( is_void(master) ) master = Om_files(1);
   if( ! file_test( master ) ) {
      write,format="Master opt. mod. file: %s was not found\n", master;
      return [];
   }

  //  Read arrays in mirror thickness file

   mthick_radius = rscol(Mirror_thickness_files(2),"radius",nomem=1,silent=1);
   mthick_thick  = rscol(Mirror_thickness_files(2),"mirror_thickness",nomem=1,silent=1);

  /*
   *! Define mirrors from inside outwards
   *
   *
   *  The photons meets first z1 at entrance and then z2 at exit
   *
   *    r1 is the radius of the reflecting surface at z1
   *    r2 is the radius of the reflecting surface at z2
   *    Therefore: z2 < z1 and r2 < r1
   *
   *  Baffle slits. Conditions are:
   *          At z = zb1i :  r > rb1i
   *      and at z = zb1o :  r < rb1o
   *      and at z = zb2i :  r > rb2i
   *      and at z = zb2o :  r < rb2o
   *
   *
   *  The mirror section defined with this code has constant
   *  mirror lengths.
   *  The aperture is to be exploited
   *  as well as possible.
   *
   */

  /*
   *   Assume that the master and slave module have no gap between them
   *   so the r1_arr of the slave is equal to the r2_arr of the master
   */

   //  Read the master data
   r1_arr = rdfitscol( master+"+1", "R2" );

   //  Calculate the mirror thicknesses
   mthick_arr = interp(mthick_thick,mthick_radius,r1_arr);

   // The hyperbolic slave section is number 2

   Fcoef = 0.5*((Zfocusarr(2) - Zfocusarr(1)) + (Z_reference(2) - Z_reference(1)));

   Zfocus = Zfocusarr(2); // mm, z-coordinate of focal point in module coord. syst.
   z1 = 0.0;              // mm, aperture is reference plane
   z2 = z1 - mirror_length;

   // Mirrors are characterized by constant length

   r1_arr = double(r1_arr);

   b2 = (z1 - Zfocus + Fcoef)^2 + Fcoef^2 + r1_arr^2; // array

   acoef_arr = sqrt(0.5*(b2 - sqrt(b2^2 - 4*(z1 - Zfocus + Fcoef)^2 * Fcoef^2)));

   r2_arr = r1_arr;
   nmir = numberof( r1_arr );
   z1_arr = array(z1, nmir);
   z2_arr = array(z2, nmir);
   for( mir = 1; mir <= nmir; mir++ ) {
      Acoef = acoef_arr(mir); // required for the 'rhyp' function
      r2_arr(mir) = rhyp( z2 );
   }
   coating = array(1,nmir);

   kwds_init;
   kwds_set,"FCLENGTH",abs(Zfocus),"[mm] Focal length";
   kwds_set,"ZFOCUS",Zfocus,"[mm] Z coordinate of focus point";
   kwds_set,"R_INNER",r1_arr(0),"[mm] Inner radius";
   kwds_set,"R_OUTER",r1_arr(1),"[mm] Outer radius";
   kwds_set,"MTCKFILE",Mirror_thickness_files(2),"Mirror thickness table file";
   kwds_set,"MIRLNGTH",mirror_length,"[mm] mirror length";
   kwds_set,"DESIGN","Constant mirror length","Telescope design type";
   kwds_set,"MODTYPE","hyperbolic","Type of mirror module";
   kwds_set,"MODSTAT","slave","Mirror positions defined by other module";
   kwds_set,"MASTER", master,"Master optical module file";
   kwds_set,"FCOEF",Fcoef,"[mm] Half distance between foci";
   kwds_set,"DATE",ndate(3),"Time of file creation";
   if( !is_void(Telescop) ) kwds_set,"TELESCOP", Telescop,"Name of telescope or mission";
   if( !is_void(Instrume) ) kwds_set,"INSTRUME", Instrume,"Name of instrument";
   kwds_set,"TUNIT1","mm","Unit for column 1";
   kwds_set,"TUNIT2","mm","Unit for column 2";
   kwds_set,"TUNIT3","mm","Unit for column 3";
   kwds_set,"TUNIT4","mm","Unit for column 4";
   kwds_set,"TUNIT5","mm","Unit for column 5";
   kwds_set,"TUNIT6","mm","Unit for column 6";
   kwds_set,"TUNIT7","mm","Unit for column 7";
    // column 8 is the coating number (dimensionless)
   kwds_set,"TUNIT9","mm","Unit for column 9";
   kwds_set,"TUNIT10","mm","Unit for column 10";
   kwds_set,"TUNIT11","mm","Unit for column 11";
   kwds_set,"TUNIT12","mm","Unit for column 12";
   kwds_set,"TUNIT13","mm","Unit for column 13";
   kwds_set,"TUNIT14","mm","Unit for column 14";
   kwds_set,"TUNIT15","mm","Unit for column 15";
   kwds_set,"TUNIT16","mm","Unit for column 16";

  /*
   *  Create the baffle parameters
   */

  /*
   *  Standard definition of rb1i is backside of previous mirror
   *  at z = z1
   */
   //+ rb1i_arr = r2_arr;
   rb1i_arr = shift(r1_arr,-1) + shift(mthick_arr,-1);
   // Special action for innermost baffle slit that is
   // assumed to have the same opening as its neighbor
   open_space = r1_arr(2) - rb1i_arr(2);
   rb1i_arr(1) = r1_arr(1) - open_space;
   rb1o_arr = r1_arr;
   zb1i_arr = z1_arr;
   zb1o_arr = z1_arr;

   rb2i_arr = shift(r2_arr,-1) + shift(mthick_arr,-1);
   // Special action for innermost baffle slit that is
   // assumed to have the same opening as its neighbor
   open_space = r2_arr(2) - rb2i_arr(2);
   rb2i_arr(1) = r2_arr(1) - open_space;
   rb2o_arr = r2_arr;
   zb2i_arr = z2_arr;
   zb2o_arr = z2_arr;

  /*
   *  create the FITS binary table
   */

   wrmfitscols, filename, "R1",r1_arr,"R2",r2_arr,"Z1",z1_arr,"Z2",z2_arr, \
       "ACOEF",acoef_arr,"MLENGTH",z1_arr-z2_arr,"MTHICK",mthick_arr, "COATING",coating, \
       "RB1I", rb1i_arr, "RB1O", rb1o_arr, "RB2I", rb2i_arr, "RB2O", rb2o_arr, \
       "ZB1I", zb1i_arr, "ZB1O", zb1o_arr, "ZB2I", zb2i_arr, "ZB2O", zb2o_arr, \
        clobber=1;
   write,format="The hyperbolic system optical module file: %s has just been created\n", filename;
   write,format="It has %i mirrors\n", numberof(r1_arr);
   write,format="and it might need a coating update: %s\n","mt_upd_om_coating,...";

}

/* Function mt_create_om_par1 */

func mt_create_om_par1( filename= )
/* DOCUMENT mt_create_om_par1, filename=

   Create Optical Module of type paraboloide 1 with constant mirror
   spacing but variable mirror lengths.

   For this function 'om_parameter' is the spacing from reflecting
   surface to the next reflecting surface i.e. the open space is
   smaller than this by the thickness of the mirror.

   The reference surface (origin of z-axis) is at module
   entrance plane.

   Takes R_outer, R_inner_design, Z_reference, Zfocus from system definition file,
   which is read by 'mt_setup_system'

   Update to version 4.0  2011-08-25/NJW
*/
{
// ynclude = zac
  extern  Dcoef,                   R_outer,                 Zfocus, \
          Mirror_thickness_files,  Z_reference,             Zfocusarr, \
          Om_parameters, Telescop, Instrume;
// yxclude =

   if( is_void(filename) ) filename = get_next_filename("om_par_???.fits");

  /*
   *   Get constants from external variables loaded by 'mt_setup_system'
   *   from system definition file
   */

   spacing = Om_parameters(1);

  //  Read arrays in mirror thickness file

   mthick_radius = rscol(Mirror_thickness_files(1),"radius",nomem=1,silent=1);
   mthick_thick  = rscol(Mirror_thickness_files(1),"mirror_thickness",nomem=1,silent=1);

  /*
   *  Define mirrors from outside inwards, but they are numbered
   *  from the inside to the outside i.e. increasing with radius.
   *
   *  The photons meets first z1 at entrance and then z2 at exit.
   *
   *    r1 is the radius of the reflecting surface at z1
   *    r2 is the radius of the reflecting surface at z2
   *    Therefore: z2 < z1 and r2 < r1
   *
   *  Baffle slits. Conditions are:
   *          At z = zb1i :  r > rb1i
   *      and at z = zb1o :  r < rb1o
   *      and at z = zb2i :  r > rb2i
   *      and at z = zb2o :  r < rb2o
   *
   *
   *  The mirror section defined with this code has longer and
   *  longer mirrors for decreasing radius. There is a constant
   *  spacing between the mirrors and the aperture is to be exploited
   *  as well as possible.
   *
   *   This is the proposal for the Gamma Ray Imager mission
   *   as of 2007-08-16
   *
   *   2007-09-21/NJW
   */

   Zfocus = Zfocusarr(1); // mm, z-coordinate of focal point in module coord. syst.
   z1 = 0.0; // mm, aperture is reference plane
   r1 = R_outer;

   // Search r2 and hence the thickness of mirror 2 in two iterations

   r2_guess = r1;

   for( i = 1; i <= 2; i++ ) {
      mirror_thickness = interp(mthick_thick,mthick_radius,r2_guess);
      open_space = spacing - mirror_thickness;

      // find where in z the radius, r2, is r1 - open_space

      z = z1;
      Dcoef = sqrt( (z1-Zfocus)^2 + r1^2 ) - (z1-Zfocus);
      r = rpar(z);
      while( r > r1 - open_space ) {
         z -= 2.;
         r = rpar(z);
      }
      while( r < r1 - open_space ) {
         z += 0.05;
         r = rpar(z);
      }
      r2_guess = r;
   }


   r2 = r;
   z2 = z;
   r1_arr = [];
   r2_arr = [];
   z1_arr = [];
   z2_arr = [];
   mthick_arr = [];
   d_arr = [];
   coating = [];
   grow, r1_arr, r1;
   grow, r2_arr, r2;
   grow, z1_arr, z1;
   grow, z2_arr, z2;
   mirror_thickness = interp(mthick_thick,mthick_radius,r1);
   grow, mthick_arr, mirror_thickness;
   grow, d_arr, Dcoef;
   grow, coating, 1;
   n_mirrors = 1;

   kwds_init;
   kwds_set,"ZREFRNCE",Z_reference(1),"[mm] Z of reference plane in tel. syst.";
   kwds_set,"ZFOCUS",Zfocus,"[mm] Z coord. of focus point in opt.mod. syst.";
   kwds_set,"R_OUTER",R_outer,"[mm] Outer radius";
   kwds_set,"R_INNER",R_inner_design,"[mm] Inner design radius";
   kwds_set,"MTCKFILE",Mirror_thickness_files(1),"Mirror thickness table file";
   kwds_set,"SPACING",spacing,"[mm] Distance refl. surface to refl. surface";
   kwds_set,"DESIGN","I - constant spacing","Telescope design type";
   kwds_set,"MODTYPE","parabolic","Type of mirror module";
   kwds_set,"DATE",ndate(3),"Time of file creation";
   if( !is_void(Telescop) ) kwds_set,"TELESCOP", Telescop,"Name of telescope or mission";
   if( !is_void(Instrume) ) kwds_set,"INSTRUME", Instrume,"Name of instrument";
   kwds_set,"TUNIT1","mm","Unit for column 1";
   kwds_set,"TUNIT2","mm","Unit for column 2";
   kwds_set,"TUNIT3","mm","Unit for column 3";
   kwds_set,"TUNIT4","mm","Unit for column 4";
   kwds_set,"TUNIT5","mm","Unit for column 5";
   kwds_set,"TUNIT6","mm","Unit for column 6";
   kwds_set,"TUNIT7","mm","Unit for column 7";
    // column 8 is the coating number (dimensionless)
   kwds_set,"TUNIT9","mm","Unit for column 9";
   kwds_set,"TUNIT10","mm","Unit for column 10";
   kwds_set,"TUNIT11","mm","Unit for column 11";
   kwds_set,"TUNIT12","mm","Unit for column 12";
   kwds_set,"TUNIT13","mm","Unit for column 13";
   kwds_set,"TUNIT14","mm","Unit for column 14";
   kwds_set,"TUNIT15","mm","Unit for column 15";
   kwds_set,"TUNIT16","mm","Unit for column 16";


   while( r1 > R_inner_design ) {

      n_mirrors++;

      z = z1;
      r1 = r2 - mirror_thickness; // mirror_thickness from prevous determination
      // Dcoef: parabolic constant; required for function 'rpar'
      Dcoef = sqrt( (z1-Zfocus)^2 + r1^2 ) - (z1-Zfocus);

     /*
      *   Start a search process in mirror length where the current
      *   mirror inner radius
      *   will intersect the next mirror's outer radius
      */

      // Search r2 and hence the thickness of mirror 2 in two iterations

      r2_guess = r1;

      for( i = 1; i <= 2; i++ ) {
         mirror_thickness = interp(mthick_thick,mthick_radius,r2_guess);
         open_space = spacing - mirror_thickness;

         r = rpar(z);
         while( r > r1 - open_space ) {
            z -= 2.;
            r = rpar(z);
         }
         while( r < r1 - open_space ) {
            z += 0.05;
            r = rpar(z);
         }
         r2_guess = r;
      }
      r2 = r;
      z2 = z;

      // when found, add to arrays

      grow, r1_arr, r1;
      grow, r2_arr, r2;
      grow, z1_arr, z1;
      grow, z2_arr, z2;
      mirror_thickness = interp(mthick_thick,mthick_radius,r1);
      grow, mthick_arr, mirror_thickness;
      grow, d_arr, Dcoef;
      grow, coating, 1;
   }

  /*
   *  Create the baffle parameters
   *    (entrance and exit slits)
   */

   //+ rb1i_arr = r1_arr - open_space;
   rb1i_arr = shift(r1_arr,1) + shift(mthick_arr,1);
   // Special action for innermost baffle slit that is
   // assumed to have the same opening as its neighbor
   open_space = r1_arr(-1) - rb1i_arr(-1);
   rb1i_arr(0) = r1_arr(0) - open_space;
   rb1o_arr = r1_arr;
   zb1i_arr = z1_arr;
   zb1o_arr = z1_arr;

   rb2i_arr = shift(r2_arr,1) + shift(mthick_arr,1);
   // Special action for innermost baffle slit that is
   // assumed to have the same opening as its neighbor
   open_space = r2_arr(-1) - rb2i_arr(-1);
   rb2i_arr(0) = r2_arr(0) - open_space;
   rb2o_arr = r2_arr;
   zb2i_arr = shift(z2_arr,1);
   zb2i_arr(0) = zb2i_arr(-1);
   zb2o_arr = z2_arr;

  /*
   *  create the FITS binary table while reversing the order of all arrays
   */

   n = numberof(r1_arr);
   i = indgen(n:1:-1);
   wrmfitscols, filename, "R1",r1_arr(i),"R2",r2_arr(i),"Z1",z1_arr(i),"Z2",z2_arr(i), \
       "DCOEF",d_arr(i),"MLENGTH",z1_arr(i)-z2_arr(i),"MTHICK",mthick_arr(i), "COATING",coating(i), \
       "RB1I", rb1i_arr(i), "RB1O", rb1o_arr(i), "RB2I", rb2i_arr(i), "RB2O", rb2o_arr(i), \
       "ZB1I", zb1i_arr(i), "ZB1O", zb1o_arr(i), "ZB2I", zb2i_arr(i), "ZB2O", zb2o_arr(i), \
        clobber=1;
   write,format="The parabolic system optical module file: %s has just been created\n", filename;
   write,format="It has %i mirrors\n", n_mirrors;
   write,format="and it might need a coating update: %s\n","mt_upd_om_coating,...";

}

/* Function mt_create_om_par2 */

func mt_create_om_par2( filename= )
/* DOCUMENT mt_create_om_par2, filename=

   Create Optical Module of type paraboloide 2
   - constant length mirrors
   - leak for on-axis radiation determined by om_parameter

   The photons move almost along the z-axis in the
   'negative direction' i.e. direction vector ~(0,0,-1).

   The reference surface (origin of z-axis) is at module
   entrance plane.

   Updated to version 4.0  2011-08-25/NJW
*/
{

// ynclude = zad
  extern  Dcoef,Telescop, Instrume,Om_parameters,           Z_reference, \
          Mirror_lengths,          R_inner_design,          Zfocus, \
          Mirror_thickness_files,  R_outer,                 Zfocusarr;
// yxclude =

   if( is_void(filename) ) filename = get_next_filename("om_par_???.fits");

   mirror_length = Mirror_lengths(1);

  /*
   * Define mirrors from outside inwards
   *
   *  The photons meets first z1 at entrance and then z2 at exit
   *
   *    r1 is the radius of the reflecting surface at z1
   *    r2 is the radius of the reflecting surface at z2
   *    Therefore: z2 < z1 and r2 < r1
   *
   *  Baffle slits. Conditions are:
   *          At z = zb1i :  r > rb1i
   *      and at z = zb1o :  r < rb1o
   *      and at z = zb2i :  r > rb2i
   *      and at z = zb2o :  r < rb2o
   *
   *
   *  The mirror section defined with this code has constant
   *  mirror lengths.
   *  The aperture is to be exploited
   *  as well as possible.
   *
   */

   Zfocus = Zfocusarr(1); // mm, z-coordinate of focal point in module coord. syst.
   z1 = 0.0; // mm, aperture is reference plane
   r1 = R_outer;

  //  Read arrays in mirror thickness file

   mthick_radius = rscol(Mirror_thickness_files(1),"radius",nomem=1,silent=1);
   mthick_thick  = rscol(Mirror_thickness_files(1),"mirror_thickness",nomem=1,silent=1);

   // Mirrors are characterized by constant length

   Dcoef = sqrt( (z1-Zfocus)^2 + r1^2 ) - (z1-Zfocus);
   z2 = z1 - mirror_length;
   r2 = rpar(z2);
   r1_arr = [];
   r2_arr = [];
   z1_arr = [];
   z2_arr = [];
   mthick_arr = [];
   d_arr = [];
   coating = [];
   grow, r1_arr, r1;
   grow, r2_arr, r2;
   grow, z1_arr, z1;
   grow, z2_arr, z2;
   mirror_thickness = interp(mthick_thick,mthick_radius,r1);
   grow, mthick_arr, mirror_thickness;
   grow, d_arr, Dcoef;
   grow, coating, 1;
   n_mirrors = 1;

   kwds_init;
   kwds_set,"ZREFRNCE",Z_reference(1),"[mm] Z coord. of refr. plane in tel.syst.";
   kwds_set,"ZFOCUS",Zfocus,"[mm] Z coordinate of focus point in opt.mod. syst.";
   kwds_set,"R_INNER",R_inner_design,"[mm] Inner design radius";
   kwds_set,"R_OUTER",R_outer,"[mm] Outer radius";
   kwds_set,"MTCKFILE",Mirror_thickness_files(1),"Mirror thickness table file";
   kwds_set,"MIRLNGTH",mirror_length,"[mm] mirror length";
   kwds_set,"DESIGN","Constant mirror length","Telescope design type";
   kwds_set,"MODTYPE","parabolic","Type of mirror module";
   kwds_set,"MODSTAT","Master","This defines the mirror positions";
   kwds_set,"PACKFACT", Om_parameters(1), "Loose packing factor, 1.0 means dense";
   kwds_set,"DATE",ndate(3),"Time of file creation";
   if( !is_void(Telescop) ) kwds_set,"TELESCOP", Telescop,"Name of telescope or mission";
   if( !is_void(Instrume) ) kwds_set,"INSTRUME", Instrume,"Name of instrument";
   kwds_set,"TUNIT1","mm","Unit for column 1";
   kwds_set,"TUNIT2","mm","Unit for column 2";
   kwds_set,"TUNIT3","mm","Unit for column 3";
   kwds_set,"TUNIT4","mm","Unit for column 4";
   kwds_set,"TUNIT5","mm","Unit for column 5";
   kwds_set,"TUNIT6","mm","Unit for column 6";
   kwds_set,"TUNIT7","mm","Unit for column 7";
    // column 8 is the coating number (dimensionless)
   kwds_set,"TUNIT9","mm","Unit for column 9";
   kwds_set,"TUNIT10","mm","Unit for column 10";
   kwds_set,"TUNIT11","mm","Unit for column 11";
   kwds_set,"TUNIT12","mm","Unit for column 12";
   kwds_set,"TUNIT13","mm","Unit for column 13";
   kwds_set,"TUNIT14","mm","Unit for column 14";
   kwds_set,"TUNIT15","mm","Unit for column 15";
   kwds_set,"TUNIT16","mm","Unit for column 16";


   while( r1 > R_inner_design ) {

      n_mirrors++;

      //+ r1 = r2 - mirror_thickness;
      // "loose" packing with factor Om_parameters(1) [NuStar value is 1.33]
      r1 = r1 - Om_parameters(1)*(r1 - r2) - mirror_thickness;
      // Dcoef: parabolic constant; required for function 'rpar'
      Dcoef = sqrt( (z1-Zfocus)^2 + r1^2 ) - (z1-Zfocus);

      r2 = rpar(z2);

      // add to arrays

      grow, r1_arr, r1;
      grow, r2_arr, r2;
      grow, z1_arr, z1;
      grow, z2_arr, z2;
      mirror_thickness = interp(mthick_thick,mthick_radius,r1);
      grow, mthick_arr, mirror_thickness;
      grow, d_arr, Dcoef;
      grow, coating, 1;
   }

  /*
   *  Create the baffle parameters
   */

   //+ rb1i_arr = r2_arr;
   rb1i_arr = shift(r1_arr,1) + shift(mthick_arr,1);
   // Special action for innermost baffle slit that is
   // assumed to have the same opening as its neighbor
   open_space = r1_arr(-1) - rb1i_arr(-1);
   rb1i_arr(0) = r1_arr(0) - open_space;
   rb1o_arr = r1_arr;
   zb1i_arr = z1_arr;
   zb1o_arr = z1_arr;

   rb2i_arr = shift(r2_arr,1) + shift(mthick_arr,1);
   // Special action for innermost baffle slit that is
   // assumed to have the same opening as its neighbor
   open_space = r2_arr(-1) - rb2i_arr(-1);
   rb2i_arr(0) = r2_arr(0) - open_space;
   rb2o_arr = r2_arr;
   zb2i_arr = z2_arr;
   zb2o_arr = z2_arr;

  /*
   *  create the FITS binary table while reversing the order of the arrays
   */

   n = numberof(r1_arr);
   i = indgen(n:1:-1);
   wrmfitscols, filename, "R1",r1_arr(i),"R2",r2_arr(i),"Z1",z1_arr(i),"Z2",z2_arr(i), \
       "DCOEF",d_arr(i),"MLENGTH",z1_arr(i)-z2_arr(i),"MTHICK",mthick_arr(i), "COATING",coating(i), \
       "RB1I", rb1i_arr(i), "RB1O", rb1o_arr(i), "RB2I", rb2i_arr(i), "RB2O", rb2o_arr(i), \
       "ZB1I", zb1i_arr(i), "ZB1O", zb1o_arr(i), "ZB2I", zb2i_arr(i), "ZB2O", zb2o_arr(i), \
        clobber=1;
   write,format="The parabolic system optical module file: %s has just been created\n", filename;
   write,format="It has %i mirrors\n", n_mirrors;
   write,format="and it might need a coating update: %s\n","mt_upd_om_coating,...";

}

/* Function mt_pre_def_photons */

func mt_pre_def_photons( fraper, energy_or_file, src_offaxis, src_azimuth, dphot=, \
                         renorm=, labxoff=, labyoff=, eqillum=, blur= )
/* DOCUMENT mt_pre_def_photons, fraper, energy_or_file, src_offaxis, src_azimuth, dphot=, \
                         renorm=, labxoff=, labyoff=, eqillum=, blur=

   Calls mt_def_photons according to the input 'energy_or_file'.

   fraper     array [inner radius, outer radius, lower azimuth, upper azimuth]
                        [mm]         [mm]          [deg]           [deg]
   energy_or_file     [keV] photon energy (monochromatic source)
           OR
              the name of a source flux file (photflux as from mk_photflux(.i))
           OR
              the name of a sky definition file (EXTNAME == SKY_DEFINITION)
           OR
              the name of a labsource definition file (EXTNAME == LABSOURCE_DEFINITION)

   src_offaxis   [arcmin] source off axis angle
   src_azimuth   [degrees] source azimuth

   Keywords:

   dphot     is the photon density (default = 1.0) applies if energy_or_file is energy
   renorm    a factor multiplied on the source flux
   labxoff   Offset [mm] of lab source in X
   labyoff   Offset [mm] of lab source in Y
   blur      [rad] Blur angle for input source
  (eqillum   set if equal mirror illumination is requested - unphysical and only for
             special investigations)

   2010-05-27/NJW
   2011-01-20/NJW Version 3.1

*/
{

// ynclude = zae
  extern  Dec_scx,   Exposure,  Phs,       Posang,    Ra_scx, \
          Focal_length, Z_reference, Z1arr;
// yxclude =

   local filepff, extno;


   if( is_void(dphot) ) dphot = 1.0;
   if(typeof(energy_or_file) != "string" ) energy = double(energy_or_file);
   if( is_void(renorm) ) renorm = 1.0;

   mode = 0; // energy, offaxis, azimuth has been input
     //   1     a single photon_flux file, offaxis, azimuth has been input
     //   2     a sky definition file has been input


  /*
   *
   *     Define the X-ray source
   *
   */

   // First the position

   // src_offaxis in arcmin,  src_azimuth in degrees
   if( !is_void(src_offaxis) ) offaxis_rad = double(src_offaxis) * (pi/180.) / 60.;
   if( !is_void(src_azimuth) ) azimuth_rad = double(src_azimuth) * (pi/180.);

   // Then see if a filename has been given in 'energy_or_file'

   if( typeof(energy_or_file) == "string" ) { // Yes, a filename is given

      filename = energy_or_file;
      if( !file_test(filename) ) error,"MT_PRE_DEF_PHOTONS, an input file is missing!";
      dol = filename+"+1";
      hdr = headfits(dol);

      extname = fxpar(hdr,"extname");

      if( extname == "PHOTON_FLUX" || extname == "DXB_PHOTON_FLUX" ) {

         // single source with spectral distribution
         // only call 'mt_def_photons' and exit

         R = [cos(azimuth_rad)*sin(offaxis_rad), \
              sin(azimuth_rad)*sin(offaxis_rad), \
              -cos(offaxis_rad)];
         mt_def_photons, fraper, filename, R, \
                  dphot=dphot, renorm=renorm, flag=1, eqillum=eqillum, blur=blur;

      } else if( extname == "SKY_DEFINITION" ) {

         x_sky = rdfitscol(dol,"x_sky"); // Unit is arcmin
         y_sky = rdfitscol(dol,"y_sky"); // Unit is arcmin
         ener_arr = rdfitscol(dol,"energy"); // Unit is keV
         dphot_arr = rdfitscol(dol,"dphot"); // Unit is /mm2
         dol_arr = rdfitscol(dol,"dol");
         renorm_arr = rdfitscol(dol,"renorm");
         if( is_void(Exposure) ) Exposure = fxpar(hdr,"exposure"); // Unit is s
         Ra_scx = fxpar(hdr,"ra_scx"); // [deg] R.A. of boresight
         Dec_scx = fxpar(hdr,"dec_scx"); // [deg] Dec. of boresight
         Posang = fxpar(hdr,"posang"); // [deg] Position angle of satellite

         nsrc = numberof(x_sky);
         for( i = 1; i <= nsrc; i++ ) {
            src_offaxis = sqrt(x_sky(i)^2 + y_sky(i)^2);
            src_azimuth = anyof([x_sky(i),y_sky(i)]) ? atan(y_sky(i),x_sky(i))*180/pi : 0.0;
            // src_offaxis in arcmin,  src_azimuth in degrees
            offaxis_rad = double(src_offaxis) * (pi/180.) / 60.;
            azimuth_rad = double(src_azimuth) * (pi/180.);
            R = [cos(azimuth_rad)*sin(offaxis_rad), \
                 sin(azimuth_rad)*sin(offaxis_rad), \
                 -cos(offaxis_rad)];
            if( strtrim(dol_arr(i)) == "none" || strlen(strtrim(dol_arr(i))) == 0 ) {
               energy = ener_arr(i);
               dphot = dphot_arr(i);
            } else { // We have a photon flux file (pff), check existence and header
               dolpff = dol_arr(i);
               get_exten_no, dolpff, filepff, extno;
               //  Changes with version 4.4.5
               //+ if( !file_test(filepff) ) error,"Not found: "+filepff;
               //+ fh = headfits( dolpff );
               //+ extnm = fxpar(fh,"extname");
               //+ if( extnm != "PHOTON_FLUX" && extnm != "DXB_PHOTON_FLUX" ) \
               //+            error,filepff+" does not contain photon flux";
               energy = filepff;
            }
            mt_def_photons, fraper, energy, R, \
                  dphot=dphot, renorm=renorm_arr(i), flag=i, cont=i-1, eqillum=eqillum, blur=blur;
         }

      } else if( extname == "LABSOURCE_DEFINITION" ) {

         // find the source position (x,y,z), a single z value but there may be
         // several (many) x and y values (to desribe a non-point source)
         zposit = fxpar( hdr, "ZPOSIT" ); // mm
         x = rdfitscol( dol, "X" );  // mm
         y = rdfitscol( dol, "Y" );  // mm
         if( !is_void(labxoff) ) x += labxoff;
         if( !is_void(labyoff) ) y += labyoff;
         // ZPOSIT is defined relative to optic center hence a shift in z
         // must be included
         dz_from_optic_center = Z_reference(1) - Focal_length + Z1arr(1);
         z = zposit - dz_from_optic_center;

         ener_arr = rdfitscol( dol, "energy"); // Unit is keV
         strength_arr = rdfitscol( dol, "strength");
         if( !is_void(renorm) ) strength_arr *= renorm;
         dol_arr = rdfitscol( dol, "dol");
         nsrc = numberof(x);
         // loop over all the sub-sources
         for( i = 1; i <= nsrc; i++ ) {
            labsource_pos = [x(i), y(i), z];
            if( strtrim(dol_arr(i)) == "none" || strlen(strtrim(dol_arr(i))) == 0 ) {
               energy = ener_arr(i);
               strength = strength_arr(i);
            } else { // We have a photon flux file, check existence and header
               dolpff = dol_arr(i);
               get_exten_no, dolpff, filepff, extno;
               if( !file_test(filepff) ) error,"Not found: "+filepff;
               fh = headfits( dolpff );
               extnm = fxpar(fh,"extname");
               if( extnm != "PHOTON_FLUX" && extnm != "DXB_PHOTON_FLUX" ) \
                          error,filepff+" does not contain photon flux";
               energy = filepff;
            }
            // call with lab source position relative to outmost mirror edge
            mt_def_photons, fraper, energy, labsource_pos, \
                  dphot=strength, flag=i, cont=i-1, lab=1, eqillum=eqillum, blur=blur;
         }

      } else {
         error,"Unsupported EXTNAME: "+extname;
      }

   } else {

      // no filename, just real energy so
      // only call 'mt_def_photons' and exit

      R = [cos(azimuth_rad)*sin(offaxis_rad), \
           sin(azimuth_rad)*sin(offaxis_rad), \
           -cos(offaxis_rad)];
      mt_def_photons, fraper, energy, R, \
                  dphot=dphot, renorm=renorm, flag=1, eqillum=eqillum, blur=blur;

   }


   local loce; eq_nocopy, loce, Phs.E;
   r = sqrt(loce(1,)^2 + loce(2,)^2);
   isr = sort(r); // sort to increasing radii to minimize
                  // the reading of scattering files
   Phs = Phs(isr);
}


/* Function mt_def_photons */

func mt_def_photons( fraper, energy_or_file, R_or_lab, \
                     dphot=, renorm=, flag=, cont=, lab=, eqillum=, blur= )
/* DOCUMENT mt_def_photons, fraper, energy_or_file, R_or_lab, \
                     dphot=, renorm=, flag=, cont=, lab=, eqillum=, blur=

   Defines photons in external array of struct 's_Ray' in Phs, arriving in
   a single specified direction.
   They will get a Z value = 0 (E(3) = 0) and be sorted with increasing
   radius. The energy distribution is evaluated here.

   fraper     Array of [inner radius, outer radius, lower azimuth, upper azimuth]
                          [mm]            [mm]         [deg]          [deg]
   energy_or_file     [keV] photon energy (monochromatic source)
                       OR
                      the name of the source flux file (photflux as from mk_photflux(.i))
   R_or_lab  directional vector of this bunch of photons
           OR
             position of laboratory source

   dphot=  is the photon density (default = 1.0) applies if energy_or_file is energy
   renorm= a factor multiplied on the source flux
   cont=   if set it will append to existing Phs
   flag=   for debugging
   lab=    set if 'R_or_lab' is lab source position
  (eqillum= set if equal mirror illumination is requested - unphysical and only for
            special investigations)
   blur=    [rad] angle to blur the source position

   SEE ALSO: mt_pre_def_photons  that deals with various photon directions

   2007-10-01/NJW
   Version 3.0   2011-01-05/NJW

*/
{

   extern Exposure, Phs;

   local energ_lo, energ_hi, photflux;

   if( is_void(dphot) ) dphot = 1.0;
   if(typeof(energy_or_file)!="string") energy = double(energy_or_file);
   if( is_void(renorm) ) renorm = 1.0;
   if( is_void(flag) ) flag = 1;


  /*
   *
   *     Define the X-ray source
   *
   */

   // First the position

   mt_log_entry,3,swrite(format="MT_DEF_PHOTONS flag = %i, R_or_lab(1,2,3) = %.6f %.6f %.6f",\
                         flag, R_or_lab(1), R_or_lab(2), R_or_lab(3));

   // Then see if there is a spectrum specification file

   if( typeof(energy_or_file) == "string" ) { // Yes, a spectrum should be used

      src_spec_file = energy_or_file;
      //  Changes with version 4.4.5
      if( file_test(src_spec_file) ) { // a disk file is found, use it

         dol = src_spec_file+"+1";
         hdr = headfits(dol);
         energ_lo = rdfitscol( dol, "energ_lo" );
         energ_hi = rdfitscol( dol, "energ_hi" );
         photflux = rdfitscol( dol, "photflux" );
         mt_log_entry,3,"MDP DOL of photflux: "+dol;

      } else {   // assume that the information is stored in memory

         // three new lines from version 4.4.5

         mem_restore,"ENERG_LO",energ_lo;
         mem_restore,"ENERG_HI",energ_hi;
         mem_restore, src_spec_file, photflux;
         mt_log_entry,3,"MDP photflux mem_restore";

      }
      if( is_void(energ_lo) ) error,"##17## energ_lo is undefined";
      if( is_void(energ_hi) ) error,"##18## energ_hi is undefined";
      if( is_void(photflux) ) error,"##19## photflux is undefined";

      if( is_void(Exposure) ) {
         write,"Notice! Exposure set to 100 s";
         Exposure = 100.;
      }
      // evaluate the photon density (include conversion to /mm2)
      dphot = 1.e-2 * Exposure * renorm * sum( photflux * (energ_hi-energ_lo));
      if(typeof(dol)=="string") mt_log_entry,3,"MDP DOL of photflux: "+dol;
      else mt_log_entry,3,"MDP photflux from memory";

   } else mt_log_entry,3,swrite(format="MDP energy = %.3f keV",energy);

   kount = 0;



  /*
   *  Set the number of photons for the telescope in such a way
   *  that the average is kept correct, i.e. no systematic
   *  truncation
   */

   ang_span = (fraper(4) - fraper(3))*pi/180;
   fnphot = 0.5 * dphot * ang_span * (fraper(2)^2 - fraper(1)^2);
   ipart = long(floor(fnphot));
   fpart = fnphot - floor(fnphot);

   nphot = random() < fpart ? ipart+1 : ipart;
   n_injected = nphot;
   fn_expected = fnphot;

   mt_log_entry,3,swrite(format="MDP nphot = %i",nphot);

   if( !cont ) Phs = [];

   if( nphot ) {

      phs = array(s_Ray, nphot);

      // Define energies

      if( typeof(energy_or_file) == "string" ) {
        /*
         *  Lines from before 2010-08-04. The multiplication
         *  by (energ_hi-energ_lo) should not have been done
         *
         * phs.energy = draw_from_dist( sqrt(energ_lo*energ_hi), \
         *       photflux*(energ_hi-energ_lo), nphot );
         *
         *      Corrected from version 2.4
         */
         phs.energy = draw_from_dist( sqrt(energ_lo*energ_hi), \
               photflux, nphot );
      } else {
         phs.energy = energy;
      }

  /*
   *  Assign arrival positions at entrance aperture
   */

      phi = fraper(3) * pi/180 + random(nphot) * ang_span;
      rd22 = fraper(2)^2;
      rd12 = fraper(1)^2;
      // make uniform distribution on the aperture by
      // inverse of normalized integrated distribution of radii
      r    = sqrt(rd12 + random(nphot)*(rd22-rd12));

      if( eqillum ) { // change distribution
         // make uniform illumination of mirrors by choosing
         // a 1/r distribution of radii obtained by
         // inverse of normalized integrated distribution of radii

         write,"Using 1/r distribution ...";
         r = fraper(1) * (fraper(2)/fraper(1))^random(nphot);
      }

      //+ (version of before 2011-08-11) phs.E = transpose([r*cos(phi),r*sin(phi),array(0.0,nphot)]);
      phs.E = transpose([r*cos(phi),r*sin(phi),array(Z1arr(1),nphot)]);

      if( lab ) { // calculate individual directions
         local cE; eq_nocopy, cE, phs.E;
         for( i = 1; i <= nphot; i++ ) {
            R = cE(,i) - R_or_lab;
            R /= sqrt(sum(R^2)); // normalize
            phs(i).R = R;
            phs(i).D1 = R; // for display of track through telescope
         }
      } else {
     /*
      *  Define all photons with same direction (R) and positions (E)
      *  on aperture with z = 0
      */

         if( is_void(blur) ) phs.R = R_or_lab(,-:1:nphot);
         else phs.R = mt_blur( R_or_lab, blur, nphot );
         phs.D1 = phs.R; // for display of track through telescope
      }
      phs.rcoef = 1.0
      phs.status = 0;

      // assign the flag value

      phs.flag = flag;

      grow, Phs, phs;
   }

}




/* Function mt_eff_area_photons */

func mt_eff_area_photons( a, earr=, samp=, outfile=, silent= )
/* DOCUMENT eff_area = mt_eff_area_photons( earr=, samp=, outfile=, silent= )

   Calculate the telescope effective area from the
   photons currently in memory in struct array 'Phs'.
   Go through all mirrors and sum the mirror contributions.
   The photons must be status==0 and bounce==2*Num_modules-1
   photons.

   Default energy array is 'E_uniq' (external variable)

   If keyword 'samp' is given then only 1 out of 'samp'
   photons are used. The process can take quite a long
   time.

   If the keyword 'outfile' is not set or set to zero then
   no output file will be written. If it is given as a string
   then this will be the output file name, else the standard
   naming as eff_area_nnnn.fits, where 'nnnn' is a serial number,
   will be used. In both cases the keyword 'clobber' is set.

   Version 1.1     2007-11-06/NJW
   Version 3.3     2011-03-10/NJW
   update to version 4.0     2011-08-25/NJW
*/

{

// ynclude = zaf
  extern  Coat_list,       Mirror_coating,  R_inner,         Src_azimuth, \
          E_uniq,          Num_modules,     R_outer,         Src_offaxis, \
          Energy,          Phs,             Scatter_file,    Version, \
          Fraper_area, Telescope, Instrume;
// yxclude =

   time_keeper = elapsed_time = array(double,3);
   timer, time_keeper;

   if( is_void(earr) ) {
      if( is_void(E_uniq) ) {
         n_earr = 50;
         earr = span(2.,79, n_earr);
      } else {
         n_earr = numberof(E_uniq);
         earr = E_uniq;
      }
   } else {
      n_earr = numberof(earr);
   }

   if( is_void(samp) ) samp = 1;
   if( samp < 1 ) samp = 1;

   n_inject = numberof(Phs);
   good_bounce = 2*Num_modules - 1;
   w = where(Phs.status == 0 & Phs.bounce == good_bounce);
   n_photons = numberof(w);
   loss_fraction = double(n_inject - n_photons) / n_inject;

   if( is_set(outfile) ) {

      if( structof(outfile) != string ) \
         outfile = get_next_filename("eff_area_????.fits");

      kwds_init;
      kwds_set,"DATE",ndate(3),"Time of file creation";
      kwds_set,"ORIGIN","mt_eff_area_photons of MT_RAYOR-"+Version,"Software used";
      kwds_set,"RESPONSI","Niels J. Westergaard","Responsible programmer";
      if( !is_void(Telescop) ) kwds_set,"TELESCOP", Telescop,"Name of telescope or mission";
      if( !is_void(Instrume) ) kwds_set,"INSTRUME", Instrume,"Name of instrument";
      kwds_set,"SITE","NSI/DTU","Institution";
      if( typeof(Energy) == "double" ) {
         kwds_set,"ENERGYIN", Energy, "[keV] Energy of injected photons";
      } else {
         kwds_set,"COMMENT","A photon flux file was used as input";
      }
      if(!is_void(Src_offaxis))kwds_set,"SRC_OFAX",Src_offaxis,"[arcmin] Source off axis angle";
      if(!is_void(Src_azimuth))kwds_set,"SRC_AZIM",Src_azimuth,"[degrees] Source azimuth angle";
      kwds_set,"TUNIT1","keV","Unit of first column: ENERGY";
      kwds_set,"TUNIT2","mm2","Unit of second column: EFF_AREA";

      kwds_set,"NPHOT_IN", n_inject,"Number of injected photons";
      kwds_set,"NPHOTONS", n_photons,"Number of accepted photons";
      kwds_set,"SAMPLING", samp, "One out of this number of photons is used";
      kwds_set,"AP_AREA", Fraper_area,"[mm2] Aperture area";
   }


   //  Sum the reflection coefficients
   //    not so useful when no scattering/refl tables have been read in

   if( !is_not_defined(Coat_list) && n_photons ) {
      rcoef = array(0.0, n_earr);
      for( iev = 1; iev <= n_photons; iev += samp ) {
         coating = Mirror_coating(Phs(w(iev)).mirror);
         q = where( Coat_list.id == coating );
         if( numberof(q) == 0 ) error,"##20## coating problem";
         required_scatfile = Coat_list(q(1)).file;
         if( Scatter_file != required_scatfile ) {
            mt_load,scatfile=required_scatfile;
         }
         angle1 = Phs(w(iev)).angle_in1;
         angle2 = Phs(w(iev)).angle_in2;

         for( iener = 1; iener <= n_earr; iener++ ) {
            ener = earr(iener);
            r = mt_get_rcoef( ener, angle1 );
            if( Num_modules > 1 ) {
               r *= mt_get_rcoef( ener, angle2 );
            }
            rcoef(iener) += r;
         }
      }
   } else rcoef = array( 1.0, n_earr );
   eff_area = n_photons ? (samp * rcoef / n_photons) * Fraper_area * (1 - loss_fraction) : array(0.0, n_earr);

   if( is_set(outfile) ) {
      wrmfitscols,outfile,"ENERGY",earr,"EFF_AREA",eff_area,clobber=1;
      write,format="%s has been written\n", outfile;
      write,"with columns ENERGY and EFF_AREA";
   }

   timer, time_keeper, elapsed_time;
   mt_log_entry,1,swrite(format="CPU time for mt_eff_area_photons was: %.3f s", elapsed_time(1));
   if(!silent)write,format="CPU time for mt_eff_area_photons was: %.3f s\n", elapsed_time(1);

   return eff_area;
}

/* Function mt_fake_scatter_data */

func mt_fake_scatter_data( filename, type=, fwhm=, angle_max=, ener_min=, ener_max=, \
             n_ener=, n_angles=, dist_angle_max=, n_dist_angles=, coat= )
/* DOCUMENT mt_fake_scatter_data, filename, type=, fwhm=, angle_max=, ener_min=, ener_max=, \
             n_ener=, n_angles=, dist_angle_max=, n_dist_angles=, coat=

   Produces a FITS scatter table as required by mt_run
   All angles (fwhm, angle_max, dist_angle_max) must be given in RADIANS

   Defaults:    type        1
                fwhm     9.7e-5 rad (20 arcsec)
              angle_max  1.5e-2 rad (51 arcmin)  Maximal grazing angle covered
              ener_min      1   keV              Minimum of energy array
              ener_max     80   keV              Maximum of energy array
              n_ener       30                    Number of energy array elements
              n_angles     20                    Number of angle array elements
                coat        1                    Coating number
           dist_angle_max 5.818e-4 rad (2 arcmin) Limit of scattering distribution
           n_dist_angles  100                    Number of angles for distribution

   2007-10-05/NJW
   2012-03-01/NJW Version 4.2 updated with n_dist_angles
   2013-07-09/NJW Version 4.5.3 updated with the type2 option
*/
{

   extern Version;

   if( is_void(type) ) type = 1; // Scatter file type
   if( is_void(n_ener) ) n_ener = 30; // number of energy values
   if( is_void(n_angles) ) n_angles = 20; // number of angle values for scattering tables
   if( is_void(fwhm) ) fwhm = 9.7e-5; // radians = 20.0 arcsec
   if( is_void(angle_max) ) angle_max = 1.5e-2; // radians = 51 arcmin
   if( is_void(ener_min) ) ener_min =  1.0; // keV
   if( is_void(ener_max) ) ener_max = 80.0; // keV
   if( is_void(coat) ) coat = 1;

   // set maximal angle for the scattering angle i.e. away from specular direction
   if( is_void(dist_angle_max) ) dist_angle_max = 5.818e-4; // radians (2 arcmin)
      // define angular values for distribution:
   if( is_void(n_dist_angles) ) n_dist_angles = 100;
   angle_dist = span(-dist_angle_max,dist_angle_max,n_dist_angles);

   if( type == 1 ) {

      str = array(string,n_angles*n_ener+1);
      str(1) = "Angle values in radians";

      energy = span(ener_min,ener_max,n_ener)(,-:1:n_angles);
      angle_in = span(0,angle_max,n_angles)(-:1:n_ener,);
      energy = reform(energy,n_angles*n_ener);
      angle_in = reform(angle_in,n_angles*n_ener);
      grow,energy,-1.0;
      grow,angle_in,-1.0;
      energy = shift(energy,-1);
      angle_in = shift(angle_in,-1);

      r_coef = array(double, n_angles*n_ener+1);
      r_coef(1) = -1.0;

      k = 0.12; // Constant product Ec * Acritical
      for( i = 2; i <= n_angles*n_ener+1; i++ ) {
         a = angle_in(i);
         e = energy(i);
         if( a == 0.0 ) {
            r_coef(i) = 1.0;
         } else {
            f = (2./pi)*atan(400.*a); // high energy suppression
            r_coef(i) = 1.0 - f * sost(e,k/a,5.);
         }
      }

      sigma = fwhm/2.35482;
      distribution = exp(-0.5*(angle_dist/sigma)^2)(,-:1:n_angles*n_ener+1);
      distribution(,1) = angle_dist;

      str(2:0) = swrite(format="Gaussian FWHM %10.3e rad (s=%10.3e rad)", fwhm, sigma);

      kwds_init;
      kwds_set,"DATE",ndate(3),"Time of file creation";
      kwds_set,"RESPONSI","Niels J. Westergaard","Responsible person";
      kwds_set,"SITE","NSI/DTU","Institution";
      kwds_set,"ORIGIN","mt_fake_scatter_data","produced this file";
      kwds_set,"VERSION", Version, "MT_RAYOR version indicator";
      kwds_set,"ENERMIN", ener_min,"[keV] Minimum energy";
      kwds_set,"ENERMAX", ener_max,"[keV] Maximum energy";
      kwds_set,"ANGLEMAX", angle_max,"[rad] Maximum grazing angle";
      kwds_set,"DISTAMAX", dist_angle_max,"[rad] Maximum scatter angle";
      kwds_set,"COATING", coat, "Coating number";
      kwds_set,"TUNIT1","keV","Photon energy";
      kwds_set,"TUNIT2","radian","Incoming angle";
      wrmfitscols, filename, "ENERGY", float(energy), "ANGLE_IN", float(angle_in), \
          "R_COEF", float(r_coef), \
          "DISTRIBUTION", float(distribution), "DATA_ORIGIN", str, \
          clobber=1, extname="SCATTER_TYPE1";
      write,format="Has written type 1 scatter file %s\n", filename;

   } else {  // must be type 2 - there is no other option

      angle_in = span( 0., angle_max, n_angles );
      energy = span( ener_min, ener_max, n_ener );

      r_coef = array(double, n_angles, n_ener);

      k = 0.12; // Constant product Ec * Acritical
      for( i = 1; i <= n_angles; i++ ) {
         a = angle_in(i);
         if( a == 0.0 ) {
            r_coef(i,) = 1.0;
         } else {
            f = (2./pi)*atan(400.*a); // high energy suppression
            r_coef(i,) = 1.0 - f * sost(energy,k/a,5.);
         }
      }

      grow, energy, -1.0;
      energy = shift(energy,-1); // Shift last element to be the first one
      grow, r_coef, angle_in;
      r_coef = shift(r_coef,0,-1); // Shift last element to be the first one

      sigma = fwhm/2.35482;
      distribution = exp(-0.5*(angle_dist/sigma)^2);

      kwds_init;
      kwds_set,"DATE",ndate(3),"Time of file creation";
      kwds_set,"RESPONSI","Niels J. Westergaard","Responsible person";
      kwds_set,"SITE","NSI/DTU","Institution";
      kwds_set,"ORIGIN","mt_fake_scatter_data","produced this file";
      kwds_set,"TUNIT1","keV","Photon energy";
      kwds_set,"TUNIT2","radian","Incoming angle";
      kwds_set,"NUM_ENER",n_ener,"Number of energies";
      kwds_set,"NUM_ANGL",n_angles,"Number of angles";
      kwds_set,"COATING", coat,"Coating number (coating type identifier)";
      kwds_set,"GRAZEMAX", angle_max,"[rad] Max of grazing angle";
      kwds_set,"DISTAMAX", dist_angle_max,"[rad] Max of distribution angle";

      write,"##21## going to write first extension:";

      fh = wrmfitscols( filename, "ENERGY", energy, "MATRIX_RCOEF", r_coef, \
          clobber=1, extname="SCATTER_TYPE2", cont=1);
      kwds_init;
      kwds_set,"TUNIT1","radian","Reference angle";
      kwds_set,"DISTAMAX", dist_angle_max,"[rad] Max of distribution angle";
      write,"##22## going to write second extension:";
      wrmfitscols, fh, "ANGLE", angle_dist, "DISTRIBUTION", distribution, \
            extname="SCATTER_TYPE2B";
      write,format="Has written type 2 scatter file %s\n", outfile;

   }

}

/* Function rpar */

/*************************************************************
 *   Functions for grazing incidence modular telescope raytracing
 *
 *   2007-08-02/NJW
 *   2007-08-13/NJW, update to apply 'rt_setup'
 *   2007-09-19/NJW, update to include Wolter I case
 *
 *   The mirror shape defining parameter 'Dcoef' must be defined
 *   externally since function 'rpar' expects it to be present.
 *   For a Wolter I system the parameters 'Fcoef' and 'Acoef'
 *   must similarly exist externally since functions 'rparw'
 *   and 'rhyp' expect them to be present.
 *
 *   Other variables expected as 'external':
 *     angles
 *     scatdist
 *
 *   3D vectors are represented as
 *   single character, upper case variable names
 *
 ****************************************************************/

func rpar( z, phi )
/* DOCUMENT radius = rpar( z, phi )

   returns the radius of a parabolic mirror with focus at
   (x,y,z) = (0,0,Zfocus) and expecting rays coming in with
   direction vectors ~ (0,0,-1)

   Dcoef and Zfocus must have been defined externally.

   If mirror deformations are to be included their value is sought
   in Mirror_deform_arr
*/
{
   extern Dcoef, Use_mdeform, Zfocus;
   r = sqrt(2*Dcoef*(z-Zfocus) + Dcoef^2);
   if( Use_mdeform ) r -= mt_mirror_deform( z, phi );
   return r;
}

/* Function rhyp */

func rhyp( z, phi )
/* DOCUMENT radius = rhyp( z, phi )

   returns the radius of a hyperbolic mirror with focus at
   (x,y,z) = (0,0,0) (second focus in (0,0,-2*Fcoef)
   and expecting rays coming in with
   direction vectors ~ (0,0,-1)

   Acoef, Fcoef and Zfocus must have been defined externally
*/
{
   extern Acoef, Fcoef, Use_mdeform, Zfocus;
   fa = Fcoef^2 - Acoef^2;
   r2 = (z - Zfocus + Fcoef)^2 * fa / Acoef^2 - fa;
   r = sqrt(r2);
   if( Use_mdeform ) r -= mt_mirror_deform( z, phi );
   return r;
}

/* Function rcon */

func rcon( z, phi )
/* DOCUMENT radius = rcon( z, phi )

   returns the radius of a conical mirror at position 'z'

   Mirror_angle, R1_mirror, and Z1_mirror must have been defined externally
*/
{
   extern Mirror_angle, R1_mirror, Use_mdeform, Z1_mirror;
   r = R1_mirror + (z - Z1_mirror)*tan(Mirror_angle);
   if( Use_mdeform ) r -= mt_mirror_deform( z, phi );
   return r;
}

/* Function mdist */

func mdist( funcname, C )
/* DOCUMENT distance = mdist( funcname, C )

   C defines a point in space that can be characterized
   by a z-value, a radius, and an azimuth angle.

   mdist returns the distance from this point to the mirror
   surface with same z and azimuth value. Under normal
   conditions this will to a very good approximation be
   the closest mirror point.
*/
{
   z = C(3);
   phi = atan( C(2), C(1) );
   r = funcname( z, phi ); // radius of mirror at this position
   // F becomes the mirror surface point at same z and azimuth
   F = [ r * cos(phi), r * sin(phi), z ];
   mdist_min = sqrt(sum((C - F)^2));
   rphot = sqrt(sum(C(1:2)^2));
   if( rphot > r ) mdist_min = -mdist_min;
   grow, F, mdist_min;
   return F;
}

/* Function impact */

func impact(funcname, z1, z2, S, R, step= )
/* DOCUMENT pos_dist = impact( funcname, z1, z2, S, R, step=  )

   calculates the interaction point of the ray defined by
   starting position S and direction vector R. The mirror
   is to be found between z1 (entrance) and z2 (exit,  z1 > z2 ).
   A vector of four elements: position plus final distance
   is returned.

   Keyword 'step' is the length step for detecting mirror surface
                  obstacles for the ray. Only active if mirror
                  deformations are applied. Defaults to 5 mm.
*/
{
   extern Use_mdeform;

   if( is_void(step) ) step = 5.0;

   // Advance photon to z1 (the start 'S2')
   t = (z1 - S(3))/R(3);
   S2 = S + t*R;
   if( mdist(funcname,S2)(4) < 0 ) return []; // Behind mirror to begin with

   u = (z2 - S2(3))/R(3); // required to get to end
   if( Use_mdeform ) {

      // Advance photon to z2 (exit) in steps of 5 mm

      upos = step;
      while( upos <= u ) {
         S1 = S2 + upos*R;
         if( mdist(funcname,S1)(4) < 0 ) { // hit the mirror at this position
                                           // or just before
         // try bisections to find the almost exact spot

            t1 = upos - step;
            t2 = upos;
            eps = 1.e29;

            while( eps > 0.0005 ) {
               t = 0.5*(t1 + t2);
               md = mdist(funcname, S2+t*R);
               if( md(4) > 0 ) {t1 = t;} else {t2 = t;}
               eps = abs(md(4));
            }
            return md; // return the point of reflection
         }
         upos += step;
      }
      return []; // no hit, no reflection

   } else { // simple test - no deformations considered

      if( mdist(funcname,S2+u*R)(4) > 0 ) return []; // never hits mirror
      // OK, the mirror was hit, so do the binary search for the exact position

      t1 = 0.0;
      t2 = u;
      eps = 1.e29;

      while( eps > TOLERANCE ) {
         t = 0.5*(t1 + t2);
         md = mdist(funcname, S2+t*R);
         if( md(4) > 0 ) {t1 = t;} else {t2 = t;}
         eps = abs(md(4));
      }
      return md; // return the point of reflection
   }
   return md;
}

/* Function deriv */

func deriv( funcname, x, phi )
/* DOCUMENT dfdx = deriv( funcname, x, phi )

   Very simple function, second argument is
   just a spectator, so it is rather a partial
   derivative.
*/
{
   dx = STEPSIZE; // Size of 'infinitisimal' step
                  // Defined in the preamble of 'mt_rayor.i'
   dfdx = (funcname(x+dx, phi)-funcname(x-dx, phi))/(2.*dx);
   return dfdx;
}

/* Function deriv2 */

func deriv2( funcname, x, phi )
/* DOCUMENT dfdphi = deriv2( funcname, x, phi )

   Very simple function, first argument is
   just a spectator, so it is rather a partial
   derivative.
*/
{
   dphi = 0.01745; // One degree
   dfdphi = (funcname(x, phi+dphi)-funcname(x, phi-dphi))/(2.*dphi);
   return dfdphi;
}

/* Function normal_vector */

func normal_vector( funcname, z, phi )
/* DOCUMENT n = normal_vector( funcname, z, phi )

   returns the normal vector to the surface given by
   'funcname' (rpar, rhyp, or rcon) that depends on the
   two variables: z and phi.
*/
{
   r = funcname(z, phi);
   drdz = deriv( funcname, z, phi );
   drdphi = deriv2( funcname, z, phi );
   sphi = sin(phi); cphi = cos(phi);
   v = [-drdphi*sphi/r - cphi, drdphi*cphi/r - sphi, drdz ];
   v /= sqrt(sum(v^2));
   return v;
}

/* Function spec_reflect */

func spec_reflect( R, N, &Q )
/* DOCUMENT angle = spec_reflect( R, N, Q )

   Calculates the specularly reflected ray as Q
   when the incoming ray is R and the surface normal is N.
   The returned value is the grazing angle [rad].
*/
{
   N = N / sqrt(sum(N^2));
   dot = sum(R*N);
   Q = R - 2*dot*N;
   return asin(-dot);
}

/* Function scat_reflect */

func scat_reflect( R, N, scat_angle, &Q )
/* DOCUMENT angle = scat_reflect( R, N, scat_angle, Q )

   Calculates the scattered reflected ray as Q
   when the incoming ray is R and the surface normal is N.
   The scattering angle scat_angle [rad] is measured from the
   specular reflection with negative values towards the reflecting
   surface and with positive values away from the reflecting surface.
   The returned value is the grazing angle [rad].
*/
{
// first get the specular reflection

   N = N / sqrt(sum(N^2));
   dot = sum(R*N);
   Q = R - 2*dot*N;

// then the direction perpendicular to the specular reflection

   dot2 = dot^2;
   V = (N + dot*R - 2*dot2*N)/sqrt(1.0-dot2);

// The two vectors are added to make the scattered direction

   Q = Q*cos(scat_angle) + V*sin(scat_angle);

   return asin(-dot);
}


/* Function mt_get_rcoef */

func mt_get_rcoef( energy, angle_in )
/* DOCUMENT refl_coef = mt_get_rcoef( energy, angle_in )

   returns bi-linearly interpolated coefficient of reflection
   under current scatter data.

   Version 1.1    2007-09-25/NJW
*/
{

// ynclude = zag
  extern  Angle_inarr,  E_uniq,       Earr,         Num_warn,     R_coefarr, \
          Angle_uniq,   Scatter_type;
// yxclude =

   if( is_void(Num_warn) ) Num_warn = 0;

   // don't let a small difference stop the process
   if( E_uniq(1) - energy > 0.0 && E_uniq(1) - energy < 0.1 ) energy = E_uniq(1);
   if( energy - E_uniq(0) > 0.0 && energy - E_uniq(0) < 1.0 ) energy = E_uniq(0);
   if( energy < E_uniq(1) || energy > E_uniq(0) ) {
      write,format="##23## Energy %13.4e exceeds table range\n", energy;
      if( energy < E_uniq(1) ) {
         write,"because energy span begins at "+ftoa(E_uniq(1),ndec=4);
      } else write,"because energy span ends at "+ftoa(E_uniq(0),ndec=4);
      return [];
   }

   if( angle_in < Angle_uniq(1) || angle_in > Angle_uniq(0) ) {
      Num_warn++;
      if(Num_warn<11)write,format="##24## Warning: Angle %13.4e rad exceeds table range\n", angle_in;
      if( angle_in < Angle_uniq(1) ) angle_in = Angle_uniq(1) == 0.0 ? 1.e-5 : 1.01*Angle_uniq(1);
      if( angle_in > Angle_uniq(0) ) angle_in = 0.99*Angle_uniq(0);
   }

   neu = numberof(E_uniq);
   nea = numberof(Angle_uniq);

   i1 = where(E_uniq <= energy)(0);
   if( i1 == neu ) i1--;
   e1 = E_uniq(i1);
   e2 = E_uniq(i1+1);
   f_e = (energy - e1)/(e2 - e1);

   j1 = where(Angle_uniq <= angle_in)(0);
   if( j1 == nea ) j1--;
   angle1 = Angle_uniq(j1);
   angle2 = Angle_uniq(j1+1);
   f_angle = (angle_in - angle1)/(angle2 - angle1);

   if( Scatter_type == 1 ) { // use TYPE1 method
      idx11 =  where( angle1 == Angle_inarr & e1 == Earr )(1);
      idx12 =  where( angle2 == Angle_inarr & e1 == Earr )(1);
      idx21 =  where( angle1 == Angle_inarr & e2 == Earr )(1);
      idx22 =  where( angle2 == Angle_inarr & e2 == Earr )(1);

      d1 = (1.0 - f_e) * R_coefarr(idx11) + f_e * R_coefarr(idx21);
      d2 = (1.0 - f_e) * R_coefarr(idx12) + f_e * R_coefarr(idx22);
   } else { // use TYPE2 method introduced with version 4.0.1
      // R_coefarr is 2D (angle,energy)
      d1 = (1.0 - f_e) * R_coefarr(j1,i1)   + f_e * R_coefarr(j1,i1+1);
      d2 = (1.0 - f_e) * R_coefarr(j1+1,i1) + f_e * R_coefarr(j1+1,i1+1);
   }
   rcoef = (1.0 - f_angle) * d1 + f_angle * d2;

   return rcoef;

}


/* Function mt_load */

func mt_load( scatfile=, omfile=, mdeffile=, detfile=, gbend=, chat=, master= )
/* DOCUMENT mt_load, scatfile=, omfile=, mdeffile=, detfile=, gbend=, chat=, master=

   Loads 1) scattering      data     by      keyword 'scatfile'
         2) the optical    module  data  by  keyword 'omfile'
         3) the mirror deformation file  by  keyword 'mdeffile'
         4) the detector description file by keyword 'detfile'
   to the memory.

   The keyword 'master' applies only for loading an optical module file
   and it ensures that R_inner will be set.

   The keyword 'gbend' (Gravitational bending) applies only for loading
   a mirror deformation file. The function mt_gravity_bend is called.

   The corresponding external variables are updated as well.

     2007-09-21/NJW
   Updated to version 4.0   2011-08-25/NJW
   Updated to version 4.4.2   2012-09-14/NJW
*/
{

// ynclude = zah
  extern  Acoefarr,             Mirror_coating,       Rb1oarr, \
          Angle_inarr,          Mirror_deform_arr,    Rb2iarr, \
          Angle_uniq,           Mirror_thicknessarr,  Rb2oarr, \
          Anglesarr,            Modtype,              Rdm, \
          Coating_scat,         N_mirrors,            Rmf_file, \
          Dcoefarr,             Ne_mnx,               Scatter_file, \
          Distributionarr,      Num_pixels1,          Xpixlims, \
          Dxb_bkg_file,         Num_pixels2,          Xpixlo, \
          E_max,                Opt_module_file,      Ypixlims, \
          E_min,                Pixel_size1,          Ypixlo, \
          E_mnx,                Pixel_size2,          Z1arr, \
          E_uniq,               Q_ener,               Z2arr, \
          Earr,                 Qeff,                 Zb1iarr, \
          Energ_hi,             R1arr,                Zb1oarr, \
          Energ_lo,             R2arr,                Zb2iarr, \
          Fcoef,                R_coefarr,            Zb2oarr, \
          Instr_bkg_file,       R_inner,              Zfocus, \
          Mirror_anglearr,      Rb1iarr,              Scatter_type, \
          Dol_dead_pixel_map,   Dead_pixel_map;
// yxclude =

   local fh, nrows;


   if( is_void(chat) ) chat = 0;

   if( !is_void(omfile) ) {

      Opt_module_file = omfile;

      if( !file_test(Opt_module_file) ) {
         write,format="The expected optical module file: %s does not exist\n", Opt_module_file;
         write,"Resolve the question and run mt_load again";
      } else {

         //  Get optical module information

         omdol = Opt_module_file+"+1";
         hdr = headfits(omdol);
         Modtype = fxpar( hdr, "MODTYPE" );

         ptr = rdfitsbin( omdol, fh, nrows );
         Z1arr = *ptr(fits_colnum(fh,"z1"));
         Z2arr = *ptr(fits_colnum(fh,"z2"));
         R1arr = *ptr(fits_colnum(fh,"r1"));
         R2arr = *ptr(fits_colnum(fh,"r2"));
         Zb1iarr = *ptr(fits_colnum(fh,"zb1i"));
         Zb2iarr = *ptr(fits_colnum(fh,"zb2i"));
         Rb1iarr = *ptr(fits_colnum(fh,"rb1i"));
         Rb2iarr = *ptr(fits_colnum(fh,"rb2i"));
         Zb1oarr = *ptr(fits_colnum(fh,"zb1o"));
         Zb2oarr = *ptr(fits_colnum(fh,"zb2o"));
         Rb1oarr = *ptr(fits_colnum(fh,"rb1o"));
         Rb2oarr = *ptr(fits_colnum(fh,"rb2o"));
         Mirror_thicknessarr = *ptr(fits_colnum(fh,"mthick"));
         fits_copy_keys, fh, tokwds=1; // Save keywords for mt_upd_om
         if( Modtype == "parabolic" ) {
            Dcoefarr = rdfitscol(omdol,"dcoef");
         } else if( Modtype == "hyperbolic" ) {
            Acoefarr = rdfitscol(omdol,"acoef");
            Fcoef = fxpar( hdr, "FCOEF" );
         } else if( Modtype == "conical" ) {
            Mirror_anglearr = rdfitscol(omdol,"mirror_angle");
         } else error,"Unrecognized type of module";
         Mirror_coating = rdfitscol(omdol,"coating");
         N_mirrors = numberof(R1arr); // avoid innermost mirror, no, not after version 3.2
         Zfocus = fxpar( hdr, "ZFOCUS" );
         if( master ) R_inner = Rb1iarr(1); // Setting R_inner to true inner radius
         if( chat > 0 ) {
            write,format="Using optical module %s\n", Opt_module_file;
            write,format="   with %i mirrors and Zfocus = %10.0f mm\n", N_mirrors, Zfocus;
         }

      }

   }

   if( !is_void(scatfile) ) {

      Scatter_file = scatfile;

      // Get scattering distribution

      if( !file_test(Scatter_file) ) {
         write,format="The expected scatter file: %s does not exist\n", Scatter_file;
         write,"Resolve the question and run mt_load again";
      } else {

         scatdol = Scatter_file+"+1";
         hdr = headfits( scatdol );
         if( is_void((extname = fxpar( hdr, "extname"))) ) error,"Missing EXTNAME in scatter file";
         if( extname != "SCATTER_TYPE1" && extname != "SCATTER_TYPE2" ) \
             error,"Bad EXTNAME of scatter file: "+Scatter_file;
         Coating_scat = fxpar( hdr, "coating" );
         if( is_void(Coating_scat) ) error,"Missing COATING keyword in scatter file";
         nrows = fxpar( hdr, "naxis2" );
         if( nrows <= 1 ) {
            write,format="Found %i rows in %s; insufficient so skip\n", \
                nrows, Scatter_file;
            write,"Resolve the question and run mt_load again";
         } else {
            if( extname == "SCATTER_TYPE1" ) {
               Scatter_type = 1;
               ptr = rdfitsbin( scatdol, hdr, nrows );
               //+ Distributionarr = rdfitscol( scatdol, "DISTRIBUTION" );
               Distributionarr = transpose(*ptr(fits_colnum( hdr, "DISTRIBUTION" )));
               Anglesarr = Distributionarr(,1);
               Distributionarr = Distributionarr(,2:0)/Distributionarr(sum,2:0)(-:1:0,);
               //+ Earr = rdfitscol( scatdol, "ENERGY" )(2:0);
               Earr = (*ptr(fits_colnum( hdr, "ENERGY" )))(2:0);
               //+ Angle_inarr = rdfitscol( scatdol, "ANGLE_IN" )(2:0);
               Angle_inarr = (*ptr(fits_colnum( hdr, "ANGLE_IN" )))(2:0);
               //+ R_coefarr = rdfitscol( scatdol, "R_COEF" )(2:0);
               R_coefarr = (*ptr(fits_colnum( hdr, "R_COEF" )))(2:0);
               // Get energy array without changing the order of Earr
               ise = sort( Earr );
               E_uniq = Earr(ise(uniq(Earr(ise))));
               // Get angle array without changing the order of Angle_inarr
               isa = sort( Angle_inarr );
               Angle_uniq = Angle_inarr(isa(uniq(Angle_inarr(isa))));
            } else { // We have a TYPE2 scatter file
               Scatter_type = 2;
               E_uniq = rdfitscol( scatdol, "energy" )(2:0); // first value is a dummy
               R_coefarr = rdfitscol( scatdol, "matrix_rcoef" );
               Angle_uniq = R_coefarr(,1); // first vector is the angle array
               R_coefarr = R_coefarr(,2:0);
               Distributionarr = rdfitscol( Scatter_file+"+2","distribution");
               Anglesarr = rdfitscol( Scatter_file+"+2","angle");
            }

            if( chat>2 ) {
               write,format="Using scatterfile %s\n", Scatter_file;
               write,format="    with %i energies and %i angles\n", \
                   numberof(E_uniq), numberof(Angle_uniq);
               write,format="  Coating type: %i\n", Coating_scat;
            }
         }
      }
   }

   if( !is_void(mdeffile) ) {

      if( !file_test(mdeffile) ) {
         write,format="The expected mirror deformation file: %s does not exist\n", mdeffile;
         write,"Resolve the question and run mt_load again";
      } else {
         Mirror_deform_arr = readfits(mdeffile+"+1");
         /*
          * Gravitational sag is introduced if requested. The phase is rotated like the optic
          * is rotated.
          */
         if( !is_void(gbend) ) mt_gravity_bend, gbend;
         dms = dimsof(Mirror_deform_arr);
         if( dms(1) != 3 ) error,"Mirror_deform_arr is not a 3D array";
         if( dms(2) < 2 ) error,"Mirror_deform_arr has too few azimuth values";
         if( dms(3) < 2 ) error,"Mirror_deform_arr has too few z values";
         if( dms(4) != N_mirrors ) error,"Mirror_deform_arr does not conform with number of mirrors";
         write,"Has loaded "+mdeffile;
         write,"      into external Mirror_deform_arr";
      }
   }

   if( !is_void(detfile) ) {

      if( !file_test(detfile) ) \
          error,"MT_LOAD did not find detector description file";

      Qeff = []; // to include a flag value
      qeff_file = comgets(detfile,"qeff_file");
      if( is_void(qeff_file) ) {
         write,"No qeff_file keyword found, assume that RMF contains the quantum efficiency";
      } else {
         if( !file_test(qeff_file) ) error,"MT_LOAD did not find the QEFF file: "+qeff_file;

         Q_ener = rdfitscol(qeff_file+"[QUANTUM EFFICIENCY]","ENERGY");
         Qeff   = rdfitscol(qeff_file+"[QUANTUM EFFICIENCY]","QUANTEFF");
         if( is_void(Q_ener) || is_void(Qeff) ) \
                error,"Extension QUANTUM EFFICIENCY or column missing";
      }

      Rmf_file = comgets(detfile,"rmf_file");
      if( is_void(Rmf_file) ) error,"MT_LOAD missing rmf_file keyword";
      if( !file_test(Rmf_file) ) error,"MT_LOAD did not find RMF file";

      rmf_strct = rd_rmf( Rmf_file ); // a function from 'rmf_funcs.i'

      Energ_lo = *rmf_strct.energ_lo;
      Energ_hi = *rmf_strct.energ_hi;
      Rdm      = *rmf_strct.rdm;
      if( is_void(Energ_lo) || is_void(Energ_hi) || is_void(Rdm) ) \
             error,"Extension MATRIX or column missing";
      E_min    = *rmf_strct.e_min;
      E_max    = *rmf_strct.e_max;
      if( is_void(E_min) || is_void(E_max) ) \
             error,"Extension EBOUNDS or column missing";
      E_mnx    = 0.5*(E_min + E_max);
      Ne_mnx = numberof(E_mnx);

      if( is_void(Qeff) ) { // assume that the quantum efficiency information
                            // is contained in the RDM
         /**************
            The statement below is changed to 'neutral' otherwise the selection
            on qeff will be done twice (see function mt_detector).
          *************/
         //+ Qeff = Rdm(sum,);
         Qeff = array(float(1),numberof(Energ_lo));
         Q_ener = 0.5*(Energ_lo+Energ_hi);
      }

      // first dimension of Rdm is over detector bins, E_min,E_max
      // second dimension of Rdm is test energies, Energ_lo, Energ_hi

      Num_pixels1 = comget(detfile,"num_pixels1",lng=1);
      if(is_void(Num_pixels1)) error,"MT_LOAD missing num_pixels1 keyword";
      Num_pixels2 = comget(detfile,"num_pixels2",lng=1);
      if(is_void(Num_pixels2)) error,"MT_LOAD missing num_pixels2 keyword";
      Pixel_size1 = comget(detfile,"pixel_size1");
      if(is_void(Pixel_size1)) error,"MT_LOAD missing pixel_size1 keyword";
      Pixel_size2 = comget(detfile,"pixel_size2");
      if(is_void(Pixel_size2)) error,"MT_LOAD missing pixel_size2 keyword";

      // Load the dead pixel map if found else set to 'all active'
      //    A pixel value of 1 (one) signals that the pixel is dead
      //    A pixel value of 0 (zero) signals that the pixel is active

      Dol_dead_pixel_map = comgets(detfile,"dead_pixel_map");
      if( is_void(Dol_dead_pixel_map)) {
         Dead_pixel_map = array(short,Num_pixels1,Num_pixels2);
      } else {
         local file_dead_pixel_map, extno_dpm;
         get_exten_no, Dol_dead_pixel_map, file_dead_pixel_map, extno_dpm;
         if( !file_test(file_dead_pixel_map) ) error,"##3## MT_LOAD did not find "+file_dead_pixel_map;
         Dead_pixel_map = readfits(Dol_dead_pixel_map);
         dms = dimsof(Dead_pixel_map);
         if( Num_pixels1 != dms(2) || Num_pixels2 != dms(3) ) error,"MT_LOAD inconsistency in dead pixel map dimensions";
      }

      Xpixlo = (-Num_pixels1/2. + indgen(0:Num_pixels1-1))*Pixel_size1;
      Ypixlo = (-Num_pixels2/2. + indgen(0:Num_pixels2-1))*Pixel_size2;
      Xpixlims = 0.5*Pixel_size1*Num_pixels1*[-1,1];
      Ypixlims = 0.5*Pixel_size2*Num_pixels2*[-1,1];

      Dxb_bkg_file = comgets( detfile, "dxb_bkg_file" );
      if( is_void(Dxb_bkg_file) ) {
         write,"Warning! MT_LOAD missing name of dxb_bkg_file";
      } else {
         if( !file_test(Dxb_bkg_file) ) {
            write,"Warning! dxb_bkg_file: "+Dxb_bkg_file+" was not found";
         }
      }
      Instr_bkg_file = comgets( detfile, "instr_bkg_file" );
      if( is_void(Instr_bkg_file) ) {
         write,"Warning! MT_LOAD missing name of instr_bkg_file";
      } else {
         if( !file_test(Instr_bkg_file) ) {
            write,"Warning! instr_bkg_file: "+Instr_bkg_file+" was not found";
         }
      }

   }
}


/* Function mt_mirror_fig */

/*
 *   This is rather a macro to sketch the mirror configuration
 */

func mt_mirror_fig( mirror_number )
/* DOCUMENT mt_mirror_fig, mirror_number

    Make a sketch of the mirror configuration around the given mirror
*/
{
// ynclude = zkk
   extern Om_files, Z_reference;
// yxclude =

   om_par_dol = Om_files(1)+"+1"
   om_hyp_dol = Om_files(2)+"+1"

   Zavg = avg(Z_reference);

   mir = mirror_number;

   R1arr = rdfitscol( om_par_dol, "r1" );
   R2arr = rdfitscol( om_par_dol, "r2" );
   Z1arr = rdfitscol( om_par_dol, "z1" );
   Z2arr = rdfitscol( om_par_dol, "z2" );
   Zb1iarr = rdfitscol( om_par_dol, "zb1i" );
   Zb1oarr = rdfitscol( om_par_dol, "zb1o" );
   Zb2iarr = rdfitscol( om_par_dol, "zb2i" );
   Zb2oarr = rdfitscol( om_par_dol, "zb2o" );
   Rb1iarr = rdfitscol( om_par_dol, "rb1i" );
   Rb1oarr = rdfitscol( om_par_dol, "rb1o" );
   Rb2iarr = rdfitscol( om_par_dol, "rb2i" );
   Rb2oarr = rdfitscol( om_par_dol, "rb2o" );

   R1p = R1arr(mir);
   R2p = R2arr(mir);
   Z1p = Z1arr(mir) + Z_reference(1) - Zavg;
   Z2p = Z2arr(mir) + Z_reference(1) - Zavg;
   Zb1ip = Zb1iarr(mir) + Z_reference(1) - Zavg;
   Zb1op = Zb1oarr(mir) + Z_reference(1) - Zavg;
   Zb2ip = Zb2iarr(mir) + Z_reference(1) - Zavg;
   Zb2op = Zb2oarr(mir) + Z_reference(1) - Zavg;
   Rb1ip = Rb1iarr(mir);
   Rb1op = Rb1oarr(mir);
   Rb2ip = Rb2iarr(mir);
   Rb2op = Rb2oarr(mir);



   R1arr = rdfitscol( om_hyp_dol, "r1" );
   R2arr = rdfitscol( om_hyp_dol, "r2" );
   Z1arr = rdfitscol( om_hyp_dol, "z1" );
   Z2arr = rdfitscol( om_hyp_dol, "z2" );
   Zb1iarr = rdfitscol( om_hyp_dol, "zb1i" );
   Zb1oarr = rdfitscol( om_hyp_dol, "zb1o" );
   Zb2iarr = rdfitscol( om_hyp_dol, "zb2i" );
   Zb2oarr = rdfitscol( om_hyp_dol, "zb2o" );
   Rb1iarr = rdfitscol( om_hyp_dol, "rb1i" );
   Rb1oarr = rdfitscol( om_hyp_dol, "rb1o" );
   Rb2iarr = rdfitscol( om_hyp_dol, "rb2i" );
   Rb2oarr = rdfitscol( om_hyp_dol, "rb2o" );

   R1h = R1arr(mir);
   R2h = R2arr(mir);
   Z1h = Z1arr(mir) + Z_reference(2) - Zavg;
   Z2h = Z2arr(mir) + Z_reference(2) - Zavg;
   Zb1ih = Zb1iarr(mir) + Z_reference(2) - Zavg;
   Zb1oh = Zb1oarr(mir) + Z_reference(2) - Zavg;
   Zb2ih = Zb2iarr(mir) + Z_reference(2) - Zavg;
   Zb2oh = Zb2oarr(mir) + Z_reference(2) - Zavg;
   Rb1ih = Rb1iarr(mir);
   Rb1oh = Rb1oarr(mir);
   Rb2ih = Rb2iarr(mir);
   Rb2oh = Rb2oarr(mir);

   r_arr = [R1p,R2p,Rb1ip,Rb1op,Rb2ip,Rb2op,R1h,R2h,Rb1ih,Rb1oh,Rb2ih,Rb2oh];
   z_arr = [Z1p,Z2p,Zb1ip,Zb1op,Zb2ip,Zb2op,Z1h,Z2h,Zb1ih,Zb1oh,Zb2ih,Zb2oh];

   xr = [min(r_arr)-1, max(r_arr)+1];
   yr = [min(z_arr)-20, max(z_arr)+20];

   plot,[R1p,R2p],[Z1p,Z2p],xr=xr,yr=yr;
   oplot,[xr(1),Rb1ip],[Zb1ip,Zb1ip],thick=7;
   oplot,[Rb1op,xr(2)],[Zb1op,Zb1op],thick=7;
   oplot,[xr(1),Rb2ip],[Zb2ip,Zb2ip],thick=7;
   oplot,[Rb2op,xr(2)],[Zb2op,Zb2op],thick=7;

   oplot,[R1h,R2h],[Z1h,Z2h];
   oplot,[xr(1),Rb1ih],[Zb1ih,Zb1ih],thick=7;
   oplot,[Rb1oh,xr(2)],[Zb1oh,Zb1oh],thick=7;
   oplot,[xr(1),Rb2ih],[Zb2ih,Zb2ih],thick=7;
   oplot,[Rb2oh,xr(2)],[Zb2oh,Zb2oh],thick=7;
}

/* Function mt_qimage */

func mt_qimage( z, size=, dim=, cen=, win=, lg=, rcoef=, bsel=, pal=, title=, aperture= )
/* DOCUMENT im = mt_qimage( z_value )
       or   mt_qimage, z_value

   Called as a function: Returns image of status==0 photon positions in extern 'Phs'
          if 'z_value' is not given, else all photons will be propagated to
          plane at z = z_value.
   Called as a subroutine: Displays the image.

   Argument:
        z_value     The photons are propagated to this distance [mm] from the focal plane
                    that has z = 0.

   Keywords:
         size    defines the image edge size [mm] (default: extreme values).
         dim     defines the dimensionality im(dim,dim) (default: 251).
         cen     is only active when 'size' is set and it can be:
                    1) a 2-element array defining the (x,y) image center in mm
                    2) a scalar 0 (zero) to center on (0,0) [shorthand]
                    3) a scalar 1 (one) to center on the average position
         win     number of display window (default is 0).
         lg      (bool) will cause a display in log-color-scale.
         rcoef   (bool) will cause a weighting with the reflection coefficients.
         aperture (bool) will cause an image of the successful photon positions at entrance.
                 If set then 'z_value' will be discarded.
         bsel    will cause a selection on the 'bounce' value.
                   A negative number means 'not' e.g.
                         mt_qimage(bsel=-3)  means select all ghost rays, where 'bounce' != 3
         pal      will change the default palette ("yarg.gp") with the given one.
         title    Like for 'plot'.

   If called as a subroutine the image will be displayed by function 'disp'

   2007-10-04/NJW
   Version 1.7   2008-12-18/NJW
   Version 4.3   2012-06-01/NJW
*/
{
   extern Phs;
   local cE, cR;

   if( is_void(win) ) win = 0;

   nphots = numberof(Phs);
   eq_nocopy, cE, Phs.E;
   eq_nocopy, cR, Phs.R;
   eq_nocopy, cE1, Phs.E1;
   if( is_void(bsel) ) {
      w = where( Phs.status == 0 );
   } else {
      if( bsel < 0 ) {
         w = where( Phs.status == 0 & Phs.bounce != (-bsel) );
      } else {
         w = where( Phs.status == 0 & Phs.bounce == bsel );
      }
   }
   if( numberof(w) == 0 ) {
      write,"Sorry, no status==0 photons present";
      return [];
   }
   cE = cE(,w);
   cR = cR(,w);
   cE1 = cE1(,w);
   //  propagate to requested z-plane
   if( !aperture ) {
      if( !is_void(z) ) {
         t = (z - cE(3,)) / cR(3,);
         cE += t(-:1:3,)*cR;
      }
   }


   if( is_void(cen) ) {
      cen = array(0.0,2);
   } else {
      if( numberof(cen) == 1 ) {
         if( cen == 0 ) {
            cen = array(0.0,2);
         } else {
            cen = aperture ? [avg(cE1(1,)),avg(cE1(2,))] : [avg(cE(1,)),avg(cE(2,))];
         }
      }
   }

   if( dim ) {
      dims = [dim,dim];
   } else {
      dims = [251,251];
   }

   if( size ) {
      xrg = size*[-0.5,0.5] + cen(1);
      yrg = size*[-0.5,0.5] + cen(2);
   } else {
      xrg = yrg = array(double,2);
      if( aperture ) {
         xrg(2) = max(cE1(1,));
         xrg(1) = min(cE1(1,));
         yrg(2) = max(cE1(2,));
         yrg(1) = min(cE1(2,));
      } else {
         xrg(2) = max(cE(1,));
         xrg(1) = min(cE(1,));
         yrg(2) = max(cE(2,));
         yrg(1) = min(cE(2,));
      }
   }
   //+ write,format="Xrange: %10.4f %10.4f\n", xrg(1), xrg(2);
   //+ write,format="Yrange: %10.4f %10.4f\n", yrg(1), yrg(2);

   weight = rcoef ? Phs(w).rcoef : [];
   if( aperture ) im = makeimageu( dims, cE1(1,), cE1(2,), xr=xrg, yr=yrg, weight=weight );
   else im = makeimageu( dims, cE(1,), cE(2,), xr=xrg, yr=yrg, weight=weight );

   if( am_subroutine() ) {
      xax = span(xrg(1),xrg(2),dims(1));
      yax = span(yrg(1),yrg(2),dims(2));
      window,win,style="boxed.gs";
      if( typeof(pal) == "string" ) {
         if( strpart(pal,-2:0) != ".gp" ) pal += ".gp";
         palette,pal;
      } else palette,"yarg.gp";
      if( lg ) {
         disp, log(1 + im), xtitle="mm", ytitle="mm", title=title, xax=xax, yax=yax;
      } else {
         disp, im, xtitle="mm", ytitle="mm", title=title, xax=xax, yax=yax;
      }
   } else {

      return im;
   }

}

/* Function mt_raytrace_module */

func mt_raytrace_module( no_scatter=, no_mdeform=, gbend=, chat= )
/* DOCUMENT mt_raytrace_module, no_scatter=, gbend=, chat=

   Updates the external array 'Phs' of structs (s_Ray) that contains the information:
      E   : endpoint (array of 3)
      R   : direction at endpoint (array of 3)
      angle_in1  : First module, grazing angle of incoming ray (rad)
      angle_out1 : First module, grazing angle of outgoing ray (rad)
      angle_in2  : Second module, grazing angle of incoming ray (rad)
      angle_out2 : Second module, grazing angle of outgoing ray (rad)
      energy : photon energy
      rcoef : reflection coefficient
      status : photon status, > 0 if obstructed, == 0 if OK
      mirror : mirror number
      bounce : bounce flag

   The photon starts at E, has direction R, will interact with
   ingoing angle 'angle_in' and outgoing angle 'angle_out'.

   Keywords:
        no_scatter=1  will disable scattering (but include Rcoef).
        no_deform=1   will disable mirror deformations.
        chat          puts info on terminal
        gbend         defines the amount of gravitational bending [mm]

   Assume that an optical module has been loaded.

   Upon exit from this function Phs.E is the point of reflection and
   Phs.R the direction of the reflected ray.

   Version 1.0   2007-10-01/NJW
   Version 3.3   2011-02-18/NJW
   Updated for version 4.0  2011-08-25/NJW
*/
{

// ynclude = zai
  extern  Acoef,                Mirror_number,        Rb2oarr, \
          Acoefarr,             Mirror_thicknessarr,  Scatter_file, \
          Anglesarr,            Modtype,              Spoke_define_files, \
          Coat_list,            Module_num,           Use_mdeform, \
          Coating_scat,         Phs,                  Z1_mirror, \
          Dcoef,                R1_mirror,            Z1arr, \
          Dcoefarr,             R1arr,                Z2arr, \
          Mirror_angle,         R2arr,                Zb1iarr, \
          Mirror_anglearr,      Rb1iarr,              Zb1oarr, \
          Mirror_coating,       Rb1oarr,              Zb2iarr, \
          Mirror_deform_files,  Rb2iarr,              Zb2oarr, \
          Mirror_length,        Full_length_spokes1;
// yxclude =

   local Q;

   if( is_void(chat) ) chat = 0;

   if( Modtype == "parabolic" ) {
      mty = 0;
      rfunction = rpar;
   } else if( Modtype == "hyperbolic" ) {
      mty = 1;
      rfunction = rhyp;
   } else if( Modtype == "conical" ) {
      mty = 2;
      rfunction = rcon;
   } else error,"Unrecognized type of module";

   nphots = numberof( Phs );
   scat_angle = 0.0; // covers case for no_scatter==1
   rcoef = 1.0;      // initialization
   mmod = (Module_num - 1)*100;

  /*
   *  See if mirror deformations should be used and load them
   *  if required
   */

   Use_mdeform = 0;
   if( !no_mdeform ) {
      if( Mirror_deform_files(Module_num) != "none" ) {
         mt_load, mdeffile=Mirror_deform_files(Module_num),gbend=gbend;
         Use_mdeform = 1;
      }
   }

  /*
   *  See if spokes have been defined and load the data
   *  if required. Check both for entrance and exit spokes
   *  and read the data if required.
   */

   include_spokes_entrance = 0;
   if( Spoke_define_files(2*Module_num-2+1) != "none" ) {
      mt_spoke_read,pos=1;
      include_spokes_entrance = 1;
   }
   include_spokes_exit = 0;
   if( Spoke_define_files(2*Module_num-2+2) != "none" ) {
      mt_spoke_read,pos=2;
      include_spokes_exit = 1;
   }

  /*
   *  Analyze z values for baffles and mirrors. If there is little
   *  variation for the entrance (exit) values then a simple
   *  and fast algorithm can be applied.
   */

   zvar = 0;
   zb1i_var = max(Zb1iarr) - min(Zb1iarr) < 1.0 ? 0 : 1;
   zb1o_var = max(Zb1oarr) - min(Zb1oarr) < 1.0 ? 0 : 1;
   z1_var = max(Z1arr) - min(Z1arr) < 1.0 ? 0 : 1;
   z2_var = max(Z2arr) - min(Z2arr) < 1.0 ? 0 : 1;
   if( anyof([zb1i_var,zb1o_var]) ) {
      write,"NOTICE: Due to variations in entrance aperture the more";
      write,"elaborate search for mirror number is applied.";
      zvar = 1;
      zb1avg = avg(0.5*(Zb1iarr+Zb1oarr));
   }

  /*
   *  All photons must be propagated to the aperture plane
   *  defined by the inner slit edge of the baffles:
   *       z = max(Zb1iarr)
   *
   *  Note: If the entrance aperture is not close to being a
   *  plane then finding the correct mirror number requires an
   *  iteration that was introduced with version 3.3 [2011-02-18/NJW]
   */

   if( !zvar ) mt_propagate, max(Zb1iarr);

   for( iphot = 1; iphot <= nphots; iphot++ ) {

      // skip if photon has already been blocked before calling this function

      if( Phs(iphot).status != 0 ) continue;

      if( Phs(iphot).mirror <= 0 ) { // then it is the first module
         Phs(iphot).E1 = Phs(iphot).E;
      } else {
         Phs(iphot).E2 = Phs(iphot).E;
      }

     /*
      *   Test for passage between spokes
      */

      if( include_spokes_entrance ) {
         res_spokes = mt_spoke_blocking( Phs(iphot).E(1:2), pos=1 );
         if( res_spokes(1) ) Phs(iphot).status = mmod + 1;
      }

      // test status
      if( Phs(iphot).status != 0 ) continue;

      r = sqrt(Phs(iphot).E(1)^2 + Phs(iphot).E(2)^2);

     /*
      *  If in a slave module then the mirror number is already
      *  defined (Phs.mirror > 0)
      *  else it must be found by the entrance baffle slit
      */

      if( Phs(iphot).mirror <= 0 ) { // mirror number is not known
         // so this must be first optical module.
         // Locate the first slit where the inner baffle radius is less than photon radius
         //  - remember that Rb1iarr decreases with index==Mirror_number
         if( zvar ) { // an iteration must be performed to assign the relevant radius
            mt_propagate, zb1avg, iphot; // propagate to average z
            r = sqrt(Phs(iphot).E(1)^2 + Phs(iphot).E(2)^2);
            zb1i_ipol = interp( Zb1iarr, Rb1iarr, r );
            zb1o_ipol = interp( Zb1oarr, Rb1oarr, r );
            mt_propagate, 0.5*(zb1i_ipol+zb1o_ipol), iphot; // propagate to local z
            r = sqrt(Phs(iphot).E(1)^2 + Phs(iphot).E(2)^2);
         }
         if( r <= min(Rb1iarr) ) {
            Phs(iphot).status = 2;
            continue;
         }
        /*
         * Assign mirror number (in first module)
         */
         //+ Mirror_number = where( Rb1iarr < r )(1);
         Mirror_number = where( Rb1iarr < r )(0);
         // if the radius is too small, then set mirror to one
         // and note in the status value
         if( numberof(Mirror_number) == 0 ) {
            Mirror_number = 1;
            Phs(iphot).status = mmod + 2;
            continue;
         }
         Phs(iphot).mirror = Mirror_number;
      } else {
         Mirror_number = Phs(iphot).mirror;
         // test entrance slit inner edge
         if( Rb1iarr(Mirror_number) > r ) {
            Phs(iphot).status = mmod + 2;
            continue;
         }
      }
      // test entrance slit outer edge
      if( Rb1oarr(Mirror_number) < r ) {
         Phs(iphot).status = mmod + 3;
         continue;
      }

     /*
      *  Now the photon has been tested against the baffle slit
      *  Test the mirror edges if not the innermost mirror
      */

      if( Mirror_number > 1 ) {
         mt_propagate, Z1arr(Mirror_number), iphot;
         r = sqrt(Phs(iphot).E(1)^2 + Phs(iphot).E(2)^2);
         if( r < R1arr(Mirror_number-1) + Mirror_thicknessarr(Mirror_number-1) ) {
            Phs(iphot).status = mmod + 4;
            continue;
         }
         if( r > R1arr(Mirror_number) ) {
            Phs(iphot).status = mmod + 5;
            continue;
         }
      }

      //+ if( !no_scatter )
      if( !is_not_defined(Coat_list) ) {
         // Make sure that the correct scatter file has been loaded - else do
         w = where( Coat_list.id == Mirror_coating(Mirror_number) );
         if( numberof(w) == 0 ) {
            ss = swrite(format="%i", Mirror_coating(Mirror_number));
            error,"##25## Scatter file not present for coating "+ss;
         }
         requested_scatfile = Coat_list(w(1)).file;
         if( Scatter_file != requested_scatfile ) {
            mt_load,scatfile=requested_scatfile,chat=chat;
         }

         // Extra check
         if( Coating_scat != Mirror_coating(Mirror_number) ) {
            error,"##26## problem with coating match";
         }
      }

      // Update Mirror_length for use in rhyp, rpar, or rcon
      Mirror_length = Z1arr(Mirror_number) - Z2arr(Mirror_number);

      // Dcoef, Acoef, Mirror_angle, or R1_mirror
      //      (extern) is used by function 'impact'
      if( mty == 1 ) { // hyperbolic
         Acoef = Acoefarr(Mirror_number);
      } else if( mty == 0 ) {    // parabolic
         Dcoef = Dcoefarr(Mirror_number);
      } else {                   // conical
         Mirror_angle = Mirror_anglearr(Mirror_number);
         R1_mirror = R1arr(Mirror_number);
         Z1_mirror = Z1arr(Mirror_number);
      }

      // find point 'I' of mirror impact

      z1 = Z1arr(Mirror_number);
      z2 = Z2arr(Mirror_number);
      // first reflection by using a step size of 5 mm
      I = impact( rfunction, z1, z2, Phs(iphot).E, Phs(iphot).R, step=5. );
      if( is_void(I) ) { // No impact on reflecting surface

         if( Module_num == 1 ) {
            Phs(iphot).D2 = Phs(iphot).R; // direction after first reflection
            Phs(iphot).I1 = Phs(iphot).E; // place for first (pseudo-) reflection
         }
         I = Phs(iphot).E; // place for pseudo-reflection
         Q = Phs(iphot).R; // same direction after pseudo-reflection

      } else { // the mirror was hit on the reflecting surface

         // update the bounce flag

         Phs(iphot).bounce += Module_num;

         // Get surface normal (N) as derivative of surface function

         phi = atan( I(2), I(1) );
         N = normal_vector( rfunction, I(3), phi );

         // The exit direction from reflection is Q

         angle = spec_reflect( Phs(iphot).R, N, Q );
         if( !no_scatter ) {
            // get the scattering distribution based on energy and angle
            scatdist = mt_sel_scatter( Phs(iphot).energy, angle, rcoef );
            scat_angle = draw_from_dist( Anglesarr, scatdist, 1)(1);
            angle = scat_reflect( Phs(iphot).R, N, scat_angle, Q );
         } else if( !is_not_defined(Coat_list) ) {
            rcoef = mt_get_rcoef( Phs(iphot).energy, angle );
         }
         if( Module_num == 1 ) {
            Phs(iphot).I1 = I(1:3);
            Phs(iphot).D2 = Q;
         } else {
            Phs(iphot).I2 = I(1:3);
            // third direction will be Phs(iphot).R, see later
         }

         Phs(iphot).E = I(1:3);
         Phs(iphot).R = Q;
         if( Module_num == 1 ) {
            Phs(iphot).angle_in1 = angle;
            Phs(iphot).angle_out1 = angle + scat_angle;
         } else {
            Phs(iphot).angle_in2 = angle;
            Phs(iphot).angle_out2 = angle + scat_angle;
         }
         Phs(iphot).rcoef *= rcoef;

        /*
         *   Test for a second reflection the same mirror
         *   (which is treated as an absorption i.e. photon loss)
         */

         // finer resolution with step = 2. mm
         //+ newI = impact( rfunction, I(3), z2, I(1:3), Q, step=2. );
         //+ if( !is_void( newI ) ) { // a second reflection did occur - reject the photon
         //+    Phs(iphot).status = mmod + 7; // error 'behind the mirror'
         //+    continue;
         //+ }

      }

      // test lower edge of next mirror if not innermost mirror
      if( Mirror_number > 1 ) {
         t = (Z2arr(Mirror_number-1) - I(3))/Q(3);
         A = I(1:3) + t*Q;
         r = sqrt(A(1)^2 + A(2)^2);
         if( r < R2arr(Mirror_number-1) + Mirror_thicknessarr(Mirror_number-1) ) {
            Phs(iphot).status = mmod + 6;
            continue;
         }
      }
      // test lower edge of mirror
      t = (Z2arr(Mirror_number) - I(3))/Q(3);
      A = I(1:3) + t*Q;
      r = sqrt(A(1)^2 + A(2)^2);
      if( r > R2arr(Mirror_number) ) {
         Phs(iphot).status = mmod + 8;
         continue;
      }

      // test inner edge of exit slit
      t = (Zb2iarr(Mirror_number) - I(3))/Q(3);
      A = I(1:3) + t*Q;
      r = sqrt(A(1)^2 + A(2)^2);
      if( r < Rb2iarr(Mirror_number) ) {
         Phs(iphot).status = mmod + 9;
         continue;
      }
      // test outer edge of exit slit
      t = (Zb2oarr(Mirror_number) - I(3))/Q(3);
      A = I(1:3) + t*Q;
      r = sqrt(A(1)^2 + A(2)^2);
      if( r > Rb2oarr(Mirror_number) ) {
         Phs(iphot).status = mmod + 10;
         continue;
      }

     /*
      *   Test for spokes at exit
      */

      if( include_spokes_exit ) {
         res_spokes_exit = mt_spoke_blocking( A(1:2), pos=2 );
         if( res_spokes_exit(1) ) {
            Phs(iphot).status = mmod + 11;
            continue;
         }
         if( Full_length_spokes1 ) {
            // Eliminate test if no 'next_spoke' number is given
            // this may happen if no spokes have been defined for a given radius
            if( res_spokes(2) >= 0 && res_spokes_exit(2) >= 0 ) {
               if( res_spokes(2) != res_spokes_exit(2) ) {
                  Phs(iphot).status = mmod + 11;
                  continue;
               }
            }
         }
      }

   }

}

/* Function mt_reflplot */

func mt_reflplot( win=, atten=, angle=, unit= )
/* DOCUMENT mt_reflplot, win=, atten=, angle=, unit=

   plots the reflection coefficients of the scattering
   data currently loaded.
   Keyword      win   indicates the plot window number (0 - 7)
                      defaults to current window.
                atten only one curve for this many angles
                angle Plot only for this angle (overrides 'atten')
                unit  Unit of 'angle'. Can be "deg", "rad" (default), "mrad",
                                       "mdeg", "arcsec", or "arcmin"

   2007-10-31/NJW, updated 2012-10-29/NJW
*/
{

// ynclude = zaj
  extern  Angle_inarr,   Coating_scat,  Earr,          R_coefarr,     Scatter_file, \
          Angle_uniq,    E_uniq;
// yxclude =

   if( is_void(win) ) win = window();
   if( is_void(atten) ) atten = 1;

   n_error = 0;
   if( Scatter_type == 1 && is_void(Earr) ) { n_error++; write,"Extern Earr is missing";}
   if( is_void(E_uniq) ) { n_error++; write,"Extern E_uniq is missing";}
   if( Scatter_type == 1 && is_void(Angle_inarr) ) { n_error++; write,"Extern Angle_inarr is missing";}
   if( is_void(Angle_uniq) ) { n_error++; write,"Extern Angle_uniq is missing";}
   if( is_void(R_coefarr) ) { n_error++; write,"Extern R_coefarr is missing";}
   if( is_void(Scatter_file) ) { n_error++; write,"Extern Scatter_file is missing";}
   if( is_void(Coating_scat) ) { n_error++; write,"Extern Coating_scat is missing";}

   if( n_error > 0 ) {
      write,"Cannot continue";
      return [];
   }

   na = numberof( Angle_uniq );

   window,win,style="boxed.gs";

   if( is_void(angle) ) {

      if( Scatter_type == 1 ) {
         plot, E_uniq, R_coefarr(where(Angle_uniq(1) == Angle_inarr)), \
            title=esc_uscore(Scatter_file), \
            xtitle="Energy [keV]",ytitle="R coef",yr=[0,1.5];
      } else { // must be of type 2
         plot, E_uniq, R_coefarr(1,), \
            title=esc_uscore(Scatter_file), \
            xtitle="Energy [keV]",ytitle="R coef",yr=[0,1.5];
      }
   
      if( na > 1 ) {
         for( i = 2; i <= na; i += atten ) {
            if( Scatter_type == 1 ) {
               oplot, E_uniq, R_coefarr(where(Angle_uniq(i) == Angle_inarr));
            } else {
               oplot, E_uniq, R_coefarr(i,);
            }
         }
      }
      str = swrite(format="Coating type : %i",Coating_scat);
      xyouts,0.2,0.83,str,ndc=1,align=0,charsize=1.1;
      str = swrite(format="Min. angle: %9.2e rad = %8.1f mdeg", \
           Angle_uniq(2), Angle_uniq(2)*(180/pi)*1000.);
      xyouts,0.2,0.80,str,ndc=1,align=0,charsize=1.1;
      str = swrite(format="Max. angle: %9.2e rad = %8.1f mdeg", \
           Angle_uniq(-1), Angle_uniq(-1)*(180/pi)*1000.);
      xyouts,0.2,0.77,str,ndc=1,align=0,charsize=1.1;

   } else {

      if( typeof(unit) == "string" ) {
         if( unit == "deg" ) angle *= pi/180.;
         if( unit == "mrad" ) angle /= 1000.;
         if( unit == "mdeg" ) angle *= pi/180000.;
         if( unit == "arcsec" ) angle *= pi/(180.*3600.);
         if( unit == "arcmin" ) angle *= pi/(180.*60.);
      }
      ne = numberof(E_uniq);
      r = array(double,ne);
      for(i = 1; i <= ne; i++ ) r(i) = mt_get_rcoef(E_uniq(i),angle);
      plot, E_uniq, r, title=esc_uscore(Scatter_file), \
            xtitle="Energy [keV]",ytitle="R coef",yr=[0,1.5];
      str = swrite(format="Coating type : %i",Coating_scat);
      xyouts,0.2,0.83,str,ndc=1,align=0,charsize=1.1;
      str = swrite(format="Angle: %9.2e rad = %8.1f mdeg", \
           angle, angle*(180/pi)*1000.);
      xyouts,0.2,0.80,str,ndc=1,align=0,charsize=1.1;
   }

}

/* Function mt_run */

func mt_run( energy_or_file, src_offaxis, src_azimuth, dphot=, renorm=, \
             exposure=, no_scatter=, no_mdeform=, chat=, flag=, fraper=, \
             labxoff=, labyoff=, roll=, eqillum=, no_par=, gbend=, blur=, silent= )
/* DOCUMENT mt_run, energy_or_file, src_offaxis, src_azimuth, dphot=, renorm=, \
             exposure=, no_scatter=, no_mdeform=, chat=, flag=, fraper=, \
             labxoff=, labyoff=, roll=, eqillum=, no_par=, gbend=, blur=, silent=

   energy_or_file   is either the photon energy in keV (scalar number),
                    a name of a source flux file e.g.
                    'photon_flux_????.fits' (e.g. made by mk_photflux(.i))
                    or a sky definition file with extension SKY_DEFINITION
   src_offaxis is source off axis angle in arcmin
   src_azimuth    is source azimuth in degrees

   dphot      is the photon density in units of /mm2   (default = 1.0)
                  Is only applicable if first argument is energy.
   renorm     the photon spectrum will be multiplied with this factor
                  Is only applicable if first argument is source flux file.
   exposure  defines the exposure time (in seconds) and creates the
                external variable 'Exposure' that will be used in subsequent
                calls of e.g. mt_run and mt_det_add_bkg.
   no_scatter=1 will disable the scattering; if no scattering files
               have been defined then this is the default choice.
               If not set: value will be taken from PFILES/mt_rayor.par
   no_mdeform=1 will disable the mirror deformations
               If not set: value will be taken from PFILES/mt_rayor.par
   chat=0      will stop screen messages
   chat=5      will produce a lot of screen messages
   flag=1      will stop processing after first module
   fraper      "front aperture" - a four-element array with [r1,r2,phi1,phi2]
               (in mm and degrees), default: [R_inner, R_outer, 0., 360.]
   labxoff     Offset [mm] of laboratory source in X direction
   labyoff     Offset [mm] of laboratory source in Y direction
                  These two only apply if a lab source is used
   roll        Roll angle [deg] around the z-axis of the optic in the counterclock
                  direction when viewed from the X-ray source towards the aperture.
  (eqillum     set if equal mirror illumination is requested - unphysical and only for
               special investigations)
   no_par      Avoid using parameter file for no_scatter and no_mdeform
   gbend       Amplitude of gravitational bending [mm] of optic
   blur        [rad] Blur angle for input source position
   silent      Suppress extra messages

   Version 1.0       2007-10-02/NJW
   Version 2.6       2010-12-09/NJW
   Version 3.0       2011-01-17/NJW
   Version 3.1       2011-01-20/NJW
   Version 3.2       2011-01-26/NJW

*/
{

// ynclude = zak
  extern  Dphot,         Num_modules,   R_inner,       Scatter_file,  Z_reference, \
          Energy,        Num_warn,      R_outer,       Src_azimuth,   Zb1iarr, \
          Exposure,      Om_files,      Scatter_files,    Src_offaxis,   Zb1oarr, \
          Module_num,    Phs,           Use_scatter,   Use_mdeform,   Fraper, Fraper_area;
// yxclude =

   time_keeper = elapsed_time = array(double,3);
   timer, time_keeper;


   if( is_void(chat) ) chat = 0;
   if( is_void(dphot) ) dphot = 1.0;
   if( is_void(energy_or_file) ) error,"First argument is void";
   if( typeof(energy_or_file) != "string" ) {
      energy = double(energy_or_file);
   } else {
      if( !file_test(energy_or_file) ) error,"Missing file: "+energy_or_file;
   }
   if( is_void(src_offaxis) ) src_offaxis = 0.0;
   if( is_void(src_azimuth) ) src_azimuth = 0.0;
   if( !is_void(exposure) ) Exposure = exposure;
   if( is_void(renorm) ) { renorm = 1.; } else { renorm = double(renorm); }
   if( typeof(energy) == "double" ) Energy = energy;
   if( !is_void(fraper) ) {
      if( numberof(fraper) == 2 ) {
         grow, fraper, [0., 360.];
      } else if( numberof(fraper) != 4 ) error,"Invalid keyword 'fraper'";
   }
   if( is_void(roll) ) roll = 0.0;
   Src_offaxis = double(src_offaxis);
   Src_azimuth = double(src_azimuth);
   Dphot = double(dphot);

   if( is_void(flag) ) flag = 0;

   if( typeof(energy_or_file) == "string" ) {
      mt_log_entry, 1, "mt_run  *****************  "+ndate(3), \
         swrite(format="%s, src_offaxis %.2f arcmin", energy_or_file, Src_offaxis), \
         swrite(format="src_azimuth = %.1f deg, dphot = %.4f", Src_azimuth, Dphot), \
         swrite(format="renorm = %.4f", renorm);
   } else {
      mt_log_entry, 1, "mt_run  ***************** "+ndate(3), \
         swrite(format="Energy %.3f keV, src_offaxis %.2f arcmin", energy, Src_offaxis), \
         swrite(format="src_azimuth = %.1f deg, dphot = %.4f", Src_azimuth, Dphot);
   }
   if( is_void(blur) ) mt_log_entry, 1, "(mt_run)        No blur has been defined";
   else mt_log_entry, 1, "(mt_run)        Blur value: "+ftoa(blur,ndec=4,sci=1);

   if( !no_par ) { 
      //  By not defining 'no_scatter' means that the value should be taken from the mt_rayor.par file
      if( is_void(no_scatter) ) {
         no_scatter = get_par( "mt_rayor.par","no_scatter" );
      } else {
         set_par, "mt_rayor.par","no_scatter", no_scatter;
      }
   } else {
      if( is_void(no_scatter) ) no_scatter = 0;
   }
   mt_log_entry, 1, "Keyword: no_scatter="+itoa(no_scatter);
   if( !no_par ) { 
      //  By not defining 'no_mdeform' means that the value should be taken from the mt_rayor.par file
      if( is_void(no_mdeform) ) {
         no_mdeform = get_par( "mt_rayor.par","no_mdeform" );
      } else {
         set_par, "mt_rayor.par","no_mdeform", no_mdeform;
      }
   } else {
      if( is_void(no_mdeform) ) no_mdeform = 0;
   }
   mt_log_entry, 1, "Keyword: no_mdeform="+itoa(no_mdeform);


  /*
   *   flag == 1 : stop after first module
   */

   Use_scatter = 1; // default condition
   if( is_void(Scatter_files) ) no_scatter = 1;
        // disable use of scatter data when none is present
   if( no_scatter ) Use_scatter = 0;

   // Generate array of struct (s_Ray) in external 'Phs' with photons

   // Load the first module in order to set the R_inner extern variable
   mt_load,omfile=Om_files(1),chat=chat,master=1; // setting the 'master' keyword
                           // ensures that R_inner will be set to Rb1iarr(1)

   if( is_void(fraper) ) {
      fraper = [R_inner, R_outer, 0., 360. ]; // default setting
   } else {  // adjust to R_inner and R_outer
      if( fraper(1) < R_inner ) fraper(1) = R_inner;
      if( fraper(2) > R_outer ) fraper(2) = R_outer;
   }
   if( fraper(3) > fraper(4) ) fraper(4) += 360.;

   Fraper = fraper;
   Fraper_area = pi * (fraper(2)^2 - fraper(1)^2) * (fraper(4)-fraper(3))/360.;

  /*
   *      GENERATE THE PHOTONS
   */
   mt_pre_def_photons, fraper, energy_or_file, src_offaxis, src_azimuth, \
                dphot=dphot,renorm=renorm,labxoff=labxoff,labyoff=labyoff,eqillum=eqillum,blur=blur;

   if( chat ) write,format="%i photons have been started\n", numberof(Phs);
   if( chat ) write,format="Rotation of the optic %.3f deg around z\n", roll;
   mt_roll, -roll*pi/180; // convert to radians, mt_roll rotates the photons
      // so the roll angle must have opposite sign

   Scatter_file = "None";

   m1 = 1;
   m2 = (flag == 1) ? 1 : Num_modules;

   for( Module_num = m1; Module_num <= m2; Module_num++ ) {

      // Load optical module

      mt_load,omfile=Om_files(Module_num),chat=chat;

      if( Module_num == 1 ) { // change z-coordinate of photons to
                              // aperture plane
         z_aperture = max(max(Zb1iarr),max(Zb1oarr));
         Phs.E(3) = z_aperture;
      }

     /*
      *  mt_raytrace_module uses a coordinate system with origin at
      *  the reference plane. The function mt_pre_def_photons produces photons
      *  with z = 0. However, mt_raytrace_module, leaves the photon
      *  with z at the impact point (reflection point) in the local
      *  coordinate system.
      *  To prepare for next module the photons must be propagated
      *  to the aperture plane and the z value reset to zero.
      */

      if( Module_num > 1 ) {

         // Propagate all photons to the reference plane of the actual module
         // and change coordinate system

         cE = Phs.E; // relative to Z_reference(Module_num-1)
         cR = Phs.R;
         znew = Z_reference(Module_num) - Z_reference(Module_num-1);
         t = (znew - cE(3,)) / cR(3,);
         cE += t(-:1:3,)*cR;
         cE(3,) = 0.0; // reset 'z' to present module system
         Phs.E = cE;

      }

      // Raytrace through optical module

      Num_warn = 0;
      mt_raytrace_module,no_scatter=no_scatter,no_mdeform=no_mdeform,gbend=gbend,chat=chat;
      if( Num_warn > 0 ) write,format="%i off-range angle warnings for module\n", Num_warn;

   }

   // Print statistics
   if( chat>0 ) write,format="%8i  photons in total\n", numberof(Phs);
   if( chat>1 ) {
      mt_stat;
   }

   // Propagate to focal plane at z = 0.0

   if( flag == 0 ) {

      cE = Phs.E;
      cR = Phs.R;
      if( Num_modules == 2 ) {
         cE(3,) += Z_reference(2); // coordinate transformation
      } else {
         cE(3,) += Z_reference(1); // coordinate transformation
      }
      t = (0.0 - cE(3,)) / cR(3,);
      cE += t(-:1:3,)*cR;
      Phs.E = cE;

   }

   //  Rotate photons back, remember the photon rotation is opposite the
   //    rotation of the optic
   mt_roll, roll*pi/180;

   timer, time_keeper, elapsed_time;
   mt_log_entry,1,swrite(format="CPU time for mt_run was: %.3f s", elapsed_time(1));
   if(!silent)write,format="CPU time for mt_run was: %.3f s\n", elapsed_time(1);

}

/* Function mt_restore */

func mt_restore( filenumber )
/* DOCUMENT mt_restore, filenumber

   Restores an MT_RAYOR session previously saved by 'mt_save'.
   The filename is 'ysession_nnnn.ysav' where 'nnnn' is the
   filenumber.

*/
{
    nnnn = swrite(format="%04i", filenumber);
    fname = "ysession_"+nnnn+".ysav";
    f = openb( fname );
    restore, f;
    close, f;
    write,"Session has been restored ...";
}

/* Function mt_save */

func mt_save( mode=, samp=, chat=, outfile=, dir=, nokwdsinit= )
/* DOCUMENT mt_save, mode=, samp=, chat=, outfile=, dir=, nokwdsinit=

   Save current content of memory (e.g. Phs) to one or more of:

       focal_plane_ssss.fits
       photons_ssss.fits
       events_ssss.evt
       session_ssss.ysav (Yorick 'save')
   ('ssss' is a serial number)

   unless 'outfile' (see below) has been given as a string.

   Keywords:
      mode   :  String with one or more of the letters
                    f  -  focal plane (FITS image file)
                    p  -  photons (FITS binary table file)
                    e  -  events (FITS binary table file)
                    s  -  Yorick save of Phs (save file format)
                Default is all of those.
      samp   :  Only one of 'samp' photons will be written to
                      photons_ssss.fits
      chat   :  Degree of verbosity
      outfile:  Is interpreted as the name basis, the final name will become:
                     f<outfile>.fits, p<outfile>.fits, e<outfile>.evt, and y<outfile>.ysav
      dir    :  Directory where output file will be placed
      nokwdsinit : Will inhibit initialization of keywords

   2007-10-02/NJW
   2007-12-20/NJW updated with keyword 'mode'
   Version 1.7  2008-12-18/NJW now saves specified extern variables

*/
{

// ynclude = zal
  extern  Dec_scx,       Exposure,      Num_pixels2,   Posang,        Scatter_file, \
          Dim_focp,      Focal_length,  Om_files,      Ra_scx,        Src_azimuth, \
          Dphot,         Use_mdeform,   Phs,           Rmf_file,      Src_offaxis, \
          Energy,        Num_modules,   Pix_focp,      Scatter_files, Version, \
          Evlist,        Num_pixels1,   Pixel_size1,   Use_scatter,   Fraper_area, \
          Telescop,      Instrume,      Fraper,        Det_offset,    Pixel_size2, \
          Dol_dead_pixel_map;
// yxclude =

   local ser_str;

   if( is_void(chat) ) chat = 0;
   dir = is_void(dir) ? "." : dir;
   dir = app_slash(dir);

   statarr = [0,1,2,3,4,5,6,7,8,9,10,11, \
              101,102,103,104,105,106,107,108,109,110,111];
   nstatarr = numberof(statarr);
   nval = array(long,nstatarr);

   if( is_void(samp) ) samp = 1;
   if( is_void(mode) ) {
      mode = "fpes";
   } else {
      if( structof(mode) != string ) error,"mode must be a string";
      if( !strmatch( mode, "f") && !strmatch( mode, "p") && !strmatch( mode, "e") \
             && !strmatch( mode, "s") ) error,"mode must include one or more of fpes";
   }

   if( typeof(outfile) == "string" ) {
      use_std = 0;
   } else {
      use_std = 1;
      if( file_test("mt_serial.txt") ) {
         ser_num = atoi(rdfile("mt_serial.txt")(1));
         ser_num++;
      } else { // start over
         ser_num = 0;
      }
      ser_str = swrite(format="%04i",ser_num);
      write_slist,"mt_serial.txt",ser_str;
   }

//  define arrays to display and output results

   focp = array(double,Dim_focp,Dim_focp);
   a_focp = double(Dim_focp+1)/2;
   focp_spill = 0;

   nall = numberof(Phs);

   for(i = 1; i <= nstatarr; i++ ) {
      nval(i) = numberof(where(Phs.status==statarr(i)));
      if(chat>1)write,format="Status %4i: %9i\n", statarr(i), nval(i);
   }

//---------------------------------------------------------------------------------
//         Set common keywords for the FITS file(s)
//---------------------------------------------------------------------------------
   if( !nokwdsinit ) kwds_init;
   kwds_set,"N_INJECT",nall,"Number of photons injected";
   if( !is_void(Fraper_area) ) kwds_set,"FRAPAREA", Fraper_area,"[mm2] Front aperture area";
   for( i = 1; i <= nstatarr; i++ ) {
      kw = swrite(format="NSTAT%03i", statarr(i));
      kwds_set,kw,nval(i),"Number of photons with status=="+itoa(statarr(i));
   }
   kwds_set,"DATE",ndate(3),"Time of file creation";
   kwds_set,"ORIGIN","MT_RAYOR-"+Version,"Software used";
   if( !is_void(Telescop) ) kwds_set,"TELESCOP", Telescop,"Name of telescope or mission";
   if( !is_void(Instrume) ) kwds_set,"INSTRUME", Instrume,"Name of instrument";
   if( !is_void(Dphot) ) kwds_set,"PHOTDENS",Dphot,"[/mm2] Input photon surface density";
   if( typeof(Energy) == "double" ) kwds_set,"ENERGY",Energy,"[keV] Energy set in latest run of mt_run";
   if( !is_void(Src_offaxis)) kwds_set,"SRC_OFAX",Src_offaxis,"[arcmin] Source off axis angle";
   if( !is_void(Src_azimuth)) kwds_set,"SRC_AZIM",Src_azimuth,"[deg] Source azimuth axis angle";
   if( !is_void(Fraper) ) {
      kwds_set,"FRAPER1",Fraper(1),"[mm] Inner radius of fraper";
      kwds_set,"FRAPER2",Fraper(2),"[mm] Outer radius of fraper";
      kwds_set,"FRAPER3",Fraper(3),"[deg] Lower angle of fraper";
      kwds_set,"FRAPER4",Fraper(4),"[deg] Upper angle of fraper";
   }
   for(i=1;i<=Num_modules;i++) {
      knam = swrite(format="OMFILE%i",i);
      kwds_set,knam,Om_files(i),"Name of opt. mod. defining file";
   }
   if(!is_void(Use_mdeform))kwds_set,"MDEFORM", Use_mdeform,"0: No mirror deformations, 1: included";
   if( Use_mdeform ) {
      for( i = 1; i <= numberof(Mirror_deform_files); i++ ) {
         knam = swrite(format="MDFILE%i",i);
         kwds_set,knam,Mirror_deform_files(i),"Name of mirror deformation file";
      }
   }
   //+ scatflag = Scatter_file != "None";
   //+ kwds_set,"SCATTER",scatflag,"0: No scatter, 1: with scatter";
   if(!is_void(Use_scatter))kwds_set,"SCATTER",Use_scatter,"0: No scatter, 1: with scatter";
   if( Use_scatter ) {
      nscat = numberof(Scatter_files);
      for(i=1;i<=nscat;i++) {
         knam = swrite(format="SCFL%i",i);
         kwds_set,knam,Scatter_files(i),"Name of scattering defining file";
      }
   }
   if( is_void(Exposure) ) {
      kwds_set,"EXPOSURE",1.0,"[s] Exposure time - default value";
   } else {
      kwds_set,"EXPOSURE",Exposure,"[s] Exposure time";
   }
   if( !is_void(Focal_length) ) kwds_set,"FOCALLEN",Focal_length,"[mm] Focal length";
   kwds_set,"SAMPLING",samp,"Sampling period";
   kwds_set,"NUMMODLS", Num_modules, "Number of optical modules in system";
   kwds_set,"TUNIT1","mm","Unit of DETX";
   kwds_set,"TUNIT2","mm","Unit of DETY";
   kwds_set,"TUNIT6","rad","Unit of ANGLE_IN1";
   kwds_set,"TUNIT7","rad","Unit of ANGLE_OUT1";
   kwds_set,"TUNIT8","rad","Unit of ANGLE_IN2";
   kwds_set,"TUNIT9","rad","Unit of ANGLE_OUT2";
   kwds_set,"TUNIT10","rad","Unit of AZIMUTH";
   kwds_set,"TUNIT12","keV","Unit of ENERGY";

   idx = indgen(1:nall:samp);
   //+ cE = Phs(idx).E;
   local cE; eq_nocopy, cE, Phs(idx).E;
   //+ cR = Phs(idx).R;
   local cR; eq_nocopy, cR, Phs(idx).R;
   local cE1; eq_nocopy, cE1, Phs(idx).E1; // for the azimuth calculation
   local cI1; eq_nocopy, cI1, Phs(idx).I1; // for saving (pos of first refl.)
   local cI2; eq_nocopy, cI2, Phs(idx).I2; // for saving (pos of second refl.)
   azimuth = zero2pi(atan(cE1(2,),cE1(1,)));

//---------------------------------------------------------------------------------
//         Make a photon file
//---------------------------------------------------------------------------------

   if( strpos( mode, "p", 1 ) ) {
      if( use_std ) {
         photon_file = dir+"photons_"+ser_str+".fits";
      } else {
         photon_file = dir+"p"+outfile+".fits";
      }
      wrmfitscols,photon_file,"DETX",cE(1,),"DETY",cE(2,), \
            "RAYX",cR(1,), "RAYY",cR(2,), "RAYZ",cR(3,), \
            "ANGLE_IN1", Phs(idx).angle_in1, \
            "ANGLE_OUT1", Phs(idx).angle_out1, \
            "ANGLE_IN2", Phs(idx).angle_in2, \
            "ANGLE_OUT2", Phs(idx).angle_out2, \
            "AZIMUTH", azimuth, \
            "MIRROR", Phs(idx).mirror, \
            "ENERGY", Phs(idx).energy, \
            "RCOEF", Phs(idx).rcoef, "STATUS", Phs(idx).status, \
            "BOUNCE", Phs(idx).bounce, \
            "FLAG", Phs(idx).flag, \
            "I1Z", cI1(3,), "I2Z", cI2(3,), extname="MT_PHOTONS",clobber=1;
      write,format="Name of photon file: %s\n", photon_file;
      write,format="Number of rows in photon file : %8i\n", numberof(idx);
   }

   write,format="Number of injected photons      : %8i\n", nall;

//---------------------------------------------------------------------------------
//         Make an event file
//---------------------------------------------------------------------------------

   if( strpos( mode, "e", 1 ) ) {
      nEvlist = numberof(Evlist);
      if( nEvlist ) {
         if( use_std ) {
            event_list_file = dir+"events_"+ser_str+".evt";
         } else {
            event_list_file = dir+"e"+outfile+".evt";
         }
         kwds_del,"TUNIT1";
         kwds_del,"TUNIT2";
         kwds_set,"TUNIT3","mm","Unit of DETX";
         kwds_set,"TUNIT4","mm","Unit of DETY";
         kwds_del,"TUNIT5";
         kwds_set,"TUNIT6","keV","Unit of ENERGY";
         kwds_del,"TUNIT7";
         kwds_del,"TUNIT8";
         kwds_del,"TUNIT9";
         kwds_del,"TUNIT10";
         kwds_del,"TUNIT12";
         kwds_set,"RESPONSE", Rmf_file,"Response file (RMF)";
         if( is_void(Dol_dead_pixel_map) ) {
            kwds_set,"DEADPXMP","none","No DOL for dead pixel map has been defined";
         } else {
            kwds_set,"DEADPXMP",Dol_dead_pixel_map,"DOL for appliied dead pixel map";
         }

         // Bring information on the coordinate system if defined

         kwds_set,"BUNIT","cts/pixel","Unit of image pixel values";
         kwds_set,"CTYPE1","RA---TAN","Coordinate system 1. axis";
         //+ kwds_set,"CRPIX1",Num_pixels1/2.,"Reference pixel coordinate 1";
         kwds_set,"CRPIX1",Num_pixels1/2. + Det_offset(1)/Pixel_size1,"Reference pixel coordinate 1";
         crval1 = is_void(Ra_scx) ? 0.0 : Ra_scx;
         kwds_set,"CRVAL1",crval1,"Reference value coordinate 1";
         kwds_set,"CUNIT1","deg","Unit of reference value coordinate 1";
         kwds_set,"CTYPE2","DEC--TAN","Coordinate system 2. axis";
         //+ kwds_set,"CRPIX2",Num_pixels2/2.,"Reference pixel coordinate 2";
         kwds_set,"CRPIX2",Num_pixels2/2. + Det_offset(2)/Pixel_size2,"Reference pixel coordinate 2";
         crval2 = is_void(Dec_scx) ? 0.0 : Dec_scx;
         kwds_set,"CRVAL2",crval2,"Reference value coordinate 2";
         kwds_set,"CUNIT2","deg","Unit of reference value coordinate 2";
         pix1 = (Pixel_size1/Focal_length)*(180./pi);
         rot = is_void(Posang) ? 0.0 : Posang * (pi/180.);
         kwds_set,"CD1_1", -pix1*cos(rot), "[deg] -pix1*cos(rot)";
         kwds_set,"CD1_2", -pix1*sin(rot), "[deg] -pix1*sin(rot)";
         kwds_set,"CD2_1", -pix1*sin(rot), "[deg] -pix1*sin(rot)";
         kwds_set,"CD2_2",  pix1*cos(rot), "[deg]  pix1*cos(rot)";

         kwds_set,"NUMPIX1", Num_pixels1, "Number of detector pixels in X";
         kwds_set,"NUMPIX2", Num_pixels2, "Number of detector pixels in Y";
         kwds_set,"PIXSIZ1", Pixel_size1, "[mm] Pixel size in X";
         kwds_set,"PIXSIZ2", Pixel_size2, "[mm] Pixel size in Y";

         wrmfitscols, event_list_file,"RAWX", Evlist.rawx,"RAWY",Evlist.rawy, \
            "DETX",Evlist.detx,"DETY",Evlist.dety,"PHA",Evlist.pha, \
            "ENERGY",Evlist.energy,"BOUNCE",Evlist.bounce,"FLAG",Evlist.flag, \
            extname="MT_EVENTS", clobber=1;
         write,format="Name of event list file: %s\n", event_list_file;
         write,format="Number of events : %8i\n", nEvlist;
      } else {
         write,"Sorry, no events in list, skip ...";
      }
   }

//---------------------------------------------------------------------------------
//         Make a focal image file
//---------------------------------------------------------------------------------

   if( strpos( mode, "f", 1 ) ) {
      for(k = 1; k <= nall; k++ ) {
         if( Phs(k).status != 0 ) continue;

         i = long(floor(Phs(k).E(1)/Pix_focp + a_focp + 0.5));
         j = long(floor(Phs(k).E(2)/Pix_focp + a_focp + 0.5));
         if( i >= 1 && i <= Dim_focp && j >= 1 && j <= Dim_focp ) {
            focp(i,j) += Phs(k).rcoef;
         } else focp_spill++;
      }

      if( use_std ) {
         focal_plane_file = dir+"focal_plane_"+ser_str+".fits";
      } else {
         focal_plane_file = dir+"f"+outfile+".fits";
      }
      kwds_set,"CRPIX1",a_focp,"Reference pixel";
      kwds_set,"CRVAL1",0.,"Reference value";
      kwds_set,"CDELT1",Pix_focp,"[mm] Pixel size";
      kwds_set,"CRPIX2",a_focp,"Reference pixel";
      kwds_set,"CRVAL2",0.,"Reference value";
      kwds_set,"CDELT2",Pix_focp,"[mm] Pixel size";
      kwds_set,"NOUTSID", focp_spill, "Number of photons outside focal image";
      kwds_set,"EXTNAME","MFOCAL_PLANE","Name of this extension";
      kwds_del,"TUNIT1";
      kwds_del,"TUNIT2";
      kwds_del,"TUNIT6";
      kwds_del,"TUNIT7";

      writefits,focal_plane_file, focp, clobber=1;
      write,format="Name of focal plane file: %s\n", focal_plane_file;
      write,format="Number of counts in focal plane : %8.1f\n", sum(focp);
   }

//---------------------------------------------------------------------------------
//         Make a save file for Phs
//---------------------------------------------------------------------------------

   if( strpos( mode, "s", 1 ) ) {
   // Save contents of Phs
      if( use_std ) {
         yorick_save_name = dir+"mt_session_"+ser_str+".ysav";
      } else {
         yorick_save_name = dir+"y"+outfile+".ysav";
      }

      f = createb(yorick_save_name);

      // avoid problems like: "WARNING bad member name in PDB chart"
      //                      "WARNING bad descriptor name in PDB chart"
      //                      "WARNING no PDB Alignments: extra -- beware of structure instances
      //                      "WARNING unknown primitive in PDB chart
      stru = s_Ray();
      save,f,stru,Phs;

      close,f;
      write,format="Saved Yorick struct Phs in file: %s\n", yorick_save_name;
   }

}


/* Function mt_scatter_data_file */

func mt_scatter_data_file( dir, template, coating, fwhm=, dist_angle_max=, \
                  unit=, outfile=, skip=, attenuate=, graze_angle_max=, gunit= )
/* DOCUMENT mt_scatter_data_file, dir, template, coating, fwhm=, \
                  dist_angle_max=, unit=, outfile=, skip=, attenuate=, \
                  graze_angle_max=, gunit=

   Produces a FITS scatter table in type 1 format from text files with two
   columns: energy [keV] and coefficient of reflection.

   Convention for the text filenames:    TTTTTTTNN_x.xxx
    where TTTTTT is the template (any number of characters)
              NN is the coating number (must be two digits in the filename)
           x.xxx is the grazing angle in degrees

   Arguments:
       dir     :   Directory name where to find the text scatter files
    template   :   Characterizing part of text scatter file names
    coating    :   Integer, coating number requested

   Keywords:
       fwhm    :   The FWHM of the scattering distribution
                        (default 1.38e-4 radians =  28.3 arcsec).
       dist_angle_max : The scattering distribution will be given from
                           -dist_angle_max to +dist_angle_max in 100 steps
                         (default is 1 arcmin).
       unit    :   The angular unit of 'fwhm' AND 'dist_angle_max'.
                   Can be "arcsec", "arcmin", "deg", or "rad" (default "rad").
     outfile   :   The output filename (will be overwritten if it exists beforehand).
        skip   :   Number of starting energy/coefficient values to skip. Defaults to 4.
   attenuate   :   Use only 1 out of 'attenuate' energy/coefficient values
                    E.g. attenuate=3 implies that only every third value is used
                    (useful to avoid creating too large files). Defaults to 3.
   graze_angle_max : The maximum grazing angle to be entered into the scattering data file.
                         Should be larger than the outer mirror angle plus the largest
                         source off-axis angle. Default is 0.05 rad (2.86 deg).
         gunit  :   The angular unit of 'graze_angle_max'.
                    Can be "arcsec", "arcmin", "deg", or "rad" (default "rad").

   Version 4.0      2011-09-02/NJW
*/
{

   if( is_void(outfile) ) outfile = get_next_filename("scatter_???.fits");

   coatstr = swrite(format="%02i", coating);
   list = file_search(template+coatstr+"_*", dir);
   nlist = numberof(list);
   if( nlist == 0 ) {
      write,"No files found by name of "+template+coatstr+"_*";
      return;
   }

   list = list(sort(list));

   energy = [];
   angle_in = [];
   rcoefarr = [];

   if( is_void(skip) ) skip = 3;
   if( is_void(attenuate) ) attenuate = 3;

   // apply the proper conversion for 'graze_angle_max'

   conv_fac = 1.0;
   if( is_void(graze_angle_max) ) {
      graze_angle_max = 0.05; // radians,  equals 2.86 degrees
   } else {
      if( !is_void(gunit) ) {
         if( gunit == "arcsec" ) {
            conv_fac = pi/(180.*60.*60.);
         } else if( gunit == "arcmin" ) {
            conv_fac = pi/(180.*60.);
         } else if( gunit == "deg" ) {
            conv_fac = pi/(180.);
         } else if( gunit == "rad" ) {
            conv_fac = 1.0;
         } else {
            write,"Illegal gunit keyword";
            return [];
         }
      }
      graze_angle_max *= conv_fac;
   }

   first = 1;
   n_angles = 0;
   for( i = 1; i <= nlist; i++ ) {
      nam = list(i);

      write,format="Doing %s ...\n", list(i);

      // find the angle value from the file name
      pus = strpos(nam,"_",rev=1);
      pd  = strpos(nam,".ysav",rev=1);
      if( pd == 0 ) pd = 1;
      angle_deg = atof(strpart(nam,pus+1:pd-1));
      angle_rad = angle_deg * pi / 180.;

// **********   NB    NB    NB   Check following lines carefully!!!!!

      //+ ener = rscol( nam, 1, silent=1 )(6:-6:3); // only use 1 of 3
      //+ rcoef = rscol( nam, 2, silent=1 )(6:-6:3);
      //+ ener  = rscol( nam, 1, silent=1, nomem=1 )(1:0:5); // only use 1 of 5
      //+ rcoef = rscol( nam, 2, silent=1, nomem=1 )(1:0:5); // only use 1 of 5
      // - the two following lines apply to finns_data_101008
      ener  = rscol( nam, 1, silent=1, nomem=1 )(skip+1:0:attenuate); // only use 1 of 'attenuate'
      rcoef = rscol( nam, 2, silent=1, nomem=1 )(skip+1:0:attenuate); // only use 1 of 'attenuate'
              // and reject the first 'skip' values

      // Check that energy values are the same throughout

      if( first ) {
         first = 0;
         ener_ref = ener;
         write,format="Using %s as energy reference\n", nam;
         n_ener = numberof(ener);
         // extra angle = 0 if not already present
         if( angle_rad != 0.0 ) {
            grow, energy, ener;
            grow, angle_in, array(0.0,n_ener);
            grow, rcoefarr, array(1.0,n_ener);
            n_angles++;
         }
      } else {
         dev = abs(ener_ref - ener);
         w = where(dev > 1.e-3);
         if( numberof(w) > 0 ) error,"Wrong e-table in "+nam;
      }

      grow, energy, ener;
      grow, angle_in, array(angle_rad,n_ener);
      grow, rcoefarr, rcoef;
      n_angles++;

   }

   // expand table to cover largest possible angle
   // 1 arcmin = 2.90888e-4 rad
   // 3 arcmin = 8.72665e-4 rad
   // 5 arcmin = 1.45444e-3 rad
   //10 arcmin = 2.90888e-3 rad
   //30 arcmin = 8.72665e-3 rad
   // 1 deg    = 1.74533e-2 rad
   // 2 deg    = 3.49066e-2 rad

// *****     NB    NB    NB  !!!!! there is a problem here
//   if max(angle_in) >= graze_angle_max
// Problem spotted 2014-07-09/NJW but not dealt with yet
//  *************************************
   grow, energy, ener;
   grow, angle_in, array(graze_angle_max,n_ener);
   grow, rcoefarr, rcoef; // repeat latest
   n_angles++;

   // Add first row values (dummy -1)

   grow, energy, -1.0;
   grow, angle_in, -1.0;
   grow, rcoefarr, -1.0;
   energy = shift(energy,-1);
   angle_in = shift(angle_in,-1);
   rcoefarr = shift(rcoefarr,-1);

   conv_fac = 1.0;
   if( is_void(fwhm) ) {
      fwhm = 1.38e-4; // radians = 28.3 arcsec
   } else {
      if( !is_void(unit) ) {
         if( unit == "arcsec" ) {
            conv_fac = pi/(180.*60.*60.);
         } else if( unit == "arcmin" ) {
            conv_fac = pi/(180.*60.);
         } else if( unit == "deg" ) {
            conv_fac = pi/(180.);
         } else if( unit == "rad" ) {
            conv_fac = 1.0;
         } else {
            write,"Illegal unit keyword";
            return [];
         }
      }
      fwhm *= conv_fac;
   }
   if( is_void(dist_angle_max) ) {
      dist_angle_max = 2.90888e-4; // radians = 1 arcmin
   } else {
      dist_angle_max *= conv_fac;
   }

   // define angular values for distribution
   angle_dist = span(-dist_angle_max,dist_angle_max,100);

   str = "Angle values in radians -1 -> 1 arcmin  ";
   s = "                    ";
   str += s+s+s+s;
   str = array(str,n_angles*n_ener+1);


   sigma = fwhm/2.35482;
   distribution = exp(-0.5*(angle_dist/sigma)^2)(,-:1:n_angles*n_ener+1);
   distribution(,1) = angle_dist;

   str(2:0) = swrite(format="Gaussian with FWHM of%10.3e rad (sigma =%10.3e rad)", fwhm, sigma);


   kwds_init;
   kwds_set,"DATE",ndate(3),"Time of file creation";
   kwds_set,"RESPONSI","Niels J. Westergaard","Responsible person";
   kwds_set,"SITE","NSI/DTU","Institution";
   kwds_set,"ORIGIN","mt_scatter_data_file-"+Version,"produced this file";
   kwds_set,"TUNIT1","keV","Photon energy";
   kwds_set,"TUNIT2","radian","Incoming angle";
   kwds_set,"NUM_ENER",n_ener,"Number of energies";
   kwds_set,"NUM_ANGL",n_angles,"Number of angles";
   kwds_set,"COATING", coating,"Coating number (coating type identifier)";
   kwds_set,"GRAZEMAX", max(angle_in),"[rad] Max of grazing angle";
   kwds_set,"DISTAMAX", dist_angle_max,"[rad] Max of distribution angle";
   kwds_set,"SKIP", skip, "Number of energies/coefficients to skip";
   kwds_set,"ATTENUAT", attenuate, "Attenuation of energies/coefficients";
   for( i = 1; i <= nlist; i++ ) {
      kwds_set,"COMMENT",list(i);
   }

   wrmfitscols, outfile, "ENERGY", energy, "ANGLE_IN", angle_in, \
       "R_COEF", rcoefarr, \
       "DISTRIBUTION", distribution, "DATA_ORIGIN", str, \
       clobber=1, extname="SCATTER_TYPE1";
   write,format="Has written scatter file %s\n", outfile;

}

/* Function mt_scatter_data_file_type2 */

func mt_scatter_data_file_type2( dir, template, coating, fwhm=, dist_angle_max=, \
                  n_dist_angle=, unit=, outfile=, skip=, attenuate=, graze_angle_max=, gunit= )
/* DOCUMENT mt_scatter_data_file_type2, dir, template, coating, fwhm=, \
                  dist_angle_max=, n_dist_angle=, unit=, outfile=, skip=, attenuate=, \
                  graze_angle_max=, gunit=

   Produces a FITS scatter table of type 2.

   Convention for the text filenames:    TTTTTTTNN_x.xxx
    where TTTTTT is the template (any number of characters)
              NN is the coating number (must be two digits)
           x.xxx is the grazing angle in degrees
   Two columns are expected, the first one with energies in keV and the
   second one the coefficient of reflection.

   Arguments:
       dir     :   Directory name where to find the text scatter files
    template   :   Characterizing part of text scatter file names (TTTTTTT)
    coating    :   Coating number requested

   Keywords:
       fwhm    :   The FWHM of the scattering distribution
                        (default 1.38e-4 radians =  28.3 arcsec).
       dist_angle_max : The scattering distribution will be given from
                        -dist_angle_max to +dist_angle_max in 'n_dist_angle' steps
                        (default is 1 arcmin).
       n_dist_angle   : Number of elements in the distribution array (default 100).
       unit    :   The angular unit of 'fwhm' AND 'dist_angle_max'.
                   Can be "arcsec", "arcmin", "deg", or "rad" (default "rad").
     outfile   :   The output filename (will be overwritten if it exists beforehand).
        skip   :   Number of starting energy/coefficient values to skip. Defaults to 3.
   attenuate   :   Use only 1 out of 'attenuate' energy/coefficient values
                    E.g. attenuate=3 implies that only every third value is used
                    (useful to avoid creating too large files). Defaults to 3.
   graze_angle_max : The maximum grazing angle to be entered into the scattering data file.
                         Should be larger than the outer mirror angle plus the largest
                         source off-axis angle. Default is 0.05 rad (2.86 deg).
         gunit  :   The angular unit of 'graze_angle_max'.
                    Can be "arcsec", "arcmin", "deg", or "rad" (default "rad").

   Version 4.0      2011-09-02/NJW cloned from mt_scatter_data_file.
*/
{

   if( is_void(outfile) ) outfile = get_next_filename("scatter_???.fits");

   coatstr = swrite(format="%02i", coating);
   list = file_search(template+coatstr+"_*", dir);
   nlist = numberof(list);
   if( nlist == 0 ) {
      write,"No files found by name of "+template+coatstr+"_*";
      return;
   }

   list = list(sort(list));

   angle_in = [];
   rcoefarr = [];

   if( is_void(skip) ) skip = 3;
   if( is_void(attenuate) ) attenuate = 3;
   if( is_void(n_dist_angle) ) n_dist_angle = 100;

   // apply the proper conversion for 'graze_angle_max'

   conv_fac = 1.0;
   if( is_void(graze_angle_max) ) {
      graze_angle_max = 0.05; // radians,  equals 2.86 degrees
   } else {
      if( !is_void(gunit) ) {
         if( gunit == "arcsec" ) {
            conv_fac = pi/(180.*60.*60.);
         } else if( gunit == "arcmin" ) {
            conv_fac = pi/(180.*60.);
         } else if( gunit == "deg" ) {
            conv_fac = pi/(180.);
         } else if( gunit == "rad" ) {
            conv_fac = 1.0;
         } else {
            write,"Illegal gunit keyword";
            return [];
         }
      }
      graze_angle_max *= conv_fac;
   }

   first = 1;
   n_angles = 0;
   for( i = 1; i <= nlist; i++ ) {
      nam = list(i);

      write,format="Doing %s ...\n", list(i);

      // find the angle value from the file name
      pus = strpos(nam,"_",rev=1);
      angle_deg = atof(strpart(nam,pus+1:0));
      angle_rad = angle_deg * pi / 180.;

      // - the two following lines apply to finns_data_101008
      // and reject the first 'skip' values
      ener  = rscol( nam, 1, silent=1, nomem=1 )(skip+1:0:attenuate); // only use 1 of 'attenuate'
      rcoef = rscol( nam, 2, silent=1, nomem=1 )(skip+1:0:attenuate); // only use 1 of 'attenuate'

      // Check that energy values are the same throughout

      if( first ) {
         first = 0;
         ener_ref = ener;
         write,format="Using %s as energy reference\n", nam;
         n_ener = numberof(ener);
         // extra angle = 0 if not already present
         if( angle_rad != 0.0 ) {
            grow, angle_in, 0.0;
            grow, rcoefarr, array(1.0,n_ener);
            n_angles++;
         }
      } else {
         dev = abs(ener_ref - ener);
         w = where(dev > 1.e-3);
         if( numberof(w) > 0 ) error,"Wrong e-table in "+nam;
      }

      grow, angle_in, angle_rad;
      grow, rcoefarr, rcoef;
      n_angles++;

   }

   // expand table to cover largest possible angle
   // 1 arcmin = 2.90888e-4 rad
   // 3 arcmin = 8.72665e-4 rad
   // 5 arcmin = 1.45444e-3 rad
   //10 arcmin = 2.90888e-3 rad
   //30 arcmin = 8.72665e-3 rad
   // 1 deg    = 1.74533e-2 rad
   // 2 deg    = 3.49066e-2 rad

   if( graze_angle_max > max(angle_in) ) {
      grow, angle_in, graze_angle_max;
      grow, rcoefarr, rcoef; // repeat latest
      n_angles++;
   }

   // Change rcoefarr to matrix  n_angles x n_ener

   rcoefarr = transpose(reform(rcoefarr,n_ener,n_angles));
   // Add first row values (dummy -1)

   grow, ener_ref, -1.0;
   grow, rcoefarr, angle_in;
   ener_ref = shift(ener_ref,-1); // Shift last element to be the first one
   rcoefarr = shift(rcoefarr,0,-1); // Shift last element to be the first one

   //+ rcoefarr = transpose( rcoefarr ); // reorganize for the file writing

   conv_fac = 1.0;
   if( is_void(fwhm) ) {
      fwhm = 1.38e-4; // radians = 28.3 arcsec
   } else {
      if( !is_void(unit) ) {
         if( unit == "arcsec" ) {
            conv_fac = pi/(180.*60.*60.);
         } else if( unit == "arcmin" ) {
            conv_fac = pi/(180.*60.);
         } else if( unit == "deg" ) {
            conv_fac = pi/(180.);
         } else if( unit == "rad" ) {
            conv_fac = 1.0;
         } else {
            write,"Illegal unit keyword";
            return [];
         }
      }
      fwhm *= conv_fac;
   }
   if( is_void(dist_angle_max) ) {
      dist_angle_max = 2.90888e-4; // radians = 1 arcmin
   } else {
      dist_angle_max *= conv_fac;
   }

   // define angular values for distribution
   angle_dist = span(-dist_angle_max,dist_angle_max,n_dist_angle);

   sigma = fwhm/2.35482;
   distribution = exp(-0.5*(angle_dist/sigma)^2);

   kwds_init;
   kwds_set,"DATE",ndate(3),"Time of file creation";
   kwds_set,"RESPONSI","Niels J. Westergaard","Responsible person";
   kwds_set,"SITE","NSI/DTU","Institution";
   kwds_set,"ORIGIN","mt_scatter_data_file_type2-"+Version,"produced this file";
   kwds_set,"TUNIT1","keV","Photon energy";
   kwds_set,"TUNIT2","radian","Incoming angle";
   kwds_set,"NUM_ENER",n_ener,"Number of energies";
   kwds_set,"NUM_ANGL",n_angles,"Number of angles";
   kwds_set,"COATING", coating,"Coating number (coating type identifier)";
   kwds_set,"GRAZEMAX", max(angle_in),"[rad] Max of grazing angle";
   kwds_set,"DISTAMAX", dist_angle_max,"[rad] Max of distribution angle";
   kwds_set,"SKIP", skip, "Number of energies/coefficients to skip";
   kwds_set,"ATTENUAT", attenuate, "Attenuation of energies/coefficients";
   for( i = 1; i <= nlist; i++ ) {
      kwds_set,"COMMENT",list(i);
   }

   write,"##27## going to write first extension:";

   fh = wrmfitscols( outfile, "ENERGY", ener_ref, "MATRIX_RCOEF", rcoefarr, \
       clobber=1, extname="SCATTER_TYPE2", cont=1);
   kwds_init;
   kwds_set,"TUNIT1","radian","Reference angle";
   kwds_set,"DISTAMAX", dist_angle_max,"[rad] Max of distribution angle";
   write,"##28## going to write second extension:";
   wrmfitscols, fh, "ANGLE", angle_dist, "DISTRIBUTION", distribution, \
         extname="SCATTER_TYPE2B";
   write,format="Has written type 2 scatter file %s\n", outfile;

}


/* Function mt_sel_scatter */

func mt_sel_scatter( energy, angle_in, &rcoef )
/* DOCUMENT distribution = mt_sel_scatter( energy, angle_in, >rcoef )

   returns bi-linearly interpolated scatter distribution, normalized to sum = 1

   Version 1.1     2007-08-13/NJW
*/
{

// ynclude = zam
  extern  Angle_inarr,      Distributionarr,  Earr,             Num_warn, \
          Angle_uniq,       E_uniq,           Logflag,          R_coefarr, \
          Anglesarr,        Scatter_type;
// yxclude =

   energy = double(energy);
   angle_in = double(angle_in);

   if( Scatter_type == 1 ) { // apply TYPE1 method (from before version 3.5)

      if( energy < E_uniq(1) || energy > E_uniq(0) ) {
         write,format="##29## Energy %13.4e exceeds table range\n", energy;
         return [];
      }

      if( angle_in < Angle_uniq(1) || angle_in > Angle_uniq(0) ) {
         Num_warn++;
         if(Num_warn<11)write,format="##30## Warning: Angle %13.4e rad exceeds table range\n", angle_in;
         if( angle_in < Angle_uniq(1) ) angle_in = Angle_uniq(1) == 0.0 ? 1.e-5 : 1.01*Angle_uniq(1);
         if( angle_in > Angle_uniq(0) ) angle_in = 0.99*Angle_uniq(0);
      }

      i1 = where(E_uniq < energy)(0);
      e1 = E_uniq(i1);
      e2 = E_uniq(i1+1);
      f_e = (energy - e1)/(e2 - e1);

      j1 = where(Angle_uniq < angle_in)(0);
      angle1 = Angle_uniq(j1);
      angle2 = Angle_uniq(j1+1);
      f_angle = (angle_in - angle1)/(angle2 - angle1);

      idx11 =  where( angle1 == Angle_inarr & e1 == Earr )(1);
      idx12 =  where( angle2 == Angle_inarr & e1 == Earr )(1);
      idx21 =  where( angle1 == Angle_inarr & e2 == Earr )(1);
      idx22 =  where( angle2 == Angle_inarr & e2 == Earr )(1);

      d1 = (1.0 - f_e) * Distributionarr(,idx11) + f_e*Distributionarr(,idx21);
      d2 = (1.0 - f_e) * Distributionarr(,idx12) + f_e*Distributionarr(,idx22);
      distri = (1.0 - f_angle) * d1 + f_angle * d2;

      d1 = (1.0 - f_e) * R_coefarr(idx11) + f_e * R_coefarr(idx21);
      d2 = (1.0 - f_e) * R_coefarr(idx12) + f_e * R_coefarr(idx22);
      rcoef = (1.0 - f_angle) * d1 + f_angle * d2; // returned in argument

      if( Logflag > 4 ) {
         coef = gaussfit(Anglesarr, \
           distri+random_n(numberof(distri))*1.e-5,[max(distri),0.,1.e-4]);
         write,format="From gaussian fit: FWHM = %10.2e radians\n",coef(3)*2.3548;
         mt_log_entry, 5, \
            swrite(format="mt_sel_scatter energy: %5.1f, angle_in: %11.3e", \
               energy, angle_in), \
            swrite(format="i1 j1 : %4i%4i, idx11 12 21 22 : %4i%4i%4i%4i", \
                i1, j1, idx11, idx12, idx21, idx22), \
            swrite(format="f_e: %6.3f, f_angle: %6.3f, FWHM: %11.3e", \
                f_e, f_angle, coef(3)*2.3548);
      }
   } else { // Scatter data was TYPE2 so just no interpolation is needed

      distri = Distributionarr;
      rcoef = mt_get_rcoef( energy, angle_in ); // returned in argument

      if( Logflag > 4 ) {
         coef = gaussfit(Anglesarr, \
           distri+random_n(numberof(distri))*1.e-5,[max(distri),0.,1.e-4]);
         write,format="From gaussian fit: FWHM = %10.2e radians (%.1f arcsec)\n", \
                  coef(3)*2.3548, coef(3)*2.3548*206265;
         mt_log_entry, 5, \
            swrite(format="Type 2 scatterfile. energy: %6.3f, angle: %6.3f, FWHM: %11.3e", \
                energy, angle_in, coef(3)*2.3548);
      }
   }

   return distri;

}

/* Function mt_setup_system */

func mt_setup_system( system_filename )
/* DOCUMENT mt_setup_system, system_filename
       status = mt_setup_system( system_filename )

   Reads the telescope system defining file (e.g. system_nustar.scm)
   and defines the external variables and the scattering files.
   Optical module files are tested for existence and for consistency
   with the requested system.

   When called as a function it returns 0 (zero) if ready to run,
   otherwise it returns the warning level - a non-zero value.

   Version 1.1   2007-10-05/NJW
   Version 1.7   2008-12-18/NJW
   Version 1.8   2010-02-09/NJW with mirror deformations
   Version 2.3   2010-06-21/NJW with detector definition
   Version 4.4.2 2012-09-14/NJW with dead pixel detector map
*/
{

// ynclude = zan
  extern  Coat_list,               Num_modules,             Scatter_files, \
          Detector_descr_file,     Om_files,                Spoke_define_files, \
          Dim_focp,                Om_functions,            Z1_setups, \
          Focal_length,            Om_parameters,           Z2_setups, \
          Mirror_deform_files,     Pix_focp,                Z_reference, \
          Mirror_lengths,          R_inner_design,          Zfocusarr, \
          Mirror_thickness_files,  R_outer, Telescop, Instrume, System_filename;
// yxclude =

   mt_log_entry, 1, "mt_setup_system", system_filename;
   if( !file_test(system_filename) ) error,system_filename+" not found";
   System_filename = system_filename;

/**************************************************************

      GENERAL INFORMATION

**************************************************************/

   warn = 0;
   erro = 0;

   Focal_length = comget(system_filename,"focal_length");
   if( is_void(Focal_length) ) {
      write,"Error: Keyword focal_length is missing in "+system_filename;
      erro += 1;
   }

   Num_modules = comget(system_filename,"num_modules",lng=1);
   if( is_void(Num_modules) ) {
      write,"Error: Keyword num_modules is missing in "+system_filename;
      erro += 1;
   }

   om_types = comgets(system_filename,"om_type");
   if( is_void(om_types) ) {
      write,"Error: Keyword om_types is missing in "+system_filename;
      erro += 1;
   }

   if( erro ) {
      write,"Very important keywords are missing!";
      return erro;
   }

/**************************************************************

      OPTICAL  MODULES

**************************************************************/

   Om_files = comgets(system_filename,"om_file");
   if( is_void(Om_files) ) {
      write,"Error: Keyword om_file is missing in "+system_filename;
      erro += 1;
   }

   Om_functions = comgets(system_filename, "om_function");
   if( is_void(Om_functions) ) {
      write,"Error: Keyword om_function is missing in "+system_filename;
      erro += 1;
   }

   Om_parameters = comget(system_filename, "om_parameter");
   if( is_void(Om_parameters) ) {
      write,"Error: Keyword om_parameter is missing in "+system_filename;
      erro += 1;
   }

   Z_reference = comget(system_filename,"z_reference");
   if( is_void(Z_reference) ) {
      write,"Error: Keyword z_reference is missing in "+system_filename;
      erro += 1;
   }

   Zfocusarr = comget(system_filename,"Zfocus");
   if( is_void(Zfocusarr) ) {
      write,"Error: Keyword Zfocus is missing in "+system_filename;
      erro += 1;
   }

   Mirror_lengths = comget(system_filename,"mirror_length");
   if( is_void(Mirror_lengths) ) {
      write,"Error: Keyword mirror_length is missing in "+system_filename;
      erro += 1;
   }

   Mirror_thickness_files = comgets(system_filename,"mirror_thickness_file");
   if( is_void(Mirror_thickness_files) ) {
      write,"Error: Keyword mirror_thickness_file is missing in "+system_filename;
      erro += 1;
   } else {
      for( i = 1; i <= numberof(Mirror_thickness_files); i++ ) {
         if( !file_test(Mirror_thickness_files(i)) ) {
            write,"Error: Missing file: "+Mirror_thickness_files(i);
            erro += 1;
         }
      }
   }

   Mirror_deform_files = comgets(system_filename,"mirror_deform_file");
   if( is_void(Mirror_deform_files) ) {
      Mirror_deform_files = array("none",Num_modules);
   } else {
      for( i = 1; i <= numberof(Mirror_deform_files); i++ ) {
         if( Mirror_deform_files(i) == "none" ) continue;
         if( !file_test(Mirror_deform_files(i)) ) {
            write,"Error: Missing file: "+Mirror_deform_files(i);
            erro += 1;
         }
      }
   }

   if( erro ) {
      write,"Missing keywords or missing files!";
      return erro;
   }

  /*
   *   Either none, a single, or 2*Num_modules occurrences of keyword
   *   'spoke_define_file'
   *   A single is expanded to 2*Num_modules:
   *          Entrance   of first module
   *          Exit       of first module
   *          Entrance   of second module
   *          Exit       of second module
   */

   Spoke_define_files = comgets(system_filename,"spoke_define_file");
   if( is_void(Spoke_define_files) ) {
      Spoke_define_files = array("none",2*Num_modules);
   }
   if( numberof(Spoke_define_files) == 1 ) { // expand to use everywhere
      Spoke_define_files = array(Spoke_define_files,2*Num_modules);
   }

   Z1_setups = comget(system_filename,"z1_setup");
   Z2_setups = comget(system_filename,"z2_setup");

   write,format="Number of modules in system :  %i\n",Num_modules;

   warn = 0;
   if( Num_modules != numberof(Om_files) ) {
      write,"Non matching number of OM files";
      warn += 1;
   }
   if( Num_modules != numberof(om_types) ) {
      write,"Non matching number of om_type keywords";
      warn += 1;
   }
   if( Num_modules != numberof(Mirror_lengths) ) {
      write,"Non matching number of mirror length keywords";
      warn += 1;
   }
   if( Num_modules != numberof(Mirror_thickness_files) ) {
      write,"Non matching number of mirror_thickness_file keywords";
      warn += 1;
   }
   if( Num_modules != numberof(Mirror_deform_files) ) {
      write,"Non matching number of mirror_deform_file keywords";
      warn += 1;
   }
   if( 2*Num_modules != numberof(Spoke_define_files) ) {
      write,"Non matching number of spoke_define_file keywords";
      warn += 1;
   }
   if( Num_modules != numberof(Z_reference) ) {
      write,"Non matching number of reference plane coordinates";
      warn += 1;
   }
   if( Num_modules != numberof(Zfocusarr) ) {
      write,"Non matching number of Zfocus values";
      warn += 1;
   }

   R_inner_design = comget(system_filename,"r_inner");
   if( numberof(R_inner_design) != 1 ) {
      write,"Missing r_inner value";
      warn += 1;
   }
   R_outer = comget(system_filename,"r_outer");
   if( numberof(R_outer) != 1 ) {
      write,"Missing r_outer value";
      warn += 1;
   }

   for(i=1;i<=Num_modules;i++) {
      if( om_types(i) != "parabolic" &&
          om_types(i) != "hyperbolic" &&
          om_types(i) != "conical" ) {
         write,format="Illegal optical module type: %s\n", om_types(i);
         warn += 10;
      }
   }

   for(i=1;i<=Num_modules;i++) {
      write,format="Optical module file :  %s\n",Om_files(i);
      if( !file_test(Om_files(i)) ) {
         write,format="                %s!\n", "not found";
         write,format="You may want to create it by '%s,filename=\"%s\"'\n", \
                  Om_functions(i), Om_files(i);
         warn += 100;
      } else { // Check consistency
         hdr = headfits( Om_files(i)+"+1" );
         if( !near(Mirror_lengths(i), -999.0, 1.e-3) ) {
            f_mirror_length = fxpar( hdr, "mirlngth" );
            if( !near(f_mirror_length, Mirror_lengths(i), 1.e-3) ) {
               write,format="Mismatching mirror length vs. %s\n", Om_files(i);
               write,format="  OM header : %.4f mm\n", f_mirror_length;
               write,format="System file : %.4f mm\n", Mirror_lengths(i);
               warn += 100;
            }
         }
         f_zfocus = fxpar( hdr, "zfocus" );
         if( !near(f_zfocus, Zfocusarr(i), 1.e-2) ) {
            write,format="Mismatching Zfocus value vs. %s\n", Om_files(i);
            write,format="  OM header : %.3f mm\n", f_zfocus;
            write,format="System file : %.3f mm\n", Zfocusarr(i);
            warn += 100;
         }
         f_r_inner = fxpar( hdr, "r_inner" );
         f_r_outer = fxpar( hdr, "r_outer" );
         if( i == 1 ) { // avoid this test for slave modules
            if( numberof(R_inner_design) == 1 ) {
               if( !near( f_r_inner, R_inner_design, 1.e1) ) {
                  write,format="Mismatching r_inner value vs. %s\n", Om_files(i);
                  write,format="  OM header : %.4f mm\n", f_r_inner;
                  write,format="System file : %.4f mm\n", R_inner_design;
                  warn += 100;
               }
            }
            if( numberof(R_outer) == 1 ) {
               if( !near( f_r_outer, R_outer, 1.e-2) ) {
                  write,format="Mismatching r_outer value vs. %s\n", Om_files(i);
                  write,format="  OM header : %.4f mm\n", f_r_outer;
                  write,format="System file : %.4f mm\n", R_outer;
                  warn += 100;
               }
            }
         }
      }

      // Reporting and checking mirror thickness file

      write,format="Mirror thickness file: %s\n", Mirror_thickness_files(i);

      if( file_test(Mirror_thickness_files(i)) ) {
         dummy = rscol(Mirror_thickness_files(i),"radius",nomem=1,silent=1);
         if( is_void(dummy) )  {
            write,format="Column %s not found in mirror thickness file\n","radius";
            warn += 5;
         }
         dummy = rscol(Mirror_thickness_files(i),"mirror_thickness",nomem=1,silent=1);
         if( is_void(dummy) )  {
            write,format="Column %s not found in mirror thickness file\n","mirror_thickness";
            warn += 5;
         }
      }

      // Reporting mirror deformation file

      if( Mirror_deform_files(i) == "none" ) {
         write,format="No mirror deformation to be used for module %i\n", i;
      } else {
         write,format="Mirror deformation file : %s\n", Mirror_deform_files(i);
         if( !file_test(Mirror_deform_files(i)) ) {
            write,format="                %s!\n", "not found";
            ttx = "You may want to create it by 'mt_mk_mdeform_file,";
            write,format=ttx+"\"%s\",mode,param,naz=,nz='\n", \
                     Mirror_deform_files(i);
            warn += 100;
         } else { // Check integrity
            fh = headfits(Mirror_deform_files(i)+"+1");
            if( fxpar(fh,"naxis") != 3 ) {
               write,"Data in mirror deformation file are not a 3D array!";
               warn += 100;
            }
         }
      }

      // Reporting spoke definition file

      for( j = 1; j <= 2; j++ ) {
         ii = (i-1)*2 + j;
         if( Spoke_define_files(ii) == "none" ) {
            write,format="No spoke definition is to be used for module %i\n", i;
         } else {
            write,format="Spoke definition file : %s\n", Spoke_define_files(ii);
            if( !file_test(Spoke_define_files(ii)) ) {
               write,format="             %s!\n","not found";
               warn += 100;
            }
         }
      }
   }

/**************************************************************

      SCATTERING INFORMATION

**************************************************************/

   Scatter_files = comgets(system_filename,"scat_file");
   nscat = numberof(Scatter_files);
   if( nscat ) {
      nacc = numberof( (w = where(Scatter_files != "none")) );
      if( nacc == 0 ) {
         Scatter_files = [];
         nscat = 0;
      } else {
         Scatter_files = Scatter_files(w);
         nscat = nacc;
      }
   }
   n_missing = 0;

   if( nscat > 0 ) { /* only if scatter files have been requested */

      Coat_list = array(s_Coat_list, nscat);

      for( i = 1; i <= nscat; i++ ) {
         // Scatter_files may be links, the target names are preferred
         //+ if( OSTYPE != "nonx" ) Scatter_files(i) = rdline(popen("targetname "+Scatter_files(i),0));
         write,format="Scatter file :  %s",Scatter_files(i);
         Coat_list(i).file = Scatter_files(i);
         if( file_test(Scatter_files(i)) ) {
            hdr = headfits( Scatter_files(i)+"+1" );
            coating = fxpar( hdr, "coating");
            if( is_void(coating) ) error,"FITS keyword COATING is missing in "+Scatter_files(i);
            Coat_list(i).id = coating;
            write,format="  coating %i\n", coating;
         } else {
            write,format="%sdoes not exist\n", " ";
            n_missing++;
            if(n_missing < 9 ) warn += 10000;
         }
      }

   } else Coat_list = []; // Flagging no scatter files included

/**************************************************************

      FOCAL PLANE STANDARD DEFINITION

**************************************************************/

   Dim_focp = comget(system_filename,"Dim_focp",lng=1);
   Pix_focp = comget(system_filename,"Pix_focp");
   if( numberof(Dim_focp) != 1 ) {
      write,"Bad definition of 'Dim_focp'";
      warn += 100000;
   }
   if( numberof(Pix_focp) != 1 ) {
      write,"Bad definition of 'Pix_focp'";
      warn += 100000;
   }

/**************************************************************

      FOCAL PLANE DETECTOR DEFINITION

**************************************************************/

   Detector_descr_file = comgets(system_filename,"detector_descr_file");
   // only a single detector descripton file is acceptable
   if( Detector_descr_file == "none" ) Detector_descr_file = []; // same as not defined
   if( !is_void(Detector_descr_file) ) {
      if( file_test(Detector_descr_file) ) {
         mt_load,detfile=Detector_descr_file;
         write,format="Has loaded detector file: %s\n", Detector_descr_file;
      } else {
         warn += 1000000;
         write,"Did not find detector_descr_file: "+Detector_descr_file;
      }
   } else write,format="%s detector has been loaded\n","No";

/**************************************************************

      OPTIONAL INFORMATION ON TELESCOP AND INSTRUME

**************************************************************/

   // Turn out void if not defined in 'system_filename'
   Telescop = comgets(system_filename,"telescop");
   Instrume = comgets(system_filename,"instrume");

/***********************************************************/

   write,format="Warning level %i\n", warn;
   if( nscat == 0 ) {
      write,"Note that the use of scattering is excluded.";
   }
   if( warn > 0 ) {
      if(warn>=1000000) write,"Detector info is erroneous!";
      warn -= (warn/1000000)*1000000;
      if(warn>=100000) write,"Focal plane info is erroneous!";
      warn -= (warn/100000)*100000;
      if(warn>=10000) write,"One or more scatter files are missing!";
      warn -= (warn/10000)*10000;
      if(warn>=1000) write,"Problematic scattering file(s)!";
      warn -= (warn/1000)*1000;
      if(warn>=100) write,"An optical module file is missing!";
      warn -= (warn/100)*100;
      if(warn>=10) write,"An illegal system type has been specified!";
      warn -= (warn/10)*10;
      if(warn>=1) write,"Problem with basic information!";
   } else {
      write,"System OK!";
   }
   return warn;

}

/* Function mt_upd_om_coating */

func mt_upd_om_coating( coat_table, opt_module_file )
/* DOCUMENT mt_upd_om_coating, coat_table, opt_module_file

     coat_table is an Sformat file with two columns:
           // colname = mirror
           // colname = coating
 
        or with two columns:
          // colname = upper_angle; In radians
          // colname = coating

     OR

     a single number (same coating for all mirrors)

     'opt_module_file' is the Optical Module FITS description file

     2007-10-05/NJW
     2014-10-07/NJW, updated to cope with upper angles
*/
{
   ttype = 0; // ttype == 1: mirror/coating
              // ttype == 2: upper_angle/coating

   if( !file_test( opt_module_file) ) {
      write,"Did not find "+opt_module_file;
      return;
   }

   if( typeof(coat_table) == "string" ) {
      if( !file_test(coat_table) ) {
         write,"Could not find: "+coat_table;
         return;
      }
      colnames = comgets( coat_table, "colname" );
      if( numberof(colnames) < 2 ) {
         write,"Too few 'colname's in "+coat_table;
         return;
      }
      if( n_columns( coat_table ) < 2 ) {
         write,"Too few columns in "+coat_table;
         return;
      }
      if( !numberof(where(colnames == "coating")) ) {
         write,"No column 'coating' found in "+coat_table;
         return;
      }
      coat = rscol(coat_table,"coating",lng=1,silent=1,nomem=1);
      if( numberof(where(colnames == "mirror")) ) {
         ttype = 1;
         mirror = rscol(coat_table,"mirror",lng=1,silent=1,nomem=1);
      } else if( numberof(where(colnames == "upper_angle")) ) {
         ttype = 2;
         upper_angle = rscol(coat_table,"upper_angle",silent=1,nomem=1);
      } else {
         write,"Neither column 'mirror' nor 'upper_angle' found in "+coat_table;
         return;
      }
   } else { // assume single value for all mirrors
      mirror = 1;
      nlines = 0;
      ttype = 1;
      coat = long(coat_table);
   }

   if( ttype > 0 ) {
      omdol = opt_module_file+"+1";
      hdr = headfits(omdol);
      nrows = fxpar(hdr,"naxis2"); // this is the number of mirrors
      if( is_void( fits_colnum( hdr, "mirror_angle", silent=1 ) ) ) {
         // no column MIRROR_ANGLE present - calculate an approximation
         // This happens for true cone sections (paraboloid, hyperboloid)
         qr1 = rdfitscol( omdol, "r1" );
         qr2 = rdfitscol( omdol, "r2" );
         qlen = rdfitscol( omdol, "mlength" );
         mirror_angle = (qr1 - qr2)/qlen; // in radians
      } else {
         mirror_angle = rdfitscol( omdol, "mirror_angle" );
      }
      allcoat = array(long, nrows);
      nlines = numberof(coat);
      modstat = strlowcase(strtrim(fxpar(hdr,"modstat")));
   }

   if( ttype != 2 ) { // Either a mirror/coating table or a single number
   
      if( mirror(1) != 1 ) {
         write,"Error! First mirror in table must be 1!";
         return [];
      }
   
      if( nlines > 1 ) {
         for( i = 2; i <= nlines; i++ ) {
            if( mirror(i) <= mirror(i-1) ) {
               write,"Error! Mirror numbers are not in increasing order!";
               return [];
            }
         }
      }
   
      iline = 1;
      nextline = 2;
      grow, mirror, 99999; // make sure that index is within limits
      grow, coat, 0;
      for( i = 1; i <= nrows; i++ ) {
         if( i == mirror(nextline) ) {
            iline++;
            nextline++;
         }
         allcoat(i) = coat(iline);
      }

   } else {  // Table with upper angles

      // Only for a master module
      if( modstat != "master" ) {
         write,"The optical module is not a master module";
         return;
      }

      stream = open( "_"+coat_table, "w" );
      write,stream,format="// colname = %s\n", "mirror";
      write,stream,format="// colname = %s\n", "coating";
      for( i = 1; i <= nrows; i++ ) {
         w = where( upper_angle > mirror_angle(i) );
         if( !numberof(w) ) {
            write,"Problem with mirror#"+itoa(i)+", where no valid angle is found.";
            close, stream;
            return;
         }
         write,stream,format="  %5i %5i\n", i, coat(w(1));
      }
      close, stream;

      write,"Now '_"+coat_table+" has been created and you have to rerun:";
      write,"> mt_upd_om_coating, \"_"+coat_table+"\", om_1a_file;";
      write,"> mt_upd_om_coating, \"_"+coat_table+"\", om_3a_file;";

      return;

   }

   fits_bintable_poke, omdol, 0, "coating", allcoat;

   write,format="%s has been updated\n", omdol;

}

/* Function get_alpha */

func get_alpha( zf, r1, mlen, &abc, &roots )
/* DOCUMENT  alpha = get_alpha( zf, r1, mlen, >abc, >roots )

   Return angle of mirror that reflects a boresight ray from
   the midpoint to an onaxis point at position 'zf' (focal point).

   See NJW notebook 29/3 2008

   2008-03-29/NJW
*/
{
   // 1 alpha section

   //+ zf = -6000.0; // mm
   //+ r1 = 200.0; // mm
   //+ mlen = 250.0; // mm - mirror length

   t1 = -mlen/2 - zf;
   t2 = r1;
   t3 = mlen/2;

   a = -t2/t3;
   b = -(t3 + 2*t1)/t3;
   c = t2/t3;

   abc = [a,b,c]; // returned as argument

   roots = cubic_eq( a, b, c ); // returned as argument
   moduli = sqrt(double(roots*conj(roots)));
   im = where( moduli == min(moduli) );

   alpha = atan(roots(im(1)).re);

   return alpha;

}

/* Function get_beta */

func get_beta( alpha, zf, r1, mlen, &abc, &roots )
/* DOCUMENT  beta = get_beta( alpha, zf, r1, mlen, >abc, >roots )

   Return angle exceeding 2*alpha of mirror that reflects a ray
   coming at an angle 'alpha' from the midpoint to an onaxis
   point at position 'zf' (focal point).

   See NJW notebook 29/3 2008

   2008-03-29/NJW
*/
{

   // 3 alpha section (or beta section)

   //+ zf = -3000.0; // mm
   //+ r1 = r1 - mlen*tan(alpha); // mm
   //+ mlen = 250.0; // mm

   t1 = -mlen/2 - zf;
   t2 = r1;
   t3 = mlen/2;

   g = tan(2*alpha);

   div = g*g*t1 - g*t2 - t3;
   a = (-3*g*t1 + t2 - 2*g*g*t2 - 3*g*t3)/div;
   b = (2*t1 - g*g*t1 + 3*g*t2 + t3 - 2*g*g*t3)/div;
   c = (g*t1 - t2 + g*t3)/div;

   abc = [a,b,c];

   roots = cubic_eq( a, b, c );
   moduli = sqrt(double(roots*conj(roots)));
   im = where( moduli == min(moduli) );

   beta = atan(roots(im(1)).re);

   return beta;

}

/* Function mt_create_om_con2 */

func mt_create_om_con2( filename= )
/* DOCUMENT mt_create_om_con2, filename=

   Create Optical Module of type conical 1-alpha
   - constant length mirrors
   - adjustable leak by Om_parameter(1)

   The photons move almost along the z-axis in the
   'negative direction' i.e. direction vector ~(0,0,-1).

   The reference surface (origin of z-axis) is at module
   entrance plane.

   Updated to version 4.0  2011-08-25/NJW
*/
{

// ynclude = zao
  extern  Mirror_angle,            R1_mirror,               Z_reference, \
          Mirror_lengths,          R_outer,                 Zfocus, \
          Mirror_thickness_files,  Z1_mirror,               Zfocusarr, \
          Om_parameters, Telescop, Instrume;
// yxclude =

   if( is_void(filename) ) filename = get_next_filename("om_con_???.fits");

   mirror_length = Mirror_lengths(1);

  /*
   * Define mirrors from outside inwards
   *
   *  The photons meets first z1 at entrance and then z2 at exit
   *
   *    r1 is the radius of the reflecting surface at z1
   *    r2 is the radius of the reflecting surface at z2
   *    Therefore: z2 < z1 and r2 < r1
   *
   *  Baffle slits. Conditions are:
   *          At z = zb1i :  r > rb1i
   *      and at z = zb1o :  r < rb1o
   *      and at z = zb2i :  r > rb2i
   *      and at z = zb2o :  r < rb2o
   *
   *
   *  The mirror section defined with this code has constant
   *  mirror lengths.
   *  The aperture is to be exploited
   *  as well as possible.
   *
   */

   Zfocus = Zfocusarr(1); // mm, z-coordinate of focal point in module coord. syst.
   z1 = 0.0; // mm, aperture is reference plane
   r1 = R_outer;

  //  Read arrays in mirror thickness file

   mthick_radius = rscol(Mirror_thickness_files(1),"radius",nomem=1,silent=1);
   mthick_thick  = rscol(Mirror_thickness_files(1),"mirror_thickness",nomem=1,silent=1);

   // Mirrors are characterized by constant length

   Mirror_angle = get_alpha( Zfocus, r1, mirror_length );
   R1_mirror = r1;
   z2 = z1 - mirror_length;
   Z1_mirror = z1; // for the benefit of 'rcon'
   r2 = rcon(z2);
   r1_arr = [];
   r2_arr = [];
   z1_arr = [];
   z2_arr = [];
   mthick_arr = [];
   ma_arr = []; // for Mirror_angle
   coating = [];
   grow, r1_arr, r1;
   grow, r2_arr, r2;
   grow, z1_arr, z1;
   grow, z2_arr, z2;
   mirror_thickness = interp(mthick_thick,mthick_radius,r1);
   grow, mthick_arr, mirror_thickness;
   grow, ma_arr, Mirror_angle;
   grow, coating, 0;
   n_mirrors = 1;

   kwds_init;
   kwds_set,"ZREFRNCE",Z_reference(1),"[mm] Z coord. of refr. plane in tel.syst.";
   kwds_set,"ZFOCUS",Zfocus,"[mm] Z coordinate of focus point in opt.mod. syst.";
   kwds_set,"R_INNER",R_inner_design,"[mm] Inner radius";
   kwds_set,"R_OUTER",R_outer,"[mm] Outer radius";
   kwds_set,"MTCKFILE",Mirror_thickness_files(1),"Mirror thickness table file";
   kwds_set,"MIRLNGTH",mirror_length,"[mm] mirror length";
   kwds_set,"DESIGN","Constant mirror length","Telescope design type";
   kwds_set,"MODTYPE","conical","Type of mirror module";
   kwds_set,"MODSTAT","Master","Defines the mirror positions";
   kwds_set,"PACKFACT", Om_parameters(1),"Loose packing factor, 1.0 means dense";
   kwds_set,"DATE",ndate(3),"Time of file creation";
   if( !is_void(Telescop) ) kwds_set,"TELESCOP", Telescop,"Name of telescope or mission";
   if( !is_void(Instrume) ) kwds_set,"INSTRUME", Instrume,"Name of instrument";
   kwds_set,"TUNIT1","mm","Unit for column 1";
   kwds_set,"TUNIT2","mm","Unit for column 2";
   kwds_set,"TUNIT3","mm","Unit for column 3";
   kwds_set,"TUNIT4","mm","Unit for column 4";
   kwds_set,"TUNIT5","rad","Unit for column 5"; // mirror angle
   kwds_set,"TUNIT6","mm","Unit for column 6";
   kwds_set,"TUNIT7","mm","Unit for column 7";
    // column 8 is the coating number (dimensionless)
   kwds_set,"TUNIT9","mm","Unit for column 9";
   kwds_set,"TUNIT10","mm","Unit for column 10";
   kwds_set,"TUNIT11","mm","Unit for column 11";
   kwds_set,"TUNIT12","mm","Unit for column 12";
   kwds_set,"TUNIT13","mm","Unit for column 13";
   kwds_set,"TUNIT14","mm","Unit for column 14";
   kwds_set,"TUNIT15","mm","Unit for column 15";
   kwds_set,"TUNIT16","mm","Unit for column 16";


   while( r1 > R_inner_design ) {

      n_mirrors++;

      //+ r1 = r2 - mirror_thickness;
      //+ "loose" packing with factor Om_parameters(1)
      r1 = r1 - Om_parameters(1)*(r1 - r2) - mirror_thickness;
      // Dcoef: parabolic constant; required for function 'rpar'
      Mirror_angle = get_alpha( Zfocus, r1, mirror_length );
      R1_mirror = r1;

      r2 = rcon(z2);

      // add to arrays

      grow, r1_arr, r1;
      grow, r2_arr, r2;
      grow, z1_arr, z1;
      grow, z2_arr, z2;
      mirror_thickness = interp(mthick_thick,mthick_radius,r1);
      grow, mthick_arr, mirror_thickness;
      grow, ma_arr, Mirror_angle;
      grow, coating, 1;
   }

  /*
   *  Create the baffle parameters
   */

   rb1i_arr = shift(r1_arr,1) + shift(mthick_arr,1);
   // Special action for innermost baffle slit that is
   // assumed to have the same opening as its neighbor
   open_space = r1_arr(-1) - rb1i_arr(-1);
   rb1i_arr(0) = r1_arr(0) - open_space;
   rb1o_arr = r1_arr;
   zb1i_arr = z1_arr;
   zb1o_arr = z1_arr;

   rb2i_arr = shift(r2_arr,1) + shift(mthick_arr,1);
   // Special action for innermost baffle slit that is
   // assumed to have the same opening as its neighbor
   open_space = r2_arr(-1) - rb2i_arr(-1);
   rb2i_arr(0) = r2_arr(0) - open_space;
   rb2o_arr = r2_arr;
   zb2i_arr = z2_arr;
   zb2o_arr = z2_arr;

  /*
   *  create the FITS binary table while reversing the order of the arrays
   */

   n = numberof(r1_arr);
   i = indgen(n:1:-1);
   wrmfitscols, filename, "R1",r1_arr(i),"R2",r2_arr(i),"Z1",z1_arr(i),"Z2",z2_arr(i), \
       "MIRROR_ANGLE",ma_arr(i),"MLENGTH",z1_arr(i)-z2_arr(i),"MTHICK",mthick_arr(i), "COATING",coating(i), \
       "RB1I", rb1i_arr(i), "RB1O", rb1o_arr(i), "RB2I", rb2i_arr(i), "RB2O", rb2o_arr(i), \
       "ZB1I", zb1i_arr(i), "ZB1O", zb1o_arr(i), "ZB2I", zb2i_arr(i), "ZB2O", zb2o_arr(i), \
        clobber=1;
   write,format="The conical system optical module file: %s has just been created\n", filename;
   write,format="It has %i mirrors\n", n_mirrors;
   write,format="and it might need a coating update: %s\n","mt_upd_om_coating,...";

}


/* Function mt_create_om_con3 */

func mt_create_om_con3( filename=, master= )
/* DOCUMENT mt_create_om_con3, filename=, master=

   Create Optical Module of type conical 3-alpha (or rather 2alpha+beta)
   - constant length mirrors

   This is a 'slave'-module so the mirror radii at entrance
   are defined by the 'R2' array of the first module.
   The master module is by default given in Om_files(1) as set up
   by 'mt_setup_system' from a system defining file.
   It can be overridden by giving another one in keyword 'master'.

   The photons move almost along the z-axis in the
   'negative direction' i.e. direction vector ~(0,0,-1).

   The reference surface (origin of z-axis) is at module
   entrance plane.

   Updated to version 4.0  2011-08-25/NJW
*/
{

// ynclude = zap
  extern  Mirror_angle,            Num_modules,             Z_reference, \
          Mirror_anglearr,         Om_files,                Zfocus, \
          Mirror_lengths,          R1_mirror,               Zfocusarr, \
          Mirror_thickness_files, Telescop, Instrume;
// yxclude =

   if( is_void(filename) ) filename = get_next_filename("om_con_???.fits");

   mirror_length = Mirror_lengths(2);

  /*
   *   Consistency check
   */

   if( is_void(Num_modules) ) {
      write,"External Num_modules not set - did you run mt_setup_system ?";
      return [];
   }
   if( numberof(Z_reference) != Num_modules ) {
      write,"Inconsistency between externals Z_reference and Num_modules";
      return [];
   }
   if( is_void(master) ) master = Om_files(1);
   if( ! file_test( master ) ) {
      write,format="Master opt. mod. file: %s was not found\n", master;
      return [];
   }

  //  Read arrays in mirror thickness file

   mthick_radius = rscol(Mirror_thickness_files(2),"radius",nomem=1,silent=1);
   mthick_thick  = rscol(Mirror_thickness_files(2),"mirror_thickness",nomem=1,silent=1);

  /*
   *! Define mirrors from inside outwards
   *
   *
   *  The photons meets first z1 at entrance and then z2 at exit
   *
   *    r1 is the radius of the reflecting surface at z1
   *    r2 is the radius of the reflecting surface at z2
   *    Therefore: z2 < z1 and r2 < r1
   *
   *  Baffle slits. Conditions are:
   *          At z = zb1i :  r > rb1i
   *      and at z = zb1o :  r < rb1o
   *      and at z = zb2i :  r > rb2i
   *      and at z = zb2o :  r < rb2o
   *
   *
   *  The mirror section defined with this code has constant
   *  mirror lengths.
   *  The aperture is to be exploited
   *  as well as possible.
   *
   *   2008-03-31/NJW
   */

  /*
   *   There is no gap between the master module and the slave module
   *   so r1_arr of the slave is equal to the r2_arr of the master
   */

   // Reading the r2_arr of the master
   r1_arr = rdfitscol( master+"+1", "R2" );
   alpha1_arr = rdfitscol( master+"+1", "MIRROR_ANGLE" );

   // The conical 3-alpha optical module is number 2

   Zfocus = Zfocusarr(2); // mm, z-coordinate of focal point in module coord. syst.
   z1 = 0.0;              // mm, aperture is reference plane
   z2 = z1 - mirror_length;

   // Mirrors are characterized by constant length

   r1_arr = double(r1_arr);

   r2_arr = r1_arr; // just array definition
   Mirror_anglearr = r1_arr; // just array definition
   nmir = numberof( r1_arr );
   z1_arr = array(z1, nmir);
   z2_arr = array(z2, nmir);
   for( mir = 1; mir <= nmir; mir++ ) {
      alpha = alpha1_arr(mir);
      beta = get_beta( alpha, Zfocus, r1_arr(mir), mirror_length );
      Mirror_anglearr(mir) = 2*alpha + beta;
      Mirror_angle = Mirror_anglearr(mir); // required by 'rcon'
      R1_mirror = r1_arr(mir); // required by 'rcon'
      r2_arr(mir) = rcon( z2 );
   }
   coating = array(1,nmir);

   mthick_arr = interp(mthick_thick,mthick_radius,r1_arr);

   kwds_init;
   kwds_set,"FCLENGTH",abs(Zfocus),"[mm] Focal length";
   kwds_set,"ZFOCUS",Zfocus,"[mm] Z coordinate of focus point";
   kwds_set,"R_INNER",r1_arr(0),"[mm] Inner radius";
   kwds_set,"R_OUTER",r1_arr(1),"[mm] Outer radius";
   kwds_set,"MODSTAT","Slave","This is dependent on the first (master) module";
   kwds_set,"MASTER", master,"The opt-mod. that acts as master";
   kwds_set,"MTCKFILE",Mirror_thickness_files(2),"Mirror thickness table file";
   kwds_set,"MIRLNGTH",mirror_length,"[mm] mirror length";
   kwds_set,"DESIGN","Constant mirror length","Telescope design type";
   kwds_set,"MODTYPE","conical","Type of mirror module";
   kwds_set,"DATE",ndate(3),"Time of file creation";
   if( !is_void(Telescop) ) kwds_set,"TELESCOP", Telescop,"Name of telescope or mission";
   if( !is_void(Instrume) ) kwds_set,"INSTRUME", Instrume,"Name of instrument";
   kwds_set,"TUNIT1","mm","Unit for column 1";
   kwds_set,"TUNIT2","mm","Unit for column 2";
   kwds_set,"TUNIT3","mm","Unit for column 3";
   kwds_set,"TUNIT4","mm","Unit for column 4";
   kwds_set,"TUNIT5","rad","Unit for column 5"; // mirror angle
   kwds_set,"TUNIT6","mm","Unit for column 6";
   kwds_set,"TUNIT7","mm","Unit for column 7";
    // column 8 is the coating number (dimensionless)
   kwds_set,"TUNIT9","mm","Unit for column 9";
   kwds_set,"TUNIT10","mm","Unit for column 10";
   kwds_set,"TUNIT11","mm","Unit for column 11";
   kwds_set,"TUNIT12","mm","Unit for column 12";
   kwds_set,"TUNIT13","mm","Unit for column 13";
   kwds_set,"TUNIT14","mm","Unit for column 14";
   kwds_set,"TUNIT15","mm","Unit for column 15";
   kwds_set,"TUNIT16","mm","Unit for column 16";


  /*
   *  Create the baffle parameters
   */

  /*
   *  Standard definition of rb1i is backside of next mirror
   *  at z = z1
   */
   rb1i_arr = shift(r1_arr,-1) + shift(mthick_arr,-1);
   // Special action for innermost baffle slit that is
   // assumed to have the same opening as its neighbor
   open_space = r1_arr(2) - rb1i_arr(2);
   rb1i_arr(1) = r1_arr(1) - open_space;
   rb1o_arr = r1_arr;
   zb1i_arr = z1_arr;
   zb1o_arr = z1_arr;

   rb2i_arr = shift(r2_arr,-1) + shift(mthick_arr,-1);
   // Special action for innermost baffle slit that is
   // assumed to have the same opening as its neighbor
   open_space = r2_arr(2) - rb2i_arr(2);
   rb2i_arr(1) = r2_arr(1) - open_space;
   rb2o_arr = r2_arr;
   zb2i_arr = z2_arr;
   zb2o_arr = z2_arr;

  /*
   *  create the FITS binary table
   */

   wrmfitscols, filename, "R1",r1_arr,"R2",r2_arr,"Z1",z1_arr,"Z2",z2_arr, \
       "MIRROR_ANGLE",Mirror_anglearr,"MLENGTH",z1_arr-z2_arr,"MTHICK", mthick_arr, "COATING",coating, \
       "RB1I", rb1i_arr, "RB1O", rb1o_arr, "RB2I", rb2i_arr, "RB2O", rb2o_arr, \
       "ZB1I", zb1i_arr, "ZB1O", zb1o_arr, "ZB2I", zb2i_arr, "ZB2O", zb2o_arr, \
        clobber=1;
   write,format="The conical system optical module file: %s has just been created\n", filename;
   write,format="It has %i mirrors\n", numberof(r1_arr);
   write,format="and it might need a coating update: %s\n","mt_upd_om_coating,...";

}

/* Function mt_mirdiag */

func mt_mirdiag( om_file, rr=, zr=, gl=, over=, z_offset=, phi= )
/* DOCUMENT mt_mirdiag, om_file, rr=, zr=, gl=, over=, z_offset=, phi=

   "Mirror Diagram" showing placements of mirrors and baffles

   Reads data from 'om_file'; if it is not given then data
   from memory will be used.

   Keywords: rr    Range in radius, a two element array
             zr    Range in z-direction, defaults to 10% outside mirrors
             gl    Show guide lines as well
             over  'Overplot' disregard 'rr' and 'zr' and plot on
                       existing frame and window
             z_offset Additive change of Z
             phi   azimuth angle [rad] - required if mirror deformations
                      are to be included  as indicated in 'Use_mdeform'

   2008-12-03/NJW
   2008-12-09/NJW, keywords zr, over, and z_offset
*/
{
// ynclude = zaq
  extern  Acoef,                N_mirrors,            Use_mdeform, \
          Acoefarr,             Opt_module_file,      Z1_mirror, \
          Dcoef,                R1_mirror,            Z1arr, \
          Dcoefarr,             R1arr,                Z2arr, \
          Fcoef,                R2arr,                Zb1iarr, \
          Mirror_angle,         Rb1iarr,              Zb1oarr, \
          Mirror_anglearr,      Rb1oarr,              Zb2iarr, \
          Mirror_number,        Rb2iarr,              Zb2oarr, \
          Mirror_thicknessarr,  Rb2oarr,              Zfocus, \
          Modtype;
// yxclude =

   local fh, nrows;

   if( is_void(Use_mdeform) ) Use_mdeform = 0;

   if( typeof(om_file) == "string" ) {
      if( !file_test(om_file) ) {
         write,format="Opt mod file: %s does not exist\n", om_file;
         return;
      }
      dol = om_file+"+1";
      ptr = rdfitsbin( dol, fh, nrows );
      Modtype = fxpar(fh,"MODTYPE");
      r1 = *ptr(fits_colnum(fh,"r1"));
      r2 = *ptr(fits_colnum(fh,"r2"));
      z1 = *ptr(fits_colnum(fh,"z1"));
      z2 = *ptr(fits_colnum(fh,"z2"));
      rb1i = *ptr(fits_colnum(fh,"rb1i"));
      rb1o = *ptr(fits_colnum(fh,"rb1o"));
      rb2i = *ptr(fits_colnum(fh,"rb2i"));
      rb2o = *ptr(fits_colnum(fh,"rb2o"));
      zb1i = *ptr(fits_colnum(fh,"zb1i"));
      zb1o = *ptr(fits_colnum(fh,"zb1o"));
      zb2i = *ptr(fits_colnum(fh,"zb2i"));
      zb2o = *ptr(fits_colnum(fh,"zb2o"));
      mthick_arr = *ptr(fits_colnum(fh,"mthick"));
      if( Modtype == "parabolic" ) {
         Dcoefarr = *ptr(fits_colnum(fh,"dcoef"));
         Zfocus = fxpar( fh, "ZFOCUS" );
      } else if( Modtype == "hyperbolic" ) {
         Acoefarr = *ptr(fits_colnum(fh,"acoef"));
         Fcoef = fxpar( fh, "FCOEF" );
         Zfocus = fxpar( fh, "ZFOCUS" );
      } else if( Modtype == "conical" ) {
         Mirror_anglearr = *ptr(fits_colnum(fh,"mirror_angle"));
      } else error,"Unrecognized type of module";
   } else {
      om_file = Opt_module_file;
      r1 = R1arr;
      r2 = R2arr;
      z1 = Z1arr;
      z2 = Z2arr;
      rb1i = Rb1iarr;
      rb1o = Rb1oarr;
      rb2i = Rb2iarr;
      rb2o = Rb2oarr;
      zb1i = Zb1iarr;
      zb1o = Zb1oarr;
      zb2i = Zb2iarr;
      zb2o = Zb2oarr;
      mthick_arr = Mirror_thicknessarr;
      nrows = N_mirrors;
   }

   if( Use_mdeform && is_void(phi) ) {
      write,"MT_MIRDIAG Warning: phi set to zero!";
      phi = 0.0;
   }
   if( is_void(z_offset) ) z_offset = 0.0;
   rmin = min(min(r1),min(r2));
   rmax = max(max(r1),max(r2));
   if( is_void(rr) ) rr = [rmin - 0.05*(rmax-rmin), rmax + 0.05*(rmax-rmin)];
   zmin = min(min(zb2i),min(zb2o));
   zmax = max(max(zb1i),max(zb1o));
   if( is_void(zr) ) zr = [zmin - 0.05*(zmax-zmin), zmax + 0.05*(zmax-zmin)];
   if( !over ) {
      //+ window,0,style="boxed.gs";
      ttl = esc_underscore(om_file);
      if( Use_mdeform ) ttl += swrite(format=", phi = %5.3f",zero2pi(phi));
      plot,[0],xr=rr, yr=zr+z_offset, title=ttl, \
             xtitle="Radius [mm]", ytitle="Z-coordinate [mm]";
   }

   if( Use_mdeform ) { // set the name of the radius function
      if( Modtype == "parabolic" ) {
         funcname = rpar;
      } else if( Modtype == "hyperbolic" ) {
         funcname = rhyp;
      } else if( Modtype == "conical" ) {
         funcname = rcon;
      } else error,"Bad Modtype string";
   }

   for( mir = 1; mir <= nrows; mir++ ) {
      Mirror_number = mir; // for communication with rpar, rhyp, or rcon functions
      // The nominal mirror itself:
      oplot,[r1(mir),r2(mir),r2(mir)+mthick_arr(mir),r1(mir)+mthick_arr(mir),r1(mir)], \
            [z1(mir),z2(mir),z2(mir),z1(mir),z1(mir)]+z_offset,li=1+Use_mdeform;
      xyouts,r1(mir)+0.5*mthick_arr(mir),z1(mir)+z_offset+0.02*(z1(mir)-z2(mir)),itoa(mir),align=0.5,charsize=0.5;
      if( Use_mdeform ) { // Add the 'real' mirror
         if( Modtype == "parabolic" ) {
            Dcoef = Dcoefarr(mir);
         } else if( Modtype == "hyperbolic" ) {
            Acoef = Acoefarr(mir);
         } else if( Modtype == "conical" ) {
            Mirror_angle = Mirror_anglearr(mir);
            R1_mirror = r1(mir);
            Z1_mirror = z1(mir);
         }
         zarr = span(z1(mir),z2(mir),200);
         rarr = array(double,200);
         for( j = 1; j <= 200; j++ ) rarr(j) = funcname( zarr(j), phi );
         oplot,rarr,zarr+z_offset;
      }
      if( gl ) { // add guidelines
         oplot,[r2(mir),r2(mir)],[z1(mir),z2(mir)]+z_offset,li=2,color="yellow";
         oplot,[r1(mir),r1(mir)]+mthick_arr(mir),[z1(mir),z2(mir)]+z_offset,li=2,color="magenta";
      }
      // Upper baffle for current mirror
      if( mir == 1 ) { // add one if first mirror
         oplot,[0,rb1i(mir)],[zb1i(mir),zb1i(mir)]+z_offset,thick=3,color="red";
      }
      //+ oplot,[rb1o(mir-1),rb1i(mir)],[zb1o(mir-1),zb1i(mir)]+z_offset,thick=3,color="red";
      if( mir < nrows ) {
         oplot,[rb1o(mir),rb1i(mir+1)],[zb1o(mir),zb1i(mir+1)]+z_offset,thick=3,color="red";
      } else {
         oplot,[rb1o(mir),r1(mir)+mthick_arr(mir)],[zb1o(mir),zb1o(mir)],thick=3,color="red";
      }
      // Lower baffle for current mirror
      if( mir == 1 ) { // add one if first mirror
         oplot,[0,rb2i(mir)],[zb2i(mir),zb2i(mir)]+z_offset,thick=3,color="blue";
      }
      //+ oplot,[rb2o(mir-1),rb2i(mir)],[zb2o(mir-1),zb2i(mir)]+z_offset,thick=3,color="blue";
      if( mir < nrows ) {
         oplot,[rb2o(mir),rb2i(mir+1)],[zb2o(mir),zb2i(mir+1)]+z_offset,thick=3,color="blue";
      } else {
         oplot,[rb2o(mir),r2(mir)+mthick_arr(mir)],[zb2o(mir),zb2o(mir)],thick=3,color="blue";
      }

   }
}

/* Function mt_upd_om */

func mt_upd_om( void )
/* DOCUMENT mt_upd_om

   Update data for an optical module in memory after
   a call of 'mt_load'

   2008-12-03/NJW
*/
{
// ynclude = zar
  extern  N_mirrors,  R2arr,      Rb1oarr,    Rb2oarr,    Zb1oarr,    Zb2oarr, \
          R1arr,      Rb1iarr,    Rb2iarr,    Zb1iarr,    Zb2iarr;
// yxclude =

   write,"Commands: czu - change z for upper baffles";
   write,"          czl - change z for lower baffles";
   write,"          csu - change slits for upper baffles";
   write,"          csl - change slits for lower baffles";

   command = "";
   read,prompt="Enter command : ... ", command;

   if( command == "czu" ) {
      delta = 0.0;
      read,prompt="Enter delta for upper baffles [mm] : ... ", delta;
      Zb1iarr += delta;
      Zb1oarr += delta;
      kwds_set,"CZU_PAR", delta,"[mm] Change Z of upper baffles";
   } else if( command == "czl" ) {
      delta = 0.0;
      read,prompt="Enter delta for lower baffles [mm] : ... ", delta;
      Zb2iarr += delta;
      Zb2oarr += delta;
      kwds_set,"CZL_PAR", delta,"[mm] Change Z of lower baffles";
   } else if( command == "csu" ) {
      write,"Define a slit opening width as a fraction in percent of the";
      write,"mirror separation and the position of the slit again as a";
      write,"percentage fraction of the mirror separation";
      frac = 0.0;
      pos  = 0.0;
      read,prompt="Enter percentage opening for upper baffles : ... ", frac;
      frac /= 100.;
      read,prompt="Enter percentage slit position upper baffles : ... ", pos;
      pos /= 100.;
      d = R1arr - R2arr;
      p = R2arr + pos*d;
      Rb1oarr = p + 0.5*frac*d;
      Rb1iarr = p - 0.5*frac*d;
      kwds_set,"CSU_PAR1", frac,"Fractional opening of upper baffle slits";
      kwds_set,"CSU_PAR2", pos,"Fractional position of upper baffle slits";
   } else if( command == "csl" ) {
      frac = 0.0;
      read,prompt="Enter percentage for lower baffles : ... ", frac;
      frac /= 100.;
      for( mir = 1; mir < N_mirrors; mir++ ) {
         opening = Rb2oarr(mir) - Rb2iarr(mir);
         Rb2iarr(mir) += frac*opening/2;
         Rb2oarr(mir) -= frac*opening/2;
      }
      kwds_set,"CSU_PAR", frac,"Fractional change upper baffle slits";
   } else {
      write,"Unknown command - quit";
      return;
   }
   kwds_set,"HISTORY","Updated with mt_upd_om";

}
/* Function mt_write_om */

func mt_write_om( out_file, updn= )
/* DOCUMENT mt_write_om, out_file, updn=

   Get data for an optical module in memory after
   a call of 'mt_load' and write to a (new) optical
   module file keeping the original keywords.

   If 'out_file' is not given then a (suitable) name
   will be constructed.

   Keyword 'updn' (Update name) will update external variable
                  'Om_files(updn)' with name of new file.

   2008-12-03/NJW
*/
{

// ynclude = zas
  extern  Mirror_anglearr,  R1arr,            Rb2iarr,          Zb1iarr, \
          Mirror_coating,   R2arr,            Rb2oarr,          Zb1oarr, \
          Om_files,         Rb1iarr,          Z1arr,            Zb2iarr, \
          Opt_module_file,  Rb1oarr,          Z2arr,            Zb2oarr;
// yxclude =

   if( is_void(out_file) ) {
      pos = strpos( Opt_module_file, ".fits", 0, rev=1 );
      if( pos ) {
         if( is_digit(strpart( Opt_module_file,pos-3:pos-1)) \
              && strpart(  Opt_module_file,pos-4:pos-4 ) == "_" ) pos -= 4;
         out_file = get_next_filename( \
          strpart(Opt_module_file,1:pos-1)+"_???.fits");
      } else out_file = get_next_filename( Opt_module_file+"_???" );
   }

   kwds_set,"HISTORY","Written by mt_write_om";
   kwds_set,"DATE",ndate(3),"Date/time of file creation";
   wrmfitscols, out_file, "R1",R1arr,"R2",R2arr,"Z1",Z1arr,"Z2",Z2arr, \
       "MIRROR_ANGLE",Mirror_anglearr,"MLENGTH",Z1arr-Z2arr,"COATING", \
        Mirror_coating, "MTHICK", \
       "RB1I", Rb1iarr, "RB1O", Rb1oarr, "RB2I", Rb2iarr, "RB2O", Rb2oarr, \
       "ZB1I", Zb1iarr, "ZB1O", Zb1oarr, "ZB2I", Zb2iarr, "ZB2O", Zb2oarr, \
        clobber=1;
   write,format="Memory OM content written to: %s\n", out_file;
   if( updn ) {
      Om_files(updn) = out_file;
      write,format="Om_files(%i) changed to: %s\n", updn, out_file;
   }
}

/* Function mt_propagate */

func mt_propagate( znew, iphot )
/* DOCUMENT mt_propagate, znew, iphot

   Propagates photon number 'iphot' to the new z-value
   if 'iphot' is given, else it will propagate all
   photons in memory i.e. struct Phs to a plane
   with given z value (znew).

   2008-12-03/NJW
*/
{
   if( is_void(iphot) ) {
      cE = Phs.E;
      cR = Phs.R;
      t = (znew - cE(3,)) / cR(3,);
      cE += t(-:1:3,)*cR;
      Phs.E = cE;
   } else {
      cE = Phs(iphot).E;
      cR = Phs(iphot).R;
      t = (znew - cE(3)) / cR(3);
      cE += t*cR;
      Phs(iphot).E = cE;
   }
}

/* Function _propa */

func _propa( startpos, direc, endz )
/* DOCUMENT endpos = _propa( startpos, direc, endz )

     startpos is a 3-elem array with initial position
     direc    is a 3-elem array with direction
     endz     is a scalar with final z position (3. dim)

     Two first arrays may be  of dimension 3 x N

     Note that backwards movement is silently accepted.

*/
{
   dms = dimsof(startpos);
   ems = dimsof(direc);
   if( anyof(dms-ems) ) error,"Imcompatible arrays";
   if( dms(2) != 3 ) error,"First dimension must be 3";
   if( dms(1) == 1 ) { // single vector
      if( direc(3) == 0.0 ) return []; // never going to make it
      t = (endz - startpos(3))/direc(3);
      return startpos + t*direc;
   } else {  // array of vectors
      if( anyof(direc(3,) == 0.0 ) ) {
         write,"Warning - one or more zero z-directions";
         return [];
      }
      t = ((endz - startpos(3,))/direc(3,))(-:1:3,);
      return startpos + t*direc
   }
}

/* Function mt_drayplot */

func mt_drayplot( iphot, rr=, over= )
/* DOCUMENT mt_drayplot, iphot, rr=, over=

   Plots the ray on both modules

   Keywords
     'over' for overplotting
       'rr' for radius range

   Version 1.3 2008-12-04/NJW
   Version 3.3 2011-03-10/NJW
   Updated to version 4.0  2011-08-25/NJW
*/
{

// ynclude = zat
  extern  N_mirrors,    R1arr,        Rb2iarr,      Z_reference,  Zb2iarr, \
          Num_modules,  R2arr,        Z1arr,        Zb1iarr,      Zb2oarr, \
          Om_files,     R2arr,        Z2arr,        Zb1oarr,      Zfocusarr, \
          Phs;
// yxclude =
  /*
   *   The coordinate system issue
   *   Presented (?) in the telescope system
   */

   if( Num_modules != 2 ) {
      write,"This function only works when there are two modules defined.";
      return;
   }
   phs = Phs(iphot);
   if( phs.mirror == 0 ) {
      write,"No mirror has been selected - quit";
      return;
   }
   mir = phs.mirror;

   // Deal with upper optical module

   mt_load,omfile=Om_files(1);
   // define the plotting range upper value
   rrflag = 0;
   if( !is_void(rr) ) {
      if( numberof(rr) != 2 ) error,"'rr' must have 2 values";
      rrflag = 1;
   }
   if( !rrflag) {
      rr = array(double,2);
      step = mir == N_mirrors ? abs(R1arr(mir) - R1arr(mir-1)) : abs(R1arr(mir+1) - R1arr(mir));
      rr(2) = max(R1arr(mir),R2arr(mir)) + 0.5*step;
   }
   z_offset = Z_reference(1) - Z_reference(2) - Zfocusarr(2);
   // set the local maximum in Z
   zmax = max(Zb1iarr(mir),Zb1oarr(mir),Z1arr(mir)) + z_offset;
   // set the local minimum in Z
   zmin_local = min(Zb2iarr(mir),Zb2oarr(mir),Z2arr(mir));
   exit_aperture = Z_reference(1) + zmin_local;

   // Deal with lower optical module

   mt_load,omfile=Om_files(2);
   if( !rrflag ) {
      // define the plotting range lower value
      rr(1) = min(Rb2iarr(mir),Rb1iarr(mir)) - 0.5*step;
   }
   z_offset =  - Zfocusarr(2);
   // set the local minimum in Z
   zmin = min(Zb2iarr(mir),Zb2oarr(mir),Z2arr(mir)) + z_offset;
   // set the local maximum in Z
   zmax_local = max(Zb1iarr(mir),Zb1oarr(mir),Z1arr(mir));
   dz = zmax - zmin;
   zr = [zmin-0.05*dz, zmax+0.05*dz];
   entrance_aperture = Z_reference(2) + zmax_local;

   gap = exit_aperture - entrance_aperture;
   xtend = gap > 0.5 ? 0.5*gap : 0.0;

   if( !over ) {
      //+ window,0,style="boxed.gs";
      plot,[0],xr=rr,yr=zr,xtitle="Radius [mm]", ytitle="Z from focal plane [mm]";
   }


   z_offset = Z_reference(1) - Z_reference(2) - Zfocusarr(2);
   mt_rayplot, iphot, 1, z_offset=z_offset, over=1, xtend=-xtend;


   z_offset =  - Zfocusarr(2);
   mt_rayplot, iphot, 2, z_offset=z_offset, over=1, xtend=xtend;

}

/* Function mt_rayplot */

func mt_rayplot( iphot, mod_num, over=, z_offset=, xtend=, rr= )
/* DOCUMENT mt_rayplot, iphot, mod_num, over=, z_offset=, xtend=, rr=

   Plots the ray on the appropriate module section

   Keyword  'over' : Overplot
            'z_offset' : Additive coordinate change in Z
            'xtend' : Extend the ingoing (xtend > 0) or outgoing ray (xtend < 0)
            'rr' : Plot range for radius

   Version 1.5 2008-12-04/NJW
   Updated to version 4.0  2011-08-25/NJW
*/
{

// ynclude = zau
  extern  Mirror_deform_files,  Om_files,             R1arr, \
          N_mirrors,            Phs,                  Use_mdeform;
// yxclude =

   phs = Phs(iphot);
   if( phs.mirror == 0 ) {
      write,"No mirror has been selected - quit";
      return;
   }

   if( is_void(z_offset) ) z_offset = 0.0;
   if( is_void(mod_num) ) mod_num = 1;

   status = phs.status;
   bounce = phs.bounce;
   mir = phs.mirror;
   mt_load,omfile=Om_files(mod_num);
   if( Use_mdeform ) mt_load,mdeffile=Mirror_deform_files(mod_num);
   if( numberof(rr) != 2 ) { // set default values
      rr = mir == 1 ? [R1arr(mir) - 3, R1arr(mir) + 2] : [R1arr(mir-1) - 2, R1arr(mir) + 2];
   }
   // determine azimuth, use E1 if no bounce, I1 if bounce
   x = phs.E1(1); y = phs.E1(2);
   if( bounce%2 == 1 ) { x = phs.I1(1); y = phs.I1(2); }
   phi = zero2pi(atan(y,x));
   xyouts,0.2,0.87,swrite(format="Phi: %8.5f rad",phi),charsize=1.,ndc=1;
   mt_mirdiag, rr=rr, over=over, z_offset=z_offset, phi=phi;

   if( mod_num == 1 ) { // first optical module
     /*
      *    Did it bounce ?
      */
      if( bounce%2 == 1 ) { // yes - it did

         // General: draw from entrance to interaction
         _odraw_ab, phs.E1, phs.D1, phs.I1(3), z_offset=z_offset;

         if( status == 0 || status > 100 ) { // valid path through first module
            // draw from interaction to exit
            _odraw_ab, phs.I1, phs.D2, Zb2oarr(mir), z_offset=z_offset;

         } else if( status == 1 ) { // blocked by entrance spoke
            error,"##31##";
         } else if( status == 2 ) { // blocked by entrance baffle, inner
            error,"##32##";
         } else if( status == 3 ) { // blocked by entrance baffle, outer
            error,"##33##";
         } else if( status == 4 ) { // blocked by next mirror edge
            error,"##34##";
         } else if( status == 5 ) { // blocked by mirror edge
            error,"##35##";
         } else if( status == 6 ) { // blocked by next mirror's backside
            // draw from interaction to exit
            _odraw_ab, phs.I1, phs.D2, Zb2oarr(mir), z_offset=z_offset;
         } else if( status == 7 ) { // blocked by second reflection on same mirror
            // draw from interaction to exit
            _odraw_ab, phs.I1, phs.D2, Zb2oarr(mir), z_offset=z_offset;
         } else if( status == 8 ) { // blocked by "behind the mirror"
            // draw from interaction to exit
            _odraw_ab, phs.I1, phs.D2, Zb2oarr(mir), z_offset=z_offset;
         } else if( status == 9 ) { // blocked by exit baffle, inner
            // draw from interaction to exit
            _odraw_ab, phs.I1, phs.D2, Zb2oarr(mir), z_offset=z_offset;
         } else if( status == 10 ) { // blocked by exit baffle, outer
            // draw from interaction to exit
            _odraw_ab, phs.I1, phs.D2, Zb2oarr(mir), z_offset=z_offset;
         } else if( status == 11 ) { // blocked by exit spoke
            // draw from interaction to exit
            _odraw_ab, phs.I1, phs.D2, Zb2oarr(mir), z_offset=z_offset;
         } else error,"##36##";

      } else {  //    No bounce situation

         if( status == 0 || status > 100 ) { // valid path through first module
            // draw from entrance to exit
            _odraw_ab, phs.E1, phs.D2, Zb2oarr(mir), z_offset=z_offset;

         } else if( status == 1 ) { // blocked by entrance spoke
            oplot,[sqrt(phs.E1(1)^2+phs.E1(2)^2)],[phs.E1(3)]+z_offset,ps=22,symsize=0.5;
         } else if( status == 2 ) { // blocked by entrance baffle, inner
            oplot,[sqrt(phs.E1(1)^2+phs.E1(2)^2)],[phs.E1(3)]+z_offset,ps=23,symsize=0.5;
         } else if( status == 3 ) { // blocked by entrance baffle, outer
            oplot,[sqrt(phs.E1(1)^2+phs.E1(2)^2)],[phs.E1(3)]+z_offset,ps=24,symsize=0.5;
         } else if( status == 4 ) { // blocked by next mirror edge
            oplot,[sqrt(phs.E1(1)^2+phs.E1(2)^2)],[phs.E1(3)]+z_offset,ps=25,symsize=0.5;
         } else if( status == 5 ) { // blocked by mirror edge
            oplot,[sqrt(phs.E1(1)^2+phs.E1(2)^2)],[phs.E1(3)]+z_offset,ps=26,symsize=0.5;
         } else if( status == 6 ) { // blocked by next mirror's backside
            // draw from entrance to exit
            _odraw_ab, phs.E1, phs.D2, Zb2oarr(mir), z_offset=z_offset;
         } else if( status == 7 ) {
            error,"##37##";
         } else if( status == 8 ) { // blocked by "behind the mirror"
            // draw from entrance to exit
            _odraw_ab, phs.E1, phs.D2, Zb2oarr(mir), z_offset=z_offset;
         } else if( status == 9 ) { // blocked by exit baffle, inner
            // draw from entrance to exit
            _odraw_ab, phs.E1, phs.D2, Zb2oarr(mir), z_offset=z_offset;
         } else if( status == 10 ) { // blocked by exit baffle, outer
            // draw from entrance to exit
            _odraw_ab, phs.E1, phs.D2, Zb2oarr(mir), z_offset=z_offset;
         } else if( status == 11 ) { // blocked by exit spoke
            // draw from entrance to exit
            _odraw_ab, phs.E1, phs.D2, Zb2oarr(mir), z_offset=z_offset;
         } else error,"##38##";

      }


   } else if( mod_num == 2 ) { // second optical module
     /*
      *    Did it bounce ?
      */
      if( bounce/2 == 1 ) { // yes - it did

         // General: draw from entrance to interaction
         _odraw_ab, phs.E2, phs.D2, phs.I2(3), z_offset=z_offset;

         if( status == 0 ) { // valid path through second module
            // draw from interaction to exit
            _odraw_ab, phs.I2, phs.R, Zb2oarr(mir), z_offset=z_offset;

         } else if( status == 101 ) { // blocked by entrance spoke
            error,"##39##";
         } else if( status == 102 ) { // blocked by entrance baffle, inner
            error,"##40##";
         } else if( status == 103 ) { // blocked by entrance baffle, outer
            error,"##41##";
         } else if( status == 104 ) { // blocked by next mirror edge
            error,"##42##";
         } else if( status == 105 ) { // blocked by mirror edge
            error,"##43##";
         } else if( status == 106 ) { // blocked by next mirror's backside
            // draw from interaction to exit
            _odraw_ab, phs.I2, phs.R, Zb2oarr(mir), z_offset=z_offset;
         } else if( status == 107 ) { // blocked by second reflections on same mirror
            // draw from interaction to exit
            _odraw_ab, phs.I2, phs.R, Zb2oarr(mir), z_offset=z_offset;
         } else if( status == 108 ) { // blocked by "behind the mirror"
            // draw from interaction to exit
            _odraw_ab, phs.I2, phs.R, Zb2oarr(mir), z_offset=z_offset;
         } else if( status == 109 ) { // blocked by exit baffle, inner
            // draw from interaction to exit
            _odraw_ab, phs.I2, phs.R, Zb2oarr(mir), z_offset=z_offset;
         } else if( status == 110 ) { // blocked by exit baffle, outer
            // draw from interaction to exit
            _odraw_ab, phs.I2, phs.R, Zb2oarr(mir), z_offset=z_offset;
         } else if( status == 111 ) { // blocked by exit spoke
            // draw from interaction to exit
            _odraw_ab, phs.I2, phs.R, Zb2oarr(mir), z_offset=z_offset;
         } else error,"##44##";

      } else {  //    No bounce situation

         if( status == 0 ) { // valid path through second module
            // draw from entrance to exit
            _odraw_ab, phs.E2, phs.D2, Zb2oarr(mir), z_offset=z_offset;

         } else if( status == 101 ) { // blocked by entrance spoke
            oplot,[sqrt(phs.E2(1)^2+phs.E2(2)^2)],[phs.E2(3)]+z_offset,ps=22,symsize=0.5;
         } else if( status == 102 ) { // blocked by entrance baffle, inner
            oplot,[sqrt(phs.E2(1)^2+phs.E2(2)^2)],[phs.E2(3)]+z_offset,ps=23,symsize=0.5;
         } else if( status == 103 ) { // blocked by entrance baffle, outer
            oplot,[sqrt(phs.E2(1)^2+phs.E2(2)^2)],[phs.E2(3)]+z_offset,ps=24,symsize=0.5;
         } else if( status == 104 ) { // blocked by next mirror edge
            oplot,[sqrt(phs.E2(1)^2+phs.E2(2)^2)],[phs.E2(3)]+z_offset,ps=25,symsize=0.5;
         } else if( status == 105 ) { // blocked by mirror edge
            oplot,[sqrt(phs.E2(1)^2+phs.E2(2)^2)],[phs.E2(3)]+z_offset,ps=26,symsize=0.5;
         } else if( status == 106 ) { // blocked by next mirror's backside
            // draw from entrance to exit
            _odraw_ab, phs.E2, phs.R, Zb2oarr(mir), z_offset=z_offset;
         } else if( status == 107 ) { // blocked by second reflection on same mirror
            // should not happen
            error,"##45##";
         } else if( status == 108 ) { // blocked by "behind the mirror"
            // draw from entrance to exit
            _odraw_ab, phs.E2, phs.R, Zb2oarr(mir), z_offset=z_offset;
         } else if( status == 109 ) { // blocked by exit baffle, inner
            // draw from entrance to exit
            _odraw_ab, phs.E2, phs.R, Zb2oarr(mir), z_offset=z_offset;
         } else if( status == 110 ) { // blocked by exit baffle, outer
            // draw from entrance to exit
            _odraw_ab, phs.E2, phs.R, Zb2oarr(mir), z_offset=z_offset;
         } else if( status == 111 ) { // blocked by exit spoke
            // draw from entrance to exit
            _odraw_ab, phs.E2, phs.R, Zb2oarr(mir), z_offset=z_offset;
         } // else draw nothing since it never entered the second module

      }

   } else error,"##46##";



}

/* Function _odraw_ab */

func _odraw_ab( startpos, direc, endz, z_offset=, mark= )
/* DOCUMENT _odraw_ab, startpos, direc, endz, z_offset=, mark=

     Overdraw from start position to end position. The latter is found
     by the interception pos(3) == endz.
*/
{
   if( is_void(xtend) ) xtend = 0.0;
   if( is_void(z_offset) ) z_offset = 0.0;
   endpos = _propa( startpos, direc, endz );
   // rotate around z-axis
   rotangle = atan(endpos(2),endpos(1));
   rotmat = [[cos(rotangle),-sin(rotangle),0.],[sin(rotangle),cos(rotangle),0.],[0.,0.,1.]];
   prot = rotmat(,+)*endpos(+);
   r2 = prot(1);
   z2 = prot(3);
   if(mark)oplot,[r2],[z2]+z_offset,ps=13,symsize=0.5;
   prot = rotmat(,+)*startpos(+);
   r1 = prot(1);
   z1 = prot(3);
   if(mark)oplot,[r2],[z2]+z_offset,ps=22,symsize=0.5;
   oplot, [r1,r2],[z1,z2]+z_offset,color="green";
}

/* Function _odraw_ray */

func _odraw_ray( p_refl, ray1, ray2, z_offset=, xtend=, mark= )
/* DOCUMENT _odraw_ray, p_refl, ray1, ray2, z_offset=, xtend=, mark=
*/
{
   if( is_void(xtend) ) xtend = 0.0;
   if( is_void(z_offset) ) z_offset = 0.0;
   p = p_refl; // point of reflection or absorption in 3D
   rotangle = atan(p(2),p(1));
   rotmat = [[cos(rotangle),-sin(rotangle)],[sin(rotangle),cos(rotangle)]];
   pxy = p(1:2); // reduced to 2D for matrix multiplication
   prot = rotmat(,+)*pxy(+);
   r = prot(1);
   z = p(3);
   if(mark)oplot,[r],[z]+z_offset,ps=13,symsize=0.5;

   // vector pointing back

   z1 = xtend > 0.0 ? Z1arr(m) + xtend : Z1arr(m);
   t = (z1 - z)/(-ray1(3));
   p1 = p + t*(-ray1); // point at aperture, original coord. syst.
   pxy = p1(1:2);
   prot = rotmat(,+)*pxy(+);
   r1 = prot(1);
   z1 = p1(3);
   oplot,[r,r1],[z,z1]+z_offset,color="green";

   // vector pointing forward

   if( !is_void(ray2) ) {

      z2 = xtend < 0.0 ? Z2arr(m) + xtend : Z2arr(m);
      t = (z2 - z)/ray2(3);
      p2 = p + t*ray2;
      pxy = p2(1:2);
      prot = rotmat(,+)*pxy(+);
      r2 = prot(1);
      z2 = p2(3);
      oplot,[r,r2],[z,z2]+z_offset,color="green";

   }
}

/* Function mt_create_om_con4 */

func mt_create_om_con4( a, filename= )
/* DOCUMENT mt_create_om_con4, filename=

   Create Optical Module of type conical 1-alpha
   - constant length mirrors
   - Om_parameter controlled leak for on-axis radiation

   The photons move almost along the z-axis in the
   'negative direction' i.e. direction vector ~(0,0,-1).

   The reference surface (origin of z-axis) is at module
   entrance plane.

   Updated to version 4.0  2011-08-25/NJW
*/
{

// ynclude = zav
  extern  Mirror_angle,            R1arr,                   Z1_setups, \
          Mirror_lengths,          R2arr,                   Z2_setups, \
          Mirror_thickness_files,  R_outer,                 Z_reference, \
          Om_parameters,           Version,                 Zfocus, \
          R1_mirror, Telescop,     Z1_mirror, Instrume,     Zfocusarr;
// yxclude =

   if( is_void(filename) ) filename = get_next_filename("om_con4_???.fits");

   mirror_length = Mirror_lengths(1);
   if( is_void(Z1_setups) ) error,"mt_create_om_con4: Z1_setups is missing";
   if( is_void(Z2_setups) ) error,"mt_create_om_con4: Z2_setups is missing";
   z1_setup = Z1_setups(1);
   z2_setup = Z2_setups(1);
   if( abs(mirror_length-(z1_setup-z2_setup)) > 0.2 ) error,"mt_create_om_con4: mismatch";

  /*
   * The reference plane is placed inside the module
   * so that z1_setup > 0 and z2_setup < 0. Mirror length = z1_setup - z2_setup
   *
   * Define mirrors from outside inwards
   *
   *  The photons meets first z1 at entrance and then z2 at exit
   *
   *    r1 is the radius of the reflecting surface at z1
   *    r2 is the radius of the reflecting surface at z2
   *    Therefore: z2 < z1 and r2 < r1
   *
   *  Baffle slits. Conditions are:
   *          At z = zb1i :  r > rb1i
   *      and at z = zb1o :  r < rb1o
   *      and at z = zb2i :  r > rb2i
   *      and at z = zb2o :  r < rb2o
   *
   *
   *  The mirror section defined with this code has constant
   *  mirror lengths.
   *  The aperture is to be exploited
   *  as well as possible.
   *
   */

   Zfocus = Zfocusarr(1); // mm, z-coordinate of focal point in module coord. syst.
   Z1_mirror = z1_setup; // mm, relative to reference plane
   r1 = R_outer;
   z1 = Z1_mirror;

  //  Read arrays in mirror thickness file

   mthick_radius = rscol(Mirror_thickness_files(1),"radius",nomem=1,silent=1);
   mthick_thick  = rscol(Mirror_thickness_files(1),"mirror_thickness",nomem=1,silent=1);

   // Mirrors are characterized by constant length

   a = r1; b = z1; c = abs(Zfocus);
   Mirror_angle = solvealpha( a, b, c);
   R1_mirror = r1;
   z2 = z1 - mirror_length;
   r2 = rcon(z2);
   r1_arr = [];
   r2_arr = [];
   z1_arr = [];
   z2_arr = [];
   mthick_arr = [];
   ma_arr = []; // for Mirror_angle
   coating = [];
   grow, r1_arr, r1;
   grow, r2_arr, r2;
   grow, z1_arr, z1;
   grow, z2_arr, z2;
   mirror_thickness = interp(mthick_thick,mthick_radius,r1);
   grow, mthick_arr, mirror_thickness;
   grow, ma_arr, Mirror_angle;
   grow, coating, 1;
   n_mirrors = 1;

   kwds_init;
   kwds_set,"ZREFRNCE",Z_reference(1),"[mm] Z coord. of refr. plane in tel.syst.";
   kwds_set,"ZFOCUS",Zfocus,"[mm] Z coordinate of focus point in opt.mod. syst.";
   kwds_set,"R_INNER",R_inner_design,"[mm] Inner radius";
   kwds_set,"R_OUTER",R_outer,"[mm] Outer radius";
   kwds_set,"Z1SETUP",z1_setup,"[mm] Position of upper edge of mirrors";
   kwds_set,"MTCKFILE",Mirror_thickness_files(1),"Mirror thickness table file";
   kwds_set,"MIRLNGTH",mirror_length,"[mm] mirror length";
   kwds_set,"DESIGN","Constant mirror length","Telescope design type";
   kwds_set,"MODTYPE","conical","Type of mirror module";
   kwds_set,"MODSTAT","Master","Defines the mirror positions";
   kwds_set,"PACKFACT", Om_parameters(1),"Loose packing factor, 1.0 means dense";
   kwds_set,"CREATOR","mt_create_om_con4","Function that created this file";
   kwds_set,"VERSION", Version, "Version of mt_rayor package";
   kwds_set,"DATE",ndate(3),"Time of file creation";
   if( !is_void(Telescop) ) kwds_set,"TELESCOP", Telescop,"Name of telescope or mission";
   if( !is_void(Instrume) ) kwds_set,"INSTRUME", Instrume,"Name of instrument";
   kwds_set,"TUNIT1","mm","Unit for column 1";
   kwds_set,"TUNIT2","mm","Unit for column 2";
   kwds_set,"TUNIT3","mm","Unit for column 3";
   kwds_set,"TUNIT4","mm","Unit for column 4";
   kwds_set,"TUNIT5","rad","Unit for column 5"; // mirror angle
   kwds_set,"TUNIT6","mm","Unit for column 6";
   kwds_set,"TUNIT7","mm","Unit for column 7";
    // column 8 is the coating number (dimensionless)
   kwds_set,"TUNIT9","mm","Unit for column 9";
   kwds_set,"TUNIT10","mm","Unit for column 10";
   kwds_set,"TUNIT11","mm","Unit for column 11";
   kwds_set,"TUNIT12","mm","Unit for column 12";
   kwds_set,"TUNIT13","mm","Unit for column 13";
   kwds_set,"TUNIT14","mm","Unit for column 14";
   kwds_set,"TUNIT15","mm","Unit for column 15";
   kwds_set,"TUNIT16","mm","Unit for column 16";


   while( r1 > R_inner_design ) {

      n_mirrors++;

      //+ r1 = r2 - mirror_thickness;
      //+ "loose" packing with factor Om_parameters(1)
      r1 = r1 - Om_parameters(1)*(r1 - r2) - mirror_thickness;
      a = r1; b = z1; c = abs(Zfocus);
      Mirror_angle = solvealpha( a, b, c);
      R1_mirror = r1;

      r2 = rcon(z2);

      // add to arrays

      grow, r1_arr, r1;
      grow, r2_arr, r2;
      grow, z1_arr, z1;
      grow, z2_arr, z2;
      mirror_thickness = interp(mthick_thick,mthick_radius,r1);
      grow, mthick_arr, mirror_thickness;
      grow, ma_arr, Mirror_angle;
      grow, coating, 1;
   }

  /*
   *  Create the baffle parameters
   */

   //+ rb1i_arr = r2_arr;
   rb1i_arr = shift(r1_arr,1) + shift(mthick_arr,1);
   // Special action for innermost baffle slit that is
   // assumed to have the same opening as its neighbor
   open_space = r1_arr(-1) - rb1i_arr(-1);
   rb1i_arr(0) = r1_arr(0) - open_space;
   rb1o_arr = r1_arr;
   zb1i_arr = z1_arr;
   zb1o_arr = z1_arr;

   rb2i_arr = shift(r2_arr,1) + shift(mthick_arr,1);
   open_space = r2_arr(-1) - rb2i_arr(-1);
   rb2i_arr(0) = r2_arr(0) - open_space;
   rb2o_arr = r2_arr;
   zb2i_arr = z2_arr;
   zb2o_arr = z2_arr;

  /*
   *  create the FITS binary table while reversing the order of the arrays
   */

   n = numberof(r1_arr);
   i = indgen(n:1:-1);
   wrmfitscols, filename, "R1",r1_arr(i),"R2",r2_arr(i),"Z1",z1_arr(i),"Z2",z2_arr(i), \
       "MIRROR_ANGLE",ma_arr(i),"MLENGTH",z1_arr(i)-z2_arr(i),"MTHICK",mthick_arr(i), "COATING",coating(i), \
       "RB1I", rb1i_arr(i), "RB1O", rb1o_arr(i), "RB2I", rb2i_arr(i), "RB2O", rb2o_arr(i), \
       "ZB1I", zb1i_arr(i), "ZB1O", zb1o_arr(i), "ZB2I", zb2i_arr(i), "ZB2O", zb2o_arr(i), \
        clobber=1;
   write,format="The conical system optical module file: %s has just been created\n", filename;
   write,format="It has %i mirrors\n", n_mirrors;
   write,format="and it might need a coating update: %s\n","mt_upd_om_coating,...";

}

/* Function mt_create_om_con5 */

func mt_create_om_con5( filename=, master= )
/* DOCUMENT mt_create_om_con5, filename=, master=

   Create Optical Module of type conical 3-alpha (or rather 2alpha+beta)
   - constant length mirrors

   This is a 'slave'-module so the mirror radii at entrance
   are defined by the the first module (the master module).
   The master module is by default given in Om_files(1) as set up
   by 'mt_setup_system' from a system defining file.
   It can be overridden by giving another one in keyword 'master'.

   The photons move almost along the z-axis in the
   'negative direction' i.e. direction vector ~(0,0,-1).

   Updated to version 4.0  2011-08-25/NJW
*/
{

// ynclude = zaw
  extern  Mirror_angle,            Om_files,                Z2_setups, \
          Mirror_anglearr,         R1_mirror,               Z_reference, \
          Mirror_lengths,          Version,                 Zfocus, \
          Mirror_thickness_files,  Z1_mirror,               Zfocusarr, \
          Num_modules,             Z1_setups, Telescop, Instrume;
// yxclude =

   if( is_void(filename) ) filename = get_next_filename("om_con_???.fits");

   mirror_length = Mirror_lengths(2);
   if( is_void(Z1_setups) ) error,"mt_create_om_con5: Z1_setups is missing";
   if( is_void(Z2_setups) ) error,"mt_create_om_con5: Z2_setups is missing";
   z1_setup = Z1_setups(2);
   z2_setup = Z2_setups(2);
   if( abs(mirror_length-(z1_setup-z2_setup)) > 0.2 ) error,"mt_create_om_con5: mismatch";

  /*
   *   Consistency check
   */

   if( is_void(Num_modules) ) {
      write,"External Num_modules not set - did you run mt_setup_system ?";
      return [];
   }
   if( numberof(Z_reference) != Num_modules ) {
      write,"Inconsistency between externals Z_reference and Num_modules";
      return [];
   }
   if( is_void(master) ) master = Om_files(1);
   if( ! file_test( master ) ) {
      write,format="Master opt. mod. file: %s was not found\n", master;
      return [];
   }

  //  Read arrays in mirror thickness file

   mthick_radius = rscol(Mirror_thickness_files(2),"radius",nomem=1,silent=1);
   mthick_thick  = rscol(Mirror_thickness_files(2),"mirror_thickness",nomem=1,silent=1);

  /*
   *! Define mirrors from inside outwards
   *
   *
   *  The photons meets first z1 at entrance and then z2 at exit
   *
   *    r1 is the radius of the reflecting surface at z1
   *    r2 is the radius of the reflecting surface at z2
   *    Therefore: z2 < z1 and r2 < r1
   *
   *  Baffle slits. Conditions are:
   *          At z = zb1i :  r > rb1i
   *      and at z = zb1o :  r < rb1o
   *      and at z = zb2i :  r > rb2i
   *      and at z = zb2o :  r < rb2o
   *
   *
   *  The mirror section defined with this code has constant mirror lengths.
   *  The aperture is to be exploited as well as possible.
   *
   */

   alpha1_arr = rdfitscol( master+"+1", "MIRROR_ANGLE" );

   // The conical 3-alpha optical module is number 2

   Zfocus = Zfocusarr(2); // mm, z-coordinate of focal point in module coord. syst.
   z1 = z1_setup;         // mm
   z2 = z1 - mirror_length;
   Z1_mirror = z1; // required by 'rcon'

   // Mirrors are characterized by constant length

   nmir = numberof( alpha1_arr );
   r1_arr = array(double,nmir);

   dz_ref_first_focus = -Zfocusarr(1) - (Z_reference(1) - Z_reference(2));

   r2_arr = array(double, nmir);
   Mirror_anglearr = array(double, nmir);
   z1_arr = array(z1, nmir);
   z2_arr = array(z2, nmir);
   for( mir = 1; mir <= nmir; mir++ ) {
      alpha = alpha1_arr(mir);
      r3 = dz_ref_first_focus*tan(2*alpha); // Radius where pilot ray at reference plane hits the mirror
      twoema = atan(r3/abs(Zfocus));
      eps = 0.5*twoema + alpha;
      Mirror_anglearr(mir) = eps;
      Mirror_angle = Mirror_anglearr(mir); // required by 'rcon'
      R1_mirror = r3 + z1*tan(eps); // required by 'rcon'
      r1_arr(mir) = R1_mirror;
      r2_arr(mir) = rcon( z2 );
   }
   coating = array(1,nmir);

   mthick_arr = interp(mthick_thick,mthick_radius,r1_arr);

   kwds_init;
   kwds_set,"FCLENGTH",abs(Zfocus),"[mm] Focal length";
   kwds_set,"ZFOCUS",Zfocus,"[mm] Z coordinate of focus point";
   kwds_set,"R_INNER",r1_arr(0),"[mm] Inner radius";
   kwds_set,"R_OUTER",r1_arr(1),"[mm] Outer radius";
   kwds_set,"Z1SETUP",z1_setup,"[mm] Position of upper mirror edges";
   kwds_set,"MODSTAT","Slave","This is dependent on the first (master) module";
   kwds_set,"MASTER", master,"The opt-mod. that acts as master";
   kwds_set,"MTCKFILE",Mirror_thickness_files(2),"Mirror thickness table file";
   kwds_set,"MIRLNGTH",mirror_length,"[mm] mirror length";
   kwds_set,"DESIGN","Constant mirror length","Telescope design type";
   kwds_set,"MODTYPE","conical","Type of mirror module";
   kwds_set,"CREATOR","mt_create_om_con5","Function that make this file";
   kwds_set,"VERSION", Version,"Version of mt_rayor package";
   kwds_set,"DATE",ndate(3),"Time of file creation";
   if( !is_void(Telescop) ) kwds_set,"TELESCOP", Telescop,"Name of telescope or mission";
   if( !is_void(Instrume) ) kwds_set,"INSTRUME", Instrume,"Name of instrument";
   kwds_set,"TUNIT1","mm","Unit for column 1";
   kwds_set,"TUNIT2","mm","Unit for column 2";
   kwds_set,"TUNIT3","mm","Unit for column 3";
   kwds_set,"TUNIT4","mm","Unit for column 4";
   kwds_set,"TUNIT5","rad","Unit for column 5"; // mirror angle
   kwds_set,"TUNIT6","mm","Unit for column 6";
   kwds_set,"TUNIT7","mm","Unit for column 7";
    // column 8 is the coating number (dimensionless)
   kwds_set,"TUNIT9","mm","Unit for column 9";
   kwds_set,"TUNIT10","mm","Unit for column 10";
   kwds_set,"TUNIT11","mm","Unit for column 11";
   kwds_set,"TUNIT12","mm","Unit for column 12";
   kwds_set,"TUNIT13","mm","Unit for column 13";
   kwds_set,"TUNIT14","mm","Unit for column 14";
   kwds_set,"TUNIT15","mm","Unit for column 15";
   kwds_set,"TUNIT16","mm","Unit for column 16";

  /*
   *  Create the baffle parameters
   */

  /*
   *  Standard definition of rb1i is backside of next mirror
   *  at z = z1
   */
   rb1i_arr = shift(r1_arr,-1) + shift(mthick_arr,-1);
   // Special action for innermost baffle slit that is
   // assumed to have the same opening as its neighbor
   open_space = r1_arr(2) - rb1i_arr(2);
   rb1i_arr(1) = r1_arr(1) - open_space;
   rb1o_arr = r1_arr;
   zb1i_arr = z1_arr;
   zb1o_arr = z1_arr;

   rb2i_arr = shift(r2_arr,-1) + shift(mthick_arr,-1);
   open_space = r2_arr(2) - rb2i_arr(2);
   rb2i_arr(1) = r2_arr(1) - open_space;
   rb2o_arr = r2_arr;
   zb2i_arr = z2_arr;
   zb2o_arr = z2_arr;

  /*
   *  create the FITS binary table
   */

   wrmfitscols, filename, "R1",r1_arr,"R2",r2_arr,"Z1",z1_arr,"Z2",z2_arr, \
       "MIRROR_ANGLE",Mirror_anglearr,"MLENGTH",z1_arr-z2_arr,"MTHICK", mthick_arr, "COATING",coating, \
       "RB1I", rb1i_arr, "RB1O", rb1o_arr, "RB2I", rb2i_arr, "RB2O", rb2o_arr, \
       "ZB1I", zb1i_arr, "ZB1O", zb1o_arr, "ZB2I", zb2i_arr, "ZB2O", zb2o_arr, \
        clobber=1;
   write,format="The conical system optical module file: %s has just been created\n", filename;
   write,format="It has %i mirrors\n", numberof(r1_arr);
   write,format="and it might need a coating update: %s\n","mt_upd_om_coating,...";

}

/* Function solvealpha */

func solvealpha ( a, b, c )
/* DOCUMENT alpha = solvealpha( a, b, c )

   Solve the equation:

   tan(2alpha) = (a - b tan(alpha))/c
      where c >> a  and c >> b

   by iterations and return the value of alpha

   2008-12-05/NJW
*/
{
   alpha = 1.0;
   alpha_prev = 0.0;

   while( abs(alpha - alpha_prev) > 1.e-10 ) {
      alpha_prev = alpha;
      tan2a = (a - b*tan(alpha_prev))/c;
      alpha = 0.5*atan(tan2a);
   }

   return alpha;
}

/* Function mt_stat */

func mt_stat( status, w= )
/* DOCUMENT s = mt_stat( status, w= )

   status is void: print statistics and return total
                         number of photons
   else
     w is void returns number of photons with given status
     else returns 'where' array

   Version 1.6   2008-12-17/NJW
   Version 3.0   2011-01-17/NJW
*/
{
   extern Phs;

   statarr = [0,1,2,3,4,5,6,7,8,9,10,11, \
              101,102,103,104,105,106,107,108,109,110,111,201];
   nstatarr = numberof(statarr);

   wstatus = Phs.status;
   s = [];
   for( i = 1; i <= nstatarr; i++ ) {
      if( numberof(where(wstatus==statarr(i))) ) grow, s, statarr(i);
   }
   statarr = s;
   nstatarr = numberof(statarr);

   if( is_void(status) ) {
      // Make condensed printout for the log file.
      s = "Stat ";
      for(i = 1; i <= (nstatarr+1)/2; i++ ) {
         s += swrite(format="%i:%i", statarr(i), numberof(where(wstatus==statarr(i))));
         if( i < (nstatarr+1)/2 ) s += ", ";
      }
      t = "Stat ";
      for(j = i; j <= nstatarr; j++ ) {
         t += swrite(format="%i:%i", statarr(j), numberof(where(wstatus==statarr(j))));
         if( j < nstatarr ) t += ", ";
      }
      mt_log_entry,2,s,t;

      // Make overview for terminal output.
      for(i = 1; i <= nstatarr; i++ ) {
         w = where(wstatus==statarr(i));
         n = numberof(w);
         write,format="Status %4i: %9i\n", statarr(i), n;
         if( statarr(i) == 0 && n > 0 ) {
            write,format="         Of these%s\n",":";
            n = numberof(where(Phs(w).bounce == 0));
            write,format="            %9i with no bounces\n", n;
            n = numberof(where(Phs(w).bounce == 1));
            write,format="            %9i with only OM1 bounce\n", n;
            n = numberof(where(Phs(w).bounce == 2));
            write,format="            %9i with only OM2 bounce\n", n;
            n = numberof(where(Phs(w).bounce == 3));
            write,format="            %9i with two bounces\n", n;
         }
      }
      return numberof(Phs);
   } else {
      if( is_void(w) ) {
         return numberof(where(Phs.status == status));
      } else {
         return where(Phs.status == status);
      }
   }
}

/* Function mt_mirror_deform */

func mt_mirror_deform( z, phi )
/* DOCUMENT  delta_r = mt_mirror_deform( z, phi )

   Returns the perturbation of the mirror due to deformations
   away from the deal shape.
   Assuming that 'z' is expressed in the local optical module
   coordinate system with origin in the reference plane.

   Computing the radius of the mirror surface

   r = r_ideal - delta_r

   2010-02-08/NJW
*/
{
// ynclude = zll
   extern Mirror_deform_arr, Mirror_length, Mirror_number;
// yxclude =

   dms = dimsof(Mirror_deform_arr);
   // Dim 1 is azimuth
   // Dim 2 is z (normalized to Mirror_length)
   // Dim 3 is number of mirror

   deform = Mirror_deform_arr(,,Mirror_number);

   phi = zero2pi(phi);
   azi_step = (2.*pi)/(dms(2)-1);
   i_azi = long(phi/azi_step) + 1;
   z_step = 1./(dms(3)-1);
   z_norm = (z - Z2arr(Mirror_number))/Mirror_length;
   i_z = long(z_norm/z_step) + 1;
   if( i_z >= dms(3) ) i_z = dms(3) - 1;
   if( i_z < 1 ) i_z = 1;

   // interpolate in azimuth
   deform = (deform(i_azi+1,)-deform(i_azi,))*(phi-azi_step*(i_azi-1))/azi_step + deform(i_azi,);
   // and then in z
   deform = (deform(i_z+1)-deform(i_z))*(z_norm-z_step*(i_z-1))/z_step + deform(i_z);

   return deform;

}

/* Function mt_mk_mdeform_file */

/*************************************************
 *
 *   Mirror deformation file generation
 *   for the MT_RAYOR package
 *
 *   2010-02-06/NJW
 *
 *   A 3D FITS image is produced with the third dimension
 *   as the mirror number. An Optical Module file is
 *   accompanied by a mirror deformation file (else "no deformation"
 *   is reported in a keyword).
 *   The first coordinate represents the azimuth angle as
 *   reported in the keywords CRPIX1, CRVAL1, CDELT1
 *   The unit must be radians.
 *   The second coordinate represents the z-value relative to the
 *   reference plane for the Optical Module. The scale given with
 *   keywords CRVAL and CDELT is normalized to -1 -> 0 since the
 *   mirror length may not be identical for all mirrors in a
 *   module. Hence a rescaling must be done in the application
 *   of the deformation values.
 *
 *   The image array value represents the deviation from the ideal
 *   surface with a negative sign.
 *
 *   Rule: With NAXISi CRPIXi = v_min, CDELTi = (v_max - v_min)/(NAXISi - 1),
 *         CRPIXi = 1  =>   v(j) = CRVALi + (j-CRPIXi)*CDELTi
 *                and     j = int((v - CRVALi)/CDELTi) + CRPIXi
 *
 *   For i == 1 (azimuth direction)
 *
 *       The first (0 radians) and last (2pi radians) values should
 *       be identical.
 *
 *   For i == 2 (along Z axis):
 *
 *       z(j) = (CRVAL2 + (j-CRPIX2)*CDELT2) * mirror_length
 *       j = int((z/mirror_length - CRVAL2)/CDELT2) + CRPIX2
 *
 ***************************************************/
func mt_mk_mdeform_file( filename, mode, param, nz=, naz=, arr=, skipinit= )
/* DOCUMENT arr = mt_mk_mdeform_file( mode, param, nz=, naz= )
     or mt_mk_mdeform_file, filename, mode, param, nz=, naz=, arr=, skipinit=

     Produce a mirror deformation datacube and/or save it to disk.
     The mirror configuration is taken from the currently loaded OM file.

     In the first form the mirror deformation array is returned for
     inspection and perhaps further manipulation.

     In the second form (called as subroutine) a file is written with
     the 3D deformation array. The array may be given by keyword 'arr'
     in which case the only action is to write it to the disk. This
     functionality gives the opportunity to define the array by e.g.:

      > my_arr = mt_mk_mdeform_file( 1, 0.002, nz=50, naz=100 );
      > ... further manipulation of 'my_arr'
     saving the result by
      > mt_mk_mdeform_file,"my_arr.fits",arr=my_arr;

     mode    param
      1      amplitude            Parabolic shape, all mirrors identical
      2      [dt_bottom, dt_top]  Uniform temperature gradient
                                  i.e. conical mirrors but with a
                                  different slope (proportional to radius)
      3      amplitude/radius     Parabolic shape, the amplitude is
                                  proportional to the radius
      4      [amplitude, wavelength, phase] Sinusoidal single frequency
                                  perturbation amp*sin(phase + (2 pi z)/wavel)

     Keywords:
       nz    Number of points in the z-direction
       naz   Number of points in the azimuth direction (0 - 2pi)
       arr   3D array for storing.
   skipinit  [bool] for skipping call of 'kwds_init' prior to file writing

   2010-02-08/NJW
   2010-06-22/NJW, updated with mode==2 temperature gradient
   2010-06-23/NJW, updated with mode==3 radius dependence
   2011-01-06/NJW, updated with mode==4 sinusoidal perturbation
*/
{
   extern N_mirrors, Z1arr, Z2arr, R1arr;

   if( is_void(nz) ) nz = 30;
   if( is_void(naz) ) naz = 5
   zarr = span(-1.0,0,nz);

   build_arr = 1;
   if( !am_subroutine() ) {
      param = mode;
      mode = filename;
      arr = []; // N/A when called as a function
   } else {
      if( numberof(arr) ) {
         dms = dimsof(arr);
         if( dms(1) != 3 ) error,"Illegal dimension of 'arr'";
         // override setting of naz and nz
         naz = dms(2);
         nz = dms(3);
         build_arr = 0;
         garr = arr;
         param = 0.0;
         mode = 99;
      }
   }

   if( build_arr ) {
      if( mode == 1 ) { // Parabolic shape with amplitude 'param'
                        // all mirrors identical
         if( numberof(param) != 1 ) error,"'param' must be a scalar";
         shape = 4*param*zarr*(zarr+1);
         garr = shape(-:1:naz,,-:1:N_mirrors); // copy to all azimuth
                                  // values and all mirrors
         kwds_set,"COMMENT","Parabolic symmetrical shape";
         kwds_set,"COMMENT","Same amplitude for all mirrors";
      } else if( mode == 2 ) { // Temperature (constant) gradient
                         // 'param' is a 2-element array with
                         // param(1) as the delta-T at bottom and
                         // param(2) as delta-T at top
                         // A negative param(i) means that radii get smaller
         if( numberof(param) != 2 ) error,"'param' must have 2 elements";
         garr = array(double,naz,nz,N_mirrors);
         alpha_Al = 23.e-6; // per degree
         for( imir = 1; imir <= N_mirrors; imir++ ) {
            auxil = -param*alpha_Al*R1arr(imir);
            shape = interp(auxil,[-1.,0.],zarr);
            garr(,,imir) = shape(-:1:naz,); // copy to all azimuth values
         }
         kwds_set,"COMMENT","Constant temperature gradient along axis";
      } else if( mode == 3 ) { // Parabolic shape with amplitude
                               // radius*param i.e. different for
                               // each mirror
         if( numberof(param) != 1 ) error,"'param' must be a scalar";
         shape = 4*param*zarr*(zarr+1);
         garr = shape(-:1:naz,,-:1:N_mirrors); // copy to all az's and mirrors
         garr *= R1arr(-:1:naz,-:1:nz,);
         kwds_set,"COMMENT","Parabolic symmetrical shape";
         kwds_set,"COMMENT","Amplitude is proportional to mirror radius";

      } else if( mode == 4 ) { // Sinusoidal perturbation
                               // param is three-element array
                               // [amplitude, wavelength, phase]
         if( numberof(param) != 3 ) error,"'param' must be a 3-element array";
         shape = param(1)*sin(param(3) + zarr(,-:1:N_mirrors)*(Z1arr-Z2arr)(-:1:nz,)*2.*pi/param(2));
         garr = shape(-:1:naz,,); // copy to all az's
         kwds_set,"COMMENT","Sinusoidal perturbation";

      } else if( mode == 5 ) { // Mirror tilt, all mirrors tilted with same delta-r
                         // 'param' is a 2-element array with
                         // param(1) as the delta-r at bottom and
                         // param(2) as delta-r at top
                         // A negative param(i) means that radii get smaller
         if( numberof(param) != 2 ) error,"'param' must have 2 elements";
         garr = array(double,naz,nz,N_mirrors);
         for( imir = 1; imir <= N_mirrors; imir++ ) {
            shape = interp(param,[-1.,0.],zarr);
            garr(,,imir) = shape(-:1:naz,); // copy to all azimuth values
         }
         kwds_set,"COMMENT","Constant tilt of all mirrors";

      } else error,"This mode is not supported";
   }

   if( am_subroutine() ) { // File writing prepared

      crval1 = 0.0; crpix1 = 1.0; cdelt1 = 2.*pi/(naz-1);
      crval2 = -1.0; crpix2 = 1.0; cdelt2 = 1.0/(nz-1);
      crval3 = 0.0; crpix3 = 1.0; cdelt3 = 1.0;

      if(!skipinit) kwds_init;
      kwds_set,"DATE",ndate(3),"Date and time of creation";
      kwds_set,"ORIGIN","mt_mk_mdeform_file","Producing software";
      kwds_set,"MODE",mode,"Parameter for deformation mode";
      if( numberof(param) == 1 ) {
         kwds_set,"PARAM",param,"Deformation parameter for applied mode";
      } else {
         for( i = 1; i <= numberof(param); i++ ) {
            kwds_set,"PARAM"+itoa(i),param(i),"Deformation parameter for applied mode";
         }
      }
      kwds_set,"RESPONSI","Niels J. Westergaard","Responsible person";
      kwds_set,"EXTNAME","MIRROR_DEFORM","Name of extension";
      kwds_set,"CRVAL1", crval1,"Reference coordinate value 1";
      kwds_set,"CRPIX1", crpix1,"Reference pixel 1";
      kwds_set,"CDELT1", cdelt1,"Increment value 1";
      kwds_set,"CRVAL2", crval2,"Reference coordinate value 2";
      kwds_set,"CRPIX2", crpix2,"Reference pixel 2";
      kwds_set,"CDELT2", cdelt2,"Increment value 2";
      kwds_set,"CRVAL3", crval3,"Reference coordinate value 3";
      kwds_set,"CRPIX3", crpix3,"Reference pixel 3";
      kwds_set,"CDELT3", cdelt3,"Increment value 3";

      writefits,filename, garr, clobber=1;

   } else {

      return garr;

   }
}
/* Function mt_mirplot */

func mt_mirplot( mirror_number, phi, yr=, nz= )
/* DOCUMENT mt_mirplot, mirror_number, phi, yr=, nz=
        res = mt_mirplot( mirror_number, phi, yr=, nz= )

   Plot the mirror surface as a function of Z in
   comparison to the straight line connecting the
   nominal mirror end radii.

   Second version - when called as a function:
      Returns a 1D array where the first half are the z-values and the second
      half the deviations from linearity.

   Keywords:    yr   plot range in y-direction
                       defaults to max(abs(deviation)) or 0.01,
                       whatever is largest
                nz   number of z-values, defaults to 100
*/
{
// ynclude = zax
  extern  Acoef,              Mirror_anglearr,    Modtype,            Use_mdeform, \
          Acoefarr,           Mirror_deform_arr,  R1_mirror,          Z1_mirror, \
          Dcoef,              Mirror_length,      R1arr,              Z1arr, \
          Dcoefarr,           Mirror_number,      R2arr,              Z2arr, \
          Mirror_angle;
// yxclude =

   if( is_void(Z1arr) ) {
      write,"No OM has been loaded - skip ...";
      return;
   }
   if( is_void(Use_mdeform) ) {
      Use_mdeform = (numberof(Mirror_deform_arr) > 0);
   }
   Mirror_number = mirror_number;
   z1 = Z2arr(mirror_number);
   z2 = Z1arr(mirror_number);
   if( Modtype == "parabolic" ) {
      funcname = rpar;
      Dcoef = Dcoefarr(Mirror_number);
   } else if( Modtype == "hyperbolic" ) {
      funcname = rhyp;
      Acoef = Acoefarr(Mirror_number);
   } else if( Modtype == "conical" ) {
      funcname = rcon;
      Mirror_angle = Mirror_anglearr(Mirror_number);
      R1_mirror = R1arr(Mirror_number);
      Z1_mirror = Z1arr(Mirror_number);
   } else error,"MT_MIRPLOT Illegal Modtype specification";

   if( is_void(nz) ) nz = 100;
   y = z = span(z1,z2,nz);
   yref = interp([R2arr(mirror_number),R1arr(mirror_number)],[z1,z2],z);

   Mirror_length = Z1arr(mirror_number) - Z2arr(mirror_number);
   for(i=1;i<=nz;i++) y(i) = funcname(z(i),double(phi));
   if( is_void(yr) ) {
      yamp = max(abs(y-yref));
      yr = (yamp < 0.01 ? 0.01 : 1.1*yamp)*[-1,1];
   }
   plot, z, yref - y,yr=yr,xtitle="Z [mm]", \
         ytitle="Line - R [mm]", title=Modtype+swrite(format=", phi = %5.3f rad",double(phi));

   if( !am_subroutine() ) {
     res = grow( z, yref - y );
     return res;
   }
}


/* Function mt_detector */

/************************************************************
 *
 *  A function to convert the photons with status==0 from
 *  the photon list (external 'Phs') to events including
 *  detection probability, finite energy resolution (by the
 *  RDM matrix), and pixellation.
 *
 *    2010-02-12/NJW
 *    2013-07-02/NJW with much better performance (speed)
 *
 *************************************************************/

func mt_detector( offset=, cont=, bkglvl=, flag= )
/* DOCUMENT mt_detector, offset=, cont=, bkglvl=, flag=

   produces an event list from the existing photon list, Phs.
   based on the detector properties as defined in a detector
   description file. The detector properties are loaded by
   function 'mt_load' with keyword 'detfile'.

   Note that the RDM plays an active role in event rejection if its
   sum over detector channels is less than 1.

   Keywords: offset  - a 2 element array [dx,dy] in mm that shifts the
                       optical axis relative to the detector center.
             cont    - will cause a catenation of the Evlist in order to
                       combine several raytracing sessions e.g. to simulate
                       a situation with more sources in the FOV such as
                       an extended source
             bkglvl  - the background level which is the number of counts
                       per pixel per keV, i.e. NOT per second!. Will be
                       added to the event list (Evlist) with a flag value
                       of zero.
             flag    - the flag value for all raytraced events (i.e. not bkg)
                       defaults to same flag value as in 'Phs'.

   SEE ALSO: mt_det_add_bkg, mt_det_image.
*/
{
// ynclude = zay
  extern  E_max,        Energ_lo,     Num_pixels2,  Qeff,         Xpixlo, \
          E_min,        Evlist,       Phs,          Rdm,          Ypixlims, \
          Energ_hi,     Num_pixels1,  Q_ener,       Xpixlims,     Ypixlo, \
          Det_offset,   Dead_pixel_map;
// yxclude =

   t0 = t1 = t2 = t3 = t4 = t5 = time_keeper = elapsed_time = array(double,3);
   tt1 = tt2 = tt3 = array(double,3);
   timer, time_keeper;
   t0 = time_keeper;

   if( is_void(offset) ) {
      Det_offset = [0.,0.];
   } else {
      if( numberof(offset) != 2 ) error,"Error in offset for mt_detector";
      Det_offset = double(offset);
   }
   if( is_void(bkglvl) ) bkglvl = 0.0;
   //+ if( is_void(flag) ) flag = 1;
   // If 'flag' is given as a keyword, then all non-bkg events will get that flag
   // Otherwise the original photon flag (source number (?)) will be copied to the events
   set_kwflag = is_void(flag) ? 0 : 1;
   //  reset the event list unless a continuation has been requested
   if(!cont) Evlist = [];


/*
 *  ------   First selection -----------------
 *
 *   The variable 'sel' follows the selections so that
 *   Phs(sel) should also be the current selection
 */
   sel = where( Phs.status == 0 );
   nphot = numberof(sel);
   if( nphot == 0 ) {
      write,"Sorry, no photons survived the status==0 selection - no action";
      return;
   }
   timer, time_keeper, t1;

/*
 *  ------   Second selection -----------------
 */
   // make a selection based on position and energy
   phs = Phs(sel);
   cE = phs.E;
   ener = phs.energy;

   // define the rectangular detector boarders
   xp = Xpixlims - Det_offset(1);
   yp = Ypixlims - Det_offset(2);

   // select the photons that fall inside the detector
   // and in the acceptable (by RMF) energy range
   w = where( cE(1,) > xp(1) & cE(1,) < xp(2) \
            & cE(2,) > yp(1) & cE(2,) < yp(2) \
            & ener > Energ_lo(min) & ener < Energ_hi(max) );
   nw = numberof(w);
   if( nw == 0 ) {
      write,"Sorry, no photons survived the 'on detector' selection - no action";
      return;
   }
   sel = sel(w);
   phs = phs(w);
   cE = phs.E;
   ener = phs.energy;
   wsel = w;
   // find quantum efficiency of the detector at all photon energies
   q = interp( Qeff, Q_ener, ener ); // now 'q' is an array
   // combine with telescope transmission coefficient
   trans = q * phs.rcoef;
   timer, time_keeper, t2;
/*
 *  ------   Third selection -----------------
 */
   // accept only if a random number is smaller than 'trans'
   r = random(nw);
   w = where( r < trans );
   nw = numberof(w);
   if( nw == 0 ) {
      write,"Sorry, no photons survived the transmission selection - no action";
      return;
   }
   wsel = wsel(w);
   sel = sel(w);
   phs = phs(w);
   cE = phs.E;
   ener = phs.energy;

   // Get the energy indices 'j'.
   bds = grow( Energ_lo, Energ_hi(0)); // assume head-to-tail
   j = digitize( ener, bds ) - 1; // since all events are within range

   // accumulate the RDM
   psu = Rdm(psum,); // is a 2D array
   ptot = psu(0,);   // pick out the end values
   timer, time_keeper, t3;

/*
 *  ------   Fourth selection -----------------
 */
   // test acceptance based on the RDM

   r = random(nw);
   w = where( r < ptot(j) ); // accepted
   nw = numberof(w);
   if( nw == 0 ) {
      write,"Sorry, no photons survived the absorption selection - no action";
      return;
   }
   wsel = wsel(w);
   sel = sel(w);
   phs = phs(w);
   j = j(w);
   cE = phs.E;
   ener = phs.energy;

   // normalize 'psu' - now the events have been accepted
   // so it is a question what energy they get assigned
   m = where(ptot > 0.0);
   if( numberof(m) == 0 ) {
      write,"Sorry, no photons survived the ptot > 0 selection - no action";
      return;
   }
   dms = dimsof(psu);
   //+ psu(,m) /= ptot(m)(-:1:dms(2),m);
   //    ---- statement changed 2016-12-15 to:
   psu(,m) /= ptot(m)(-:1:dms(2),);
   timer, time_keeper, tt1;

   // the following loop seems necessary
   // another possibility is:
   //    for( i ..) {
   //       k(i) = digitize( r(i), psu(,j(i) )
   //    }
   k = array(long,nw);
   r = random(nw);
   for( i = 1; i <= nw; i++ ) {
      k(i) = where( psu(,j(i)) > r(i) )(1); // channel number
   }
   timer, time_keeper, tt2;

   // Alternative method

   //+ is = sort(j);
   //+ js = j(is);
   //+ us = uniq(js);
   //+ nus = numberof(us);
   //+ usp = grow(us,numberof(j)+1);
   //+
   //+ for(i=1;i<=nus;i++) {
   //+    m = where( j == js(us(i)) );

   // now determine the pixel

   bds = grow(Xpixlo, 2*Xpixlo(0) - Xpixlo(-1));
   kk = digitize( cE(1,), bds - Det_offset(1) ) - 1;
   detx = cE(1,) + Det_offset(1);
   bds = grow(Ypixlo, 2*Ypixlo(0) - Ypixlo(-1));
   ll = digitize( cE(2,), bds - Det_offset(2) ) - 1;
   dety = cE(2,) + Det_offset(2);
   timer, time_keeper, t4;

/*
 *  ------   Fifth selection -----------------
 */
   // Allow only events in the active detector pixels
   idx = (kk-1) * dimsof(Dead_pixel_map)(2) + ll;
   w = where( !Dead_pixel_map(idx) );
   nw = numberof(w);
   if( nw == 0 ) {
      write,"Sorry, no photons survived the dead pixel selection - no action";
      return;
   }
   wsel = wsel(w);
   sel = sel(w);
   k = k(w);

   evt = array(s_MTEvent,nw);
   evt.rawx   = kk(w);
   evt.rawy   = ll(w);
   evt.detx   = detx(w);
   evt.dety   = dety(w);
   evt.pha    = k;
   r = random(nw);
   evt.energy = E_min(k) + r*(E_max(k) - E_min(k));
   if( set_kwflag ) {
      evt.flag   = flag;
   } else evt.flag = Phs(sel).flag;
   evt.bounce = Phs(sel).bounce;
   evt.phs_index = sel;

   grow, Evlist, evt;
   timer, time_keeper, t5;

   if( bkglvl > 0.0 ) {
      nbkg = long(bkglvl * (E_max(0) - E_min(1)) * Num_pixels1 * Num_pixels2+0.5);
      bkg = array(s_MTEvent,nbkg);
      bkg.detx = random(nbkg)*Num_pixels1 - Num_pixels1/2.;
      bkg.rawx = long(bkg.detx + Num_pixels1/2. + 1);
      bkg.dety = random(nbkg)*Num_pixels2 - Num_pixels2/2.;
      bkg.rawy = long(bkg.dety + Num_pixels2/2. + 1);
      dead = where( Dead_pixel_map );
      if( numberof(dead) ) {
         idx = (bkg.rawy - 1) * Num_pixels1 + bkg.rawx;
         idx = whereany( idx, dead );
         bkg = rem_elem( bkg, idx );
         nbkg = numberof( bkg );
      }
      bkg.flag = 0;
      bkg.phs_index = -2;
      bkg.energy = random(nbkg)*(E_max(0) - E_min(1)) + E_min(1);
      for( i = 1; i <= nbkg; i++ ) {
         w = where( E_min < bkg(i).energy );
         bkg(i).pha = numberof(w) ? w(0) : 1;
      }
      grow, Evlist, bkg;
   }

   timer, t0, elapsed_time;
   mt_log_entry,1,swrite(format="CPU time for mt_detector was: %.3f s", elapsed_time(1));
   write,format="CPU time for mt_detector was: %.3f s\n", elapsed_time(1);
   //+ write,format="CPU time for selection 1 was: %.3f s\n", t1(1);
   //+ write,format="CPU time for selection 2 was: %.3f s\n", t2(1);
   //+ write,format="CPU time for selection 3 was: %.3f s\n", t3(1);
   //+ write,format="CPU time for RDM         was: %.3f s\n", tt1(1);
   //+ write,format="CPU time for event loop  was: %.3f s\n", tt2(1);
   //+ write,format="CPU time for selection 4 was: %.3f s\n", t4(1);
   //+ write,format="CPU time for selection 5 was: %.3f s\n", t5(1);

}

/* Function mt_det_image */

/************************************************************
 *
 *  A function to make a detector image in a specified energy
 *  interval
 *
 *    2010-02-12/NJW
 *
 *************************************************************/

func mt_det_image( void, emin=, emax=, outfile=, bkglvl=, flagsel= )
/* DOCUMENT mt_det_image, emin=, emax=, outfile=, bkglvl=, flagsel=
       or   image = mt_det_image( emin=, emax=, outfile=, bkglvl=, flagsel= )

   Forms a detector image based on the event list Evlist
          and the detector size as defined by a call of 'mt_detector'.

   Called as subroutine: displays image in a plotting window.
   Called as function  : returns the image


   Keywords: emin    - minimum energy, defaults to E_min(1)
             emax    - maximum energy, defaults to E_max(0)
             outfile - name of output file for FITS image
             bkglvl  - average number of background counts per pixel
                       and per keV i.e. NOT per second will be
                       randomly distributed in position
             flagsel - selects the indicated 'flag' value


*/
{
// ynclude = zaz
  extern  Dec_scx,       Evlist,        Num_pixels1,   Pixel_size1,   Ra_scx, \
          E_max,         Exposure,      Num_pixels2,   Posang,        Version, \
          E_min,         Focal_length,  Det_offset,    Pixel_size2;
// yxclude =

   nevs = numberof(Evlist);
   if( nevs == 0 ) {
      write,"No events in event list - skip";
      return [];
   }
   if( is_void(emin) ) emin = E_min(1);
   if( is_void(emax) ) emax = E_max(0);

   sel = where( Evlist.energy >= emin & Evlist.energy <= emax );
   if( (nsel = numberof(sel)) == 0 ) {
      write,"No events survived the energy selection - skip";
      return [];
   }
   if( !is_void(flagsel) ) {
      sel = where( Evlist.energy >= emin & Evlist.energy <= emax & Evlist.flag == flagsel );
      if( (nsel = numberof(sel)) == 0 ) {
         write,"No events survived the combined energy and flag selection - skip";
         return [];
      }
   }

   im = array(int,Num_pixels1,Num_pixels2);
   for( i = 1; i <= nsel; i++ ) im(Evlist(sel(i)).rawx,Evlist(sel(i)).rawy)++;

   // Add background events to the image if requested

   if( !is_void(bkglvl) ) {
      npix = Num_pixels1*Num_pixels2;
      nbkgcts = long(bkglvl*npix*(emax-emin)+0.5);
      x = long(random(nbkgcts)*Num_pixels1 + 1);
      y = long(random(nbkgcts)*Num_pixels2 + 1);
      for( i = 1; i <= nbkgcts; i++ ) im(x(i),y(i))++;
   } else bkglvl = 0.0;

   if( !is_void(outfile) ) {
      kwds_init;
      kwds_set,"DATE",ndate(3),"Date and time of creation";
      kwds_set,"ORIGIN","mt_rayor(mt_det_image)","Software used";
      kwds_set,"VERSION",Version,"Version of mt_rayor";
      kwds_set,"BUNIT","cts/pixel","Unit of image pixel values";
      kwds_set,"CTYPE1","RA---TAN","Coordinate system 1. axis";
      kwds_set,"CRPIX1",Num_pixels1/2. + Det_offset(1)/Pixel_size1,"Reference pixel coordinate 1";
      crval1 = is_void(Ra_scx) ? 0.0 : Ra_scx;
      kwds_set,"CRVAL1",crval1,"Reference value coordinate 1";
      kwds_set,"CUNIT1","deg","Unit of reference value coordinate 1";
      kwds_set,"CTYPE2","DEC--TAN","Coordinate system 2. axis";
      kwds_set,"CRPIX2",Num_pixels2/2. + Det_offset(2)/Pixel_size2,"Reference pixel coordinate 2";
      crval2 = is_void(Dec_scx) ? 0.0 : Dec_scx;
      kwds_set,"CRVAL2",crval2,"Reference value coordinate 2";
      kwds_set,"CUNIT2","deg","Unit of reference value coordinate 2";
      pix1 = (Pixel_size1/Focal_length)*(180./pi);
      rot = is_void(Posang) ? 0.0 : Posang * (pi/180.);
      kwds_set,"CD1_1", -pix1*cos(rot), "[deg] -pix1*cos(rot)";
      kwds_set,"CD1_2", -pix1*sin(rot), "[deg] -pix1*sin(rot)";
      kwds_set,"CD2_1", -pix1*sin(rot), "[deg] -pix1*sin(rot)";
      kwds_set,"CD2_2",  pix1*cos(rot), "[deg]  pix1*cos(rot)";

      kwds_set,"E_MIN",emin,"[keV] Lower energy limit";
      kwds_set,"E_MAX",emax,"[keV] Upper energy limit";
      if( !is_void(Exposure) ) kwds_set,"EXPOSURE", Exposure,"[s] Exposure time";
      kwds_set,"BKGLVL", bkglvl,"[cts/s/keV] Background counts added to the image";
      writefits,outfile,im,clobber=1;
   }

   if( am_subroutine() ) {
      disp, im;
   } else return im;

}

/* Function mt_upd_scatter_width */

func mt_upd_scatter_width( filename, width, expon=, typ=, angle_max=, unit=, outfile= )
/* DOCUMENT mt_upd_scatter_width, filename, width, expon=, typ=, angle_max=, unit=, outfile=

   Updates a FITS scatter table with (super-)gaussian scatter tables
   characterized by 'width'.
   Keyword 'expon'
        is the exponent in the supergaussian expression: exp(-0.5*abs(x/sigma)^expon)
        defaults to 2 - the true gaussian case.
   Keyword 'typ' most be one of
       "f" : FWHM (default)
       "s" : sigma
       "h" : HPD
   Keyword angle_max is max angle of distribution
           If not given it will be kept as previously
   Keyword 'unit' must be one of
       "arcsec" : arcsec
       "arcmin" : arcmin
       "deg" : degrees
       "rad" : radians (default)
   Defaults to:    FWHM =  9.7e-5 rad (20 arcsec)

   NB! 'unit' applies BOTH to 'width' and to 'angle_max'

   Keyword 'outfile' is the name of the output file, if not given then the
       input file will be updated.

   2010-04-13/NJW
   2013-07-04/NJW, updated wih supergaussian option
*/
{
   extern Version;

   dol = filename+"+1";
   fh = headfits( dol );
   if( is_void(outfile) ) {
      outfile = filename;
      flag_newfile = 0;
   } else flag_newfile = 1;

   extname = fxpar(fh, "EXTNAME");
   is_type2 = (extname == "SCATTER_TYPE2");

   if( is_void(expon) ) expon = 2;

   // See if 'width' has been given

   if( is_void(width) ) {
      width = 9.7e-5; // FWHM in radians
      typ = "f";
      unit = "rad";
   } else width = double(width);

   // Check for width being FWHM, sigma, or HPD

   if( is_void(typ) ) {
      typ = "f";
   } else {
      typ = strpart(typ,1:1);
      if( typ != "f" && typ != "s" && typ != "h" ) error,"Bad 'typ' keyword";
   }

   // Check for unit being arcsec, arcmin, degree, or radian

   if( is_void(unit) ) {
      unit = "rad";
   } else {
      if( unit != "arcsec" && unit != "arcmin" && unit != "deg" && unit != "rad" ) error,"Bad 'unit' keyword";
   }

   if( typ == "s" ) {
      sigma = width;
   } else if( typ == "f" ) {
      sigma = width / 2.35482;
   } else { // typ == "h"
      sigma = width / 1.34898;
   }
   conv_fac = 1.0;
   if( unit == "arcsec" ) {
      conv_fac = 1./206265.;
   } else if( unit == "arcmin" ) {
      conv_fac = 1./3437.75;
   } else if( unit == "deg" ) {
      conv_fac = 1./57.2958;
   } // do nothing if unit == "rad"
   sigma *= conv_fac; // now in radians

  /*
   ********    Act differently for a type 2 scatter file
   */

   if( is_type2 ) {

      if( flag_newfile ) cp, filename, outfile;

      dol = outfile+"[SCATTER_TYPE2B]";
      angle_dist = rdfitscol( dol, "ANGLE" );
      atype = typeof(angle_dist);
      old_distr = rdfitscol( dol, "DISTRIBUTION" );

      if( !is_void(angle_max) ) {
         angle_max *= conv_fac;
         angle_dist = span(-angle_max,angle_max,numberof(angle_dist));
      }
      distribution = exp(-0.5*abs(angle_dist/sigma)^expon);
      if( structof(old_distr) == float ) distribution = float(distribution);
      if( atype == "float" ) angle_dist = float(angle_dist);

      fits_bintable_poke, dol, 1, "DISTRIBUTION", distribution;
      fits_bintable_poke, dol, 1, "ANGLE", angle_dist;
      write,format="Has updated type 2 scatter file %s\n", outfile;

   } else {

      ener = rdfitscol( dol, "energy" );
      angi = rdfitscol( dol, "angle_in" );
      rcoe = rdfitscol( dol, "r_coef" );
      coat = rdfitscol( dol, "coating" );
      dori = rdfitscol( dol, "data_origin" );

      scat = rdfitscol( dol, "distribution" );
      dms = dimsof( scat );
      if( is_void(angle_max) ) {
         angle_dist = scat(,1);
      } else {
         angle_max *= conv_fac;
         angle_dist = span(-angle_max,angle_max,dms(2));
      }

      distribution = exp(-0.5*abs(angle_dist/sigma)^expon)(,-:1:dms(3));
      distribution(,1) = angle_dist;

      dori(2:0) = swrite(format="Gaussian with FWHM of %10.3e rad (sigma=%10.3e rad)",\
                 sigma*2.35482 , sigma);

      kwds_init;
      fits_copy_keys, fh, tokwds=1;
      kwds_set,"DATE",ndate(3),"Time of file creation";
      kwds_set,"RESPONSI","Niels J. Westergaard","Responsible person";
      kwds_set,"SITE","NSI/DTU","Institution";
      kwds_set,"ORIGIN","mt_upd_scatter_width","produced this file";
      kwds_set,"ORIGFILE",filename,"Original file";
      kwds_set,"VERSION", Version, "MT_RAYOR version indicator";
      kwds_set,"COATING", coat(0), "Coating number";
      kwds_set,"SIGMA", sigma, "[rad] Sigma of gaussian distribution";
      kwds_set,"TUNIT1","keV","Photon energy";
      kwds_set,"TUNIT2","radian","Incoming angle";
      wrmfitscols, outfile, "ENERGY", ener, "ANGLE_IN", angi, \
          "R_COEF", rcoe, "COATING", coat, \
          "DISTRIBUTION", distribution, "DATA_ORIGIN", dori, \
          clobber=1, extname="UPD_SCAT_DIST";
      write,format="Has written scatter file %s\n", outfile;

   }

}

/* Function mt_scatfrac */

func mt_scatfrac( filename, &scatter_fraction, get=, outfile= )
/* DOCUMENT mt_scatfrac, filename, scatter_fraction, outfile=
            mt_scatfrac, filename, >scatter_fraction, get=1

   Updates a FITS scatter table with central one or two elements (symmetry achieved)
   so that the scattered part (fraction of total) fulfills the 'scatter_fraction'.

       Arguments:

     filename    Name of scatter file to use. If not given the previously loaded
                 'Scatter_file' is used.
     scatter_fraction  First version: Input to update the scattered fraction in the
                 scatter distribution. Either a scalar (for a scatter file of type2
                 this is expected; for a scatter file of type 1 this is expanded to an
                 array of appropriate number of elements), or an array (for a scatter
                 file of type 1) with as many elements as there are rows in the scatter
                 file, note that the first one is not used.
                 Second version: Returns the derived value - or, in case of a scatter file
                 of type 1, an array with values.

       Keywords:

     outfile     name of the output file, if not given then the
                 input file will be updated.
     get         (bool) when set (second version) the distribution will be read and the
                 scattered fraction returned in the 'scatter_fraction' argument.
                 Else first version.

   2015-09-01/NJW
*/
{
   extern Version;

   if( is_void(filename) ) filename = Scatter_file;
   if( !file_test(filename) ) {
      write,"MT_SCATFRAC: Did not find: "+filename+", no action!";
      return;
   }
   dol = filename+"+1";
   fh = headfits( dol );
   if( is_void(outfile) ) {
      outfile = filename;
      flag_newfile = 0;
   } else flag_newfile = 1;

   extname = fxpar(fh, "EXTNAME");
   is_type2 = (extname == "SCATTER_TYPE2");

   if( !get ) if( is_void(scatter_fraction) ) scatter_fraction = 0.99;

  /*
   ********    Act differently for a type 2 scatter file
   */

   if( is_type2 ) {

      if( flag_newfile ) cp, filename, outfile;

      dol = outfile+"[SCATTER_TYPE2B]"; // get extension with scatter distribution
      angle_dist = rdfitscol( dol, "ANGLE" );
      atype = typeof(angle_dist);
      old_distr = rdfitscol( dol, "DISTRIBUTION" );
      distribution = old_distr;
      sum_dist = sum(distribution);

      n = numberof(angle_dist);
      i0 = (abs(angle_dist))(mnx);
      if( near( abs(angle_dist(i0)) , abs(angle_dist(i0+1)), 1.e-16 ) ) {
         i0 = [i0,i0+1];
      } else if( near( abs(angle_dist(i0)) , abs(angle_dist(i0-1)), 1.e-16 ) ) i0 = [i0-1,i0];
      ni0 = numberof(i0);

      old_distr(i0) = 0.0;
      if( get ) {
         scatter_part = old_distr(sum);
         scatter_fraction = scatter_part / sum_dist;
         return;
      } else {
         scatter_part = old_distr(sum);
         specular_part = (1.0 - scatter_fraction) * scatter_part / scatter_fraction;
         distribution(i0) = specular_part / ni0;
      }

      fits_bintable_poke, dol, 1, "DISTRIBUTION", distribution;
      fits_bintable_poke, dol, 1, "ANGLE", angle_dist;
      write,format="Has updated type 2 scatter file %s\n", outfile;

   } else {

      ener = rdfitscol( dol, "energy" );
      angi = rdfitscol( dol, "angle_in" );
      rcoe = rdfitscol( dol, "r_coef" );
      coat = rdfitscol( dol, "coating" );
      dori = rdfitscol( dol, "data_origin" );

      old_distr = distribution = rdfitscol( dol, "distribution" );
      dms = dimsof( old_distr );
      angle_dist = old_distr(,1);

      i0 = (abs(angle_dist))(mnx);
      if( near( abs(angle_dist(i0)) , abs(angle_dist(i0+1)), 1.e-16 ) ) {
         i0 = [i0,i0+1];
      } else if( near( abs(angle_dist(i0)) , abs(angle_dist(i0-1)), 1.e-16 ) ) i0 = [i0-1,i0];
      ni0 = numberof(i0);

      if( get ) {
         scatter_fraction = array(double, dms(3) );
      } else {
         nsf = numberof( scatter_fraction );
         if( nsf == 1 ) {
            scatter_fraction = array( scatter_fraction, dms(3) );
         } else {
            if( nsf != dms(3) ) error,"Bad dimension of scatter_fraction";
         }
      }

      for( i = 2; i <= dms(3); i++ ) {
         sum_dist = old_distr(sum,i);
         old_distr(i0,i) = 0.0;
         if( get ) {
            scatter_part = old_distr(sum,i);
            scatter_fraction(i) = scatter_part / sum_dist;
         } else {
            scatter_part = old_distr(sum,i);
            specular_part = (1.0 - scatter_fraction(i)) * scatter_part / scatter_fraction(i);
            distribution(i0,i) = specular_part / ni0;
         }
      }

      if( get ) return;

      kwds_init;
      fits_copy_keys, fh, tokwds=1;
      kwds_set,"DATE",ndate(3),"Time of file creation";
      kwds_set,"RESPONSI","Niels J. Westergaard","Responsible person";
      kwds_set,"SITE","NSI/DTU","Institution";
      kwds_set,"ORIGIN","mt_scatfrac","produced this file";
      kwds_set,"ORIGFILE",filename,"Original file";
      kwds_set,"VERSION", Version, "MT_RAYOR version indicator";
      kwds_set,"COATING", coat(0), "Coating number";
      kwds_set,"SCATFRAC", scatter_fraction, "Scatter/specular fraction";
      kwds_set,"TUNIT1","keV","Photon energy";
      kwds_set,"TUNIT2","radian","Incoming angle";
      wrmfitscols, outfile, "ENERGY", ener, "ANGLE_IN", angi, \
          "R_COEF", rcoe, "COATING", coat, \
          "DISTRIBUTION", distribution, "DATA_ORIGIN", dori, \
          clobber=1, extname="UPD_SCAT_DIST";
      write,format="Has written scatter file %s\n", outfile;

   }

}

/* Function mt_det_add_bkg */

func mt_det_add_bkg( instr=, dxb=, dxb_adjust= )
/* DOCUMENT mt_det_add_bkg, instr=, dxb=, dxb_adjust=

     Add both instrument and diffuse background to current
     event list (Evlist).

     Override selections in the detector description file by
     keywords instr resp. dxb that each is the name of a file.
     Keyword 'dxb_adjust' is a factor to enhance the DXB contribution.

     The external variable 'Exposure' together with the information
     in the detector description file will define the number of
     background counts.

     SEE ALSO: mt_det_add_instr_bkg, mt_det_add_dxb_bkg

   2010-06-19/NJW
*/
{
   mt_det_add_instr_bkg, instr;
   mt_det_add_dxb_bkg, dxb, adjust=dxb_adjust;
}

/* Function mt_det_add_instr_bkg */

func mt_det_add_instr_bkg( filename )
/* DOCUMENT mt_det_add_instr_bkg[, filename]

   Add the detector background as described in the s-format file
   with two columns, energy (keV) and background (cts/s/keV/cm2).

   If 'filename' is given then this will be read as Instr_bkg_file
   else the file defined in the detector definition file
   (keyword instr_bkg_file) and read in by command: 'mt_detector,init=1'.

   2010-05-17/NJW
*/
{

// ynclude = zba
  extern  E_min,           Exposure,        Num_pixels1,     Pixel_size1, \
          Evlist,          Instr_bkg_file,  Num_pixels2,     Pixel_size2;
// yxclude =

   local ener, bkg;

   if( !is_void(filename) ) {
      if( !file_test(filename) ) {
         write,"Warning: instr_bkg_file: "+filename+" was not found";
         write,"     cancel operation";
         return;
      }
      Instr_bkg_file = filename;
   }
   if( is_void(Instr_bkg_file) ) {
      write,"Warning! No instr_bkg_file has been defined, cancel operation";
      return;
   }

   rstab,Instr_bkg_file,2,ener,bkg,typ="dd",silent=1;

   // Detector area (in cm2) :
   det_area = 0.01 * Num_pixels1 * Num_pixels2 * Pixel_size1 * Pixel_size2;

   //+ int_ener = integ( bkg, ener, E_max(0) ) - integ( bkg, ener, E_min(1));
   int_ener = integ( bkg, ener, ener(0) );

   if( is_void(Exposure) ) {
      write,"  NB: 'Exposure' is set to 100 s";
      Exposure = 100.0;
   }
   n_bkg_counts = long(int_ener * det_area * Exposure + 0.5);

   ener_bkg = draw_from_dist( ener, bkg, n_bkg_counts );

   bkg_evlist = array(s_MTEvent,n_bkg_counts);
   bkg_evlist.detx = random(n_bkg_counts)*Num_pixels1 - Num_pixels1/2.;
   bkg_evlist.rawx = long(bkg_evlist.detx + Num_pixels1/2. + 1);
   bkg_evlist.dety = random(n_bkg_counts)*Num_pixels2 - Num_pixels2/2.;
   bkg_evlist.rawy = long(bkg_evlist.dety + Num_pixels2/2. + 1);
   bkg_evlist.flag = 0;
   bkg_evlist.phs_index = -1;
   bkg_evlist.energy = ener_bkg;
   for( i = 1; i <= n_bkg_counts; i++ ) {
      w = where( E_min < bkg_evlist(i).energy );
      bkg_evlist(i).pha = numberof(w) ? w(0) : 1;
   }
   grow, Evlist, bkg_evlist;
   write,format="%i instr bkg events have been added to 'Evlist'\n", n_bkg_counts;

}

/* Function mt_det_add_dxb_bkg */

func mt_det_add_dxb_bkg( filename, adjust= )
/* DOCUMENT mt_det_add_dxb_bkg[, filename][, adjust=]

   Add the DXB background from the DXB raytracing
   event list in 'filename'. The events are selected randomly
   from this file.

   If 'filename' is given then this will be read as Dxb_bkg_file
   else the file defined in the detector definition file
   (keyword dxb_bkg_file) and read in by command: 'mt_detector,init=1'.

   2010-06-19/NJW
*/
{

// ynclude = zbb
  extern  Dxb_bkg_file,  E_min,         Evlist,        Exposure;
// yxclude =

   if( !is_void(filename) ) {
      Dxb_bkg_file = filename;
      if( !file_test(Dxb_bkg_file) ) {
         write,"Warning! dxb_bkg_file: "+Dxb_bkg_file+" was not found";
         write,"    processing was cancelled";
         return;
      }
   }
   if( is_void(Dxb_bkg_file) ) {
      write,"Warning! No dxb_bkg_file has been define, cancel processing";
      return;
   }
   if( is_void(adjust) ) adjust = 1.0;

   hdr = headfits(Dxb_bkg_file+"+1");
   pool_exposure = fxpar(hdr,"EXPOSURE");
   if( is_void(pool_exposure) ) error,"No EXPOSURE keyword in DXB event file";
   npool = fxpar(hdr,"naxis2");
   // how many events to pick?
   n_dxb = long(npool*(Exposure/pool_exposure)*adjust);
   if( n_dxb < 1 || n_dxb > npool ) error,"mt_det_add_dxb_bkg error, n_dxb is illegal";
   write,format="Add %i DXB events to 'Evlist' ...\n", n_dxb;

   r = random(n_dxb);
   i = long(r*npool) + 1;

   if( mem_restore( Dxb_bkg_file+"+1_rawx", rawx ) < 0 ) {
      write,format="Reading %s ...\n", Dxb_bkg_file+"+1_rawx";
      rawx = rdfitscol(Dxb_bkg_file+"+1","rawx");
      mem_save, Dxb_bkg_file+"+1_rawx", rawx;
   }
   if( mem_restore( Dxb_bkg_file+"+1_rawy", rawy ) < 0 ) {
      write,format="Reading %s ...\n", Dxb_bkg_file+"+1_rawy";
      rawy = rdfitscol(Dxb_bkg_file+"+1","rawy");
      mem_save, Dxb_bkg_file+"+1_rawy", rawy;
   }
   if( mem_restore( Dxb_bkg_file+"+1_detx", detx ) < 0 ) {
      write,format="Reading %s ...\n", Dxb_bkg_file+"+1_detx";
      detx = rdfitscol(Dxb_bkg_file+"+1","detx");
      mem_save, Dxb_bkg_file+"+1_detx", detx;
   }
   if( mem_restore( Dxb_bkg_file+"+1_dety", dety ) < 0 ) {
      write,format="Reading %s ...\n", Dxb_bkg_file+"+1_dety";
      dety = rdfitscol(Dxb_bkg_file+"+1","dety");
      mem_save, Dxb_bkg_file+"+1_dety", dety;
   }
   if( mem_restore( Dxb_bkg_file+"+1_energy", ener ) < 0 ) {
      write,format="Reading %s ...\n", Dxb_bkg_file+"+1_energy";
      ener = rdfitscol(Dxb_bkg_file+"+1","energy");
      mem_save, Dxb_bkg_file+"+1_energy", ener;
   }

   dxb_evlist = array(s_MTEvent,n_dxb);
   dxb_evlist.detx = detx(i);
   dxb_evlist.rawx = rawx(i);
   dxb_evlist.dety = dety(i);
   dxb_evlist.rawy = rawy(i);
   dxb_evlist.flag = 0;
   dxb_evlist.phs_index = 0;
   dxb_evlist.energy = ener(i);
   for( i = 1; i <= n_dxb; i++ ) {
      w = where( E_min < dxb_evlist(i).energy );
      dxb_evlist(i).pha = numberof(w) ? w(0) : 1;
   }
   grow, Evlist, dxb_evlist;

}

/* Function mt_get_scatter_hpd */

func mt_get_scatter_hpd( energy, angle_in )
/* DOCUMENT hpd_rad = mt_get_scatter_hpd( energy, angle_in )

   Returns HPD in radians.
   (For a gaussian sigma = HPD / 1.35).

   Arguments:
      energy    [keV]
    angle_in    [rad]

*/
{
   d = mt_sel_scatter( energy, angle_in );
   z = integ(d,Anglesarr,Anglesarr);
   z /= z(0);
   i = 0;
   da = 1.e33;
   do {
      da_prev = da;
      i++;
      // Find angle so half of distribution is covered
      q = interp(Anglesarr,z-z(i),0.5);
      da = q - Anglesarr(i);
   } while( da < da_prev );
   i--;
   return interp(Anglesarr,z-z(i),0.5) - Anglesarr(i);
}

/* Function mt_spoke_read */

func mt_spoke_read( filename, pos= )
/* DOCUMENT mt_spoke_read, filename, pos=

   'filename' is assumed to be an s-format file with at least the
   columns 'angle' and 'width'. Optional columns are 'rstart' and
   'rstop' (always in mm).

   Required keywords: reverse, angle_unit and width_unit
      angle_unit can only be one of: deg  rad
      width_unit can only be one of: mm   cm

*/
{
// ynclude = zbc
  extern  Module_num,          Reverse_spokes2,     Rstop_spokes2, \
          Phi_spokes1,         Rstart_spokes1,      Spoke_define_files, \
          Phi_spokes2,         Rstart_spokes2,      Width_spokes1, \
          Reverse_spokes1,     Rstop_spokes1,       Width_spokes2, \
          Full_length_spokes1, Full_length_spokes2;
// yxclude =

   if( is_void(pos) ) pos = 1;
   if( pos != 1 && pos != 2 ) error,"Bad value of keyword pos";
   if( is_void(filename) ) filename = Spoke_define_files(2*Module_num-2+pos);

   if( filename != "none" ) {

      // define unit conversion factors

      au = comgets( filename, "angle_unit" );
      aconv_factor = au == "deg" ? pi/180 : 1.0;
      wu = comgets( filename, "width_unit" );
      wconv_factor = wu == "cm" ? 10.0 : 1.0;

      // see if radius defining columns are present
      colnames = comgets(filename,"colname");
      if( noneof( colnames == "angle" ) ) error,"Spoke angle column is missing";
      if( noneof( colnames == "width" ) ) error,"Spoke width column is missing";
      rdef = anyof( colnames == "rstart" ) & anyof( colnames == "rstop" );

      // read the table values
      phi_spokes = rscol(filename,"angle",silent=1,nomem=1)*aconv_factor;
      width_spokes = rscol(filename,"width",silent=1,nomem=1)*wconv_factor;
      n_spokes = numberof(phi_spokes);
      if( rdef ) {
         rstart_spokes = rscol(filename,"rstart",silent=1,nomem=1); // must be in mm
         rstop_spokes = rscol(filename,"rstop",silent=1,nomem=1); // must be in mm
      } else {
         rstart_spokes = array(0.0, n_spokes);
         rstop_spokes = array(1.e6, n_spokes); // ridiculously large number
      }

      // make sure that all angles are positive (or zero) and sorted
      phi_spokes = zero2pi(phi_spokes);
      is = sort(phi_spokes);
      phi_spokes = phi_spokes(is);
      width_spokes = width_spokes(is);
      rstart_spokes = rstart_spokes(is);
      rstop_spokes = rstop_spokes(is);

      // see if spokes are blocking or slits

      reverse_spokes = comget( filename, "reverse", lng=1 );
      if( is_void(reverse_spokes) ) reverse_spokes = 0;
      if( reverse_spokes != 0 ) reverse_spokes = 1;

      // see if spokes are blocking in full length

      full_length_spokes = comget( filename, "full_length", lng=1 );
      if( is_void(full_length_spokes) ) full_length_spokes = 0;

      if( pos == 1 ) {
         Phi_spokes1 = phi_spokes;
         Width_spokes1 = width_spokes;
         Rstart_spokes1 = rstart_spokes;
         Rstop_spokes1 = rstop_spokes;
         Reverse_spokes1 = reverse_spokes;
         Full_length_spokes1 = full_length_spokes;
      } else {
         Phi_spokes2 = phi_spokes;
         Width_spokes2 = width_spokes;
         Rstart_spokes2 = rstart_spokes;
         Rstop_spokes2 = rstop_spokes;
         Reverse_spokes2 = reverse_spokes;
         Full_length_spokes2 = full_length_spokes;
      }
   }

}

/* Function mt_spoke_blocking */

func mt_spoke_blocking( p, pos= ) // p is coordinate in (x,y)
/* DOCUMENT flag = mt_spoke_blocking( p, pos= )

   Returns [1,0] if the photon at position p (2 element array (x,y))
   is blocked by a spoke. Returns [0,next_spoke] if allowed.
   This makes it possible to test against 'passage through same opening'.
   The spokes are defined in the
   file given in variable 'Spoke_define_file'.

   Keyword pos: 1 for entrance and 2 for exit of optical module.

*/
{

// ynclude = zbd
  extern  Phi_spokes1,      Reverse_spokes2,  Rstop_spokes1,    Width_spokes1, \
          Phi_spokes2,      Rstart_spokes1,   Rstop_spokes2,    Width_spokes2, \
          Reverse_spokes1,  Rstart_spokes2,   Full_length_spokes1;
// yxclude =

   if( is_void(pos) ) pos = 1;
   if( pos != 1 && pos != 2 ) error,"Bad value of keyword pos";

  /*
   *   Assume that the spoke definition is such that
   *   Phi_spokes is never decreasing (assured by function
   *   mt_spoke_read). There may be several spokes with the same angle.
   */

   phi = zero2pi(atan( p(2), p(1) )); // phi angle of photon
   rad = sqrt(p(1)^2 + p(2)^2);       // radius of photon
   //+ write,format="SB: phi = %8.4f\n", phi;

   // use temporary storage to do a radius selection
   if( pos == 1 ) {
      phi_spokes = Phi_spokes1;
      width_spokes = Width_spokes1;
      rstart_spokes = Rstart_spokes1;
      rstop_spokes = Rstop_spokes1;
      reverse_spokes = Reverse_spokes1;
   } else {
      phi_spokes = Phi_spokes2;
      width_spokes = Width_spokes2;
      rstart_spokes = Rstart_spokes2;
      rstop_spokes = Rstop_spokes2;
      reverse_spokes = Reverse_spokes2;
   }

   w = where( rstart_spokes <= rad & rad <= rstop_spokes );
   nw = numberof(w);
   if( nw == 0 ) return [reverse_spokes,-1]; // no spokes at this radius
   phi_spokes = phi_spokes(w);
   width_spokes = width_spokes(w);
   rstart_spokes = rstart_spokes(w);
   rstop_spokes = rstop_spokes(w);

   // Locate next spoke

   w = where( phi_spokes > phi );
   nw = numberof(w);
   next_spoke = nw == 0 ? 1 : w(1);
   //+ write,format="SB: next_spoke = %i\n", next_spoke;
   previous_spoke = next_spoke - 1; // exploits that Yorick has 'rolling' indices
   //+ write,format="SB: previous_spoke = %i\n", previous_spoke;
   //+ write,format="SB: Phi prev/next = %8.4f  %8.4f\n", Phi_spokes(previous_spoke), Phi_spokes(next_spoke);
   // oplot,[p(1)],[p(2)],ps=12,symsize=0.5,color="red";
   // distance to previous spoke
   //+ epsilon = abs(phi - phi_spokes(previous_spoke));
   epsilon = zero2pi(phi - phi_spokes(previous_spoke));
   dprev = rad * sin(epsilon);

   // distance to next spoke
   //+ epsilon = abs(phi - phi_spokes(next_spoke));
   epsilon = zero2pi(phi_spokes(next_spoke) - phi);
   dnext = rad * sin(epsilon);

   //+ write,format="SB: dist prev/next = %8.4f  %8.4f mm\n", dprev, dnext;

   if( dprev < 0.5*width_spokes(previous_spoke) || dnext < 0.5*width_spokes(next_spoke) ) {
      // It has hit a spoke
      if( reverse_spokes ) {
         return [0,next_spoke];
      } else {
         return [1,0];
      }
   } else {
      // It did not hit a spoke
      if( reverse_spokes ) {
         return [1,0];
      } else {
         return [0,next_spoke];
      }
   }

}

/* Function mt_clear */

func mt_clear
/* DOCUMENT mt_clear

     Erases Exposure information
*/
{
   extern Exposure;
   Exposure = [];
}

/* Function mt_dxb2skydef */

func mt_dxb2skydef( skydefname, dol_dxbflux, wfov, n, exposure=,
            mission=, instrume= )
/* DOCUMENT mt_dxb2skydef, skydefname, dol_dxbflux, wfov, n, exposure=, \
            mission=, instrume=

     Produces a skydefinition file for DXB determination
     Put sources in a regular mesh around (RA,Dec) = (180,0)

     Arguments:
        skydefname   :  filename of sky definition file to be made
        dol_dxbflux  :  DOL of DXB flux information in proper energy
                        range and in photons /cm2/s/keV/sr
        wfov         :  [deg] full width of the FOV
        n            :  The number of 'sources' from center to edge
                        implying that the total number of sources
                        will become (2*n + 1)^2
     Keywords:
        exposure     :  [s] exposure time to be written to the output file
        mission      :  A scalar string for the FITS keyword 'MISSION'
        instrume     :  A scalar string for the FITS keyword 'INSTRUME'


     After this has been run, a call of mt_run (if the raytracing has been
     properly initiated) will generate the DXB photons:

      > mt_run,dol_dxbflux

    2010-08-13/NJW
*/
{

   extern Version;

   local x, y;

   if( is_void(exposure) ) exposure= 1000.0; // s

   w_of_fov_deg = wfov; // degrees
   w_of_fov_arcmin = w_of_fov_deg * 60.; // arcmin
   // 'n' is number of steps to one side
   N = 2*n + 1; // making it odd
   step_arcmin = w_of_fov_arcmin/(N-1);
   // 8.4616e-8 sr per arcmin^2
   omega = 8.4616e-8 * step_arcmin^2;

   ra_scx = 180.;
   dec_scx = 0.0;
   posang = 0.0;
   rad2arcmin = 60*(180./pi); // rad to arcmin

   ra_obj = reform(span(180-w_of_fov_deg/2,180+w_of_fov_deg/2,N)(,-:1:N),N*N);
   dec_obj = reform(span(-w_of_fov_deg/2,w_of_fov_deg/2,N)(-:1:N,),N*N);

   fluxdols = array( dol_dxbflux, N*N);
   enera = dphota = x_sky = y_sky = array(1., N*N);
   renorma = array( omega, N*N);


   for( i = 1; i <= N*N; i++ ) {
      jconv_coord, ra_scx, dec_scx, posang, ra_obj(i), dec_obj(i), \
                   x, y, to_sc=1;
      x_sky(i) = -x * rad2arcmin;
      y_sky(i) = y * rad2arcmin;
   }

   kwds_init;
   kwds_set,"EXTNAME","SKY_DEFINITION","Name of this extension";
   if( !is_void(mission) ) kwds_set,"MISSION", mission, "Name of mission";
   if( !is_void(instrume) ) kwds_set,"INSTRUME",instrume,"Name of instrument";
   kwds_set,"DATE",ndate(3),"Date of creation";
   kwds_set,"ORIGIN","MT_RAYOR-"+Version,"Software package/version";
   kwds_set,"FUNCTION","mt_dxb2skydef (MT_RAYOR)","Software that created this file";
   kwds_set,"VERSION", Version,"MT_RAYOR version";
   kwds_set,"EXPOSURE", exposure,"[s] Exposure time";
   kwds_set,"RA_SCX", ra_scx,"[deg] R.A. of satellite pointing axis";
   kwds_set,"DEC_SCX", dec_scx,"[deg] Decl. of satellite pointing axis";
   kwds_set,"POSANG", posang,"[deg] Position angle of satellite pointing";
   kwds_set,"COMMENT","Mesh of sources for DXB determination";

   wrmfitscols, skydefname, "X_SKY", x_sky, "Y_SKY", y_sky, \
                "ENERGY", enera, "DPHOT", dphota, "RENORM", renorma, \
                "DOL", fluxdols, "RA_OBJ", ra_obj, "DEC_OBJ", dec_obj, \
                clobber=1;

   write,format="%s has been written\n", skydefname;

}

/* Function mt_skyspec2skydef */

/*************************************************************************
 *
 *   Use a FITS sky spectral definition file to make a MT_RAYOR
 *   sky definition file (SDF) from a sky spectral definition file (SSDF).
 *
 *   The SSDF consists of a norm map (NORM_MAP),
 *   and then -
 *     in XSPEC mode a number of maps for the mapped spectral parameters
 *     in SIMPLE mode a map of PARAM1 (P1MAP) or a PARAM1 keyword,
 *            and a map of NH (NHMAP) or a NH keyword. The spectral code is
 *            given as a keyword.
 *
 *   In both cases the last extension is the ARF for the instrument that
 *   produced the original countrate image.
 *
 *   The SSDF can be prepared with function 'skyima2skyspec'
 *
 *   2013-05-24/NJW
 *
 ************************************************************************/

func mt_skyspec2skydef( sky_spec_def_file, skydef_file, ra_scx, dec_scx, posang, \
            fraclim=, exposure=, e1=, e2=, nchan=, \
            radius=, fluxdir=, mission=, instrume=, telescop= )
/* DOCUMENT mt_skyspec2skydef, sky_spec_def_file, skydef_file, ra_scx, dec_scx, posang

  or   nsources = mt_skyspec2skydef( sky_spec_def_file, skydef_file, ra_scx, dec_scx, posang )

    Arguments:
       sky_spec_def_file  - name of the input sky spectral definition file (FITS format)
       skydef_file - Name of the output sky definition file
       ra_scx      - [deg] RA of telescope boresight or DOL of previous sky definition file.
       dec_scx     - [deg] Dec of telescope boresight
       posang      - [deg] Position angle of telescope

    Keywords:
       fraclim  - is fractional limit of image maximum for inclusion
                  (if set too low a LARGE number of sources may be defined)
                  default is 0.5.
       exposure - [s] exposure time, default 1000.
       e1       - [keV] lower energy boundary for photon spectra, default 0.5.
       e2       - [keV] upper energy boundary for photon spectra, default 10.0.
       nchan    - number of energy channels in photon spectra (default is 200).
       radius   - [deg] only pixels within this radius from the pointing direction
                  will be used, default 0.5.
       fluxdir  - directory for storing the photon flux files (default is current).
                  If given, then the flux files will be written to disk.
       mission  - name of mission (becomes FITS keyword 'MISSION')
       instrume - name of instrument (becomes FITS keyword 'INSTRUME')
       telescop - name of 'telescope' (becomes FITS keyword 'TELESCOP')

    If the argument 'ra_scx' is a string, then it is interpreted as a DOL to a previous
    SDF and the keywords except 'fluxdir' will be copied from that. If some are given
    nevertheless, they will be neglected.


    SEE ALSO: mt_skyima2skyspec

    2010-08-18/NJW
*/
{

   extern Version;

   local ra, dec;
   local x, y;

   mt_log_entry, 5, "Entered mt_skyspec2skydef";

   rad2arcmin = 60*(180./pi); // rad to arcmin

   hdr_norm = headfits(sky_spec_def_file+"[NORM_MAP]");
   norm_map = readfits(sky_spec_def_file+"[NORM_MAP]");
   dms = dimsof(norm_map);
   mt_log_entry, 5, "##47## mt_skyspec2skydef, has read norm_map with dms: "+swrite(dms(2),dms(3),format="%ix%i");

   if( typeof( ra_scx ) == "string" ) {
      hdr = headfits( ra_scx+"[SKY_DEFINITION]" );
      ra_scx = fxpar( hdr, "ra_scx" );
      dec_scx = fxpar( hdr, "dec_scx" );
      posang = fxpar( hdr, "posang" );
      fraclim = fxpar( hdr, "fraclim" );
      radius = fxpar( hdr, "radius" );
      e1 = fxpar( hdr, "e1" );
      e2 = fxpar( hdr, "e2" );
      nchan = fxpar( hdr, "nchan" );
      exposure = fxpar( hdr, "exposure" );
   }
   if( is_void(nchan) ) nchan = 200;
   if( is_void(radius) ) radius = 0.5; // deg
   if( is_void(exposure) ) exposure = 1000.0; // s
   if( is_void(fraclim) ) fraclim = 0.05;
   if( is_void(e1) ) e1 = 0.5;
   if( is_void(e2) ) e2 = 10.0;
   if( is_void(fluxdir) ) {
      fluxdir = "";
      // Reset the Energ_lo, Energ_hi, and Flux content of memory
      mem_reset;
      usemem = 1;
   } else {
      fluxdir = app_slash(fluxdir);
      nof = 0; // force writing photflux file(s)
      usemem = 0;
   }

   // get the mode ("XSPEC" or "SIMPLE") from the header of the SSDF

   mode = fxpar( hdr_norm, "MODE" );
   in_xspecmode = ( mode == "XSPEC" );

   map_index = 0; // 0 for value and third index in 'parmscube' for map
                  // normalisation is first sheet: parmscube(,,1)
   mapnum = 1; // There's always the normalisation map
   tolmap = []; // should grow for each map

   if( in_xspecmode ) {

      mt_log_entry, 5, "##48## mt_skyspec2skydef: in_xspecmode==1";

      xspecmodel = fxpar( hdr_norm, "XSPMODEL" );
      idxnorm    = fxpar( hdr_norm, "IDXNORM" );
      // go through the PARAM1, PARAM2, .. keywords to locate
      // values and maps
      // save the maps in 'parmscube'
      parmscube = float(norm_map(,,-));
      v_or_map = []; // 0 for value and 1 for map
      nparms = 0;
      xspecparms = [];
      while( !is_void((pval = fxpar(hdr_norm,"PARAM"+itoa(++nparms)))) ) {
         if( typeof(pval) == "string" ) {
            mt_log_entry, 5, "##49## mt_skyspec2skydef: Next (string) pval = "+pval;
            grow, v_or_map, 1;
            grow, map_index, ++mapnum;
            grow,xspecparms,-99.;
            write,"Reading "+sky_spec_def_file+"[PARAM"+itoa(nparms)+"]";
            mt_log_entry, 5, "##50## mt_skyspec2skydef, Reading "+sky_spec_def_file+"[PARAM"+itoa(nparms)+"]";
            hdr = headfits(sky_spec_def_file+"[PARAM"+itoa(nparms)+"]");
            tolmap_c = fxpar(hdr,"TOLMAP");
            grow,tolmap, tolmap_c;
            grow, parmscube, float(readfits(sky_spec_def_file+"[PARAM"+itoa(nparms)+"]"));
         } else {
            mt_log_entry, 5, "##51## mt_skyspec2skydef: Next (double) pval = "+ftoa(double(pval),ndec=3,sci=1);
            grow,v_or_map, 0;
            grow, map_index, 0;
            grow,xspecparms,pval;
         }
      }
      nparms--;
      write,"Found "+itoa(nparms)+" parameters";
      write,"  and built cube with "+itoa(dimsof(parmscube)(4))+" layers";
      mt_log_entry, 5, "##52## mt_skyspec2skydef: Found "+itoa(nparms)+" parameters", \
              "  and built cube with "+itoa(dimsof(parmscube)(4))+" layers";

   } else { // in 'simple' mode

      mt_log_entry, 5, "##53## mt_skyspec2skydef: Simple mode";

      // get spectral code from the NORM_MAP header of the SSDF

      sc = fxpar( hdr_norm, "SC" );
      if( is_void(sc) ) error,"SC not defined in NORM_MAP header";

      // is p1 a scalar or a map?

      local dummy, extno;
      get_exten_no, sky_spec_def_file+"[P1MAP]", dummy, extno;
      if( extno >= 0 ) { // There is a P1MAP in an extension
         grow, map_index, ++mapnum;
         p1dol = sky_spec_def_file+"+"+itoa(extno);
         hdr = headfits(p1dol);
         tolmap_c = fxpar(hdr, "TOLMAP");
         grow, tolmap, tolmap_c;
         p1map = readfits(p1dol);
         dmsmap = dimsof(p1map);
         if( dms(2) != dmsmap(2) || dms(3) != dmsmap(3) ) {
            error,"Error in file - p1map has bad dimensions ...";
         }
      } else { // a keyword must exist
         grow, map_index, 0;
         p1 = fxpar(hdr_norm,"PARAM1");
         if( is_void(p1) ) error,"Error in file - no PARAM1 keyword found ...";
         p1map = [];
      }

      // is nh a scalar or a map?

      get_exten_no, sky_spec_def_file+"[NHMAP]", dummy, extno;
      if( extno >= 0 ) { // There is a NHMAP in an extension
         grow, map_index, ++mapnum;
         nhdol = sky_spec_def_file+"+"+itoa(extno);
         nhmap = readfits(nhdol);
         hdr = headfits(nhdol);
         tolmap_c = fxpar(hdr, "TOLMAP");
         grow, tolmap, tolmap_c;
         dmsmap = dimsof(nhmap);
         if( dms(2) != dmsmap(2) || dms(3) != dmsmap(3) ) {
            error,"Error in file - nhmap has bad dimensions ...";
         }
      } else { // a keyword must exist
         grow, map_index, 0;
         nh = fxpar(hdr_norm,"NH");
         if( is_void(nh) ) error,"Error in file - no NH keyword found ...";
         nhmap = [];
      }
   }

   mt_log_entry, 5, "##54## mt_skyspec2skydef: numberof(tolmap) = "+itoa(numberof(tolmap)), \
                  "     mapnum = "+itoa(mapnum);

   if( numberof(tolmap) < mapnum-1 ) tolmap = []; // TOLMAP must be defined for all
                           // maps else it is discarded completely.

  /*
   *   Default rule: include pixels with value > fraclim*max_value
   */

   maxval = max(norm_map);
   allowed = where(norm_map > fraclim*maxval);
   nallowed = numberof(allowed);
   write,format="%i allowed pixels found\n", nallowed;
   mt_log_entry, 5, "##55## mt_skyspec2skydef: "+itoa(nallowed)+" allowed pixels found";


//////   Define groups with almost same parameter value in the maps applied ///////////

   gn = indgen(nallowed); // default, in case no grouping has been requested

   if( in_xspecmode ) {
      num_maps = numberof(where(map_index));
      if( mapnum-1 != num_maps ) error,"##56## wrong number of maps";
      if( num_maps ) {
         if( !is_void(tolmap) ) { // only do the grouping if the tolerance has been defined
            mt_log_entry, 5, "##57## mt_skyspec2skydef: is_void(tolmap) is false";
            k = 1;
            gn = [];
            for( j = 1; j <= numberof(xspecparms); j++ ) {
               if( map_index(j) ) { // there is a map for this parameter
                  xxx = parmscube(,,map_index(j));
                  if( is_void(gn) ) {
                     gn = groupn( xxx, allowed, tolmap(k));
                  } else {
                     gn2 = groupn( xxx, allowed, tolmap(k));
                     Ngn2 = max(gn2);
                     gn = (gn-1)*Ngn2 + gn2;
                     // reduce the grouping number
                     ggg = gn(sort(gn));
                     hhh = ggg(uniq(ggg));
                     for(i = 1; i <= numberof(hhh); i++ ) {
                        gn( where( gn == hhh(i) ) ) = i;
                     }
                  }
                  k++;
               }
            }
         }
      } else { // no maps have been defined - only one group
         mt_log_entry, 5, "##58## mt_skyspec2skydef: Define single grup";
         gn = array(int(1),nallowed);
      }

   } else { // in simple mode

      num_maps = !is_void(nh_map) + !is_void(p1_map);
      if( mapnum-1 != num_maps ) error,"##59## wrong number of maps";

      if( num_maps ) {
         if( !is_void(tolmap) ) { // only do grouping if the tolerance has been defined
            // NH map is first, p1 map is second
            k = 1;
            gn = [];
            if( !is_void(nh_map) ) {
               gn = groupn( nh_map, allowed, tolmap(k) );
               k++;
            }
            if( !is_void(p1_map) ) {
               if( is_void(gn) ) {
                  gn = groupn( p1_map, allowed, tolmap(k));
               } else {
                  gn2 = groupn( p1_map, allowed, tolmap(k));
                  Ngn2 = max(gn2);
                  gn = (gn-1)*Ngn2 + gn2;
                  // reduce the grouping number
                  ggg = gn(sort(gn));
                  hhh = ggg(uniq(ggg));
                  for(i = 1; i <= numberof(hhh); i++ ) {
                     gn( where( gn == hhh(i) ) ) = i;
                  }
               }
            }
         }
      } else { // no maps have been defined - only one group
         gn = array(int(1),nallowed);
      }
   }

//////   End of: Define groups with almost same parameter value in the maps applied ///////////



   // Walk through all allowed image pixels to produce a
   // photon flux file for each
   // but do it in order of increasing group numbers to take advantage of
   // reusing the spectral shape

   dols = x_sky = y_sky = enera = renorma = dphota = \
           name = ra_obj = dec_obj = intens = nharr = param1 = [];
   nsour = 0;
   sx_sum = 0.0;
   memflag = usemem ? 1 : [];

   ign = sort(gn);
   group_num_prev = 0; // make sure to define a non-existing group number

   // basic filenaming string for uniqueness:
   basicstr = ranstr(8);

   log_count = 0;

   for( m = 1; m <= nallowed; m++ ) {
      // group_number = gn(ign(m))
      group_num_now = gn(ign(m));
      i = ign(m);

      ij = indices(norm_map,allowed(i));
      skypos_fits,hdr_norm,ij(1),ij(2),ra,dec,to_sky=1,silent=1;

      // reject if more than 'radius' away from pointing

      if( arcdist(ra,dec,ra_scx,dec_scx) > radius ) {
         if( log_count < 10 ) {
            mt_log_entry, 5, "##60## mt_skyspec2skydef: (rejected) radius = "+ftoa(radius,ndec=3,sci=1), \
                  "ra,dec,ra_scx,dec_scx = "+swrite(ra,dec,ra_scx,dec_scx,format="%8.3f%8.3f%8.3f%8.3f");
            log_count++;
         }
         continue;
      }
      nsour++;

      if( group_num_prev != group_num_now ) {

         group_num_prev = group_num_now;

         // with version 4.6.0 make a unique file name to be able to run
         // several processes simultaneously
         //+ fname = swrite(format=fluxdir+"photflux_%06i.fits",1000*ij(1)+ij(2));
         fname = swrite(format=fluxdir+"pf_"+basicstr+"_%06i.fits",1000*ij(1)+ij(2));
         if( in_xspecmode ) {
            k = 1;
            for( j = 1; j <= numberof(xspecparms); j++ ) {
               if( v_or_map(j) ) { // this is a map
                  k++; // increase layer number
                  xspecparms(j) = parmscube(ij(1),ij(2),k);
               } // no action if not a map since the value has already been set
            }
            //+ xspecparms(idxnorm) = norm_map(ij(1),ij(2));
            xspecparms(idxnorm) = 1.0;

         } else {
            if( !is_void(p1map) ) p1 = p1map(allowed(i));
            if( !is_void(nhmap) ) nh = nhmap(allowed(i));
         }
         // Change with version 4.4.5
         //+ mk_photflux,sc=sc,nh=nh,norm=norm_map(allowed(i)), \
         //+     p1=p1,e1=e1,e2=e2,nchan=nchan,outfile=fname,silent=1;
   
         //+ mk_photflux,sc=sc,nh=nh,norm=norm_map(allowed(i)), \
         mk_photflux,sc=sc,nh=nh,norm=1.0, \
             p1=p1,e1=e1,e2=e2,nchan=nchan, \
             xspecmodel=xspecmodel, xspecparms=xspecparms, \
             outfile=fname,nof=nof,mem=memflag,silent=1;
         sx_sum += Sx_photflux;
         if(usemem)memflag++;

      } // if no need for a new calculation save the same filename
      // and only update the 'renorm' parameter

      grow, dols, fname+"+1";
      grow, name, strpadd(itoa(i),40,truncate=1);
      grow, ra_obj, ra;
      grow, dec_obj, dec;
      grow, intens, Sx_photflux;
      grow, nharr, nh;
      grow, param1, p1;

      jconv_coord, ra_scx, dec_scx, posang, ra, dec, \
                   x, y, to_sc=1;

      grow, x_sky, -x * rad2arcmin;
      grow, y_sky, y * rad2arcmin;
      grow, enera, 1.0;
      //+ grow, renorma, 1.0;
      grow, renorma, norm_map(allowed(i));
      grow, dphota, 1.0;
   }
   mt_log_entry, 5, "##61## mt_skyspec2skydef: (number of sources) nsour = "+itoa(nsour);
   if( nsour == 0 ) error,"##3## No sources within selected radius";

   kwds_init;
   kwds_set,"EXTNAME","SKY_DEFINITION","Name of this extension";
   if( !is_void(mission) ) kwds_set,"MISSION", mission, "Name of mission";
   if( !is_void(instrume) ) kwds_set,"INSTRUME", instrume,"Name of instrument";
   if( !is_void(telescop) ) kwds_set,"TELESCOP", telescop,"Name of telescope";
   kwds_set,"DATE",ndate(3),"Date of creation of this file";
   kwds_set,"ORIGIN","MT_RAYOR-"+Version, "Software/version for creation of this file";
   kwds_set,"FUNCTION","mt_skyspec2skydef", "Function used for creation of this file";
   kwds_set,"EXPOSURE", exposure, "[s] Exposure time";
   kwds_set,"RA_SCX", ra_scx, "[deg] R.A. of telescope bore sight";
   kwds_set,"DEC_SCX", dec_scx, "[deg] Decl. of telescope bore sightsatellite pointing axis";
   kwds_set,"E1", e1, "[keV] Lower energy for photon spectrum";
   kwds_set,"E2", e2, "[keV] Upper energy for photon spectrum";
   kwds_set,"NCHAN", nchan, "Number of energy channels in photon spectrum";
   kwds_set,"POSANG", posang, "[deg] Position angle of telescope bore sight";
   kwds_set,"INPUTFIL",sky_spec_def_file, "Sky spectral definition file";
   kwds_set,"FRACLIM", fraclim, "Fractional limit for pixel inclusion";
   strran = swrite(format="%.1f-%.1f", e1, e2);
   kwds_set,"SX_SUM", sx_sum,"[erg/cm2/s] ("+strran+" keV) Actual summed source intensity";
   kwds_set,"RADIUS", radius, "[deg] Inclusion radius";
   if( in_xspecmode ) {
      kwds_set,"MODE","XSPEC","Used XSPEC generated photon flux";
      kwds_set,"XSPMODEL",xspecmodel,"XSPEC spectral model";
   } else {
      kwds_set,"SC", sc, "Spectral code (used by mk_photflux)";
      if( is_void(nhmap) ) {
         kwds_set,"NH", nh, "[atoms/cm2] Column density for entire image";
      } else {
         kwds_set,"NHMAPDOL",nhdol,"Map of column densities used";
      }
      if( is_void(p1map) ) {
         kwds_set,"PARAM1",p1,"Spectral parameter used for entire image";
      } else {
         kwds_set,"P1MAPDOL",p1dol,"Map of spectral parameters used";
      }
   }

   wrmfitscols, skydef_file, "X_SKY", x_sky, "Y_SKY", y_sky, \
                "ENERGY", enera, "INTENS", intens, "DPHOT", dphota, "RENORM", renorma, \
                "DOL", dols, "RA_OBJ", ra_obj, "DEC_OBJ", dec_obj, \
                "NAME", name, clobber=1;
   write,format="%i rows in %s have been written\n", numberof(x_sky), skydef_file;
   if( am_subroutine() ) write,format="%s has been written\n", skydef_file;

   mt_log_entry, 5, "Going to exit mt_skyspec2skydef";

   return nsour;

}


/* Function mt_skyima2skyspec */

/************************************************
 *
 *   Convert count or countrate image to sky spectral
 *   definition file.
 *
 *    Information needed:
 *        ARF (effective area)
 *        Energy limits
 *        Spectral models i.e. spectral code, param1, nh
 *
 *    ARF must be in a FITS table with ENERG_LO, ENERG_HI
 *        and SPECRESP columns.
 ************************************************/


func mt_skyima2skyspec( dol_skyima, dol_arf, emin, emax, outfile, \
             fraclim=, xspecmodel=, xspecparlist=, tolmap=, sc=, nh=, p1=, enx= )
/* DOCUMENT mt_skyima2skyspec, dol_skyima, dol_arf, emin, emax, outfile,
            fraclim=, xspecmodel=, xspecparlist=, tolmap=, sc=, nh=, p1=, enx=

   Uses a skyimage, 'dol_skyima', with countrates per pixel to produce
   a sky spectral definition file (SSDF) - in short: a skyspec file,
   which defines all spectral parameters either as keywords or as maps.

   The ancillary response function (ARF) pertinent for the input sky image
   is copied to the output SSDF.

   The energy limits, 'emin' and 'emax', are also those for the input image.

   Keyword 'fraclim' defines the selection:
                    where(skyima > fraclim*max(skyima))
                    (defaults to 0.02)
   Keyword 'tolmap'  defines the tolerance for the applied parameter maps.
                    It is a 1D array with as many elements as there are
                    parameter maps, taken in the same order as the map appearance.
                    Recalculation of source photon flux is not done if
                    the map values changes less than the given tolerance.

   The keywords define a spectral model:
      For the 'simple' mode:
            sc   :  spectral code "PL" (default), "BB", or "TB"
            nh   :  column density, (defaults to 1e21), a scalar
                      value or a DOL to a map
            p1   :  spectral parameter, (defaults to 2.0), a scalar
                      value or a DOL to a map
                      -> for "PL"  the photon index
                      -> for "TB" and "BB" the temperature (kT) in keV
     For the XSPEC mode:
         xspecmodel  :  a string such as "phabs*power" (see XSPEC manual)

         xspecparlist  :  a yorick list with as many elements as the required
                          number of model parameters PLUS one. The first element
                          must be the number of the normalisation parameter.
                          The remainging elements are either
                          a (double) number for a constant value or a string
                          which is the DOL to map of its values.

                          Example 1: Using a 'mekal' model requires 6 parameters
                          where the last one is 'K' (in xspec parlance), that
                          acts as normalisation.
                          Assuming you have a temperature map then
                          xspecmodel = "mekal";
                          xspecparlist = _lst(6, "temp_map.fits+1", 0.21, 0.35, 0.02, 0, 1.);
                               - here the normalisation is the 6'th parameter and its value
                               - must be given as a scalar numeric value

                          Example 2: A 'tbabs*bremss' model
                          xspecmodel = "tbabs*bremss";
                          xspecparlist = _lst(3, "nh_map.fits+1", "temp_map.fits+1", 1.);
                          There is no way at the moment to change abundances (under planning, though).

                   The presence of both these keywords indicates that the
                   XSPEC mode is requested.

   Keyword for extra Sx calculation:
           enx    :   [e1,e2] (2 element array)
                      makes only sense if emin < e1 < e2 < emax

   SEE ALSO: mt_skyspec2skydef

   2010-08-18/NJW, 2013-05-23/NJW updated with XSPEC option

*/
{

   mt_log_entry, 5, "Entered mt_skyima2skyspec";

   // Read the input sky image and set up the normalisation map ('norm_map')
   //  The question is if this should really be the last map since the XSPEC
   //  convention seems to have 'K' as the last parameter in all spectral models.
   //   Can I use 'EXTNAME' for defining?

   skyima = readfits(dol_skyima);
   hdr_skyima = headfits(dol_skyima);
   wcs_skyima = get_wcs( hdr_skyima );
   dms = dimsof(skyima);
   norm_map = array(float,dms);

   mt_log_entry, 5, "##62## mt_skyima2skyspec, skyima has maximum "+ftoa(max(skyima),ndec=3,sci=1);

   // Locate the pixels that are going to be used (above a certain limit)

   if( is_void(fraclim) ) fraclim = 0.02;
   allowed = where(skyima > fraclim*max(skyima));
   nallowed = numberof(allowed);
   if( nallowed == 0 ) error,"A blank skyimage has been input!!";
   mt_log_entry, 5, "##63## mt_skyima2skyspec, nallowed = "+itoa(nallowed);

   // See if XSPEC mode is requested

   in_xspecmode = ( typeof(xspecmodel) == "string" && typeof(xspecparlist) == "list" );

   if( in_xspecmode ) {

      mt_log_entry, 5, "##64## mt_skyima2skyspec, in_xspecmod==1";
      nxlist = _len(xspecparlist); // since the first list elements points to the
                                   // parameter that is the normalisation (this avoids
                                   // an extra keyword) the number of elements is one
                                   // larger than the number of parameters.
      xspecparms = array(double, nxlist-1); // an array that will hold the parameters
                                            // relevant for the current pixel
      map_index = array(int,nxlist-1); // this index is zero if no map is assigned to a
                                     // parameter, otherwise it is the index in the
                                     // list which means that it may values between 2
                                     // and nxlist (both included)
      // Read in the parameter maps as they are encountered
      // the variable is 'parmscube', the third dimension is for the parameter number
      // Each layer in this cube will later be written to the SSDF in a separate extension.
      parmscube = norm_map(,,-); // prepare for adding a map to the cube
      for( i = 2; i <= nxlist; i++ ) {
         if( typeof(_car(xspecparlist,i)) == "string" ) { // we have a map
            write,"Going to read the map in "+_car(xspecparlist,i);
            // There are two options:
            //     1) There is no WCS information and the dimensions match the input image
            //     2) There _is_ WCS information and remapping is executed
            mapdol = _car(xspecparlist,i);
            map = readfits( mapdol );
            mapdms = dimsof( map );
            wcs_map = get_wcs( mapdol, nocheck=1 );
            if( wcs_map.flag == 0 ) { // Found no WCS info, test dimensions
               if( anyof(dms-mapdms) ) { // Exact dimension mismatch - reject
                  error,"MT_SKYIMA2SKYSPEC problem with map: "+mapdol;
               }
               mapout = map;
            } else {  // Yes, WCS was found, remap to image dimensions
               mapout = remap_wcs( map, wcs_map,"none" , wcs_skyima, dms(2:3) );
            }
            grow, parmscube, float(mapout);
            map_index(i-1) = dimsof(parmscube)(0);
            // map_index(j) is the third index of the parmscube for parameter j
         }
      }

   } else {

       // the 'simple' mode as in mt_rayor-4.4.6 and earlier versions

      if( is_void(sc) ) sc = "PL";
      if( is_void(nh) ) nh = 1.e21;
      if( typeof(nh) == "string" ) { // it must be a DOL
         nhdol = nh;
         write,"Going to read the map in "+nhdol;
         // There are two options:
         //     1) There is no WCS information and the dimensions match the input image
         //     2) There _is_ WCS information and remapping is executed
         nhmap = readfits( nhdol );
         nhdms = dimsof( nhmap );
         wcs_nh = get_wcs( nhdol, nocheck=1 );
         if( wcs_nh.flag == 0 ) { // Found no WCS info, test dimensions
            if( anyof(dms-nhdms) ) { // Exact dimension mismatch - reject
               error,"MT_SKYIMA2SKYSPEC problem with nhmap: "+nhdol;
            }
            mapout = nhmap;
         } else {  // Yes, WCS was found, remap to image dimensions
            mapout = remap_wcs( nhmap, wcs_nh, "none", wcs_skyima, dms(2:3) );
         }
         nhmap = mapout;
      } else {
         nhmap = [];
      }
      if( is_void(p1) ) p1 = 2.;
      if( typeof(p1) == "string" ) { // it must be a DOL
         p1dol = p1;
         write,"Going to read the map in "+p1dol;
         // There are two options:
         //     1) There is no WCS information and the dimensions match the input image
         //     2) There _is_ WCS information and remapping is executed
         p1map = readfits( p1dol );
         p1dms = dimsof( p1map );
         wcs_p1 = get_wcs( p1dol, nocheck=1 );
         if( wcs_p1.flag == 0 ) { // Found no WCS info, test dimensions
            if( anyof(dms-p1dms) ) { // Exact dimension mismatch - reject
               error,"MT_SKYIMA2SKYSPEC problem with p1map: "+p1dol;
            }
            mapout = p1map;
         } else {  // Yes, WCS was found, remap to image dimensions
            mapout = remap_wcs( p1map, wcs_p1, "none", wcs_skyima, dms(2:3) );
         }
         p1map = mapout;
      } else {
         p1map = [];
      }

   }

   energ_lo = rdfitscol(dol_arf,"energ_lo");
   energ_hi = rdfitscol(dol_arf,"energ_hi");
   specresp = rdfitscol(dol_arf,"specresp");
   eline = sqrt(energ_lo*energ_hi);

//////   Define groups with almost same parameter value in the maps applied ///////////

   gn = indgen(nallowed); // default, in case no grouping has been requested

   if( in_xspecmode ) {
      num_maps = numberof(where(map_index));
      if( num_maps ) {
         if( !is_void(tolmap) ) { // only do the grouping if the tolerance has been defined
            k = 1;
            gn = [];
            for( j = 1; j <= numberof(xspecparms); j++ ) {
               if( map_index(j) ) { // there is a map for this parameter
                  xxx = parmscube(,,map_index(j));
                  if( is_void(gn) ) {
                     gn = groupn( xxx, allowed, tolmap(k));
                  } else {
                     gn2 = groupn( xxx, allowed, tolmap(k));
                     Ngn2 = max(gn2);
                     gn = (gn-1)*Ngn2 + gn2;
                     // reduce the grouping number
                     ggg = gn(sort(gn));
                     hhh = ggg(uniq(ggg));
                     for(i = 1; i <= numberof(hhh); i++ ) {
                        gn( where( gn == hhh(i) ) ) = i;
                     }
                  }
                  k++;
               }
            }
         }
      } else { // no maps have been defined - only one group
         gn = array(int(1),nallowed);
      }

   } else { // in simple mode

      num_maps = !is_void(nh_map) + !is_void(p1_map);

      if( num_maps ) {
         if( !is_void(tolmap) ) { // only do grouping if the tolerance has been defined
            // NH map is first, p1 map is second
            k = 1;
            gn = [];
            if( !is_void(nh_map) ) {
               gn = groupn( nh_map, allowed, tolmap(k) );
               k++;
            }
            if( !is_void(p1_map) ) {
               if( is_void(gn) ) {
                  gn = groupn( p1_map, allowed, tolmap(k));
               } else {
                  gn2 = groupn( p1_map, allowed, tolmap(k));
                  Ngn2 = max(gn2);
                  gn = (gn-1)*Ngn2 + gn2;
                  // reduce the grouping number
                  ggg = gn(sort(gn));
                  hhh = ggg(uniq(ggg));
                  for(i = 1; i <= numberof(hhh); i++ ) {
                     gn( where( gn == hhh(i) ) ) = i;
                  }
               }
            }
         }
      } else { // no maps have been defined - only one group
         gn = array(int(1),nallowed);
      }
   }

//////   End of: Define groups with almost same parameter value in the maps applied ///////////
         
   first = 1;
   tot_flux = [];

   // Loop over all 'active' pixels in the order of increasing group number
   //  in order to skip duplicating calculation with almost identical parameters

   ign = sort(gn);
   group_num_prev = 0;

   for( k = 1; k <= nallowed; k++ ) { // the 'k' index runs with increasing group number
      // group number = gn(ign(k)) starts with smallest group number
      group_num_now = gn(ign(k));
      i = ign(k);
      ixx = (allowed(i)-1)%dms(2)+1;
      iyy = (allowed(i)-1)/dms(2)+1;
      if( in_xspecmode ) {
         idxnorm = _car(xspecparlist,1); // first element is index of norm
         for( j = 1; j <= numberof(xspecparms); j++ ) {
            if( map_index(j) ) { // there is a map for this parameter
               xspecparms(j) = parmscube(ixx,iyy,map_index(j));
            } else xspecparms(j) = _car(xspecparlist,j+1);
         }
      } else {
         if( !is_void(nhmap) ) nh = nhmap(allowed(i));
         if( !is_void(p1map) ) p1 = p1map(allowed(i));
      }
      // generate a photon flux in the energy interval relevant for
      // the instrument that has produced the sky image.
      // Set normalisation to one in order to calculate the proper
      // one to save into the SSDF normalisation extension.

      if( in_xspecmode ) xspecparms(idxnorm) = 1.0;

      if( group_num_prev != group_num_now ) { // Only do recalculation when
                  // a new group is encountered; else reuse previous 'Flux'

         group_num_prev = group_num_now;

         mk_photflux,sc=sc,nh=nh,p1=p1,norm=1.,e1=emin,e2=emax,\
               xspecmodel=xspecmodel,xspecparms=xspecparms, \
               nchan=100,nof=1,silent=1;

         // produces Eline, Flux, Sx_photflux
         if( first ) arf = interp(specresp,eline,Eline);
         // calculate countrate expected from a norm=1 spectrum
         I = sum(Flux(zcen)*arf(zcen)*Eline(dif));

      }
      k_norm = skyima(allowed(i))/I;
      norm_map(allowed(i)) = k_norm;
      if( in_xspecmode ) parmscube(ixx,iyy,1) = k_norm;
      if( is_void(tot_flux) ) {
         tot_flux = k_norm * Flux;
      } else {
         tot_flux += k_norm * Flux;
      }
      first = 0;
   }
   sxtot = sflux(emin,emax,Eline,tot_flux);
   write,format="Sx from entire image: %.3e erg/s/cm2 in %.1f-%.1f keV\n", \
                 sxtot, emin, emax;
   sxtotx = [];
   if( numberof(enx) == 2 ) {
      sxtotx = sflux(enx(1),enx(2),Eline,tot_flux);
      write,format="Sx from entire image: %.3e erg/s/cm2 in %.1f-%.1f keV\n", \
              sxtotx, enx(1), enx(2);
   }

   kwds_init;
   kwds_set,"EXTNAME","NORM_MAP","Name of this extension";
   kwds_set,"SKYIMAIN",dol_skyima,"DOL of input sky ima";
   kwds_set,"ARFDOL",dol_arf,"DOL of applied ARF";
   kwds_set,"FRACLIM",fraclim,"selects: > fraclim*max(ima)";
   kwds_set,"E_MIN",emin,"[keV] Lower energy limit";
   kwds_set,"E_MAX",emax,"[keV] Upper energy limit";
   if( in_xspecmode ) {
      kwds_set,"MODE","XSPEC","Mode chosen for the simulation";
      kwds_set,"XSPMODEL", xspecmodel,"XSPEC spectral model";
      kwds_set,"IDXNORM", idxnorm,"Index of parameter list for norm";
      for( i = 1; i <= numberof(xspecparms); i++ ) {
         kwds_set,"PARAM"+itoa(i),_car(xspecparlist,i+1),"Spectral parameter "+itoa(i);
      }
   } else {
      kwds_set,"MODE","SIMPLE","Mode chosen for the simulation";
      kwds_set,"SC",sc,"Spectral code PL, BB, or TB";
      if( !is_void(p1map) ) {
         kwds_set,"P1MAP",p1dol,"DOL of applied p1 map";
      } else {
         kwds_set,"PARAM1", p1,"Value of param1 for entire image";
      }
      if( !is_void(nhmap) ) {
         kwds_set,"NHMAP",nhdol,"DOL of applied nh map";
      } else {
         kwds_set,"NH", nh,"Value of nh for entire image";
      }
   }
   kwds_set,"SXTOTAL",sxtot,"[erg/cm2/s] from E_MIN to E_MAX";
   if( !is_void(sxtotx) ) {
      s1 = swrite(format="%.2f", enx(1));
      s2 = swrite(format="%.2f", enx(2));
      kwds_set,"SXTOTALX",sxtotx,"[erg/cm2/s] from "+s1+" to "+s2+" keV";
   }

   // Copy coordinate defining keywords

   clist = "CRVAL1,CRVAL2,CRPIX1,CRPIX2,CDELT1,CDELT2,";
   clist += "CROTA2,CTYPE1,CTYPE2,CUNIT1,CUNIT2,";
   clist += "CD1_1,CD1_2,CD2_1,CD2_2";
   fits_copy_keys,hdr_skyima,list=clist, tokwds=2;
       // tokwds is given a value of 2 to avoid initialization of keywords

   fh = writefits(outfile,norm_map,clobber=1,cont=1);

   if( in_xspecmode ) {

      // Add the maps from parmscube

      mapnum = 0;
      for( i = 1; i <= numberof(xspecparms); i++ ) {
         if( map_index(i) ) {
            mapnum++;
            if( !is_void(tolmap) ) kwds_set,"TOLMAP", tolmap(mapnum),"Tolerance for this map";
            kwds_set,"EXTNAME","PARAM"+itoa(i),"Name of this extension";
            fh = writefits( fh, parmscube(,,map_index(i)),cont=1);
         }
      }

   } else {

      // Add the p1map array (if defined)

      mapnum = 0;
      if( !is_void(p1map) ) {
         mapnum++;
         if( !is_void(tolmap) ) kwds_set,"TOLMAP", tolmap(mapnum),"Tolerance for this map";
         kwds_set,"EXTNAME","P1MAP","Name of this extension";
         fh = writefits( fh, p1map,cont=1);
      }

      // Add the nhmap array (if defined)

      if( !is_void(nhmap) ) {
         mapnum++;
         if( !is_void(tolmap) ) kwds_set,"TOLMAP", tolmap(mapnum),"Tolerance for this map";
         kwds_set,"EXTNAME","NHMAP","Name of this extension";
         fh = writefits( fh, nhmap, cont=1);
      }

   }

   // Add the ARF

   kwds_set,"EXTNAME","SPECRESP","Name of this extension";
   fh = wrmfitscols( fh, "ENERG_LO",energ_lo,"ENERG_HI",energ_hi,"SPECRESP",specresp);
}


/* Function mt_roll */

func mt_roll( roll_angle )
/* DOCUMENT mt_roll, roll_angle

   Executes a rotation of all the photons in 'Phs' around the z-axis.
   roll_angle is in radians. Please note that this is the opposite angle
   of the optic rotation.

   The roll angle is saved as an external variable: Roll_phot (in radians).

   Update 2017-12-13/NJW : Added the rotation of Phs.E1 and Phs.E2 as well.
*/
{
   extern Phs, Roll_phot;

   Roll_phot = roll_angle;

   eq_nocopy, cE, Phs.E;
   eq_nocopy, cR, Phs.R;
   eq_nocopy, cE1, Phs.E1;
   eq_nocopy, cE2, Phs.E2;

   omega = [[cos(roll_angle),sin(roll_angle),0], \
            [-sin(roll_angle),cos(roll_angle),0], \
             [0,0,1]]; // defines the rotation around z
   Phs.E = omega(,+)*cE(+,);
   Phs.R = omega(,+)*cR(+,);
   Phs.E1 = omega(,+)*cE1(+,);
   Phs.E2 = omega(,+)*cE2(+,);
}

/* Function mt_photpr */

func mt_photpr( iphot )
/* DOCUMENT mt_photpr, iphot

    Nice screen print of photon number 'iphot' or s_Ray element
*/
{
   if( typeof(iphot) == "struct_instance" ) {
      p = iphot;
   } else if( typeof(iphot) == "int" || typeof(iphot) == "long" ) {
      p = Phs(iphot);
   } else {
      write,"Illegal argument";
      return;
   }
   write,"Element  Meaning                 Value(s)";
   write,format="(status)    Status                : %6i\n", p.status;
   write,format="(mirror)    Mirror                : %6i\n", p.mirror;
   write,format="(bounce)    Bounce                : %6i\n", p.bounce;
   write,format="(flag)      Flag                  : %6i\n", p.flag;
   write,format="(energy)    Energy                : %10.3f keV\n", p.energy;
   write,format="(rcoef)     Rcoef                 :   %10.5f\n\n", p.rcoef;
   write,format="(E)         Current position      : %10.3f%10.3f%10.3f\n", p.E(1), p.E(2), p.E(3);
   write,format="(E1)        Entrance OM1 position : %10.3f%10.3f%10.3f\n", p.E1(1), p.E1(2), p.E1(3);
   write,format="( - )       Azimuth               : %12.5f rad\n", zero2pi(atan(p.E1(2),p.E1(1)));
   write,format="(E2)        Entrance OM2 position : %10.3f%10.3f%10.3f\n", p.E2(1), p.E2(2), p.E2(3);
   write,format="( - )       Azimuth               : %12.5f rad\n", zero2pi(atan(p.E2(2),p.E2(1)));
   write,format="(I1)        Reflection OM1 point  : %10.3f%10.3f%10.3f\n", p.I1(1), p.I1(2), p.I1(3);
   if( p.I1(1) != 0.0 || p.I1(2) != 0.0 ) \
   write,format="( - )       Azimuth               : %12.5f rad\n", zero2pi(atan(p.I1(2),p.I1(1)));
   write,format="(I2)        Reflection OM2 point  : %10.3f%10.3f%10.3f\n", p.I2(1), p.I2(2), p.I2(3);
   if( p.I2(1) != 0.0 || p.I2(2) != 0.0 ) \
   write,format="( - )       Azimuth               : %12.5f rad\n", zero2pi(atan(p.I2(2),p.I2(1)));
   write,"";
   write,format="(R)         Current direction     :   %10.5f%10.5f%10.5f\n", p.R(1), p.R(2), p.R(3);
   write,format="(D1)        Entrance OM1 direction:   %10.5f%10.5f%10.5f\n", p.D1(1), p.D1(2), p.D1(3);
   write,format="(D2)        Entrance OM2 direction:   %10.5f%10.5f%10.5f\n\n", p.D2(1), p.D2(2), p.D2(3);
   write,format="(angle_in1) Reflection OM1 angles :   %10.5f%10.5f mrad\n", p.angle_in1*1000., \
                       p.angle_out1*1000.;
   write,format="(angle_in2) Reflection OM2 angles :   %10.5f%10.5f mrad\n", p.angle_in2*1000., \
                       p.angle_out2*1000.;
}

/* Function is_not_defined */

func is_not_defined ( x )
{
   if( is_void(x) ) return 1;
   if( structof(x) == string ) {
      if( x == "[]" ) return 1;
   }
   return 0;
}

/* Function mt_eff_area_quick */

func mt_eff_area_quick( earr=, outfile= )
/* DOCUMENT eff_area = mt_eff_area_quick( earr=, outfile= )

   Calculate the telescope effective area from the optical module
   files and scatter+reflection informations.
   Go through all mirrors and sum the mirror contributions.

   Default energy array is 'E_uniq' (external variable)

   If the keyword 'outfile' is not set or set to zero then
   no output file will be written. If it is given as a string
   then this will be the output file name, else the standard
   naming as eff_area_quick_nnnn.fits, where 'nnnn' is a serial number,
   will be used.

   The correction for the blocking by spokes is included if
   'mt_get_mirror_eff_factors' is run before this function.

   Version 3.5     2011-08-04/NJW
*/

{

// ynclude = zbe
  extern  Om_files, Num_modules, N_mirrors, Coat_list, Mirror_coating, \
          R1arr, R2arr, Z1arr, Z2arr, E_uniq, Mirror_eff_factors, \
          Telescop, Instrume, Mirror_area;
// yxclude =

   if( is_void(earr) ) {
      if( is_void(E_uniq) ) {
         n_earr = 50;
         earr = span(2.,79, n_earr);
      } else {
         n_earr = numberof(E_uniq);
         earr = E_uniq;
      }
   } else {
      n_earr = numberof(earr);
   }



   //  Sum the reflection coefficients
   //    not so useful when no scattering/refl tables have been read in

   mt_load, omfile=Om_files(1);
   if( is_not_defined(Coat_list) ) {
      write,"Coat_list has not been defined, no action! ...";
      return [];
   }

   Mirror_area = array(double, N_mirrors);
   eff_area_m = array(double, n_earr, N_mirrors);
   mir_angle1 = array(double, N_mirrors);

   //  Loop over all mirrors

   for( mirror = 1; mirror <= N_mirrors; mirror++ ) {
      coating = Mirror_coating(mirror);
      q = where( Coat_list.id == coating );
      if( numberof(q) == 0 ) error,"##65## coating problem";
      required_scatfile = Coat_list(q(1)).file;
      if( Scatter_file != required_scatfile ) {
         mt_load,scatfile=required_scatfile;
      }
      mir_angle1(mirror) = \
        atan((R1arr(mirror) - R2arr(mirror))/(Z1arr(mirror) - Z2arr(mirror)));
      Mirror_area(mirror) = 2 * pi * (R1arr(mirror) - R2arr(mirror)) \
                    * (0.5*(R1arr(mirror) + R2arr(mirror)));

      // Loop over energies

      for( j = 1; j <= n_earr; j++ ) {
         refl_coef = mt_get_rcoef( earr(j), mir_angle1(mirror) );
         eff_area_m( j, mirror ) = Mirror_area(mirror) * refl_coef;
      }
   }

   if( Num_modules > 1 ) { // include the second reflection

      mt_load, omfile=Om_files(2);

      //  Loop over all mirrors

      for( mirror = 1; mirror <= N_mirrors; mirror++ ) {
         coating = Mirror_coating(mirror);
         q = where( Coat_list.id == coating );
         if( numberof(q) == 0 ) error,"##66## coating problem";
         required_scatfile = Coat_list(q(1)).file;
         if( Scatter_file != required_scatfile ) {
            mt_load,scatfile=required_scatfile;
         }
         // account for the first reflection
         mir_angle = \
            atan((R1arr(mirror) - R2arr(mirror))/(Z1arr(mirror) - Z2arr(mirror)));
         angle = mir_angle - 2*mir_angle1(mirror);

         // Loop over energies

         for( j = 1; j <= n_earr; j++ ) {
            refl_coef = mt_get_rcoef( earr(j), angle );
            eff_area_m( j, mirror ) *= refl_coef;
         }
      }
   }

   if( numberof(Mirror_eff_factors) == N_mirrors ) {
      eff_area_m *= Mirror_eff_factors(-:1:n_earr,);
   } else {
      write,"NOT corrected for shielding by spokes";
      write,"You can run 'mt_get_mirror_eff_factors' to introduce";
      write,"the correction and then 'mt_eff_area_quick' again.";
   }
   eff_area = eff_area_m(,sum);

   if( is_set(outfile) ) {

      if( structof(outfile) != string ) \
         outfile = get_next_filename("eff_area_quick_????.fits");

      kwds_init;
      kwds_set,"DATE",ndate(3),"Time of file creation";
      kwds_set,"ORIGIN","mt_eff_area_quick of MT_RAYOR-"+Version,"Software used";
      kwds_set,"RESPONSI","Niels J. Westergaard","Responsible programmer";
      kwds_set,"SITE","NSI/DTU","Institution";
      if( !is_void(Telescop) ) kwds_set,"TELESCOP", Telescop,"Name of telescope or mission";
      if( !is_void(Instrume) ) kwds_set,"INSTRUME", Instrume,"Name of instrument";
      kwds_set,"TUNIT1","keV","Unit of first column: ENERGY";
      kwds_set,"TUNIT2","mm2","Unit of second column: EFF_AREA";
      if( numberof(Mirror_eff_factors) == N_mirrors ) {
         kwds_set,"COMMENT","Corrected for shielding by spokes";
      } else {
         kwds_set,"COMMENT","NOT corrected for shielding by spokes";
      }

      wrmfitscols,outfile,"ENERGY",earr,"EFF_AREA",eff_area,clobber=1;
      write,format="%s has been written\n", outfile;
      write,"with columns ENERGY and EFF_AREA";
   }

   return eff_area;
}

/* Function mt_get_mirror_eff_factors */

func mt_get_mirror_eff_factors
/* DOCUMENT mt_get_mirror_eff_factors

    Fills the external array 'Mirror_eff_factors' which holds
    each individual relative throughput not blocked by spokes.
    Only for on-axis rays so no concern about 'full length' spokes.

    Takes no arguments. Mainly useful for 'mt_eff_area_quick'.

    SEE ALSO: mt_eff_area_quick
*/
{
// ynclude = zbf
  extern N_mirrors, Mirror_eff_factors, Spoke_define_files, Om_files, \
         R1arr, R2arr, Module_num;
// yxclude =

   Mirror_eff_factors = array(1.0,N_mirrors); // Unity if no spokes are defined

   if( Spoke_define_files(1) != "none" ) {
      Module_num = 1;      // used by 'mt_spoke_read'
      mt_spoke_read,pos=1; // Load the entrance spokes
      mt_load,omfile=Om_files(1); // Load the first optical module


      // Loop over mirrors

      for( mirror = 1; mirror <= N_mirrors; mirror++ ) {
         radius = 0.5*(R1arr(mirror) + R2arr(mirror));
         // make 0.1 mm steps along circumference
         step = 0.1; // mm
         phi = span(0., 2*pi, long(2*pi*radius/step)+1)(1:-1);
         x = radius * cos(phi);
         y = radius * sin(phi);
         nphi = numberof(phi);
         block = 0;
         for( i = 1; i <= nphi; i++ ) {
            if( mt_spoke_blocking([x(i),y(i)],pos=1)(1) ) block++;
         }
         Mirror_eff_factors(mirror) = 1. - double(block)/nphi;
      }
   }
}

/* Function mt_bg_run */

func mt_bg_run( system_file, energy, offaxis, azimuth, no_scatter=, no_mdeform=, \
                dphot=, renorm=, fraper=, labxoff=, labyoff=, roll=, gbend=, \
                exposure=, ncases=, dir=, nproc=, blur= )
/* DOCUMENT filelist_file = mt_bg_run( system_file, energy, offaxis, azimuth, \
                no_scatter=, no_mdeform=, \
                dphot=, renorm=, fraper=, labxoff=, labyoff=, roll=, gbend=, \
                exposure=, ncases=, dir=, nproc=, blur= )

     This function will start 'mt_run' in the background in several incarnation
     and then assemble all the Phs arrays finally. This takes advantage of a Linux
     machine with several CPUs. Returns name of file that holds the list of
     individual batch calls.

     Arguments:
     'system_file' is the argument for mt_setup_system
     'energy', 'offaxis', 'azimuth' as for an ordinary call of mt_run

     Keywords:
     'no_scatter', 'no_mdeform', 'gbend', 'dphot', and 'renorm' as for mt_run
     'fraper', 'labxoff', 'labyoff', and 'roll' as for mt_run
     'ncases' is the number of individual mt_run's (default: 20)
     'dir'    is the directory for the batch files and save files
     'nproc'  is the number of simultaneous processes (default: 10)
     'exposure' is the exposure when a flux file is given rather than an energy.
                The value is divided by 'ncases' before subprocesses are called.
     'blur'   [rad] Blur angle for the input source.

     'dphot' and 'renorm' apply to each individual run.

     Upon completion the saved 'Phs' arrays are assembled and saved into
     the external variable 'Phs'.

*/
{

// ynclude = eehd
   extern Phs, R_inner_design, R_outer, Fraper, Fraper_area, Dphot, \
          Src_offaxis, Src_azimuth, Use_mdeform, Use_scatter, Exposure, \
          Dec_scx, Ra_scx, Posang;
// yxclude =

   if( OSTYPE == "nonx" ) {
      write,"Sorry, this command is not valid for current OS";
      return [];
   }
   ncases = is_void(ncases) ? 10 : ncases;
   dir = is_void(dir) ? "." : dir;
   dir = app_slash( dir);
   nproc = is_void(nproc) ? 12 : nproc;

   if( !is_void(dphot) ) dphot = double(dphot);
   if( !is_void(renorm) ) renorm = double(renorm);
   list_name = get_next_filename("bg_run_????.list",dir=dir);
   fh = open( list_name, "w" );

   if( is_void(fraper) ) {
      mt_setup_system,system_file;
      fraper = [R_inner_design, R_outer, 0., 360. ];
   }
   Fraper = fraper;
   Fraper_area = pi * (fraper(2)^2 - fraper(1)^2) * (fraper(4)-fraper(3))/360.;

   // Set externals for later call of e.g. mt_save

   if( !is_void(dphot) ) Dphot = dphot * ncases;
   if( !is_void(offaxis) ) Src_offaxis = offaxis;
   if( !is_void(azimuth) ) Src_azimuth = azimuth;
   Use_scatter = no_scatter ? 0 : 1;
   Use_mdeform = no_mdeform ? 0 : 1;

   // If "energy" is the name of a sky definition file, then the action will be
   // to divide it into ncases shorter files and run each of these.

   is_skydefinition = 0;
   if( typeof(energy) == "string" ) { // we have a file
      if( !file_test(energy) ) error,"File not found: "+energy;
      hdr = headfits(energy+"+1");
      is_skydefinition = (fxpar(hdr,"extname") == "SKY_DEFINITION");
   }

   if( is_skydefinition ) {
      // split into sub-files
      stem = dir+"sub_skydef_"+ranstr(6);
      newsubnames = fits_binsplit( energy+"+1", ncases, stem=stem, colname="renorm" );
      ncases = numberof(newsubnames); // if 'renorm' has a strange distribution of values
                           // fewer files might have been produced
      // Extract values for the WCS to use by mt_save:
      hdr = headfits(energy+"+1");
      Ra_scx = fxpar(hdr,"ra_scx");
      Dec_scx = fxpar(hdr,"dec_scx");
      Posang = fxpar(hdr,"posang");

      Exposure = exposure;

      for( i = 1; i <= ncases; i++ ) {
         outstr = "_"+ranstr(15);
         write,fh,format="%s\n", dir+"y"+outstr+".ysav";
         _mt_bg_run, system_file, outstr, dir, \
                  newsubnames(i), \
                  no_scatter=no_scatter, no_mdeform=no_mdeform, gbend=gbend, \
                  dphot=dphot, renorm=renorm, fraper=fraper, labxoff=labxoff, \
                  labyoff=labyoff, roll=roll, exposure=exposure, blur=blur;
         while( n_yoricks() > nproc ) {
            write,"Wait a bit, currently too many yorick processes ...";
            pause, 30000;
         }
      }

   } else { // Standard running

      if( is_void(exposure) ) {
         expo = exposure;
      } else {
         expo = exposure/ncases;
         Exposure = exposure;
      }
   
      for( i = 1; i <= ncases; i++ ) {
         outstr = "_"+ranstr(15);
         write,fh,format="%s\n", dir+"y"+outstr+".ysav";
         _mt_bg_run, system_file, outstr, dir, \
                  energy, offaxis, azimuth, \
                  no_scatter=no_scatter, no_mdeform=no_mdeform, gbend=gbend, \
                  dphot=dphot, renorm=renorm, fraper=fraper, labxoff=labxoff, \
                  labyoff=labyoff, roll=roll, exposure=expo, blur=blur;
         while( n_yoricks() > nproc ) {
            write,"Wait a bit, currently too many yorick processes ...";
            pause, 30000;
         }
      }

   }
   close, fh;

   // Re assemble the 'Phs' array

   filenames = rdfile(list_name);
   n = numberof(filenames);

   assembled_Phs = [];

   while( n > 0 ) {
      for( i = 1; i <= n; i++ ) {
         if( !file_test(filenames(i)) ) continue;
         // heureca ! found one
         pause, 2000; // give time to finish writing
         write,"Getting data from "+filenames(i)+" ...";
         fh = openb( filenames(i) );
         restore, fh, Phs;
         close, fh;
         grow, assembled_Phs, Phs;
         filenames = rem_elem( filenames, i );
         n--;
      }
   }

   //+ eq_nocopy, Phs, assembled_Phs;
   Phs = assembled_Phs;
   write,"Got all photon data ...";

   return list_name;
}

/* Function _mt_bg_run */

func _mt_bg_run( system_file, outfile, dir, energy, offaxis, azimuth, \
                no_scatter=, no_mdeform=, dphot=, renorm=, fraper=, gbend=, \
                labxoff=, labyoff=, roll=, exposure=, blur= )
/* DOCUMENT _mt_bg_run, system_file, outfile, dir, energy, offaxis, azimuth, \
                no_scatter=, no_mdeform=, dphot=, renorm=, fraper=, gbend=, \
                labxoff=, labyoff=, roll=, exposure=, blur=

    Produces a batch file '<dir>/ymbr_<random string>.ymbr' and a yorick save
    file '<dir>/ytmp_<random string>.ytmp' holding the resulting Phs array.

    The batch file is launched in the background.

    Note that 'outfile' is only the basic name, becomes y<outfile>.ysav

    SEE ALSO: mt_bg_run

    2011-09-22/NJW
*/
{
   fname = app_slash(dir)+"ymbr_"+ranstr(15)+".ymbr";
   fn = open(fname,"w");
   badir = get_env("HOSTSITE") == "CALTECH" ? "/users/njw/yorick" : "/home/njw/yorick";
   write,fn,format="#include \"%s/common.id\"%s",badir,"\n";
   write,fn,format="#include \"%s/mt_rayor.i\"%s",badir,"\n";
   write,fn,format="%s\n","mt_setup_system,\""+system_file+"\";";

   // build the command

   comm = "mt_run,";
   s = typeof(energy) == "string" ? "\""+energy+"\"" : ftoa(energy,ndec=3);
   // First mt_run argument: energy
   comm += (s+","); // remember comma between parts
   // Second and third argument: offaxis and azimuth
   if( is_void(offaxis) ) offaxis = 0.0;
   if( is_void(azimuth) ) azimuth = 0.0;
   comm += (ftoa(offaxis,ndec=3)+","+ftoa(azimuth,ndec=3));
   // From now on we have only keywords with preceding commas
   if( no_scatter ) comm += (",no_scatter="+itoa(no_scatter));
   if( no_mdeform ) comm += (",no_mdeform="+itoa(no_mdeform));
   if( numberof(dphot) ) comm += (",dphot="+ftoa(dphot,ndec=3));
   if( numberof(exposure) ) comm += (",exposure="+ftoa(exposure,ndec=3));
   if( numberof(renorm) ) comm += (",renorm="+ftoa(renorm,ndec=3));
   if( numberof(gbend) ) comm += (",gbend="+ftoa(gbend,ndec=4));
   if( numberof(blur) ) comm += (",blur="+ftoa(blur,ndec=4,sci=1));
   if( numberof(fraper) ) {
      n = numberof(fraper);
      s = ",fraper=[";
      for(i=1;i<=n;i++) {
         s += ftoa(fraper(i),ndec=4);
         if( i < n ) s += ",";
      }
      s += "]";
      comm += s;
   }
   if( numberof(labxoff) ) comm += (",labxoff="+ftoa(labxoff,ndec=3));
   if( numberof(labyoff) ) comm += (",labyoff="+ftoa(labyoff,ndec=3));
   if( numberof(roll) ) comm += (",roll="+ftoa(roll,ndec=3));

   write,fn,format="%s,no_par=1;\n", comm;

   write,fn,format="mt_save,mode=\"s\",outfile=\"%s\",dir=\"%s\";\n", outfile, rem_slash(dir);
   write,fn,format="%s\n","quit";
   close, fn;
   system,"yorick -batch "+fname+" &";
}

/* Function mt_add_ysaves */

func mt_add_ysaves( file=, list=, reset= )
/* DOCUMENT mt_add_ysaves, file=, list=, reset=

    Add Phs's saved as 'save,fh,Phs'.

    Keywords:
       file     :  String (scalar or array) with text file(s)
                   containing ysave filenames.
       list     :  String (scalar or array) with ysave filenames.
       reset    :  Will reset the Phs, else Phs will be expanded.

    2011-09-23/NJW
*/
{
   extern Phs;

   if( numberof(file) && numberof(list) ) error,"Use 'file' xor 'list'";
   if( !(numberof(file) || numberof(list)) ) error,"Use one of 'file' or 'list'";

   filenames = [];

   if( !is_void(file) ) {
      for( i = 1; i <= numberof(file); i++ ) {
         if( !file_test(file(i)) ) {
            write,"Did not find "+file(i);
            continue;
         }
         grow, filenames, rdfile(file(i));
      }
   } else {
      filenames = list;
   }

   // Re assemble the 'Phs' array

   n = numberof(filenames);

   if( reset ) Phs = [];
   assembled_Phs = Phs;

   for( i = 1; i <= n; i++ ) {
      if( !file_test(filenames(i)) ) {
         write,"Failed to find "+filenames(i);
         continue;
      }
      write,"Getting data from "+filenames(i)+" ...";
      fh = openb( filenames(i) );
      restore, fh, Phs;
      close, fh;
      grow, assembled_Phs, Phs;
   }

   Phs = assembled_Phs;

}

/* Function mt_bg_run_eff_area */

func mt_bg_run_eff_area( energy, offaxis, azimuth, system_file, \
              no_scatter, no_mdeform, dphot, earr, outfile=, ncases=, dir=, nproc=, aps= )
/* DOCUMENT eff_area = mt_bg_run_eff_area( energy, offaxis, azimuth, system_file, \
              no_scatter, no_mdeform, dphot, earr, outfile=, \
              ncases=, dir=, nproc=, aps= )

    Running a number of background processes for the effective area determination.

    Keywords:
      outfile :  Name of output file. Can be given as '1' in which case the name
                      becomes the next of eff_area_????.fits.
      ncases  :  Number of runs of the eff. area determination [default 20]
         dir  :  Directory for temporary files [default '.']
       nproc  :  Maximum number of simultaneous processes [default 20]
         aps  :  Arguments for the 'mt_aperture_stop' call, e.g. "833.187,29.0"
*/
{
// ynclude = rwwe
   extern Telescop, Instrume;
// yxclude =

   if( is_void(ncases) ) ncases = 20;
   if( is_void(nproc) ) nproc = 20;
   if( is_void(dir) ) dir = ".";
   dir = app_slash(dir);
   energy = double(energy);
   offaxis = double(offaxis);
   azimuth = double(azimuth);
   earr = double(earr);
   dphot = double(dphot);

   local serstrf;
   first_file = get_next_filename("mbr_eff_area_?????.fits",serstrf,dir=dir);
   s = atoi(serstrf);
   ss = indgen(s:s+ncases-1);
   intermediate_files = dir+"mbr_eff_area_"+itoa(ss,5)+".fits";

   for( i = 1; i <= ncases; i++ ) {
      _mt_bg_run_eff_area, energy, offaxis, azimuth, system_file, \
               no_scatter, no_mdeform, dphot, earr, intermediate_files(i), dir, aps=aps;
      while( n_yoricks() > nproc ) {
         write,"Waiting for free processes ...";
         pause, 30000;
      }
   }

   // Add up the individual effective area files

   n = ncases;

   assembled_Phs = [];

   while( n > 0 ) {
      for( i = 1; i <= n; i++ ) {
         if( !file_test(intermediate_files(i)) ) continue;
         // heureca ! found one
         pause, 2000; // give time to finish writing
         write,"Getting data from "+intermediate_files(i)+" ...";

         hdr = headfits( intermediate_files(i)+"+1" );
         area = rdfitscol(intermediate_files(i)+"+1","EFF_AREA");
         if( n == ncases ) {
            nphot_in = fxpar( hdr, "NPHOT_IN");
            nphotons = fxpar( hdr, "NPHOTONS");
            eff_area = area(,-);
            ap_area = fxpar( hdr, "AP_AREA");
         } else {
            grow, eff_area, area;
            nphot_in += fxpar( hdr, "NPHOT_IN");
            nphotons += fxpar( hdr, "NPHOTONS");
         }

         intermediate_files = rem_elem( intermediate_files, i );
         n--;
      }
      write,"Waiting for eff. area file ...";
      pause,2000;
   }

   // Got all data, make average and write to FITS file

   eff_area_avg = eff_area(,avg); // is now average
   eff_area_err = eff_area(,rms)/sqrt(ncases); // is now error

   if( is_set(outfile) ) {

      if( structof(outfile) != string ) \
         outfile = get_next_filename("eff_area_????.fits");

      kwds_init;
      kwds_set,"DATE",ndate(3),"Time of file creation";
      kwds_set,"ORIGIN","mt_bg_run_eff_area of MT_RAYOR-"+Version,"Software used";
      kwds_set,"RESPONSI","Niels J. Westergaard","Responsible programmer";
      kwds_set,"SITE","NSI/DTU","Institution";
      if( !is_void(Telescop) ) kwds_set,"TELESCOP", Telescop,"Name of telescope or mission";
      if( !is_void(Instrume) ) kwds_set,"INSTRUME", Instrume,"Name of instrument";
      if( typeof(energy) == "double" ) {
         kwds_set,"ENERGYIN", energy, "[keV] Energy of injected photons";
      } else {
         kwds_set,"COMMENT","A photon flux file was used as input";
      }
      if( !is_void(offaxis) ) kwds_set,"SRC_OFAX",offaxis,"[arcmin] Source off axis angle";
      if( !is_void(azimuth) ) kwds_set,"SRC_AZIM",azimuth,"[degrees] Source azimuth angle";
      kwds_set,"NPHOT_IN",nphot_in,"Total number of injected photons";
      kwds_set,"NPHOTONS",nphotons,"Total number of accepted photons";
      kwds_set,"AP_AREA",ap_area,"[mm2] Aperture area of telescope";
      kwds_set,"TUNIT1","keV","Unit of first column: ENERGY";
      kwds_set,"TUNIT2","mm2","Unit of second column: EFF_AREA";

      wrmfitscols,outfile,"ENERGY",earr,"EFF_AREA",eff_area_avg, \
            "EFF_AREA_ERR", eff_area_err, clobber=1;
      write,format="%s has been written\n", outfile;
      write,"with columns ENERGY, EFF_AREA, and EFF_AREA_ERR";

   }

   write,format="Geometrical active area: %7.2f cm2\n", 0.01 * (ap_area * nphotons) / nphot_in;

   return eff_area_avg;
}

/* Function _mt_bg_run_eff_area */

func _mt_bg_run_eff_area( energy, offaxis, azimuth, system_file, no_scatter, \
                no_mdeform, dphot, earr, outfile, dir, aps= )
/* DOCUMENT _mt_bg_run_eff_area, energy, offaxis, azimuth, system_file, no_scatter, \
                no_mdeform, dphot, earr, outfile, dir, aps=

    Produces a batch file 'mbr_<random string>.ymac' and an FITS file with the
    energy array: 'tmp_<random string>.fits', both in directory 'dir', for the
    production of an effective area file in 'outfile'. The batch file is
    launched in the background.

    Keyword
       aps    arguments for 'mt_aperture_stop'
               Ex: aps="833.187,29.0"

    2011-09-04/NJW
*/
{
   if( is_void(dir) ) dir = "."; // defaults to current directory
   dir = app_slash(dir);
   fname = dir+"mbr_"+ranstr(15)+".ymac";
   tmpfil = dir+"tmp_"+ranstr(15)+".fits";
   kwds_init;
   wrmfitscols,tmpfil,"EARR",earr;
   fn = open(fname,"w");
   write,fn,format="#include \"/home/njw/yorick/common.id\"%s","\n";
   write,fn,format="#include \"/home/njw/yorick/mt_rayor.i\"%s","\n";
   write,fn,format="%s\n","mt_setup_system,\""+system_file+"\";";
   write,fn,format="mt_run,%.2f,%.2f,%.2f,no_scatter=%i,no_mdeform=%i,dphot=%.4f;\n", \
       energy, offaxis, azimuth, no_scatter, no_mdeform, dphot;
   if( structof(aps) == string ) write,fn,format="mt_aperture_stop,%s;\n", aps;
   write,fn,format="earr = rdfitscol(\"%s+1\",\"EARR\");\n", tmpfil;
   write,fn,format="mt_eff_area_photons,earr=earr,outfile=\"%s\";\n",outfile;
   write,fn,format="%s\n","quit";
   close, fn;
   system,"yorick -batch "+fname+" &";
}

/* Function mt_aperture_stop */

func mt_aperture_stop( z_position, open_radius, cen_dx=, cen_dy=, photfile=, undo=, chat= )
/* DOCUMENT mt_aperture_stop, z_position, open_radius, cen_dx=, cen_dy=, photfile=, undo=, chat=

    will update Phs.status if equal to zero to 201 for photons
    passing outside of a circle centered on the optical axis
    with radius 'open_radius' at a z-position 'z_position'
    above the focal plane.  For NuSTAR z_position is 833.187 mm,
    and open_radius = 29 mm.

    To take effect this subroutine should be called after 'mt_run' has finished.

    When called with arguments the external variables 'Z_position_as' and
    'Open_radius_as' are defined so for following calls no arguments are required.

    Keywords
        cen_dx      [mm] displacement in x-direction
        cen_dy      [mm] displacement in y-direction
        photfile    Operate on the named photonfile
        undo        Undo the operation on the named photonfile
        chat        Display certain extra information

    2011-10-24/NJW
*/
{
// ynclude = udts
   extern Phs, Z_position_as, Open_radius_as, Cen_dx_as, Cen_dy_as;
// yxclude =

   // When neither the argument nor the external variable is defined:
   if( is_void(z_position) && is_void(Z_position_as) ) error,"First argument must be defined";
   if( is_void(open_radius) && is_void(Open_radius_as) ) error,"Second argument must be defined";

   if( !is_void(z_position) ) {
      Z_position_as = double(z_position);
   }
   if( !is_void(open_radius) ) {
      Open_radius_as = double(open_radius);
   }

   // Shifting center position in x/y directions

   if( is_void(cen_dx) ) {
      if( is_void(Cen_dx_as) ) Cen_dx_as = 0.0;
   } else {
      Cen_dx_as = double(cen_dx);
   }
   if( is_void(cen_dy) ) {
      if( is_void(Cen_dy_as) ) Cen_dy_as = 0.0;
   } else {
      Cen_dy_as = double(cen_dy);
   }

   if( is_void(photfile) ) {

      if( is_void(Phs) ) {
         write,"External 'Phs' does not exist, no action.";
         return; // If no photons then only set externals
      }
      if( undo ) {
         w102 = where( Phs.status == 201 );
         if( numberof(w102) ) Phs(w102).status = 0;
      } else {
         w0 = where( Phs.status == 0 );
         if( numberof(w0) == 0 ) {
            if( chat ) write,"Found no status zero photons, no action.";
            return; // No good photons, simply skip this step
         }

         mt_propagate, Z_position_as;
         eq_nocopy, cE, Phs.E;
         r = sqrt((cE(1,w0)-Cen_dx_as)^2 + (cE(2,w0)-Cen_dy_as)^2);
         w = where( r > Open_radius_as );
         if( numberof(w) ) Phs(w0(w)).status = 201;
         mt_propagate, 0.0; // set photon positions (back) to focal plane
      }

   } else {  // Operate on a photon file

      if( !file_test(photfile) ) {
         write,"Photfile: "+photfile+" was not found, no action.";
         return;
      }

      local hdr, nrows;
      ptr = rdfitsbin( photfile+"+1", hdr, nrows );
      colstat = fits_colnum(hdr,"status");
      status = *ptr(colstat);
      detx = *ptr(fits_colnum(hdr,"detx"));
      dety = *ptr(fits_colnum(hdr,"dety"));
      rayx = *ptr(fits_colnum(hdr,"rayx"));
      rayy = *ptr(fits_colnum(hdr,"rayy"));
      rayz = *ptr(fits_colnum(hdr,"rayz"));

      if( undo ) {
         nw = numberof( (w = where( status == 201 ) ) );
         if( nw == 0 ) {
            if( chat ) write,"Found no status 201 photons, no action.";
            return; // No 201 photons, simply skip this step
         }
         status(w) = 0; // Remove aperture stop signature
         if( chat ) write,"Updating "+photfile+" with "+itoa(nw)+" reversals to  status zero ...";
         fits_bintable_poke, photfile+"+1", 1, colstat, status;
         if( chat ) write,"done";

      } else {

         nw = numberof( (w = where( status == 0 ) ) );
         if( nw == 0 ) {
            if( chat ) write,"Found no status zero photons, no action.";
            return; // No good photons, simply skip this step
         }

         kount = 0;
         for( i = 1; i <= nw; i++ ) {
            p = _propa([detx(w(i)),dety(w(i)),0.],[rayx(w(i)),rayy(w(i)),rayz(w(i))],Z_position_as);
            if( sqrt( (p(1) - Cen_dx_as)^2 + (p(2) - Cen_dy_as)^2) > Open_radius_as ) {status(w(i)) = 201; kount++;}
         }
         if( kount ) {
            if( chat ) write,"Updating "+photfile+" with "+itoa(kount)+" times status 201 ...";
            fits_bintable_poke, photfile+"+1", 1, colstat, status;
            if( chat ) write,"done";
         } else {
            if( chat ) write,"No update of "+photfile+" was required.";
         }
      }
   }
}

/* Function mt_translate */

func mt_translate( iphot, dx=, dy= )
/* DOCUMENT mt_translate, iphot, dx=, dy=

   Translates photon(s) in X and Y direction as indicated
   by keywords 'dx' and 'dy'.
   'iphot' is an array of indices to Phs. If not given
   'mt_translate' will operate on all photons.

   The struct element 'E' will be updated.

   2011-11-10/NJW
*/
{
   if( is_void(iphot) ) {
      cE = Phs.E;
      if( !is_void(dx) ) cE(1,) += dx;
      if( !is_void(dy) ) cE(2,) += dy;
      Phs.E = cE;
   } else {
      cE = Phs(iphot).E;
      if( !is_void(dx) ) cE(1,iphot) += dx;
      if( !is_void(dy) ) cE(2,iphot) += dy;
      Phs(iphot).E = cE;
   }
}

/* Function mt_info */

func mt_info( void )
/* DOCUMENT mt_info

   Takes no arguments. Displays basic information on the current telescope under
   investigation.
*/
{
// ynclude = zzxf
   extern Om_files, System_filename, Scat_files;
// yxclude =

   write," --- Information about currently loaded system ---";
   if( structof(System_filename) == string ) {
      write,"System file: "+System_filename;
   } else write,"System file is badly or not defined";
   if( structof(Om_files) == string ) {
      if( strlen(Om_files(1)) ) {
         write,"OM file 1  : "+Om_files(1);
      } else write,"OM file 1 is not defined"
      if( numberof(Om_files) > 1 ) {
         if( strlen(Om_files(1)) ) {
            write,"OM file 2  : "+Om_files(2);
         } else write,"OM file 2 is not defined"
      }
   } else write,"No OM files are defined";
}

/* Function mt_gravity_bend */

func mt_gravity_bend( gbend, mlength= )
/* DOCUMENT mt_gravity_bend, gbend, mlength=

    Make mirror deformation from circular bending of an X-ray telescope
    horizontally suspended in a gravitational field giving largest deviation
    in the middle.

    deform = -delta_r = -(gbend/mlength^2) * (z + mlength) * (z - mlength)

    for mirror parts above the axis and with reversed sign below the axis.
    Sign reversal is taken care of by the sine function in azimuth.

        For 1-alpha   z c [ 0., mlength]
        For 3-alpha   z c [-mlength, 0.]

    The 'gbend' parameter is the distance (same unit as 'mlength', often mm)
    between the cord and the circular arc where it is at a maximum,
    i.e. right between the 1alpha and 3alpha mirror sections.

    Use dimensions from already existing deformation cube Mirror_deform_arr.

    The mirror length can be set with keyword 'mlength' (defaults to 225. mm)

    2012-06-04/NJW

*/
{
   extern Mirror_deform_arr, Module_num, Roll_phot;

   //   Module_num is either 1 (U or 1-alpha section) or 2 (L or 3-alpha section)

   if( is_void(mlength) ) mlength = 225.; // mm - mirror length

   dms = dimsof( Mirror_deform_arr );

   naz = dms(2);
   nz =  dms(3);
   nlayers = dms(4);

   // Ensure that the largest deviation is found for low indices
   // for the 1-alpha section

   if( Module_num == 1 ) {

      z = span(0.,mlength,nz);
      yz = -gbend * (z + mlength) * (z - mlength) / mlength^2;

      ya = sin(span(0.,2*pi,naz) - Roll_phot);

      Mirror_deform_arr += ((ya(,-:1:nz)) * (yz(-:1:naz,)))(,,-:1:nlayers);

   } else {

   // Ensure that the largest deviation is found for high indices
   // for the 3-alpha section

      z = span(-mlength,0.,nz);
      yz = -gbend * (z + mlength) * (z - mlength) / mlength^2;

      ya = sin(span(0.,2*pi,naz) - Roll_phot);

      Mirror_deform_arr += ((ya(,-:1:nz)) * (yz(-:1:naz,)))(,,-:1:nlayers);

   }

}

/* Function mt_substr_volume */

func mt_substr_volume( void )
{
// ynclude = jdhf
   extern Mirror_thicknessarr, Num_modules, R1arr, R2arr, Om_files, Z1arr, Z2arr;
// yxclude =

   // Loading first mirror module;
   mt_load,omfile=Om_files(1);

   // vol = Mirror_thicknessarr * 0.5 * (R1arr + R2arr) * 2 * pi * (Z1arr - Z2arr);
   vol = pi * sum( Mirror_thicknessarr * (R1arr + R2arr) * (Z1arr - Z2arr) );

   if( Num_modules > 1 ) {
      mt_load,omfile=Om_files(2);
      vol += pi * sum( Mirror_thicknessarr * (R1arr + R2arr) * (Z1arr - Z2arr) );
   }

   write,format="Substrate volume = %10.2e mm3 = %10.2e liters\n", vol, vol*1.e-6;

   return vol;

}

/* Function mt_e1image */

func mt_e1image( z, size=, dim=, offset=, cen=, win=, lg=, rcoef=, \
                    bsel=, statsel=, pal=, title= )
/* DOCUMENT im = mt_e1image( z_value )
       or   mt_e1image, z_value

   Returns image of status==0 *) photon positions in extern 'Phs.E1'
   if 'z_value' is not given, else all photons will be propagated
   along the entrance direction 'Phs.D1' to
   plane at z = z_value.

   *) see keyword 'statsel' however

   Keyword 'size' defines the image edge size [mm] (default: extreme values).
   Keyword 'dim'  defines the dimensionality im(dim,dim) (default: 251).
   Keyword 'offset' or 'cen'  is only active when 'size' is set and it can be:
                    1) a 2-element array defining the (x,y) image center in mm
                    2) a scalar 0 (zero) to center on (0,0) [shorthand]
                    3) a scalar 1 (one) to center on the average position
   Keyword 'win'  number of display window (default is 0)
   Keyword 'lg'   display in log-color-scale
   Keyword 'rcoef' will cause a weighting with the reflection coefficients
   Keyword 'statsel'  will override the default 'status==0' selection. Expects
                      a string compatible with the 'where' function based on the
                      'status' Phs element.
   Keyword 'bsel'  will make an additional selection on this 'bounce' value
   Keyword 'pal'   will change the default palette ("yarg.gp") with the given one
   Keyword 'title'  Like for 'plot'

   If called as a subroutine the image will be displayed by function 'disp'

   SEE ALSO: mt_qimage

   2013-06-27/NJW
*/
{
   extern Phs;
   local cE1, cD1;

   if( is_void(win) ) win = 0;

   nphots = numberof(Phs);
   eq_nocopy, cE1, Phs.E1;
   eq_nocopy, cD1, Phs.D1;
   if( is_void(statsel) ) {
      w = where( Phs.status == 0 );
   } else {
      // Prepare for call of 'dwhere' that requires the use of 'var'
      sel = strstrrepl(strlowcase(statsel),"status","var");
      w = dwhere( Phs.status, sel );
   }
   if( numberof(w) == 0 ) {
      write,"Sorry, no photons from status selection";
      return [];
   }
   m = indgen(numberof(w));
   if( !is_void(bsel) ) m = where( Phs(w).bounce == bsel );
   if( numberof(m) == 0 ) {
      write,"Sorry, no photons from bounce selection";
      return [];
   }
   w = w(m);
   cE1 = cE1(,w);
   cD1 = cD1(,w);
   //  propagate to requested z-plane
   if( !is_void(z) ) {
      t = (z - cE1(3,)) / cD1(3,);
      cE1 += t(-:1:3,)*cD1;
   }

   if( is_void(cen) && !is_void(offset) ) cen = offset;

   if( is_void(cen) ) {
      cen = array(0.0,2);
   } else {
      if( numberof(cen) == 1 ) {
         if( cen == 0 ) {
            cen = array(0.0,2);
         } else {
            cen = [avg(cE1(1,)),avg(cE1(2,))];
         }
      }
   }

   if( dim ) {
      dims = [dim,dim];
   } else {
      dims = [251,251];
   }

   if( size ) {
      xrg = size*[-0.5,0.5] + cen(1);
      yrg = size*[-0.5,0.5] + cen(2);
   } else {
      xrg = yrg = array(double,2);
      xrg(2) = max(cE1(1,));
      xrg(1) = min(cE1(1,));
      yrg(2) = max(cE1(2,));
      yrg(1) = min(cE1(2,));
   }
   //+ write,format="Xrange: %10.4f %10.4f\n", xrg(1), xrg(2);
   //+ write,format="Yrange: %10.4f %10.4f\n", yrg(1), yrg(2);

   weight = rcoef ? Phs(w).rcoef : [];
   im = makeimageu( dims, cE1(1,), cE1(2,), xr=xrg, yr=yrg, weight=weight );

   if( am_subroutine() ) {
      xax = span(xrg(1),xrg(2),dims(1));
      yax = span(yrg(1),yrg(2),dims(2));
      window,win,style="boxed.gs";
      if( typeof(pal) == "string" ) {
         if( strpart(pal,-2:0) != ".gp" ) pal += ".gp";
         palette,pal;
      } else palette,"yarg.gp";
      if( lg ) {
         disp, log(1 + im), xtitle="mm", ytitle="mm", title=title, xax=xax, yax=yax;
      } else {
         disp, im, xtitle="mm", ytitle="mm", title=title, xax=xax, yax=yax;
      }
   } else {

      return im;
   }

}

/* Function mt_catalog2skydef */

/*************************************************************************
 *
 *   Use an X-ray source catalog to prepare a
 *   sky definition file (SDF) that can be used as input to 'mt_run'
 *
 *    Columns required in the catalog:
 *       RA_SRC (RA_OBJ ?)
 *       DEC_SRC (DEC_OBJ ?)
 *       SIMPLE                   XSPEC
 *       spectrcode               spectral model
 *       param1                   params
 *       nh
 *
 *       A keyword 'MODE' (model type) must be 'SIMPLE' or 'XSPEC'
 *             So far all sources in a catalog must have the same
 *             spectral model definition.
 *
 *   2013-06-12/NJW
 *
 ************************************************************************/

func mt_catalog2skydef( catalog_file, skydef_file, ra_scx, dec_scx, posang, \
             exposure=, e1=, e2=, nchan=, fluxdir=, \
            radius=, mission=, instrume=, telescop= )
/* DOCUMENT mt_catalog2skydef, catalog_file, skydef_file, ra_scx, dec_scx, posang

  or   nsources = mt_catalog2skydef( catalog_file, skydef_file, ra_scx, dec_scx, posang )

    Writes a sky definition file using sources and source spectral parameters from a
    catalog file, most conviently produced by 'mt_build_cat'.

    Arguments:
       catalog_file  - name of the input sky spectral definition file (FITS format)
       skydef_file - Name of the output sky definition file
       ra_scx      - [deg] RA of telescope boresight or DOL of previous sky definition file.
       dec_scx     - [deg] Dec of telescope boresight
       posang      - [deg] Position angle of telescope

    Keywords:
       exposure - [s] exposure time, default 1000.
       e1       - [keV] lower energy boundary for photon spectra, default 0.5.
       e2       - [keV] upper energy boundary for photon spectra, default 10.0.
       nchan    - number of energy channels in photon spectra (default is 200).
       radius   - [deg] only pixels within this radius from the pointing direction
                  will be used, default 0.5.
       fluxdir  - directory for storing the flux tables. If set the fluxes will be saved
                  to disk in the named directory, else the flux tables will only exist in
                  the actual Yorick session in memory.
       mission  - name of mission (becomes FITS keyword 'MISSION')
       instrume - name of instrument (becomes FITS keyword 'INSTRUME')
       telescop - name of 'telescope' (becomes FITS keyword 'TELESCOP')

    SEE ALSO: mt_skyspec2skydef, mt_build_cat

    2013-06-12/NJW, updated 2017-06-27/NJW
*/
{

   extern Version;

   // Communication with 'mk_photflux'
   extern Energy_start, Energy_stop, NumChans;

   rad2arcmin = (180/pi) * 60.;

   if( is_void(posang) ) posang = 0.0; // default position angle

   if( is_void(exposure) ) {
      exposure = 1000.; // Defaults to 1000 s
      write,"'exposure' has been set to 1000 s";
   }
   // Source position RA_OBJ | RA | RA_SRC
   //  assume that the catalog is in extension 1
   if( !file_test(catalog_file) ) error,"Did not find: "+catalog_file;
   dol = catalog_file+"+1";
   hdr = headfits(dol);
   srcnames = rdfitscol( dol, "NAME" );
   colnum = fits_colnum( hdr, "RA_OBJ", silent=1 );
   if( is_void(colnum) ) { // did not find RA_OBJ, try RA_SRC
      colnum = fits_colnum( hdr, "RA_SRC", silent=1 );
      if( is_void(colnum) ) { // did not find RA_SRC, try just RA
         colnum = fits_colnum( hdr, "RA", silent=1 );
         if( is_void(colnum) ) { // did not find RA either - give up
            error,"None of RA_OBJ, RA_SRC, and RA were found";
         } else {
            dec_col_name = "DEC";
         }
      } else {
         dec_col_name = "DEC_SRC";
      }
   } else {
      dec_col_name = "DEC_OBJ";
   }
   ra_obj = rdfitscol( dol, colnum );
   dec_obj = rdfitscol( dol, dec_col_name ); // should work
   nsrcs = numberof(ra_obj);

   // what spectral model to use?

   mode = strupcase(fxpar( hdr, "MODE" ));
   in_xspecmode = ( mode == "XSPEC" );

   if( in_xspecmode ) { // read columns XSPMODEL, IDXNORM, PARAMS

      xspmodel = strtrim(rdfitscol( dol, "XSPMODEL" ));
      idxnorm = rdfitscol( dol, "IDXNORM" );
      // + NB use fits_bintable_peek to read this : params = rdfitscol( dol, "PARAMS" );

   } else { // read columns SC, PARAM1, NH, and NORM

      sc = rdfitscol( dol, "SC" );
      param1 = rdfitscol( dol, "PARAM1" );
      nh = rdfitscol( dol, "NH" );
      norm = rdfitscol( dol, "NORM" );

   }

   // Generate a spectrum for each source

   dols = intens = x_sky = y_sky = enera = renorma = dphota = [];
   if( !is_void(fluxdir) ) {
      fluxdir = app_slash(fullpath(fluxdir));
   }

   for( isrc = 1; isrc <= nsrcs; isrc++ ) {

      fname = "flux"+itoa(isrc,4)+".fits";

      if( in_xspecmode ) {
         params = fits_bintable_peek( dol, isrc, "PARAMS" );
         if( is_void(fluxdir) ) {
            mk_photflux, e1=e1,e2=e2,nchan=nchan, \
                xspecmodel=xspmodel(isrc), xspecparms=params, \
                outfile=fname,nof=1,mem=isrc,silent=1;
         } else {
            mk_photflux, e1=e1,e2=e2,nchan=nchan, \
                xspecmodel=xspmodel(isrc), xspecparms=params, \
                outfile=fluxdir+fname,silent=1;
         }
      } else {
         if( is_void(fluxdir) ) {
            mk_photflux,sc=sc(isrc),nh=nh(isrc),norm=norm(isrc), \
                p1=param1(isrc),e1=e1,e2=e2,nchan=nchan, \
                outfile=fname,nof=1,mem=isrc,silent=1;
         } else {
            mk_photflux,sc=sc(isrc),nh=nh(isrc),norm=norm(isrc), \
                p1=param1(isrc),e1=e1,e2=e2,nchan=nchan, \
                outfile=fluxdir+fname,silent=1;
         }
      }
      e1 = Energy_start; // from mk_photflux
      e2 = Energy_stop; // from mk_photflux
      nchan = NumChans; // from mk_photflux

      grow, dols, fname+"+1";
      grow, intens, Sx_photflux;
      //+ grow, nharr, nh;
      //+ grow, param1, p1;
      jconv_coord, ra_scx, dec_scx, posang, ra_obj(isrc), dec_obj(isrc), \
                   x, y, to_sc=1;
      grow, x_sky, -x * 60.; // convert degrees to arcmin
      grow, y_sky,  y * 60.; // convert degrees to arcmin
      grow, enera, 1.0;
      grow, renorma, 1.0;
      grow, dphota, 1.0;

   }

   kwds_init;
   kwds_set,"EXTNAME","SKY_DEFINITION","Name of this extension";
   if( !is_void(mission) ) kwds_set,"MISSION", mission, "Name of mission";
   if( !is_void(instrume) ) kwds_set,"INSTRUME", instrume,"Name of instrument";
   if( !is_void(telescop) ) kwds_set,"TELESCOP", telescop,"Name of telescope";
   kwds_set,"DATE",ndate(3),"Date of creation of this file";
   kwds_set,"ORIGIN","MT_RAYOR-"+Version, "Software/version for creation of this file";
   kwds_set,"FUNCTION","mt_catalog2skydef", "Function used for creation of this file";
   kwds_set,"EXPOSURE", exposure, "[s] Exposure time";
   kwds_set,"RA_SCX", ra_scx, "[deg] R.A. of telescope bore sight";
   kwds_set,"DEC_SCX", dec_scx, "[deg] Decl. of telescope bore sightsatellite pointing axis";
   kwds_set,"E1", e1, "[keV] Lower energy for photon spectrum";
   kwds_set,"E2", e2, "[keV] Upper energy for photon spectrum";
   kwds_set,"NCHAN", nchan, "Number of energy channels in photon spectrum";
   kwds_set,"POSANG", posang, "[deg] Position angle of telescope bore sight";
   kwds_set,"INPUTFIL",catalog_file, "Catalog input file";
   if( in_xspecmode ) {
      kwds_set,"MODE","XSPEC","Used XSPEC generated photon flux";
   } else {
      kwds_set,"MODE","SIMPLE","Used simple mk_photflux generated photon flux";
   }

   wrmfitscols, skydef_file, "X_SKY", x_sky, "Y_SKY", y_sky, \
                "ENERGY", enera, "INTENS", intens, "DPHOT", dphota, "RENORM", renorma, \
                "DOL", dols, "RA_OBJ", ra_obj, "DEC_OBJ", dec_obj, \
                "NAME", srcnames, clobber=1;
   write,format="%i rows in %s have been written\n", numberof(x_sky), skydef_file;
   if( am_subroutine() ) write,format="%s has been written\n", skydef_file;

   return nsour;

}

/* Function mt_build_cat */

func mt_build_cat( catfile, mode=, import= )
/* DOCUMENT mt_build_cat, catfile, mode=, import=

     Make or update a source catalog to be used by 'mt_catalog2skydef'
     with one source per call of this subroutine.

     The catalog can either define source spectra by the simple
     spectral definition (homemade) or by XSPEC, but not both.

     If the catalog file exists the mode (SIMPLE or XSPEC) will be
     read from its header else the keyword 'mode' must be supplied as
     a string "SIMPLE" or "XSPEC" (case insensitive), and the
     catalog will be started.

     Keyword
       mode         'SIMPLE' or 'XSPEC' (see above)
     import         import the name, RA, and Dec from this instance
                    of the struct 's_Src', as it is returned from
                    e.g. 'nearest_source', 'nearby_source', or
                    'find_src_by_name' (all functions in 'jemx.i').

     SEE ALSO: mt_catalog2skydef

     2013-07-14/NJW
*/
{
   NAMESIZE = 25;
   MODELNAMESIZE = 50;

   if( !is_void(mode) ) {
      mode = strupcase(mode);
      if( mode != "SIMPLE" && mode != "XSPEC" ) error,"Illegal mode keyword";
   }

   // Accept both a filename and a DOL

   local fname, extno;
   get_exten_no, catfile, fname, extno;
   if( extno > 0 ) {
      dol = catfile;
      catfile = fname;
   } else dol = catfile+"+1";

   if( file_test(catfile) ) {
      mk_new = 0;
      write,"Catalog file exists - add a source or replace an existing source";
      hdr = headfits( dol );
      modef = strupcase(fxpar(hdr,"MODE"));
      if( !is_void(mode) ) {
         if( mode != modef ) error,"Conflicting mode indications";
      } else mode = modef;
      if( mode != "SIMPLE" && mode != "XSPEC" ) error,"Illegal mode keyword in cat file";
   } else { // define new catalog file
      if( is_void(mode) ) error,"Keyword 'mode' must be given";
      mk_new = 1;
      write,"New file: The catalog will be created";
   }

   // Get info for new (or updated) source
   if( is_void(import) ) { // Get user input for name, RA, and Dec
      srcname = strpadd(rdline(prompt="Enter source name : "),NAMESIZE);
      ra = atof(rdline(prompt="Enter RA : "));
      dec = atof(rdline(prompt="Enter Dec : "));
   } else {
      t = structof(s_Src());
      if( t != structof(import) ) error,"import has wrong type";
      write,"Importing source "+import.name;
      write,"    RA,Dec = "+ftoa(import.ra_obj,ndec=3)+", "+ftoa(import.dec_obj,ndec=3);
      srcname = strpadd(import.name, NAMESIZE);
      ra = import.ra_obj;
      dec = import.dec_obj;
   }
   if( mode == "SIMPLE" ) {
      sc = rdline(prompt="SIMPLE Enter spectral code : ");
      p1 = atof(rdline(prompt="SIMPLE Enter parameter : "));
      nh = atof(rdline(prompt="SIMPLE Enter NH (in 1e22 /cm2) : "));
      norm = atof(rdline(prompt="SIMPLE Enter norm : "));
   } else {
      model = strpadd(rdline(prompt="XSPEC Enter spectral model : "), MODELNAMESIZE);
      s = "asdf";
      i = 0;
      pars = [];
      while( strlen(s) > 0 ) {
         prom = "par"+itoa(++i);
         s = rdline(prompt="XSPEC "+prom+" : ");
         if(strlen(s) > 0 ) grow, pars, atof(s);
      }
      idxnorm = i - 1;
   }


   if( mk_new ) { // get info
      kwds_init;
      kwds_set, "EXTNAME", "MT_CATALOG", "Source catalog for MT_RAYOR";
      kwds_set, "MODE", mode, "Spectral mode definition";
      kwds_set, "DATE", ndate(3),"Date/time of file creation";
      kwds_set, "ORIGIN", "mt_build_cat.i","Software used";
      write,"Writing the catalog file: "+catfile;
      if( mode == "SIMPLE" ) {
         wrmfitscols, catfile, "NAME", [srcname], "RA_OBJ", [ra], "DEC_OBJ", [dec], "SC", [sc], \
                      "PARAM1", [p1], "NH", [nh], "NORM", [norm];
      } else { // must be mode XSPEC
         wrmfitscols, catfile, "NAME", [srcname], "RA_OBJ", [ra], "DEC_OBJ", [dec], "XSPMODEL", [model], \
                      "IDXNORM", [idxnorm], "PARAMS", double(1), var=6;
         fits_bintable_poke, dol, 1, "PARAMS", pars;
      }
   } else {
      fits_copy_keys, hdr, tokwds=1;
      srcnames = rdfitscol( dol ,"NAME");
      nsrcs = numberof(srcnames);
      write,"List of sources already in the catalog:";
      for( i = 1; i <= nsrcs; i++ ) {
         write,format="%4i  %s\n", i, strtrim(srcnames(i));
      }
      write,"\nFor a source update: enter the appropriate number";
      write,"For adding the source: enter 0 or just RTRN";
      n = nsrcs+1;
      while( n > nsrcs ) {
         n = atoi(rdline(prompt="Enter source number : "));
         if( n < 0 ) n = 0;
      }
      nsrcs += (n==0);

      // if 'n' is zero then add new source, else replace
      workfile = catfile+"_";
      ra_arr = rdfitscol( dol, "RA_OBJ");
      dec_arr = rdfitscol( dol, "DEC_OBJ");
      if( n == 0 ) { // add the source
         grow, srcnames, srcname;
         grow, ra_arr, ra;
         grow, dec_arr, dec;
      } else {
         srcnames(n) = srcname;
         ra_arr(n) = ra;
         dec_arr(n) = dec;
      }
    
      if( mode == "SIMPLE" ) {
         sc_arr = rdfitscol( dol, "SC" );
         p1_arr = rdfitscol( dol, "PARAM1" );
         nh_arr = rdfitscol( dol, "NH" );
         norm_arr = rdfitscol( dol, "NORM" );
         if( n == 0 ) { // add the source
            grow, sc_arr, sc;
            grow, p1_arr, p1;
            grow, nh_arr, nh;
            grow, norm_arr, norm;
         } else {       // replace
            sc_arr(n) = sc;
            p1_arr(n) = p1;
            nh_arr(n) = nh;
            norm_arr(n) = norm;
         }
         wrmfitscols, workfile, "NAME", srcnames, "RA_OBJ", ra_arr, "DEC_OBJ", dec_arr, "SC", sc_arr, \
                      "PARAM1", p1_arr, "NH", nh_arr, "NORM", norm_arr, clobber=1;
      } else {
         model_arr = rdfitscol( dol, "XSPMODEL" );
         idxnorm_arr = rdfitscol( dol, "IDXNORM" );
         if( n == 0 ) { // add the source
            grow, model_arr, model;
            grow, idxnorm_arr, idxnorm;
         } else {       // replace
            model_arr(n) = model;
            idxnorm_arr(n) = idxnorm;
         }
         wrmfitscols, workfile, "NAME", srcnames, "RA_OBJ", ra_arr, "DEC_OBJ", dec_arr, "XSPMODEL", model_arr, \
                      "IDXNORM", idxnorm_arr, "PARAMS", double(1), var=6;
         if( n == 0 ) n = nsrcs;
         for( i = 1; i <= nsrcs; i++ ) {
            spepars = i == n ? pars : fits_bintable_peek( dol, i, "PARAMS" );
            fits_bintable_poke, workfile+"+1", i, "PARAMS", spepars;
         }
      }
      back, catfile;
      mv, workfile, catfile;
   }
}

/* Function mt_jitter */

func mt_jitter( max_angle, chat= )
/* DOCUMENT mt_jitter, max_angle, chat=

    Distributes the photons in the focal plane as is the
    bore sight jitters like a harmonic oscillator. The
    maximum angle deviation must be given in radians.

    This function is to be called after 'mt_run' and assumes
    that the photon positions in 'Phs' refer to the focal plane.

    It operates only on the status == 0 photons.

    2013-07-05/NJW
*/
{
// ynclude = udts
   extern Phs, Focal_length;
// yxclude =

   local cE;

   if( is_void(Phs) ) {
      write,"External 'Phs' does not exist, no action.";
      return; // If no photons then only set externals
   }
   w0 = where( Phs.status == 0 );
   if( (nw0 = numberof(w0)) == 0 ) {
      if( chat ) write,"Found no status zero photons, no action.";
      return; // No good photons, simply skip this step
   }

   cE = Phs(w0).E; // position at focal plane
   // use expression for the distribution in a harmonic oscillator
   dradius = Focal_length * max_angle * sin( pi * (random(nw0) - 0.5) );
   azimuth = 2 * pi * random(nw0);
   cE(1,) += dradius * cos(azimuth);
   cE(2,) += dradius * sin(azimuth);
   // Put these values back
   for(i=1;i<=nw0;i++) Phs(w0(i)).E = cE(,i); // there seems to be no way of
                                              // of doing this in a vector-way
}

/* Function mt_tel_design */

func mt_tel_design( sysname )
/* DOCUMENT mt_tel_design, sysname

    Supports the construction of a Wolter 1 conical approximation
    telescope with constant mirror thickness and constant mirror
    lengths. A gap may exist between modules.

    Outputs to "system_"+sysname+".scm";

    The user is prompted for the various basic dimensions. The
    inputs are saved in a special file: 'mt_tel_design.spar'. The
    system description file 'system_<sysname>.scm' is created if it does
    not exist, else the keyword values are transferred to the new file.
*/
{
   write,"Supports the construction of a Wolter 1 conical approximation";
   write,"telescope with constant mirror thickness and constant mirror lengths";
   write,"";

   if( is_void(sysname) ) sysname = "tmp";

  /* -----------
       'outfile' is the final product that is compatible with mt_setup_system
       'parfile' is the intermediate product that will only hold the information
                 due to the assumptions in this function.
       It has been constructed this way to avoid conflicts with keywords appearing
       twice in the outfile.
   --------------- */
   outfile = "system_"+sysname+".scm"; // will be overwritten
   parfile = "mt_tel_design.spar"; // will be updated
   // Standard initial lines:
   text = "//";
   grow,text,"// Product of mt_tel_design from the MT_RAYOR package";
   grow,text,"// Initiated "+ndate(3);
   grow,text,"//";
   grow,text,"// The system setup file for a Wolter 1 conical approximation";
   grow,text,"// telescope with a gap between the mirror sets (optical modules)";
   grow,text,"//";
   scat_file = detector_descr_file = Dim_focp = Pix_focp = [];
   telescop = instrume = spoke_define_file = mirror_deform_file = [];
   if( file_test(outfile) ) { // save some keyword values that may have been edited
      scat_file = comgets( outfile, "scat_file" );
      detector_descr_file = comgets( outfile, "detector_descr_file" );
      Dim_focp = comget( outfile, "Dim_focp" );
      Pix_focp = comget( outfile, "Pix_focp" );
      telescop = comgets( outfile, "telescop" );
      instrume = comgets( outfile, "instrume" );
      spoke_define_file = comgets( outfile, "spoke_define_file" );
      mirror_deform_file = comgets( outfile, "mirror_deform_file" );
   }
   write_slist,outfile,text; // a new beginning
   if( !file_test(parfile) ) write_slist,parfile,text;

   // Get the focal length
   focal_length = comget( parfile, "focal_length" );
   if( is_void(focal_length) ) comput, parfile, "focal_length", 10000.0, comment="mm";
   focal_length = get_spar( parfile, "focal_length" );

   // Get the gap size
   gap_size = comget( parfile, "gap_size" );
   if( is_void(gap_size) ) comput, parfile, "gap_size", 5.0, comment="mm";
   gap_size = get_spar( parfile, "gap_size" );

   // Get the mirror length
   mirror_length = comget( parfile, "mirror_length" );
   if( is_void(mirror_length) ) comput, parfile, "mirror_length", 200.0, comment="mm";
   mirror_length = get_spar( parfile, "mirror_length" );

   // Get the om_parameter alias packing density
   om_parameter = comget( parfile, "om_parameter" );
   if( is_void(om_parameter) ) comput, parfile, "om_parameter", 1.0, comment="packing density";
   om_parameter = get_spar( parfile, "om_parameter" );

   // Get the mirror thickness - and save to special file
   mirror_thickness = comget( parfile, "mirror_thickness" );
   if( is_void(mirror_thickness) ) comput, parfile, "mirror_thickness", 0.4, comment="mm";
   mirror_thickness = get_spar( parfile, "mirror_thickness" );
   mirror_thickness_file = "mthick_"+sysname+".scm";
   text = ["// Mirror thickness file for "+sysname, \
           "// colname = radius; mm", "// colname = mirror_thickness; mm", \
           "   0.0   "+ftoa(mirror_thickness,ndec=3), " 10000.0 "+ftoa(mirror_thickness,ndec=3)];
   write_slist, mirror_thickness_file, text;

   // get outer radius
   r_outer = comget( parfile, "r_outer" );
   if( is_void(r_outer) ) comput, parfile, "r_outer", 200.0, comment="mm";
   r_outer = get_spar( parfile, "r_outer" );

   // get inner radius
   r_inner = comget( parfile, "r_inner" );
   if( is_void(r_inner) ) comput, parfile, "r_inner", 200.0, comment="mm";
   r_inner = get_spar( parfile, "r_inner" );


   //====   All ncessary information has been acquired, writing the outfile can start

   text = "// num_modules = 2";
   grow, text, "// om_type = conical";
   z_reference = focal_length + (mirror_length + gap_size)/2;
   grow, text, "// z_reference = "+ftoa(z_reference,ndec=3)+" ; mm";
   grow, text, "// Zfocus = "+ftoa(-2*z_reference,ndec=3)+" ; mm";
   grow, text, "// mirror_length = "+ftoa(mirror_length,ndec=3)+" ; mm";
   grow, text, "// mirror_thickness_file = "+mirror_thickness_file;
   grow, text, "// r_outer = "+ftoa(r_outer,ndec=3)+" ; mm";
   grow, text, "// r_inner = "+ftoa(r_inner,ndec=3)+" ; mm";
   grow, text, "// z1_setup = "+ftoa(mirror_length/2,ndec=3)+" ; mm";
   grow, text, "// z2_setup = "+ftoa(-mirror_length/2,ndec=3)+" ; mm";
   grow, text, "// om_function = mt_create_om_con4";
   grow, text, "// om_parameter = "+ftoa(om_parameter,ndec=3)+"; packing density";
   grow, text, "// om_file = om_"+sysname+"_1a.fits";
   if( is_void(mirror_deform_file) ) {
      grow, text, "// mirror_deform_file = none";
   } else {
      grow, text, "// mirror_deform_file = "+mirror_deform_file(1);
   }
   n = numberof(spoke_define_file);
   if( n != 2 && n != 4 ) {
      grow, text, "// spoke_define_file = none";
      grow, text, "// spoke_define_file = none";
   } else {
      grow, text, "// spoke_define_file = "+spoke_define_file(1);
      if( n == 4 ) grow, text, "// spoke_define_file = "+spoke_define_file(2);
   }
   grow, text, "//";
   grow, text, "// om_type = conical";
   z_reference = focal_length - (mirror_length + gap_size)/2;
   grow, text, "// z_reference = "+ftoa(z_reference,ndec=3)+" ; mm";
   grow, text, "// Zfocus = "+ftoa(-z_reference,ndec=3)+" ; mm";
   grow, text, "// mirror_length = "+ftoa(mirror_length,ndec=3)+" ; mm";
   grow, text, "// mirror_thickness_file = "+mirror_thickness_file;
   grow, text, "// z1_setup = "+ftoa(mirror_length/2,ndec=3)+" ; mm";
   grow, text, "// z2_setup = "+ftoa(-mirror_length/2,ndec=3)+" ; mm";
   grow, text, "// om_function = mt_create_om_con5";
   grow, text, "// om_parameter = -999.; no effect for slave module";
   grow, text, "// om_file = om_"+sysname+"_3a.fits";
   if( is_void(mirror_deform_file) ) {
      grow, text, "// mirror_deform_file = none";
   } else {
      if( numberof(mirror_deform_file) > 1 ) grow, text, "// mirror_deform_file = "+mirror_deform_file(2);
   }
   n = numberof(spoke_define_file);
   if( n != 2 && n != 4 ) {
      grow, text, "// spoke_define_file = none";
      grow, text, "// spoke_define_file = none";
   } else {
      if( n == 2 ) {
         grow, text, "// spoke_define_file = "+spoke_define_file(2);
      } else { // 'n' must be 4
         grow, text, "// spoke_define_file = "+spoke_define_file(3);
         grow, text, "// spoke_define_file = "+spoke_define_file(4);
      }
   }
   grow, text, "//";
   grow, text, "// focal_length = "+ftoa(focal_length,ndec=3)+" ; mm";
   grow, text, "//";
   if( is_void(scat_file) ) {
      grow, text, "// scat_file = none";
   } else {
      for(i = 1; i <= numberof(scat_file); i++ ) {
         grow, text, "// scat_file = "+scat_file(i);
      }
   }
   if( is_void(detector_descr_file) ) grow, text, "// detector_descr_file = none";
   else grow, text, "// detector_descr_file = "+detector_descr_file;

   if( is_void(Dim_focp) ) grow, text, "// Dim_focp = 441; For use by 'mt_save' alone";
   else grow, text, "// Dim_focp = "+itoa(Dim_focp)+" ; For use by 'mt_save' alone";
   if( is_void(Pix_focp) ) grow, text, "// Pix_focp = 0.1; mm  For use by 'mt_save' alone";
   else grow, text, "// Pix_focp = "+ftoa(Pix_focp,ndec=3)+" ; mm For use by 'mt_save' alone";
   if( is_void(telescop) ) grow, text, "// telescop = "+sysname;
   else grow, text, "// telescop = "+telescop;
   if( is_void(instrume) ) grow, text, "// instrume = "+sysname;
   else grow, text, "// instrume = "+instrume;

   write_slist, outfile, text, app=1;

   //==========    Section for preparing new setup  ================================

   // Remove previous OM files

   remove,"om_"+sysname+"_1a.fits"
   remove,"om_"+sysname+"_3a.fits"
   text = ["mt_setup_system, \""+outfile+"\";", "mt_create_om_con4, filename=\"om_"+sysname+"_1a.fits\";", \
           "mt_create_om_con5, filename=\"om_"+sysname+"_3a.fits\";"];
   write_slist,"mk_om.i", text;

   include,"mk_om.i";

}

/* Function mt_simulation */

func mt_simulation( sim_name, no_scatter=, no_mdeform= )
/* DOCUMENT mt_simulation, sim_name, no_scatter=, no_mdeform=

    Runs a complete NuSTAR observation simulation with input from the file
    <sim_name>.par that must exist in the current directory before running
    this function. The parameter file can be created by copying the
    'mt_simulation.par' file (found in /home/njw/pfiles) and updating the parameters.

    Keywords: no_scatter and no_mdeform as in 'mt_run'

    2015-01-06/NJW
*/
{

   parfilename = sim_name+".par";
   if( !file_test(parfilename) ) {
      write,"The required parameter file: "+parfilename+" does not exist.";
      write,"You may create it by copying 'mt_simulation.par' to '"+parfilename;
      write,"and update it suitably.";
      return;
   }
   if( is_void(no_scatter) ) no_scatter = 0;
   if( is_void(no_mdeform) ) no_mdeform = 0;

   // Setup telescope system

   system_file = comgets( parfilename, "system_file" );
   mt_setup_system, system_file;

   /*
    *  Defining pointing and exposure time (in seconds)
    */
   ra_point = comget( parfilename, "ra_point" );
   dec_point = comget( parfilename, "dec_point" );
   ra_det_center = comget( parfilename, "ra_det_center" );
   dec_det_center = comget( parfilename, "dec_det_center" );
   exposure = comget( parfilename, "exposure");
   posang = comget( parfilename, "posang");
   offset = get_offset( ra_point, dec_point, ra_det_center, dec_det_center, posang );

   RUN_IN_BKG = comget( parfilename, "RUN_IN_BKG", lng=1);
   MULTIPLICITY = comget( parfilename, "MULTIPLICITY", lng=1); // number of times to run
   exposure /= MULTIPLICITY;
   /*
    *  Make a log file with essential information
    *     Remember that 'skyspecfile' (Sky Spectral File) is a description of the
    *     object to be simulated
    *     'skydeffile' (Sky Definition File) sets the pointing and where the pixels are
    *     positioned in the FOV. It defines all the individual point sources (including spectra)
    *     that make up the extended source.
    */
   lg = open(sim_name+".log","w");
   lg = gen_log_init(sim_name+".log",logflag=5,erase=1,reset=1);
   gen_log_entry,1,"// Run start time "+ndate(3),id=lg;

   skyspecfile = "skyspec_"+sim_name+".fits";
   skydeffile = "skydef_"+sim_name+".fits";
   gen_log_entry,1,"// skyspecfile = "+skyspecfile,id=lg;
   gen_log_entry,1,"// ra_point = "+ftoa(ra_point,ndec=3)+"; deg", id=lg;
   gen_log_entry,1,"// dec_point = "+ftoa(dec_point,ndec=3)+"; deg", id=lg;
   gen_log_entry,1,"// ra_det_center = "+ftoa(ra_det_center,ndec=3)+"; deg", id=lg;
   gen_log_entry,1,"// dec_det_center = "+ftoa(dec_det_center,ndec=3)+"; deg", id=lg;
   gen_log_entry,1,"// posang = "+ftoa(posang,ndec=3)+"; deg", id=lg;
   gen_log_entry,1,"// offset1 = "+ftoa(offset(1),ndec=3)+"; mm", id=lg;
   gen_log_entry,1,"// offset2 = "+ftoa(offset(2),ndec=3)+"; mm", id=lg;
   gen_log_entry,1,"// exposure = "+ftoa(exposure,ndec=2)+"; s", id=lg;

   remove,skyspecfile;
   remove,skydeffile;

   fraclim = comget( parfilename, "fraclim");
   xspecmodel = comgets( parfilename, "xspecmodel");
   num_model_params = comget( parfilename,"num_model_params", lng=1);
   // Initiate the parameter list
   xspecparlist = _lst(num_model_params); // index of normalization parameter
   // now you know how many to read, but you must also know which of them can be
   // strings
   param = comgets( parfilename, "param" );

   maybe_string = comgets( parfilename, "maybe_string");
   if( maybe_string == "none" || maybe_string == "0" ) {
      maybe_string = 0;
   } else maybe_string = str2arr( maybe_string );

  /*
   *   NB NB NB A new way to organize the 'tolmap' functionality is required
   *       e.g. by keywords in the *.par file
   *   because a hardcoded value does not work well for both NH and kT, for example.
   */

   tolmap = [];
   for( i = 1; i <= num_model_params; i++ ) {
      gen_log_entry,1,swrite(format="// param%i = %s", i, param(i)),id=lg;
      if( allof( i - maybe_string ) ) { // this parameter must be a number
         xspecparlist = _cat( xspecparlist, atof(param(i)) );
      } else { // this parameter may be a number
         if( is_number( param(i) ) ) {
            xspecparlist = _cat( xspecparlist, atof(param(i)) );
         } else {
            xspecparlist = _cat( xspecparlist, param(i) );
            grow, tolmap, 0.05;
         }
      }
   }

   //                      NH     kT   e-densi  metal    z    interpolate norm
   // xspecparlist = _lst( 7, 0.041, tmap, 0.1,    0.23, 0.058, 1,           1. );


   if( !file_test(skyspecfile) ) {
      dol_skyima = comgets( parfilename, "dol_skyima" );
      dol_arf = comgets( parfilename, "dol_arf" );
      emin = comget( parfilename, "emin" ); // keV - applies to dol_skyima, input image energy limit
      emax = comget( parfilename, "emax" ); // keV - applies to dol_skyima, input image energy limit
      gen_log_entry,1,"// emin = "+ftoa(emin,ndec=2)+"; keV",id=lg;
      gen_log_entry,1,"// emax = "+ftoa(emax,ndec=2)+"; keV",id=lg;
      mt_skyima2skyspec, dol_skyima, dol_arf, emin, emax, skyspecfile, \
                  fraclim=fraclim,xspecmodel=xspecmodel,xspecparlist=xspecparlist, \
                  tolmap=tolmap;
   }

   e1 = comget( parfilename, "e1"); // keV - applies to simulated telescope
   e2 = comget( parfilename, "e2"); // keV - applies to simulated telescope
   radius_lim = comget( parfilename, "radius_lim"); // deg - max angle in simulated telescope FOV
   blur = comget( parfilename, "blur" ); // rad - blur angle, should correspond to input image pixel size
   nchan = comget( parfilename, "nchan", lng=1);
   gen_log_entry,1,"// e1 = "+ftoa(e1,ndec=2)+"; keV",id=lg;
   gen_log_entry,1,"// e2 = "+ftoa(e2,ndec=2)+"; keV",id=lg;
   gen_log_entry,1,"// radius_lim = "+ftoa(radius_lim,ndec=4)+"; deg",id=lg;
   mt_skyspec2skydef, skyspecfile, skydeffile, ra_point, dec_point, \
          posang, fraclim=fraclim, exposure=exposure, \
          e1=e1, e2=e2, nchan=nchan, radius=radius_lim, \
          telescop="NuSTAR",instrume="MT_RAYOR-4.6.6 sim";

   // Do you want to include background?

   add_bkg = strpart(strlowcase(comgets( parfilename, "add_bkg" )),1:1);

   for( MUL = 1; MUL <= MULTIPLICITY; MUL++ ) {
      mt_clear;
      if( RUN_IN_BKG ) {
         ncases = 15;
         mt_bg_run,system_file,skydeffile,exposure=exposure, \
            ncases=ncases,dir="/r12/scratch/njw/mt",nproc=ncases+10,blur=blur;
      } else {
         mt_run,skydeffile,no_scatter=no_scatter,no_mdeform=no_mdeform,blur=blur,chat=2;
      }
      //+ NuSTAR: mt_aperture_stop, 833.187, 29.0;
      //+ NuSTAR: mt_jitter, 4.e-5; // 8"
      mt_detector, offset=offset;

      num_events = numberof(Evlist);
      gen_log_entry,1,"// num_events = "+itoa(num_events)+"; excluding background", id=lg;

      // then add the background

      if( add_bkg == "y" ) {
         mt_det_add_instr_bkg;
         num_events_with_bg = numberof(Evlist);
         gen_log_entry,1,"// num_bg_events = "+itoa(num_events_with_bg - num_events)+"; background events", id=lg;
      }

      namstr = MULTIPLICITY > 1 ? "_"+itoa(MUL,2)+itoa(MULTIPLICITY,2) : "";
      mt_save,mode="e",outfile=sim_name+namstr;
      mt_evt_ds9,"e"+sim_name+namstr+".evt","e"+sim_name+namstr+"p.evt",skydef=skydeffile;
      mt_trace_events, "e"+sim_name+namstr+"p.evt", clobber=1;
      //+ remove,"e"+sim_name+namstr+".evt";

   }

   gen_log_entry,1,"// Run end time "+ndate(3),id=lg;

}

/* Function mt_evt_ds9 */

func mt_evt_ds9( mt_evt_file, new_evt_file, imdim=, skydef=, skyspec= )
/* DOCUMENT   mt_evt_ds9, mt_evt_file, new_evt_file, imdim=, skydef=, skyspec=

     Copies the eventfile with the addition of two columns:
     'X' and 'Y' with corresponding coordinate defining
     keywords so that the file can be read directly into 'ds9' and 'xselect'.
     NB: for the latter an MDB file similar to the one used for Astrosat/XST
         is appropriate.

     This function assumes that the 'mt_setup_system' has been run with the
     proper telescope definition so that the external variables Focal_length,
     Num_pixels1, Num_pixels2, Pixel_size1, and Pixel_size2 are defined.

     Keywords:
        'skydef' and 'skyspec'
              causes those filenames to be written in the output header as
              required by 'mt_trace_events'.
        'imdim'  final image dimension (same in both directions)

   2014-10-08/NJW

*/
{
   if( is_void(Focal_length) ) error, "Focal_length has not been defined ...";
   if( is_void(Num_pixels1) ) error, "Num_pixels1 has not been defined ...";
   if( is_void(Num_pixels2) ) error, "Num_pixels2 has not been defined ...";
   if( is_void(Pixel_size1) ) error, "Pixel_size1 has not been defined ...";
   if( is_void(Pixel_size2) ) error, "Pixel_size2 has not been defined ...";

   if( is_void(imdim) ) imdim = 1000;

   local hdr, nrows;
   ptr_evts = rdfitsbin( mt_evt_file+"+1", hdr, nrows );
   n_evts = numberof(*ptr_evts(1));
   exposure = livetime = ontime = fxpar( hdr, "exposure" );
   datti = ndate(3);
   dateobs = strpart(datti,1:10);
   timeobs = strpart(datti,12:);
   tstart = double(timeconv( datti, in="utc", out="nst" ));
   tstop  = tstart + exposure;
   dattiend = timeconv( tstop, in="nst", out="utc" );
   dateend = strpart(dattiend,1:10);
   timeend = strpart(dattiend,12:);
   timecol = span(tstart,tstop,n_evts);

   /* ------------
       The WCS in 'hdr' is for the detector (RAWX, RAWY). When
       writing the output file they are omitted because 'get_wcs'
       gets mixed up if they are present together with TCRVLi, TCRPXi,
       etc.

       Because of the efforts to re-establish the conversion from RAWX(Y)
       to sky coordinates (as in fit_crpix_pa) it might be useful to
       bring the coordinates but under different (though related) keywords.

       CRPIXi -> RWRPIXi, CRVALi -> RWRVALi, CD1_1 -> RWCD1_1 etc.

   -----
   */

   coords = get_wcs( hdr );

   nrx = fits_colnum( hdr, "rawx", silent=1 );
   nry = fits_colnum( hdr, "rawy", silent=1 );
   // the following RA,DEC are relative to CRVAL1, CRVAL2
   ra  = ( (*ptr_evts(nrx)) - coords.crpix1 ) * coords.cd1_1 \
       + ( (*ptr_evts(nry)) - coords.crpix2 ) * coords.cd1_2;
   dec = ( (*ptr_evts(nrx)) - coords.crpix1 ) * coords.cd2_1 \
       + ( (*ptr_evts(nry)) - coords.crpix2 ) * coords.cd2_2;

   // produce a test plot

   plot,ra,dec,ps=1,xr=[max(ra),min(ra)];

   //+ cdelt1 = -6.828e-4; // deg/pix
   //+ cdelt2 =  6.828e-4; // deg/pix
   //+ crpix1 = crpix2 = 250; // > sqrt(2)*(324/2)
   cdelt1 = -(Pixel_size1 / Focal_length) * 180. / pi; // deg/pix
   cdelt2 =  (Pixel_size2 / Focal_length) * 180. / pi; // deg/pix
   //+ crpix1 = (Num_pixels1 + 100) / 2;
   //+ crpix2 = (Num_pixels2 + 100) / 2;
   crpix1 = imdim / 2;
   crpix2 = imdim / 2;
   x = int( floor( crpix1 + ra/cdelt1 + 0.5 ));
   y = int( floor( crpix2 + dec/cdelt2 + 0.5 ));

   kwds_init;

   fits_copy_keys, hdr, skip="CDELT1,CDELT2,CRVAL1,CRVAL2,CRPIX1,CRPIX2," \
      +"CTYPE1,CTYPE2,CD1_1,CD1_2,CD2_1,CD2_2", tokwds=1;

   kwds_set,"EXTNAME", "EVENTS","Name of this extension";
   kwds_set,"TUNIT3", "mm", "Unit for column 3";
   kwds_set,"TUNIT4", "mm", "Unit for column 4";
   kwds_set,"TUNIT6", "keV", "Unit for column 6";
   kwds_set,"TUNIT9", "pixel", "Unit for column 9";
   kwds_set,"TUNIT10", "pixel", "Unit for column 10";
   kwds_set,"TUNIT12", "s", "Unit for column 12 (time)";
   kwds_set,"TLMIN9", 1, "Lower limit for X";
   kwds_set,"TLMAX9", imdim, "Upper limit for X";
   kwds_set,"TCRPX9", crpix1, "Equivalent of CRPIX1";
   kwds_set,"TCRVL9", coords.crval1, "Equivalent of CRVAL1";
   kwds_set,"TCDLT9", cdelt1, "Equivalent of CDELT1";
   kwds_set,"TCTYP9", "RA---TAN", "Equivalent of CTYPE1";
   kwds_set,"TCUNI9", "deg", "Equivalent of CUNIT1";
   kwds_set,"TLMIN10", 1, "Lower limit for Y";
   kwds_set,"TLMAX10", imdim, "Upper limit for Y";
   kwds_set,"TLMIN11", 1, "Lower limit for PI";
   kwds_set,"TLMAX11", 1024, "Upper limit for PI";
   kwds_set,"TCRPX10", crpix2, "Equivalent of CRPIX2";
   kwds_set,"TCRVL10", coords.crval2, "Equivalent of CRVAL2";
   kwds_set,"TCDLT10", cdelt2, "Equivalent of CDELT2";
   kwds_set,"TCTYP10", "DEC--TAN", "Equivalent of CTYPE2";
   kwds_set,"TCUNI10", "deg", "Equivalent of CUNIT2";
   kwds_set,"RWTYPE1", "RA---TAN","Projection";
   kwds_set,"RWTYPE2", "DEC--TAN","Projection";
   kwds_set,"RWRVAL1", coords.crval1, "[deg] Reference value RA";
   kwds_set,"RWRVAL2", coords.crval2, "[deg] Reference value DEC";
   kwds_set,"RWRPIX1", coords.crpix1, "Reference pixel axis#1";
   kwds_set,"RWRPIX2", coords.crpix2, "Reference pixel axis#2";
   kwds_set,"RWDELT1", coords.cdelt1, "[deg] Pixel size axis#1";
   kwds_set,"RWDELT2", coords.cdelt2, "[deg] Pixel size axis#2";
   kwds_set,"RWROTA2", coords.crota2, "[deg] Rotation angle";
   kwds_set,"RWCD1_1", coords.cd1_1,  "[deg] Part of transformation matrix";
   kwds_set,"RWCD2_1", coords.cd2_1,  "[deg] Part of transformation matrix";
   kwds_set,"RWCD1_2", coords.cd1_2,  "[deg] Part of transformation matrix";
   kwds_set,"RWCD2_2", coords.cd2_2,  "[deg] Part of transformation matrix";
   kwds_set,"TELESCOP", "ASTROSAT", "Name of the mission";
   kwds_set,"INSTRUME", "SXT", "Name of the instrument";
   kwds_set,"OBJECT", "Tycho SNR", "Name of the celestial object";
   kwds_set,"ORB_NUM", 1234, "Orbit number";
   kwds_set,"DATAMODE", "PC", "Data mode";
   kwds_set,"CREATOR", "MT_RAYOR", "Simulation software";
   kwds_set,"EXPOSURE", exposure, "Exposure time";
   kwds_set,"LIVETIME", exposure, "Livetime";
   kwds_set,"ONTIME", exposure, "Ontime";
   kwds_set,"TIMEUNIT", "s", "Time unit";
   kwds_set,"TSTART", tstart, "Time in NST for data start";
   kwds_set,"TSTOP", tstop, "Time in NST for data stop";
   kwds_set,"MJDREFI", 55197, "MJD reference day";
   kwds_set,"MJDREFF", 0.0, "MJD reference, fraction of day";
   kwds_set,"DATE-OBS", dateobs, "Date of observation beginning";
   kwds_set,"TIME-OBS", timeobs, "Time of observation beginning";
   kwds_set,"DATE-END", dateend, "Date of observation end";
   kwds_set,"TIME-END", timeend, "Time of observation end";
   kwds_set,"TIMESYS", "UTC", "Time system";

   if( typeof(skydef) == "string" ) kwds_set, "SKYDEF", skydef, "Sky definition file";
   if( typeof(skyspec) == "string" ) kwds_set, "SKYSPEC", skydef, "Sky spectral file";

   kwds_set,"COMMENT","Updated copy of: "+mt_evt_file;
   kwds_set,"COMMENT","by yorick function 'mt_evt_ds9' "+ndate(3);

   // Dump events to new file but prepare to append the GTI extension

   fh = wrmfitscols( new_evt_file, "RAWX", *ptr_evts(1), "RAWY", *ptr_evts(2), \
                "DETX", *ptr_evts(3), "DETY", *ptr_evts(4), "PHA", *ptr_evts(5), \
                "ENERGY", *ptr_evts(6), "BOUNCE", *ptr_evts(7), "FLAG", *ptr_evts(8), \
                "X", x, "Y", y, "PI", *ptr_evts(5), "TIME", timecol, clobber=1, cont=1 );

   kwds_set,"EXTNAME", "STDGTI","Standard GTI extension";
   kwds_set,"HDUCLASS", "OGIP";
   kwds_set,"HDUCLAS1", "GTI";
   kwds_set,"HDUCLAS2", "STANDARD";
   wrmfitscols, fh, "START", [tstart], "STOP", [tstop];

}

/* Function mt_trace_events */

func mt_trace_events( eventfilename, outfile=, clobber= )
/* DOCUMENT mt_trace_events, eventfilename, outfile=, clobber=

    Make a binary FITS table with two columns: RA_ORIG and DEC_ORIG for all events
    in a MT_RAYOR event file that has been produced with 'mt_simulation'.

    The resulting file will have an 'o' put just before the postfix
    example:   myevents.evt -> myeventso.evt

    Arguments:
     eventfilename     file name of the event file as produced by MT_RAYOR
                       NB: The SKYDEF filename and the SKYSPEC filename must
                           appear in the header of the event file.

    Keywords:
     outfile           If set then the interim file will be saved to this name
     clobber           (bool) as usual

     2015-02-12/NJW
*/
{
   local hdr, nrows;

   if( !file_test(eventfilename) ) error,"File not found: "+eventfilename;

   dot = strpos( eventfilename, ".", rev=1 ); // locate last dot
   part1 = strpart( eventfilename, 1:dot-1 );
   part2 = strpart( eventfilename, dot: );
   resultfile = part1+"o"+part2;
   if( file_test(resultfile) ) {
      if( clobber ) {
         remove, resultfile;
      } else {
         write,resultfile+" exists already, no action ...";
         return;
      }
   }

   evdol = eventfilename+"+1";

   ptr = rdfitsbin( evdol, hdr, nrows );

   rawx = *ptr(fits_colnum(hdr,"rawx"));
   rawy = *ptr(fits_colnum(hdr,"rawy"));
   flag = *ptr(fits_colnum(hdr,"flag"));

   skydef = fxpar(hdr,"skydef");
   if( is_void(skydef) ) {
      write,"Keyword SKYDEF is missing, can't do this one ...";
      return;
   }

   ra_src = rdfitscol( skydef+"+1","RA_OBJ" );
   dec_src = rdfitscol( skydef+"+1","DEC_OBJ" );

   ra_orig = dec_orig = array( float, nrows );

   for( i = 1; i <= nrows; i++ ) {
      if( flag(i) ) { // skip flag==0 that are bgd events
         ra_orig(i)  = ra_src(flag(i));
         dec_orig(i) = dec_src(flag(i));
      }
   }

   kwds_init;
   kwds_set,"DATE", ndate(3), "Date/time of production";
   kwds_set,"ORIGIN", "mt_trace_events of MT_RAYOR","Software used";
   kwds_set,"EVNTFILE", eventfilename, "Input eventfile";
   kwds_set,"SKYDEF", skydef, "Sky definition file";

   if( is_void(outfile) ) {
      outfile = ranstr(20)+".fits";
      keep = 0;
   } else keep = 1;
   wrmfitscols, outfile, "RA_ORIG", ra_orig, "DEC_ORIG", dec_orig, clobber=clobber;

   // use ftool 'faddcol'

   system,"cp "+eventfilename+" "+resultfile;
   system,"faddcol "+resultfile+" "+outfile+" \"ra_orig dec_orig\""

   if( !keep ) remove, outfile;

}

/* Function mt_blur */

func mt_blur( direc, angle, nphot )
/* DOCUMENT res = mt_blur( direc, angle, nphot )

    Returns an array (3 x nphot) with blurred values around 'direc'

    Arguments:
       direc      vector (x,y,z) where it is expected that z ~ -1
       angle      blur angle in radians
       nphot      number of photons

     2016-12-14/NJW
*/
{
   if( nphot <= 0 ) return []; // nothing to do

   // use expression for the distribution in a harmonic oscillator
   dradius = angle * sin( pi * (random(nphot) - 0.5) );
   azimuth = 2 * pi * random(nphot);
   res = direc(,-:1:nphot);
   res(1,) += dradius * cos(azimuth);
   res(2,) += dradius * sin(azimuth);
   res(3,) = -sqrt(1.0 - res(1,)^2 - res(2,)^2);
   return res;
}
